/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties;

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

/**
 * Abstract base clase for a {@link de.fu_berlin.ties.Processor} that collects
 * all the input arguments and processes the collected arguments when shutting
 * down. The {@link #close(int)} method delegates to the abstract
 * {@link #process(List, ContextMap)} method to process the collected arguments.
 *
 * <p>Instances of this class are not thread-safe and must be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2004/09/06 17:21:46 $, $Author: siefkes $
 */
public abstract class CollectingProcessor
    extends ConfigurableProcessor
    implements Closeable {

    /**
     * Used to collect the input arguments.
     */
    private final List<String> collectedArgs = new LinkedList<String>();

    /**
     * Creates a new instance.
     *
     * @param conf used to configure this instance; if <code>null</code>,
     * the {@linkplain TiesConfiguration#CONF standard configuration} is used
     */
    public CollectingProcessor(final TiesConfiguration conf) {
        super(conf);
    }

    /**
     * {@inheritDoc}
     */
    public final void close(final int errorCount)
            throws IOException, ProcessingException {
        // delegate to abstract method
        process(collectedArgs, new ContextMap());
    }

    /**
     * {@inheritDoc}
     */
    public final void process(final String input) {
        // collect in list
        collectedArgs.add(input);
    }

    /**
     * Processes the collected input arguments.
     *
     * @param collected a list of Strings containing the collected input
     * arguments
     * @param context a map of objects that are made available for processing;
     * will be empty when called from the {@link #close(int)} method in this
     * class
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public abstract void process(final List<String> collected,
            final ContextMap context)
            throws IOException, ProcessingException;

}
