/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * Abstract base class for a {@link de.fu_berlin.ties.Processor} that operates
 * on the contents of directories.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/02/15 17:46:56 $, $Author: siefkes $
 */
public abstract class DirectoryProcessor extends ConfigurableProcessor {

    /**
     * The filter used to decide which files to accept.
     */
    private final FileFilter fileFilter;

    /**
     * Creates a new instance.
     *
     * @param filter the filter used to decide which files to accept
     * @param conf used to configure this instance; if <code>null</code>,
     * the {@linkplain TiesConfiguration#CONF standard configuration} is used
     */
    public DirectoryProcessor(final FileFilter filter,
            final TiesConfiguration conf) {
        super(conf);
        fileFilter = filter;
    }

    /**
     * Processes an array of files.
     *
     * @param files the array of files to process
     * @param context a map of objects that are made available for processing;
     * will be empty when called from the implemented <code>process</code>
     * methods in this class
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public abstract void process(final File[] files, final ContextMap context)
            throws IOException, ProcessingException;


    /**
     * Processes the contents of a directory.
     *
     * @param directory the the directory to process
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public final void process(final File directory)
            throws IOException, ProcessingException {
        if (!directory.isDirectory() || !directory.canRead()) {
            throw new IllegalArgumentException(
                directory + " is not a readable directory");
        }

        // filter files + delegate to abstract method
        final File[] files = directory.listFiles(fileFilter);
        final ContextMap context = new ContextMap();
        process(files, context);
    }

    /**
     * Processes an input argument that must specify a directory.
     *
     * @param inputName the name of the directory to process
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public final void process(final String inputName)
            throws IOException, ProcessingException {
        final File directory = new File(inputName);
        process(directory);
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("file filter", fileFilter)
            .toString();
    }

}
