/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;

import org.dom4j.Document;
import org.dom4j.DocumentException;

import de.fu_berlin.ties.xml.dom.DOMUtils;

/**
 * Abstract base class for a {@link de.fu_berlin.ties.Processor} that read
 * {@linkplain org.dom4j.Document XML documents}.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.4 $, $Date: 2004/03/11 16:18:22 $, $Author: siefkes $
 */
public abstract class DocumentReader extends TextProcessor {

    /**
     * Creates a new instance.
     *
     * @param outExt the extension to use for output files
     * @param conf used to configure this instance; if <code>null</code>,
     * the {@linkplain TiesConfiguration#CONF standard configuration} is used
     */
    public DocumentReader(final String outExt, final TiesConfiguration conf) {
        super(outExt, conf);
    }

    /**
     * Processes the contents of a reader, writing a modified version to a
     * writer. This implementation delegates to
     * {@link #process(Document, Writer, ContextMap)}. The reader must contain a
     * well-formed XML document.
     *
     * @param reader reader containing the text to process; not closed
     * by this method
     * @param writer the writer to write the processed text to; flushed
     * but not closed by this method
     * @param context a map of objects that are made available for processing
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    protected final void doProcess(final Reader reader, final Writer writer,
            final ContextMap context) throws IOException, ProcessingException {
        final Document document;

        try {
            // reader should contain a well-formed document
            document = DOMUtils.readDocument(reader);
        } catch (DocumentException de) {
            // wrap exception
            throw new ParsingException(de);
        }

        // delegate to abstract method
        process(document, writer, context);
    }

    /**
     * Reads an XML document and stores the results of processing in a writer.
     *
     * @param document the document to read
     * @param writer the writer to write the processed text to; flushed
     * but not closed by this method
     * @param context a map of objects that are made available for processing
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public abstract void process(final Document document, final Writer writer,
            final ContextMap context) throws IOException, ProcessingException;

}
