/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties;

import java.util.Iterator;
import java.util.List;

import org.apache.commons.lang.ArrayUtils;

import de.fu_berlin.ties.util.TaskRunner;
import de.fu_berlin.ties.util.Util;

/**
 * A static class that provides a main external
 * entry point to TIE. The {@link #main(String[])} method executes a specified
 * {@link de.fu_berlin.ties.Processor} on a series of input arguments (typically
 * files or URLs).
 *
 * <p>No instances of this class can be created, only the static members
 * should be used.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.22 $, $Date: 2004/12/10 11:48:47 $, $Author: siefkes $
 */
public final class Main {

    /**
     * The name of this software.
     */
    public static final String NAME = "Trainable Information Extractor";

    /**
     * The current version number of this software.
     */
    public static final String VERSION_NO = "0.2";

    /**
     * Main method: execute the given goal (first argument) on the given
     * arguments.
     *
     * @param args the command-line parameters: first element must be the name
     * of the {@link Processor} to execute (mapped as configured in "goal"
     * section), each further element is given as input to an invocation of the
     * {@link Processor#process(String)} method; elements starting with "+" or
     * "-" are processed as configuration options (cf.
     * {@link TiesConfiguration#configureFromArgs(String[])}
     */
    public static void main(final String[] args) {
        final long startTime = System.currentTimeMillis();
        Util.LOG.debug("Starting execution, running on "
                + System.getProperty("java.vm.name") + ' '
                + System.getProperty("java.vm.version") + " installed in "
                + System.getProperty("java.home"));

        // ensure the default task runner is always available
        TaskRunner.registerInterest();

        try {
            // modify standard configuration + collect remaining arguments
            List arguments = null;
            try {
                arguments = TiesConfiguration.CONF.configureFromArgs(args);
            } catch (Exception e) {
                Util.LOG.fatalError("Illegal option", e);
                return;
            }

            if (arguments.size() < 2) {
                // not enough remaining arguments: print version info +
                // short help message to standard out and exit
                System.out.println(NAME + ' ' + VERSION_NO);
                System.out.println(
                    "Specify processing goal as first argument and the files "
                    + "or URLs to process as further arguments");
                return;
            }
            final Iterator argIter = arguments.iterator();

            // reconfigure logger
            Util.configureLog(TiesConfiguration.CONF);

            // check goal (we check above that there are enough arguments)
            final String goalName = (String) argIter.next();
            final String goalParam = TiesConfiguration.joinKey(
                TiesConfiguration.CONFIG_GOAL_PREFIX, goalName);

            if (!TiesConfiguration.CONF.containsKey(goalParam)) {
                // unknown goal: exit
                Util.LOG.error(goalName + " is not a registered goal");
                return;
            }

            // first argument is class name, any further arguments are
            // constructor parameters
            final String[] goalDefinition =
                TiesConfiguration.CONF.getStringArray(goalParam);

            // try to load goal processor with specified arguments
            final Processor goal;
            String input;

            try {
                goal = (Processor) Util.createObject(goalDefinition);
                Util.LOG.debug("Executing " + goalName + " goal (definition: "
                    + ArrayUtils.toString(goalDefinition) + ")");
            } catch (Exception e) {
                // unable to load goal: exit
                Util.LOG.error("Could not initialize processor for " + goalName
                    + " goal (definition: "
                    + ArrayUtils.toString(goalDefinition) + ")", e);
                return;
            }
            int errorCount = 0;

            // feed each further argument to the goal's process method
            try {
                while (argIter.hasNext()) {
                    input = (String) argIter.next();
                    try {
                        goal.process(input);
                    } catch (Exception e) {
                        Util.LOG.error("Error while processing " + input, e);
                        errorCount++;
                        if (errorCount >= 3) {
                            // exit after third error
                            Util.LOG.fatalError("Too many errors");
                            return;
                        }
                    }
                }
            } finally {
                // close the processor if it implements the Closable interface
                if (goal instanceof Closeable) {
                    final Closeable closeableGoal = (Closeable) goal;

                    try {
                        closeableGoal.close(errorCount);
                    } catch (Exception e) {
                        Util.LOG.error("Error while closing goal processor", e);
                        return;
                    }
                }
            }

            // print run time
            Util.LOG.debug("Done (" + Util.showDuration(startTime) + ")");
        } catch (Throwable t) {
            Util.LOG.fatalError("Exiting due to fatal error", t);
        } finally {
            // always release the default task runner
            TaskRunner.deregisterInterest();
        }
    }

    /**
     * Private constructor prevents creation of instances.
     */
    private Main() {
        super();
    }

}
