/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.io.ContentType;
import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.util.Util;

/**
 * Abstract base class for a {@link de.fu_berlin.ties.Processor} that operates
 * on text documents. Input is read from a file or URL or
 * {@link java.io.Reader}, output is written to a file or
 * {@link java.io.Writer}.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.22 $, $Date: 2004/10/12 14:17:19 $, $Author: siefkes $
 */
public abstract class TextProcessor extends ConfigurableProcessor {

    /**
     * Configuration prefix for post-processors.
     */
    public static final String CONFIG_POST = "post";

    /**
     * Context key referring to the local name of the processed document.
     */
    public static final String KEY_LOCAL_NAME = "local";

    /**
     * Context key referring to the directory of the processed document,
     * if it is a local file.
     */
    public static final String KEY_DIRECTORY = "directory";

    /**
     * Context key referring output directory; if missing, the value of
     * {@link #KEY_DIRECTORY} is used instead.
     */
    public static final String KEY_OUT_DIRECTORY = "outdir";

    /**
     * Context key referring to the URL of the processed document, if loaded
     * from an URL.
     */
    public static final String KEY_URL = "url";

    /**
     * The extension used for output files.
     */
    private final String outFileExt;

    /**
     * The configuration key for the preprocessor (or <code>null</code> if no
     * <code>outFileExt</code> is given).
     */
    private final String postProcKey;

    /**
     * A map of post-processors for different languages, if configured.
     * Maps from localized or global {@link #CONFIG_POST} key to initialized
     * {@link Processor} instance (or to <code>null</code> if no post-processor
     * exists). Synchronized on itself.
     */
    private final Map<String, Processor> postProcessors =
        new HashMap<String, Processor>();

    /**
     * Creates a new instance.
     *
     * @param outExt the extension to use for output files
     * @param conf used to configure this instance; if <code>null</code>,
     * the {@linkplain TiesConfiguration#CONF standard configuration} is used
     */
    public TextProcessor(final String outExt, final TiesConfiguration conf) {
        super(conf);
        outFileExt = outExt;

        if (outExt != null) {
            postProcKey = TiesConfiguration.joinKey(CONFIG_POST, outExt);
        } else {
            postProcKey = null;
        }
    }

    /**
     * Processes the contents of a reader, writing a modified version to a
     * writer.
     *
     * @param reader reader containing the text to process; should not be closed
     * by this method
     * @param writer the writer to write the processed text to; might be flushed
     * but not closed by this method; if this method does not use the writer,
     * the underlying file will be deleted afterwards
     * @param context a map of objects that are made available for processing;
     * when called from the implemented <code>process</code> methods in this
     * class, it will contain mappings from {@link IOUtils#KEY_LOCAL_CHARSET}
     * to the character set of the output writer; from
     * {@link ContentType#KEY_MIME_TYPE} to the document's MIME type; from
     * {@link #KEY_LOCAL_NAME} to the local name (String) and either from
     * {@link #KEY_DIRECTORY} to the directory ({@link File}), in case of a
     * local file) or from {@link #KEY_URL} to the {@link URL} (otherwise) of
     * the processed document
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    protected abstract void doProcess(final Reader reader, final Writer writer,
            final ContextMap context) throws IOException, ProcessingException;

    /**
     * Returns the extension used for output files.
     * @return the value of the attribute
     */
    public String getOutFileExt() {
        return outFileExt;
    }

    /**
     * Delegates to the abstract {@link #doProcess(Reader, Writer, ContextMap)}
     * method and invokes a post-processor, if configured.
     *
     * @param reader reader containing the text to process; should not be closed
     * by this method
     * @param writer the writer to write the processed text to; might be flushed
     * but not closed by this method; if this method does not use the writer,
     * the underlying file will be deleted afterwards
     * @param context a map of objects that are made available for processing;
     * when called from the implemented <code>process</code> methods in this
     * class, it will contain mappings from {@link IOUtils#KEY_LOCAL_CHARSET}
     * to the character set of the output writer; from
     * {@link ContentType#KEY_MIME_TYPE} to the document's MIME type; from
     * {@link #KEY_LOCAL_NAME} to the local name (String) and either from
     * {@link #KEY_DIRECTORY} to the directory ({@link File}), in case of a
     * local file) or from {@link #KEY_URL} to the {@link URL} (otherwise) of
     * the processed document
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public final void process(final Reader reader, final Writer writer,
            final ContextMap context) throws IOException, ProcessingException {
        final TextProcessor postProc;

        // determine whether there is a post-processor to invoke
        if (postProcKey != null) {
            final String localizedKey = getConfig().localizeKey(postProcKey);

            synchronized (postProcessors) {
                if (postProcessors.containsKey(localizedKey)) {
                    // load stored instance resp. "null" if there is none
                    postProc = (TextProcessor) postProcessors.get(localizedKey);
                } else {
                    // try to load instance from configuration
                    if (getConfig().containsKey(localizedKey)) {
                        // first argument is class name, any further arguments
                        // are constructor parameters
                        final String[] classDefinition =
                            TiesConfiguration.CONF.getStringArray(localizedKey);

                        try {
                            postProc = (TextProcessor)
                                Util.createObject(classDefinition);
                            Util.LOG.debug(getClass().getName()
                                + ": Initialized post-processor from "
                                + localizedKey + " (definition: "
                                + ArrayUtils.toString(classDefinition) + ")");
                        } catch (Exception e) {
                            // wrap and rethrow
                            throw new ProcessingException(
                                "Could not initialize post-processor from "
                                + localizedKey + " (definition: "
                                + ArrayUtils.toString(classDefinition) + ")",
                                e);
                        }
                    } else {
                        // remember that there is no post-processor
                        postProc = null;
                    }

                    // store initialized resp. null processor in map
                    postProcessors.put(localizedKey, postProc);
                }
            } // synchronized
        } else {
            postProc = null;
        }

        if (postProc != null) {
            final StringWriter interimWriter = new StringWriter();

            // delegate to abstract method and then to post-processor
            doProcess(reader, interimWriter, context);
            postProc.process(new StringReader(interimWriter.toString()),
                writer, context);
        } else {
            // delegate to abstract method
            doProcess(reader, writer, context);
        }
    }

    /**
     * Processes the contents of a file, delegating to the
     * {@link #process(File, Writer, ContextMap)} method.
     *
     * @param file the file to process
     * @param writer the writer to write the processed text to; not closed by
     * this method
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public final void process(final File file, final Writer writer)
    throws IOException, ProcessingException {
        // create a context and store local charset, if configured
        final ContextMap context = new ContextMap();
        if (getConfig().containsKey(IOUtils.KEY_LOCAL_CHARSET)) {
            context.put(IOUtils.KEY_LOCAL_CHARSET,
                getConfig().getString(IOUtils.KEY_LOCAL_CHARSET));
        }

        process(file, writer, context);
    }

    /**
     * Processes the contents of a file, delegating to the
     * {@link #process(Reader, Writer, ContextMap)} method. Stores a mapping
     * from {@link ContentType#KEY_MIME_TYPE} to the document's MIME type in the
     * {@link ContextMap context}.
     *
     * @param file the file to process
     * @param writer the writer to write the processed text to; not closed by
     * this method
     * @param context a map of objects that are made available for processing;
     * should contain a mapping from {@link IOUtils#KEY_LOCAL_CHARSET} to the
     * character set to use for local files
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public final void process(final File file, final Writer writer,
            final ContextMap context)
    throws IOException, ProcessingException {
        // rely on file extension for content type
        final ContentType contentType =
            ContentType.determineContentType(file, getConfig());

        // store the MIME type + directory in the context
        context.put(ContentType.KEY_MIME_TYPE, contentType.getMimeType());
        context.put(KEY_DIRECTORY, file.getParentFile());
        final String charset = (String) context.get(IOUtils.KEY_LOCAL_CHARSET);

        // open reader using configured charset
        final InputStreamReader reader = IOUtils.openReader(file, charset);

        if (charset == null) {
            // no charset given-- store the platform-specific default charset
            context.put(IOUtils.KEY_LOCAL_CHARSET,
                    IOUtils.determineCharsetName(reader));
        }

        try {
            // delegate to actual method
            process(reader, writer, context);
        } finally {
            IOUtils.tryToClose(reader);
        }
    }

    /**
     * Processes a file or URL given as input argument, delegating to the
     * appropriate <code>process</code> method. A warning is logged if the input
     * is neither a readable file nor a readable URL. Stores a mapping
     * from {@link IOUtils#KEY_LOCAL_CHARSET} to the character set of the
     * output writer in the created {@link ContextMap context}.
     *
     * @param inputName the name of a readable file or URL to process
     * @throws IOException if an I/O error occurs during processing
     * @throws ProcessingException if an error occurs during processing
     */
    public final void process(final String inputName)
            throws IOException, ProcessingException {
        final long startTime = System.currentTimeMillis();
        URL inURL = null;
        URLConnection inConn = null;
        String dirName = null;
        String localName = null;
        boolean isOkay;
        Util.LOG.debug("Starting to process " + inputName);

        // check whether it is a file or an URL
        final File inFile = new File(inputName);
        if (inFile.exists() && inFile.canRead()) {
            // for determining output file name
            dirName = inFile.getParent();
            localName = inFile.getName();
            isOkay = true;
        } else {
            // if it's not a file check whether it's an URL
            try {
                inURL = new URL(inputName);
                inConn = inURL.openConnection();

                // for determining output file name
                dirName = null;
                localName = IOUtils.getLocalName(inURL, true);
                isOkay = true;
            } catch (IOException ioe) {
                isOkay = false;
            }
        }

        // where to write the output file
        final String outDirName =
            getConfig().getString(KEY_OUT_DIRECTORY, dirName);
        final File outDir = (outDirName != null) ? new File(outDirName) : null;

        if (isOkay) {
            final File outFile = IOUtils.createOutFile(outDir,
                    localName, getOutFileExt());
            final OutputStreamWriter writer;
            final String localCharset;

            // check whether a local charset is configured
            if (getConfig().containsKey(IOUtils.KEY_LOCAL_CHARSET)) {
                // use configured charset
                localCharset = getConfig().getString(IOUtils.KEY_LOCAL_CHARSET);
                writer = new OutputStreamWriter(
                    new FileOutputStream(outFile), localCharset);
            } else {
                // use platform default charset for output
                localCharset = null;
                writer = new FileWriter(outFile);
            }

            // create a context and store local charset + local name
            final ContextMap context = new ContextMap();
            final String usedCharset = (localCharset != null)
                ? localCharset
                : IOUtils.determineCharsetName(writer);
            context.put(IOUtils.KEY_LOCAL_CHARSET, usedCharset);
            context.put(KEY_LOCAL_NAME, localName);

            try {
                if (inConn != null) {
                    // delegate to URL processing method
                    process(inConn, writer, context);
                } else {
                    // delegate to file processing method
                    process(inFile, writer, context);
                }

                writer.flush();
            } finally {
                IOUtils.tryToClose(writer);

                // delete the out file if no output was written
                if (outFile.length() == 0) {
                    if (!outFile.delete()) {
                        Util.LOG.warn("Could not delete empty output file "
                            + outFile);
                    }
                }
            }

            if (outFile.exists() && (outFile.length() > 0)) {
                Util.LOG.info("Results of processing " + inputName
                    + " stored in " + outFile + " ("
                    + Util.showDuration(startTime) + ")");
            } else {
                Util.LOG.info("Processed " + inputName + " ("
                    + Util.showDuration(startTime) + ")");
            }
        } else {
            Util.LOG.warn(inputName + " is neither a readable "
                + "file nor a readable URL -- ignoring it");
        }
    }

    /**
     * Processes the contents of an URL connection, delegating to the
     * {@link #process(URLConnection, Writer, ContextMap)} method.
     * Assumed the {@linkplain IOUtils#STANDARD_HTTP_CHARSET standard HTTP
     * character set} ("ISO-8859-1") if no {@link IOUtils#KEY_LOCAL_CHARSET} has been configured.
     *
     * @param urlConn the URL connection to process
     * @param writer the writer to write the processed text to; not closed by
     * this method
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public final void process(final URLConnection urlConn, final Writer writer)
    throws IOException, ProcessingException {
        // create a context and store local charset (default to HTTP standard)
        final ContextMap context = new ContextMap();
        final String charset = getConfig().getString(IOUtils.KEY_LOCAL_CHARSET,
                IOUtils.STANDARD_HTTP_CHARSET);
        context.put(IOUtils.KEY_LOCAL_CHARSET, charset);

        process(urlConn, writer, context);
    }

    /**
     * Processes the contents of an URL connection, delegating to the
     * {@link #process(Reader, Writer, ContextMap)} method. Stores a mapping
     * from {@link ContentType#KEY_MIME_TYPE} to the document's MIME type in the
     * {@link ContextMap context}.
     *
     * @param urlConn the URL connection to process
     * @param writer the writer to write the processed text to; not closed by
     * this method
     * @param context a map of objects that are made available for processing;
     * must contain a mapping from {@link IOUtils#KEY_LOCAL_CHARSET} to the
     * character set to use for local files
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public final void process(final URLConnection urlConn, final Writer writer,
            final ContextMap context) throws IOException, ProcessingException {
        // use Content-Type header + URL extension to determine content type
        final URL url = urlConn.getURL();
        final String contentTypeHeader = urlConn.getContentType();
        final ContentType contentType = ContentType.determineContentType(
                url, contentTypeHeader, getConfig());

        // store the MIME type + URL in the context
        context.put(ContentType.KEY_MIME_TYPE,
            contentType.getMimeType());
        context.put(KEY_URL, url);

        // determine charset + open reader
        final String charset;
        if (contentType.getCharset() != null) {
            // use determined charset
            charset = contentType.getCharset();
        } else {
            // use configured local charset
            charset = (String) context.get(IOUtils.KEY_LOCAL_CHARSET);
        }

        final InputStreamReader reader =
            new InputStreamReader(urlConn.getInputStream(), charset);

        try {
            // delegate to actual method
            process(reader, writer, context);
        } finally {
            IOUtils.tryToClose(reader);
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("output file extension", outFileExt)
            .toString();
    }

}
