/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties;

import java.io.File;
import java.io.InputStream;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.configuration.CompositeConfiguration;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.ConfigurationKey;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.apache.commons.configuration.XMLConfiguration;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.util.Util;

/**
 * A composite configuration that allows to localize of keys and to querying
 * types and descriptions of entries.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.24 $, $Date: 2004/10/30 17:25:28 $, $Author: siefkes $
 */
public class TiesConfiguration extends CompositeConfiguration {

    /**
     * The main configuration object for TIES, loaded via
     * {@link #TiesConfiguration(String)} using "ties" as base name.
     * <strong>You should NOT modify the configuration properties stored in this
     * object, especially since it is not fully thread-safe (unless you know
     * what you're doing and are sure there are no other objects around that
     * might try to access the configuration, e.g. at the begin of a
     * program's <code>main</code> method).</strong>
     */
    public static final TiesConfiguration CONF = new TiesConfiguration("ties");

    /**
     * The subdirectory in the class path containing config files.
     */
    public static final String CONF_DIR = "conf";

    /**
     * The extension of config files in {@link PropertiesConfiguration}
     * format.
     */
    public static final String CONF_EXTENSION = ".cfg";

    /**
     * The extension of config files in {@linkplain XMLConfiguration XML
     * format}.
     */
    public static final String XML_EXTENSION = ".xml";

    /**
     * The extension of
     * {@linkplain #addDescriptorConfig(Configuration) descriptor configuration}
     * files (in {@link PropertiesConfiguration} format).
     */
    public static final String DESC_EXTENSION = ".desc";

    /**
     * Configuration key prefix mapping goals to fully qualified class names.
     */
    public static final String CONFIG_GOAL_PREFIX = "goal";

    /**
     * Special configuration key: the language of documents, using the <a
     * href="http://www.loc.gov/standards/iso639-2/englangn.html">ISO 639</a>
     * language codes (2-letter codes where available, e.g. "en" (English) or
     * "de" (German).
     */
    public static final String CONFIG_LANG = "lang";

    /**
     * The character separating tokens in keys.
     */
    private static final char PROPERTY_DELIM = '.';


    /**
     * Creates a adapted (caller-specific) key by joining a base name
     * with a suffix. If the adapted form of the key doesn't exist in this
     * configuration, the base form is returned instead.
     *
     * @param baseKey the basic key
     * @param suffix the suffix of the key; if <code>null</code> the
     * <code>baseKey</code> is returned
     * @return the full key, adapted by joining with the {@link #CONFIG_LANG}
     * key; resp. the base form if the full key doesn't exist or
     * <code>suffix</code> is <code>null</code>
     * @see #joinKey(String, String)
     */
    public String adaptKey(final String baseKey, final String suffix) {
        if (suffix != null) {
            final String adaptedKey = joinKey(baseKey, suffix);

            // return adapted key only if it exists, base form otherwise
            if (containsKey(adaptedKey)) {
                return adaptedKey;
            } else {
                return baseKey;
            }
        } else {
            return baseKey;
        }
    }

    /**
     * Utility method that thorougly checks whether a string array is empty.
     * It returns <code>true</code> in three cases:
     *
     * <ol>
     * <li>the specified array is <code>null</code>;
     * <li>the specified array is empty (0 elements);
     * <li>the specified array contains a single element which is
     * <code>null</code> or an empty or blank (only whitespace) string.
     * </ol>
     * 
     * This method is useful because some reconfigurations can cause a
     * <code>TiesConfiguration</code> to return a single-element list containing
     * only an empty string. Usually such a list should be treated as empty.
     *
     * @param array the array to check
     * @return <code>true</code> iff the list is empty as described above
     */
    public static boolean arrayIsEmpty(final String[] array) {
        final boolean result;

        if ((array == null) || array.length == 0) {
            // null or empty array
            result = true;
        } else if (array.length == 1) {
            // only one element: check if it's null, empty, or whitespace
            result = StringUtils.isBlank(array[0]);
        } else {
            result = false;
        }

        return result;
    }

    /**
     * Utility method that thorougly checks whether a list is empty. It returns
     * <code>true</code> in three cases:
     *
     * <ol>
     * <li>the specified list is <code>null</code>;
     * <li>the specified list is empty (0 elements);
     * <li>the specified list contains a single element which is
     * <code>null</code> or an empty or blank (only whitespace) string.
     * </ol>
     * 
     * This method is useful because some reconfigurations can cause a
     * <code>TiesConfiguration</code> to return a single-element list containing
     * only an empty string. Usually such a list should be treated as empty.
     *
     * @param list the list to check
     * @return <code>true</code> iff the list is empty as described above
     */
    public static boolean listIsEmpty(final List list) {
        final boolean result;

        if ((list == null) || list.isEmpty()) {
            // null or empty list
            result = true;
        } else if (list.size() == 1) {
            // only one element: check if it's the empty string (or null)
            final Object first = list.get(0);

            if (first == null) {
                // single element is null
                result = true;
            } else if (first instanceof String) {
                // true iff single element is an empty or blank string
                result = StringUtils.isBlank((String) first);
            } else {
                result = false;
            }
        } else {
            result = false;
        }

        return result;
    }

    /**
     * Utility method that thorougly checks whether a property (as returned
     * by {@link Configuration#getProperty(String)} is empty. It returns
     * <code>true</code> in two cases:
     *
     * <ol>
     * <li>the specified object is <code>null</code>;
     * <li>the specified object is an empty or blank (only whitespace) string.
     * </ol>
     * 
     * This method is useful because some reconfigurations can cause a
     * <code>TiesConfiguration</code> to return a empty strings.
     *
     * @param object the object to check
     * @return <code>true</code> iff the object is empty as described above
     */
    public static boolean propertyIsEmpty(final Object object) {
        final boolean result;

        if (object == null) {
            // null object
            result = true;
        } else if (object instanceof String) {
            // test if string is empty
            result = StringUtils.isBlank((String) object);
        } else {
            result = false;
        }

        return result;
    }

    /**
     * Creates a full key by joining a prefix and a suffix string, separated by
     * the property delimiter.
     *
     * @param prefix the prefix of the key
     * @param suffix the suffix of the key
     * @return the full key (prefix + delimiter + suffix)
     */
    public static String joinKey(final String prefix, final String suffix) {
        return new ConfigurationKey(prefix).append(suffix).toString();
    }


     /**
     * An inner class wrapping descriptor information on an entry: type
     * of the entry, whether it is optional or a list, a description of the
     * entry.
     */
    public final class EntryDescriptor {

        /**
         * The key (name) of this entry.
         */
        private final String key;

        /**
         * The base type of the entry (String, Integer etc.).
         */
        private final String type;

        /**
         * Whether the entry is optional.
         */
        private final boolean optional;

        /**
         * Whether the entry is a list (can contain multiple values).
         */
        private final boolean list;

        /**
         * A textual description of the entry.
         */
        private final String description;

        /**
         * Returns a textual description of the entry.
         * @return the value of the attribute
         */
        public String getDescription() {
            return description;
        }

        /**
         * Creates a new instance.
         *
         * @param keyName the key (name) of this entry
         * @param typ the base type of the entry
         * @param isOpt whether the entry is optional
         * @param isList whether the entry is a list
         * @param desc a textual description of the entry
         */
        private EntryDescriptor(final String keyName, final String typ,
                final boolean isOpt, final boolean isList, final String desc) {
            key = keyName;
            type = typ;
            optional = isOpt;
            list = isList;
            description = desc;
        }

        /**
         * Reads the value of this property as a string list and returns the
         * element at the specified position.
         *
         * @param index the index of element to return
         * @return the element at the specified position in this list; or
         * <code>null</code> if the position does not exist in the list or
         * no value exists for this key
         * @throws IndexOutOfBoundsException if <code>index</code> &lt; 0
         */
        public String getElement(final int index)
                throws IndexOutOfBoundsException {
            // ensure value is present
            if (containsKey(key)) {
                final List value = getList(key);

                // ensure element is present
                if (index < value.size()) {
                    return (String) value.get(index);
                } else {
                    return null;
                }
            } else {
                return null;
            }
        }

        /**
         * Reads the value of this property as a string list and returns a
         * sublist starting at the specified position.
         *
         * @param fromIndex the index of the first element to return (inclusive)
         * @return a sublist starting at the specified position in this list; or
         * <code>null</code> if the position does not exist in the list or
         * no value exists for this key
         * @throws IndexOutOfBoundsException if <code>fromIndex</code> &lt; 0
         */
        public List getElementsFrom(final int fromIndex)
                throws IndexOutOfBoundsException {
            // ensure value is present
            if (containsKey(key)) {
                final List value = getList(key);

                // ensure there is at least one element to return
                if (fromIndex < value.size()) {
                    return value.subList(fromIndex, value.size());
                } else {
                    return null;
                }
            } else {
                return null;
            }
        }

        /**
         * Returns the key (name) of this entry.
         * @return the value of the attribute
         */
        public String getKey() {
            return key;
        }

        /**
         * Returns the base type of the entry (String, Integer etc.).
         * @return the value of the attribute
         */
        public String getType() {
            return type;
        }

        /**
         * Returns the value of this property (as an object).
         * @return the value of this property; or <code>null</code> if no
         * value is set
         */
        public Object getValue() {
            // delegate to outer class if key is present
            return containsKey(key) ? getProperty(key) : null;
        }

        /**
         * Whether the entry is a list (can contain multiple values).
         * @return the value of the attribute
         */
        public boolean isList() {
            return list;
        }

        /**
         * Whether the entry is optional.
         * @return the value of the attribute
         */
        public boolean isOptional() {
            return optional;
        }

        /**
         * Returns a string representation of this object.
         *
         * @return a textual representation
         */
        public String toString() {
            return new ToStringBuilder(this)
                .append("key", key)
                .append("type", type)
                .append("optional", optional)
                .append("list", list)
                .append("description", description)
                .toString();
        }

    }


    /**
     * Describes the entries in this config. Each value must be a two-array,
     * specifying the type of the entry and a description of the entry.
     */
    private final CompositeConfiguration descriptorConfig =
        new CompositeConfiguration();


    /**
     * Creates a new empty instance.
     */
    public TiesConfiguration() {
        super();
    }

    /**
     * Creates a new instance, delegating to
     * {@link #addConfiguration(Configuration, Configuration)}.
     *
     * @param config the initial configuration to wrap
     * @param desc the initial descriptor config to wrap (cf.
     * {@link #addDescriptorConfig(Configuration)} for the required contents)
     */
    public TiesConfiguration(final Configuration config,
                             final Configuration desc) {
        this();
        addConfiguration(config, desc);
    }

    /**
     * Creates a new instance, delegating to {@link #load(String)}.
     *
     * @param baseName the base name of the configuration
     */
    public TiesConfiguration(final String baseName) {
        this();
        load(baseName);
    }

    /**
     * Adds a configuration and a corresponding descriptor config.
     *
     * @param config the configuration to add
     * @param desc the descriptor config to add (cf.
     * {@link #addDescriptorConfig(Configuration)} for the required contents)
     */
    public void addConfiguration(final Configuration config,
                                 final Configuration desc) {
        addConfiguration(config);
        addDescriptorConfig(desc);
    }

    /**
     * Adds a descriptor configuration that can be consulted to query the
     * type and use of a entries. Each value must be a two-array specifying
     * the type (first element) and a description (second element) of the key
     * as used in the main configuration. Append '?' to the type if it is
     * optional (zero or one values), '*' if it is an optional list (zero or
     * more values), or '+' if it is a required list (one or more values).
     *
     * @param desc the descriptor configuration to add
     */
    public void addDescriptorConfig(final Configuration desc) {
        descriptorConfig.addConfiguration(desc);
    }

    /**
     * Modifies configuration properties from <code>[+|-]key[=value]</code>
     * pairs in a string array. Delegates to
     * {@link #modifyProperty(String, boolean)} for each array element
     * starting with "-" (using overwrite mode) or "+" (using append mode).
     * Other array elements are collected and returned.
     *
     * @param args the array of strings to parse
     * @return a list of all arguments that do not start with "+" or "-"
     * @throws IllegalArgumentException if one of the
     * <code>[+|-]key[=value]</code> pairs doesn't contain a key ("=" is second
     * character)
     */
    public List configureFromArgs(final String[] args)
            throws IllegalArgumentException {
        final List<String> otherArgs = new LinkedList<String>();
        for (int i = 0; i < args.length; i++) {
            if (args[i].startsWith("-")) {
                modifyProperty(args[i].substring(1), true);
            } else if (args[i].startsWith("+")) {
                modifyProperty(args[i].substring(1), false);
            } else {
                // adds to other arguments
                otherArgs.add(args[i]);
            }
        }
        return otherArgs;
    }

    /**
     * Returns the descriptor for a given key, if any is given in the
     * {@linkplain #addDescriptorConfig(Configuration) descriptor
     * configuration}.
     *
     * @param key the key to describe
     * @return a descriptor object for this key, or <code>null</code> is none
     * is given
     */
    public EntryDescriptor getDescriptor(final String key) {
        final EntryDescriptor result;
        final String[] descEntry;
        final String suffixDesc;

        if (descriptorConfig.containsKey(key)) {
            descEntry = descriptorConfig.getStringArray(key);
            suffixDesc = null;
        } else {
            // split last part from key (language or format-specific) + add
            // specific description in paranthesis
            final int lastDelim = key.lastIndexOf(PROPERTY_DELIM);

            if (lastDelim > 0) {
                final String mainKey = key.substring(0, lastDelim);
                final String suffix = key.substring(lastDelim);
                descEntry = descriptorConfig.getStringArray(mainKey);
                suffixDesc = descriptorConfig.getString(suffix, null);
            } else {
                descEntry = null;
                suffixDesc = null;
            }
        }

        if (descEntry != null) {
            if (descEntry.length >= 2) {
                // first element contains type info
                final String firstElem = descEntry[0].trim();
                final String type;
                final boolean optional;
                final boolean list;

                // check for (and remove) special chars ?*+ at end of string
                if (firstElem.endsWith("?")) {
                    optional = true;
                    list = false;
                    type = StringUtils.chop(firstElem);
                } else if (firstElem.endsWith("*")) {
                    optional = true;
                    list = true;
                    type = StringUtils.chop(firstElem);
                } else if (firstElem.endsWith("+")) {
                    optional = false;
                    list = true;
                    type = StringUtils.chop(firstElem);
                } else {
                    optional = false;
                    list = false;
                    type = firstElem;
                }

                // second element contains actual description
                String description = descEntry[1];

                // append suffix description if paranthesis, if given
                if (!StringUtils.isEmpty(suffixDesc)) {
                    description += " (" + suffixDesc + ")";
                }

                result =
                    new EntryDescriptor(key, type, optional, list, description);
            } else {
                result = null;
            }

            // log warning if there are too few or to many elements
            if ((descEntry.length == 1) || (descEntry.length > 2)) {
                Util.LOG.error("Descriptor entry for " + key
                    + " has " + descEntry.length + " elements instead of 2");
            }
        } else {
            result = null;
        }

        return result;
    }

    /**
     * Read a property from this configuration. If the given key does not exist
     * in this configuration, it is looked up as a
     * {@linkplain System#getProperty(java.lang.String) Java system property}.
     *
     * @param key key to use for mapping
     * @return object associated with the given configuration key
     */
    protected Object getPropertyDirect(final String key) {
        Object result = super.getPropertyDirect(key);

        // if the superclass doesn't know it, maybe it's a Java system property?
        if (result == null) {
            try {
                result = System.getProperty(key);
            } catch (SecurityException se) {
                // not allowed to look it up
                result = null;
            }
        }

        return result;
    }

    /**
     * Get an array of strings associated with the given configuration key.
     *
     * @param key The configuration key
     * @return The associated string array if key is found.
     */
    public String[] getStringArray(final String key) {
        final String[] result = super.getStringArray(key);

        if ((result.length == 1) && (StringUtils.isBlank(result[0]))) {
            // convert array containg a single null/empty/blank element into
            // an empty array
            return new String[0];
        } else {
            return result;
        }
    }

    /**
     * Copies all properties contained in this instance to a given
     * configuration. This method can be used for storing this configuration in
     * any format desired, by creating an empty configuration of the
     * requested type and passing it to this method. Keys are sorted in
     * alphabetic order.
     *
     * @param appender the configuration to append all properties to
     */
    public void flatten(final Configuration appender) {
        final Iterator keys = sortedKeys().iterator();
        String key;
        Object value;

        while (keys.hasNext()) {
            key = (String) keys.next();
            value = getProperty(key);
            appender.addProperty(key, value);
        }
    }

    /**
     * Loads configuration in {@link PropertiesConfiguration} or
     * {@linkplain org.apache.commons.configuration.XMLConfiguration XML}
     * format. Combines configuration properties from several resources if
     * they exist:
     *
     * <ol>
     * <li>The files <code><em>baseName</em>.xml</code> resp.
     * <code><em>baseName</em>.cfg</code> in the current working
     * directory, if they exist and are readable files</li>
     * <li>The file <code><em>baseName</em>.xml</code> resp.
     * <code><em>baseName</em>.cfg</code> in the user's home
     * directory, if they exist and are readable files</li>
     * <li>The file <code><em>baseName</em>.cfg</code> in the <code>conf</code>
     * subdirectory in the classpath (e.g. from a jar file) -- <strong>this
     * classpath resource should always exist (otherwise an warning will be
     * logged and later errors are likely)</strong></li>
     * </ol>
     *
     * <p>For shared keys, the first match will be returned. If corresponding
     * <code><em>baseName</em>.desc</code> exists, they are assumed to
     * contain {@linkplain #addDescriptorConfig(Configuration) descriptors}.
     *
     * <p>*.cfg and *.desc files must must use the
     * {@link PropertiesConfiguration} format, *.xml files must use the
     * {@linkplain org.apache.commons.configuration.XMLConfiguration XML}
     * format.
     *
     * <p>The last resource loaded form the classpath must use UTF-8 character
     * set (otherwise it could not be shared between systems). The other
     * *.cfg and *.desc (local files) are assumed to use the platform's default
     * character set. The character set of XML files is determined in the
     * standard way.
     *
     * @param baseName the base name of the configuration
     */
    public void load(final String baseName) {
        // directories to check
        final String[] dirCandidates = {
            System.getProperty("user.dir"),
            System.getProperty("user.home")
        };
        File file;

        // check .xml + .cfg + .desc files in each directory
        for (int i = 0; i < dirCandidates.length; i++) {
            // check .xml file (XML format)
            file = new File(dirCandidates[i] + File.separator + baseName
                    + XML_EXTENSION);
            try {
                if (file.canRead() && file.isFile()) {
                    addConfiguration(new XMLConfiguration(file));
                }

                // check .cfg file (PropertiesConfig format)
                file = new File(dirCandidates[i] + File.separator + baseName
                        + CONF_EXTENSION);
                if (file.canRead() && file.isFile()) {
                    addConfiguration(new PropertiesConfiguration(
                            file.getAbsolutePath()));
                }

                // check .desc file (descriptor config in PropertiesConf format)
                file = new File(dirCandidates[i] + File.separator + baseName
                        + DESC_EXTENSION);
                if (file.canRead() && file.isFile()) {
                    addDescriptorConfig(new PropertiesConfiguration(
                            file.getAbsolutePath()));
                }
            } catch (Exception e) {
                Util.LOG.error("Couldn't load configuarion file "
                    + file.getAbsolutePath() + ": " + e);
            }
        }

        // load default values + descriptors from class path (class path always
        // uses '/' as separator)
        String classPathLocation
            = CONF_DIR + '/' + baseName + CONF_EXTENSION;
        InputStream in =
            ClassLoader.getSystemResourceAsStream(classPathLocation);

        try {
            if (in != null) {
                final PropertiesConfiguration defaults =
                    new PropertiesConfiguration();
                defaults.load(in, IOUtils.STANDARD_UNICODE_CHARSET);
                addConfiguration(defaults);
                IOUtils.tryToClose(in);
            } else {
                Util.LOG.warn("Didn't find default configuration "
                    + classPathLocation + " in classpath");
            }

            classPathLocation = CONF_DIR + File.separator + baseName
                + DESC_EXTENSION;
            in = ClassLoader.getSystemResourceAsStream(classPathLocation);

            if (in != null) {
                    final PropertiesConfiguration defaultDesc =
                        new PropertiesConfiguration();
                    defaultDesc.load(in, IOUtils.STANDARD_UNICODE_CHARSET);
                    addDescriptorConfig(defaultDesc);
            }
        } catch (ConfigurationException ce) {
            Util.LOG.error("Couldn't load default configuration "
                + classPathLocation + " from classpath: " + ce);
        } finally {
            IOUtils.tryToClose(in);
        }

    }

    /**
     * Creates a localized (language-specific) key by joining a base name
     * with the configured language suffix (value mapped to the
     * {@link #CONFIG_LANG}} key (if this key doesn't exist, the language of
     * default locale used by the Java Virtual Machine is used).
     * If the localized form of the key doesn't exist in the specified
     * configuration, the base form is returned instead. Implemented by
     * delegating to {@link #adaptKey(String, String)}.
     *
     * @param baseKey the basic key
     * @return the full key, localized by joining with the {@link #CONFIG_LANG}
     * key
     */
    public String localizeKey(final String baseKey) {
        final String lang = getString(CONFIG_LANG,
                Locale.getDefault().getLanguage());
        return adaptKey(baseKey, lang);
    }

    /**
     * Modifies a configuration property, parsing a <code>key[=value]</code>
     * pair. Both key and value are trimmed. {@link Boolean#TRUE} is used as
     * default value if the <code>=value</code> part is omitted.
     * If "=" is the last character (empty value) and <code>overwrite</code> is
     * <code>true</code>, the property is removed from the configuration
     * ({@link Configuration#clearProperty(java.lang.String)}).
     *
     * @param keyValue the <code>key=value</code> pair to parse
     * @param overwrite whether to overwrite (replacing old value: {@link
     * Configuration#setProperty(java.lang.String, java.lang.Object)}) or
     * add (append to list, {@link
     * Configuration#addProperty(java.lang.String, java.lang.Object)}) the
     * value
     * @throws IllegalArgumentException if <code>keyValue</code> doesn't contain
     * a key ("=" is first character)
     */
    public void modifyProperty(final String keyValue,
            final boolean overwrite) throws IllegalArgumentException {
        // parse string
        final String key;
        final Object value;
        final int equalIndex = keyValue.indexOf('=');

        if (equalIndex > 0) {
            key = keyValue.substring(0, equalIndex).trim();
            value = keyValue.substring(equalIndex + 1).trim();
        } else if (equalIndex < 0) {
            // boolean "true" can be omitted (default value)
            key = keyValue.trim();
            // specify String instead of boolean to avoid problems during save
            value = "true";
        } else {
            // '=' is first character
            throw new IllegalArgumentException("Missing key: " + keyValue);
        }

        if (overwrite) {
            // empty string means: delete value
            if ("".equals(value)) {
                clearProperty(key);
                Util.LOG.debug("Cleared property " + key);
            } else {
                setProperty(key, value);
                Util.LOG.debug("Set property " + key + " = " + value);
            }
        } else {
            addProperty(key, value);
            Util.LOG.debug("Added to property " + key + ": " + value);
        }
    }

    /**
     * Saves the contents of this configuration in a file, storing them in
     * {@link PropertiesConfiguration} format. The platform's default
     * character set is used.
     *
     * @param file the file to use
     * @throws ConfigurationException if an I/O error occurs during saving
     */
    public void save(final File file) throws ConfigurationException {
        // currently [Base]PropertiesConfig except a String instead of a File
        save(file.getAbsolutePath());
    }

    /**
     * Saves the contents of this configuration in a file, storing them in
     * {@link PropertiesConfiguration} format. The platform's default
     * character set is used.
     *
     * @param filename the name of the file
     * @throws ConfigurationException if an I/O error occurs during saving
     */
    public void save(final String filename) throws ConfigurationException {
        final PropertiesConfiguration outConf = new PropertiesConfiguration();
        flatten(outConf);
        outConf.save(filename);
    }

    /**
     * Returns the list of keys contained in this configuration, sorted in
     * alphabetic order. Delegates to {@link #sortedKeys(boolean)}, setting
     * <code>inclNotSet</code> to <code>false</code>.
     *
     * @return an immutable set containing the sorted list of keys
     */
    public SortedSet sortedKeys() {
        return sortedKeys(false);
    }

    /**
     * Returns the list of keys contained in this configuration, sorted in
     * alphabetic order.
     *
     * @param inclNotSet if <code>true</code>, keys that are not contained in
     * this configuration but that described in the
     * {@linkplain #addDescriptorConfig(Configuration) descriptor configuration}
     * are also included
     * @return an immutable set containing the sorted list of keys
     */
    public SortedSet<Object> sortedKeys(final boolean inclNotSet) {
        final SortedSet<Object> result = new TreeSet<Object>();
        final Iterator keys = getKeys();

        while (keys.hasNext()) {
            result.add(keys.next());
        }

        if (inclNotSet) {
            // include additional keys from descriptor config
            final Iterator descKeys = descriptorConfig.getKeys();
            String key;

            while (descKeys.hasNext()) {
                key = (String) descKeys.next();

                // don't re-add key if it is contained in this config or if
                // it a prefix of contained keys or if it is a prefix descriptor
                // (starting with a dot)
                if (!containsKey(key) && super.subset(key).isEmpty()
                        && (key.indexOf(PROPERTY_DELIM) != 0)) {
                    result.add(key);
                }
            }
        }

        return Collections.unmodifiableSortedSet(result);
    }

    /**
     * Create an Configuration object that is a subset of this one. The new
     * Configuration object contains every key from the current Configuration
     * that starts with prefix. The prefix is removed from the keys in the
     * subset.
     *
     * @param prefix The prefix used to select the properties
     * @return a subset of this configuration; the returned object will be
     * a {@link TiesConfiguration} instance
     */
    public Configuration subset(final String prefix) {
        return new TiesConfiguration(super.subset(prefix),
            descriptorConfig.subset(prefix));
    }

}
