/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify;

import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TextProcessor;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.feature.DefaultFeatureVector;
import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.io.FieldContainer;
import de.fu_berlin.ties.io.FieldMap;
import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.text.TextTokenizer;
import de.fu_berlin.ties.text.TokenizerFactory;
import de.fu_berlin.ties.util.Util;

/**
 * Classifies a list of files, training the classifier on each error. See
 * {@link #classifyAndTrain(FieldContainer, File, String)} for a description of
 * input and output formats.
 *
 * <p>This class does not calculate statistics; you can do so be calling e.g.
 * <code>tail -q --lines 500 <em>FILENAME</em>|grep -v "|+"|wc</code> on the
 * output serialized in {@link de.fu_berlin.ties.io.DelimSepValues} format to
 * get the number of errors during the last 500 classifications (assuming that
 * classes to not start with a "+").
 *
 * <p>Instances of this class are thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.13 $, $Date: 2004/11/17 09:15:10 $, $Author: siefkes $
 */
public class ClassTrain extends TextProcessor {

    /**
     * Configuration key: The extension to append to file names given via the
     * {@linkplain #KEY_FILE File key} (if any).
     */
    public static final String CONFIG_FILE_EXT = "file.ext";

    /**
     * Serialization key for the name of the file to classify.
     */
    public static final String KEY_FILE = "File";

    /**
     * Serialization key for the correct class.
     */
    public static final String KEY_CLASS = "Class";

    /**
     * Serialization key for the result of the classification: either
     * {@link #CORRECT_CLASS} if the correct class was predicted or the
     * wrongly predicted class in case of an error.
     */
    public static final String KEY_CLASSIFICATION = "Classification";

    /**
     * Value of the {@link #KEY_CLASSIFICATION} field for correct predictions:
     * {@value}.
     */
    public static final String CORRECT_CLASS = "+";

    /**
     * Used to create tokenizers.
     */
    private final TokenizerFactory tFactory;

    /**
     * The extension to append to file names given via the
     * {@linkplain #KEY_FILE File key}; empty string if none.
     */
    private final String fileExtension;

    /**
     * Creates a new instance using a default extension and the
     * {@link TiesConfiguration#CONF standard configuration}.
     */
    public ClassTrain() {
        this("cls");
    }

    /**
     * Creates a new instance using the
     * {@link TiesConfiguration#CONF standard configuration}.
     *
     * @param outExt the extension to use for output files
     */
    public ClassTrain(final String outExt) {
        this(outExt, TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance from the provided configuration.
     *
     * @param outExt the extension to use for output files
     * @param conf used to configure this instance; if <code>null</code>,
     * the {@linkplain TiesConfiguration#CONF standard configuration} is used
     */
    public ClassTrain(final String outExt, final TiesConfiguration conf) {
        this(outExt, conf,
            new TokenizerFactory(conf, Classifier.CONFIG_CLASSIFIER),
            conf.getString(CONFIG_FILE_EXT, ""));
    }

    /**
     * Creates a new instance.
     *
     * @param outExt the extension to use for output files
     * @param conf used to configure this instance; if <code>null</code>,
     * the {@linkplain TiesConfiguration#CONF standard configuration} is used
     * @param factory used to create tokenizers
     * @param fileExt the extension to append to file names given via the
     * {@linkplain #KEY_FILE File key}; <code>null</code> or the empty string
     * if none should be appended
     */
    public ClassTrain(final String outExt, final TiesConfiguration conf,
        final TokenizerFactory factory, final String fileExt) {
        super(outExt, conf);
        tFactory = factory;
        fileExtension = (fileExt != null) ? fileExt : "";
    }

    /**
     * Classifies a list of files, training the classifier on each error.
     *
     * @param filesToClassify a field container of the files to process; each
     * entry must contain a {@link #KEY_FILE} field giving the name of the file
     * to classify and {@link #KEY_CLASS} giving the true class of the file
     * @param directory file names are relative to this directory; if
     * <code>null</code> they are relative to the working directory
     * @param charset the character set of the files to process
     * @return a field container of the classification results; in addition to
     * the fields given above, each entry will contain the classification result
     * in a {@link #KEY_CLASSIFICATION} field: {@link #CORRECT_CLASS} in
     * case of a correct classification, the name of the wrongly predicted
     * class otherwise
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public FieldContainer classifyAndTrain(final FieldContainer filesToClassify,
            final File directory, final String charset)
            throws IOException, ProcessingException {
        final FieldContainer result = FieldContainer.createFieldContainer();
        final int numFiles = filesToClassify.size();
        FieldMap inMap;
        FieldMap outMap;
        String currentClass;
        String[] filenames = new String[numFiles];
        String[] classes = new String[numFiles];
        final Iterator fileIter = filesToClassify.entryIterator();
        final Set<String> classSet = new HashSet<String>();
        int i = 0;

        // collect files to process and determine set of classes
        while (fileIter.hasNext()) {
            inMap = (FieldMap) fileIter.next();
            filenames[i] = (String) inMap.get(KEY_FILE);
            currentClass = (String) inMap.get(KEY_CLASS);
            classSet.add(currentClass);
            classes[i] = currentClass;
            i++;
        }

        // initialize classifier and tokenizer
        final TrainableClassifier classifier =
            TrainableClassifier.createClassifier(classSet, getConfig());
        final TextTokenizer tokenizer = tFactory.createTokenizer("");

        // ensure that prediction model is empty
        classifier.reset();

        Reader reader;
        String contents;
        FeatureVector features;
        PredictionDistribution predDist;
        Prediction best;

        // classify and if necessary train each listed file
        for (i = 0; i < numFiles; i++) {
            // read contents of file (relative to specified directory if any)
            reader = IOUtils.openReader(
                new File(directory, filenames[i] + fileExtension), charset);

            try {
                // create feature vector by tokenizing contents
                contents = IOUtils.readToString(reader);
                features = new DefaultFeatureVector();
                features.addAllTokens(contents, tokenizer);

                // delegate to classifier
                predDist =
                    classifier.trainOnError(features, classes[i], classSet);

                // log + store results
                outMap = new FieldMap();
                outMap.put(KEY_FILE, filenames[i]);
                outMap.put(KEY_CLASS, classes[i]);

                if (predDist == null) {
                    Util.LOG.debug("Processed " + filenames[i] + fileExtension
                            + ": classification as " + classes[i]
                            + " was correct");
                    outMap.put(KEY_CLASSIFICATION, CORRECT_CLASS);
                } else {
                    best = predDist.best();
                    Util.LOG.debug("Processed " + filenames[i] + fileExtension
                            + ": misclassified as " + best.getType()
                            + " instead of " + classes[i]);
                    outMap.put(KEY_CLASSIFICATION, best.getType());
                }
                result.add(outMap);
            } finally {
                IOUtils.tryToClose(reader);
            }
        } // for

        Util.LOG.debug("Finished classifying and training using "
                + classifier + " and " + tokenizer);
        return result;
    }

    /**
     * Delegates to {@link #classifyAndTrain(FieldContainer, File, String)}.
     *
     * @param reader the {@link FieldContainer} of files to classify is read
     * from this reader; not closed by this method
     * @param writer the resulting {@link FieldContainer} containing
     * classification results is serialized to this writer; not closed by
     * this method
     * @param context a map of objects that are made available for processing;
     * the {@link IOUtils#KEY_LOCAL_CHARSET} is used to determine the character
     * set of the listed files; the {@link TextProcessor#KEY_DIRECTORY}
     * {@link File} determines the source of relative file names, if given
     * (otherwise the current working directory is used)
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    protected void doProcess(final Reader reader, final Writer writer,
                             final ContextMap context)
            throws IOException, ProcessingException {
        // read input + determine charset + directory (if any)
        final FieldContainer filesToClassify =
            FieldContainer.createFieldContainer(reader);
        final String charset = (String) context.get(IOUtils.KEY_LOCAL_CHARSET);
        final File directory = (File) context.get(KEY_DIRECTORY);

        // delegate to classifyAndTrain
        final FieldContainer result =
            classifyAndTrain(filesToClassify, directory, charset);

        // serialize results
        result.store(writer);
    }

}
