/*
 * Copyright (C) 2003-2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.feature.FeatureTransformer;
import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.util.ExternalCommand;
import de.fu_berlin.ties.util.Util;

/**
 * A proxy that provides a trainable classifier by communicating with an
 * external (non-Java) program. Program name and command line options of the
 * external classifier can be configured.
 *
 * <p>Instances of this class are thread-safe if and only if several instances
 * of the external classifier can reliably run in parallel.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.29 $, $Date: 2004/11/19 14:04:19 $, $Author: siefkes $
 */
public class ExternalClassifier extends TrainableClassifier {

    /**
     * Configuration key: the directory to run the classifier in (optional,
     * defaults to current working directory).
     */
    public static final String CONFIG_DIR = "classifier.ext.directory";

    /**
     * Configuration key: Command name + arguments to call for
     * classification (list of possible target classes will be second argument,
     * feature vector will be provided as standard input).
     */
    private static final String CONFIG_CMD_CLASSIFY = "classifier.ext.classify";

    /**
     * Configuration key: Command name + arguments to call for class
     * initialization (class to initialize will be last arg).
     */
    private static final String CONFIG_CMD_INIT = "classifier.ext.init";

    /**
     * Configuration key: Command name + arguments to call for resetting the
     * classifier by deleting the prediction model (class to reset will be last
     * arg).
     */
    private static final String CONFIG_CMD_RESET = "classifier.ext.reset";

    /**
     * Configuration key: Command name + arguments to call for
     * training (expected target class will be second argument, feature vector
     * will be provided as standard input).
     */
    private static final String CONFIG_CMD_TRAIN = "classifier.ext.train";

    /**
     * Configuration key: the suffix to append to classes for the classifier
     * (optional).
     */
    private static final String CONFIG_CLASS_SUFFIX = "classifier.ext.suffix";

    /**
     * Configuration key:  regular expression to extract the predicted class
     * (group 1) and the probability (group 2) from the classifier's standard
     * output; for all classes or at least for the best one.
     */
    private static final String CONFIG_REGEX = "classifier.ext.regex";

    /**
     * External command called for classification.
     */
    private final ExternalCommand extClassifier;

    /**
     * External command called for initialization; might be <code>null</code>.
     */
    private final ExternalCommand extInitializer;

    /**
     * External command called for resetting the prediction model; might be
     * <code>null</code>.
     */
    private final ExternalCommand extResetter;

    /**
     * External command called for training.
     */
    private final ExternalCommand extTrainer;

    /**
     * The suffix to append to classes for the classifier (might be null).
     */
    private final String classSuffix;

    /**
     * The directory to run the classifier in (if null, the current working
     * directory is used).
     */
    private final File workDir;

    /**
     * Regular expression to extract the predicted class (group 1) and the
     * probability (group 2) from the classifier's standard output.
     */
    private final Pattern predictionPattern;

    /**
     * The classifier is trained if the pR is below this value as well as
     * on errors ("thick threshold" heuristic). {@link Double#NaN} if not used.
     */
    private final double thickThresholdPR;

    /**
     * The classifier is trained if the probability is below this value as well
     * as on errors ("thick threshold" heuristic). {@link Double#NaN} if not
     * used.
     */
    private final double thickThresholdProb;

    /**
     * Whether this instance has been initialized.
     */
    private boolean initialized = false;

    /**
     * Creates a new instance based on the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @param allValidClasses the set of all valid classes
     * @throws ProcessingException if an I/O error occurs during initialization
     */
    public ExternalClassifier(final Set<String> allValidClasses)
            throws ProcessingException {
        this(allValidClasses, TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance based on the provided configuration.
     *
     * @param allValidClasses the set of all valid classes
     * @param config contains configuration properties
     * @throws ProcessingException if an I/O error occurs during initialization
     */
    public ExternalClassifier(final Set<String> allValidClasses,
            final TiesConfiguration config) throws ProcessingException {
        this(allValidClasses, FeatureTransformer.createTransformer(config),
            null, config);
    }

    /**
     * Creates a new instance based on the provided arguments.
     *
     * @param allValidClasses the set of all valid classes
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param runDirectory the directory to run the classifier in; used instead
     * of the {@linkplain #CONFIG_DIR configured directory} if not
     * <code>null</code>
     * @param config contains configuration properties
     * @throws ProcessingException if an I/O error occurs during initialization
     */
    public ExternalClassifier(final Set<String> allValidClasses,
            final FeatureTransformer trans, final File runDirectory,
            final TiesConfiguration config)
            throws ProcessingException {
        super(allValidClasses, trans, config);

        // configure instance
        final String[] cmdClassify = config.getStringArray(CONFIG_CMD_CLASSIFY);
        final String[] cmdInit = config.getStringArray(CONFIG_CMD_INIT);
        final String[] cmdReset = config.getStringArray(CONFIG_CMD_RESET);
        final String[] cmdTrain = config.getStringArray(CONFIG_CMD_TRAIN);
        predictionPattern = Pattern.compile(config.getString(CONFIG_REGEX));

        // handle optional properties
        classSuffix = config.getString(CONFIG_CLASS_SUFFIX, null);

        final String rawThresholdProb =
            config.getString("classifier.ext.threshold.prob", null);
        if (StringUtils.isNotEmpty(rawThresholdProb)) {
            // we do it like this to cope with empty strings
            thickThresholdProb = Util.asDouble(rawThresholdProb);
        } else {
            thickThresholdProb = Double.NaN;
        }

        final String rawThresholdPR =
            config.getString("classifier.ext.threshold.pR", null);
        if (StringUtils.isNotEmpty(rawThresholdPR)) {
            // we do it like this to cope with empty strings
            thickThresholdPR = Util.asDouble(rawThresholdPR);
        } else {
            thickThresholdPR = Double.NaN;
        }

        // set directory
        if (runDirectory != null) {
            workDir = runDirectory;
        } else if (config.containsKey(CONFIG_DIR)) {
            workDir = new File(config.getString(CONFIG_DIR));
        } else {
            workDir = null;
        }

        // create wrappers for external programs
        extClassifier = new ExternalCommand(cmdClassify, workDir);
        extTrainer = new ExternalCommand(cmdTrain, workDir);

        if (cmdInit.length > 0) {
            extInitializer = new ExternalCommand(cmdInit, workDir);
        } else {
            extInitializer = null;
        }

        if (cmdReset.length > 0) {
            extResetter = new ExternalCommand(cmdReset, workDir);
        } else {
            extResetter = null;
        }
    }

    /**
     * Helper method building a class name by appending the configured suffix,
     * if any.
     *
     * @param baseName the base name of the class
     * @return the complete class name
     */
    private String buildClassName(final String baseName) {
        // append suffix, if exists
        if (classSuffix != null) {
            return baseName + classSuffix;
        } else {
            return baseName;
        }
    }

    /**
     * Classifies an item that is represented by a feature vector by choosing
     * the most probable class among a set of candidate classes.
     *
     * @param features the feature vector to consider
     * @param candidateClasses an array of the classes that are allowed for
     * this item
     * @param context ignored by this implementation
     * @return the result of the classification; you can call
     * {@link PredictionDistribution#best()} to get the most probably class;
     * this classifier returns only the best prediction, so
     * {@link PredictionDistribution#size()} will be 1
     * @throws ProcessingException if an I/O error occurs during communication
     * with the external program
     */
    protected PredictionDistribution doClassify(final FeatureVector features,
            final Set candidateClasses, final ContextMap context)
    throws ProcessingException {
        if (!initialized) {
            initialize();
        }

        // build list of candidate classes as last argument
        final StringBuffer candidateBuffer = new StringBuffer();
        final Iterator candidateIter = candidateClasses.iterator();

        while (candidateIter.hasNext()) {
            candidateBuffer.append(
                buildClassName((String) candidateIter.next()));

            if (candidateIter.hasNext()) {
                // no whitespace required after the last candidate
                candidateBuffer.append(' ');
            }
        }
        final String[] furtherArg = new String[] {candidateBuffer.toString()};
        final String output;

        // run the external program on the flattened features
        try {
            output = extClassifier.execute(furtherArg, features.flatten());
        } catch (IOException ioe) {
            // wrap and rethrow exception
            throw new ProcessingException("I/O error while classifying", ioe);
        }

        // match output to regex to extract predicted class + probability + pR
        final Matcher outputMatcher = predictionPattern.matcher(output);
        final PredictionDistribution predDist = new PredictionDistribution();
        final Set<String> addedClasses = new HashSet<String>();
        Prediction pred;
        String rawPredictedClass, predictedClass;
        String rawProbability;
        double probability;
        String rawPR;

        // ensure regex matched okay
        while (outputMatcher.find()) {
/*            throw new IllegalArgumentException(
                    "No match found for extraction pattern '"
                    + predictionPattern.pattern() + "' in classifier output: '"
                    + output + "'"); */
            if (outputMatcher.groupCount() < 2) {
                throw new IllegalArgumentException("Extraction pattern '"
                        + predictionPattern.pattern()
                        + "' should match 2 or 3 subgroups but matched only "
                        + outputMatcher.groupCount()  + " in classifier output: '"
                        + output + "'");
            }
    
            // match and convert subgroups
            rawPredictedClass = outputMatcher.group(1);
    
            if ((classSuffix != null)
                    && rawPredictedClass.endsWith(classSuffix)) {
                // strip suffix from predicted class
                predictedClass = rawPredictedClass.substring(0,
                        rawPredictedClass.length() - classSuffix.length());
            } else {
                predictedClass = rawPredictedClass;
            }
    
            rawProbability = outputMatcher.group(2);
            probability = Util.asDouble(rawProbability);
    
            // pR is given in 3rd group (optional), otherwise NaN
            final double pR;
            if (outputMatcher.groupCount() >= 3) {
                rawPR = outputMatcher.group(3);
                pR = Util.asDouble(rawPR);
            } else {
                pR = Double.NaN;
            }
    
            // create and store prediction (they will be sorted automatically)
            if (!addedClasses.contains(predictedClass)) {
                // avoid adding the same class multiple times
                addedClasses.add(predictedClass);
                pred = new Prediction(predictedClass,
                        new Probability(probability, pR));
                predDist.add(pred);
            }
        }

        // check that we found at least one class
        if (predDist.size() < 1) {
            throw new IllegalArgumentException(
                    "No match found for extraction pattern '"
                    + predictionPattern.pattern() + "' in classifier output: '"
                    + output + "'");
        }

        //Util.LOG.debug(predDist + "; classifier output: " + output);

        return predDist;
    }

    /**
     * {@inheritDoc}
     */
    protected void doTrain(final FeatureVector features,
            final String targetClass, final ContextMap context)
    throws ProcessingException {
        if (!initialized) {
            initialize();
        }

        // pass target class as last argument
        final String[] furtherArg = new String[] {buildClassName(targetClass)};

        // run the external program on the flattened features
        // (output can be ignored)
        try {
            extTrainer.execute(furtherArg, features.flatten());
        } catch (IOException ioe) {
            // wrap and rethrow exception
            throw new ProcessingException("I/O error while training", ioe);
        }
    }

    /**
     * Initializes all classes.
     * @throws ProcessingException if an I/O error occurs during initialization
     */
    private void initialize() throws ProcessingException {
        if (extInitializer != null) {
            // initialize all classes
            final Iterator classIter = getAllClasses().iterator();

            while (classIter.hasNext()) {
                init((String) classIter.next());
            }
        }

        initialized = true;
    }

    /**
     * Initializes a class. This method is called once for each class.
     *
     * @param cls the class to initialize
     * @throws ProcessingException if an I/O error occurs during initialization
     */
    private void init(final String cls) throws ProcessingException {
        // pass target class as last argument
        final String[] furtherArg = new String[] {buildClassName(cls)};

        // run the external program (output can be ignored)
        try {
            extInitializer.execute(furtherArg);
        } catch (IOException ioe) {
            // wrap and rethrow exception
            throw new ProcessingException("I/O error while initializing the "
                    + cls + " class", ioe);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void reset() throws ProcessingException {
        if (extResetter != null) {
            // reset all classes
            final Iterator classIter = getAllClasses().iterator();

            while (classIter.hasNext()) {
                resetClass((String) classIter.next());
            }
        }

        // re-initialization necessary
        initialized = false;
    }

    /**
     * Resets a class.
     *
     * @param cls the class to reset
     * @throws ProcessingException if an I/O error occurs during reset
     */
    private void resetClass(final String cls) throws ProcessingException {
        // pass target class as last argument
        final String[] furtherArg = new String[] {buildClassName(cls)};

        // run the external program (output can be ignored)
        try {
            extResetter.execute(furtherArg);
        } catch (IOException ioe) {
            // wrap and rethrow exception
            throw new ProcessingException("I/O error while resetting the "
                    + cls + " class", ioe);
        }
    }

    /**
     * This implementation uses reinforcement training, if a thick threshold
     * is configured.
     * {@inheritDoc}
     */
    protected boolean shouldTrain(final String targetClass,
            final PredictionDistribution predDist, final ContextMap context) {
        final Prediction best = predDist.best();
        final Probability bestProb = best.getProbability();

        if (super.shouldTrain(targetClass, predDist, context)) {
            // normal train-on-error
            return true;
        } else if (!Double.isNaN(thickThresholdPR)
                && !Double.isNaN(bestProb.getPR())
                && bestProb.getPR() < thickThresholdPR) {
            Util.LOG.debug("Reinforcement training because pR " 
                    + bestProb.getPR() + " is below the pR threshold "
                    + thickThresholdPR);
            return true;
        } else if (!Double.isNaN(thickThresholdProb)
                && bestProb.getProb() < thickThresholdProb) {
            Util.LOG.debug("Reinforcement training because probability "
                    + best.getProbability() + " is below the threshold "
                    + thickThresholdProb);
            return true;            
        } else {
            // no need to train
            return false;
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder builder = new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("classify command", extClassifier)
            .append("train command", extTrainer)
            .append("init command", extInitializer)
            .append("reset command", extResetter)
            .append("class suffix", classSuffix)
            .append("prediction pattern", predictionPattern.pattern());

        if (!Double.isNaN(thickThresholdProb)) {
            builder.append("thick threshold prob.", thickThresholdProb);
        }
        if (!Double.isNaN(thickThresholdPR)) {
            builder.append("thick threshold pR", thickThresholdPR);
        }

        return builder.toString();
    }

}
