/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify;

import java.io.File;
import java.util.Set;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.feature.FeatureTransformer;
import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.filter.TrainableFilter;
import de.fu_berlin.ties.util.Util;

/**
 * A meta classifier combines several layers of classifiers. For each layer
 * (except the last one), there is a "judge" that decides whether or not the
 * decision of this classifier is likely to be correct. If the judge decides
 * that it is likely to be wrong, the next layer is invoked to correct the
 * decision.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/12/09 18:09:14 $, $Author: siefkes $
 */
class MetaClassifier extends TrainableClassifier {

    /**
     * Key used to store the contexts of the inner classifiers.
     */
    private static final String KEY_INNER_CONTEXTS = "inner-context";

    /**
     * Key used to store the prediction distributions returned by the inner
     * classifiers.
     */
    private static final String KEY_INNER_DISTS = "inner-dist";

    /**
     * Key used to store the contexts of the judges.
     */
    private static final String KEY_JUDGE_CONTEXTS = "judge-context";

    /**
     * Key used to store the prediction distributions returned by jthe udges.
     */
    private static final String KEY_JUDGE_DISTS = "judge-dist";

    /**
     * Key used to store the decisions made by the judges.
     */
    private static final String KEY_JUDGE_DECISIONS = "judge-decision";


    /**
     * The array of inner classifiers managed by this instance.
     */
    private final TrainableClassifier[] inner;

    /**
     * The array of judges managed by this instance.
     */
    private final TrainableClassifier[] judges;

    /**
     * Creates a new instance.
     *
     * @param allValidClasses the set of all valid classes; the first member
     * of this set is considered as the "background" class, all further members
     * are considered as "foreground" classes
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param runDirectory optional run directory passed to inner classifiers
     * of the {@link ExternalClassifier} type
     * @param innerSpec the specification used to initialize the inner
     * classifiers, passed to the
     * {@link TrainableClassifier#createClassifier(Set, File,
     * FeatureTransformer, String[], TiesConfiguration)} factory method
     * @param conf used to configure this instance and the inner classifiers
     * @throws ProcessingException if an error occurred while creating this
     * classifier or one of the wrapped classifiers
     */
    public MetaClassifier(Set<String> allValidClasses,
            FeatureTransformer trans, final File runDirectory,
            final String[] innerSpec, TiesConfiguration conf)
    throws ProcessingException {
        this(allValidClasses, trans, runDirectory, innerSpec,
                conf.getInt("classifier.meta.layers"),
                conf.getStringArray("classifier.meta.judge"), conf);
    }

    /**
     * Creates a new instance.
     *
     * @param allValidClasses the set of all valid classes; the first member
     * of this set is considered as the "background" class, all further members
     * are considered as "foreground" classes
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param runDirectory optional run directory passed to inner classifiers
     * of the {@link ExternalClassifier} type
     * @param innerSpec the specification used to initialize the inner
     * classifiers, passed to the
     * {@link TrainableClassifier#createClassifier(Set, File,
     * FeatureTransformer, String[], TiesConfiguration)} factory method
     * @param layers the number of layers to use, must be at least one
     * @param judgeSpec the specification used to initialize the judges,
     * passed to the {@link TrainableClassifier#createClassifier(Set, File,
     * FeatureTransformer, String[], TiesConfiguration)} factory method
     * @param conf used to configure this instance as well as the inner
     * classifiers and judges
     * @throws ProcessingException if an error occurred while creating this
     * classifier or one of the wrapped classifiers
     */
    public MetaClassifier(Set<String> allValidClasses,
            FeatureTransformer trans, final File runDirectory,
            final String[] innerSpec, final int layers,
            final String[] judgeSpec, TiesConfiguration conf)
    throws ProcessingException {
        super(allValidClasses, trans, conf);

        if (layers < 1) {
            throw new IllegalArgumentException(
                "MetaClassifier requires at least 1 layer instead of "
                    + layers);
        }

        // Init classifiers for each layer
        inner = new TrainableClassifier[layers];
        for (int i = 0; i < inner.length; i++) {
            // Transformer is set to null because features shouldn't be
            // transformed twice
            inner[i] = TrainableClassifier.createClassifier(allValidClasses,
                runDirectory, null, innerSpec, conf);
            
        }

        // Init judges for each layer except the very last one
        judges = new TrainableClassifier[layers - 1];
        for (int i = 0; i < judges.length; i++) {
            // Judges classify between "true" and "false".
            // Transformer is set to null, as above
            judges[i] = TrainableClassifier.createClassifier(
                    TrainableFilter.BOOLEAN_CLASSES, runDirectory, null,
                    judgeSpec, conf);
            
        }
    }

    /**
     * {@inheritDoc}
     */
    protected PredictionDistribution doClassify(FeatureVector features,
            Set candidateClasses, ContextMap context)
            throws ProcessingException {
        // Used to populate the context
        PredictionDistribution[] innerDists =
            new PredictionDistribution[inner.length];
        ContextMap[] innerContexts = new ContextMap[inner.length];
        PredictionDistribution[] judgeDists =
            new PredictionDistribution[judges.length];
        ContextMap[] judgeContexts = new ContextMap[judges.length];
        boolean[] judgements = new boolean[judges.length];

        PredictionDistribution innerDist = null;
        PredictionDistribution judgeDist;
        ContextMap innerContext, judgeContext;
        boolean judgement = false;
        int i = 0;

        // invoke inner classifier until judge trusts a decision to be correct
        // or all layers are exhausted
        while ((i < inner.length) || !judgement) {
            if (i > 0) {
                Util.LOG.debug("Invoking layer " + i
                        + " classifier since decision of previous layer was "
                        + "judged to be incorrect");
            }

            // invoke i-th inner classifier
            innerContext = new ContextMap();
            innerDist = inner[i].doClassify(features, candidateClasses,
                    innerContext);
            innerContexts[i] = innerContext;
            innerDists[i] = innerDist;

            // call judge if available (not for last layer)
            if (i < judges.length) {
                judgeContext = new ContextMap();
                judgeDist = judges[i].doClassify(features,
                        judges[i].getAllClasses(), judgeContext);
                judgeContexts[i] = judgeContext;
                judgeDists[i] = judgeDist;

                // convert best class to boolean
                judgement = Util.asBoolean(judgeDist.best().getType());
                judgements[i] = judgement;
            }

            i++;
        }

        // store judgements, distributions, and contexts
        context.put(KEY_INNER_CONTEXTS, innerContexts);
        context.put(KEY_INNER_DISTS, innerDists);
        context.put(KEY_JUDGE_CONTEXTS, judgeContexts);
        context.put(KEY_JUDGE_DISTS, judgeDists);
        context.put(KEY_JUDGE_DECISIONS, judgements);

        // return decision of last invoked layer
        return innerDist;
    }

    /**
     * {@inheritDoc}
     */
    protected void doTrain(FeatureVector features, String targetClass,
            ContextMap context) throws ProcessingException {
        // TODO throw exception + add note in header similar to Winnow
    }

    /**
     * {@inheritDoc}
     */
    public void reset() throws ProcessingException {
        // TODO delegate to all layers and judges
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("inner classifiers", ArrayUtils.toString(inner))
            .append("judges", ArrayUtils.toString(judges))
            .toString();
    }

    /**
     * {@inheritDoc}
     */
    protected boolean trainOnErrorHook(final PredictionDistribution predDist,
            final FeatureVector features, final String targetClass,
            final Set candidateClasses, final ContextMap context)
    throws ProcessingException {
        // TODO implement, delegating to layers and judges as appropriate
        return false;
    }

}
