/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.feature.FeatureTransformer;
import de.fu_berlin.ties.classify.feature.FeatureVector;

/**
 * This classifier converts an multi-class classification task into a several
 * binary (two-class) classification task. It wraps several instances of
 * another classifier that are used to perform the binary classifications and
 * combines their results.
 *
 * <p>The first class from the set of classes passed to the constructor is
 * considered as the "background" class, while all further members are
 * considered as "foreground" classes.
 *
 * <p>Instances of this class are thread-safe if and only if instances of the
 * wrapped classifier are.
 *
 * <p><em><strong>WARNING:</strong> The current implementation does
 * <strong>not</strong> query the {@link
 * de.fu_berlin.ties.classify.TrainableClassifier#shouldTrain(String,
 * PredictionDistribution, ContextMap) shouldTrain} method of inner classifiers.
 * Because of this, classifiers overwriting <code>shouldTrain</code> might not
 * work correctly within this classifier.</em>
 *
 * @author Christian Siefkes
 * @version $Revision: 1.19 $, $Date: 2004/12/09 18:09:14 $, $Author: siefkes $
 */
public class MultiBinaryClassifier extends TrainableClassifier {

    /**
     * Key prefix used to store the contexts of inner classifiers.
     */
    private static final String PREFIX_CONTEXT = "context-";

    /**
     * Key prefix used to store the prediction distributions returned by inner
     * classifiers.
     */
    private static final String PREFIX_DIST = "dist-";

    /**
     * The "background" class of this classifier.
     */
    private final String backgroundClass;

    /**
     * Maps from the names of the foreground classes to the binary classifiers
     * used to decide between this foreground class and the background class.
     */
    private final Map<String, TrainableClassifier> innerClassifiers =
        new HashMap<String, TrainableClassifier>();

    /**
     * Creates a new instance.
     *
     * @param allValidClasses the set of all valid classes; the first member
     * of this set is considered as the "background" class, all further members
     * are considered as "foreground" classes
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param runDirectory optional run directory passed to inner classifiers
     * of the {@link ExternalClassifier} type
     * @param innerSpec the specification used to initialize the inner
     * classifiers, passed to the
     * {@link TrainableClassifier#createClassifier(Set, File,
     * FeatureTransformer, String[], TiesConfiguration)} factory method
     * @param conf used to configure this instance and the inner classifiers
     * @throws IllegalArgumentException if there are fewer than three classes
     * (in which case you should use the inner classifier directly since there
     * is no need to wrap several instances)
     * @throws ProcessingException if an error occurred while creating this
     * classifier or one of the wrapped classifiers
     */
    public MultiBinaryClassifier(final Set<String> allValidClasses,
                                 final FeatureTransformer trans,
                                 final File runDirectory,
                                 final String[] innerSpec,
                                 final TiesConfiguration conf)
    throws IllegalArgumentException, ProcessingException {
        super(allValidClasses, trans, conf);

        if (allValidClasses.size() < 3) {
            throw new IllegalArgumentException(
                "MultiBinaryClassifier requires at least 3 classes instead of "
                + allValidClasses.size());
        }

        // store background class (we checked that there are enough classes)
        final Iterator<String> classIter= allValidClasses.iterator();
        backgroundClass = classIter.next();

        // create inner classifiers via factory method
        String foregroundClass;
        TrainableClassifier innerClassifier;
        Set<String> innerSet;

        while (classIter.hasNext()) {
            // each inner classifier decides between a foreground class and the
            // background class
            foregroundClass = classIter.next();
            innerSet = createBinarySet(foregroundClass);

            // transformer is set to null because features shouldn't be
            // transformed twice
            innerClassifier = TrainableClassifier.createClassifier(innerSet,
                runDirectory, null, innerSpec, conf);
            innerClassifiers.put(foregroundClass, innerClassifier);
        }
    }

    /**
     * Helper method that creates a set containing the two classes of a
     * binary classifier.
     *
     * @param foregroundClass the "foreground" class to use
     * @return a set containing the {@linkplain #getBackgroundClass()
     * "background" class} and the specified <code>foregroundClass</code>;
     * this implementation returns the two classes in alphabetic order
     */
    protected Set<String> createBinarySet(final String foregroundClass) {
        final SortedSet<String> result = new TreeSet<String>();
        result.add(backgroundClass);
        result.add(foregroundClass);
        return result;
    }

    /**
     * {@inheritDoc}
     * This implementation combines the predictions for the foreground
     * of all involved inner classifiers.
     *
     * <p>If the {@linkplain #getBackgroundClass() background class} is part of
     * the <code>candidateClasses</code>, the classifier whose background
     * probability is closest to 0.5 determines the probability of the
     * background class. In this way all classes will be sorted the right way
     * (foreground classes with a higher than 0.5 before the background class,
     * those with a lower probability after it). pR values are <em>not</em>
     * considered for this purpose, so you should be careful when combination
     * classifiers that mainly rely on the pR value in a multi-binary setup.
     *
     * <p><em>The probability estimates returned by each classifier are used
     * "as is", so the result will <strong>not</strong> be a real probability
     * distribution because sum of all probabilities will be more than 1.
     * If you want to work on a real probability distribution you have
     * normalize it yourself.</em>
     */
    protected PredictionDistribution doClassify(final FeatureVector features,
                                                final Set candidateClasses,
                                                final ContextMap context)
    throws ProcessingException {
        final List<Prediction> unnormalizedPreds =
            new ArrayList<Prediction>(candidateClasses.size());
        final Iterator classIter = candidateClasses.iterator();
        String currentClass;
        TrainableClassifier innerClassifier;
        PredictionDistribution currentDist;
        Prediction currentPred;
        ContextMap currentContext;
        Iterator predIter;
//        double summedProb = 0.0;

        // for determining which prediction to use for the background class
        Prediction backgroundMostNearToFifty = null;
        double lowestBackgroundDist = Double.MAX_VALUE;
        double currentBackgroundDist;

        // delegate to doClassify methods of all foreground candidates
        while (classIter.hasNext()) {
            currentClass = (String) classIter.next();

            if (!backgroundClass.equals(currentClass)) {
                // classify all foreground candidates
                currentContext = new ContextMap();
                innerClassifier = innerClassifiers.get(currentClass);
                currentDist = innerClassifier.doClassify(features,
                    innerClassifier.getAllClasses(), currentContext);

                // store prediction distribution and context map of this
                // classifier in my own context
                context.put(PREFIX_CONTEXT + currentClass, currentContext);
                context.put(PREFIX_DIST + currentClass, currentDist);

                predIter = currentDist.iterator();

                // iterate prediction distribution of this classifier
                while (predIter.hasNext()) {
                    currentPred = (Prediction) predIter.next();

                    // prediction for background class
                    if (backgroundClass.equals(currentPred.getType())) {
                        // calculate distance to 0.5 (absolute value)
                        currentBackgroundDist = Math.abs(
                                currentPred.getProbability().getProb()- 0.5);

                        if (currentBackgroundDist < lowestBackgroundDist) {
/*                            if (backgroundMostNearToFifty != null) {
                                Util.LOG.debug("Using background prob. "
                                    + currentPred + " instead of "
                                    + backgroundMostNearToFifty
                                    + " because it's closer to 0.5");
                            } */

                            // prob. is nearer to 0.5 than other background
                            // predictions
                            backgroundMostNearToFifty = currentPred;
                            lowestBackgroundDist = currentBackgroundDist;
                        }
                    } else {
                        // prediction for a foreground class
                        unnormalizedPreds.add(currentPred);
//                        summedProb += currentPred.getProbability();
                    }
                }
            }
        }

        if ((backgroundMostNearToFifty != null)
                && candidateClasses.contains(backgroundClass)) {
            // background class is a candidate:
            // add it using the prob. that is most near to 0.5
            unnormalizedPreds.add(backgroundMostNearToFifty);
//            summedProb += backgroundMostNearToFifty.getProbability();
        }

        final PredictionDistribution result = new PredictionDistribution();
        predIter = unnormalizedPreds.iterator();

        // build full distribution
        while (predIter.hasNext()) {
            currentPred = (Prediction) predIter.next();
            result.add(currentPred);

/*            // normalize probabilities so they sum to 1
            result.add(new Prediction(currentPred.getType(),
                currentPred.getSource(),
                currentPred.getProbability() / summedProb,
                currentPred.getPR(), currentPred.getEvalStatus())); */
        }

        //Util.LOG.debug("Combined probability distribution: "
        //    + result.toString());
        //Util.LOG.debug("Current context: " + context);

        return result;
    }

    /**
     * {@inheritDoc}
     */
    protected void doTrain(final FeatureVector features,
                           final String targetClass,
                           final ContextMap context)
    throws ProcessingException {
        final Iterator classIter = innerClassifiers.keySet().iterator();
        String currentClass;
        TrainableClassifier classifier;
        String classToTrain;
        String contextKey;
        ContextMap currentContext;

        // delegate to doTrain methods of all classifiers
        while (classIter.hasNext()) {
            currentClass = (String) classIter.next();
            classifier = innerClassifiers.get(currentClass);

            if (currentClass.equals(targetClass)) {
                // classifier for the target class
                classToTrain = targetClass;
            } else {
                // other classifier: train background class
                classToTrain = backgroundClass;
            }

            // retrieve original context map, if stored
            contextKey = PREFIX_CONTEXT + currentClass;
            if (context.containsKey(contextKey)) {
                currentContext = (ContextMap) context.get(contextKey);
            } else {
                currentContext = new ContextMap();
            }

            classifier.doTrain(features, classToTrain, currentContext);
        }
    }

    /**
     * Returns the "background" class of this classifier.
     * @return the value of the attribute
     */
    public String getBackgroundClass() {
        return backgroundClass;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("background class", backgroundClass)
            .append("inner classifiers", innerClassifiers.values())
            .toString();
    }

    /**
     * {@inheritDoc}
     */
    protected boolean trainOnErrorHook(final PredictionDistribution predDist,
            final FeatureVector features, final String targetClass,
            final Set candidateClasses, final ContextMap context)
    throws ProcessingException {
        boolean result = false;
        final Iterator predIter = predDist.iterator();
        Prediction currentPred;
        String currentClass;
        TrainableClassifier classifier;
        ContextMap currentContext;
        PredictionDistribution currentDist;
        String classToTrain;

        // delegate to all involved classifiers, returning ORing of results
        // (i.e. return true if at least one of them returns true)
        while (predIter.hasNext()) {
            currentPred = (Prediction) predIter.next();
            currentClass = currentPred.getType();

            if (!backgroundClass.equals(currentClass)) {
                classifier = innerClassifiers.get(currentClass);

                // retrieve original context map + prediction distribution
                currentContext = (ContextMap)
                    context.get(PREFIX_CONTEXT + currentClass);
                currentDist = (PredictionDistribution)
                    context.get(PREFIX_DIST + currentClass);

                if (currentClass.equals(targetClass)) {
                    // classifier for the target class
                    classToTrain = targetClass;
                } else {
                    // other classifier: train background class
                    classToTrain = backgroundClass;
                }

                result = classifier.trainOnErrorHook(currentDist, features,
                    classToTrain, classifier.getAllClasses(), currentContext)
                    || result;
            }
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public void reset() throws ProcessingException {
        final Iterator innerIter = innerClassifiers.values().iterator();
        TrainableClassifier classifier;

        // delegate to reset methods of all inner classifiers
        while (innerIter.hasNext()) {
            classifier = (TrainableClassifier) innerIter.next();
            classifier.reset();
        }
    }

}
