/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.feature.FeatureTransformer;
import de.fu_berlin.ties.classify.feature.FeatureVector;

/**
 * This classifier converts an multi-class classification task into a several
 * binary (two-class) classification task. It wraps several instances of
 * another classifier that are used to perform the binary classifications and
 * combines their results.
 *
 * <p>When used with <em>N</em> classes, this classifier created and wrapped
 * <em>N</em> inner binary classifiers. Each inner classifier classifiers
 * between members (positive instances) and non-members (negative instances) of
 * a specified class. The inner classifier returning the highest
 * {@linkplain de.fu_berlin.ties.classify.Prediction#getProbability()
 * probability} for its (positive) class wins, even if the negative class of
 * that classifier would be still higher. This is the classical
 * "one-against-the-rest" (or "one-against-all") setup.
 *
 * <p>This classifier differs from
 * {@link de.fu_berlin.ties.classify.MultiBinaryClassifier} in not using a
 * "background class" and using <em>N</em> inner classifiers, while
 * <code>MultiBinaryClassifier</code> uses <em>N-1</em>.
 *
 * <p>Instances of this class are thread-safe if and only if instances of the
 * wrapped classifier are.
 *
 * <p><em><strong>WARNING:</strong> The current implementation does
 * <strong>not</strong> query the {@link
 * de.fu_berlin.ties.classify.TrainableClassifier#shouldTrain(String,
 * PredictionDistribution, ContextMap) shouldTrain} method of inner classifiers.
 * Because of this, classifiers overwriting <code>shouldTrain</code> might not
 * work correctly within this classifier.</em>
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/11/19 14:04:19 $, $Author: siefkes $
 */
public class OneAgainstTheRestClassifier extends TrainableClassifier {

    /**
     * Key prefix used to store the contexts of inner classifiers.
     */
    private static final String PREFIX_CONTEXT = "context-";

    /**
     * Key prefix used to store the prediction distributions returned by inner
     * classifiers.
     */
    private static final String PREFIX_DIST = "dist-";

    /**
     * The suffix appended to negative classes.
     */
    private static final String NEG_SUFFIX = "-neg";

    /**
     * Maps from the names of the classes to the binary classifiers
     * used to decide between positive and negative instances of this class.
     */
    private final Map<String, TrainableClassifier> innerClassifiers =
        new HashMap<String, TrainableClassifier>();

    /**
     * Creates a new instance.
     *
     * @param allValidClasses the set of all valid classes
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param runDirectory optional run directory passed to inner classifiers
     * of the {@link ExternalClassifier} type
     * @param innerSpec the specification used to initialize the inner
     * classifiers, passed to the
     * {@link TrainableClassifier#createClassifier(Set, File,
     * FeatureTransformer, String[], TiesConfiguration)} factory method
     * @param conf used to configure this instance and the inner classifiers
     * @throws IllegalArgumentException if there are fewer than three classes
     * (in which case you should use the inner classifier directly since there
     * is no need to wrap several instances)
     * @throws ProcessingException if an error occurred while creating this
     * classifier or one of the wrapped classifiers
     */
    public OneAgainstTheRestClassifier(final Set<String> allValidClasses,
                                 final FeatureTransformer trans,
                                 final File runDirectory,
                                 final String[] innerSpec,
                                 final TiesConfiguration conf)
    throws IllegalArgumentException, ProcessingException {
        super(allValidClasses, trans, conf);

        if (allValidClasses.size() < 3) {
            throw new IllegalArgumentException("OneAgainstTheRestClassifier "
                    + "requires at least 3 classes instead of "
                    + allValidClasses.size());
        }

        // create inner classifiers via factory method
        final Iterator<String> classIter= allValidClasses.iterator();
        String baseClass;
        TrainableClassifier innerClassifier;
        Set<String> innerSet;

        while (classIter.hasNext()) {
            // each inner classifier decides between a positive class and a
            // negative class
            baseClass = classIter.next();
            innerSet = createBinarySet(baseClass);

            // transformer is set to null because features shouldn't be
            // transformed twice
            innerClassifier = TrainableClassifier.createClassifier(innerSet,
                runDirectory, null, innerSpec, conf);
            innerClassifiers.put(baseClass, innerClassifier);
        }
    }

    /**
     * Helper method that creates a set containing the two classes of a
     * binary classifier.
     *
     * @param baseClass the base class to use
     * @return a set containing a positive and a negative class derived from the
     * base class; this implementation returns the two classes in alphabetic
     * order which means that the negative class comes first
     */
    protected Set<String> createBinarySet(final String baseClass) {
        final SortedSet<String> result = new TreeSet<String>();
        result.add(baseClass);
        result.add(baseClass + NEG_SUFFIX);
        return result;
    }

    /**
     * {@inheritDoc}
     * This implementation combines the predictions for the positive class
     * of all involved inner classifiers.
     *
     * <p><em>The probability estimates returned by each classifier are used
     * "as is", so the result will <strong>not</strong> be a real probability
     * distribution because sum of all probabilities will be more than 1.
     * If you want to work on a real probability distribution you have
     * normalize it yourself.</em>
     */
    protected PredictionDistribution doClassify(final FeatureVector features,
                                                final Set candidateClasses,
                                                final ContextMap context)
    throws ProcessingException {
        final List<Prediction> unnormalizedPreds =
            new ArrayList<Prediction>(candidateClasses.size());
        final Iterator classIter = candidateClasses.iterator();
        String currentClass;
        TrainableClassifier innerClassifier;
        PredictionDistribution currentDist;
        Prediction currentPred;
        String predClass;
        ContextMap currentContext;
        Iterator predIter;
//        double summedProb = 0.0;

        // delegate to doClassify methods of all candidat classes
        while (classIter.hasNext()) {
            currentClass = (String) classIter.next();

            // classify all candidat classeses
            currentContext = new ContextMap();
            innerClassifier = innerClassifiers.get(currentClass);
            currentDist = innerClassifier.doClassify(features,
                innerClassifier.getAllClasses(), currentContext);

            // store prediction distribution and context map of this
            // classifier in my own context
            context.put(PREFIX_CONTEXT + currentClass, currentContext);
            context.put(PREFIX_DIST + currentClass, currentDist);

            predIter = currentDist.iterator();

            // iterate prediction distribution of this classifier
            while (predIter.hasNext()) {
                currentPred = (Prediction) predIter.next();
                predClass = currentPred.getType();

                if (predClass.equals(currentClass)) {
                    // use predictions of positive classes only
                    unnormalizedPreds.add(currentPred);
//                  summedProb += currentPred.getProbability();                    
                } else if (!predClass.equals(currentClass + NEG_SUFFIX)) {
                    // just to make sure that everything is as it should be
                    throw new RuntimeException("Implementation error: "
                            + "unexpected class name " + predClass + " used by "
                            + currentClass + " classifier");
                }
            }
        }

        final PredictionDistribution result = new PredictionDistribution();
        predIter = unnormalizedPreds.iterator();

        // build full distribution for positive classes
        while (predIter.hasNext()) {
            currentPred = (Prediction) predIter.next();
            result.add(currentPred);

/*            // normalize probabilities so they sum to 1
            result.add(new Prediction(currentPred.getType(),
                currentPred.getSource(),
                currentPred.getProbability() / summedProb,
                currentPred.getPR(), currentPred.getEvalStatus())); */
        }

        //Util.LOG.debug("Combined probability distribution: "
        //        + result.toString());
        //Util.LOG.debug("Current context: " + context);

        return result;
    }

    /**
     * {@inheritDoc}
     */
    protected void doTrain(final FeatureVector features,
                           final String targetClass,
                           final ContextMap context)
    throws ProcessingException {
        final Iterator classIter = innerClassifiers.keySet().iterator();
        String currentClass;
        TrainableClassifier classifier;
        String classToTrain;
        String contextKey;
        ContextMap currentContext;

        // delegate to doTrain methods of all classifiers
        while (classIter.hasNext()) {
            currentClass = (String) classIter.next();
            classifier = innerClassifiers.get(currentClass);

            if (currentClass.equals(targetClass)) {
                // classifier for the target class
                classToTrain = targetClass;
            } else {
                // other classifier: train negative class
                classToTrain = currentClass + NEG_SUFFIX;
            }

            // retrieve original context map, if stored
            contextKey = PREFIX_CONTEXT + currentClass;
            if (context.containsKey(contextKey)) {
                currentContext = (ContextMap) context.get(contextKey);
            } else {
                currentContext = new ContextMap();
            }

            classifier.doTrain(features, classToTrain, currentContext);
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("inner classifiers", innerClassifiers.values())
            .toString();
    }

    /**
     * {@inheritDoc}
     */
    protected boolean trainOnErrorHook(final PredictionDistribution predDist,
            final FeatureVector features, final String targetClass,
            final Set candidateClasses, final ContextMap context)
    throws ProcessingException {
        boolean result = false;
        final Iterator predIter = predDist.iterator();
        Prediction currentPred;
        String currentClass;
        TrainableClassifier classifier;
        ContextMap currentContext;
        PredictionDistribution currentDist;
        String classToTrain;

        // delegate to all involved classifiers, returning ORing of results
        // (i.e. return true if at least one of them returns true)
        while (predIter.hasNext()) {
            currentPred = (Prediction) predIter.next();
            currentClass = currentPred.getType();
            classifier = innerClassifiers.get(currentClass);

            // retrieve original context map + prediction distribution
            currentContext = (ContextMap)
                context.get(PREFIX_CONTEXT + currentClass);
            currentDist = (PredictionDistribution)
                context.get(PREFIX_DIST + currentClass);

            if (currentClass.equals(targetClass)) {
                // classifier for the target class
                classToTrain = targetClass;
            } else {
                // other classifier: train negative class
                classToTrain = currentClass + NEG_SUFFIX;
            }

            result = classifier.trainOnErrorHook(currentDist, features,
                classToTrain, classifier.getAllClasses(), currentContext)
                || result;
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public void reset() throws ProcessingException {
        final Iterator innerIter = innerClassifiers.values().iterator();
        TrainableClassifier classifier;

        // delegate to reset methods of all inner classifiers
        while (innerIter.hasNext()) {
            classifier = (TrainableClassifier) innerIter.next();
            classifier.reset();
        }
    }

}
