/*
 * Copyright (C) 2003-2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify;

import java.util.Iterator;
import java.util.LinkedList;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

import de.fu_berlin.ties.eval.EvalStatus;
import de.fu_berlin.ties.io.BaseStorable;
import de.fu_berlin.ties.io.FieldMap;
import de.fu_berlin.ties.util.Util;

/**
 * A prediction, wrapping the predicted class and the probability of the
 * prediction.
 *
 * <p>Instances of this class are not thread-safe and must be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.22 $, $Date: 2004/11/25 13:35:55 $, $Author: siefkes $
 */
public class Prediction extends BaseStorable {

    /**
     * Serialization key for the predicted class.
     */
    public static final String KEY_TYPE = "Type";

    /**
     * Serialization key for the source identifier.
     */
    public static final String KEY_SOURCE = "Source";

    /**
     * Serialization key for the probability.
     */
    public static final String KEY_PROB = "Probability";

    /**
     * Serialization key for the pR.
     */
    public static final String KEY_PR = "pR";

    /**
     * The predicted type.
     */
    private final String type;

    /**
     * The probabilities combinied of the prediction.
     */
    private LinkedList<Probability> probs = new LinkedList<Probability>();

    /**
     * The {@linkplain EvalStatus evaluation status} of this instance.
     */
    private EvalStatus evalStatus;

    /**
     * An optional identifier of the source of this prediction (e.g., the file
     * name).
     */
    private String source;

    /**
     * Creates a new instance from a field map, fulfilling the
     * {@link de.fu_berlin.ties.io.Storable} contract.
     *
     * @param fieldMap map containing the serialized fields
     */
    public Prediction(final FieldMap fieldMap) {
        this(Util.asString(fieldMap.get(KEY_TYPE)),
            Util.asString(fieldMap.get(KEY_SOURCE)),
            new Probability(
                    // this will set -1 (confirmed) if the value is missing
                    Util.asDouble(fieldMap.get(KEY_PROB)),
                    // pR default to not-a-number
                    fieldMap.containsKey(KEY_PR)
                        ? Util.asDouble(fieldMap.get(KEY_PR)) : Double.NaN),
            // this will use UNKNOWN if the field is missing or invalid
            EvalStatus.parse(Util.asString(fieldMap.get(
                    EvalStatus.KEY_EVAL_STATUS)), true));
    }

    /**
     * Creates a new instance, setting the evaluation status to
     * {@link EvalStatus#UNKNOWN}.
     *
     * @param predicted the predicted type
     * @param prob the probability of the prediction
     */
    public Prediction(final String predicted, final Probability prob) {
        this(predicted, prob, EvalStatus.UNKNOWN);
    }

    /**
     * Creates a new instance.
     *
     * @param predicted the predicted type
     * @param prob the probability of the prediction
     * @param status the {@linkplain EvalStatus evaluation status} of this
     * instance
     */
    public Prediction(final String predicted, final Probability prob,
            final EvalStatus status) {
        this(predicted, null, prob, status);
    }

    /**
     * Creates a new instance.
     *
     * @param predicted the predicted type
     * @param sourceID an identifier of the source of this prediction (e.g., the
     * file name), might be <code>null</code>
     * @param prob the probability of the prediction
     * @param status the {@linkplain EvalStatus evaluation status} of this
     * instance
     */
    public Prediction(final String predicted, final String sourceID,
            final Probability prob, final EvalStatus status) {
        super();
        type = predicted;
        source = sourceID;
        probs.add(prob);
        evalStatus = status;
    }

    /**
     * Adds a new probability for the prediction. The actual probabilities
     * are averaged over the combined probabilities.
     *
     * @param prob the new probability; might be <code>null</code>
     * if the overall probability of the extraction should not be changed
     * @param atEnd whether to add the new probability at the end or at the
     * start
     */
    protected void addProb(final Probability prob, final boolean atEnd) {
        if (atEnd) {
            probs.addLast(prob);
        } else {
            probs.addFirst(prob);
        }
    }

    /**
     * Indicates whether some other object is "equal to" this one, fulfulling
     * the {@link Object#equals(java.lang.Object)} contract.
     * {@linkplain #getEvalStatus() Evaluation status} and
     * {@linkplain #getProbability() probability}
     * are ignored when checking equality, only {@linkplain #getType() type}
     * and {@linkplain #getSource() source} are compared.
     *
     * @param obj the reference object with which to compare
     * @return <code>true</code> iff the specified object is a
     * {@link Prediction} equal to this instance
     */
    public boolean equals(final Object obj) {
        if (obj == this) {
            return true;
        } else if ((obj != null) && (getClass().equals(obj.getClass()))) {
            // used getClass instead of instanceof because otherwise subclasses
            // with additional fields would break the contract
            final Prediction other = (Prediction) obj;
            return new EqualsBuilder()
                .append(getType(), other.getType())
                .append(getSource(), other.getSource())
                .isEquals();
        } else {
            return false;
        }
    }

    /**
     * Returns the {@linkplain EvalStatus evaluation status} of this instance.
     * @return the value of the attribute
     */
    public EvalStatus getEvalStatus() {
        return evalStatus;
    }

    /**
     * Calculates and returns the average probability.
     *
     * @return the average probability
     */
    public Probability getProbability() {
        if (probs.size() == 1) {
            return probs.getFirst();
        } else {
            // calculate average
            double summedProb = 0.0;
            double summedPR = 0.0;
            int nonNullProbs = 0;
            Probability currentProb;
            final Iterator<Probability> probIter = probs.iterator();

            while (probIter.hasNext()) {
                currentProb = probIter.next();

                // jump over probabilities that are null
                if (currentProb != null) {
                    nonNullProbs++;

                    // using logarithm to multiply prob.s instead of adding
                    summedProb += Math.log(currentProb.getProb());

                     if (!Double.isNaN(summedPR)) {
                        if (Double.isNaN(currentProb.getPR())) {
                            // set PR NaN if there are one of more NaN values
                            summedPR = Double.NaN;
                        } else {
                            summedPR += currentProb.getPR();
                        }
                    }
                }
            } // while

            // using exp. to multiply prob.s instead of adding (undo log)
            final double averageProb = Math.exp(summedProb / nonNullProbs);

            if (Double.isNaN(summedPR)) {
                return new Probability(averageProb);
            } else {
                return new Probability(averageProb, summedPR / nonNullProbs);
            }
        }
    }

    /**
     * Returns the identifier of the source of this prediction (e.g., the file
     * name).
     *
     * @return the value of the attribute; or <code>null</code> if none
     * specified
     */
    public String getSource() {
        return source;
    }

    /**
     * Returns the predicted type (class).
     * @return the predicted type (class)
     */
    public String getType() {
        return type;
    }

    /**
     * Returns a hash code value for this object, fulfulling the
     * {@link Object#hashCode()} contract.
     * @return a hash code value for this object
     */
    public int hashCode() {
        // you pick two hard-coded, randomly chosen, non-zero, odd numbers
        // (preferably primes); ideally different for each class
        return new HashCodeBuilder(3, 5)
            .append(getType())
            .append(getSource())
            .toHashCode();
    }

    /**
     * Returns the number of probabilities combined in this prediction.
     *
     * @return the number of probabilities
     */
    protected int probCount() {
        return probs.size();
    }

    /**
     * Deletes one of the probabilities used for the prediction. At least one
     * probability must always remain, i.e. {@link #probCount()} must be 2 or
     * more prior to calling this method.
     *
     * @param atEnd whether to delete the first or the last probability
     * @return the removed probability
     * @throws IllegalStateException if there is only one probability left
     */
    protected Probability removeProb(final boolean atEnd)
    throws IllegalStateException {
        // check that at least one prob will remain
        if (probs.size() < 2) {
            throw new IllegalStateException(
                    "Cannot delete the last remaining probabity");
        }

        if (atEnd) {
            return probs.removeLast();
        } else {
            return probs.removeFirst();
        }
    }

    /**
     * Modifies the {@linkplain EvalStatus evaluation status} of this instance.
     *
     * @param newEvalStatus the new value of the attribute
     */
    public void setEvalStatus(final EvalStatus newEvalStatus) {
        evalStatus = newEvalStatus;
    }

    /**
     * Sets the identifier of the source of this prediction (e.g., the file
     * name).
     *
     * @param newSource the new value of the attribute; or <code>null</code> to
     * delete the current value
     */
    public void setSource(final String newSource) {
        source = newSource;
    }

    /**
     * Stores all relevant fields of this object in a field map for
     * serialization. An equivalent object can be created by calling
     * {@link de.fu_berlin.ties.io.FieldMap#createObject(Class)} on the created
     * field map.
     *
     * @return the created field map
     */
    public FieldMap storeFields() {
        final FieldMap result = new FieldMap();
        result.put(KEY_TYPE, getType());

        // omit source if null
        if (source != null) {
            result.put(KEY_SOURCE, getSource());
        }

        // omit probability if negative, i.e. for confirmed values/answer keys
        final Probability prob = getProbability();
        if (prob.getProb() >= 0) {
            result.put(KEY_PROB, new Double(prob.getProb()));
        }

        // omit pR if not-a-number, i.e. unknown
        final double pR = prob.getPR();
        if (!Double.isNaN(pR)) {
            result.put(KEY_PR, new Double(pR));
        }

        // omit evaluation status if unknown
        if (evalStatus != EvalStatus.UNKNOWN) {
            result.put(EvalStatus.KEY_EVAL_STATUS, getEvalStatus());
        }
        return result;
    }

}
