/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify;

import java.util.Comparator;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.classify.winnow.WinnowPrediction;

/**
 * A comparison function that compares
 * {@link de.fu_berlin.ties.classify.Prediction}s based on their
 * {@linkplain de.fu_berlin.ties.classify.Prediction#getProbability()
 * probabilities}.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.6 $, $Date: 2004/11/19 14:04:19 $, $Author: siefkes $
 */
public class PredictionComparator implements Comparator<Prediction> {

    /**
     * Creates a new instance.
     */
    public PredictionComparator() {
        super();
    }

    /**
     * Compares two {@link Prediction}s. This method checks whether the
     * {@linkplain Probability#getPR() pR} values of both predictions are given
     * and compares them if they are. Otherwise the
     * {@linkplain Probability#getProb() probabilities} are compared.
     * In case of a tie the {@linkplain WinnowPrediction#getRawScore() raw
     * scores} are compared if both are {@link WinnowPrediction}s.
     * Otherwise the {@linkplain Prediction#getType() types} are compared
     * lexicographically (but in inverted order, so in case of a tie the
     * alphabetically first prediction "wins").
     *
     * @param pred1 the first prediction to be compared
     * @param pred2 the second prediction to be compared
     * @return a negative integer, zero, or a positive integer if the first
     * argument is less than, equal to, or greater than of the second
     * @throws ClassCastException if either of the arguments is not a
     * {@link Prediction}
     */
    public int compare(final Prediction pred1, final Prediction pred2)
            throws ClassCastException {
        final Probability prob1= pred1.getProbability();
        final Probability prob2 = pred2.getProbability();
        final double pR1 = prob1.getPR();
        final double pR2 = prob2.getPR();
        int result = 0;

        // compare pRs if both are given
        if (!Double.isNaN(pR1) && !Double.isNaN(pR2)) {
            result = Double.compare(pR1, pR2);
        }

        // otherwise or in case of a tie compare probabilities
        if (result == 0) {
            result = Double.compare(prob1.getProb(),
                     prob2.getProb());
        }

        // consult raw Winnow score if still tie
        if ((result == 0) && (pred1 instanceof WinnowPrediction)
                && (pred2 instanceof WinnowPrediction)) {
            result = Float.compare(((WinnowPrediction) pred1).getRawScore(),
                ((WinnowPrediction) pred2).getRawScore());
        }

        // let lexicographical order decide if all else fails -- invert order
        if (result == 0) {
            result = -pred1.getType().compareTo(pred2.getType());
        }

        return result;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this).toString();
    }

}
