/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify;

import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.util.InvertedComparator;

/**
 * A distribution over the classes predicted by a classifier. The contained
 * predictions are sorted in descending order, so the most probably prediction
 * comes first.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.7 $, $Date: 2004/09/06 17:21:59 $, $Author: siefkes $
 */
public class PredictionDistribution {

    /**
     * Stores the predictions, in inverted order, so the most probably
     * prediction comes first.
     */
    private final SortedSet<Prediction> store = new TreeSet<Prediction>
            (new InvertedComparator<Prediction>(new PredictionComparator()));

    /**
     * Creates a new instance.
     */
    public PredictionDistribution() {
        super();
    }

    /**
     * Creates a new instance, adding a prediction.
     * @param pred the prediction to add
     */
    public PredictionDistribution(final Prediction pred) {
        this();
        add(pred);
    }

    /**
     * Adds a new prediction to this distribution.
     *
     * @param pred the prediction to add
     * @throws IllegalArgumentException if the distribution already contains
     * a prediction considered equal by {@link PredictionComparator}
     */
    public void add(final Prediction pred) throws IllegalArgumentException {
        final boolean didAdd = store.add(pred);

        if (!didAdd) {
            throw new IllegalArgumentException("Cannot add " + pred
                + " because another prediction in this distribution is "
                + " considered equal");
        }
    }

    /**
     * Returns the best (most probably) prediction in this distribution.
     * @return the best prediction
     */
    public Prediction best() {
        return store.first();
    }

    /**
     * Returns an iterator over the predictions in this distribution,
     * in descending order (most probably class comes first).
     *
     * @return an iterator over the predictions
     */
    public Iterator iterator() {
        return store.iterator();
    }

    /**
     * Returns the number of predictions (classes) in this distribution.
     *
     * @return the number of predictions
     */
    public int size() {
        return store.size();
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("stored predictions", store)
            .toString();
    }

}
