/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * Wraps a probability. Optionally also wraps a "pR" value that can be used
 * to decide between probabilities that are very similar
 * (e.g. ~1.0), because it is less sensitive to rounding issues.
 *
 * Instance of this class are immutable and thus thread-safe.
 * 
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/11/19 14:04:19 $, $Author: siefkes $
 */
public class Probability {

    /**
     * The actual probability; must be in value in the range from 0 to 1; or -1
     * if the actual value is unknown/not applicable.
     */
    private double prob;

    /**
     * An optional pR: <em>pR</em> = log(<em>p</em> /
     * (1-<em>p</em>)); or {@link Double#NaN} if not known/not relevant.
     */
    private double pR;


    /**
     * Creates a new instance, setting the {@linkplain #getPR() pR} to
     * {@link Double#NaN} (unknown).
     *
     * @param theProb the probability of the prediction (must be in the range from
     * 0.0 to 1.0; or -1 if this is a confirmed prediction or an answer key)
     */
    public Probability(final double theProb) {
        this(theProb, Double.NaN);
    }

    /**
     * Creates a new instance.
     *
     * @param theProb the probability of the prediction (must be in the range from
     * 0.0 to 1.0; or -1 if this is a confirmed prediction or an answer key)
     * @param thePR the {@linkplain #getPR() pR} of the prediction;
     * or {@link Double#NaN} if not known
     */
    public Probability(final double theProb, final double thePR) {
        super();
        // check arguments
        if (((theProb < 0.0) && (theProb != -1.0)) || (theProb > 1.0)) {
            throw new IllegalArgumentException("Not a probability: " + theProb);
        }
        prob = theProb;
        pR = thePR;
    }

    
    /**
     * Returns the optional pR: <em>pR</em> = log(<em>p</em> /
     * (1-<em>p</em>)); or {@link Double#NaN} if not known/not relevant.
     *
     * @return the value of the attribute
     */
    public double getPR() {
        return pR;
    }

    /**
     * Returns the actual probability; will be in value in the range from
     * 0 to 1; or -1 if the actual value is unknown/not applicable.
     *
     * @return the value of the attribute
     */
    public double getProb() {
        return prob;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder builder = new ToStringBuilder(this)
            .append(prob);

        // omit pR if not-a-number, i.e. unknown
        if (!Double.isNaN(pR)) {
            builder.append("pR", pR);
        }

        return builder.toString();
    }

}
