/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.util.Util;

/**
 * Reranks the predictions in a distribution by multiplying the {@linkplain
 * de.fu_berlin.ties.classify.Prediction#getProbability() probabilities} of each
 * of them with a bias, if specified for the type of the prediction.
 * 
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2004/11/19 14:04:19 $, $Author: siefkes $
 */
public class Reranker {

    /**
     * The map of biases used to rerank predictions. Keys are
     * {@linkplain Prediction#getType() prediction types} (Strings); values are
     * biases (Doubles). Predictions of types not contained in this map are left
     * unchanged.
     */
    private final Map<String, Double> biasMap;

    /**
     * Creates a new instance.
     * 
     * @param config the biases are read from keys starting with "bias" in
     * this configuration (<code>rerank.classname = bias</code>, e.g.
     * <code>rerank.spam = 0.9</code>)
     */
    public Reranker(final Configuration config) {
        super();
        biasMap = new HashMap<String, Double>();
        final Configuration biasSubset = config.subset("bias");
        final Iterator biasKeys = biasSubset.getKeys();
        String type;
        Object rawValue;
        double value;

        // populate map from subset of configuration
        while (biasKeys.hasNext()) {
            type = (String) biasKeys.next();
            rawValue = biasSubset.getProperty(type);

            // ignore empty/null values
            if (!TiesConfiguration.propertyIsEmpty(rawValue)) {
                value = Util.asDouble(rawValue);
                biasMap.put(type, new Double(value));
                Util.LOG.debug("Reranker: Bias of " + type + " set to "
                        + value);
            }
        }
    }

    /**
     * Creates a new instance.
     * 
     * @param biases a map from {@linkplain Prediction#getType() prediction
     * types} (Strings) to biases (Doubles); predictions of types not contained
     * in this map are left unchanged
     */
    public Reranker(final Map<String, Double> biases) {
        super();
        biasMap = biases;
    }

    /**
     * Reranks the predictions in a distribution.
     *
     * <p><em>The reranked probability estimates are not renormalized, so the
     * result will <strong>not</strong> be a real probability distribution
     * because sum of all probabilities will typically be different from 1.
     * If you want to work on a real probability distribution you have to
     * normalize it yourself.</em>
     * 
     * @param origDist the prediction distribution to rerank
     * @return a new prediction distribution containing the reranked
     * predictions; if no biases are defined at all (bias map is empty), the
     * original distribution is returned
     */
    public PredictionDistribution rerank(
            final PredictionDistribution origDist) {
        if (biasMap.isEmpty()) {
            return origDist;
        } else {
            final PredictionDistribution result = new PredictionDistribution();
            final Iterator origIter = origDist.iterator();
            Prediction origPred, newPred;
            String type;
            Probability origProbability;
            double origProb, origPR;
            double newProb, newPR;
            double bias;

            while (origIter.hasNext()) {
                origPred = (Prediction) origIter.next();
                type = origPred.getType();

                if (biasMap.containsKey(type)) {
                    // multiply pR and prob. with bias
                    bias = biasMap.get(type).doubleValue();
                    origProbability = origPred.getProbability();
                    origProb = origProbability.getProb();
                    origPR = origProbability.getPR();

                    // change probability only if given (non-negative)
                    if (origProb >= 0) {
                        newProb = origProb * bias;
                    } else {
                        newProb = origProb;
                    }

                    // change PR only if given, not if it's NaN
                    if (!Double.isNaN(origPR)) {
                        newPR = origPR * bias;
                    } else {
                        newPR = origPR;
                    }

                    newPred = new Prediction(type, origPred.getSource(),
                            new Probability(newProb, newPR),
                            origPred.getEvalStatus());
                } else {
                    // keep prediction unchanged
                    newPred = origPred;
                }

                // predictions are automatically ordered by the distribution
                result.add(newPred);
            }

            return result;
        }
    }

    /**
     * Returns a string representation of this object.
     * 
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this).append(biasMap).toString();
    }

}
