/*
 * Copyright (C) 2003-2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify;

import java.io.File;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.feature.FeatureTransformer;
import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.classify.winnow.UltraconservativeWinnow;
import de.fu_berlin.ties.classify.winnow.Winnow;
import de.fu_berlin.ties.util.Util;
import de.fu_berlin.ties.xml.io.ObjectElement;
import de.fu_berlin.ties.xml.io.XMLStorable;

/**
 * Classifiers extending this abstract class must provide a training mechanism
 * by implementing the {@link #doTrain(FeatureVector, String, ContextMap)}
 * method. This class supports error-driven learning ("train only errors")
 * which often leads to better prediction models than brute-force training.
 *
 * <p>The code in this class is thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.31 $, $Date: 2004/12/09 18:09:14 $, $Author: siefkes $
 */
public abstract class TrainableClassifier implements Classifier, XMLStorable {

    /**
     * Flag used to load the {@link MetaClassifier}.
     */
    public static final String META_CLASSIFIER = "meta";

    /**
     * Flag used to load the {@link MultiBinaryClassifier}.
     */
    public static final String MULTI_CLASSIFIER = "multi";

    /**
     * Flag used to load the {@link OneAgainstTheRestClassifier}.
     */
    public static final String OAR_CLASSIFIER = "oar";

    /**
     * Names of classifiers wrapping inner classifiers.
     */
    private static final Set<String> wrappingClassifiers;

    /**
     * Static initialization of set of wrapping classifiers.
     */
    static {
        final Set<String> wrapping = new HashSet<String>();
        wrapping.add(META_CLASSIFIER);
        wrapping.add(MULTI_CLASSIFIER);
        wrapping.add(OAR_CLASSIFIER);
        wrappingClassifiers = Collections.unmodifiableSet(wrapping);
    }

    /**
     * Factory method that delegates to
     * {@link #createClassifier(Set, TiesConfiguration)} using the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @param allValidClasses the set of all valid classes
     * @return the created classifier
     * @throws IllegalArgumentException if the value of the
     * {@link #CONFIG_CLASSIFIER} key is missing or invalid
     * @throws ProcessingException if an error occurred while creating the
     * classifier
     */
    public static TrainableClassifier createClassifier(
            final Set<String> allValidClasses)
            throws IllegalArgumentException, ProcessingException {
        return createClassifier(allValidClasses, TiesConfiguration.CONF);
    }

    /**
     * Factory method that delegates to
     * {@link #createClassifier(Set, File, TiesConfiguration)} without
     * specifying an run directory.
     *
     * @param allValidClasses the set of all valid classes
     * @param conf the configuration to use
     * @return the created classifier
     * @throws IllegalArgumentException if the value of the
     * {@link #CONFIG_CLASSIFIER} key is missing or invalid
     * @throws ProcessingException if an error occurred while creating the
     * classifier
     */
    public static TrainableClassifier createClassifier(
            final Set<String> allValidClasses, final TiesConfiguration conf)
            throws IllegalArgumentException, ProcessingException {
        return createClassifier(allValidClasses, null, conf);
    }

    /**
     * Factory method that delegates to
     * {@link #createClassifier(Set, File, FeatureTransformer, String[],
     * TiesConfiguration)}. It reads the specification of the classifier from
     * the {@link #CONFIG_CLASSIFIER} key in the provided configuration. It
     * calls {@link FeatureTransformer#createTransformer(TiesConfiguration)} to
     * initialize a transformer chain, if configured.
     *
     * @param allValidClasses the set of all valid classes
     * @param runDirectory the directory to run the classifier in; used for
     * {@link ExternalClassifier} instead of the
     * {@linkplain ExternalClassifier#CONFIG_DIR configured directory}
     * if not <code>null</code>; ignored otherwise
     * @param conf the configuration to use
     * @return the created classifier
     * @throws IllegalArgumentException if the value of the
     * {@link #CONFIG_CLASSIFIER} key is missing or invalid
     * @throws ProcessingException if an error occurred while creating the
     * classifier
     */
    public static TrainableClassifier createClassifier(
            final Set<String> allValidClasses, final File runDirectory,
            final TiesConfiguration conf)
            throws IllegalArgumentException, ProcessingException {
        // read specification
        final String[] spec = conf.getStringArray(CONFIG_CLASSIFIER);

        // create transformer chain from the configuration, if specified
        final FeatureTransformer trans =
            FeatureTransformer.createTransformer(conf);
        return createClassifier(allValidClasses, runDirectory, trans,
                spec, conf);
    }

    /**
     * Factory method that creates a trainable classifier based on the
     * provided specification.
     *
     * <p>Currently supported values in the first element of the specification:
     *
     * <ul>
     * <li>"Ext" for {@link ExternalClassifier}
     * <li>"Winnow" for {@link Winnow}
     * <li>"ucWinnow" for {@link UltraconservativeWinnow}
     * <li>"Meta" followed by the specification of the inner
     * classifiers as further element(s) for {@link MetaClassifier}
     * <li>"Multi" or "OAR" followed by the specification of the inner
     * classifiers as further element(s) for {@link MultiBinaryClassifier} resp.
     * {@link OneAgainstTheRestClassifier} (if there are only two classes
     * to classify, the outer classifer is skipped and the inner classifier is
     * used directly).
     * </ul>
     *
     * <p>Otherwise the first element must be the qualified name of a
     * TrainableClassifier subclass accepting a {@link Set} (of all valid class
     * names) as first argument, a {@link FeatureTransformer} as second argument
     * and a {@link TiesConfiguration} as third argument.
     *
     * @param allValidClasses the set of all valid classes
     * @param runDirectory the directory to run the classifier in; used for
     * {@link ExternalClassifier} instead of the
     * {@linkplain ExternalClassifier#CONFIG_DIR configured directory}
     * if not <code>null</code>; ignored otherwise
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param spec the specification used to initialize the classifier, as
     * described above
     * @param conf passed to the created classifier to configure itself
     * @return the created classifier
     * @throws IllegalArgumentException if the value of the
     * {@link #CONFIG_CLASSIFIER} key is missing or invalid
     * @throws ProcessingException if an error occurred while creating the
     * classifier
     */
    public static TrainableClassifier createClassifier(
            final Set<String> allValidClasses, final File runDirectory,
            final FeatureTransformer trans, final String[] spec,
            final TiesConfiguration conf)
            throws IllegalArgumentException, ProcessingException {
        if ((spec == null) || spec.length < 1) {
            throw new IllegalArgumentException(
                "Cannot create classifier -- specification is null or empty");
        }


        // convert 1st array element to lower case to compare with defined types
        final String lowerValue = spec[0].toLowerCase();
        final TrainableClassifier result;

        if ("ext".equals(lowerValue)) {
            // load external classifier
            result = new ExternalClassifier(allValidClasses, trans,
                runDirectory, conf);
        } else if ("winnow".equals(lowerValue)) {
            // load Winnow classifier
            result = new Winnow(allValidClasses, trans, conf);
        } else if ("ucwinnow".equals(lowerValue)) {
            // load Ultraconservative Winnow
            result =
                new UltraconservativeWinnow(allValidClasses, trans, conf);
        } else if (wrappingClassifiers.contains(lowerValue)) {
            // Load classifier wrapping inner classifiers
            final String[] innerSpec = new String[spec.length - 1];

            // Use rest of array as specification for the inner classifiers
            for (int i = 0; i < innerSpec.length; i++) {
                innerSpec[i] = spec[i + 1];
            }

            if (META_CLASSIFIER.equals(lowerValue)) {
                // load meta classifier
                result = new MetaClassifier(allValidClasses, trans,
                        runDirectory, innerSpec, conf);
            } else if (allValidClasses.size() > 2) {
                // must be multi-binary or one-against-the-rest classifier
                // whch are unnecessary for 2 classes

                if (MULTI_CLASSIFIER.equals(lowerValue)) {
                    // multi-binary classifier (with background class)
                    result = new MultiBinaryClassifier(allValidClasses, trans,
                            runDirectory, innerSpec, conf);
                } else if (OAR_CLASSIFIER.equals(lowerValue))  {
                    // one-against-the-rest classifier (no background class)
                    result = new OneAgainstTheRestClassifier(allValidClasses,
                            trans, runDirectory, innerSpec, conf);                    
                } else {
                    // not supposed to happen
                    throw new RuntimeException("Implementation error: "
                            + "unknown wrapping classifier" + spec[0]);
                }
            } else {
                // no need to use outer classifier: recursively call this method
                // to return an instance of the inner classifier
                result = createClassifier(allValidClasses, runDirectory, trans,
                    innerSpec, conf);
            }
        } else {
            // should be the qualified name of a TrainableClassifier subclass
            // accepting a Set (of valid class names) as 1st argument, a
            // FeatureTransformer as 2nd and a TiesConfiguration as 3rd argument
            try {
                result = (TrainableClassifier) Util.createObject(
                    Class.forName(spec[0]),
                    new Object[] {allValidClasses, trans, conf},
                    new Class[] {Set.class, FeatureTransformer.class,
                        TiesConfiguration.class});
            } catch (ClassNotFoundException cnfe) {
                // convert and rethrow exception
                throw new ProcessingException(
                    "Cannot create classifier from specification "
                    + ArrayUtils.toString(spec) + ": " + cnfe.toString());
            } catch (InstantiationException ie) {
                // convert and rethrow exception
                throw new ProcessingException(
                    "Cannot create classifier from specification "
                    + ArrayUtils.toString(spec), ie);
            }
        }
        return result;
    }

    /**
     * The immutable set of all valid classes. Each target or candidate class
     * must be contained in this set.
     */
    private final Set<String> allClasses;

    /**
     * Used to configure this instance.
     */
    private final TiesConfiguration config;

    /**
     * If <code>true</code> the classifier considers all classes for
     * error-driven training, not only the candidate classes (results are
     * filtered to the candidate classes prior to returning them).
     */
    private final boolean trainingAll;

    /**
     * The last transformer in a transformer chain, or <code>null</code> if
     * no feature transformers are used.
     */
    private final FeatureTransformer transformer;

    /**
     * Creates a new instance.
     *
     * @param allValidClasses the set of all valid classes
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param conf used to configure this instance
     */
    public TrainableClassifier(final Set<String> allValidClasses,
            final FeatureTransformer trans, final TiesConfiguration conf) {
        super();

        // make set immutable
        allClasses = Collections.unmodifiableSet(allValidClasses);
        config = conf;
        transformer = trans;
        trainingAll = conf.getBoolean("classifier.train.all");
    }

    /**
     * Ensure that all candidate classes are valid (contained in the set of all
     * classes), throwing an exception otherwise.
     *
     * @param candidateClasses an set of classes that are allowed for this item
     * (the actual <code>targetClass</code> must be one of them)
     * @throws IllegalArgumentException if not all candidate classes are members
     * of the {@linkplain #getAllClasses() set of valid classes}
     */
    private void checkCandidateClass(final Set candidateClasses)
            throws IllegalArgumentException {
        final Iterator classIter = candidateClasses.iterator();
        String currentClass;

        while (classIter.hasNext()) {
            currentClass = (String) classIter.next();
            if (!allClasses.contains(currentClass)) {
                throw new IllegalArgumentException("Candidate class "
                    + currentClass
                    + " is not in the set of valid classes: " + allClasses);
            }
        }
    }

    /**
     * Ensure that the target class is valid (contained in the set of all
     * classes), throwing an exception otherwise.
     *
     * @param targetClass the expected class of an instance; must be
     * contained in the set of <code>candidateClasses</code>
     * @throws IllegalArgumentException if the target class is not a member of
     * the {@linkplain #getAllClasses() set of valid classes}
     */
    private void checkTargetClass(final String targetClass)
            throws IllegalArgumentException {
        // ensure that the target class is valid
        if (!allClasses.contains(targetClass)) {
            throw new IllegalArgumentException("Target class " + targetClass
                + " is not in the set of valid classes: " + allClasses);
        }
    }

    /**
     * Classifies an item that is represented by a feature vector by choosing
     * the most probable class among a set of candidate classes. Delegates to
     * the abstract {@link #doClassify(FeatureVector, Set, ContextMap)} method.
     *
     * @param features the feature vector to consider
     * @param candidateClasses an set of classes that are allowed for this item
     * @return the result of the classification; you can call
     * {@link PredictionDistribution#best()} to get the most probably class
     * @throws IllegalArgumentException if the
     * {@linkplain #getAllClasses() set of valid classes} does not contain all
     * candidate classes
     * @throws ProcessingException if an error occurs during classification
     */
    public final PredictionDistribution classify(final FeatureVector features,
            final Set candidateClasses)
            throws IllegalArgumentException, ProcessingException {
        // ensure that all candidate classes are valid
        checkCandidateClass(candidateClasses);

        // transform features
        final FeatureVector actualFeatures = (transformer != null)
            ? transformer.transform(features) : features;

        // delegate to abstract method
        final ContextMap context = new ContextMap();
        return doClassify(actualFeatures, candidateClasses, context);
    }

    /**
     * Classifies an item that is represented by a feature vector by choosing
     * the most probable class among a set of candidate classes.
     *
     * @param features the feature vector to consider
     * @param candidateClasses an set of classes that are allowed for this item
     * @param context can be used to transport implementation-specific
     * contextual information between the
     * {@link #doClassify(FeatureVector, Set, ContextMap)},
     * {@link #doTrain(FeatureVector, String, ContextMap)}, and
     * {@link #trainOnErrorHook(PredictionDistribution, FeatureVector, String,
     * Set, ContextMap)} methods
     * @return the result of the classification; you can call
     * {@link PredictionDistribution#best()} to get the most probably class
     * @throws ProcessingException if an error occurs during classification
     */
    protected abstract PredictionDistribution doClassify(
            final FeatureVector features, final Set candidateClasses,
            final ContextMap context)
            throws ProcessingException;

    /**
     * Incorporates an item that is represented by a feature vector into the
     * classification model.
     *
     * @param features the feature vector to consider
     * @param targetClass the class of this feature vector
     * @param context can be used to transport implementation-specific
     * contextual information between the
     * {@link #doClassify(FeatureVector, Set, ContextMap)},
     * {@link #doTrain(FeatureVector, String, ContextMap)}, and
     * {@link #trainOnErrorHook(PredictionDistribution, FeatureVector, String,
     * Set, ContextMap)} methods
     * @throws ProcessingException if an error occurs during training
     */
    protected abstract void doTrain(final FeatureVector features,
            final String targetClass, final ContextMap context)
            throws ProcessingException;

    /**
     * Returns the set of all valid classes. Each target or candidate class
     * must be contained in this set.
     *
     * @return an immutable set containing all valid class names
     */
    public Set getAllClasses() {
        return allClasses;
    }

    /**
     * Returns the configuration used by this instance.
     * @return the used configuration
     */
    public TiesConfiguration getConfig() {
        return config;
    }

    /**
     * Resets the classifer, completely deleting the prediction model.
     * @throws ProcessingException if an error occurs during reset
     */
    public abstract void reset() throws ProcessingException;

    /**
     * Invoked by {@link #trainOnError(FeatureVector, String, Set)} to decide
     * whether to train an instance. The default behavior is to train if the
     * best prediction was wrong or didn't yield a positive probability
     * ("train only errors"). Subclasses can override this method to
     * add their own behavior, e.g. reinforcement training (thick threshold
     * heuristic).
     *
     * @param targetClass the expected class of this feature vector; must be
     * contained in the set of <code>candidateClasses</code>
     * @param predDist the prediction distribution returned by
     * {@link #doClassify(FeatureVector, Set, ContextMap)}
     * @param context can be used to transport implementation-specific
     * contextual information between the
     * {@link #doClassify(FeatureVector, Set, ContextMap)},
     * {@link #doTrain(FeatureVector, String, ContextMap)}, and
     * {@link #trainOnErrorHook(PredictionDistribution, FeatureVector, String,
     * Set, ContextMap)} methods
     * @return whether to train this instance
     */
    protected boolean shouldTrain(final String targetClass,
            final PredictionDistribution predDist, final ContextMap context) {
        final Prediction best = predDist.best();
        final double bestProb = best.getProbability().getProb();
        return !best.getType().equals(targetClass) || Double.isNaN(bestProb)
            || (bestProb <= 0.0);
    }

    /**
     * {@inheritDoc}
     * Subclasses of {@link TrainableClassifier} should extend this method and
     * the corresponding constructor from {@link org.dom4j.Element} to
     * ensure (de)serialization works as expected.
     */
    public ObjectElement toElement() {
        final ObjectElement result =
            new ObjectElement("classifier", this.getClass());
        // TODO implement this method + Element constructor
        return result;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder builder = new ToStringBuilder(this);
        builder.append("all classes", allClasses);

        if (trainingAll) {
            builder.append("training all classes", trainingAll);
        }

        if (transformer != null) {
            builder.append("transformer", transformer);
        }

        return builder.toString();
    }

    /**
     * Incorporates an item that is represented by a feature vector into the
     * classification model. Delegates to the abstract
     * {@link #doTrain(FeatureVector, String, ContextMap)} method.
     *
     * @param features the feature vector to consider
     * @param targetClass the class of this feature vector
     * @throws IllegalArgumentException if the target class is not in the
     * {@linkplain #getAllClasses() set of valid classes}
     * @throws ProcessingException if an error occurs during training
     */
    public final void train(final FeatureVector features,
            final String targetClass)
            throws IllegalArgumentException, ProcessingException {
        // ensure that the target class is valid
        checkTargetClass(targetClass);

        // transform features
        final FeatureVector actualFeatures = (transformer != null)
            ? transformer.transform(features) : features;

        // delegate to abstract method
        final ContextMap context = new ContextMap();
        doTrain(actualFeatures, targetClass, context);
    }

    /**
     * Handles error-driven learning ("train only errors"): the specified
     * feature vector is trained into the model only if the predicted class
     * for the feature vector differs from the specified target class. If the
     * prediction was correct, the model is not changed.
     *
     * @param features the feature vector to consider
     * @param targetClass the expected class of this feature vector; must be
     * contained in the set of <code>candidateClasses</code>
     * @param candidateClasses an set of classes that are allowed for this item
     * (the actual <code>targetClass</code> must be one of them)
     * @return the original prediction distribution if the best prediction was
     * wrong, i.e. if training was necessary; or <code>null</code> if no
     * training was necessary (the prediction was already correct)
     * @throws ProcessingException if an error occurs during training
     */
    public final PredictionDistribution trainOnError(
            final FeatureVector features, final String targetClass,
            final Set candidateClasses) throws ProcessingException {
        // transform features after checking that all classes are valid
        checkTargetClass(targetClass);
        checkCandidateClass(candidateClasses);
        final FeatureVector actualFeatures = (transformer != null)
            ? transformer.transform(features) : features;

        // consider all classes for training if configured
        final Set consideredClasses = trainingAll
            ? allClasses : candidateClasses;
        final ContextMap context = new ContextMap();

        // call abstract method to classify
        final PredictionDistribution predDist =
            doClassify(actualFeatures, consideredClasses, context);

        // call hook for subclasses
        final boolean hookHandledTraining = trainOnErrorHook(predDist,
            actualFeatures, targetClass, consideredClasses, context);

        if (shouldTrain(targetClass, predDist, context)) {
            // don't train if the hook kicked in
            if (!hookHandledTraining) {
                doTrain(actualFeatures, targetClass, context);
            }

            // filter prediction distribution to candidate classes
            if (trainingAll) {
                final Iterator predIter = predDist.iterator();
                Prediction pred;

                while (predIter.hasNext()) {
                    pred = (Prediction) predIter.next();

                    if (!candidateClasses.contains(pred.getType())) {
                        // not a candidate class
                        predIter.remove();
                    }
                }
            }

            // check whether best prediction is correct after filtering
            if (((predDist.size() > 0)
                    && !predDist.best().getType().equals(targetClass))) {
                // distribution was wrong -- return for analysis
                return predDist;
            } else {
                // null signals that prediction was fine
                return null;
            }
        } else {
            // null signals that prediction was fine
            return null;
        }
    }

    /**
     * Subclasses can implement this hook for more refined error-driven
     * learning. It is called from the
     * {@link #trainOnError(FeatureVector, String, Set)} method after
     * classifying. This method can do any necessary training itself and
     * return <code>true</code> to signal that no further action is necessary.
     * This implementation is just a placeholder that always returns
     * <code>false</code>.
     *
     * @param predDist the prediction distribution returned by
     * {@link #classify(FeatureVector, Set)}
     * @param features the feature vector to consider
     * @param targetClass the expected class of this feature vector; must be
     * contained in the set of <code>candidateClasses</code>
     * @param candidateClasses an set of classes that are allowed for this item
     * (the actual <code>targetClass</code> must be one of them)
     * @param context can be used to transport implementation-specific
     * contextual information between the
     * {@link #doClassify(FeatureVector, Set, ContextMap)},
     * {@link #doTrain(FeatureVector, String, ContextMap)}, and
     * {@link #trainOnErrorHook(PredictionDistribution, FeatureVector, String,
     * Set, ContextMap)} methods
     * @return this implementation always returns <code>false</code>; subclasses
     * can return <code>true</code> to signal that any error-driven learning was
     * already handled
     * @throws ProcessingException if an error occurs during training
     */
    protected boolean trainOnErrorHook(final PredictionDistribution predDist,
            final FeatureVector features, final String targetClass,
            final Set candidateClasses, final ContextMap context)
            throws ProcessingException {
        return false;
    }

}
