/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.feature;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * Default implementation of the
 * {@link de.fu_berlin.ties.classify.feature.Feature} class. Instances of this
 * class are immutable and thus thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/05/31 19:13:49 $, $Author: siefkes $
 */
public class DefaultFeature extends Feature {

    /**
     * The strenght of this feature -- higher values might indicate more
     * important or more frequent features.
     */
    private final float strength;

    /**
     * The representation of this feature, to be used for classification.
     */
    private final String representation;

    /**
     * Creates a new instance, without storing a comment.
     *
     * @param rep the representation of this feature; or <code>null</code> to
     * create a comment-only feature
     */
    public DefaultFeature(final String rep) {
        this(rep, null);
    }

    /**
     * Creates a new instance, setting the strength to 1.0.
     *
     * @param rep the representation of this feature; or <code>null</code> to
     * create a comment-only feature
     * @param myComment a comment on this feature or the features to follow;
     * might be <code>null</code>
     */
    public DefaultFeature(final String rep, final String myComment) {
        this(rep, myComment, 1.0f);
    }

    /**
     * Creates a new instance.
     *
     * @param rep the representation of this feature; or <code>null</code> to
     * create a comment-only feature
     * @param myComment a comment on this feature or the features to follow;
     * might be <code>null</code>
     * @param myStrength The strenght of this feature -- higher values might
     * indicate more important or more frequent features
     */
    public DefaultFeature(final String rep, final String myComment,
                          final float myStrength) {
        super(myComment);
        representation = rep;
        strength = myStrength;
    }

    /**
     * Returns the representation of this feature, to be used for
     * classification.
     *
     * @return the feature representation, or <code>null</code> if this feature
     * contains only a comment
     */
    public String getRepresentation() {
        return representation;
    }

    /**
     * Returns a strength value for this feature.
     *
     * <p>This method is non-public because it should only be accessed through
     * {@link FeatureVector#strength(Feature)} to allow feature vectors to
     * modify the strenghts of the stored features.
     *
     * @return a strength value for this feature
     */
    public float getStrength() {
        return strength;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("representation", representation)
            .append("strength", strength)
            .toString();
    }

}
