/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.feature;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * An abstract base class for immutable feature representation used for
 * classification.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/05/31 19:13:49 $, $Author: siefkes $
 */
public abstract class Feature {

    /**
     * An optional comment on this feature, ignored for classification.
     */
    private final String comment;

    /**
     * Creates a new instance, without storing a comment.
     */
    public Feature() {
        this(null);
    }

    /**
     * Creates a new instance.
     *
     * @param myComment a comment on this feature, ignored for classification;
     * might be <code>null</code>
     */
    public Feature(final String myComment) {
        super();
        comment = myComment;
    }

    /**
     * Indicates whether some other object is "equal to" this one, fulfulling
     * the {@link Object#equals(java.lang.Object)} contract. To be considered
     * equal, the object must be a {@link Feature}. If the
     * {@linkplain #getRepresentation() representation} of this instance is not
     * <code>null</code>, the representations of the two features are compared.
     * Otherwise the {@linkplain #getFullRepresentation() full representations}
     * are compared. Thus for normal features comments and other fields will
     * be ignored. For comment-only features, comments will be considered,
     * other fields will be ignored.
     *
     * @param obj the reference object with which to compare
     * @return <code>true</code> iff the specified object is a
     * {@link Feature} equal to this instance
     */
    public boolean equals(final Object obj) {
        if (obj == this) {
            return true;
        } else if (obj instanceof Feature) {
            final Feature other = (Feature) obj;
            final String myRep = getRepresentation();
            if (myRep != null) {
                // compare representations if present
                return myRep.equals(other.getRepresentation());
            } else {
                // otherwise compare full representations (incl. comments)
                return getFullRepresentation().equals(
                    other.getFullRepresentation());
            }
        } else {
            return false;
        }
    }

    /**
     * Returns the comment attached to this feature, if any.
     *
     * @return the comment, or <code>null</code> is no comment was stored
     */
    public String getComment() {
        return comment;
    }

    /**
     * Prints a full representation of this feature that contains both
     * representation (if any) and comment (if any). The comment is preceded
     * by a "#" character.
     *
     * @return the full representation
     */
    public String getFullRepresentation() {
        return getFullRepresentation("#");
    }

    /**
     * Prints a full representation of this feature that contains both
     * representation (if any) and comment (if any). The comment is preceded by
     * the specified <code>separator</code> (surrounded by spaces).
     *
     * @param separator used to introduce the comment
     * @return the full representation
     */
    public String getFullRepresentation(final String separator) {
        final StringBuffer result = new StringBuffer();
        final String representation = getRepresentation();
        if ((representation != null) && (representation.length() > 0)) {
            result.append(representation);
        }

        if ((comment != null) && (comment.length() > 0)) {
            if (result.length() > 0) {
                result.append(' ');
            }
            result.append(separator).append(' ').append(comment);
        }
        return result.toString();
    }

    /**
     * Abstract method for returning the representation of this feature, to be
     * used for classification.
     *
     * @return the feature representation, or <code>null</code> if this feature
     * contains only a comment
     */
    public abstract String getRepresentation();

    /**
     * Returns a strength value for this feature. This method can be
     * overwritten to assign higher strenghts to more important or more frequent
     * features. This implementation always returns 1.0.
     *
     * <p>Typically you should call {@link FeatureVector#strength(Feature)}
     * instead of this method to allow feature vectors to modify the strenghts
     * of the stored features.
     *
     * @return a strength value for this feature -- 1.0 in this implementation
     */
    public float getStrength() {
        return 1.0f;
    }

    /**
     * Returns a hash code value for this object, fulfulling the
     * {@link Object#hashCode()} contract.
     * @return a hash code value for this object
     */
    public int hashCode() {
        final String myRep = getRepresentation();
        if (myRep != null) {
            // hash representation if present
            return myRep.hashCode();
        } else {
            // otherwise hash full representation (incl. comments)
            return getFullRepresentation().hashCode();
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder builder = new ToStringBuilder(this);
        if ((comment != null) && (comment.length() > 0)) {
            builder.append("comment", comment);
        }
        return builder.toString();
    }

}
