/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.feature;

import java.util.Collection;
import java.util.Iterator;

import org.apache.commons.collections.Bag;
import org.apache.commons.collections.bag.HashBag;

/**
 * This feature vector implementation stores a multi-set of features.
 * The order of features is not preserved. Internally it uses a
 * {@link org.apache.commons.collections.bag.HashBag} as storage.
 *
 * <p>Instances of this class are not thread-safe and must be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.11 $, $Date: 2004/09/06 17:22:10 $, $Author: siefkes $
 */
public class FeatureSet extends FeatureVector {

    /**
     * Constants specifying that feature frequencies are not considered
     * when calculating {@linkplain FeatureVector#strength(Feature) strength}
     * values.
     */
    public static final String STRENGTH_CONSTANT = "constant";

    /**
     * Constants specifying that the logarithm of feature frequencies is
     * considered when calculating {@linkplain FeatureVector#strength(Feature)
     * strength} values (1.0 is added to log(<em>f</em>) to avoid the result
     * becoming 0).
     */
    public static final String STRENGTH_LOG = "log";

    /**
     * Constants specifying that the square root of feature frequencies is
     * considered when calculating {@linkplain FeatureVector#strength(Feature)
     * strength} values.
     */
    public static final String STRENGTH_SQUARE_ROOT = "sqrt";

    /**
     * Constants specifying that feature frequencies are considered linear
     * (as is) when calculating {@linkplain FeatureVector#strength(Feature)
     * strength} values (a feature occurring twice as frequently will be
     * twice as strong).
     */
    public static final String STRENGTH_LINEAR = "linear";


    /**
     * Store features in an multi-set (bag).
     */
    private final Bag store = new HashBag();

    /**
     * The type of method used to consider feature frequencies when determining
     * {@linkplain #strength(Feature) strength} values.
     */
    private final String strengthType;

    /**
     * Creates a new instance.
     *
     * @param strengthMethod The type of method used to consider feature
     * frequencies when determining {@linkplain #strength(Feature) strength}
     * values -- should be one of the <code>STRENGTH</code> constants
     * defined in this class
     */
    public FeatureSet(final String strengthMethod) {
        super();

        // convert to lower case + internalize to allow efficient comparison
        strengthType = strengthMethod.toLowerCase().intern();
    }

    /**
     * {@inheritDoc}
     */
    public Iterator iterator() {
        // iterate each feature only once
        return store.uniqueSet().iterator();
    }

    /**
     * Returns the type of method used to consider feature frequencies when
     * determining {@linkplain #strength(Feature) strength} values.
     *
     * @return the value of the attribute -- should be one of the
     * <code>STRENGTH</code> constants defined in this class
     */
    public String getStrengthType() {
        return strengthType;
    }

    /**
     * Implementation of the hook provided by the superclass to ensure that
     * the {@linkplain FeatureVector#getSummedStrength() summed strength}
     * is updated correctly.
     *
     * @param feature the feature to add
     */
    protected void preAddHook(final Feature feature) {
        if (store.contains(feature)) {
            // subtract old feature strength from sum -- the superclass will
            // re-add the new (overall) strength after adding the feature
            setSummedStrength(getSummedStrength() - strength(feature));
        }
    }

    /**
     * {@inheritDoc}
     */
    public int size() {
        // count each feature only once
        return store.uniqueSet().size();
    }

    /**
     * {@inheritDoc}
     */
    protected Collection<Feature> store() {
        return store;
    }

    /**
     * Returns a strength value for a feature contained in this vector.
     * This implementation simply delegates to {@link Feature#getStrength()}.
     *
     * @param feature the feature to consider
     * @return a strength value for the specified feature
     */
    public double strength(final Feature feature) {
        final double baseStrength = super.strength(feature);
        final int frequency = store.getCount(feature);
        final double result;

        // strengthType is internalized so we can do identity comparisons
        if (strengthType == STRENGTH_CONSTANT) {
            result = baseStrength;
        } else if (strengthType == STRENGTH_LOG) {
            // adding 1 so the result will never be 0.0 (but 1.0 for singletons)
            result = baseStrength * (Math.log(frequency) + 1.0);
        } else if (strengthType == STRENGTH_SQUARE_ROOT) {
            result = baseStrength * Math.sqrt(frequency);
        } else if (strengthType == STRENGTH_LINEAR) {
            result = baseStrength * frequency;
        } else {
            throw new IllegalArgumentException("Unsupported strength type: "
                    + strengthType);
        }

/*        if (result != 1.0) {
            Util.LOG.debug("Strength of feature " + feature + ": " + result
                    + " (base strength: " + baseStrength + ", frequency: "
                    + frequency + ", calculation method: " + strengthType);
        } */

        return result;
    }

}
