/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.feature;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.util.Util;

/**
 * An feature transformer transforms a feature vector in a suitable way.
 * Feature transformer can be used for feature selection (discarding irrelevant
 * features), for combining features etc. Feature transforms can be
 * <em>chained</em> -- in this case each feature transformer will work on the
 * featured returned by the previous one.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/07/08 08:42:18 $, $Author: siefkes $
 */
public abstract class FeatureTransformer {

    /**
     * Configuration key specifying the list of transformers to chain.
     */
    public static final String CONFIG_TRANSFORMERS = "transformer.chain";

    /**
     * Factory method that delegates to {@link #createTransformer()} using the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @return the last transformer in the created transformer chain; or
     * <code>null</code> is no transformers are specified (parameter value is
     * <code>null</code> or empty)
     * @throws IllegalArgumentException if the value of the
     * {@link #CONFIG_TRANSFORMERS} key is invalid
     * @throws ProcessingException if an error occurred while creating the
     * transformer
     */
    public static FeatureTransformer createTransformer()
    throws IllegalArgumentException, ProcessingException {
        return createTransformer(TiesConfiguration.CONF);
    }

    /**
     * Factory method that creates a feature transformer based on the
     * {@link #CONFIG_TRANSFORMERS} key in the provided configuration.
     * This parameter must contain a list fully specified names of subclasses
     * of this class) used in a chain.
     *
     * <p>Eeach specified transformer must provide a constructor accepting a
     * preceding {@link FeatureTransformer} as first argument and a
     * {@link TiesConfiguration} as second argument.
     *
     * @param config the configuration to use
     * @return the last transformer in the created transformer chain; or
     * <code>null</code> is no transformers are specified (parameter value is
     * <code>null</code> or empty)
     * @throws IllegalArgumentException if the value of the
     * {@link #CONFIG_TRANSFORMERS} key is invalid
     * @throws ProcessingException if an error occurred while creating the
     * transformer
     */
    public static FeatureTransformer createTransformer(
            final TiesConfiguration config)
    throws IllegalArgumentException, ProcessingException {
        final String[] paramValue = config.getStringArray(CONFIG_TRANSFORMERS);
        FeatureTransformer transformer = null;
        int i = 0;

        try {
            for (; i < paramValue.length; i++) {
                // initialize each of the specified transformers, handing over
                // the previous transformer as first single argument and the
                // TiesConfiguration as second argument
                if (StringUtils.isNotEmpty(paramValue[i])) {
                    transformer = (FeatureTransformer) Util.createObject(
                        Class.forName(paramValue[i]),
                        new Object[] {transformer, config},
                        new Class[] {FeatureTransformer.class,
                            TiesConfiguration.class});
                }
            }
        } catch (ClassNotFoundException cnfe) {
             // convert and rethrow exception
             throw new ProcessingException(
                 "Cannot create transformer chain from key "
                 + CONFIG_TRANSFORMERS + " because the class " + paramValue[i]
                 + " is missing: " + cnfe.toString());
         } catch (InstantiationException ie) {
             // convert and rethrow exception
             throw new ProcessingException(
                 "Cannot create transformer chain from key "
                 + CONFIG_TRANSFORMERS + " because instantation of the class "
                 + paramValue[i] + " failed: ", ie);
        }

        // return the last transformer in the chain (if any)
        return transformer;
    }

    /**
     * The preceding transformer used if this transformer is part of a
     * <em>chain</em>; <code>null</code> otherwise.
     */
    private final FeatureTransformer precedingTransformer;

    /**
     * Creates a new instance.
     *
     * @param precTrans the preceding transformer to use if this transformer
     * is part of a <em>chain</em>; <code>null</code> otherwise
     */
    public FeatureTransformer(final FeatureTransformer precTrans) {
        super();
        precedingTransformer = precTrans;
    }

    /**
     * Transforms a feature vector.
     *
     * @param orgFeatures the original feature vector to transform
     * @return a new feature vector containing the transformed features
     */
    protected abstract FeatureVector doTransform(
        final FeatureVector orgFeatures);

    /**
     * Returns the preceding transformer used if this transformer is part of a
     * <em>chain</em>.
     *
     * @return the preceding transformer, if any; or <code>null</code> if this
     * transformer is not part of a chain resp. is the first transformer in a
     * chain
     */
    public FeatureTransformer getPrecedingTransformer() {
        return precedingTransformer;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder builder =  new ToStringBuilder(this);

        if (precedingTransformer != null) {
            builder.append("preceding transformer", precedingTransformer);
        }
        return builder.toString();
    }

    /**
     * Transforms a feature vector. This method calls itself on the
     * {@linkplain #getPrecedingTransformer() preceding transformer} (if any)
     * prior to delegating to the abstract {@link #doTransform(FeatureVector)}
     * method.
     *
     * @param orgFeatures the original feature vector to transform
     * @return a new feature vector containing the transformed features
     */
    public FeatureVector transform(final FeatureVector orgFeatures) {
        final FeatureVector result;

        if (precedingTransformer != null) {
            // call preceding transformer prior to abstract method
            result = doTransform(precedingTransformer.transform(orgFeatures));
        } else {
            // no preceding transformer -- just call abstract method
            result = doTransform(orgFeatures);
        }

//      Util.LOG.debug("Length of original feature vector: "
//          + orgFeatures.size() + ", of transformed vector: " + result.size());
        return result;
    }

}
