/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.feature;

import java.util.Collection;
import java.util.Iterator;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.text.TextTokenizer;
import de.fu_berlin.ties.text.TextUtils;

/**
 * A feature vector contains the features representing an instance. Subclasses
 * must implement the {@link #store()} method to provide a collection for
 * storing the features.
 *
 * <p>Instances of this class are thread-safe if and only if the provided
 * collection is thread-safe (normally this won't be the case).
 *
 * @author Christian Siefkes
 * @version $Revision: 1.6 $, $Date: 2004/09/06 17:22:10 $, $Author: siefkes $
 */
public abstract class FeatureVector {

    /**
     * The sum of the {@linkplain #strength(Feature) strength values}
     * of all features currently contained in this vector.
     */
    private double summedStrength = 0.0;

    /**
     * Creates a new instance.
     */
    public FeatureVector() {
        super();
    }

    /**
     * Adds a feature to this vector.
     *
     * @param feature the feature to add
     */
    public void add(final Feature feature) {
        preAddHook(feature);
        store().add(feature);

        // add feature strength to summed strength
        summedStrength += strength(feature);
    }

    /**
     * Adds all of the features in the specified Collection to this vector,
     * in the order they are returned by the specified Collection's Iterator.
     *
     * @param coll a collection of features to add
     * @throws ClassCastException if the collection contains elements that are
     * not {@link Feature}s
     */
    public void addAll(final Collection coll) throws ClassCastException {
        Feature feature;
        final Iterator iter = coll.iterator();

        while (iter.hasNext()) {
            feature = (Feature) iter.next();
            add(feature);
        }
    }

    /**
     * Adds all of the features in the specified feature vector to this vector,
     * in the order they are returned by the specified feature vector.
     *
     * @param fv the vector of features to add
     */
    public void addAll(final FeatureVector fv) {
        addAll(fv.store());
    }

    /**
     * Tokenizes a text, creating and adding a feature for each token.
     *
     * @param text the text to tokenize
     * @param tokenizer the tokenizer to use
     */
    public void addAllTokens(final CharSequence text,
                             final TextTokenizer tokenizer) {
        tokenizer.reset(text);
        String token;

        while ((token = tokenizer.nextToken()) != null) {
            add(new DefaultFeature(token));
        }
    }

    /**
     * Returns a sum of the {@linkplain #strength(Feature) strength values}
     * of all features contained in this vector.
     *
     * @return the summed strength of all features
     */
    public double getSummedStrength() {
        return summedStrength;
    }

    /**
     * Flattens the contained features into a single character sequence, without
     * including comments. Features are separated by newlines.
     *
     * @return the resulting character sequence, created by printing calling
     * {@link Feature#getRepresentation()} on each feature in the vector; each
     * feature representation is followed by a newline
     */
    public CharSequence flatten() {
        return flatten(false);
    }

    /**
     * Flattens a list of features into a single character sequence. Features
     * are separated by newlines.
     *
     * @param inclComments whether or not to include comments
     * @return the resulting character sequence, created by printing calling
     * {@link Feature#getRepresentation()} on each feature in the vector; each
     * feature representation is followed by a newline
     */
    public CharSequence flatten(final boolean inclComments) {
        final StringBuffer result = new StringBuffer();
        final Iterator iter = iterator();
        Feature currentFeature;
        String currentRep;

        while (iter.hasNext()) {
            currentFeature = (Feature) iter.next();
            if (inclComments) {
                currentRep = currentFeature.getFullRepresentation();
            } else {
                currentRep = currentFeature.getRepresentation();
            }

            // representation w/o comment might be null
            if (currentRep != null) {
                result.append(currentRep);
                result.append(TextUtils.LINE_SEPARATOR);
            }
        }
        return result;
    }

    /**
     * Returns an iterator over the {@link Feature}s stored in this vector.
     *
     * @return an iterator over the stored features
     */
    public Iterator iterator() {
        return store().iterator();
    }

    /**
     * Empty method that can by implemented by child classes, for example if
     * they calculate strength values in some special way. This method is
     * called at the start of each {@link #add(Feature)} operation.
     *
     * @param feature the feature to add
     */
    protected void preAddHook(final Feature feature) {
    }

    /**
     * Modifies the summed strength of all features contained in this instance.
     * This method exists for the convenience of child classes that calculate
     * strength values in some special way.
     *
     * @param newSum the new value of the attribute
     */
    protected void setSummedStrength(final double newSum) {
        summedStrength = newSum;
    }

    /**
     * Returns the number of features stored in this vector.
     *
     * @return the number of features
     */
    public int size() {
        return store().size();
    }

    /**
     * Returns a strength value for a feature contained in this vector.
     * This implementation simply delegates to {@link Feature#getStrength()}.
     *
     * @param feature the feature to consider
     * @return a strength value for the specified feature
     */
    public double strength(final Feature feature) {
        return feature.getStrength();
    }

    /**
     * Returns the collection used for storing the features. The properties
     * of a feature vector depend on the kind of collection that is used.
     *
     * @return the collection used for storing the features.
     */
    protected abstract Collection<Feature> store();

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("feature store", store())
            .toString();
    }

}
