/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.feature;

import java.util.Iterator;

import org.apache.commons.collections.buffer.CircularFifoBuffer;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.util.CollectionUtils;
import de.fu_berlin.ties.util.Util;

/**
 * Transforms a feature vector using the <em>orthogonal sparse bigrams
 * (OSB)</em> technique developed by Fidelis Assis. This transformer
 * discard all comment-only features. It slides of window of
 * {@linkplain #getLength() length <em>N</em>} over the remaining original
 * features. At each window position it generates <em>N</em>-1 joint features
 * as exemplified above (assumping the pipe character "|" is used as
 * {@linkplain #getSeparator() separator} and <em>N</em>=5:
 *
 * <pre>
 *    -   -   -  w4 | w5
 *    -   -   w3  | | w5
 *    -  w2   |   | | w5
 *   w1   |   |   | | w5
 * </pre>
 *
 * <p>If {@link #isPreserving()}, the original features are preserved as well;
 * otherwise they are discarded.
 *
 * <p>Instances of this class are thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.6 $, $Date: 2004/12/06 17:57:30 $, $Author: siefkes $
 */
public class OSBTransformer extends FeatureTransformer {

    /**
     * The maximum number of original features joined.
     */
    private final int length;

    /**
     * Whether to preserve the original features as well or only to use joint
     * features.
     */
    private final boolean preserving;

    /**
     * The string used to separate original features (by default a space
     * character). This string <strong>should never occur</strong> within
     * original features.
     */
    private final String separator;

    /**
     * Array of strength values used for bigrams with different distances.
     */
    private final float[] strengths;

    /**
     * Strength value used for unigrams (single tokens) if they are
     * {@linkplain #isPreserving() preserved}.
     */
    private final float unigramStrength;

    /**
     * Creates a new instance.
     *
     * @param precTrans the preceding transformer to use if this transformer
     * is part of a <em>chain</em>; <code>null</code> otherwise
     * @param len the maximum number of original features joined; minimum value
     * is 2
     * @param sep the string used to separate original features -- this string
     * <strong>should never occur</strong> within original features
     * @param preserve whether to preserve the original features as well or
     * only to use joint features
     * @param strengthArray Array of strength values used for bigrams with
     * different distances
     * @param singleTokenStrength Strength value used for unigrams (single
     * tokens); ignored if <code>preserve</code> is <code>false</code>
     * @throws IllegalArgumentException if <code>len &lt; 2</code> or if
     * <code>strengthArray</code> is empty
     */
    public OSBTransformer(final FeatureTransformer precTrans, final int len,
                          final String sep, final boolean preserve,
                          final float[] strengthArray,
                          final float singleTokenStrength)
    throws IllegalArgumentException {
        super(precTrans);

        // check + store arguments
        if (len < 2) {
            throw new IllegalArgumentException(
                "OSB length must be at least 2:" + len);
        }
        length = len;
        separator = sep;
        preserving = preserve;
        unigramStrength = singleTokenStrength;

        // store strengh array, ensuring there are sufficient values
        if (strengthArray.length >= len - 1) {
            strengths = strengthArray;
        } else if (strengthArray.length > 0) {
            strengths = new float[len - 1];
            int i = 0;

            while (i < strengthArray.length) {
                strengths[i] = strengthArray[i];
                i++;
            }

            while (i < strengths.length) {
                // re-use last specified value for rest of array
                strengths[i] = strengthArray[strengthArray.length - 1];
                i++;
            }

            if (strengthArray.length > 1) {
                Util.LOG.warn("Strength array has only " + strengthArray.length
                        + " elements instead of " + len
                        + " ones -- re-using the last one");
            } else {
                Util.LOG.debug("Using uniform strength " + strengthArray[0]
                        + " for all bigrams");
            }
        } else {
            throw new IllegalArgumentException(
                "Empty strength array");
        }
    }

    /**
     * Creates a new instance.
     *
     * @param precTrans the preceding transformer to use if this transformer
     * is part of a <em>chain</em>; <code>null</code> otherwise
     * @param config used to configure this instance
     */
    public OSBTransformer(final FeatureTransformer precTrans,
            final TiesConfiguration config) {
        this(precTrans, config.getInt("transformer.osb.length"),
            config.getString("transformer.osb.separator",
                SBPHTransformer.DEFAULT_SEPARATOR),
            config.getBoolean("transformer.osb.preserve"),
            CollectionUtils.asFloatArray(config.getStringArray(
                                "transformer.osb.strengths")),
            config.getFloat("transformer.osb.strength.unigram"));
    }

    /**
     * {@inheritDoc}
     */
    protected FeatureVector doTransform(final FeatureVector orgFeatures) {
        final FeatureVector result = new DefaultFeatureVector();
        final Iterator orgIter = orgFeatures.iterator();
        final CircularFifoBuffer priorFeatureReps =
            new CircularFifoBuffer(length - 1);
        Iterator bufferIter;
        String orgRep;
        Feature orgF;
        StringBuffer newRep = new StringBuffer();
        int distance;
        int i;

        while (orgIter.hasNext()) {
            orgF = (Feature) orgIter.next();
            orgRep = orgF.getRepresentation();

            // ignore comment-only features
            if (orgRep != null) {
                // append original feature if configured
                if (preserving) {
                    if (orgF.getStrength() == unigramStrength) {
                        result.add(orgF);
                    } else {
                        // create copy of feature with specified strength
                        result.add(new DefaultFeature(orgF.getRepresentation(),
                                        orgF.getComment(), unigramStrength));
                    }
                }

                distance = priorFeatureReps.size();
                bufferIter = priorFeatureReps.iterator();

                // create feature pairs
                while (bufferIter.hasNext()) {
                    // reset string buffer + append prior feature from buffer
                    newRep.setLength(0);
                    newRep.append(bufferIter.next());

                    // append suitable number of separators
                    for (i = 0; i < distance; i++) {
                        newRep.append(separator);
                    }

                    // decrement distance for next run and to look up the
                    // current strength (stored in array from 0 to N-1, while
                    // 1 to N separators are used)
                    distance--;
                    // append currently processed feature
                    newRep.append(orgRep);
                    // store result with specified strength
                    result.add(new DefaultFeature(newRep.toString(), null,
                                    strengths[distance]));
                }

                // store processed feature
                priorFeatureReps.add(orgRep);
            }
        }

/*        // tracing for Bill
        Util.LOG.debug("Input features: " + orgFeatures.size());
        Util.LOG.debug("OSB features: " + result.size()); */

        return result;
    }

    /**
     * Returns the maximum number of original features joined.
     * @return the value of the attribute
     */
    public int getLength() {
        return length;
    }

    /**
     * Returns the string used to separate original features (by default a space
     * character). This string <strong>should never occur</strong> within
     * original features.
     *
     * @return the value of the attribute
     */
    public String getSeparator() {
        return separator;
    }

    /**
     * Whether original features are preserved as well in addition to the
     * generated joint features.
     *
     * @return the value of the attribute
     */
    public boolean isPreserving() {
        return preserving;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder builder = new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("length", length)
            .append("separator", separator)
            .append("preserving originals", preserving)
            .append("strengths", strengths);

        if (preserving) {
            builder.append("unigram strength", unigramStrength);
        }
        return builder.toString();
    }

}
