/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.feature;

import java.util.Iterator;

import org.apache.commons.collections.buffer.CircularFifoBuffer;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.TiesConfiguration;

/**
 * Transforms a feature vector using a simple implementation of the <em>sparse
 * binary polynomial hashing (SBPH)</em> technique introduced by
 * <a href="http://crm114.sourceforge.net/">CRM114</a>. This transformer
 * discard all comment-only features (indeed all comments). It slides of window
 * of {@linkplain #getLength() length <em>N</em>} over the remaining original
 * features. For each window position, it counts in binary from 1 to
 * 2<sup><em>N</em></sup>. For each odd number, a joint feature is generated
 * where original features at "1" positions are visible and original features at
 * "0" positions are hidden. Separators prior to the first feature are
 * discarded, but all inner separators are kept. E.g. if <em>N</em>=3 and a
 * pipe character "|" is used as {@linkplain #getSeparator() separator}, from
 * the original features "a", "b", "c", four joint features will be generated
 * at the last position: "c" (binary 1=001), "b|c" (binary 3=011), "a||c"
 * (binary 5=101), "a|b|c" (binary 7=111). Thus 2<sup><em>N</em>-1</sup>
 * joint features are generated for each original (non-comment) feature (except
 * for the very first features).
 *
 * <p>Instances of this class are thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/08/23 17:08:21 $, $Author: siefkes $
 */
public class SBPHTransformer extends FeatureTransformer {

    /**
     * The {@linkplain #getSeparator() separator} used by default (a space
     * character).
     */
    public static final String DEFAULT_SEPARATOR = " ";

    /**
     * The maximum number of original features joined.
     */
    private final int length;

    /**
     * The string used to separate original features (by default a space
     * character). This string <strong>should never occur</strong> within
     * original features.
     */
    private final String separator;

    /**
     * Stores bit values up to {@linkplain #getLength() length <em>N</em>},
     * for bit operations: 1, 2, 4, 8, 16 etc. The <em>n</em>-th element in this
     * vector contains the value 2<sup>n+1</sup>.
     */
    private final int[] bitValues;

    /**
     * The maximum number of new features to generate for each position:
     * 2<sup>(<em>{@linkplain #getLength() N} - 1</em>)</sup>.
     */
    private final int maxNumber;

    /**
     * Creates a new instance.
     *
     * @param precTrans the preceding transformer to use if this transformer
     * is part of a <em>chain</em>; <code>null</code> otherwise
     * @param len the maximum number of original features joined
     * @param sep the string used to separate original features -- this string
     * <strong>should never occur</strong> within original features
     */
    public SBPHTransformer(final FeatureTransformer precTrans, final int len,
            final String sep) {
        super(precTrans);
        length = len;
        separator = sep;

        // pre-calculate values for bit operations (1, 2, 4, 8...)
        bitValues = new int[length];
        int bitVal = 1;

        for (int i = 0; i < length; i++) {
            bitValues[i] = bitVal;
            bitVal *= 2;
        }

        // calculate number of features to generate at each position
        maxNumber = numFeatures(length);
    }

    /**
     * Creates a new instance.
     *
     * @param precTrans the preceding transformer to use if this transformer
     * is part of a <em>chain</em>; <code>null</code> otherwise
     * @param config used to configure this instance
     */
    public SBPHTransformer(final FeatureTransformer precTrans,
            final TiesConfiguration config) {
        this(precTrans, config.getInt("transformer.sbph.length"),
            config.getString("transformer.sbph.separator", DEFAULT_SEPARATOR));
    }

    /**
     * {@inheritDoc}
     */
    protected FeatureVector doTransform(final FeatureVector orgFeatures) {
        final FeatureVector result = new DefaultFeatureVector();
        final Iterator orgIter = orgFeatures.iterator();
        final CircularFifoBuffer lastNFeatureReps =
            new CircularFifoBuffer(length);
        Iterator bufferIter;
        int bufferSize;
        String orgRep;
        Feature orgF;
        StringBuffer[] newReps;
        int numNew;
        int currentBit;
        int i;

        while (orgIter.hasNext()) {
            orgF = (Feature) orgIter.next();
            orgRep = orgF.getRepresentation();

            // ignore comment-only features, keep only representations
            if (orgRep != null) {
                lastNFeatureReps.add(orgRep);
                bufferSize = lastNFeatureReps.size();

                // how many features to generate?
                if (bufferSize == length) {
                    // use pre-calculated value
                    numNew = maxNumber;
                } else {
                    numNew = numFeatures(bufferSize);
                }

                // count bits from bufferSize down to 1
                currentBit = bitValues[bufferSize - 1];
                newReps = new StringBuffer[numNew];
                bufferIter = lastNFeatureReps.iterator();
                boolean bufferHasNext = bufferIter.hasNext();

                // create polynomial feature representations, all in parallel
                while (bufferHasNext) {
                    orgRep = (String) bufferIter.next();

                    for (i = 0; i < newReps.length; i++) {
                    // i-th buffer corresponds to binary number 2i+1
                        if ((((2 * i) + 1) & currentBit) > 0) {
                            // bit is set: add feature representation
                            if (newReps[i] == null) {
                                newReps[i] = new StringBuffer(orgRep);
                            } else {
                                newReps[i].append(orgRep);
                            }
                        }

                        bufferHasNext = bufferIter.hasNext();

                        // add separator, except prior to first included feature
                        // and after last feature
                        if ((newReps[i] != null) && bufferHasNext) {
                            newReps[i].append(separator);
                        }
                    }

                    currentBit /= 2;
                }

                // append created features to result vector
                for (i = 0; i < newReps.length; i++) {
                    result.add(new DefaultFeature(newReps[i].toString()));
                }
            }
        }
        return result;
    }

    /**
     * Returns the maximum number of original features joined.
     * @return the value of the attribute
     */
    public int getLength() {
        return length;
    }

    /**
     * Returns the string used to separate original features (by default a space
     * character). This string <strong>should never occur</strong> within
     * original features.
     *
     * @return the value of the attribute
     */
    public String getSeparator() {
        return separator;
    }

    /**
     * Returns the number of jointed features to generate for a given number of
     * features to join.
     *
     * @param n the number of features to join
     * @return 2<sup>(<em>n</em>-1)</sup>
     */
    private int numFeatures(final int n) {
        return (int) Math.pow(2, n - 1);
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("length", length)
            .append("separator", separator)
            .toString();
    }

}
