/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.winnow;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.feature.FeatureTransformer;

/**
 * A combination of Winnow with the "ultraconservative" approach proposed by
 * Koby Crammer and Yoram Singer. In case of a loss (mistake) the target class
 * is promoted and all classes whose score is greater than or equal to the
 * target class (the "error set") are demoted.
 *
 * <p>Instances of this class are thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.12 $, $Date: 2004/09/06 17:22:20 $, $Author: siefkes $
 */
public class UltraconservativeWinnow extends Winnow {

    /**
     * Optional prefix used to give values for the configuration
     * parameters of this classifier that differ from the values used by Winnow.
     */
    public static final String UC_SUFFIX = "uc";

    /**
     * Creates a new instance by delegating to the corresponding super
     * constructor.
     *
     * @param allValidClasses the set of all valid classes
     * @throws IllegalArgumentException if one of the parameters is outside
     * the allowed range
     * @throws ProcessingException if an error occurred while creating
     * the feature transformer(s)
     */
    public UltraconservativeWinnow(final Set<String> allValidClasses)
    throws IllegalArgumentException, ProcessingException {
        super(allValidClasses, UC_SUFFIX);
    }

    /**
     * Creates a new instance by delegating to the corresponding super
     * constructor.
     *
     * @param allValidClasses the set of all valid classes
     * @param config contains configuration properties
     * @throws IllegalArgumentException if one of the parameters is outside
     * the allowed range
     * @throws ProcessingException if an error occurred while creating
     * the feature transformer(s)
     */
    public UltraconservativeWinnow(final Set<String> allValidClasses,
            final TiesConfiguration config)
    throws IllegalArgumentException, ProcessingException {
        super(allValidClasses, config, UC_SUFFIX);
    }

    /**
     * Creates a new instance by delegating to the corresponding super
     * constructor.
     *
     * @param allValidClasses the set of all valid classes
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param config contains configuration properties
     * @throws IllegalArgumentException if one of the parameters is outside
     * the allowed range
     * @throws ProcessingException if an error occurred while creating
     * the feature transformer(s)
     */
    public UltraconservativeWinnow(final Set<String> allValidClasses,
            final FeatureTransformer trans, final TiesConfiguration config)
    throws IllegalArgumentException, ProcessingException {
        super(allValidClasses, trans, config, UC_SUFFIX);
    }

    /**
     * Creates a new instance by delegating to the corresponding super
     * constructor.
     *
     * @param allValidClasses the set of all valid classes
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param balance whether to use the Balanced Winnow or the standard
     * Winnow algorithm
     * @param promotionFactor the promotion factor used by the algorithm
     * @param demotionFactor the demotion factor used by the algorithm
     * @param thresholdThick the thickness of the threshold if the "thick
     * threshold" heuristic is used (must be &lt; 1.0), 0.0 otherwise
     * @param config contains configuration properties
     * @param configSuffix optional suffix appended to the configuration keys
     * when configuring this instance; might be <code>null</code>
     * @throws IllegalArgumentException if one of the parameters is outside
     * the allowed range
     */
    public UltraconservativeWinnow(final Set<String> allValidClasses,
            final FeatureTransformer trans, final boolean balance,
            final float promotionFactor, final float demotionFactor,
            final float thresholdThick, final TiesConfiguration config,
            final String configSuffix) throws IllegalArgumentException {
        super(allValidClasses, trans, balance, promotionFactor, demotionFactor,
            thresholdThick, config, configSuffix);
    }

    /**
     * Chooses the classes to promote and the classes to demote. This
     * "ultraconservative" implementation demotes all classes whose score is
     * greater than or equal to the target class (the "error set") and promotes
     * the target class if the error set is not empty.
     *
     * @param winnowDist the prediction distribution returned by
     * {@link de.fu_berlin.ties.classify.TrainableClassifier#classify}
     * @param targetClass the expected class of this instance; must be
     * contained in the set of <code>candidateClasses</code>
     * @param classesToPromote the classes to promote are added to this set
     * @param classesToDemote the classes to demote are added to this set
     */
    protected void chooseClassesToAdjust(final WinnowDistribution winnowDist,
            final String targetClass, final Set<String> classesToPromote,
            final Set<String> classesToDemote) {
        final Iterator predIter = winnowDist.iterator();
        WinnowPrediction pred;
        float dynamicThreshold = Float.NaN;
        boolean goOn = true;
        final Set<String> errorSet = new HashSet<String>();

        // classes are sorted according to probability/score in decreasing order
        while (predIter.hasNext() && goOn) {
            pred = (WinnowPrediction) predIter.next();

            if (targetClass.equals(pred.getType())) {
                // this is the target class: remember score
                final float scoreOfTarget = pred.getRawScore();

                // use thick threshold heuristic if configured, dynamically
                // calculated from the target score
                dynamicThreshold =
                    minorThreshold(scoreOfTarget, winnowDist.getRawThreshold());
            } else {
                if (!Float.isNaN(dynamicThreshold)
                        && (pred.getRawScore() < dynamicThreshold)) {
                    // class with lower score: break loop
                    goOn = false;
                } else {
                    // score is higher than or equal to dynamic threshold
                    errorSet.add(pred.getType());
                }
            }
        }

        if (!errorSet.isEmpty()) {
            // promote target class because of erroneous classifications
            classesToPromote.add(targetClass);
            classesToDemote.addAll(errorSet);
        }
    }

}
