/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.winnow;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.PredictionDistribution;
import de.fu_berlin.ties.classify.TrainableClassifier;
import de.fu_berlin.ties.classify.feature.Feature;
import de.fu_berlin.ties.classify.feature.FeatureSet;
import de.fu_berlin.ties.classify.feature.FeatureTransformer;
import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.util.Util;

/**
 * Classifier implementing the Winnow algorithm (Nick Littlestone). <b>Winnow
 * supports <em>only</em> error-driven training, so you always have to use the
 * {@link #trainOnError(FeatureVector, String, Set)} method. Trying to
 * call the {@link
 * de.fu_berlin.ties.classify.TrainableClassifier#train(FeatureVector, String)}
 * method instead will result in an
 * {@link java.lang.UnsupportedOperationException}.</b>
 *
 * <p>Instances of this class are thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.34 $, $Date: 2004/11/17 09:16:27 $, $Author: siefkes $
 */
public class Winnow extends TrainableClassifier {

    /**
     * Configuration key: How feature frequencies are considered when
     * calculating strength values.
     */
    private static final String CONFIG_STRENGTH_METHOD =
        "classifier.winnow.strength.frequency";

    /**
     * Whether the Balanced Winnow or the standard Winnow algorithm is used.
     * Balanced Winnow keeps <code>two</code> weights per feature and class,
     * a positive and a negative one.
     */
    private final boolean balanced;

    /**
     * The promotion factor used by the algorithm.
     */
    private final float promotion;

    /**
     * The demotion factor used by the algorithm.
     */
    private final float demotion;

    /**
     * The thickness of the threshold if the "thick threshold" heuristic is used
     * (must be &lt; 1.0), 0.0 otherwise.
     */
    private final float thresholdThickness;

    /**
     * Stores the feature weights, using a variation of the LRU mechanism for
     * pruning surplus features. Every access should be synchronized on
     * <strong>this</strong>.
     */
    private final WinnowStore store;

    /**
     * Creates a new instance based on the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @param allValidClasses the set of all valid classes
     * @throws IllegalArgumentException if one of the parameters is outside
     * the allowed range
     * @throws ProcessingException if an error occurred while creating
     * the feature transformer(s)
     */
    public Winnow(final Set<String> allValidClasses)
    throws IllegalArgumentException, ProcessingException {
        this(allValidClasses, (String) null);
    }

    /**
     * Creates a new instance based on the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @param allValidClasses the set of all valid classes
     * @param configSuffix optional suffix appended to the configuration keys
     * when configuring this instance; might be <code>null</code>
     * @throws IllegalArgumentException if one of the parameters is outside
     * the allowed range
     * @throws ProcessingException if an error occurred while creating
     * the feature transformer(s)
     */
    protected Winnow(final Set<String> allValidClasses, final String configSuffix)
    throws IllegalArgumentException, ProcessingException {
        this(allValidClasses, TiesConfiguration.CONF, configSuffix);
    }

    /**
     * Creates a new instance based on the provided configuration.
     *
     * @param allValidClasses the set of all valid classes
     * @param config contains configuration properties
     * @throws IllegalArgumentException if one of the parameters is outside
     * the allowed range
     * @throws ProcessingException if an error occurred while creating
     * the feature transformer(s)
     */
    public Winnow(final Set<String> allValidClasses,
            final TiesConfiguration config)
    throws IllegalArgumentException, ProcessingException {
        this(allValidClasses, config, null);
    }

    /**
     * Creates a new instance based on the provided configuration.
     *
     * @param allValidClasses the set of all valid classes
     * @param config contains configuration properties
     * @param configSuffix optional suffix appended to the configuration keys
     * when configuring this instance; might be <code>null</code>
     * @throws IllegalArgumentException if one of the parameters is outside
     * the allowed range
     * @throws ProcessingException if an error occurred while creating
     * the feature transformer(s)
     */
    protected Winnow(final Set<String> allValidClasses,
            final TiesConfiguration config, final String configSuffix)
    throws IllegalArgumentException, ProcessingException {
        this(allValidClasses, FeatureTransformer.createTransformer(config),
            config, configSuffix);
    }

    /**
     * Creates a new instance based on the provided configuration.
     *
     * @param allValidClasses the set of all valid classes
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param config contains configuration properties
     * @throws IllegalArgumentException if one of the parameters is outside
     * the allowed range
     * @throws ProcessingException if an error occurred while creating
     * the feature transformer(s)
     */
    public Winnow(final Set<String> allValidClasses,
            final FeatureTransformer trans, final TiesConfiguration config)
    throws IllegalArgumentException, ProcessingException {
        this(allValidClasses, trans, config, null);
    }

    /**
     * Creates a new instance based on the provided configuration.
     *
     * @param allValidClasses the set of all valid classes
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param config contains configuration properties
     * @param configSuffix optional suffix appended to the configuration keys
     * when configuring this instance; might be <code>null</code>
     * @throws IllegalArgumentException if one of the parameters is outside
     * the allowed range
     * @throws ProcessingException if an error occurred while creating
     * the feature transformer(s)
     */
    protected Winnow(final Set<String> allValidClasses,
            final FeatureTransformer trans, final TiesConfiguration config,
            final String configSuffix)
    throws IllegalArgumentException, ProcessingException {
        this(allValidClasses, trans,
            config.getBoolean(
                config.adaptKey("classifier.winnow.balanced", configSuffix)),
            config.getFloat(
                config.adaptKey("classifier.winnow.promotion", configSuffix)),
            config.getFloat(
                config.adaptKey("classifier.winnow.demotion", configSuffix)),
            config.getFloat(config.adaptKey(
                "classifier.winnow.threshold.thickness", configSuffix)),
            config, configSuffix);
    }

    /**
     * Creates a new instance.
     *
     * @param allValidClasses the set of all valid classes
     * @param trans the last transformer in the transformer chain to use, or
     * <code>null</code> if no feature transformers should be used
     * @param balance whether to use the Balanced Winnow or the standard
     * Winnow algorithm
     * @param promotionFactor the promotion factor used by the algorithm;
     * must be &gt; 1.0
     * @param demotionFactor the demotion factor used by the algorithm; must
     * be &lt; 1.0
     * @param thresholdThick the thickness of the threshold if the "thick
     * threshold" heuristic is used (must be &lt; 1.0), 0.0 otherwise
     * @param config contains configuration properties
     * @param configSuffix optional suffix appended to the configuration keys
     * when configuring this instance; might be <code>null</code>
     * @throws IllegalArgumentException if one of the parameters is outside
     * the allowed range
     */
    public Winnow(final Set<String> allValidClasses,
            final FeatureTransformer trans, final boolean balance,
            final float promotionFactor, final float demotionFactor,
            final float thresholdThick, final TiesConfiguration config,
            final String configSuffix) throws IllegalArgumentException {
        // wrap classes in a tree set to ensure that they are sorted
        super(new TreeSet<String>(allValidClasses), trans, config);

        // check arguments
        if ((promotionFactor <= 1.0)) {
            throw new IllegalArgumentException("Promotion factor must be > 1: "
                + promotionFactor);
        }
        if ((demotionFactor >= 1.0) || (demotionFactor <= 0.0)) {
            throw new IllegalArgumentException(
                "Demotion factor must be in ]0, 1[ range:" + demotionFactor);
        }
        if ((thresholdThick >= 1.0) || (thresholdThick < 0.0)) {
            throw new IllegalArgumentException("Threshold thickness must be "
                + "in [0, 1[ range: " + thresholdThick);
        }

        // store arguments + initialize feature cache of specified size
        balanced = balance;
        promotion = promotionFactor;
        demotion = demotionFactor;
        thresholdThickness = thresholdThick;
        store = new WinnowStore(initWeight(), config, configSuffix);
    }

    /**
     * Adjusts the weights of a feature for all classes. This method should be
     * called in a synchronized context.
     *
     * @param feature the feature to process
     * @param directions an array specifying for each class (in alphabetic
     * order) whether it should be promoted (positive value), demoted (negative
     * value) or left unmodified (0)
     */
    protected void adjustWeights(final Feature feature,
            final short[] directions) {
        // use hashcode of feature to retrieve weights
        final Integer featureHash = new Integer(feature.hashCode());
        float[] weights = store.getWeights(featureHash);
        final int length = getAllClasses().size();

        // check argument
        if (directions.length != length) {
            throw new IllegalArgumentException("Array of directions has "
                + directions.length + " members instead of one for each of the "
                + length + " classes");
        }

        if (weights == null) {
            // initialize and store weights
            weights = initWeightArray();
            store.putWeights(featureHash, weights);
        }

        // promote/demote (positive) weights as specified
        // (stored in fields from 0 to n-1)
        for (int i = 0; i < length; i++) {
            if (directions[i] < 0) {
                // demote (positive) weight
                weights[i] *= demotion;

                if (balanced) {
                    // promote negative weight
                    weights[i + length] *= promotion;
                }
            } else if (directions[i] > 0) {
                // promote positive weight
                weights[i] *= promotion;

                if (balanced) {
                    // demote negative weight
                    weights[i + length] *= demotion;
                }
            }
            // otherwise (0) there is nothing to do
        }
    }

    /**
     * Chooses the classes to promote and the classes to demote. This class
     * chooses the <code>targetClass</code> for promotion if its score is
     * less or equal to the {@linkplain #threshold(float) threshold}.
     * It chooses all other classes for demotion if their score is greather
     * than the threshold.
     *
     * @param winnowDist the prediction distribution returned by
     * {@link #classify(FeatureVector, Set)}
     * @param targetClass the expected class of this instance; must be
     * contained in the set of <code>candidateClasses</code>
     * @param classesToPromote the classes to promote are added to this set
     * @param classesToDemote the classes to demote are added to this set
     */
    protected void chooseClassesToAdjust(final WinnowDistribution winnowDist,
            final String targetClass, final Set<String> classesToPromote,
            final Set<String> classesToDemote) {
        final Iterator predIter = winnowDist.iterator();
        WinnowPrediction pred;

        // use thick threshold heuristic if configured
        final float minorThreshold = minorThreshold(winnowDist.getThreshold(),
            winnowDist.getRawThreshold());
        final float majorThreshold = majorThreshold(winnowDist.getThreshold(),
            winnowDist.getRawThreshold());

/*        // tracing for Bill
        Util.LOG.debug("Threshold=" + winnowDist.getThreshold()
                + " (minor=" + minorThreshold
                + " major=" + majorThreshold + ')'); */

        while (predIter.hasNext()) {
            pred = (WinnowPrediction) predIter.next();

            if (targetClass.equals(pred.getType())) {
                // this is the target class
                if (pred.getRawScore() <= majorThreshold) {
                    classesToPromote.add(pred.getType());
                }
            } else {
                if (pred.getRawScore() > minorThreshold) {
                    // this class should be demoted
                    classesToDemote.add(pred.getType());
                }
            }
        }
    }

    /**
     * Converts a {@linkplain #sigmoid(float, float, float) sigmoid activation
     * value} into a confidence estimate.
     *
     * @param sigmoid the {@linkplain #sigmoid(float, float, float) sigmoid
     * activation value} to convert
     * @param sum the sum of all sigmoid activation values
     * @return the estimated confidence: <code>sigmoid / sum</code>
     */
    protected double confidence(final float sigmoid, final float sum) {
        return (double) sigmoid / sum;
    }

    /**
     * Returns the default weight to use if a feature is unknown. This
     * implementation returns 0.0 in case of {@link #isBalanced() Balanced
     * Winnow} (where positive and negative weights should cancel each other
     * out), {@link #initWeight()} otherwise.
     *
     * @return the default weight
     */
    protected float defaultWeight() {
        if (balanced) {
            return 0.0f;
        } else {
            return initWeight();
        }
    }

    /**
     * {@inheritDoc}
     */
    protected PredictionDistribution doClassify(final FeatureVector features,
            final Set candidateClasses, final ContextMap context) {
        // convert to (multi-)set if required
        final FeatureSet featureSet = featureSet(features);
        final float[] scores = initScores();
        final Iterator featureIter = featureSet.iterator();
        Feature currentFeature;

/*        // tracing for Bill
        Util.LOG.debug("Unique OSB features: " + featureSet.size());
        Util.LOG.debug("Stored features: " + store.size()); */

        synchronized (this) {
            // iterate over all feature to calculate score for each class
            while (featureIter.hasNext()) {
                currentFeature = (Feature) featureIter.next();
                updateScores(currentFeature, features.strength(currentFeature),
                        scores);
            }
        }

        // calculate threshold and sigmoid activation values
        final float rawThreshold = rawThreshold(featureSet);
        final float threshold = threshold(rawThreshold);
        final float[] sigmoids = new float[scores.length];
        float sigmoidSum = 0.0f;
        int i;

        for (i = 0; i < scores.length; i++) {
            sigmoids[i] = sigmoid(scores[i], threshold, rawThreshold);
            sigmoidSum += sigmoids[i];
        }

        // create winnow distribution to store results
        final WinnowDistribution result =
            new WinnowDistribution(threshold, rawThreshold);

        // add WinnowPredictions in alphabetic order; the will be automatically
        // sorted according to the confidence values
        final Iterator classIter = candidateClasses.iterator();
        String className;
        i = 0;

        while (classIter.hasNext()) {
            className = (String) classIter.next();
            result.add(new WinnowPrediction(className,
                confidence(sigmoids[i], sigmoidSum), scores[i], sigmoids[i]));
            i++;
        }

/*        // tracing for Bill
        final StringBuilder scoreTrace = new StringBuilder("Scores:");
        final Iterator predIter = result.iterator();
        WinnowPrediction pred;
        while (predIter.hasNext()) {
            pred = (WinnowPrediction) predIter.next();
            scoreTrace.append(' ' + pred.getType() + '=' + pred.getRawScore());            
        }
        Util.LOG.debug(scoreTrace.toString()); */

        return result;
    }

    /**
     * <b>Winnow supports <em>only</em> error-driven training, so you always
     * have to use the {@link #trainOnError(FeatureVector, String, Set)} method
     * instead of this one. Trying to call this method instead will result in an
     * {@link java.lang.UnsupportedOperationException}.</b>
     *
     * @param features ignored by this method
     * @param targetClass ignored by this method
     * @param context ignored by this method
     * @throws UnsupportedOperationException always thrown by this method;
     * use {@link #trainOnError(FeatureVector, String, Set)} instead
     */
    protected void doTrain(final FeatureVector features,
            final String targetClass, final ContextMap context)
    throws UnsupportedOperationException {
        // we cannot support this method because we need to know all
        // candidate classes for updating
        throw new UnsupportedOperationException("Winnow supports only "
            + "error-driven training -- call trainOnError instead of train");
    }

    /**
     * Converts a feature vector into a {@link FeatureSet} (a multi-set of
     * features). If the provided vector already is a <code>FeatureSet</code>
     * instance, it is casted and returned. Otherwise a new
     * <code>FeatureSet</code> with the same contents is created, reading the
     * used method for considering feature frequencies in strength values
     * from the "classifier.winnow.strength.frequency" configuration key.
     *
     * @param fv the feature vector to convert
     * @return a feature set with the same contents as the provided vector
     */
    protected FeatureSet featureSet(final FeatureVector fv) {
        final FeatureSet result;

        if (fv instanceof FeatureSet) {
            result = (FeatureSet) fv;
        } else {
            result = new FeatureSet(getConfig().getString(
                CONFIG_STRENGTH_METHOD));
            result.addAll(fv);
        }

        return result;
    }

    /**
     * Returns the promotion factor used by the algorithm.
     *
     * @return the value of the attribute
     */
    public float getDemotion() {
        return demotion;
    }

    /**
     * Returns the demotion factor used by the algorithm.
     *
     * @return the value of the attribute
     */
    public float getPromotion() {
        return promotion;
    }

    /**
     * Whether the Balanced Winnow or the standard Winnow algorithm is used.
     * Balanced Winnow keeps <em>two</em> weights per feature and class,
     * a positive and a negative one.
     *
     * @return the value of the attribute
     */
    public boolean isBalanced() {
        return balanced;
    }

    /**
     * Initializes the score (activation values) to use for all classes.
     *
     * @return an array of floats containing the initial score for each class;
     * the value of each float will be 0.0
     */
    protected float[] initScores() {
        // each float is automatically initialized to 0.0
        final float[] result = new float[getAllClasses().size()];
        return result;
    }

    /**
     * Returns the thickness of the threshold if the "thick threshold"
     * heuristic is used.
     *
     * @return the value of the attribute, will be &lt; 1.0; 0.0 if no
     * thick threshold is used
     */
    public float getThresholdThickness() {
        return thresholdThickness;
    }

    /**
     * Returns the initial weight to use for each feature per class. This
     * implementation returns 1.0.
     *
     * @return the initial weight
     */
    protected float initWeight() {
        return 1.0f;
    }

    /**
     * Returns the initial weight array to use for a feature for all classes.
     * The array returns by this implementation fill contain one weight for
     * each class in case of normal Winnow, two weights in case of
     * {@link #isBalanced() Balanced} Winnow. Each element is initialized to
     * {@link #initWeight()}.
     *
     * @return the initial weight array
     */
    protected float[] initWeightArray() {
        final float[] result;

        if (balanced) {
            // store positive and negatives weight for each class
            result = new float[getAllClasses().size() * 2];
        } else {
            // store weight for each class
            result = new float[getAllClasses().size()];
        }

        // initialize each value
        final float initWeight = initWeight();
        for (int i = 0; i < result.length; i++) {
            result[i] = initWeight;
        }

        return result;
    }

    /**
     * Calculates the major theshold (<em>theta+</em>) to use for classification
     * with the "thick threshold" heuristic. This
     * implementation multiplies <em>theta<sub>r</sub></em> with the
     * {@linkplain #getThresholdThickness() threshold thickness} and adds
     * the result to <em>theta</em>. Subclasses can overwrite this method to
     * calculate the major theshold in a different way.
     *
     * @param threshold the {@linkplain #threshold(float) threshold}
     * <em>theta</em>
     * @param rawThreshold the {@linkplain #rawThreshold(FeatureSet) raw
     * threshold} <em>theta<sub>r</sub></em>
     * @return the major theshold (<em>theta+</em>) to use for classification
     * @see #minorThreshold(float, float)
     */
    protected float majorThreshold(final float threshold,
            final float rawThreshold) {
       final float result =  threshold + getThresholdThickness() * rawThreshold;
       return result;
    }

    /**
     * Calculates the minor theshold (<em>theta-</em>) to use for classification
     * with the "thick threshold" heuristic. This
     * implementation multiplies <em>theta<sub>r</sub></em> with the
     * {@linkplain #getThresholdThickness() threshold thickness} and subtracts
     * the result from <em>theta</em>. Subclasses can overwrite this method to
     * calculate the minor theshold in a different way.
     *
     * @param threshold the {@linkplain #threshold(float) threshold}
     * <em>theta</em>
     * @param rawThreshold the {@linkplain #rawThreshold(FeatureSet) raw
     * threshold} <em>theta<sub>r</sub></em>
     * @return the minor theshold (<em>theta-</em>) to use for classification
     * @see #majorThreshold(float, float)
     */
    protected float minorThreshold(final float threshold,
            final float rawThreshold) {
        final float result = threshold - getThresholdThickness() * rawThreshold;
        return result;
    }

    /**
     * Calculates the theshold (theta) to use for classification, based on the
     * number of active features. This implementation returns the
     * {@linkplain FeatureVector#getSummedStrength() summed strength} of all
     * features. Subclasses can overwrite this method to calculate the theshold
     * in a different way.
     *
     * @param features the feature set to consider
     * @return the raw theshold (theta) to use
     */
    protected float rawThreshold(final FeatureSet features) {
        return (float) features.getSummedStrength();
    }

    /**
     * {@inheritDoc}
     */
    public void reset() {
        store.reset();
    }

    /**
     * Converts the raw <em>score</em> (activation value) to a value in the
     * range from 0 to 1 via a sigmoid function depending on the threshold
     * <em>theta</em>. In this implementation this is calculed as follows:
     *
     * <ul>
     * <li>Normal Winnow (not balanced): <em>sigma</em>(<em>score</em>,
     * <em>theta</em>) = 1 / (1 + (<em>theta</em> / <em>score</em>))</li>
     * <li>{@linkplain #isBalanced() Balanced} Winnow:
     * <em>sigma</em>(<em>score</em>, <em>theta</em>,
     * <em>theta<sub>r</sub></em>) = 1 / (1 +
     * e^((<em>theta</em> - <em>score</em>) / <em>theta<sub>r</sub></em>))</li>
     * </ul>
     *
     * @param score the raw <em>score</em> (activation value); must be a
     * positive value in case of normal (non-balanced) Winnow
     * @param threshold the {@linkplain #threshold(float) threshold}
     * <em>theta</em> used for this instance
     * @param rawThreshold the {@linkplain #rawThreshold(FeatureSet) raw
     * threshold} <em>theta<sub>r</sub></em> used for this instance
     * @return the sigmoid score calculated as described above; will be in range
     * from 0 to 1
     * @throws IllegalArgumentException if normal Winnow is used and
     * <code>score &lt;= 0</code>
     */
    protected float sigmoid(final float score, final float threshold,
            final float rawThreshold) throws IllegalArgumentException {
        if (balanced) {
            return (float) (1.0 / (1.0
                            + Math.exp((threshold - score) / rawThreshold)));
        } else {
            // ensure that the score is positive
            if (score <= 0.0f) {
                throw new IllegalArgumentException(
                    "Activation value must be positive in normal Winnow: "
                    + score);
            }

            return 1.0f / (1.0f + (threshold / score));
        }
    }

    /**
     * Calculates the theshold (theta) to use for classification. This
     * implementation returns the <code>rawThreshold</code> multiplied with
     * the {@linkplain #defaultWeight() default weight}. Subclasses can
     * overwrite this method to calculate the theshold in a different way.
     *
     * @param rawThreshold the {@linkplain #rawThreshold(FeatureSet) raw
     * threshold}
     * @return the theshold (theta) to use for classification
     */
    protected float threshold(final float rawThreshold) {
        return rawThreshold * defaultWeight();
    }

    /**
     * Hook implementing error-driven learning, promoting and demoting weights
     * as required.
     *
     * @param predDist the prediction distribution returned by
     * {@link #classify(FeatureVector, Set)}; must be a
     * {@link WinnowDistribution}
     * @param features the feature vector to consider
     * @param targetClass the expected class of this feature vector; must be
     * contained in the set of <code>candidateClasses</code>
     * @param candidateClasses an set of classes that are allowed for this item
     * (the actual <code>targetClass</code> must be one of them)
     * @param context ignored by this implementation
     * @return this implementation always returns <code>true</code> to signal
     * that any error-driven learning was already handled
     * @throws ProcessingException if an error occurs during training
     */
    protected boolean trainOnErrorHook(final PredictionDistribution predDist,
            final FeatureVector features, final String targetClass,
            final Set candidateClasses, final ContextMap context)
    throws ProcessingException {
        // convert to (multi-)set if required
        final FeatureSet featureSet = featureSet(features);

        final Set<String> classesToPromote = new HashSet<String>();
        final Set<String> classesToDemote = new HashSet<String>();

        // decide which classes to promote and demote
        chooseClassesToAdjust((WinnowDistribution) predDist, targetClass,
            classesToPromote, classesToDemote);

        if (!(classesToPromote.isEmpty() && classesToDemote.isEmpty())) {
            // there are class(es) to promote/demote
            Util.LOG.debug("Promoting classes: " + classesToPromote
                + "; demoting classes: " + classesToDemote);

            // iterate classes in alphabetic order to specify directions
            final short[] directions = new short[getAllClasses().size()];
            final Iterator classIter = getAllClasses().iterator();
            String currentClass;
            int i = 0;

            while (classIter.hasNext()) {
                currentClass = (String) classIter.next();
                if (classesToDemote.contains(currentClass)) {
                    // demote this class
                    directions[i] = -1;
                } else if (classesToPromote.contains(currentClass)) {
                    // promote this class
                    directions[i] = 1;
                } else {
                    // don't modify class
                    directions[i] = 0;
                }
                i++;
            }

            final Iterator featureIter = featureSet.iterator();
            Feature currentFeature;

            synchronized (this) {
                // iterate over all feature to promote/demote weights
                while (featureIter.hasNext()) {
                    currentFeature = (Feature) featureIter.next();
                    adjustWeights(currentFeature, directions);
                }
            }
        }

        // signal that we handled the training
        return true;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("balanced", balanced)
            .append("promotion", promotion)
            .append("demotion", demotion)
            .append("threshold thickness", thresholdThickness)
            .append("frequency-based strength",
                getConfig().getString(CONFIG_STRENGTH_METHOD))
            .append("feature store", store)
            .toString();
    }

    /**
     * Updates the score (activation values) for all classes by adding the
     * weights of a feature. This method should be called in a synchronized
     * context.
     *
     * @param feature the feature to process
     * @param strength the strength of this feature
     * @param scores an array of floats containing the scores for each
     * class; will be updated by this method
     */
    protected void updateScores(final Feature feature, final double strength,
                                final float[] scores) {
        // use hashcode of feature to retrieve weights
        final float[] weights =
            store.getWeights(new Integer(feature.hashCode()));
        final int length = getAllClasses().size();

        // check argument
        if (scores.length != length) {
            throw new IllegalArgumentException("Array of scores has "
                + scores.length + " members instead of one for each of the "
                + length + " classes");
        }

        if (weights != null) {
            // add (positive) weighted strength to each score
            // (weights stored in fields from 0 to n-1)
            for (int i = 0; i < length; i++) {
                scores[i] += weights[i] * strength;
            }

            if (balanced) {
                // Balanced Winnow: subtract negative weighted strengths
                // (weights stored in the fields from n to 2n-1)
                for (int i = 0; i < length; i++) {
                    scores[i] -= weights[i + length] * strength;
                }
            }
        } else {
            final float defaultWeight = defaultWeight();

            // add default weight * strength to each score
            if (defaultWeight != 0.0f) {
                for (int i = 0; i < length; i++) {
                    scores[i] += defaultWeight * strength;
                }
            }
        }
    }

}
