/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.winnow;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.classify.PredictionDistribution;

/**
 * A distribution over the classes predicted by a
 * {@link de.fu_berlin.ties.classify.winnow.Winnow Winnow classifier}.
 * Extends the superclass by also storing the
 * {@linkplain de.fu_berlin.ties.classify.winnow.Winnow#threshold(float)
 * threshold} and raw threshold used by Winnow for this instance.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.7 $, $Date: 2004/08/23 17:08:41 $, $Author: siefkes $
 */
public class WinnowDistribution extends PredictionDistribution {

    /**
     * The {@linkplain Winnow#threshold(float) threshold} used
     * by the Winnow classifier for this instance.
     */
    private final float threshold;

    /**
     * The {@linkplain Winnow#rawThreshold raw threshold} used
     * by the Winnow classifier for this instance.
     */
    private final float rawThreshold;

    /**
     * Creates a new instance.
     * @param theta the {@linkplain Winnow#threshold(float) threshold}
     * used by Winnow
     * @param rawTheta the {@linkplain Winnow#rawThreshold raw
     * threshold} used by Winnow
     */
    public WinnowDistribution(final float theta, final float rawTheta) {
        super();
        threshold = theta;
        rawThreshold = rawTheta;
    }

    /**
     * Returns the {@linkplain Winnow#rawThreshold raw threshold}
     * used by the Winnow classifier for this instance.
     * @return the value of the attribute
     */
    public float getRawThreshold() {
        return rawThreshold;
    }

    /**
     * Returns the {@linkplain Winnow#threshold(float) threshold}
     * used by the Winnow classifier for this instance.
     * @return the value of the attribute
     */
    public float getThreshold() {
        return threshold;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder builder = new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("threshold", threshold);

        // omit raw threshold if equal to threshold
        if (rawThreshold != threshold) {
            builder.append("raw threshold", rawThreshold);
        }
        return builder.toString();
    }

}
