/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.winnow;

import de.fu_berlin.ties.classify.Prediction;
import de.fu_berlin.ties.classify.Probability;
import de.fu_berlin.ties.eval.EvalStatus;
import de.fu_berlin.ties.io.FieldMap;
import de.fu_berlin.ties.util.Util;

/**
 * A prediction that also stores a raw score (activation value) and a sigmoid
 * score, as calculated by the {@link de.fu_berlin.ties.classify.winnow.Winnow}
 * algorithm. The {@linkplain de.fu_berlin.ties.classify.Probability#getPR() pR}
 * is never set by this class.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.8 $, $Date: 2004/11/19 14:04:28 $, $Author: siefkes $
 */
public class WinnowPrediction extends Prediction {

    /**
     * Serialization key for the predicted class.
     */
    public static final String KEY_RAW = "Raw Score";

    /**
     * Serialization key for the source identifier.
     */
    public static final String KEY_SIGMOID = "Sigmoid Score";

    /**
     * The raw score (activation value) of this prediction.
     */
    private final float rawScore;

    /**
     * The {@link Winnow#sigmoid(float, float, float) sigmoid score} (activation
     * value) of this prediction; a value in the range from 0 to 1.
     */
    private final float sigmoidScore;

    /**
     * Creates a new instance from a field map, fulfilling the
     * {@link de.fu_berlin.ties.io.Storable} contract.
     *
     * @param fieldMap map containing the serialized fields
     */
    public WinnowPrediction(final FieldMap fieldMap) {
        super(fieldMap);
        rawScore = Util.asFloat(fieldMap.get(KEY_RAW));
        sigmoidScore = Util.asFloat(fieldMap.get(KEY_SIGMOID));

        // check argument
        if ((sigmoidScore < 0.0) || (sigmoidScore > 1.0)) {
            throw new IllegalArgumentException(
                "Illegal sigmoid score (out of range from 0 to 1): "
                + sigmoidScore);
        }
    }

    /**
     * Creates a new instance, setting the evaluation status to
     * {@link EvalStatus#UNKNOWN}.
     *
     * @param predicted the predicted type
     * @param prob the probability of the prediction (must be in the range from
     * 0.0 to 1.0; or -1 if this is a confirmed prediction or an answer key)
     * @param raw the raw score (activation value) of this prediction
     * @param sigmoid the {@link Winnow#sigmoid(float, float, float) sigmoid
     * score} (activation value) of this prediction; must be a value in the
     * range from 0 to 1
     */
    public WinnowPrediction(final String predicted, final double prob,
            final float raw, final float sigmoid) {
        this(predicted, prob, raw, sigmoid, EvalStatus.UNKNOWN);
    }

    /**
     * Creates a new instance, without setting a source ID.
     *
     * @param predicted the predicted type
     * @param prob the probability of the prediction (must be in the range from
     * 0.0 to 1.0; or -1 if this is a confirmed prediction or an answer key)
     * @param raw the raw score (activation value) of this prediction
     * @param sigmoid the {@link Winnow#sigmoid(float, float, float) sigmoid
     * score} (activation value) of this prediction; must be a value in the
     * range from 0 to 1
     * @param status the {@linkplain EvalStatus evaluation status} of this
     * instance
     */
    public WinnowPrediction(final String predicted, final double prob,
            final float raw, final float sigmoid, final EvalStatus status) {
        this(predicted, null, prob, raw, sigmoid, status);
    }

    /**
     * Creates a new instance.
     *
     * @param predicted the predicted type
     * @param sourceID an identifier of the source of this prediction (e.g., the
     * file name), might be <code>null</code>
     * @param prob the probability of the prediction (must be in the range from
     * 0.0 to 1.0; or -1 if this is a confirmed prediction or an answer key)
     * @param raw the raw score (activation value) of this prediction
     * @param sigmoid the {@link Winnow#sigmoid(float, float, float) sigmoid
     * score} (activation value) of this prediction; must be a value in the
     * range from 0 to 1
     * @param status the {@linkplain EvalStatus evaluation status} of this
     * instance
     */
    public WinnowPrediction(final String predicted, final String sourceID,
            final double prob, final float raw, final float sigmoid,
            final EvalStatus status) {
        super(predicted, sourceID, new Probability(prob), status);

        // check argument
        if ((sigmoid < 0.0) || (sigmoid > 1.0)) {
            throw new IllegalArgumentException(
                "Illegal sigmoid score (out of range from 0 to 1): " + sigmoid);
        }

        rawScore = raw;
        sigmoidScore = sigmoid;
    }

    /**
     * Returns the raw score (activation value) of this prediction.
     * @return the value of the attribute
     */
    public float getRawScore() {
        return rawScore;
    }

    /**
     * Returns the {@link Winnow#sigmoid(float, float, float) sigmoid score}
     * (activation value) of this prediction.
     *
     * @return the value of the attribute; a value in the range from 0 to 1
     */
    public float getSigmoidScore() {
        return sigmoidScore;
    }

    /**
     * Stores all relevant fields of this object in a field map for
     * serialization. An equivalent object can be created by calling
     * {@link de.fu_berlin.ties.io.FieldMap#createObject(Class)} on the created
     * field map.
     *
     * @return the created field map
     */
    public FieldMap storeFields() {
        final FieldMap result = super.storeFields();
        result.put(KEY_RAW, new Float(rawScore));
        result.put(KEY_SIGMOID, new Float(sigmoidScore));
        return result;
    }

}
