/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.classify.winnow;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.util.AdaptableLRUMap;
import de.fu_berlin.ties.util.MultiValueMap;
import de.fu_berlin.ties.util.Pruner;

/**
 * Feature store used by {@link Winnow}. Controls the number of stored
 * features and implements pruning based on the LRU (least recently used)
 * principle.
 *
 * <p>Instances of this class are synchronized externally by Winnow.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.15 $, $Date: 2004/12/06 17:58:06 $, $Author: siefkes $
 */
public class WinnowStore implements Pruner {

    /**
     * The initial weight of each feature. This implementation prunes
     * the one of the candidate features whose weights deviate least from this
     * initial weight.
     */
    private final float initWeight;

    /**
     * Map storing the feature weights. Uses the hash codes (Integer) of
     * features as keys and an array of floats storing the weight for each
     * class as values.
     */
    private final AdaptableLRUMap store;

    /**
     * Creates a new instance.
     *
     * @param initialWeight The initial weight of each feature -- this
     * implementation prunes features whose weights deviate least from this
     * initial weight
     * @param config Used to configure this instance
     * @param configSuffix Optional suffix appended to the configuration keys
     * when configuring this instance; might be <code>null</code>
     */
    public WinnowStore(final float initialWeight,
                       final TiesConfiguration config,
                       final String configSuffix) {
        this(initialWeight,
             config.getInt(config.adaptKey(
                                "classifier.winnow.features", configSuffix)),
             config.getInt(config.adaptKey(
                                 "prune.candidates", configSuffix)),
             config.getInt(config.adaptKey(
                                 "prune.num", configSuffix)));
    }

    /**
     * Creates a new instance.
     *
     * @param initialWeight The initial weight of each feature -- this
     * implementation prunes features whose weights deviate least from this
     * initial weight
     * @param featureNum The number of features to store
     * @param candidates The number of candidates to consider for each pruning
     * operation
     * @param pruneNum The number of elements to remove by each pruning
     * operation, must not be larger than <code>candidates</code>
     */
    public WinnowStore(final float initialWeight, final int featureNum,
                       final int candidates, final int pruneNum) {
        super();
        initWeight = initialWeight;
        store = new AdaptableLRUMap(featureNum, this, candidates, pruneNum);
    }

    /**
     * Returns the weights of a feature.
     *
     * @param featureHash The hash code integer identifying the feature
     * @return The weights of this feature
     */
    public float[] getWeights(final Integer featureHash) {
        return (float[]) store.get(featureHash);
    }

    /**
     * Returns the maximum number of features that can be stored by this
     * instance.
     *
     * @return the value of the attribute
     */
    public int maxSize() {
        return store.maxSize();
    }

    /**
     * Stores new weights for a feature.
     *
     * @param featureHash The hash code integer identifying the feature
     * @param weights The new weights of this feature
     */
    public void putWeights(final Integer featureHash, final float[] weights) {
        store.put(featureHash, weights);
    }

    /**
     * Resets the store, completely deleting the prediction model.
     */
    public void reset() {
        store.clear();
    }

    /**
     * Returns the number of features currently stored by this instance.
     *
     * @return the value of the attribute
     */
    public int size() {
        return store.size();
    }

    /**
     * {@inheritDoc}
     * This implementation sorts the candidate by the deviation of their weights
     * from the initial weights, so candidates with lower deviation will be
     * pruned first. This are candidates that were involved in fewer unbalanced
     * promotions and demotions than the other candidate.
     */
    public Map.Entry[] sortForPruning(final Map.Entry[] candidates) {
        // sort candidate by weight deviation
        Map.Entry candidate;
        float[] weights;
        float deviation;
        int j;
        MultiValueMap<Float, Integer> sortedMap =
            new MultiValueMap<Float, Integer>(
                    new TreeMap<Float, Collection<Integer>>());

        for (int i = 0; i < candidates.length; i++) {
            candidate = candidates[i];
            //Util.LOG.debug("Retrieving weights for " + i + "-th candidate");
            weights = (float[]) candidate.getValue();
            //Util.LOG.debug("Retrieved weights for candidate");
            deviation = 0;

            for (j = 0; j < weights.length; j++) {
                deviation += Math.abs(weights[j] - initWeight);
            }

            // map from weight deviations to collections of positions in the map
            sortedMap.put(new Float(deviation), Integer.valueOf(i));
        }

        // convert collected values into array; TreeMap sorts them by weight
        final Collection<Integer> sortedValues = sortedMap.values();
        Iterator<Integer> valueIter = sortedValues.iterator();
        final Map.Entry[] result = new Map.Entry[candidates.length];
        int i = 0;

        while (valueIter.hasNext()) {
            result[i] = candidates[valueIter.next().intValue()];
            i++;
        }

        // ensure that we retrieved all candidates
        if (i != candidates.length) {
            throw new RuntimeException("Implementation error: Retrieved "
                + i + " pruning candidates " + " instead of "
                + candidates.length + " ones");
        }

        return result;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("current size", size())
            .append("maximum size", maxSize())
            .append("prune number", store.getPruneNumber())
            .append("prune candidates", store.getCandidateNumber())
            .toString();
    }

}
