/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.combi;

import java.util.Collections;
import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import de.fu_berlin.ties.classify.PredictionDistribution;

/**
 * A combination strategy using begin/after tagging (also called "BIA" tagging
 * due to the prefixed used). Classes are prefixed with "B-" (first element of
 * an instance), "I-" (inner element of an instance), "A-" (first element
 * <em>after</em> an instance that is not part of an other instance).
 * "A" is used for anything else/outside of any class (except where the "A-"
 * prefix is used).
 *
 * <p>{@link de.fu_berlin.ties.combi.CombinationState#isEnd()} is not
 * supported by this class and will always return <code>false</code>.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.9 $, $Date: 2004/09/16 11:12:33 $, $Author: siefkes $
 */
public class BeginAfterStrategy extends CombinationStrategy {

    /**
     * The character terminating each prefix.
     */
    private static final char PREFIX_TERMINATOR = '-';

    /**
     * Prefix used to classify the first element of an instance.
     */
    private static final String BEGIN_PREFIX = "B" + PREFIX_TERMINATOR;

    /**
     * Prefix used to classify elements within instances.
     */
    private static final String IN_PREFIX = "I" + PREFIX_TERMINATOR;

    /**
     * Prefix used to classify the first element <em>after</em> an instance
     * that is not part of an other instance.
     */
    private static final String AFTER_PREFIX = "A" + PREFIX_TERMINATOR;

    /**
     * String used to classify outside/other (except where "A-" is used).
     * Actually called "A" so it will be picked in case of a tie ("A" is
     * sorted before "A-" etc.).
     */
    private static final String OUTSIDE = "A";

    /**
     * The set of all classes prefixed with {@link #BEGIN_PREFIX}, stored for
     * efficiency.
     */
    private final TreeSet<String> bClasses = new TreeSet<String>();

    /**
     * An immutable set of all classes (Strings) that can possible occur
     * during classification.
     */
    private final SortedSet[] allClasses;

    /**
     * Creates a new instance.
     *
     * @param theClasses a set of valid class names (String)
     */
    public BeginAfterStrategy(final Set<String> theClasses) {
        super(theClasses);
        final SortedSet<String> allValidClasses = new TreeSet<String>();
        final Iterator<String> iter = theClasses.iterator();
        String currentClass;

        // initialize immutable set of all classes + set of B classes
        while (iter.hasNext()) {
            currentClass = iter.next();
            bClasses.add(BEGIN_PREFIX + currentClass);
            allValidClasses.add(IN_PREFIX + currentClass);
            allValidClasses.add(AFTER_PREFIX + currentClass);
        }

        allValidClasses.addAll(bClasses);
        allValidClasses.add(OUTSIDE);
        allClasses = new SortedSet[] {
            Collections.unmodifiableSortedSet(allValidClasses)
        };
    }

    /**
     * Builds a set of class names (Strings) to pass to the classifier
     * to consider for the next decision. This implementation returns the
     * classes in alphabetic order (using a {@link SortedSet}).
     *
     * @return a set of class names
     */
    public Set<String>[] activeClasses() {
        // all B classes are always allowed
        final SortedSet<String> result = new TreeSet<String>(bClasses);
        final String lastType = state().getType();

        if ((lastType == null)) {
            // O is allowed as well
            result.add(OUTSIDE);
        } else {
            // I + A for current class are allowed as well
            result.add(IN_PREFIX + lastType);
            result.add(AFTER_PREFIX + lastType);
        }
        return new Set[] {result};
    }

    /**
     * {@inheritDoc}
     */
    public Set<String>[] allClasses() {
        return allClasses;
    }

    /**
     * {@inheritDoc}
     */
    public String[] translateCurrentState(final CombinationState currentState)
            throws IllegalArgumentException {
        final String result;
        final String type = currentState.getType();
        final String lastType = state().getType();

        if (type == null) {
            if (lastType != null) {
                // mark as after last type
                result = AFTER_PREFIX + lastType;
            } else {
                // mark as completely outside
                result = OUTSIDE;
            }
        } else {
            if (currentState.isBegin()) {
                result = BEGIN_PREFIX + type;
            } else {
                result = IN_PREFIX + type;
            }
        }
        return new String[] {result};
    }

    /**
     * {@inheritDoc}
     */
    public CombinationState translateResult(
            final PredictionDistribution[] predictions)
    throws IllegalArgumentException {
        final CombinationState result;
        final String predictedClass = predictions[0].best().getType();

        if (OUTSIDE.equals(predictedClass)) {
            // outside any instance
            result = CombinationState.OUTSIDE;
        } else {
            final int prefixEndPos =
                predictedClass.indexOf(PREFIX_TERMINATOR);

            if (prefixEndPos >= 0) {
                final String prefix;
                final String newClass;

                prefix = predictedClass.substring(0, prefixEndPos + 1);
                newClass = predictedClass.substring(prefixEndPos + 1);

                if (BEGIN_PREFIX.equals(prefix)) {
                    result = new CombinationState(newClass, true, false);
                } else if (IN_PREFIX.equals(prefix)) {
                    result = new CombinationState(newClass, false, false);
                } else if (AFTER_PREFIX.equals(prefix)) {
                    // now outside
                    result = CombinationState.OUTSIDE;
                } else {
                    throw new IllegalArgumentException(
                        "Prefix " + prefix + " of predicted class "
                            + predictedClass + " is invalid");
                }
            } else {
                throw new IllegalArgumentException(
                    "Invalid predicted class (no prefix): " + predictedClass);
            }
        }
        return result;
    }

}
