/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.combi;

import java.util.Collections;
import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import de.fu_berlin.ties.classify.Prediction;
import de.fu_berlin.ties.classify.PredictionComparator;
import de.fu_berlin.ties.classify.PredictionDistribution;
import de.fu_berlin.ties.util.Util;

/**
 * A combination strategy that uses two classifiers, one to recognize the begin
 * of extractions and one to recognize the end.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.7 $, $Date: 2004/09/15 16:08:45 $, $Author: siefkes $
 */
public class BeginEndStrategy extends CombinationStrategy {

    /**
     * The character terminating each prefix.
     */
    private static final char PREFIX_TERMINATOR = '-';

    /**
     * Prefix used for the begin classifier.
     */
    private static final String BEGIN_PREFIX = "B" + PREFIX_TERMINATOR;

    /**
     * Prefix used for the end classifier.
     */
    private static final String END_PREFIX = "E" + PREFIX_TERMINATOR;

    /**
     * String used to classify other (not begin resp. end). Actually called
     * "A" so it will be picked in case of a tie.
     */
    private static final String OTHER = "A";

    /**
     * The set of all classes for the begin classifier, stored for efficiency.
     */
    private final SortedSet<String> beginClasses;

    /**
     * The set of all classes for the end classifier, stored for efficiency.
     */
    private final SortedSet<String> endClasses;

    /**
     * An immutable set of all classes (Strings) that can possible occur
     * during classification.
     */
    private final SortedSet[] allClasses;

    /**
     * Creates a new instance.
     *
     * @param theClasses a set of valid class names (String)
     */
    public BeginEndStrategy(final Set<String> theClasses) {
        super(theClasses);

        final TreeSet<String> bClasses = new TreeSet<String>();
        final TreeSet<String> eClasses = new TreeSet<String>();
        final Iterator iter = theClasses.iterator();
        String currentClass;

        bClasses.add(OTHER);
        eClasses.add(OTHER);

        // full sets of classes
        while (iter.hasNext()) {
            currentClass = (String) iter.next();
            bClasses.add(BEGIN_PREFIX + currentClass);
            eClasses.add(END_PREFIX + currentClass);
        }

        // initialize unmodifyable sets + set array
        beginClasses = Collections.unmodifiableSortedSet(bClasses);
        endClasses = Collections.unmodifiableSortedSet(eClasses);
        allClasses = new SortedSet[] {beginClasses, endClasses};
    }

    /**
     * {@inheritDoc}
     */
    public Set<String>[] activeClasses() {
        // all classes are always active
        return allClasses;
    }

    /**
     * {@inheritDoc}
     */
    public Set<String>[] allClasses() {
        return allClasses;
    }

    /**
     * {@inheritDoc}
     */
    protected boolean resetHook() {
        final boolean result;

        if (!state().isEnd() && (state().getType() != null)) {
            // discard not-yet-finished extraction
            result = true;
        } else {
            return false;
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public String[] translateCurrentState(final CombinationState currentState)
            throws IllegalArgumentException {
        final String beginResult;
        final String endResult;

        if (currentState.isBegin()) {
            // positive instance for the begin classifier
            beginResult = BEGIN_PREFIX + currentState.getType();
        } else {
            beginResult = OTHER;
        }

        if (currentState.isEnd()) {
            // positive instance for the end classifier
            endResult = END_PREFIX + currentState.getType();
        } else {
            endResult = OTHER;
        }

        return new String[] {beginResult, endResult};
    }

    /**
     * {@inheritDoc}
     */
    public CombinationState translateResult(
            final PredictionDistribution[] predictions)
    throws IllegalArgumentException {
        if (predictions.length != 2) {
            throw new IllegalArgumentException(
                "Illegal number of classifiers: " + predictions.length
                + " instead of 2");
        }

        final CombinationState result;
        final Prediction beginBest = predictions[0].best();
        final Prediction endBest = predictions[1].best();
        final PredictionComparator predComp = new PredictionComparator();

        boolean isBegin = !beginBest.getType().equals(OTHER);
        boolean isEnd = !endBest.getType().equals(OTHER);
        final String newClass;

        // previous state
        final String lastType = state().getType();
        final boolean lastEnd = state().isEnd();

        if (isBegin) {
            // strip prefix to determine actual class
            final String beginClass =
                beginBest.getType().substring(BEGIN_PREFIX.length());

            if (isEnd) {
                final String endClass =
                    endBest.getType().substring(END_PREFIX.length());

                if (beginClass.equals(endClass)) {
                    // start new single-token extraction
                    newClass = beginClass;
                } else if (endClass.equals(lastType) && !lastEnd) {
                    // end class is valid: trust the more confident classifier
                    // (end classifier in case of a type)
                    if (predComp.compare(beginBest, endBest) <= 0) {
                        // end classifier is more or equally confident
                        newClass = endClass;
                        isBegin = false;
                        Util.LOG.debug("BeginEnd: end classification "
                            + endBest + " wins over begin classification "
                            + beginBest + "due to higher/equal confidence");
                    } else {
                        // begin classifier is more confident
                        newClass = beginClass;
                        isEnd = false;
                        Util.LOG.debug("BeginEnd: begin classification "
                            + beginBest + " wins over end classification "
                            + endBest + "due to higher confidence");
                    }
                } else {
                    // end class is invalid: use the begin classifier
                    newClass = beginClass;
/*                    Util.LOG.debug("BeginEnd: end classification " + endBest
                        + " is invalid -- using begin classification"); */
                }
            } else {
                // begin of new instance
                newClass = beginClass;
            }
        } else {
            if (isEnd) {
                final String endClass =
                    endBest.getType().substring(END_PREFIX.length());

                // check whether end prediction is valid
                if (endClass.equals(lastType) && !lastEnd) {
                    newClass = endClass;
                } else {
                    // discard invalid prediction
                    isEnd = false;
/*                    Util.LOG.debug("BeginEnd: end classification " + endBest
                        + " is invalid -- using state of preceding token"); */

                    if (lastEnd) {
                        // set type to null because last extraction is finished
                        newClass = null;
                    } else {
                        // re-use type from last state
                        newClass = lastType;
                    }
                }
            } else {
                // neither begin nor end: use state of preceding token
                if (lastEnd) {
                    // set type to null because last extraction is finished
                    newClass = null;
                } else {
                    // re-use type from last state
                    newClass = lastType;
                }
            }
        }

        final boolean discardPrevious;

        if (isBegin && !lastEnd && (lastType != null)
                && !lastType.equals(newClass)) {
            // starting a new extraction but the last extraction is unfinished
            // and has an different type: discard it
            discardPrevious = true;
        } else {
            discardPrevious = false;
        }

        if (newClass != null) {
            if (getValidClasses().contains(newClass)) {
                result = new CombinationState(newClass, isBegin, isEnd,
                    discardPrevious);
            } else {
                throw new IllegalArgumentException(
                    "Type " + newClass + " derived from predictions "
                        + beginBest + " and " + endBest + " is invalid");
            }
        } else {
            result = CombinationState.OUTSIDE;
        }

        return result;
    }

}
