/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.combi;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * Immutable class contains state information used by
 * {@linkplain de.fu_berlin.ties.combi.CombinationStrategy combination
 * strategies}.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2004/07/15 14:35:48 $, $Author: siefkes $
 */
public class CombinationState {

    /**
     * Combination state to be used for outside/other (type is set to
     * <code>null</code>). Begin and end are always <code>false</code>.
     */
    public static final CombinationState OUTSIDE =
        new CombinationState(null, false, false);

    /**
     * The type (class) of the current instance; or <code>null</code> if we're
     * outside of any instance ({@link #OUTSIDE}).
     */
    private final String type;

    /**
     * Whether this is the begin of an instance.
     */
    private final boolean begin;

    /**
     * Whether this is the end of an instance.
     */
    private final boolean end;

    /**
     * Signals that the extraction in progress during the <em>preceding</em>
     * combination state should be discarded. If this combination state starts
     * a new extraction, the new extraction should <em>not</em> be discarded.
     */
    private final boolean discardPreceding;

    /**
     * Creates a new instance, setting {@link #isDiscardPreceding()} to
     * <code>false</code>.
     *
     * @param myType the type (class) of this instance; or <code>null</code> if
     * we're outside of any instance
     * @param myBegin whether this is the begin of an instance
     * @param myEnd whether this is the end of an instance
     */
    public CombinationState(final String myType, final boolean myBegin,
            final boolean myEnd) {
        this(myType, myBegin, myEnd, false);
    }

    /**
     * Creates a new instance.
     *
     * @param myType the type (class) of this instance; or <code>null</code> if
     * we're outside of any instance
     * @param myBegin whether this is the begin of an instance
     * @param myEnd whether this is the end of an instance
     * @param discardPrevious whether the extraction in progress during the
     * <em>preceding</em> combination state should be discarded
     */
    public CombinationState(final String myType, final boolean myBegin,
            final boolean myEnd, final boolean discardPrevious) {
        super();
        type = myType;
        begin = myBegin;
        end = myEnd;
        discardPreceding = discardPrevious;
    }

    /**
     * Indicates whether some other object is "equal to" this one, fulfulling
     * the {@link Object#equals(java.lang.Object)} contract.
     *
     * @param obj the reference object with which to compare
     * @return <code>true</code> iff the specified object is a
     * {@link CombinationState} equal to this instance
     */
    public boolean equals(final Object obj) {
        if (obj == this) {
            return true;
        } else if ((obj != null) && (getClass().equals(obj.getClass()))) {
            // used getClass instead of instanceof because otherwise subclasses
            // with additional fields would break the contract
            final CombinationState other = (CombinationState) obj;
            return new EqualsBuilder()
                .append(type, other.type)
                .append(begin, other.begin)
                .append(end, other.end)
                .append(discardPreceding, other.discardPreceding)
                .isEquals();
        } else {
            return false;
        }
    }

    /**
     * Returns the type (class) of the current instance; or <code>null</code>
     * if we're outside of any instance ({@link #OUTSIDE}).
     *
     * @return the value of the attribute
     */
    public String getType() {
        return type;
    }

    /**
     * Returns a hash code value for this object, fulfulling the
     * {@link Object#hashCode()} contract.
     *
     * @return a hash code value for this object
     */
    public int hashCode() {
        // you pick two hard-coded, randomly chosen, non-zero, odd numbers
        // (preferably primes); ideally different for each class
        return new HashCodeBuilder(29, 31)
            .append(type)
            .append(begin)
            .append(end)
            .append(discardPreceding)
            .toHashCode();
    }

    /**
     * Returns whether this is the begin of an instance.
     * @return the value of the attribute
     */
    public boolean isBegin() {
        return begin;
    }

    /**
     * Signals that the extraction in progress during the <em>preceding</em>
     * combination state should be discarded. If this combination state starts
     * a new extraction, the new extraction should <em>not</em> be discarded.
     *
     * @return the value of the attribute
     */
    public boolean isDiscardPreceding() {
        return discardPreceding;
    }

    /**
     * Returns whether this is the end of an instance.
     * @return the value of the attribute
     */
    public boolean isEnd() {
        return end;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("type", type)
            .append("begin", begin)
            .append("end", end)
            .append("discard preceding", discardPreceding)
            .toString();
    }

}
