/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.combi;

import java.util.Set;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.PredictionDistribution;
import de.fu_berlin.ties.util.Util;

/**
 * A combination strategy combines sequential classifications. Instances of
 * this class are not thread-safe. Subclasses must also provide a constructor
 * that accepts a {@link java.util.Set} of valid class names as single argument;
 * otherwise they cannot be loaded via reflection.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.14 $, $Date: 2004/09/16 11:12:53 $, $Author: siefkes $
 */
public abstract class CombinationStrategy {

    /**
     * Configuration key specifying the combination strategy to use.
     */
    public static final String CONFIG_COMBINATION = "combination.strategy";

    /**
     * Factory method that delegates to
     * {@link #createStrategy(Set, TiesConfiguration)} using the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @param classNames a set of valid class names (String)
     * @return the created strategy
     * @throws IllegalArgumentException if the value of the
     * {@link #CONFIG_COMBINATION} key is missing or invalid
     * @throws ProcessingException if an error occurred while creating the
     * strategy
     */
    public static CombinationStrategy createStrategy(
            final Set<String> classNames)
    throws IllegalArgumentException, ProcessingException {
        return createStrategy(classNames, TiesConfiguration.CONF);
    }

    /**
     * Factory method that creates a combination strategy based on the
     * {@link #CONFIG_COMBINATION} key in the provided configuration.
     *
     * <p>Currently supported values:
     * "BE" for {@link BeginEndStrategy},
     * "BIA" for {@link BeginAfterStrategy},
     * "BIE" for {@link OpenCloseStrategy},
     * "IOB1" for {@link InsideOutsideStrategy} with
     * {@link InsideOutsideStrategy#isBStartingAll()} set to <code>false</code>,
     * "IOB2" for {@link InsideOutsideStrategy} with
     * {@link InsideOutsideStrategy#isBStartingAll()} set to <code>true</code>;
     * "Triv" for {@link TrivialStrategy}.
     *
     * <p>Otherwise the value must be the qualified name of a
     * CombinationStrategy subclass accepting a {@link Set} (of valid class
     * names) as single argument.
     *
     * @param classNames a set of valid class names (String)
     * @param config the configuration to use
     * @return the created strategy
     * @throws IllegalArgumentException if the value of the
     * {@link #CONFIG_COMBINATION} key is missing or invalid
     * @throws ProcessingException if an error occurred while creating the
     * strategy
     */
    public static CombinationStrategy createStrategy(
            final Set<String> classNames, final TiesConfiguration config)
    throws IllegalArgumentException, ProcessingException {
        if (!config.containsKey(CONFIG_COMBINATION)) {
            throw new IllegalArgumentException(
                "Cannot create combination strategy -- missing key "
                + CONFIG_COMBINATION);
        }

        final String combinationValue =
            config.getString(CONFIG_COMBINATION);
        final String upperValue = combinationValue.toUpperCase();
        final CombinationStrategy result;

        if ("BE".equals(upperValue)) {
            // load Begin/End strategy
            result = new BeginEndStrategy(classNames);
        } else if ("BIA".equals(upperValue)) {
            // load Begin/After strategy
            result = new BeginAfterStrategy(classNames);
        } else if ("BIE".equals(upperValue)) {
            // load Open/Close strategy
            result = new OpenCloseStrategy(classNames);
        } else if ("IOB1".equals(upperValue)) {
            // load Inside/Outside with B only where strictly required
            result = new InsideOutsideStrategy(classNames, false);
        } else if ("IOB2".equals(upperValue)) {
            // load InsideOutside with B starting all instances
            result = new InsideOutsideStrategy(classNames, true);
        } else if ("TRIV".equals(upperValue)) {
            // load trivial strategy
            result = new TrivialStrategy(classNames);
        } else {
            // should be the qualified name of a CombinationStrategy subclass
            // accepting a Set (of valid class names) as single argument
            try {
                result = (CombinationStrategy) Util.createObject(
                    Class.forName(combinationValue),
                    new Object[] {classNames}, Set.class);
            } catch (ClassNotFoundException cnfe) {
                // convert and rethrow exception
                throw new ProcessingException(
                    "Cannot create combination strategy from key "
                    + CONFIG_COMBINATION + "=" + combinationValue,
                    cnfe);
            } catch (InstantiationException ie) {
                // convert and rethrow exception
                throw new ProcessingException(
                    "Cannot create combination strategy from key "
                    + CONFIG_COMBINATION + "=" + combinationValue,
                    ie);
            }
        }
        return result;
    }

    /**
     * The current state (state of the last item); initially set via
     * {@link #reset()}.
     */
    private CombinationState state = CombinationState.OUTSIDE;

    /**
     * The array of valid class names (Strings).
     */
    private final Set<String> validClasses;

    /**
     * Creates a new instance.
     * @param theClasses a set of valid class names (String)
     */
    public CombinationStrategy(final Set<String> theClasses) {
        super();
        validClasses = theClasses;
        state = CombinationState.OUTSIDE;
    }

    /**
     * Builds a set array of class names (Strings) to pass to each classifier
     * to consider for the next decision. Must be a subset of
     * {@link #allClasses()}.
     *
     * @return a set of class names for each classifier
     */
    public abstract Set<String>[] activeClasses();

    /**
     * Returns an array of immutable sets of all classes (Strings) that can
     * possible occur during classification for each classifier.
     * The "background" class (outside any extractions) should be the first
     * member of each set.
     *
     * @return an array of immutable sets of class names
     */
    public abstract Set<String>[] allClasses();

    /**
     * Translates an expected class to a class name to feed to the classifiers
     * in training mode. This method reverses the direction of
     * {@link #translateResult(PredictionDistribution[])}.
     *
     * <p>The return value can be feed to the classifiers for training it.
     *
     * @param currentState the current state
     * @return the array of classes to feed to each classifier for training
     * @throws IllegalArgumentException if the specified argument is invalid
     * (non-empty string that is not in the set of valid classes, or empty
     * string when <code>null</code> was given in the preceding call)
     */
    public abstract String[] translateCurrentState(
            final CombinationState currentState)
            throws IllegalArgumentException;

    /**
     * Translates the classes predicted by the classifiers in a decision for the
     * caller. The {@linkplain CombinationState#isEnd() end} information might
     * be unreliably/absent for same classes.
     *
     * @param predictions array containing the prediction distributions of
     * all classifier
     * @return the translated result
     * @throws IllegalArgumentException if the specified argument was invalid
     * (not in the set of currently {@link #activeClasses() active classes}
     */
    public abstract CombinationState translateResult(
            final PredictionDistribution[] predictions)
    throws IllegalArgumentException;

    /**
     * Returns the set of valid class names.
     * @return a set of Strings containing the valid class names
     */
    protected Set getValidClasses() {
        return validClasses;
    }

    /**
     * Resets the {@link #state()} of this instance to the initial value
     * {@link CombinationState#OUTSIDE}. Also checks whether the last extraction
     * should be discarded, analogously to
     * {@link CombinationState#isDiscardPreceding()}. The default behavior
     * is to always return <code>false</code>, but subclasses can modify
     * this by overwriting the {@link #resetHook()} method.
     *
     * @return <code>true</code> iff the last extraction should be discarded
     */
    public final boolean reset() {
        final boolean result = resetHook();
        state = CombinationState.OUTSIDE;
        return result;
    }

    /**
     * Hook called by the {@link #reset()} method to query whether the last
     * extraction should be discarded, analogously to
     * {@link CombinationState#isDiscardPreceding()}.
     *
     * @return the value of the attribute
     */
    protected boolean resetHook() {
        return false;
    }

    /**
     * The current state (state of the last item).
     * @return the current state
     */
    public CombinationState state() {
        return state;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("valid classes", validClasses)
            .append("last state", state)
          .toString();
    }

    /**
     * Updates the current state (the class of the last item). You must always
     * call this method after classifying
     * ({@link #translateResult(PredictionDistribution[])}) and/or training
     * ({@link #translateCurrentState(CombinationState)}) an item.
     *
     * @param newState the state of the last item
     * @throws IllegalArgumentException if the
     * {@linkplain CombinationState#getType() type} of the argument is invalid
     * (not in the set of {@link #getValidClasses() valid classes or
     * <code>null</code>}
     */
    public void updateState(final CombinationState newState)
            throws IllegalArgumentException {
        // check that it's a valid state (if not null)
        final String type = newState.getType();
        if ((type != null) && (!validClasses.contains(type))) {
            throw new IllegalArgumentException("Not a valid type: "
                + newState);
        }
        state = newState;
    }

}
