/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.combi;

import java.util.Collections;
import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.classify.PredictionDistribution;

/**
 * A combination strategy using inside/outside tagging. Classes are prefixed
 * with "I-" (in this class) or "B-" (begin of a new instance of this class),
 * "A" is used for anything else/outside of any class.
 *
 * <p>{@link de.fu_berlin.ties.combi.CombinationState#isEnd()} is not
 * supported by this class and will always return <code>false</code>.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.9 $, $Date: 2004/09/15 16:08:45 $, $Author: siefkes $
 */
public class InsideOutsideStrategy extends CombinationStrategy {

    /**
     * Prefix used to classify the begin of a new instance.
     */
    private static final String BEGIN_PREFIX = "B-";

    /**
     * Prefix used to classify items within instances.
     */
    private static final String IN_PREFIX = "I-";

    /**
     * String used to classify outside/other. Actually called "A" so it will be
     * picked in case of a tie.
     */
    private static final String OUTSIDE = "A";

    /**
     * Whether the "B-" prefix is used for the first item of each instance
     * ("O/B/I" mode) or only for the first item of instances immediately
     * following an instance of the same class ("O/I/B" mode).
     */
    private final boolean bStartingAll;

    /**
     * The set of all classes prefixed with {@link #BEGIN_PREFIX}, stored for
     * efficiency.
     */
    private final TreeSet<String> bClasses;

    /**
     * The set of all classes prefixed with {@link #IN_PREFIX}, stored for
     * efficiency.
     */
    private final TreeSet<String> iClasses;

    /**
     * An immutable set of all classes (Strings) that can possible occur
     * during classification.
     */
    private final SortedSet[] allClasses;

    /**
     * Creates a new instance, using "O/I/B" mode: the "B-" is only used
     * where required for disambiguation, otherwise the "I-" prefix is used.
     *
     * @param theClasses a set of valid class names (String)
     */
    public InsideOutsideStrategy(final Set<String> theClasses) {
        this(theClasses, false);
    }

    /**
     * Creates a new instance, using "O/I/B" mode (setting
     * {@link #isBStartingAll()} to <code>false</code>). In this mode, the "B-"
     * is only used where required for disambiguation, otherwise the "I-" prefix
     * is used.
     *
     * @param theClasses a set of valid class names (String)
     * @param bStartingAllClasses if <code>true</code>, the "B-" prefix is used
     * for the first item of each instance ("O/B/I" mode); otherwise it is used
     * only for first item of instances immediately following an instance of the
     * same class ("O/I/B" mode)
     */
    public InsideOutsideStrategy(final Set<String> theClasses,
                                 final boolean bStartingAllClasses) {
        super(theClasses);
        bStartingAll = bStartingAllClasses;
        bClasses = new TreeSet<String>();
        iClasses = new TreeSet<String>();
        final Iterator iter = theClasses.iterator();
        String currentClass;

        // prepend prefixes to all classes
        while (iter.hasNext()) {
            currentClass = (String) iter.next();
            bClasses.add(BEGIN_PREFIX + currentClass);
            iClasses.add(IN_PREFIX + currentClass);
        }

        // initialize immutable set of all classes
        final SortedSet<String> allValidClasses = new TreeSet<String>();
        allValidClasses.addAll(bClasses);
        allValidClasses.addAll(iClasses);
        allValidClasses.add(OUTSIDE);
        allClasses = new SortedSet[] {
            Collections.unmodifiableSortedSet(allValidClasses)
        };
    }

    /**
     * Builds a set of class names (Strings) to pass to the classifier
     * to consider for the next decision. This implementation returns the
     * classes in alphabetic order (using a {@link SortedSet}).
     *
     * @return a set of class names
     */
    public Set<String>[] activeClasses() {
        final SortedSet<String> result;
        if (bStartingAll) {
            // all B classes allowed
            result = new TreeSet<String>(bClasses);
        } else {
            // all I classes allowed
            result = new TreeSet<String>(iClasses);
        }

        if (state().getType() != null) {
            if (bStartingAll) {
                // I is used to continue current instance
                result.add(IN_PREFIX + state().getType());
            } else {
                // B is used to start new instance of same type
                result.add(BEGIN_PREFIX + state().getType());
            }
        }

        // O is always allows
        result.add(OUTSIDE);
        return new Set[] {result};
    }

    /**
     * {@inheritDoc}
     */
    public Set<String>[] allClasses() {
        return allClasses;
    }

    /**
     * Whether the "B-" prefix is used for the first item of each instance
     * ("O/B/I" mode) or only for the first item of instances immediately
     * following an instance of the same class ("O/I/B" mode).
     *
     * @return the value of the attribute
     */
    public boolean isBStartingAll() {
        return bStartingAll;
    }

    /**
     * {@inheritDoc}
     */
    public String[] translateCurrentState(final CombinationState currentState)
            throws IllegalArgumentException {
        final String result;
        final String type = currentState.getType();

        if (type == null) {
            // null: outside
            result = OUTSIDE;
        } else if (!currentState.isBegin()) {
            // continuation of last instance (must have same type)
            if (type.equals(state().getType())) {
                result = IN_PREFIX + type;
            } else {
                throw new IllegalArgumentException(
                    "Illegal continuation state: last state "
                        + state().getType() + " differs from current state "
                        + type);
            }
        } else if (type.equals(state().getType())) {
            // new instance of same type
            result = BEGIN_PREFIX + type;
        } else {
            // new instance, last class was different
            if (bStartingAll) {
                result = BEGIN_PREFIX + type;
            } else {
                result = IN_PREFIX + type;
            }
        }
        return new String[] {result};
    }

    /**
     * {@inheritDoc}
     */
    public CombinationState translateResult(
            final PredictionDistribution[] predictions)
    throws IllegalArgumentException {
        final CombinationState result;
        final String predictedClass = predictions[0].best().getType();

        if (OUTSIDE.equals(predictedClass)) {
            // outside any instance
            result = CombinationState.OUTSIDE;
        } else {
            final String newClass;
            final boolean isBegin;

            if (predictedClass.startsWith(BEGIN_PREFIX)) {
                // begin of a instance
                isBegin = true;
                newClass = predictedClass.substring(BEGIN_PREFIX.length());
            } else if (predictedClass.startsWith(IN_PREFIX)) {
                // within a instance
                isBegin = false;
                newClass = predictedClass.substring(IN_PREFIX.length());
            } else {
                throw new IllegalArgumentException(
                    "Invalid predicted class: " + predictedClass);
            }

            if (isBegin) {
                if (!bStartingAll) {
                    // check that this is allowed (last item had same class)
                    if (!newClass.equals(state().getType())) {
                        throw new IllegalArgumentException("'O/I/B' mode: "
                            + "B must occur after item of same class, but"
                            + "old class (" + state().getType()
                            + ") and new class (" + newClass + ") differ");
                    }
                }

                // start of new instance
                result = new CombinationState(newClass, true, false);
            } else {
                if (bStartingAll) {
                    // check that this is allowed (last item had same class)
                    if (!newClass.equals(state().getType())) {
                        throw new IllegalArgumentException("'O/B/I' mode: "
                            + "I must occur after item of same class, but"
                            + "old class (" + state().getType()
                            + ") and new class (" + newClass + ") differ");
                    }

                    // continue last instance
                    result = new CombinationState(newClass, false, false);
                } else {
                    // O/I/B mode:
                    if (newClass.equals(state().getType())) {
                        // continue last instance
                        result = new CombinationState(newClass, false, false);
                    } else {
                        // start of new instance
                        result = new CombinationState(newClass, true, false);
                    }
                }
            }
        }
        return result;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("B starts all", bStartingAll)
            .toString();
    }

}
