/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.combi;

import java.util.Collections;
import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import de.fu_berlin.ties.classify.PredictionDistribution;

/**
 * A combination strategy using open/close tagging (also called "BIE" tagging
 * due to the prefixed used). Classes are prefixed with "B-" (first element of
 * an instance), "I-" (inner element of an instance), "E-" (last element of an
 * instance), or "BE-" (only element of an instance -- begin and end).
 * "A" is used for anything else/outside of any class.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.7 $, $Date: 2004/09/15 16:08:45 $, $Author: siefkes $
 */
public class OpenCloseStrategy extends CombinationStrategy {

    /**
     * The character terminating each prefix.
     */
    private static final char PREFIX_TERMINATOR = '-';

    /**
     * Prefix used to classify the first element of an instance.
     */
    private static final String BEGIN_PREFIX = "B" + PREFIX_TERMINATOR;

    /**
     * Prefix used to classify inner elements of an instance.
     */
    private static final String INNER_PREFIX = "I" + PREFIX_TERMINATOR;

    /**
     * Prefix used to classify the last element of an instance.
     */
    private static final String END_PREFIX = "E" + PREFIX_TERMINATOR;

    /**
     * Prefix used to classify the only element of an instance (begin and end).
     */
    private static final String BEGIN_END_PREFIX = "BE" + PREFIX_TERMINATOR;

    /**
     * String used to classify outside/other. Actually called "A" so it will be
     * picked in case of a tie.
     */
    private static final String OUTSIDE = "A";

    /**
     * The set of all classes prefixed with {@link #BEGIN_PREFIX}, stored for
     * efficiency.
     */
    private final TreeSet<String> bClasses = new TreeSet<String>();

    /**
     * The set of all classes prefixed with {@link #BEGIN_END_PREFIX}, stored
     * for efficiency.
     */
    private final TreeSet<String> beClasses = new TreeSet<String>();

    /**
     * An immutable set of all classes (Strings) that can possible occur
     * during classification.
     */
    private final SortedSet[] allClasses;

    /**
     * Creates a new instance.
     *
     * @param theClasses a set of valid class names (String)
     */
    public OpenCloseStrategy(final Set<String> theClasses) {
        super(theClasses);
        final SortedSet<String> allValidClasses = new TreeSet<String>();
        final Iterator<String> iter = theClasses.iterator();
        String currentClass;

        // initialize immutable set of all classes + sets of B/BE classes
        while (iter.hasNext()) {
            currentClass = iter.next();
            bClasses.add(BEGIN_PREFIX + currentClass);
            beClasses.add(BEGIN_END_PREFIX + currentClass);
            allValidClasses.add(INNER_PREFIX + currentClass);
            allValidClasses.add(END_PREFIX + currentClass);
        }

        allValidClasses.addAll(bClasses);
        allValidClasses.addAll(beClasses);
        allValidClasses.add(OUTSIDE);
        allClasses = new SortedSet[] {
            Collections.unmodifiableSortedSet(allValidClasses)
        };
    }

    /**
     * Builds a set of class names (Strings) to pass to the classifier
     * to consider for the next decision. This implementation returns the
     * classes in alphabetic order (using a {@link SortedSet}).
     *
     * @return a set of class names
     */
    public Set<String>[] activeClasses() {
        final SortedSet<String> result;
        final String lastType = state().getType();

        if ((lastType == null) || state().isEnd()) {
            // all B + BE classes allowed + O
            result = new TreeSet<String>(bClasses);
            result.addAll(beClasses);
            result.add(OUTSIDE);
        } else {
            // only I + E for current class allowed
            result = new TreeSet<String>();
            result.add(INNER_PREFIX + lastType);
            result.add(END_PREFIX + lastType);
        }
        return new Set[] {result};
    }

    /**
     * {@inheritDoc}
     */
    public Set<String>[] allClasses() {
        return allClasses;
    }

    /**
     * {@inheritDoc}
     */
    public String[] translateCurrentState(final CombinationState currentState)
            throws IllegalArgumentException {
        final String result;
        final String type = currentState.getType();

        if (type == null) {
            // null: outside
            result = OUTSIDE;
        } else {
            final String prefix;

            if (currentState.isBegin()) {
                if (currentState.isEnd()) {
                    prefix = BEGIN_END_PREFIX;
                } else {
                    prefix = BEGIN_PREFIX;
                }
            } else {
                if (currentState.isEnd()) {
                    prefix = END_PREFIX;
                } else {
                    prefix = INNER_PREFIX;
                }
            }
            result = prefix + type;
        }
        return new String[] {result};
    }

    /**
     * {@inheritDoc}
     */
    public CombinationState translateResult(
            final PredictionDistribution[] predictions)
    throws IllegalArgumentException {
        final CombinationState result;
        final String predictedClass = predictions[0].best().getType();

        if (OUTSIDE.equals(predictedClass)) {
            // outside any instance
            result = CombinationState.OUTSIDE;
        } else {
            final int prefixEndPos =
                predictedClass.indexOf(PREFIX_TERMINATOR);

            if (prefixEndPos >= 0) {
                final String prefix;
                final String newClass;
                final boolean isBegin;
                final boolean isEnd;

                prefix = predictedClass.substring(0, prefixEndPos + 1);
                newClass = predictedClass.substring(prefixEndPos + 1);

                if (BEGIN_PREFIX.equals(prefix)) {
                    isBegin = true;
                    isEnd = false;
                } else if (BEGIN_END_PREFIX.equals(prefix)) {
                    isBegin = true;
                    isEnd = true;
                } else if (INNER_PREFIX.equals(prefix)) {
                    isBegin = false;
                    isEnd = false;
                } else if (END_PREFIX.equals(prefix)) {
                    isBegin = false;
                    isEnd = true;
                } else {
                    throw new IllegalArgumentException(
                        "Prefix " + prefix + " of predicted class "
                            + predictedClass + " is invalid");
                }

                if (getValidClasses().contains(newClass)) {
                    result = new CombinationState(newClass, isBegin, isEnd);
                } else {
                    throw new IllegalArgumentException(
                        "Type " + newClass + " of predicted class "
                            + predictedClass + " is invalid");
                }
            } else {
                throw new IllegalArgumentException(
                    "Invalid predicted class (no prefix): " + predictedClass);
            }
        }
        return result;
    }

}
