/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.combi;

import java.util.Collections;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import de.fu_berlin.ties.classify.PredictionDistribution;

/**
 * The trivial combination strategy simply uses a single tag for each class;
 * no prefixes are used. Thus is cannot express the instances of the same
 * class immediately following one after other -- they would be joined in a
 * single instance. "A" is used for anything else/outside of any class (thus it
 * cannot be used as the name of a class).
 *
 * <p>{@link de.fu_berlin.ties.combi.CombinationState#isEnd()} is not
 * supported by this class and will always return <code>false</code>.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.7 $, $Date: 2004/09/16 11:12:53 $, $Author: siefkes $
 */
public class TrivialStrategy extends CombinationStrategy {

    /**
     * String used to classify outside/other. Actually called "A" so it will be
     * picked in case of a tie.
     */
    private static final String OUTSIDE = "A";

    /**
     * An immutable set of all classes (Strings) that can possible occur
     * during classification.
     */
    private final SortedSet[] allClasses;

    /**
     * Creates a new instance.
     *
     * @param theClasses a set of valid class names (String)
     */
    public TrivialStrategy(final Set<String> theClasses) {
        super(theClasses);
        final SortedSet<String> allValidClasses =
            new TreeSet<String>(theClasses);
        allValidClasses.add(OUTSIDE);
        allClasses = new SortedSet[] {
            Collections.unmodifiableSortedSet(allValidClasses)
        };
    }

    /**
     * Builds a set of class names (Strings) to pass to the classifier
     * to consider for the next decision. This implementation returns the
     * classes in alphabetic order (using a {@link SortedSet}).
     *
     * @return a set of class names
     */
    public Set<String>[] activeClasses() {
        // all classes are always allowed
        return allClasses();
    }

    /**
     * {@inheritDoc}
     */
    public Set<String>[] allClasses() {
        return allClasses;
    }

    /**
     * {@inheritDoc}
     */
    public String[] translateCurrentState(final CombinationState currentState)
            throws IllegalArgumentException {
        final String result;
        final String type = currentState.getType();

        if (type == null) {
            // null: outside
            result = OUTSIDE;
        } else {
            // return type as is
            result = type;
        }
        return new String[] {result};
    }

    /**
     * {@inheritDoc}
     */
    public CombinationState translateResult(
            final PredictionDistribution[] predictions)
    throws IllegalArgumentException {
        final CombinationState result;
        final String predictedClass = predictions[0].best().getType();

        if (OUTSIDE.equals(predictedClass)) {
            // outside any instance
            result = CombinationState.OUTSIDE;
        } else {
            final boolean isBegin;
            if (predictedClass.equals(state().getType())) {
                isBegin = false;
            } else {
                isBegin = true;
            }

            if (getValidClasses().contains(predictedClass)) {
                result = new CombinationState(predictedClass, isBegin, false);
            } else {
                throw new IllegalArgumentException(
                    "Predicted type " + predictedClass + " is invalid");
            }
        }
        return result;
    }

}
