/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.context;

import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Element;

import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.util.MutableInt;
import de.fu_berlin.ties.util.Util;
import de.fu_berlin.ties.xml.dom.DOMUtils;

/**
 * Provides basic functionality shared by different representations.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/09/06 17:22:41 $, $Author: siefkes $
 */
public abstract class AbstractRepresentation extends Representation {

    /**
     * Configuration key: The number of preceding recognitions to represent.
     */
    public static final String CONFIG_RECOGN_NUM = "representation.recogn.num";

    /**
     * Configuration key: The maximum number of subsequences to keep when a
     * feature value must be split.
     */
    public static final String CONFIG_SPLIT_MAXIMUM =
        "representation.split.maximum";

    /**
     * Configuration key: Each <em>n</em>-th context representation is stored
     * for debugging and inspection purposes,if &gt; 0.
     */
    public static final String CONFIG_STORE_NTH = "representation.store.nth";

    /**
     * The maximum number of subsequences to keep when a feature value must
     * be split (at whitespace).
     */
    private final int splitMaximum;

    /**
     * Counts how many representations were already generated (if some of them
     * are {@linkplain #storeN stored}).
     */
    private long repCount = 0;

    /**
     * Each <em>storeN</em>-th context representation is stored for debugging
     * and inspection purposes (if &gt; 0, otherwise no representation is
     * stored).
     */
    private final int storeN;

    /**
     * Last integer suffix used to store a context representation, if enabled
     * ({@link #storeN} &gt; 0). Also used to synchronize the {@link #repCount}.
     */
    private final MutableInt lastStoredSuffix = new MutableInt(0);

    /**
     * The output character set to use (only used to
     * {@linkplain #storeIfNth(FeatureVector, String) store} some configurations
     * for inspection purposes).
     */
    private final String outputCharset;

    /**
     * Creates a new instance.
     *
     * @param recogNum the number of preceding recognitions to represent
     * @param splitMax the maximum number of subsequences to keep when
     * a feature value must be split (at whitespace).
     * @param n Each <em>n</em>-th context representation is stored if &gt; 0;
     * otherwise no representation is stored
     * @param outCharset the output character set to use (only used to
     * store some configurations for inspection purposes, if <code>n</code>
     * &gt; 0); if <code>null</code>, the default charset of the current
     * platform is used
     */
    public AbstractRepresentation(final int recogNum, final int splitMax,
            final int n, final String outCharset) {
        super(recogNum);
        splitMaximum = splitMax;
        storeN = n;
        outputCharset = outCharset;
    }

    /**
     * Builds the context representation of text in an element. Returns a
     * feature vector of all context features considered relevant for
     * representation.
     *
     * @param element the element whose context should be represented
     * @param leftText textual content to the left of (preceding)
     * <code>mainText</code>, might be empty
     * @param mainText the main textual content to represent, might be empty
     * @param rightText textual content to the right of (following)
     * <code>mainText</code>, might be empty
     * @param priorRecognitions a buffer of the last {@link Recognition}s from
     * the document, created by calling {@link #createRecognitionBuffer()};
     * might be <code>null</code>
     * @param featureCache a cache of (local) feature, should be re-used
     * between all calls for the nodes in a single document (but must not be
     * re-used when building the context of nodes in different documents!)
     * @param logPurpose the type of contexts of main interest to the caller
     * (e.g. "Token" or "Sentence"), used for logging
     * @return a vector of features considered relevant for representation
     * @throws ClassCastException if the <code>priorRecognitions</code> buffer
     * contains objects that aren't {@link Recognition}s
     * @throws IllegalArgumentException if the specified node is of an
     * unsupported type
     */
    public final FeatureVector buildContext(final Element element,
            final String leftText, final String mainText,
            final String rightText, final PriorRecognitions priorRecognitions,
            final Map<Element, List<LocalFeature>> featureCache,
            final String logPurpose)
    throws ClassCastException, IllegalArgumentException {
        FeatureVector result = null;

        try {
            result = doBuildContext(element, leftText, mainText, rightText,
                    priorRecognitions, featureCache, logPurpose);

            // store each n-th representation, if configured
            if (storeN > 0) {
                storeIfNth(result, logPurpose);
            }

            return result;
        } catch (Error e) {
            // log details and rethrow
            Util.LOG.error("Error while building " + logPurpose 
                    + " context representation for " 
                    + DOMUtils.showToken(element, mainText)
                    + "(current size of rep.: " + result.size()
                    + ", size of feature cache: " + featureCache.size() + ")",
                    e);
            Util.LOG.debug("Snapshot of context rep.: " + result);
            //Util.LOG.debug("Snapshot of feature cache: " + featureCache);
            throw e;
        }
    }

    /**
     * Builds the context representation of text in an element. Returns a
     * feature vector of all context features considered relevant for
     * representation.
     *
     * @param element the element whose context should be represented
     * @param leftText textual content to the left of (preceding)
     * <code>mainText</code>, might be empty
     * @param mainText the main textual content to represent, might be empty
     * @param rightText textual content to the right of (following)
     * <code>mainText</code>, might be empty
     * @param priorRecognitions a buffer of the last {@link Recognition}s from
     * the document, created by calling {@link #createRecognitionBuffer()};
     * might be <code>null</code>
     * @param featureCache a cache of (local) feature, should be re-used
     * between all calls for the nodes in a single document (but must not be
     * re-used when building the context of nodes in different documents!)
     * @param logPurpose the type of contexts of main interest to the caller
     * (e.g. "Token" or "Sentence"), used for logging
     * @return a vector of features considered relevant for representation
     * @throws ClassCastException if the <code>priorRecognitions</code> buffer
     * contains objects that aren't {@link Recognition}s
     * @throws IllegalArgumentException if the specified node is of an
     * unsupported type
     */
    protected abstract FeatureVector doBuildContext(final Element element,
            final String leftText, final String mainText,
            final String rightText, final PriorRecognitions priorRecognitions,
            final Map<Element, List<LocalFeature>> featureCache,
            final String logPurpose)
    throws ClassCastException, IllegalArgumentException;

    /**
     * Returns the maximum number of subsequences to keep when a feature
     * value must be split (at whitespace).
     *
     * @return the maximum number
     */
    public int getSplitMaximum() {
        return splitMaximum;
    }

    /**
     * Each <em>storeN</em>-th context representation is stored for debugging
     * and inspection purposes (if &gt; 0, otherwise no representation is
     * stored).
     *
     * @return the value of the attribute
     */
    public int getStoreN() {
        return storeN;
    }

    /**
     * Checks whether the given representation should be stored for inspection
     * purposes and stores it, if required. This method should be only called
     * if {@link #storeN} &gt; 0.
     *
     * @param representation the representation to store, if required
     * @param type the type of contexts of main interest to the caller
     * (e.g. "Token" or "Sentence"), used when storing the representation
     */
    private void storeIfNth(final FeatureVector representation,
            final String type) {
        File storeFile = null;
        // locally store count because the global value might be unstable
        long localCount = -1;

        // must synchronize while checking count and determining file name
        synchronized (lastStoredSuffix) {
            ++repCount;
            localCount = repCount;
            if (repCount % storeN == 0) {
                // we should store this one -- determine next available
                // file name in default directory
                storeFile = IOUtils.createOutFile(IOUtils.getDefaultDirectory(),
                    type + "Context", "rep", lastStoredSuffix);
            }
        }

        // no need to synchronize actual storing process (if required)
        if (storeFile != null) {
            Writer writer = null;
            try {
                writer = IOUtils.openWriter(storeFile, outputCharset);
                final CharSequence flatRep =
                    representation.flatten(true);
                writer.write(flatRep.toString());
                writer.flush();
                Util.LOG.info("Stored " + type + " representation no. "
                        + localCount + " in file " + storeFile
                        + " for inspection purposes");
            } catch (IOException ie) {
                Util.LOG.error("Error while storing " + type
                        + " representation no. " + localCount + " in file "
                        + storeFile + " for inspection purposes", ie);
            } finally {
                IOUtils.tryToClose(writer);
            }
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("split maximum", splitMaximum)
            .append("store n-th", storeN)
            .toString();
    }

}
