/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.context;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.collections.Bag;
import org.apache.commons.collections.bag.HashBag;
import org.apache.commons.collections.KeyValue;
import org.apache.commons.collections.MultiHashMap;
import org.apache.commons.collections.MultiMap;
import org.apache.commons.collections.SortedBag;
import org.apache.commons.collections.bag.TreeBag;
import org.apache.commons.collections.map.LinkedMap;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Attribute;
import org.dom4j.Element;

import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.xml.dom.DOMUtils;
import de.fu_berlin.ties.classify.feature.DefaultFeatureVector;
import de.fu_berlin.ties.classify.feature.Feature;
import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.context.sensor.BaseSensor;
import de.fu_berlin.ties.context.sensor.Sensor;
import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.text.TextUtils;
import de.fu_berlin.ties.util.Util;

/**
 * The context representation used by default. This class is thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.60 $, $Date: 2004/10/12 15:55:47 $, $Author: siefkes $
 */
public class DefaultRepresentation extends AbstractRepresentation {

    /**
     * Ancestor axis.
     */
    protected static final String AXIS_ANCESTOR = "an";

    /**
     * Descendant-or-self axis.
     */
    protected static final String AXIS_DESC_OR_SELF = "ds";

    /**
     * Following sibling axis.
     */
    protected static final String AXIS_FOLLOW_SIBLING = "fs";

    /**
     * Preceeding sibling axis.
     */
    protected static final String AXIS_PREC_SIBLING = "ps";

    /**
     * The pseudo-axis of prior recognitions.
     */
    protected static final String AXIS_PRIOR = "prior";

    /**
     * A sequence map mapping used by
     * {@link #calculateValuesFromText(String, String, List)} to determine the
     * "tokenType" value. Maps description strings to {@link Pattern}s to
     * match. The description string for the first matching pattern is used.
     * Initialized in the static constuctor. The value "mixed" is reserved for
     * tokens that are not matched by any of the patterns.
     */
    protected static final LinkedMap TOKEN_TYPE_PATTERNS;

    /**
     * Static constructor: initializes the map of {@link #TOKEN_TYPE_PATTERNS}.
     */
    static {
        // some constants for defining the patterns:
        // lowercase or modifier letter
        final String lowercaseLetter = "[\\p{Ll}\\p{Lm}]";
        // uppercase, titlecase, or modifier letter
        final String uppercaseLetter = "[\\p{Lu}\\p{Lt}\\p{Lm}]";
        // any Unicode letter or mark
        final String anyLetter = "[\\p{L}\\p{M}]";

        TOKEN_TYPE_PATTERNS = new LinkedMap();

        // different kinds of words (letters):
        TOKEN_TYPE_PATTERNS.put("lower",
            Pattern.compile(lowercaseLetter + "+"));
        TOKEN_TYPE_PATTERNS.put("CAPS",
            Pattern.compile(uppercaseLetter + "+"));
        // lowercase letters after a single uppercase or titlecase letter
        TOKEN_TYPE_PATTERNS.put("Cap",
            Pattern.compile(uppercaseLetter + lowercaseLetter + "+"));
        TOKEN_TYPE_PATTERNS.put("mixedCase", Pattern.compile(anyLetter + "+"));
        // letters + trailing dot
        TOKEN_TYPE_PATTERNS.put("CAPSDot",
            Pattern.compile(uppercaseLetter + "+\\."));
        TOKEN_TYPE_PATTERNS.put("alphaDot",
            Pattern.compile(anyLetter + "+\\."));
        // letters (incl. marks) + dashes ("-" etc.)
        TOKEN_TYPE_PATTERNS.put("alpha+Punc",
            Pattern.compile("[\\p{L}\\p{M}\\p{Pd}]+\\."));

        // different kinds of punctation
        // opening and closing puncuation
        TOKEN_TYPE_PATTERNS.put("openPunc", Pattern.compile("\\p{Ps}+"));
        TOKEN_TYPE_PATTERNS.put("closePunc", Pattern.compile("\\p{Pe}+"));
        /* not used -- unknown in JDK 1.4.2
        // puncation for opening or closing quotes
        TOKEN_TYPE_PATTERNS.put("initialQuote", Pattern.compile("\\p{Pi}+"));
        TOKEN_TYPE_PATTERNS.put("finalQuote", Pattern.compile("\\p{Pf}+"));
        */
        // any punctuation
        TOKEN_TYPE_PATTERNS.put("punc", Pattern.compile("\\p{P}+"));

        // different kinds of numbers
        // numbers cannot start with 0 (a single 0 is ok)
        TOKEN_TYPE_PATTERNS.put("num", Pattern.compile("(?:0|[1-9]\\d*)"));
        // digits can start with '0' and can also contain other Unicode digits
        TOKEN_TYPE_PATTERNS.put("digits", Pattern.compile("\\p{N}+"));
        TOKEN_TYPE_PATTERNS.put("digits+.", Pattern.compile("[\\p{N}.]+"));
        TOKEN_TYPE_PATTERNS.put("digits+.,",
            Pattern.compile("[\\p{N}.,]+"));
        //TOKEN_TYPE_PATTERNS.put("digits+:", Pattern.compile("[\\p{N}:]+"));

        // alphanumeric also includes dot + comma + dashes
        TOKEN_TYPE_PATTERNS.put("alphanum",
            Pattern.compile("[\\p{L}\\p{M}\\p{N}\\p{Pd}.,]+"));

        // different kinds of symbols
        // mathematical and currency symbols
        TOKEN_TYPE_PATTERNS.put("math", Pattern.compile("\\p{Sm}+"));
        TOKEN_TYPE_PATTERNS.put("curr", Pattern.compile("\\p{Sc}+"));
        // any symbols
        TOKEN_TYPE_PATTERNS.put("symb", Pattern.compile("\\p{S}+"));

        // control and other characters
        TOKEN_TYPE_PATTERNS.put("ctrl", Pattern.compile("\\p{C}+"));
    }

    /**
     * The maximum number of ancestors to include in the context representation.
     */
    private final int ancestorNumber;

    /**
     * The name of the attribute to use for calculating head values,
     * cf. {@link #calculateHeadValues(Element, List)}. The name must be
     * compatible to {@link DOMUtils#name(Attribute)}.
     */
    private final String headAttribute;

    /**
     * The name of the element to use for calculating head values,
     * cf. {@link #calculateHeadValues(Element, List)}. The name must be
     * compatible to {@link DOMUtils#name(Element)}.
     */
    private final String headElement;

    /**
     * An unmodifiable set of names of default attributes. Contains Strings
     * in a format compatible to {@link DOMUtils#name(Attribute)}.
     */
    private final Set defaultAttributes;

    /**
     * The number of preceding recognitions to represent in detail.
     */
    private final int detailedRecognitions;

    /**
     * The basic number of preceding and following siblings to include in the
     * context representation.
     */
    private final int siblingNumber;

    /**
     * The maximum length of prefixed and suffixes.
     */
    private final int maximumPrefixLength;

    /**
     * An array of sensors used to add semantic information.
     */
    private final Sensor[] sensors;

    /**
     * Creates a new instance based on the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @throws ProcessingException if an error occurs while initializing this
     * instance
     */
    public DefaultRepresentation() throws ProcessingException {
        this(TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance based on the provided configuration.
     *
     * @param config used to configure this instance
     * @throws ProcessingException if an error occurs while initializing this
     * instance
     */
    public DefaultRepresentation(final TiesConfiguration config)
    throws ProcessingException {
        this(config.getInt(CONFIG_RECOGN_NUM),
            config.getInt("representation.recogn.detailed"),
            config.getInt("representation.ancestor.num"),
            config.getInt("representation.sibling.num"),
            config.getInt(CONFIG_SPLIT_MAXIMUM),
            config.getInt("representation.prefix.maxlength"),
            config.getString("representation.head.element"),
            config.getString("representation.head.attrib"),
            config.getStringArray("representation.default.attribs"),
            config.getInt(CONFIG_STORE_NTH),
            config.getString(IOUtils.KEY_LOCAL_CHARSET, null),
            config.getStringArray("representation.sensors"),
            config);
    }

    /**
     * Creates a new instance. All element/attribute names must be compatible
     * to {@link DOMUtils#name(Element)}.
     *
     * @param recogNum the number of preceding recognitions to represent
     * @param detailedRecogs the number of preceding recognitions to represent
     * in detail
     * @param numberOfAncestors the maximum number of ancestors to include in
     * the context representation
     * @param numberOfSiblings the basic number of preceding and following
     * siblings to include
     * @param splitMax the maximum number of subsequences to keep when
     * a feature value must be split (at whitespace).
     * @param prefixMax the maximum length of prefixed and suffixes
     * @param headElementName the name of the element to use for
     * calculating head values (cf. {@link #calculateHeadValues(Element, List)})
     * @param headAttribName the name of the attribute to use for
     * calculating head values
     * @param defaultAttribs array of names of default attributes
     * @param outCharset the output character set to use (only used to
     * store some configurations for inspection purposes, if <code>n</code>
     * &gt; 0); if <code>null</code>, the default charset of the current
     * platform is used
     * @param n Each <em>n</em>-th context representation is stored if &gt; 0;
     * otherwise no representation is stored
     * @param sensorNames array of fully specified names of classes
     * implementing the {@link Sensor} interface to be used to look up semantic
     * information
     * @param config used to configure the sensors
     * @throws ProcessingException if an error occurs while initializing this
     * instance
     */
    public DefaultRepresentation(final int recogNum, final int detailedRecogs,
            final int numberOfAncestors, final int numberOfSiblings,
            final int splitMax, final int prefixMax,
            final String headElementName, final String headAttribName,
            final String[] defaultAttribs, final int n,
            final String outCharset, final String[] sensorNames,
            final TiesConfiguration config) throws ProcessingException {
        super(recogNum, splitMax, n, outCharset);
        detailedRecognitions = detailedRecogs;
        ancestorNumber = numberOfAncestors;
        siblingNumber = numberOfSiblings;
        maximumPrefixLength = prefixMax;

        headElement = headElementName;
        headAttribute = headAttribName;

        // create unmodifyable set of default attributes
        final Set<String> defAttribSet = new HashSet<String>();
        for (int i = 0; i < defaultAttribs.length; i++) {
            defAttribSet.add(defaultAttribs[i]);
        }
        defaultAttributes = Collections.unmodifiableSet(defAttribSet);
        // initialize sensors
        sensors = BaseSensor.createSensors(sensorNames, config);
    }

    /**
     * Builds the context representation of text in an element. Returns a
     * feature vector of all context features considered relevant for
     * representation.
     *
     * @param element the element whose context should be represented
     * @param leftText textual content to the left of (preceding)
     * <code>mainText</code>, might be empty
     * @param mainText the main textual content to represent, might be empty
     * @param rightText textual content to the right of (following)
     * <code>mainText</code>, might be empty
     * @param priorRecognitions a buffer of the last {@link Recognition}s from
     * the document, created by calling {@link #createRecognitionBuffer()};
     * might be <code>null</code>
     * @param featureCache a cache of (local) feature, should be re-used
     * between all calls for the nodes in a single document (but must not be
     * re-used when building the context of nodes in different documents!)
     * @param logPurpose the type of contexts of main interest to the caller
     * (e.g. "Token" or "Sentence"), used for logging
     * @return a vector of features considered relevant for representation
     * @throws ClassCastException if the <code>priorRecognitions</code> buffer
     * contains objects that aren't {@link Recognition}s
     * @throws IllegalArgumentException if the specified node is of an
     * unsupported type
     */
    protected FeatureVector doBuildContext(final Element element,
            final String leftText, final String mainText,
            final String rightText, final PriorRecognitions priorRecognitions,
            final Map<Element, List<LocalFeature>> featureCache,
            final String logPurpose)
            throws ClassCastException, IllegalArgumentException {
        LinkedList<Feature> selfFeatures = new LinkedList<Feature>();
        LinkedList<Feature> precedingFeatures = new LinkedList<Feature>();
        LinkedList<Feature> followingFeatures = new LinkedList<Feature>();
        LinkedList<Feature> ancestorFeatures = new LinkedList<Feature>();
        LinkedList<Feature> ancestorSiblingFeatures = new LinkedList<Feature>();
        final FeatureVector fullRep = new DefaultFeatureVector();

        // build pseudo-axis of prior recognitions, if any
        List priorFeatures;
        if ((getRecognitionNumber() > 0) && (priorRecognitions != null)) {
            priorFeatures = buildPrior(priorRecognitions);
        } else {
            priorFeatures = null;
        }

        // build sibling axes
        ElementPosition position =
            handleSiblings("", element, getSiblingNumber(),
            precedingFeatures, followingFeatures, featureCache);
/*           Util.LOG.debug("Build preceding (size: " + precedingFeatures.size()
                + ") + followingFeatures (size: + " + followingFeatures.size()
                + "; cache size: " + featureCache.size() + ")"); */

        // build descendant-or-self axis, prepending start marker
        selfFeatures.add(new GlobalFeature("Descendant-or-self axis"));
        selfFeatures.add(new GlobalFeature(AXIS_DESC_OR_SELF,
            LocalFeature.MARKER_START));
        buildFeatures(AXIS_DESC_OR_SELF, element, position, true,
            selfFeatures, true, featureCache);
        buildTextFeatures(AXIS_DESC_OR_SELF, element, leftText.trim(),
            mainText.trim(), rightText.trim(), selfFeatures);
/*            Util.LOG.debug("Build descendant-or-self (size: "
                + selfFeatures.size() + ", cache size: " + featureCache.size()
                + ")"); */

        // build ancestor (prepending start marker) + ancestor-siblings axes
        handleAncestors(element, getAncestorNumber(),
            getSiblingNumber() - 1, ancestorFeatures,
            ancestorSiblingFeatures, new HashBag(), featureCache);
        ancestorFeatures.addFirst(new GlobalFeature(AXIS_ANCESTOR,
            LocalFeature.MARKER_START));
        ancestorFeatures.addFirst(new GlobalFeature("Ancestor axis"));
/*            Util.LOG.debug("Build ancestors (size: " + ancestorFeatures.size()
                + ") incl. siblings (size: + " + ancestorSiblingFeatures.size()
                + "; cache size: " + featureCache.size() + ")"); */

        // combine all lists into full representation, start w/prior recog.s
        if (priorFeatures != null) {
            fullRep.addAll(priorFeatures);
            priorFeatures = null;
        }

        // preceding siblings + self + following siblings
        fullRep.addAll(precedingFeatures);
        precedingFeatures = null;
        fullRep.addAll(selfFeatures);
        fullRep.addAll(followingFeatures);
        followingFeatures = null;

        // ancestors + self + ancestor-siblings:
        fullRep.addAll(ancestorFeatures);
        ancestorFeatures = null;
        fullRep.addAll(selfFeatures);
        selfFeatures = null;
        fullRep.addAll(ancestorSiblingFeatures);
        ancestorSiblingFeatures = null;
/*            Util.LOG.debug("Combined axes into single representation (size: "
                + fullRep.size() + ", cache size: " + featureCache.size()
                + ")"); */

        // create and append filtered representations
        final List filteredRep = filterRepresentation(fullRep);

        // remove unnecessary markers from filtered rep.s and add them
        removeExtraMarkers(filteredRep);
        fullRep.addAll(filteredRep);

        Util.LOG.debug("Build " + logPurpose + " representation for "
            + DOMUtils.showToken(element, mainText) + " (full size: "
            + fullRep.size() + ", cache size: " + featureCache.size()
            + ")");

        return fullRep;
    }

    /**
     * Builds the features of an element and appends them to the
     * specified <code>featureList</code>. Handles attributes and calculated
     * values and the element itself. Child elements are only handled when
     * <code>recurseInsteadOfText</code> it <code>true</code> -- the axis name
     * is not changed for child elements.
     *
     * @param axisName the name of the axis, used to start each feature
     * @param element the element to process
     * @param position wraps the position of the element within its parent
     * element and related data, used for calculating positional features;
     * if <code>null</code>, no positional features are calculated
     * @param recurseInsteadOfText if <code>true</code> child elements
     * are recursively processed, but no
     * {@linkplain #calculateValuesFromText(String, String, List) values from
     * text} are calculated; otherwise text is processed but child elements are
     * not
     * @param featureList the list of {@link GlobalFeature}s to add features to
     * @param addAtEnd whether to add the new features at the end of at the
     * beginning of the <code>featureList</code>
     * @param cache a cache of local feature, mapping {@link Element}s to
     * {@link List}s of {@link LocalFeature}
     */
    protected void buildFeatures(final String axisName, final Element element,
            final ElementPosition position, final boolean recurseInsteadOfText,
            final LinkedList<Feature> featureList, final boolean addAtEnd,
            final Map<Element, List<LocalFeature>> cache) {
        final List<LocalFeature> localFeatures;

        if ((!recurseInsteadOfText) && cache.containsKey(element)) {
            // we can only use the cache when textual contents are included
            localFeatures = cache.get(element);
/*            Util.LOG.debug("Re-used cached features for "
                + DOMUtils.showElement(element)); */
        } else {
            // build local features, with or without textual features
            localFeatures = buildLocalFeatures(element, position,
                recurseInsteadOfText);

            // cache them if the are complete (including textual features
            if (!recurseInsteadOfText) {
                cache.put(element, localFeatures);
            }
        }

        // convert to global features and append to list
        final Iterator<LocalFeature> localIter = localFeatures.iterator();
        GlobalFeature.globalize(axisName, localIter, featureList, addAtEnd);
        //Util.LOG.debug("Globalized local features");

        if (recurseInsteadOfText) {
            // process child elements, without calculating positional values
            final Iterator childIter = element.elementIterator();
            Element child;
            while (childIter.hasNext()) {
                child = (Element) childIter.next();
                buildFeatures(axisName, child, null, recurseInsteadOfText,
                    featureList, addAtEnd, cache);
            }
        }
    }

    /**
     * Builds the local features of an element. Handles attributes and
     * calculated values and the element itself, but no child elements.
     * The {@link LocalFeature}s can be stored in a cache for re-use; they must
     * be combined with an axis name for getting a global feature.
     *
     * @param element the element to process
     * @param position wraps the position of the element within its parent
     * element and related data, used for calculating positional features;
     * if <code>null</code>, no positional features are calculated
     * @param ignoreText if <code>true</code>, no {@linkplain
     * #calculateValuesFromText(String, String, List) values from text} or
     * {@linkplain #calculateHeadValues(Element, List) head values} are
     * calculated
     * @return returns a List of created {@link LocalFeature}s
     */
    protected List<LocalFeature> buildLocalFeatures(final Element element,
            final ElementPosition position, final boolean ignoreText) {
        final List<LocalFeature> result = new LinkedList<LocalFeature>();
        final String name = DOMUtils.name(element);

        // Process attributes sorted according to their names.
        // Map contains name/value mappings.
        final SortedMap<String, String> attribMap =
            new TreeMap<String, String>();
        final Iterator attribIter = element.attributeIterator();
        Attribute currentAttrib;
        String attName;
        boolean foundDefaultAttrib = false;

        while (attribIter.hasNext()) {
            currentAttrib = (Attribute) attribIter.next();
            attName = DOMUtils.name(currentAttrib);
            attribMap.put(attName, currentAttrib.getValue());

            if (!foundDefaultAttrib
                    && getDefaultAttributes().contains(attName)) {
                // at least one of the default attributes is present
                foundDefaultAttrib = true;
            }
        }

        // no default attribute present: include stand-alone element name
        if (!foundDefaultAttrib) {
            result.add(LocalFeature.createElementFeature(name));
            //Util.LOG.debug("Created element feature for " + name);
        }

        final Iterator<String> sortedIter = attribMap.keySet().iterator();
        String attValue;

        while (sortedIter.hasNext()) {
            attName = sortedIter.next();
            attValue = attribMap.get(attName);
            //Util.LOG.debug("Creating " + attName + " attribute feature");
            result.addAll(LocalFeature.createAttributeFeatures(name, attName,
                attValue, getSplitMaximum()));
            //Util.LOG.debug("Created " + attName + " attribute feature");
        }

        //Util.LOG.debug("Created attribute features");

        // calculate positional values, if there is a parent
        if (position != null) {
            calculatePositionalValues(name, position, result);
            //Util.LOG.debug("Calculated positional values");
        }

        if (!ignoreText) {
            // check textual content
            final String trimmedText = element.getTextTrim();

            if ((trimmedText != null) && (trimmedText.length() > 0)) {
                // calculate values from text incl. text itself
                calculateValuesFromText(name, trimmedText, result);
            } else {
                // check whether there are head values instead
                 calculateHeadValues(element, result);
            }
        }

        return result;
    }

    /**
     * Builds the pseudo-axis of prior recognitions.
     *
     * @param priorRecognitions a buffer of the last {@link Recognition}s from
     * the document, created by calling {@link #createRecognitionBuffer()}
     * @return a list of {@link GlobalFeature}s representing prior recognitions
     * @throws ClassCastException if the <code>priorRecognitions</code> buffer
     * contains objects that aren't {@link Recognition}s
     */
    protected List<Feature> buildPrior(
            final PriorRecognitions priorRecognitions)
            throws ClassCastException {
        final LinkedList<Feature> result =
            new LinkedList<Feature>();

        // start with comment and start marker
        result.add(new GlobalFeature("Prior recognitions"));
        result.add(new GlobalFeature(AXIS_PRIOR, LocalFeature.MARKER_START));

        // prepare to iterate prior recognitions
        final Iterator recogIter = priorRecognitions.iterator();
        PriorRecognitions.Pair currentPair;
        Recognition currentRecognition;
        String type;
        String trimmedText;
        LinkedList<Feature> recogFeatures;
        final List<LocalFeature> calculatedValues =
            new LinkedList<LocalFeature>();

        // number of recognitions to describe without details
        final int briefRecogs =
            priorRecognitions.size() - getDetailedRecognitions();

        // log an error if the buffer is too large
        if (priorRecognitions.size() > getRecognitionNumber()) {
            Util.LOG.error("Buffer of prior recognitions too large ("
                + priorRecognitions.size() + " instead of "
                + getRecognitionNumber() + " elements) -- use "
                + "Representation.createRecognitionBuffer() to create it");
        }

        // process prior recognitions
        for (int i = 0; recogIter.hasNext(); i++) {
            currentPair = (PriorRecognitions.Pair) recogIter.next();
            currentRecognition = currentPair.getRecognition();
            type = currentRecognition.getType();

            // check whether features are cached
            recogFeatures = currentPair.getCachedFeatures();
            if (recogFeatures == null) {
                // features are not yet cached
                recogFeatures = new LinkedList<Feature>();

                // include feature type for all recognitions
                recogFeatures.add(new GlobalFeature(AXIS_PRIOR,
                    LocalFeature.createElementFeature(type)));

                if (i >= briefRecogs) {
                    // include further details for last n=detailedRecognitions:
                    // calculate values from the text incl. text itself (if any)
                    trimmedText = currentRecognition.getText().trim();
                    if (StringUtils.isNotEmpty(trimmedText)) {
                        calculateValuesFromText(type, trimmedText,
                                calculatedValues);

                            // store as global features and clear list
                            GlobalFeature.globalize(AXIS_PRIOR,
                                    calculatedValues.iterator(), recogFeatures,
                                    true);
                            calculatedValues.clear();
                    }
                }

                // cache features if the recognition is sealed
                if (currentRecognition.isSealed()) {
                    currentPair.setCachedFeatures(recogFeatures);
                }
            } else {
                // features are cached

                if (i < briefRecogs) {
                    // only include type (very first feature) for oldest
                    // recognitions -- discard all other features
                    while (recogFeatures.size() > 1) {
                        recogFeatures.removeLast();
                    }
                }
            }

            // add features of recognition to result list
            result.addAll(recogFeatures);

            if (!recogIter.hasNext()) {
                // last recognition: check whether it is finished (sealed)
                result.addLast(new GlobalFeature(AXIS_PRIOR,
                    LocalFeature.createCalculatedFeature(type, "finished",
                        String.valueOf(currentRecognition.isSealed()))));
            }
        }
        return result;
    }

    /**
     * Builds the context representation of text in an element, differentiating
     * between three kinds of textual contents: a left part, a main part, and
     * a right part. Each text part is optional (if all of them are empty,
     * this method does nothing).
     *
     * @param axisName the name of the axis, used to start each feature
     * @param element the element whose context should be represented
     * @param trimmedLeft trimmed textual content to the left of (preceding)
     * <code>trimmedMain</code>, might be empty
     * @param trimmedMain trimmed main textual content to represent, might
     * be empty
     * @param trimmedRight trimmed textual content to the right of (following)
     * <code>trimmedMain</code>, might be empty
     * @param featureList a list of {@link GlobalFeature}s to add the values to
     */
    protected void buildTextFeatures(final String axisName,
            final Element element, final String trimmedLeft,
            final String trimmedMain, final String trimmedRight,
            final LinkedList<Feature> featureList) {
        final String name = DOMUtils.name(element);
        final List<LocalFeature> localTextFeatures =
            new LinkedList<LocalFeature>();

        // the feature mark for calculated values to used to introduce left
        // + right features
        final String innerMarker = FeatureType.CALCULATED.getMark();

        if (trimmedLeft.length() > 0) {
            calculateValuesFromText(name + innerMarker + "left", trimmedLeft,
                localTextFeatures);
        }
        if (trimmedMain.length() > 0) {
            calculateValuesFromText(name, trimmedMain, localTextFeatures);
        }
        if (trimmedRight.length() > 0) {
            calculateValuesFromText(name + innerMarker + "right", trimmedRight,
                localTextFeatures);
        }

        // globalize + append features
        GlobalFeature.globalize(axisName, localTextFeatures.iterator(),
            featureList, true);
    }

    /**
     * Creates values that depend on "head" children of an element, if the
     * element contains any of them. This method is meant for element that
     * don't contain any textual content, so their children are checked
     * instead.
     *
     * This implementation proceeds as follows:
     * <ul>
     * <li>If the specified element is of type {@link #getHeadElement()}, the
     * value of the {@link #getHeadAttribute()} or else the textual content of
     * its right-most child element is stored (in a calculated value named
     * "head"). Child elements are iterated from right to left unless one
     * containing an appropriate attribute or textual content is found (or none
     * are left).
     * <li>Otherwise, if the element contains children of the type
     * {@link #getHeadElement()}, the first and last child element of this type
     * are recursively processed and the results stored in calculated names
     * named "lhead" and "rhead". Of there is only a single child of appropriate
     * type, the "lhead" value is omitted.
     * <li>Otherwise no values are calculated.
     * </ul>
     *
     * <p>If a value contains whitespace, only the final subsequence following
     * all whitespace is preserved.
     *
     * @param element the element to process
     * @param values a list of {@link LocalFeature}s to add the calculated
     * values to
     */
    protected void calculateHeadValues(final Element element,
            final List<LocalFeature> values) {
        final String elementName = DOMUtils.name(element);
        if (getHeadElement().equals(elementName)) {
            // element is of suitable type
            final String headValue = determineHeadValue(element);

            // store if not empty
            if (headValue.length() > 0) {
                values.add(LocalFeature.createCalculatedFeature(elementName,
                    "head", headValue));
            }
        } else {
            // process children of suitable type
            final List suitableChildren = DOMUtils.elementsByName(element,
                getHeadElement());
            final int numberOfChildren = suitableChildren.size();

            if (numberOfChildren >= 1) {
                // process right-most child
                final Element rightMostChild =
                    (Element) suitableChildren.get(numberOfChildren - 1);
                final String rHeadValue = determineHeadValue(rightMostChild);

                // store if not empty
                if (rHeadValue.length() > 0) {
                    values.add(LocalFeature.createCalculatedFeature(
                        elementName, "rhead", rHeadValue));
                }

                if (numberOfChildren >= 2) {
                    // process left-most child
                    final Element leftMostChild =
                        (Element) suitableChildren.get(0);
                    final String lHeadValue = determineHeadValue(leftMostChild);

                    // store if not empty
                    if (lHeadValue.length() > 0) {
                        values.add(LocalFeature.createCalculatedFeature(
                            elementName, "lhead", lHeadValue));
                    }
                }
            }
        }
    }

    /**
     * Calculates values that depend on the position of an element within its
     * parent.
     *
     * @param elementName the name of the element to process, as returned by
     * {@link DOMUtils#name(Element)}
     * @param position wraps the position of the element within its parent
     * element and related data, must not be <code>null</code>
     * @param values a list to add the calculated values to
     */
    protected void calculatePositionalValues(final String elementName,
            final ElementPosition position, final List<LocalFeature> values) {
        // store exact position, counting all children
        values.add(LocalFeature.createCalculatedFeature(elementName, "pos",
            Integer.toString(position.getOverallPosition())));

        // store exact position, counting only children with same type
        values.add(LocalFeature.createCalculatedFeature(elementName,
            "typedPos", Integer.toString(position.getTypedPosition())));

        // store rough positions
        values.add(LocalFeature.createCalculatedFeature(elementName,
            "roughPos", determineRoughPosition(position.getOverallPosition(),
                position.getAllChildren())));
        values.add(LocalFeature.createCalculatedFeature(elementName,
            "roughTypedPos", determineRoughPosition(position.getTypedPosition(),
                position.getTypedChildren())));
    }

    /**
     * Calculates values that depend on the textual content of an element:
     * prefixes, suffixes, length data, and token type. Also includes the text
     * itself.
     *
     * @param elementName the name of the element to process, as returned by
     * {@link DOMUtils#name(Element)}
     * @param trimmedText the trimmed textual content of the element to
     * process, must not be empty
     * @param values a list of {@link LocalFeature}s to add the calculated
     * values to
     * @throws IllegalArgumentException if the empty string was given as
     * <code>trimmedText</code>
     */
    protected void calculateValuesFromText(final String elementName,
            final String trimmedText, final List<LocalFeature> values)
            throws IllegalArgumentException {
        final int textLength = trimmedText.length();
        if (textLength == 0) {
            throw new IllegalArgumentException(
                "Specifed trimmed text must not be empty!");
        }

        // split at whitespace, if any
        final String[] tokens = TextUtils.splitString(trimmedText, -1);
        final String firstTokenLC = tokens[0].toLowerCase();
        final String lastTokenLC;

        if (tokens.length == 1) {
            // no whitespace -- single token
            lastTokenLC = firstTokenLC;
        } else {
            lastTokenLC = tokens[tokens.length - 1].toLowerCase();
        }

        // store starting 1--x letters from first token (but only if the
        // token contains enough letters; using lower case)
        final int maxPrefixLength = Math.min(maximumPrefixLength,
            firstTokenLC.length() - 1);
        for (int i = 1; i <= maxPrefixLength; i++) {
            values.add(LocalFeature.createCalculatedFeature(elementName,
                 "prefix" + i, firstTokenLC.substring(0, i)));
        }

        // store trailing 1--x letters from last token (but only if the
        // token contains enough letters; using lower case)
        final int maxSuffixLength = Math.min(maximumPrefixLength,
            lastTokenLC.length() - 1);
        for (int i = 1; i <= maxSuffixLength; i++) {
            values.add(LocalFeature.createCalculatedFeature(elementName,
                "suffix" + i, lastTokenLC.substring(lastTokenLC.length() - i)));
        }

        // store length
        values.add(LocalFeature.createCalculatedFeature(elementName, "length",
            Integer.toString(textLength)));

        // also store rounded square root of length as a less sparse feature
        final long sqrtLength = Math.round(Math.sqrt(textLength));
        values.add(LocalFeature.createCalculatedFeature(elementName,
            "sqrtLength", Long.toString(sqrtLength)));

        // determine the token type (of last token, if there are several)
        final Set typeSet = TOKEN_TYPE_PATTERNS.keySet();
        final Iterator typeIter = typeSet.iterator();
        String currentDesc;
        Pattern currentPattern;
        Matcher currentMatcher;
        boolean foundMatch = false;
        String matchingType = "mixed"; // use if no match is found

        // try each pattern until one of them matches (or none are left)
        while (typeIter.hasNext() && !foundMatch) {
            currentDesc = (String) typeIter.next();
            currentPattern = (Pattern) TOKEN_TYPE_PATTERNS.get(currentDesc);
            currentMatcher = currentPattern.matcher(tokens[tokens.length - 1]);

            if (currentMatcher.matches()) {
                foundMatch = true;
                matchingType = currentDesc; // overwrite value
            }
        }

        // append "+multi" if there are several whitespace-separated tokens
        final String fullType = (tokens.length > 1)
            ? matchingType + "+multi" : matchingType;
        values.add(LocalFeature.createCalculatedFeature(elementName,
            "tokenType", fullType));

        // lookup information from sensors (for the whole text, not per token)
        KeyValue[] sensorData;
        String prefix;
        int j;

        for (int i = 0; i < sensors.length; i++) {
            // prefix keys of each sensor with "i$" (e.g. "0$") toi avoid
            // naming collisions
            prefix = i + FeatureType.CALCULATED.getMark();
            sensorData = sensors[i].lookup(trimmedText);

            for (j = 0; j < sensorData.length; j++) {
                values.add(LocalFeature.createCalculatedFeature(elementName,
                    prefix + sensorData[j].getKey(),
                    Util.asString(sensorData[j].getValue())));
            }
        }

        // add textual content itself -- in lower case and "as is"
        values.addAll(LocalFeature.createCalculatedFeatures(elementName, "lc",
            trimmedText.toLowerCase(), getSplitMaximum()));
        values.addAll(LocalFeature.createTextFeatures(elementName, trimmedText,
            getSplitMaximum()));
    }

    /**
     * Helper method for determining the head value for an element of type
     * {@link #getHeadElement()}. See
     * {@link #calculateHeadValues(Element, List)} for a description of the
     * algorithm.
     *
     * @param element the element to process, must be of type
     * {@link #getHeadElement()}
     * @return the calculated head value, or the empty string if no value
     * could be calculated (the specified element doesn't have any children,
     * or the right-most child contains neither {@link #getHeadAttribute()} nor
     * textual content)
     */
    protected String determineHeadValue(final Element element) {
        String fullResult = null;
        final List children = element.elements();
        final ListIterator childIter = children.listIterator(children.size());
        Element currentChild;
        Attribute headAttrib;
        boolean done = false;

        // iterate children from right to left unless required attribute or
        // textual content is present
        while (childIter.hasPrevious() && !done) {
            currentChild = (Element) childIter.previous();
            headAttrib = DOMUtils.attributeByName(currentChild,
                getHeadAttribute());

            // check attribute value or else trimmed textual content
            if (headAttrib != null) {
                fullResult = headAttrib.getText();
            } else {
                fullResult = element.getTextTrim();
            }

            // we're done when we got an non-empty string
            done = (fullResult != null) && (fullResult.length() > 0);
        }

        final String result;
        if (fullResult != null) {
            // split text, only keeping last subsequence after any whitespace
            final String[] splits = TextUtils.splitString(fullResult, 1);
            result = splits[0];
        } else {
            result = "";
        }
        return result;
    }

    /**
     * Helper method called by
     * {@link #calculatePositionalValues(String, ElementPosition, List)} to
     * collapse a position in to one of five values.
     * <dl>
     * <dt>first</dt>
     * <dd>for the first element</dd>
     * <dt>early</dt>
     * <dd>if position is within the first third of all elements (but not the
     * first one), upper limit included</dd>
     * <dt>middle</dt>
     * <dd>if position is within the second third of all elements, limits
     * excluded</dd>
     * <dt>late</dt>
     * <dd>if position is within the last third of all elements (but not the
     * last one), lower limit included</dd>
     * <dt>last</dt>
     * <dd>for the last element</dd>
     * </dl>
     *
     * @param position the position counted from 0, should be non-negative and
     * smaller than <code>elementCount</code> (otherwise the results are
     * undefined)
     * @param elementCount the number of elements
     * @return a String containing one of the five values described above
     */
    protected String determineRoughPosition(final int position,
            final int elementCount) {
        // for easier counting
        final int realPosition = position + 1;
        final String result;

        if (realPosition == 1) {
            result = "first";
        } else if (realPosition == elementCount) {
            result = "last";
        } else {
            final double oneThird = elementCount / 3.0;
            if (realPosition <= oneThird) {
                result = "early";
            } else if (realPosition >= (2 * oneThird)) {
                result = "late";
            } else {
                result = "middle";
            }
        }
        return result;
    }

    /**
     * Creates a filtered view of a context representation. A filtered
     * representation is created for each {@linkplain FeatureType type of
     * features} that occurs at least twice;
     * attributes and calculated values occuring only once are omitted.
     * For textual content, another filtered representation is created.
     *
     * <p>Features representing markers ({@link FeatureType#MARKER}),
     * stand-alone elements ({@link FeatureType#ELEMENT}) and default attributes
     * ({@link #getDefaultAttributes()}) are included in all filtered
     * representations. Comment-only features are ignored.
     *
     * @param originalRep a feature vector containing the representation
     * to filter
     * @return a list of {@link de.fu_berlin.ties.classify.feature.Feature}s
     * combining the filtered representations created by this method
     */
    protected List<Feature> filterRepresentation(
            final FeatureVector originalRep) {
        // features to include in all views (markers, stand-alone elements,
        // default attributes)
        final List<Feature> universalFeatures = new LinkedList<Feature>();
        // textual features
        final LinkedList<Feature> textualFeatures = new LinkedList<Feature>();
        // multi-map mapping mark + name of attributes/calculated values to
        // LinkedLists of filtered features
        final MultiMap featureMap = new MultiHashMap();
        // count features of each attribute/calculated type
        final SortedBag featureCount = new TreeBag();

        final Iterator iter = originalRep.iterator();
        GlobalFeature currentGlobal;
        LocalFeature currentLocal;
        FeatureType currentType;
        Iterator keyIter;
        String currentKey;
        Iterator<Feature> universalIter;
        Iterator innerIter;
        Feature innerFeature;

        // iterate original representation
        while (iter.hasNext()) {
            currentGlobal = (GlobalFeature) iter.next();
            currentLocal = currentGlobal.getLocalFeature();

            if (currentLocal != null) { // otherwise it's a comment -- ignored
                currentType = currentLocal.getType();
                if (((currentType == FeatureType.ATTRIBUTE)
                            && (getDefaultAttributes().contains(
                                currentLocal.getName())))
                        || (currentType == FeatureType.MARKER)) {
                    // include in all views
                    universalFeatures.add(currentGlobal);
                    textualFeatures.add(currentGlobal);
                    keyIter = featureMap.keySet().iterator();
                    while (keyIter.hasNext()) {
                        currentKey = (String) keyIter.next();
                        featureMap.put(currentKey, currentGlobal);
                    }
                // use attributes, elements and calculated ...head values
                } else if ((currentType == FeatureType.ATTRIBUTE)
                          || (currentType == FeatureType.ELEMENT)
                          || ((currentType == FeatureType.CALCULATED)
                              && currentLocal.getName().endsWith("head"))) {
                    // determine key: use type marker
                    currentKey = currentType.getMark();

                    if (!featureMap.containsKey(currentKey)) {
                        // create new view, copying collected universal entries
                        universalIter = universalFeatures.iterator();
                        while (universalIter.hasNext()) {
                            featureMap.put(currentKey, universalIter.next());
                        }
                    }

                    // include in appropriate view + increase count
                    featureMap.put(currentKey, currentGlobal);
                    featureCount.add(currentKey);
                } else if (currentType == FeatureType.TEXT) {
                    // include in textual view
                    textualFeatures.add(currentGlobal);
                }
                // elements are ignored in filtered rep.s
            }
        }

        // combine all views
        final List<Feature> result = new LinkedList<Feature>();
        final Iterator sortedIter = featureCount.uniqueSet().iterator();
        Collection currentView;

        // add view for each suitable attribute/valculated value
        while (sortedIter.hasNext()) {
            currentKey = (String) sortedIter.next();

            // ignore attributes/calculated values that occur only once or twice
            if (featureCount.getCount(currentKey) > 2) {
                result.add(new GlobalFeature("Filtered view for " + currentKey
                    + " features"));
                currentView = (Collection) featureMap.get(currentKey);
                innerIter = currentView.iterator();

                while (innerIter.hasNext()) {
                    innerFeature = (Feature) innerIter.next();
                    result.add(innerFeature);                    
                }
            }
        }

        // add view for textual content
        result.add(new GlobalFeature("Filtered view for textual features"));
        result.addAll(textualFeatures);

        return result;
    }

    /**
     * Returns the maximum number of ancestors to include in the context
     * representation.
     *
     * @return the number of ancestors to include
     */
    public int getAncestorNumber() {
        return ancestorNumber;
    }

    /**
     * Returns the unmodifiable set of names of default attributes. The
     * contained names are Strings in a format compatible to
     * {@link DOMUtils#name(Attribute)}.
     *
     * @return the set of default attributes
     */
    public Set getDefaultAttributes() {
        return defaultAttributes;
    }

    /**
     * Returns the number of preceding recognitions to represent in detail.
     * @return the value of the attribute
     */
    public int getDetailedRecognitions() {
        return detailedRecognitions;
    }

    /**
     * Returns the name of the element to use for calculating head values.
     *
     * @return the name of the attribute
     * @see #calculateHeadValues(Element, List)
     */
    public String getHeadAttribute() {
        return headAttribute;
    }

    /**
     * Returns the name of the attribute to use for calculating head values.
     *
     * @return the name of the element
     * @see #calculateHeadValues(Element, List)
     */
    public String getHeadElement() {
        return headElement;
    }

    /**
     * Returns the basic number of preceding and following siblings to include
     * in the context representation.
     *
     * @return the number of siblings to include
     */
    public int getSiblingNumber() {
        return siblingNumber;
    }

    /**
     * Handles ancestors and ancestor siblings of an element.
     *
     * @param element the element to process
     * @param ancestorsToAdd the number of ancestors to add, must be &gt; 0;
     * if &gt; 1, this method calls itself recursively, decreasing the number
     * by 1
     * @param ancestorSiblingsToAdd the number of preceding/following siblings
     * of the current ancestors to add; if 0 or negative, no siblings are added;
     * a recursive call decreases this parameter by 1 if any ancestor siblings
     * were found (if no siblings were found, the number passes unchanged)
     * @param ancestorFeatures the list of {@link GlobalFeature}s to prepend
     * the features on the ancestors to
     * @param ancestorSiblingFeatures the list of {@link GlobalFeature}s to
     * append the features on the ancestors siblings to
     * @param processedAncestorNames a bag that will typically be empty when
     * first calling this method (will be filled by recursive calls)
     * @param cache a cache of local feature, mapping {@link Element}s to
     * {@link List}s of {@link LocalFeature}
     * @throws IllegalArgumentException if <code>ancestorsToAdd</code> is 0 or
     * negative
     */
    protected void handleAncestors(final Element element,
            final int ancestorsToAdd, final int ancestorSiblingsToAdd,
            final LinkedList<Feature> ancestorFeatures,
            final LinkedList<Feature> ancestorSiblingFeatures,
            final Bag processedAncestorNames,
            final Map<Element, List<LocalFeature>> cache)
            throws IllegalArgumentException {
        if (ancestorsToAdd <= 0) {
            throw new IllegalArgumentException(
                "Cannot add a non-positive number of ancestors ("
                    + ancestorsToAdd + ")!");
        }

        // handle parent
        final Element parent = element.getParent();
        if (parent != null) {
            // otherwise there is nothing to do (reached highest node)
            final ElementPosition position;

            // process siblings of parent, if required
            if (ancestorSiblingsToAdd > 0) {
                // use bag of names to determine axis prefix:
                final String parentName = DOMUtils.name(parent);
                processedAncestorNames.add(parentName);
                final int typeCount
                    = processedAncestorNames.getCount(parentName);
                final String nameToQuote;

                if (typeCount > 2) {
                    // repeated occurrence of this element type:
                    // prepend repetition counter to element name
                    nameToQuote = Integer.toString(typeCount) + parentName;
                } else {
                    // first occurrence of this element type: just use the name
                    nameToQuote = parentName;
                }
                final String axisPrefix =
                    AXIS_ANCESTOR + LocalFeature.quote(nameToQuote);

                // store both sides in shared temporary list which is then
                // appended to the specified list
                final LinkedList<Feature> siblingFeatures =
                    new LinkedList<Feature>();
                position = handleSiblings(axisPrefix, parent,
                    ancestorSiblingsToAdd, siblingFeatures, siblingFeatures,
                    cache);
                ancestorSiblingFeatures.addAll(siblingFeatures);
            } else {
                position = null;
            }

            // process parent itself, prepending to list
            buildFeatures(AXIS_ANCESTOR, parent, position, false,
                ancestorFeatures, false, cache);

            // recursively handle ancestors of parent, if required
            if (ancestorsToAdd > 1) {
                final int parentSilbingsToAdd;

                // determine how many siblings to process at the next step
                if ((position != null) && (position.getProcessedPreceding()
                        + position.getProcessedPreceding() > 0)) {
                    // processed one or more ancestor siblings: decrease number
                    parentSilbingsToAdd = ancestorSiblingsToAdd - 1;
                } else {
                    // no ancestor siblings processed: pass number unchange
                    parentSilbingsToAdd = ancestorSiblingsToAdd;
                }

                handleAncestors(parent, ancestorsToAdd - 1 ,
                    parentSilbingsToAdd, ancestorFeatures,
                    ancestorSiblingFeatures, processedAncestorNames, cache);
            }
        }
    }

    /**
     * Adds the preceding and following siblings of an element.
     *
     * @param axisPrefix the prefix of the axis name, used to start each
     * feature; specify the empty string if no prefix should be used
     * @param element the element to process
     * @param baseNumber the basic number of siblings to keep; the actual
     * number of siblings kept might vary
     * @param precedingFeatures the list of {@link GlobalFeature}s to prepend
     * the features on the preceding siblings to
     * @param followingFeatures the list of {@link GlobalFeature}s to append
     * the features on the following siblings to
     * @param cache a cache of local feature, mapping {@link Element}s to
     * {@link List}s of {@link LocalFeature}
     * @return a wrapper of the position of the element with its parent element
     * and related data, required by
     * {@link #calculatePositionalValues(String, ElementPosition, List)}, or
     * <code>null</code> if there is no parent element
     */
    protected ElementPosition handleSiblings(final String axisPrefix,
            final Element element, final int baseNumber,
            final LinkedList<Feature> precedingFeatures,
            final LinkedList<Feature> followingFeatures,
            final Map<Element, List<LocalFeature>> cache) {
        final Element parent = element.getParent();
        final String name = DOMUtils.name(element);
        final ElementPosition result;

        if (parent != null) {
            // otherwise there is nothing to do
            final Iterator childIter = parent.elementIterator();
            Element currentChild;
            boolean foundElement = false;
            LinkedList<Element> allPrecedingSiblings =
                new LinkedList<Element>();
            LinkedList<Element> allFollowingSiblings =
                new LinkedList<Element>();

            // counters for object to return
            int allChildren = 0;
            int overallPosition = 0;
            int typedChildren = 0;
            int typedPosition = 0;
            boolean sameType;

            while (childIter.hasNext()) {
                currentChild = (Element) childIter.next();

                // update child count
                allChildren++;
                if (name.equals(DOMUtils.name(currentChild))) {
                    sameType = true;
                    typedChildren++;
                } else {
                    sameType = false;
                }

                if (currentChild == element) {
                    foundElement = true;
                } else {
                    if (!foundElement) {
                        // add to preceding siblings
                        allPrecedingSiblings.addLast(currentChild);

                        // update position count
                        overallPosition++;
                        if (sameType) {
                            typedPosition++;
                        }
                    } else {
                        // add to following siblings
                        allFollowingSiblings.addLast(currentChild);
                    }
                }
            }

            // select siblings to keep
            final List selectedPrecedingSiblings = selectPrecedingSiblings(
                   element, allPrecedingSiblings, baseNumber);
            final List selectedFollowingSiblings = selectFollowingSiblings(
                    element, allFollowingSiblings, baseNumber);

            // create object to return
            result = new ElementPosition(allChildren, overallPosition,
                typedChildren, typedPosition, selectedPrecedingSiblings.size(),
                selectedFollowingSiblings.size());

            // name axes
            final StringBuffer precedingAxisName = new StringBuffer(axisPrefix);
            final StringBuffer followingAxisName = new StringBuffer(axisPrefix);
            final String prefixComment;
            if (axisPrefix.length() > 0) {
                prefixComment = " with '" + axisPrefix + "' prefix";
                precedingAxisName.append(GlobalFeature.SEP);
                followingAxisName.append(GlobalFeature.SEP);
            } else {
                prefixComment = ""; // no prefix
            }
            precedingAxisName.append(AXIS_PREC_SIBLING);
            followingAxisName.append(AXIS_FOLLOW_SIBLING);

            ListIterator siblingIter;
            Element currentSibling;

            // prepend preceding siblings, without calculating positional values
            siblingIter = selectedPrecedingSiblings.listIterator(
                selectedPrecedingSiblings.size());
            while (siblingIter.hasPrevious()) {
                currentSibling = (Element) siblingIter.previous();
                buildFeatures(precedingAxisName.toString(), currentSibling,
                        null, false, precedingFeatures, false, cache);
            }

            precedingFeatures.addFirst(
                new GlobalFeature(precedingAxisName.toString(),
                    LocalFeature.MARKER_START));
            precedingFeatures.addFirst(new GlobalFeature(
                "Preceding siblings axis" + prefixComment));

            // append following siblings, without calculating positional values
            followingFeatures.addLast(new GlobalFeature(
                            "Following siblings axis" + prefixComment));
            followingFeatures.addLast(
                new GlobalFeature(followingAxisName.toString(),
                    LocalFeature.MARKER_START));

            siblingIter = selectedFollowingSiblings.listIterator();
            while (siblingIter.hasNext()) {
                currentSibling = (Element) siblingIter.next();
                buildFeatures(followingAxisName.toString(), currentSibling,
                        null, false, followingFeatures, true, cache);
            }
        } else {
            // no parent
            result = null;
        }
        return result;
    }

    /**
     * Modifies a list of {@link GlobalFeature}s to remove extraneous
     * {@link FeatureType#MARKER} features. Keeps only the last one of several
     * sequential marker features; trailing marker features are removed as well.
     *
     * @param features the list of features to modify
     */
    protected void removeExtraMarkers(final List features) {
        // iterate backward for easier operation
        final ListIterator backwardIter =
            features.listIterator(features.size());
        GlobalFeature currentGlobal;
        LocalFeature currentLocal;
        FeatureType currentType;

        // initialized this way to remove trailing marker features
        FeatureType lastType = FeatureType.MARKER;

        while (backwardIter.hasPrevious()) {
            currentGlobal = (GlobalFeature) backwardIter.previous();
            currentLocal = currentGlobal.getLocalFeature();

            if (currentLocal != null) { // otherwise it's a comment -- ignored
                currentType = currentLocal.getType();

                if ((currentType == FeatureType.MARKER)
                        && (lastType == FeatureType.MARKER)) {
                    // remove extraneous marker feature
                    backwardIter.remove();
                }

                lastType = currentType;
            }
        }
    }

    /**
     * Selects the siblings to keep among all following siblings. This
     * implementation always keeps the <code>baseNumber</code> first siblings.
     * One of the selected siblings may have a different type (name as returned
     * by {@link DOMUtils#name(Element)}) than the main element --
     * if there are more with different types, they are skipped.
     *
     * @param mainElement the element whose siblings should be selected
     * @param allFollowingSiblings the list of all following siblings
     * @param baseNumber the basic number of siblings to keep; the actual
     * number of siblings kept might vary
     * @return the subset of following siblings to keep; an empty list if
     * there are no siblings to keep
     */
    protected List<Element> selectFollowingSiblings(final Element mainElement,
            final LinkedList<Element> allFollowingSiblings,
            final int baseNumber) {
        final String elementType = DOMUtils.name(mainElement);
        final LinkedList<Element> selectedFollowingSiblings =
            new LinkedList<Element>();
        final Iterator<Element> siblingIter = allFollowingSiblings.iterator();
        Element currentSibling;
        boolean addedSiblingOfDifferentType = false;
        boolean doAdd;

        // keep up to |baseNumber| siblings from start of list
        while (siblingIter.hasNext()
                && (selectedFollowingSiblings.size() < baseNumber)) {
            currentSibling = siblingIter.next();

            if (elementType.equals(DOMUtils.name(currentSibling))) {
                doAdd = true;
            } else {
                // only add if this is the first differently typed sibling
                doAdd = !addedSiblingOfDifferentType;
                addedSiblingOfDifferentType = true;
            }

            if (doAdd) {
                selectedFollowingSiblings.addLast(currentSibling);
            }
        }

        return selectedFollowingSiblings;
    }

    /**
     * Selects the siblings to keep among all preceding siblings. This
     * implementation always keeps the first sibling and the
     * <code>baseNumber</code> last siblings. One of the last siblings may
     * have a different type (name as returned by
     * {@link DOMUtils#name(Element)}) than the main element --
     * if there are more with different types, they are skipped.
     * If none of the <code>baseNumber</code> last siblings has a different
     * types, the last siblings with a different type (name) is also kept.
     *
     * @param mainElement the element whose siblings should be selected
     * @param allPrecedingSiblings the list of all preceding siblings
     * @param baseNumber the basic number of siblings to keep; the actual
     * number of siblings kept might vary
     * @return the subset of preceding siblings to keep; an empty list if
     * there are no siblings to keep
     */
    protected List<Element> selectPrecedingSiblings(final Element mainElement,
            final LinkedList<Element> allPrecedingSiblings,
            final int baseNumber) {
        final String elementType = DOMUtils.name(mainElement);
        final LinkedList<Element> selectedPrecedingSiblings =
            new LinkedList<Element>();
        final ListIterator<Element> siblingIter =
            allPrecedingSiblings.listIterator(allPrecedingSiblings.size());
        Element currentSibling;
        boolean addedSiblingOfDifferentType = false;
        boolean doAdd;

        // keep up to |baseNumber| siblings from end of list
        while (siblingIter.hasPrevious()
                && (selectedPrecedingSiblings.size() < baseNumber)) {
            currentSibling = siblingIter.previous();

            if (elementType.equals(DOMUtils.name(currentSibling))) {
                doAdd = true;
            } else {
                // only add if this is the first differently typed sibling
                doAdd = !addedSiblingOfDifferentType;
                addedSiblingOfDifferentType = true;
            }

            if (doAdd) {
                selectedPrecedingSiblings.addFirst(currentSibling);
            }
        }

        // add first sibling, if not yet done
        if (!allPrecedingSiblings.isEmpty()) {
            final Element firstSibling = allPrecedingSiblings.getFirst();
            final Element firstSelected = selectedPrecedingSiblings.getFirst();

            // no need to check size of selectedPrecedingSiblings: must be > 0
            if (firstSelected != firstSibling) {
                selectedPrecedingSiblings.addFirst(firstSibling);
            }
        }

        return selectedPrecedingSiblings;
    }

    /**
      * Returns a string representation of this object.
      *
      * @return a textual representation
      */
     public String toString() {
         return new ToStringBuilder(this)
             .appendSuper(super.toString())
             .append("detailed recognitions", detailedRecognitions)
             .append("number of ancestors", ancestorNumber)
             .append("number of siblings", siblingNumber)
             .append("maximum prefix length", maximumPrefixLength)
             .append("head element", headElement)
             .append("head attribute", headAttribute)
             .append("default attributes", defaultAttributes)
             .append("sensors", ArrayUtils.toString(sensors))
             .toString();
     }

}
