/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.context;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * A small immutable wrapper class that bundles the position of an element
 * within a parent element and the number of children of the parent.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/02/02 18:35:58 $, $Author: siefkes $
 */
public class ElementPosition {

    /**
     * The number of all children elements in the parent element.
     */
    private final int allChildren;

    /**
     * The position of the represented element in the parent element,
     * counting all children.
     */
    private final int overallPosition;

    /**
     * The number of children elements in the parent element with the type
     * (qualified name) of the represented element.
     */
    private final int typedChildren;

    /**
     * The position of the represented element in the parent element,
     * counting only children with the same type (qualified name).
     */
    private final int typedPosition;

    /**
     * The number of preceding siblings that were processed by the context
     * representation.
     */
    private final int processedPreceding;

    /**
     * The number of following siblings that were processed by the context
     * representation.
     */
    private final int processedFollowing;

    /**
     * Creates a new instance.
     *
     * @param allChildElements the number of all children elements in the
     * parent element
     * @param overallPos the position of the represented element in the
     * parent element, counting all children
     * @param typedChildElements the number of children elements in the
     * parent element with the type (qualified name) of the represented
     * element
     * @param typedPos the position of the represented element in the
     * parent element, counting only children with the same type
     * (qualified name)
     * @param processedPrecedingSiblings the number of following siblings that
     * were processed by the context representation
     * @param processedFollowingSiblings the number of preceding siblings that
     * were processed by the context representation
     */
    public ElementPosition(final int allChildElements,
            final int overallPos, final int typedChildElements,
            final int typedPos, final int processedPrecedingSiblings,
            final int processedFollowingSiblings) {
        allChildren = allChildElements;
        overallPosition = overallPos;
        typedChildren = typedChildElements;
        typedPosition = typedPos;
        processedPreceding = processedPrecedingSiblings;
        processedFollowing = processedFollowingSiblings;
    }

    /**
     * Returns the number of all children elements in the parent element.
     *
     * @return the number
     */
    public int getAllChildren() {
        return allChildren;
    }

    /**
     * Returns the position of the represented element in the parent
     * element, counting all children.
     *
     * @return the position
     */
    public int getOverallPosition() {
        return overallPosition;
    }

    /**
     * Returns the number of preceding siblings that were processed by the
     * context representation.
     *
     * @return the number
     */
    public int getProcessedFollowing() {
        return processedFollowing;
    }

    /**
     * Returns the number of following siblings that were processed by the
     * context representation.
     *
     * @return the number
     */
    public int getProcessedPreceding() {
        return processedPreceding;
    }

    /**
     * Returns the number of children elements in the parent element with
     * the type (qualified name) of the represented element.
     *
     * @return the number
     */
    public int getTypedChildren() {
        return typedChildren;
    }

    /**
     * Returns the position of the represented element in the parent
     * element, counting only children with the same type (qualified name).
     *
     * @return the position
     */
    public int getTypedPosition() {
        return typedPosition;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("all child elements", allChildren)
            .append("overall position", overallPosition)
            .append("typed child elements", typedChildren)
            .append("typed position", typedPosition)
            .append("processed preceding siblings", processedPreceding)
            .append("processed following siblings", processedFollowing)
            .toString();
    }

}