/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.context;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * Type-safe enumeration of the types of features used by the default
 * representation.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/02/02 18:35:58 $, $Author: siefkes $
 */
public final class FeatureType {

    /**
     * Constant for features representing attribute values.
     */
    public static final FeatureType ATTRIBUTE = new FeatureType("@");

    /**
     * Constant for features representing calculated values.
     */
    public static final FeatureType CALCULATED = new FeatureType("$");

    /**
     * Constant for features representing an element itself (marker is empty
     * just as for text).
     */
    public static final FeatureType ELEMENT = new FeatureType("");

    /**
     * Constant for features marking a special position or situation.
     */
    public static final FeatureType MARKER = new FeatureType("---");

    /**
     * Constant for features representing textual content (marker is empty just
     * as for elements).
     */
    public static final FeatureType TEXT = new FeatureType("");

    /**
     * The mark used to introduce features of this type.
     */
    private final String mark;

    /**
     * Private constructor to prevent creation of further instances.
     * The static constants defined in this class are the only instantiations.
     *
     * @param featureMark the mark used to introduce features of this type
     */
    private FeatureType(final String featureMark) {
        super();
        mark = featureMark;
    }

    /**
     * Returns the mark used to introduce features of this type.
     *
     * @return the mark
     */
    public String getMark() {
        return mark;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("mark", mark)
            .toString();
    }

}
