/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.context;

import java.util.Iterator;
import java.util.LinkedList;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.classify.feature.Feature;

/**
 * An immutable representation of a feature that can be used for classification.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/09/06 17:22:41 $, $Author: siefkes $
 */
public class GlobalFeature extends Feature {

    /**
     * Converts a series of {@link LocalFeature}s into global features,
     * adding the created global features to a linked list.
     *
     * @param axisName the name of the global axis
     * @param localIter an iterator over the local features to globalize
     * @param globalFeatures the created global features are added to this
     * linked list
     * @param addAtEnd whether to add the created features at the end of at the
     * beginning of the linked list
     * @throws ClassCastException if the iterator refers to objects that aren't
     * {@link LocalFeature}
     */
    public static void globalize(final String axisName,
            final Iterator<LocalFeature> localIter,
            final LinkedList<Feature> globalFeatures,
            final boolean addAtEnd)
    throws ClassCastException {
        GlobalFeature newFeature;
        LocalFeature localValue;

        while (localIter.hasNext()) {
            localValue = localIter.next();
            newFeature = new GlobalFeature(axisName, localValue);

            if (addAtEnd) {
                globalFeatures.addLast(newFeature);
            } else {
                globalFeatures.addFirst(newFeature);
            }
        }
    }

    /**
     * Separator string between axis and element names, used if neither or
     * them is empty.
     */
    public static final String SEP = ".";

    /**
     * The local feature wrapped by this global feature.
     */
    private final LocalFeature localFeature;

    /**
     * The axis prefix wrapped by this feature.
     */
    private final String prefix;

    /**
     * The separator actually used between axis and element names -- either
     * the {@link #SEP} constants, or the empty string if one of them is empty
     * as well (so there is no need for a separator).
     */
    private final String separator;

    /**
     * Creates a new instance that combines a local feature with an axis prefix.
     *
     * @param axisPrefix the axis prefix wrapped by this feature
     * @param local the local feature wrapped by this global feature
     */
    public GlobalFeature(final String axisPrefix,
            final LocalFeature local) {
        super();
        prefix = axisPrefix;
        localFeature = local;

        // determine whether we need a separator
        if ((prefix.length() > 0)
                && (localFeature.getElementName().length() > 0)) {
            separator = SEP;
        } else {
            separator = "";
        }
    }

    /**
     * Creates a new instance that contains only a comment.
     *
     * @param myComment a comment on the features to follow, ignored for
     * classification
     */
    public GlobalFeature(final String myComment) {
        super(myComment);
        // set all other fields to null
        prefix = null;
        localFeature = null;
        separator = null;
    }

    /**
     * Returns the comment attached to this feature, if any.
     *
     * @return the comment, or <code>null</code> is no comment was stored
     */
    public String getComment() {
        if (localFeature != null) {
            // delegate to local feature
            return localFeature.getComment();
        } else {
            // delegate to superclass
            return super.getComment();
        }
    }

    /**
     * Returns the local feature wrapped by this global feature.
     *
     * @return the local feature; or <code>null</code>if this features doesn't
     * wrap a local feature (but only a comment)
     */
    public LocalFeature getLocalFeature() {
        return localFeature;
    }

    /**
     * Returns the representation of this feature, to be used for
     * classification.
     *
     * @return the feature representation, or <code>null</code> if this feature
     * contains only a comment
     */
    public String getRepresentation() {
        if (localFeature != null) {
            return prefix + separator + localFeature.getRepresentation();
        } else {
            // this is a comment-only feature
            return null;
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("prefix", prefix)
            .append("local feature", localFeature)
            .toString();
    }

}
