/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.context;

import java.util.LinkedList;
import java.util.List;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.classify.feature.Feature;
import de.fu_berlin.ties.text.TextUtils;

/**
 * An immutable representation of the local part of a feature, useful to cache
 * and re-use the parts of feature representation that do not depend on the
 * relative position of the element to classify. Must be wrapped in a
 * {@link de.fu_berlin.ties.context.GlobalFeature} for classification.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.10 $, $Date: 2004/09/06 17:22:41 $, $Author: siefkes $
 */
public class LocalFeature extends Feature {

    /**
     * Constant for a marker feature that marks the start of an axis.
     */
    protected static final LocalFeature MARKER_START = new
        LocalFeature("", FeatureType.MARKER, "Start", null, "Start of axis");

    /**
     * Character opening a quoted value.
     */
    protected static final char OPEN = '{';

    /**
     * Character closing a quoted value.
     */
    protected static final char CLOSE = '}';

    /**
     * Factory method creating a feature that wraps an attribute value.
     * It the <code>featureValue</code> contains whitespace, it is
     * split and up to <code>splitMaximum</code> different features are created
     * for the subsequences (discarding initial surplus sequences, if any).
     *
     * @param element the name of the element this feature is attached to
     * @param attributeName the name of the attribute; must not be empty
     * @param featureValue the textual content to wrap
     * @param splitMaximum the maximum number of features to generate
     * @return a list of created feature
     * @throws IllegalArgumentException if the specified attribute name is
     * empty (and thus not a valid name)
     */
    public static List<LocalFeature> createAttributeFeatures(
            final String element, final String attributeName,
            final String featureValue, final int splitMaximum)
            throws IllegalArgumentException {
        if (attributeName.length() == 0) {
            throw new IllegalArgumentException("Attribute name is empty!");
        }
        return createFeatures(element, attributeName, featureValue,
            splitMaximum, true);
    }

    /**
     * Factory method creating a feature that wraps a calculated value.
     *
     * @param element the name of the element this feature is attached to
     * @param calcName the name identifying the calculated value
     * @param featureValue the calculated value to wrap; should not contain
     * whitespace (otherwise the classifier will get confesed)
     * @return the created feature
     */
    public static LocalFeature createCalculatedFeature(final String element,
            final String calcName, final String featureValue) {
        return new LocalFeature(element, FeatureType.CALCULATED, calcName,
            featureValue, null);
    }

    /**
     * Factory method creating a feature that wraps a calculated value that
     * might contain whitespace.
     * It the <code>featureValue</code> contains whitespace, it is
     * split and up to <code>splitMaximum</code> different features are created
     * for the subsequences (discarding initial surplus sequences, if any).
     *
     * @param element the name of the element this feature is attached to
     * @param calcName the name identifying the calculated value
     * @param featureValue the textual content to wrap
     * @param splitMaximum the maximum number of features to generate
     * @return a list of created features
     * @throws IllegalArgumentException if the specified <code>calcName</code>
     * is empty (and thus not a valid name)
     */
    public static List<LocalFeature> createCalculatedFeatures(
            final String element, final String calcName,
            final String featureValue, final int splitMaximum)
            throws IllegalArgumentException {
        if (calcName.length() == 0) {
            throw new IllegalArgumentException("Calculated name is empty!");
        }
        return createFeatures(element, calcName, featureValue,
            splitMaximum, false);
    }

    /**
     * Factory method creating a feature that just represents an element itself.
     *
     * @param element the name of the element this feature is attached to
     * @return the created feature
     */
    public static LocalFeature createElementFeature(final String element) {
        return new LocalFeature(element, FeatureType.ELEMENT, "", null, null);
    }

    /**
     * Helper method creating feature(s) that wrap an attribute or calculated
     * value or textual content. It the <code>featureValue</code> contains
     * whitespace, it is split and up to <code>splitMaximum</code> different
     * features are created for the subsequences (discarding any initial
     * surplus sequences).
     *
     * @param element the name of the element this feature is attached to
     * @param attributeName the name of the attribute; or the empty string when
     * wrapping textual content
     * @param featureValue the attribute value or textual content to wrap
     * @param splitMaximum the maximum number of features to generate
     * @return a list of created feature
     * @param isAttribute whether this is a attribute or calculated feature
     * (ignored for text features, i.e. when <code>attributeName</code> is
     * empty)
     */
    private static List<LocalFeature> createFeatures(final String element,
            final String attributeName, final String featureValue,
            final int splitMaximum, final boolean isAttribute) {
        final List<LocalFeature> result = new LinkedList<LocalFeature>();
        final String[] splitValues = TextUtils.splitString(featureValue,
            splitMaximum);

        // determine feature type
        final FeatureType featureType;
        if (attributeName.length() == 0) {
            featureType = FeatureType.TEXT;
        } else if (isAttribute) {
            featureType = FeatureType.ATTRIBUTE;
        } else {
            featureType = FeatureType.CALCULATED;
        }

        for (int i = 0; i < splitValues.length; i++) {
            result.add(new LocalFeature(element, featureType, attributeName,
                splitValues[i], null));
        }
        return result;
    }

    /**
     * Factory method creating a feature that wraps textual content of an
     * element. It the <code>featureValue</code> contains whitespace, it is
     * split and up to <code>splitMaximum</code> different features are created
     * for the subsequences (discarding initial surplus sequences, if any).
     *
     * @param element the name of the element this feature is attached to
     * @param featureValue the textual content to wrap
     * @param splitMaximum the maximum number of features to generate
     * @return a list of created feature
     */
    public static List<LocalFeature> createTextFeatures(final String element,
            final String featureValue, final int splitMaximum) {
        return createFeatures(element, "", featureValue,
            splitMaximum, false);
    }

    /**
     * Quotes a value text (by enclosing it between {@link #OPEN} and
     * {@link #CLOSE} character).
     *
     * @param text the text to quote (might be <code>null</code>)
     * @return the quoted value, or the empty string if <code>text</code> is
     * <code>null</code>
     */
    public static String quote(final String text) {
        final String result;
        if (text != null) {
            result = OPEN + text + CLOSE;
        } else {
            result = "";
        }
        return result;
    }

    /**
     * The name of the element this feature is attached to.
     */
    private final String elementName;

    /**
     * The type of this feature.
     */
    private final FeatureType type;

    /**
     * The name of this feature (empty for textual features).
     */
    private final String name;

    /**
     * The value wrapped by this feature.
     */
    private final String value;

    /**
     * The representation of this feature, calculated by combining the other
     * fields. Stored for efficiency reasons.
     */
    private final String representation;

    /**
     * Non-public constructor. Use the factory methods to create new
     * instances; or use one of the predefined marker constants.
     *
     * @param element the name of the element this feature is attached to
     * (should be empty for marker features)
     * @param myType the type of this feature
     * @param myName the name of this feature (should be empty for textual
     * or element features)
     * @param myValue the value wrapped by this feature (might be
     * <code>null</code> for marker or element features)
     * @param myComment a comment on this feature, ignored for classification;
     * might be <code>null</code>
     */
    protected LocalFeature(final String element, final FeatureType myType,
            final String myName, final String myValue, final String myComment) {
        super(myComment);
        elementName = element;
        type = myType;
        name = myName;
        value = myValue;
        representation = elementName + type.getMark() + name + quote(value);
    }

    /**
     * Returns the name of the element this feature is attached to.
     *
     * @return the name of the element
     */
    public String getElementName() {
        return elementName;
    }

    /**
     * Returns the name of this feature.
     *
     * @return the name of this feature (the empty string if this is a textual
     * feature)
     */
    public String getName() {
        return name;
    }

    /**
     * Returns the representation of this feature, to be used for
     * classification.
     *
     * @return the feature representation, or <code>null</code> if this feature
     * contains only a comment
     */
    public String getRepresentation() {
        return representation;
    }

    /**
     * Returns the type of this feature.
     *
     * @return the type
     */
    public FeatureType getType() {
        return type;
    }

    /**
     * Returns the value wrapped by this feature.
     *
     * @return the value
     */
    public String getValue() {
        return value;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation combining all relevant fields
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("representation", representation)
            .toString();
    }

}
