/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.context;

import java.util.Iterator;
import java.util.LinkedList;

import org.apache.commons.collections.buffer.CircularFifoBuffer;
import org.apache.commons.collections.iterators.EmptyIterator;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.classify.feature.Feature;

/**
 * A buffer of prior {@link de.fu_berlin.ties.context.Recognition}s that should
 * be considered in the context representation. The buffer adjusts its size
 * automatically, discarding the least recently added recognitions when enough
 * recognitions are stored.
 *
 * <p>Allows to cache feature vectors representing recognitions (caching should
 * only be used for
 * {@linkplain de.fu_berlin.ties.context.Recognition#isSealed() sealed}
 * recognitions because otherwise the cached features might get out-of-date).
 *
 * <p>Instances of this class and the nested class
 * {@link PriorRecognitions.Pair} are not thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2004/09/06 17:22:41 $, $Author: siefkes $
 */
public class PriorRecognitions {

    /**
     * An inner class wrapping a recognition and optionally a feature vector
     * representing the recognition.
     */
    public static class Pair {

        /**
         * The recognition wrapped in this pair.
         */
        private final Recognition recognition;

        /**
         * A list of features representing the recognition,
         * initially <code>null</code>.
         */
        private LinkedList<Feature> cachedFeatures = null;

        /**
         * Creates a new instance.
         * @param recog the recognition to wrap
         */
        public Pair(final Recognition recog) {
            recognition = recog;
        }

        /**
         * Returns the list of features representing the recognition,
         * initially <code>null</code>.
         * @return the value of the attribute
         */
        public LinkedList<Feature> getCachedFeatures() {
            return cachedFeatures;
        }

        /**
         * Returns the recognition wrapped in this pair.
         * @return the value of the attribute
         */
        public Recognition getRecognition() {
            return recognition;
        }

        /**
         * Sets the list of features representing the recognition.
         * @param features the new value of the attribute
         */
        public void setCachedFeatures(final LinkedList<Feature> features) {
            cachedFeatures = features;
        }

        /**
         * Returns a string representation of this object.
         *
         * @return a textual representation
         */
        public String toString() {
            return new ToStringBuilder(this)
                .append("recognition", recognition)
                .append("features are cached", (cachedFeatures != null))
                .toString();
        }
    }

    /**
     * The buffer used to store the pairs; <code>null</code> if the number of
     * recognitions to store is 0.
     */
    private final CircularFifoBuffer pairBuffer;

    /**
     * Creates a new instance.
     * @param size the number of prior recognitions to store; if 0 this is
     * just a dummy instance that throws away any recognitions that are
     * {@linkplain #add(Recognition) added}
     */
    public PriorRecognitions(final int size) {
        super();
        pairBuffer = (size > 0) ? new CircularFifoBuffer(size) : null;
    }

    /**
     * Adds the given recognition to this buffer, discarding the least recently
     * added recognition if appropriate.
     * @param recognition the recognition to add
     */
    public void add(final Recognition recognition) {
        if (pairBuffer != null) {
            pairBuffer.add(new Pair(recognition));
        }
    }

    /**
     * Returns an iterator over the elements stored in this instance.
     * <em>Note that the objects referred by the iterator will be
     * {@link PriorRecognitions.Pair}s (<strong>not</strong>
     * {@link Recognition}s). These pairs can be used to access the recognition
     * and to retrieve and store feature vectors representing the recognition.
     * Feature vectors should be cached only for
     * {@linkplain Recognition#isSealed() sealed} recognitions because otherwise
     * the cached features might get out-of-date</em>.
     *
     * @return an iterator over the stored {@link PriorRecognitions.Pair}s
     */
    public Iterator iterator() {
        if (pairBuffer != null) {
            return pairBuffer.iterator();
        } else {
            return EmptyIterator.INSTANCE;
        }
    }

    /**
     * Removes the most recently inserted recognition from this buffer.
     *
     * @return the removed recognition; or <code>null</code> if this container
     * is empty
     */
    public Recognition removeLast() {
        if (size() > 0) {
            // we have use the iterator because we cannot access the last
            // element directly
            final Iterator pairIter = pairBuffer.iterator();
            Pair pair = null;

            while (pairIter.hasNext()) {
                pair = (Pair) pairIter.next();
            }

            // remove the last pair and recogn its recognition
            pairIter.remove();
            return pair.getRecognition();
        } else {
            // empty/no buffer
            return null;
        }
    }

    /**
     * Returns the number of recognitions stored in the buffer.
     * @return the number of recognitions stored
     */
    public int size() {
        if (pairBuffer != null) {
            return pairBuffer.size();
        } else {
            return 0;
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("pair buffer", pairBuffer)
            .toString();
    }

}
