/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.context;

import java.util.List;
import java.util.Map;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Document;
import org.dom4j.Element;

import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.xml.dom.DOMUtils;

/**
 * Abstract class that manages context representations for entity recognition
 * and information extraction. Subclasses must implement the
 * {@link #buildContext(Element, String, String, String, PriorRecognitions,
 * Map, String)} method for building representations.
 * 
 * @author Christian Siefkes
 * @version $Revision: 1.13 $, $Date: 2004/11/04 15:26:51 $, $Author: siefkes $
 */
public abstract class Representation {

    /**
     * The number of preceding recognitions to represent.
     */
    private final int recognitionNumber;

    /**
     * Creates a new instance.
     * 
     * @param recogNum the number of preceding recognitions to represent
     */
    public Representation(final int recogNum) {
        super();
        recognitionNumber = recogNum;
    }

    /**
     * Builds the context representation of a document. The default
     * implementation delegates to the {@link #buildContext(Element,
     * PriorRecognitions, Map, String)} method, using the root element of the
     * document.
     * 
     * @param document the XML document whose context should be represented
     * @param priorRecognitions a buffer of the last {@link Recognition}s from
     * the document, created by calling {@link #createRecognitionBuffer()};
     * might be <code>null</code>
     * @param featureCache a cache of (local) feature, should be re-used between
     * all calls for the nodes in a single document (but must not be re-used
     * when building the context of nodes in different documents!)
     * @param logPurpose the type of contexts of main interest to the caller
     * (e.g. "Token" or "Sentence"), used for logging
     * @return a vector of features considered relevant for representation
     * @throws ClassCastException if the <code>priorRecognitions</code> buffer
     * contains objects that aren't {@link Recognition}s
     */
    public FeatureVector buildContext(final Document document,
            final PriorRecognitions priorRecognitions,
            final Map<Element, List<LocalFeature>> featureCache,
            final String logPurpose) throws ClassCastException {
        // delegate using root element
        return buildContext(document.getRootElement(), priorRecognitions,
                featureCache, logPurpose);
    }

    /**
     * Builds the context representation of an element. The default
     * implementation delegates to the {@link #buildContext(Element, String,
     * String, String, PriorRecognitions, Map, String)} method, using the
     * {@linkplain DOMUtils#collectText(org.dom4j.Branch) full textual content}
     * of the element as <code>mainText</code> and empty strings as
     * <code>leftText</code> and <code>rightText</code>.
     * 
     * @param element the element whose context should be represented
     * @param priorRecognitions a buffer of the last {@link Recognition}s from
     * the document, created by calling {@link #createRecognitionBuffer()};
     * might be <code>null</code>
     * @param featureCache a cache of (local) feature, should be re-used between
     * all calls for the nodes in a single document (but must not be re-used
     * when building the context of nodes in different documents!)
     * @param logPurpose the type of contexts of main interest to the caller
     * (e.g. "Token" or "Sentence"), used for logging
     * @return a vector of features considered relevant for representation
     * @throws ClassCastException if the <code>priorRecognitions</code> buffer
     * contains objects that aren't {@link Recognition}s
     */
    public FeatureVector buildContext(final Element element,
            final PriorRecognitions priorRecognitions,
            final Map<Element, List<LocalFeature>> featureCache,
            final String logPurpose) throws ClassCastException {
        // delegate to abstract method
        return buildContext(element, "", DOMUtils.collectText(element), "",
                priorRecognitions, featureCache, logPurpose);
    }

    /**
     * Builds the context representation of text in an element.
     * Returns a feature vector of all context features considered relevant for
     * representation.
     * 
     * @param element the element whose context should be represented
     * @param leftText textual content to the left of (preceding)
     * <code>mainText</code>, might be empty
     * @param mainText the main textual content to represent, might be empty
     * @param rightText textual content to the right of (following)
     * <code>mainText</code>, might be empty
     * @param priorRecognitions a buffer of the last {@link Recognition}s from
     * the document, created by calling {@link #createRecognitionBuffer()};
     * might be <code>null</code>
     * @param featureCache a cache of (local) feature, should be re-used between
     * all calls for the nodes in a single document (but must not be re-used
     * when building the context of nodes in different documents!)
     * @param logPurpose the type of contexts of main interest to the caller
     * (e.g. "Token" or "Sentence"), used for logging
     * @return a vector of features considered relevant for representation
     * @throws ClassCastException if the <code>priorRecognitions</code> buffer
     * contains objects that aren't {@link Recognition}s
     */
    public abstract FeatureVector buildContext(final Element element,
            final String leftText, final String mainText,
            final String rightText, final PriorRecognitions priorRecognitions,
            final Map<Element, List<LocalFeature>> featureCache,
            final String logPurpose) throws ClassCastException;

    /**
     * Creates a buffer to be filled with prior {@link Recognition}s and passed
     * as argument to the {@link #buildContext(Element, String, String, String,
     * PriorRecognitions, Map, String)} method. The caller must
     * {@link PriorRecognitions#add(Recognition)}new recognitions to the buffer
     * but it is not necessary to remove them -- the buffer will automatically
     * delete the oldest recognitions when appropriate.
     * 
     * @return a buffer to be used for collecting prior {@link Recognition}s
     */
    public PriorRecognitions createRecognitionBuffer() {
        return new PriorRecognitions(getRecognitionNumber());
    }

    /**
     * Returns the number of preceding recognitions to represent.
     * 
     * @return the value of the attibute
     */
    public int getRecognitionNumber() {
        return recognitionNumber;
    }

    /**
     * Returns a string representation of this object.
     * 
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this).append("number of recognitions",
                recognitionNumber).toString();
    }

}