/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.context;

import java.util.Iterator;
import java.util.Map;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Attribute;
import org.dom4j.Element;

import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.feature.DefaultFeature;
import de.fu_berlin.ties.classify.feature.DefaultFeatureVector;
import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.text.TextTokenizer;
import de.fu_berlin.ties.text.TextUtils;
import de.fu_berlin.ties.text.TokenizerFactory;
import de.fu_berlin.ties.xml.dom.DOMUtils;

/**
 * A simple representation of an text in an element in an XML document.
 * Instances of this class are thread-safe.
 * 
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/09/02 16:30:14 $, $Author: siefkes $
 */
public class SimpleRepresentation extends AbstractRepresentation {

    /**
     * Used to tokenize the textual content of the elements to represent.
     * Synchronized on itself.
     */
    private final TextTokenizer tokenizer;

    /**
     * Creates a new instance based on the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @throws ProcessingException if an error occurs while initializing this
     * instance
     */
    public SimpleRepresentation() throws ProcessingException {
        this(TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance based on the provided configuration.
     *
     * @param config used to configure this instance
     * @throws ProcessingException if an error occurs while initializing this
     * instance
     */
    public SimpleRepresentation(final TiesConfiguration config)
    throws ProcessingException {
        this(config, null);
    }

    /**
     * Creates a new instance based on the provided configuration.
     *
     * @param config used to configure this instance
     * @param suffix this suffix can be appended to the used configuration
     * parameters to give values that are specific for this instance; may be
     * <code>null</code>
     * @throws ProcessingException if an error occurs while initializing this
     * instance
     */
    public SimpleRepresentation(final TiesConfiguration config,
            final String suffix) throws ProcessingException {
        this(config.getInt(config.adaptKey(CONFIG_RECOGN_NUM, suffix)),
            config.getInt(config.adaptKey(CONFIG_SPLIT_MAXIMUM, suffix)),
            config.getInt(config.adaptKey(CONFIG_STORE_NTH, suffix)),
            config.getString(config.adaptKey(IOUtils.KEY_LOCAL_CHARSET, suffix),
                    null),
            new TokenizerFactory(config, suffix).createTokenizer(""));
    }

    /**
     * Creates a new instance.
     *
     * @param recogNum the number of preceding recognitions to represent
     * @param splitMax the maximum number of subsequences to keep when
     * a feature value must be split (at whitespace)
     * @param n Each <em>n</em>-th context representation is stored if &gt; 0;
     * otherwise no representation is stored
     * @param outCharset the output character set to use (only used to
     * store some configurations for inspection purposes, if <code>n</code>
     * &gt; 0); if <code>null</code>, the default charset of the current
     * platform is used
     * @param textTokenizer the tokenizer to use
     */
    public SimpleRepresentation(final int recogNum, final int splitMax,
            final int n, final String outCharset,
            final TextTokenizer textTokenizer) {
        super(recogNum, splitMax, n, outCharset);
        tokenizer = textTokenizer;
    }

    /**
     * Creates a feature and adds it to a feature vector. The feature is
     * created by joining <code>prefix</code> and <code>value</code>, separated
     * by a colon as separator character.
     *
     * @param features the feature vector to append to
     * @param prefix the prefix of the new feature
     * @param value the main value of the new feature
     */
    protected void addFeature(final FeatureVector features, final String prefix,
            final String value) {
        features.add(new DefaultFeature(prefix + ':' + value));
    }

    /**
     * Adds feature(s) representing text to a feature vector, using the
     * instance tokenizer for splitting the text into tokens.
     *
     * @param features the feature vector to append to
     * @param prefix the prefix of the new feature(s)
     * @param text to text to tokenize and add
     */
    protected void addText(final FeatureVector features, final String prefix,
            final String text) {
        // synchronize on the tokenizer because it isn't thread-safe
        synchronized (tokenizer) {
            tokenizer.reset(text);
            String token;

            while ((token = tokenizer.nextToken()) != null) {
                addFeature(features, prefix, token);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    protected FeatureVector doBuildContext(final Element element,
            final String leftText, final String mainText,
            final String rightText, final PriorRecognitions priorRecognitions,
            final Map featureCache, final String logPurpose)
    throws ClassCastException {
        final FeatureVector result = new DefaultFeatureVector();

        // store name of element
        addFeature(result, "type", DOMUtils.name(element));

        // store attributes, splitting values at whitespace
        final Iterator attribIter = element.attributeIterator();
        Attribute attrib;
        String name;
        String[] splittedVal;
        int i;

        while (attribIter.hasNext()) {
            attrib = (Attribute) attribIter.next();
            name = DOMUtils.name(attrib);
            splittedVal = TextUtils.splitString(attrib.getValue(),
                    getSplitMaximum());

            for (i = 0; i < splittedVal.length; i++) {
                addFeature(result, '@' + name, splittedVal[i]);
            }
        }

        // tokenize + store left + main + right text
        addText(result, "left", leftText);
        addText(result, "text", mainText);
        addText(result, "right", rightText);

        // add prior recognitions (only the types, not the text)
        PriorRecognitions.Pair pair;
        Recognition recognition;
        Iterator priorIter = priorRecognitions.iterator();

        while (priorIter.hasNext()) {
            pair = (PriorRecognitions.Pair) priorIter.next();
            recognition = pair.getRecognition();
            addFeature(result, "prior", recognition.getType());
        }

        return result;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("tokenizer", tokenizer)
            .toString();
    }

}
