/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.context.sensor;

import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.util.Util;

/**
 * Abstract implementation of the
 * {@link de.fu_berlin.ties.context.sensor.Sensor} interface that stores a
 * configuration and provides a factory method to initialize a set of sensors.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/07/08 08:45:33 $, $Author: siefkes $
 */
public abstract class BaseSensor implements Sensor {

    /**
     * Factory method that initializes an array of sensors.
     *
     * @param classNames array of fully specified names of classes
     * implementing the {@link Sensor} interface; each of them must provide
     * a constructor that accepts a {@link TiesConfiguration} as single argument
     * @param conf used to configure the sensors
     * @return an array of newly created sensors
     * @throws ProcessingException if an error occurred while creating the
     * classifier
     */
    public static Sensor[] createSensors(final String[] classNames,
            final TiesConfiguration conf) throws ProcessingException {
        final Sensor[] result = new Sensor[classNames.length];

        // use the configuration as single constructor parameter
        final Object[] params = new Object[] {conf};

        for (int i = 0; i < result.length; i++) {
            try {
                result[i] = (Sensor) Util.createObject(
                    Class.forName(classNames[i]), params,
                    TiesConfiguration.class);
            } catch (ClassNotFoundException cnfe) {
                // convert and rethrow exception
                throw new ProcessingException(
                    "Cannot initialize sensor " + classNames[i],
                    cnfe);
            } catch (InstantiationException ie) {
                // convert and rethrow exception
                throw new ProcessingException(
                    "Cannot initialize sensor " + classNames[i],
                    ie);
            }
        }

        return result;
    }


    /**
     * The configuration used by this instance.
     */
    private final TiesConfiguration config;

    /**
     * Creates a new instance.
     * @param conf the configuration to use
     */
    public BaseSensor(final TiesConfiguration conf) {
        super();
        config = conf;
    }

    /**
     * Returns the configuration used by this instance.
     * @return the configuration
     */
    public TiesConfiguration getConfig() {
        return config;
    }

}
