/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.context.sensor;

import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.collections.KeyValue;
import org.apache.commons.collections.keyvalue.DefaultKeyValue;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.text.TextUtils;
import de.fu_berlin.ties.util.MultiValueMap;
import de.fu_berlin.ties.util.SortedMultiValueMap;
import de.fu_berlin.ties.util.Util;

/**
 * A list sensor uses a one or several gazetteers to look up semantic
 * information. Gazetteers are text files in a very simple format, they just
 * contain one entry per line. Gazetters must be in the character set configured
 * by the {@link de.fu_berlin.ties.io.IOUtils#KEY_LOCAL_CHARSET} (or in the
 * local default character set if there is no entry for this key).
 *
 * @author Christian Siefkes
 * @version $Revision: 1.7 $, $Date: 2004/11/05 12:15:08 $, $Author: siefkes $
 */
public class ListSensor extends BaseSensor {

    /**
     * Gazetteer files are resolved relative to this path, if given.
     */
    private final File basePath;

    /**
     * Gazetteer entries are looked up case-sensitive if this is
     * <code>true</code>.
     */
    private final boolean caseSensitive;

    /**
     * The key/value pair to use as negative marker if no positive information
     * is present for a token; or <code>null</code> if no negative marker
     * should be used.
     */
    private final KeyValue negativeMarker;

    /**
     * The keys of all gazetteer types managed by this instance; only stored
     * if "false" should be added for all negative (not present) types.
     */
    private final SortedSet<String> typeSet;

    /**
     * Sorted multi-map containing mapping each known token to a sorted
     * collection of values to be returned for this token.
     */
    private final MultiValueMap<String, String> store =
        new SortedMultiValueMap<String, String>();

    /**
     * Creates a new instance.
     *
     * @param conf the configuration to use
     * @throws IOException if an I/O error occurs while reading the gazetteers
     * to use
     */
    public ListSensor(final TiesConfiguration conf) throws IOException {
        super(conf);

        // init fields
        caseSensitive = conf.getBoolean("sensor.list.case");
        final String basePathName = conf.getString("sensor.list.basepath",
            null);
        basePath = (basePathName == null)
            ? null
            : new File(basePathName);

        // initialize negative marker resp. type set if it should be used
        if (conf.getBoolean("sensor.list.negative")) {
            final String negativeMarkerKey =
                conf.getString("sensor.list.negative.value", null);

            if (StringUtils.isNotEmpty(negativeMarkerKey)) {
                // use marker key if _no_ positive information is present
                negativeMarker = new DefaultKeyValue(negativeMarkerKey, null);
                typeSet = null;
            } else {
                // otherwise use "true" or "false" for _every_ type
                typeSet = new TreeSet<String>();
                negativeMarker = null;
            }
        } else {
            // use neither
            negativeMarker = null;
            typeSet = null;
        }

        // load gazetteer files
        final Configuration fileMappings = conf.subset("sensor.list.map");
        final Iterator keyIter = fileMappings.getKeys();
        String key;
        String fileName;
        File gazetteerFile;
        Reader fileReader;
        String fileContents;
        String[] entries;
        String currentEntry;
        HashSet<String> entryHash;

        while (keyIter.hasNext()) {
            key = (String) keyIter.next();
            fileName = fileMappings.getString(key).trim();

            // ignore empty values (could be used to reset a mapping)
            if (fileName.length() > 0) {
                gazetteerFile = new File(fileName);

                if ((basePath != null) && (!gazetteerFile.isAbsolute())) {
                    // non-absolute paths are resolved relative to the base path
                    gazetteerFile = new File(basePath, fileName);
                }

                // open file using local resp. default charset
                fileReader = IOUtils.openReader(gazetteerFile, conf);
                fileContents = IOUtils.readToString(fileReader);
                IOUtils.tryToClose(fileReader);

                // split file into entries (one per non-empty line)
                entries = TextUtils.splitLines(fileContents);
                entryHash = new HashSet<String>(entries.length);

                // add key to collection of values for each entry
                for (int i = 0; i < entries.length; i++) {
                    // entries are converted to lower case if case is ignored
                    currentEntry = caseSensitive
                        ? entries[i]
                        : entries[i].toLowerCase();

                    // use hash to avoid that entries are added twice (e.g. if
                    // present in different capitalizations and case is ignored)
                    if (!entryHash.contains(currentEntry)) {
                        store.put(currentEntry, key);
                        entryHash.add(currentEntry);
                    }
                }

                if (typeSet != null) {
                    typeSet.add(key);
                }

                Util.LOG.debug("Loaded " + gazetteerFile + " file: "
                    + entries.length + " entries");
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public KeyValue[] lookup(final String token) {
        final KeyValue[] result;
        // look up the token in the store (in lower case if case is ignored)
        final Collection<String> mappings = store.get(
            caseSensitive ? token : token.toLowerCase());

        if (typeSet != null) {
            result = new KeyValue[typeSet.size()];
            final Iterator typeIter = typeSet.iterator();
            String currentKey;
            boolean foundKey;

            // add "true" or "false" for every list type
            for (int i = 0; typeIter.hasNext(); i++) {
                currentKey = (String) typeIter.next();
                foundKey = (mappings != null) && mappings.contains(currentKey);
                result[i] = new DefaultKeyValue(currentKey,
                    Boolean.toString(foundKey));
            }
        } else {
            // only use positive types (and single negative marker if specified)
            if ((mappings != null) && (mappings.size() > 0)) {
                // found at least one mapping
                result = new KeyValue[mappings.size()];
                final Iterator mappingsIter = mappings.iterator();
                String currentKey;

                for (int i = 0; mappingsIter.hasNext(); i++) {
                    currentKey = (String) mappingsIter.next();
                    // use null values for each key
                    result[i] = new DefaultKeyValue(currentKey, null);
                }
            } else {
                if (negativeMarker != null) {
                    // return negative Marker as only element
                    result = new KeyValue[] {negativeMarker};
                } else {
                    // return empty array
                    result = new KeyValue[0];
                }
            }
        }

        return result;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("base path", basePath)
            .append("case sensitive", caseSensitive)
            .append("negative marker key", negativeMarker.getKey())
            .append("store size", store.size())
            .toString();
    }

}
