/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.eval;

import de.fu_berlin.ties.io.BaseStorable;
import de.fu_berlin.ties.io.FieldMap;
import de.fu_berlin.ties.util.Util;

/**
 * Counts <em>true</em> and <em>false</em> items and measures the
 * <em>accuracy</em>: <em>A</em> = <em>true</em> /
 * (<em>true</em> + <em>false</em>).
 *
 * <p>Instances of this class are not thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/05/18 13:46:10 $, $Author: siefkes $
 */
public class Accuracy extends BaseStorable implements AccuracyView {

    /**
     * Serialization key for the accuracy.
     */
    public static final String KEY_ACCURACY = "Accuracy";

    /**
     * Serialization key for the number of false items.
     */
    public static final String KEY_FALSE_COUNT = "False Count";

    /**
     * Serialization key for the number of true items.
     */
    public static final String KEY_TRUE_COUNT = "True Count";

    /**
     * Optional prefix used for (de)serializing this instance; should typically
     * end in a space; empty string if not used.
     */
    private final String prefix;

    /**
     * The number of false decisions.
     */
    private long falseCount = 0;

    /**
     * The number of true decisions.
     */
    private long trueCount = 0;

    /**
     * Creates a new empty instance, without using a prefix.
     */
    public Accuracy() {
        this("");
    }

    /**
     * Creates a new empty instance.
     *
     * @param prefixString An optional prefix used for (de)serializing this
     * instance; should typically end in a space; empty string if not used
     */
    public Accuracy(final String prefixString) {
        super();
        prefix = prefixString;
    }

    /**
     * Creates a new instance from a field map, fulfilling the
     * {@link de.fu_berlin.ties.io.Storable} contract.
     *
     * @param fieldMap map containing the serialized fields
     * @throws IllegalArgumentException if at least one of the parameters is
     * negative or missing
     */
    public Accuracy(final FieldMap fieldMap) throws IllegalArgumentException {
        this(fieldMap, "");
    }

    /**
     * Creates a new instance from a field map, using a prefix.
     *
     * @param fieldMap map containing the serialized fields
     * @param prefixString An optional prefix used for (de)serializing this
     * instance; should typically end in a space; empty string if not used
     * @throws IllegalArgumentException if at least one of the parameters is
     * negative or missing
     */
    public Accuracy(final FieldMap fieldMap, final String prefixString)
            throws IllegalArgumentException {
        prefix = prefixString;
        trueCount = Util.asLong(fieldMap.get(prefix + KEY_TRUE_COUNT));
        falseCount = Util.asLong(fieldMap.get(prefix + KEY_FALSE_COUNT));

        if ((trueCount < 0) || (falseCount < 0)) {
            throw new IllegalArgumentException("Both '" + prefix
                + KEY_TRUE_COUNT + "' (" + trueCount + ") and '" + prefix
                + KEY_FALSE_COUNT + "' (" + falseCount + ") must be > 0");
        }
    }

    /**
     * Calculates and returns the <em>accuracy</em>:
     * <em>A</em> = <em>true</em> / (<em>true</em> + <em>false</em>).
     *
     * @return the accuracy
     */
    public double getAccuracy() {
        return (double) trueCount / (trueCount + falseCount);
    }

    /**
     * Returns the number of false decisions.
     * @return the number of false decisions
     */
    public long getFalseCount() {
        return falseCount;
    }

    /**
     * Returns the number of true decisions.
     * @return the number of true decisions
     */
    public long getTrueCount() {
        return trueCount;
    }

    /**
     * Increases the number of false decisions by one.
     */
    public void incFalseCount() {
        falseCount++;
    }

    /**
     * Increases the number of true decisions by one.
     */
    public void incTrueCount() {
        trueCount++;
    }

    /**
     * {@inheritDoc}
     */
    public FieldMap storeFields() {
        final FieldMap result = new FieldMap();
        result.put(prefix + KEY_ACCURACY, new Double(getAccuracy()));
        result.put(prefix + KEY_TRUE_COUNT, new Long(trueCount));
        result.put(prefix + KEY_FALSE_COUNT, new Long(falseCount));
        return result;
    }

}
