/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.eval;

import java.util.HashMap;
import java.util.Map;

import org.apache.commons.lang.StringUtils;

/**
 * Type-safe enumeration of possible evaluation states for predictions
 * (unknown, correct, spurious etc.) and answer keys (missing etc.).
 *
 * @author Christian Siefkes
 * @version $Revision: 1.4 $, $Date: 2004/11/07 18:21:27 $, $Author: siefkes $
 */
public final class EvalStatus {

    /**
     * A map of all instantiated instances of this class. The name of the
     * instance is mapped to the instance itself. There is no need to
     * synchronize this map as it is only modified during the static
     * initialization of this class.
     */
    private static final Map<String, EvalStatus> ALL_INSTANCES =
        new HashMap<String, EvalStatus>();

    /**
     * Serialization key recommended for instances of this class.
     */
    public static final String KEY_EVAL_STATUS = "EvalStatus";

    /**
     * Constant for predictions whose evaluation status is unknown (not
     * yet evaluated).
     */
    public static final EvalStatus UNKNOWN =
        new EvalStatus("unknown", true, false);

    /**
     * Constant for items defined to be true (answer keys, "gold standard").
     */
    public static final EvalStatus TRUTH = new EvalStatus("truth", false, true);

    /**
     * Constant for correct predictions (true positives) resp for answer
     * keys matched by a prediction. This is the only instance that is suitable
     * for both {@link #isAnswerState() answer keys} and
     * {@link #isPredictionState() predictions}.
     */
    public static final EvalStatus CORRECT =
        new EvalStatus("correct", true, true);

    /**
     * Constant for missing answer keys (false negatives).
     */
    public static final EvalStatus MISSING =
        new EvalStatus("missing", false, true);

    /**
     * Constant for spurious predictions (false positives).
     */
    public static final EvalStatus SPURIOUS =
        new EvalStatus("spurious", true, false);

    /**
     * Constant for answer keys that could have proposed as predictions but
     * were not. When there is only a single instance of each type to predict
     * ("best match" mode), one of the answer keys should be tagged as
     * {@link #CORRECT} resp. {@link #MISSING}; the others should be tagged
     * as {@link #ALTERNATIVE}s.
     */
    public static final EvalStatus ALTERNATIVE =
        new EvalStatus("alternative", false, true);

    /**
     * Constant for predictions that were ignored. When there is only a single
     * instance of each type to predict ("best match" mode), the most probably
     * prediction will be evaluated, while the others of the same type will be
     * ignored.
     */
    public static final EvalStatus IGNORED =
        new EvalStatus("ignored", true, false);

    /**
     * Parses a textual representation into an eval status, without using a
     * default object. This method can convert to results of {@link #getName()}
     * and {@link #toString()} back into the corresponding instance.
     *
     * @param representation the textual representation to parse
     * @return the eval status object corresponding to the given string
     * @throws IllegalArgumentException if a given string does not represent a
     * known evaluation status
     */
    public static EvalStatus parse(final String representation)
            throws IllegalArgumentException {
        return parse(representation, false);
    }

    /**
     * Parses a textual representation into an eval status. This method can
     * convert to results of {@link #getName()} and {@link #toString()} back
     * into the corresponding instance.
     *
     * @param representation the textual representation to parse
     * @param useDefault if <code>true</code>, {@link #UNKNOWN} is returned if
     * the representation is an unparsable (invalid) or <code>null</code>;
     * otherwise an exception is thrown in this case
     * @return the eval status object corresponding to the given string
     * @throws IllegalArgumentException if a given string does not represent a
     * known evaluation status and <code>useDefault</code> is <code>false</code>
     */
    public static EvalStatus parse(final String representation,
            final boolean useDefault) throws IllegalArgumentException {
        // using StringUtils so null is handled
        final EvalStatus result =
            ALL_INSTANCES.get(StringUtils.trim(representation));

        if (result != null) {
            return result;
        } else {
            if (useDefault) {
                return EvalStatus.UNKNOWN;
            }  else {
                throw new IllegalArgumentException(
                        "Cannot parse to eval status: " + representation);
            }
        }
    }

    /**
     * The name of this instance.
     */
    private final String name;

    /**
     * Whether this state is suitable for an answer key.
     */
    private final boolean answerState;

    /**
     * Whether this state is suitable for a prediction.
     */
    private final boolean predictionState;

    /**
     * Private constructor to prevent creation of further instances.
     * The static constants defined in this class are the only instantiations.
     *
     * @param newName the name of this instance
     * @param predState whether this state is suitable for a prediction
     * @param ansState whether this state is suitable for an answer key
     * @throws IllegalArgumentException if the specified name is already used
     * by another instance
     */
    private EvalStatus(final String newName, final boolean predState,
            final boolean ansState) throws IllegalArgumentException {
        super();
        name = newName;
        predictionState = predState;
        answerState = ansState;

        // register myself
        final EvalStatus previousValue = ALL_INSTANCES.put(newName, this);
        if (previousValue != null) {
            throw new IllegalArgumentException(
                "Name " + newName + " already in use!");
        }
    }

    /**
     * Returns the name of this instance.
     * @return the value of the attribute
     */
    public String getName() {
        return name;
    }

    /**
     * Returns a string representation, printing the
     * {@linkplain #getName() name} of this instance.
     *
     * @return a textual representation
     */
    public String toString() {
        return name;
    }

    /**
     * Whether this state is suitable for an answer key.
     * @return <code>true</code> iff this state is suitable for an answer key
     */
    public boolean isAnswerState() {
        return answerState;
    }

    /**
     * Whether this state is suitable for a prediction.
     * @return <code>true</code> iff this state is suitable for a prediction
     */
    public boolean isPredictionState() {
        return predictionState;
    }

}
