/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.eval;

import de.fu_berlin.ties.io.BaseStorable;
import de.fu_berlin.ties.io.FieldMap;
import de.fu_berlin.ties.util.Util;

/**
 * This class manages and updates evaluation results, calculating
 * <em>precision (P)</em>, <em>recall (R)</em> and <em>F-measure (F)</em>.
 * It counts <em>true positives (tp)</em>, <em>false negatives (tn)</em> and
 * <em>false positives (fp)</em>.
 *
 * <p>The metrics are defines as follows (cf. Manning and Schtze (1999):
 * <em>Foundations of Statistical Natural Language Processing</em>,
 * p.&nbsp;268f):
 *
 * <dl>
 * <dt>Precision <em>P</em> =</dt>
 * <dd><em>tp</em> / (<em>tp</em> + <em>fp</em>)</dd>
 * <dt>Recall <em>R</em> =</dt>
 * <dd><em>tp</em> / (<em>tp</em> + <em>fn</em>)</dd>
 * <dt>F-measure F =</dt>
 * <dd>1 / (<em>alpha</em> * (1/<em>P</em>) +
 *         (1-<em>alpha</em>) * (1/<em>R</em>))<br>
 * We define <em>F</em> to be 0 if <em>P</em> = 0 or <em>R</em> = 0 (not
 * discussed by Manning and Schtze).<br>
 * The factor <em>alpha</em> defines the weighting of precision and recall;
 * <em>P</em> and <em>R</em> are weighted equal if <em>alpha</em> = 0.5
 * (the default value).</dd>
 * </dl>
 *
 * <p>Instances of this class are not thread-safe and must be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2004/10/11 15:30:41 $, $Author: siefkes $
 */
public class FMetrics extends BaseStorable implements FMetricsView {

    /**
     * Serialization key for true positives.
     */
    public static final String KEY_TRUE_POS = "True Positives";

    /**
     * Serialization key for false negatives.
     */
    public static final String KEY_FALSE_NEG = "False Negatives";

    /**
     * Serialization key for false positives.
     */
    public static final String KEY_FALSE_POS = "False Positives";

    /**
     * Serialization key for the precision.
     */
    public static final String KEY_PRECISION = "Precision";

    /**
     * Serialization key for the recall.
     */
    public static final String KEY_RECALL = "Recall";

    /**
     * Serialization key for the F1 measure.
     */
    public static final String KEY_F1_MEASURE = "F1-Measure";

    /**
     * The number of true positives (correct recognitions).
     */
    private long truePos = 0;

    /**
     * The number of false negatives (false rejections).
     */
    private long falseNeg = 0;

    /**
     * The number of false positives (false acceptances).
     */
    private long falsePos = 0;

    /**
     * Cached precision value, stored for efficiency; <code>-1</code> if not
     * known or out-of-date.
     */
    private double cachedPrecision = -1.0;

    /**
     * Cached recall value, stored for efficiency; <code>-1</code> if not
     * known or out-of-date.
     */
    private double cachedRecall = -1.0;

    /**
     * Creates a new empty instance.
     */
    public FMetrics() {
        super();
    }

    /**
     * Creates a new instance.
     *
     * @param input contains the initial number of true and false positives and
     * false negatives
     * @throws IllegalArgumentException if at least one of the input value is
     * negative
     */
    public FMetrics(final EvalInput input) throws IllegalArgumentException {
        this(input.getTruePos(), input.getFalseNeg(), input.getFalsePos());
    }

    /**
     * Creates a new instance from a field map, fulfilling the
     * {@link de.fu_berlin.ties.io.Storable} contract.
     *
     * @param fieldMap map containing the serialized fields
     * @throws IllegalArgumentException if at least one of the parameters is
     * negative or missing
     */
    public FMetrics(final FieldMap fieldMap) throws IllegalArgumentException {
        this(Util.asLong(fieldMap.get(KEY_TRUE_POS)),
            Util.asLong(fieldMap.get(KEY_FALSE_NEG)),
            Util.asLong(fieldMap.get(KEY_FALSE_POS)));
    }

    /**
     * Creates a new instance.
     *
     * @param initTruePos the initial number of true positives
     * @param initFalseNeg the initial number of false negatives
     * @param initFalsePos the initial number of false positives
     * @throws IllegalArgumentException if at least one of the parameters is
     * negative
     */
    public FMetrics(final long initTruePos, final long initFalseNeg,
            final long initFalsePos) throws IllegalArgumentException {
        this();
        checkValues(initTruePos, initFalseNeg, initFalsePos);
        truePos = initTruePos;
        falseNeg = initFalseNeg;
        falsePos = initFalsePos;
    }

    /**
     * Checks the provided parameters, throwing an exception if at least one
     * of them is negative.
     *
     * @param theTruePos the number of true positives to check
     * @param theFalseNeg the number of false negatives to check
     * @param theFalsePos the number of false positives to check
     * @throws IllegalArgumentException if at least one of the parameters is
     * negative
     */
    private void checkValues(final long theTruePos,
            final long theFalseNeg, final long theFalsePos)
            throws IllegalArgumentException {
        if (theTruePos < 0) {
            throw new IllegalArgumentException(
                "Negative number of true positives: " + theTruePos);
        }
        if (theFalseNeg < 0) {
            throw new IllegalArgumentException(
                "Negative number of false negatives: " + theFalseNeg);
        }
        if (theFalsePos < 0) {
            throw new IllegalArgumentException(
                "Negative number of false positives: " + theFalsePos);
        }
    }

    /**
     * Helper method that resets the cached precision + recall values. Must be
     * called whenever input values change.
     */
    private void clearCache() {
        cachedPrecision = -1.0;
        cachedRecall = -1.0;
    }

    /**
     * Returns the number of false negatives (false rejections).
     * @return the number of false negatives
     */
    public final long getFalseNeg() {
        return falseNeg;
    }

    /**
     * Returns the number of false positives (false acceptances).
     * @return the number of false positives
     */
    public final long getFalsePos() {
        return falsePos;
    }

    /**
     * Returns the F-measure, setting <em>alpha</em> = 0.5 so <em>P</em> and
     * <em>R</em> are weighted equal ("F1 measure").
     * F1 = (2 * <em>P</em> * <em>R</em>) / (<em>P</em> + <em>R</em>).
     *
     * @return the F1 measure; a value in the range from 0.0 to 1.0
     */
    public final double getF1Measure() {
        return getFMeasure(0.5);
    }

    /**
     * Returns the F-measure: F = 1 / (<em>alpha</em> * (1/<em>P</em>)
     * + (1-<em>alpha</em>) * (1/<em>R</em>)). <br>
     * <em>F</em> is defined to be 0 if <em>P</em> = 0 or <em>R</em> = 0.
     *
     * @param alpha a factor in the range from 0.0 to 1.0 defining the weighting
     * of precision and recall
     * @return the F-measure; a value in the range from 0.0 to 1.0
     * @throws IllegalArgumentException if <code>alpha</code> is smaller than
     * 0.0 or larger than 1.0
     */
    public final double getFMeasure(final double alpha)
            throws IllegalArgumentException {
        if ((alpha < 0.0) || (alpha > 1.0)) {
            throw new IllegalArgumentException(
                "alpha factor outside of 0.0 to 1.0 range: " + alpha);
        }
        final double inverted = alpha * (1.0 / getPrecision())
            + (1.0 - alpha) * (1.0 / getRecall());

        if (Double.isNaN(inverted)) {
            // we define F-measure to be 0 if precision = 0 or recall = 0
            return 0.0;
        } else {
            // inverted can never be 0
            return 1.0 / inverted;
        }
    }

    /**
     * Returns the precision: <em>P</em> =
     * <em>tp</em> / (<em>tp</em> + <em>fp</em>).
     * @return the precision; a value in the range from 0.0 to 1.0
     */
    public final double getPrecision() {
        if (cachedPrecision < 0.0) {
            // calculate and cache value
            cachedPrecision = ((double) truePos) / (truePos + falsePos);
        } // otherwise the value is already cached

        return cachedPrecision;
    }

    /**
     * Returns the recall: <em>R</em> =
     * <em>tp</em> / (<em>tp</em> + <em>fn</em>).
     * @return the precision; a value in the range from 0.0 to 1.0
     */
    public final double getRecall() {
        if (cachedRecall < 0.0) {
            // calculate and cache value
            cachedRecall = ((double) truePos) / (truePos + falseNeg);
        } // otherwise the value is already cached

        return cachedRecall;
    }

    /**
     * Returns the number of true positives (correct recognitions).
     * @return the number of true positives
     */
    public final long getTruePos() {
        return truePos;
    }

    /**
     * Increases the number of false negatives by 1.
     */
    public final void incFalseNeg() {
        clearCache();
        falseNeg++;
    }

    /**
     * Increases the number of false positives by 1.
     */
    public final void incFalsePos() {
        clearCache();
        falsePos++;
    }

    /**
     * Increases the number of true positives by 1.
     */
    public final void incTruePos() {
        clearCache();
        truePos++;
    }

    /**
     * Stores all relevant fields of this object in a field map for
     * serialization. An equivalent object can be created by calling
     * {@link de.fu_berlin.ties.io.FieldMap#createObject(Class)} on the created
     * field map. The calculated values precision, recall, and F-measure are
     * also stored (they are ignored when
     * {@linkplain #FMetrics(FieldMap) deserializing} a stored instance).
     *
     * @return the created field map
     */
    public FieldMap storeFields() {
        final FieldMap result = new FieldMap();
        result.put(KEY_PRECISION, new Double(getPrecision()));
        result.put(KEY_RECALL, new Double(getRecall()));
        result.put(KEY_F1_MEASURE, new Double(getF1Measure()));
        result.put(KEY_TRUE_POS, new Long(truePos));
        result.put(KEY_FALSE_NEG, new Long(falseNeg));
        result.put(KEY_FALSE_POS, new Long(falsePos));
        return result;
    }

    /**
     * Updates the statistics, increasing the stored values as specified.
     *
     * @param input contains the number of true and false positives and
     * false negatives to add
     * @throws IllegalArgumentException if at least one of the input values is
     * negative
     */
    public final void update(final EvalInput input) throws
            IllegalArgumentException {
        update(input.getTruePos(), input.getFalseNeg(), input.getFalsePos());
    }

    /**
     * Updates the statistics, increasing the stored values as specified.
     *
     * @param addTruePos the number of new true positives to add
     * @param addFalseNeg the number of new false negatives to add
     * @param addFalsePos the number of new false positives to add
     * @throws IllegalArgumentException if at least one of the parameters is
     * negative
     */
    public void update(final long addTruePos, final long addFalseNeg,
            final long addFalsePos) throws IllegalArgumentException {
        checkValues(addTruePos, addFalseNeg, addFalsePos);
        clearCache();
        truePos += addTruePos;
        falseNeg += addFalseNeg;
        falsePos += addFalsePos;
    }

}
