/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.eval;

import de.fu_berlin.ties.io.Storable;

/**
 * A read-only view of the evaluation results calculated by the
 * {@link de.fu_berlin.ties.eval.FMetrics} class and the underlying raw counts.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/10/11 15:30:41 $, $Author: siefkes $
 */
public interface FMetricsView extends EvalInput, Storable {

    /**
     * Returns the F-measure, setting <em>alpha</em> = 0.5 so <em>P</em> and
     * <em>R</em> are weighted equal ("F1 measure").
     * F1 = (2 * <em>P</em> * <em>R</em>) / (<em>P</em> + <em>R</em>).
     *
     * @return the F1 measure; a value in the range from 0.0 to 1.0
     */
    double getF1Measure();

    /**
     * Returns the F-measure: F = 1 / (<em>alpha</em> * (1/<em>P</em>)
     * + (1-<em>alpha</em>) * (1/<em>R</em>)). <br>
     * <em>F</em> is defined to be 0 if <em>P</em> = 0 or <em>R</em> = 0.
     *
     * @param alpha a factor in the range from 0.0 to 1.0 defining the weighting
     * of precision and recall
     * @return the F-measure; a value in the range from 0.0 to 1.0
     * @throws IllegalArgumentException if <code>alpha</code> is smaller than
     * 0.0 or larger than 1.0
     */
    double getFMeasure(final double alpha) throws IllegalArgumentException;

    /**
     * Returns the precision: <em>P</em> =
     * <em>tp</em> / (<em>tp</em> + <em>fp</em>).
     * @return the precision; a value in the range from 0.0 to 1.0
     */
    double getPrecision();

    /**
     * Returns the recall: <em>R</em> =
     * <em>tp</em> / (<em>tp</em> + <em>fn</em>).
     * @return the precision; a value in the range from 0.0 to 1.0
     */
    double getRecall();

}
