/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.eval;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import de.fu_berlin.ties.classify.feature.Feature;
import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.io.BaseStorable;
import de.fu_berlin.ties.io.FieldMap;
import de.fu_berlin.ties.util.Util;

/**
 * Keeps track of the average number of features and of unique features in
 * context representations and of the average number of contexts in documents.
 * Comment-only features are excluded when counting features; comments are
 * ignored when comparing features.
 *
 * <p>Instances of this class are not thread-safe and must be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.8 $, $Date: 2004/09/06 17:23:00 $, $Author: siefkes $
 */
public class FeatureCount extends BaseStorable implements FeatureCountView {

    /**
     * Serialization key for context representations.
     */
    public static final String KEY_CONTEXTS = "Contexts";

    /**
     * Serialization key for documents.
     */
    public static final String KEY_DOCUMENTS = "Documents";

    /**
     * Serialization key for features.
     */
    public static final String KEY_FEATURES = "Features";

    /**
     * Serialization key for characters.
     */
    public static final String KEY_CHARS = "Characters";

    /**
     * Serialization key for unique features.
     */
    public static final String KEY_UNIQUE_FEATURES = "Unique Features";

    /**
     * Serialization key for the average number of context representations in a
     * document.
     */
    public static final String KEY_AVERAGE_CONTEXTS = "Average Contexts";

    /**
     * Serialization key for the average number of features in a context
     * representation.
     */
    public static final String KEY_AVERAGE_FEATURES = "Average Features";

    /**
     * Serialization key for the average number of unique features in a context
     * representation.
     */
    public static final String KEY_AVERAGE_UNIQUE_FEATURES =
        "Average Unique Features";

    /**
     * Serialization key for the average number of characters in a context
     * representation.
     */
    public static final String KEY_CHARS_PER_CONTEXT = "Characters per Context";

    /**
     * Serialization key for the average number of characters in a feature.
     */
    public static final String KEY_CHARS_PER_FEATURE = "Characters per Feature";

    /**
     * The number of characters counted so far.
     */
    private long characters = 0;

    /**
     * The number of context representations evaluated so far.
     */
    private long contexts = 0;

    /**
     * The number of documents counted so far.
     */
    private long documents = 0;

    /**
     * The number of non-comment features encountered so far.
     */
    private long featureSum = 0;

    /**
     * The number of non-comment non-duplicate features encountered so far.
     * Duplicates within the same context representation are ignored; but
     * equal features in different representations are not recognized as
     * duplicate.
     */
    private long uniqueFeatureSum = 0;

    /**
     * Creates a new instance.
     */
    public FeatureCount() {
        super();
    }

    /**
     * Creates a new instance from a field map, fulfilling the
     * {@link de.fu_berlin.ties.io.Storable} contract.
     *
     * @param fieldMap map containing the serialized fields
     * @throws IllegalArgumentException if at least one of the parameters is
     * negative or missing
     */
    public FeatureCount(final FieldMap fieldMap)
            throws IllegalArgumentException {
        this();

        // read values
        documents = Util.asLong(fieldMap.get(KEY_DOCUMENTS));
        contexts = Util.asLong(fieldMap.get(KEY_CONTEXTS));
        featureSum = Util.asLong(fieldMap.get(KEY_FEATURES));
        uniqueFeatureSum = Util.asLong(fieldMap.get(KEY_UNIQUE_FEATURES));
        characters = Util.asLong(fieldMap.get(KEY_CHARS));

        // ensure that no value is negative or missing
        Util.ensureNonNegative(contexts, KEY_DOCUMENTS);
        Util.ensureNonNegative(contexts, KEY_CONTEXTS);
        Util.ensureNonNegative(featureSum, KEY_FEATURES);
        Util.ensureNonNegative(uniqueFeatureSum, KEY_UNIQUE_FEATURES);
        Util.ensureNonNegative(characters, KEY_CHARS);
    }

    /**
     * Counts a document (increases the {@linkplain #getDocuments() number of
     * documents} by one.
     */
    public void countDocument() {
        documents++;
    }

    /**
     * Calculates and returns the average number of context representations in a
     * document.
     *
     * @return the average number of context representations
     */
    public double getAverageContexts() {
        return (double) contexts / documents;
    }

    /**
     * Calculates and returns the average number of non-comment features in a
     * context representation.
     *
     * @return the average number of features
     */
    public double getAverageFeatures() {
        return (double) featureSum / contexts;
    }

    /**
     * Calculates and returns the average number of unique non-comment features
     * in a context representation.
     *
     * @return the average number of features
     */
    public double getAverageUniqueFeatures() {
        return (double) uniqueFeatureSum / contexts;
    }

    /**
     * Returns the number of characters counted so far. Only characters
     * <em>within</em> features are counted; separators between different
     * features are ignored.
     *
     * @return the value of the attribute
     */
    public long getCharacters() {
        return characters;
    }

    /**
     * Calculates and returns the average number of characters in a context
     * representation. Only characters <em>within</em> features are considered;
     * separators between different features are ignored.
     *
     * @return the average number of characters in a context
     */
    public double getCharactersPerContext() {
        return (double) characters / contexts;
    }

    /**
     * Calculates and returns the average number of characters in a feature.
     *
     * @return the average number of characters in a feature
     */
    public double getCharactersPerFeature() {
        return (double) characters / featureSum;
    }

    /**
     * Returns the number of representations evaluated so far.
     * @return the value of the attribute
     */
    public long getContexts() {
        return contexts;
    }

    /**
     * Returns the number of documents counted so far.
     * @return the value of the attribute
     */
    public long getDocuments() {
        return documents;
    }


    /**
     * Returns the number of non-comment features encountered so far.
     * @return the value of the attribute
     */
    public long getFeatureSum() {
        return featureSum;
    }

    /**
     * Returns the number of non-comment non-duplicate features encountered so
     * far. Duplicates within the same context representation are ignored; but
     * equal features in different representations are not recognized as
     * duplicate.
     *
     * @return the value of the attribute
     */
    public long getUniqueFeatureSum() {
        return uniqueFeatureSum;
    }

    /**
     * Stores all relevant fields of this object in a field map for
     * serialization. An equivalent object can be created by calling
     * {@link de.fu_berlin.ties.io.FieldMap#createObject(Class)} on the created
     * field map. The calculated averages are also stored (they are ignored when
     * {@linkplain #FeatureCount(FieldMap) deserializing} a stored instance).
     *
     * @return the created field map
     */
    public FieldMap storeFields() {
        final FieldMap result = new FieldMap();
        result.put(KEY_AVERAGE_CONTEXTS, new Double(getAverageContexts()));
        result.put(KEY_AVERAGE_FEATURES, new Double(getAverageFeatures()));
        result.put(KEY_AVERAGE_UNIQUE_FEATURES,
            new Double(getAverageUniqueFeatures()));
        result.put(KEY_CHARS_PER_CONTEXT,
            new Double(getCharactersPerContext()));
        result.put(KEY_CHARS_PER_FEATURE,
            new Double(getCharactersPerFeature()));
        result.put(KEY_DOCUMENTS, new Long(documents));
        result.put(KEY_CONTEXTS, new Long(contexts));
        result.put(KEY_FEATURES, new Long(featureSum));
        result.put(KEY_UNIQUE_FEATURES, new Long(uniqueFeatureSum));
        result.put(KEY_CHARS, new Long(characters));
        return result;
    }

    /**
     * Evaluates a feature vector and updates the statistics accordingly.
     * Comment-only features are excluded when counting features; comments are
     * ignored when comparing features.
     *
     * @param features a feature vector representing a context
     * @throws ClassCastException if the list contains objects that aren't
     * {@link Feature}s
     */
    public void update(final FeatureVector features) throws ClassCastException {
        final Iterator featureIter = features.iterator();
        Feature currentFeature;
        String currentRep;
        int newFeatures = 0;
        Set<String> uniqueFeatureReps =
            new HashSet<String>(features.size() / 2);

        while (featureIter.hasNext()) {
            currentFeature = (Feature) featureIter.next();
            currentRep = currentFeature.getRepresentation();

            // ignore comment-only features
            if (currentRep != null) {
                newFeatures++;

                // ignore comments by storing only the representation
                uniqueFeatureReps.add(currentRep);

                // increase character count
                characters += currentRep.length();
            }
        }

        // update the statistics
        contexts++;
        featureSum += newFeatures;
        uniqueFeatureSum += uniqueFeatureReps.size();
    }

}
