/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.eval;

import de.fu_berlin.ties.io.Storable;

/**
 * Provides a read-only view on the statistics calculated by the
 * {@link de.fu_berlin.ties.eval.FeatureCount} class and the underlying raw
 * counts.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/02/17 18:10:28 $, $Author: siefkes $
 */
public interface FeatureCountView extends Storable {

    /**
     * Calculates and returns the average number of context representations in a
     * document.
     *
     * @return the average number of context representations
     */
    double getAverageContexts();

    /**
     * Calculates and returns the average number of non-comment features in a
     * context representation.
     *
     * @return the average number of features
     */
    double getAverageFeatures();

    /**
     * Calculates and returns the average number of unique non-comment features
     * in a context representation.
     *
     * @return the average number of features
     */
    double getAverageUniqueFeatures();

    /**
     * Returns the number of characters counted so far. Only characters
     * <em>within</em> features are counted; separators between different
     * features are ignored.
     *
     * @return the value of the attribute
     */
    long getCharacters();

    /**
     * Calculates and returns the average number of characters in a context
     * representation. Only characters <em>within</em> features are considered;
     * separators between different features are ignored.
     *
     * @return the average number of characters in a context
     */
    double getCharactersPerContext();

    /**
     * Calculates and returns the average number of characters in a feature.
     *
     * @return the average number of characters in a feature
     */
    double getCharactersPerFeature();

    /**
     * Returns the number of representations evaluated so far.
     * @return the value of the attribute
     */
    long getContexts();

    /**
     * Returns the number of documents counted so far.
     * @return the value of the attribute
     */
    long getDocuments();


    /**
     * Returns the number of non-comment features encountered so far.
     * @return the value of the attribute
     */
    long getFeatureSum();

    /**
     * Returns the number of non-comment non-duplicate features encountered so
     * far. Duplicates within the same context representation are ignored; but
     * equal features in different representations are not recognized as
     * duplicate.
     *
     * @return the value of the attribute
     */
    long getUniqueFeatureSum();

}
