/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.eval;

import java.util.Collections;
import java.util.Iterator;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.io.FieldContainer;
import de.fu_berlin.ties.io.FieldMap;
import de.fu_berlin.ties.io.RestorableContainer;
import de.fu_berlin.ties.util.Util;

/**
 * Instances of this class manage multiple
 * {@link de.fu_berlin.ties.eval.FMetrics} for different types. They also
 * calculate average statistics over all types.
 *
 * <p>Instances of this class are not thread-safe and must be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.8 $, $Date: 2004/09/06 17:23:00 $, $Author: siefkes $
 */
public class MultiFMetrics implements MultiFMetricsView, RestorableContainer {

    /**
     * Reserved type used for the calculated sums and averages over all types:
     * <code>&lt;All&gt;</code>. This string is reserved and cannot be used to
     * mark regular types.
     */
    public static final String ALL = "<All>";

    /**
     * Serialization key for the type.
     */
    public static final String KEY_TYPE = "Type";

    /**
     * The sums and averages over all types.
     */
    private final FMetrics all;

    /**
     * A sorted map from type Strings to the corresponding {@link FMetrics}.
     */
    private final SortedMap<String, FMetrics> typedMetrics =
        new TreeMap<String, FMetrics>();

    /**
     * Whether {@linkplain FMetricsSummary summaries} are calculated by this
     * type.
     */
    private final boolean calculatingSummaries;

    /**
     * Extension recommended for files storing metrics.
     */
    public static final String EXT_METRICS = "metrics";

    /**
     * Creates a new instance, without calculating
     * {@linkplain FMetricsSummary summaries}.
     */
    public MultiFMetrics() {
        this(false);
    }

    /**
     * Creates a new instance.
     *
     * @param calcSummaries whether this type should calculate
     * {@linkplain FMetricsSummary summaries}
     */
    public MultiFMetrics(final boolean calcSummaries) {
        super();
        calculatingSummaries = calcSummaries;
        all = createMetrics();
    }

    /**
     * Helper factory method that creates a new {@link FMetrics} instance of the
     * required type.
     *
     * @return a {@link SummaryFMetrics} instance if
     * {@link #isCalculatingSummaries()}; a simple {@link FMetrics} instance
     * otherwise
     */
    protected FMetrics createMetrics() {
        if (calculatingSummaries) {
            return new SummaryFMetrics();
        } else {
            return new FMetrics();
        }
    }

    /**
     * Helper factory method that creates a new {@link FMetrics} instance of the
     * required type from a field map.
     *
     * @param fieldMap field map used to initialize the new instance
     * @return a {@link SummaryFMetrics} instance if
     * {@link #isCalculatingSummaries()}; a simple {@link FMetrics} instance
     * otherwise
     * @throws IllegalArgumentException if at least one of the parameters is
     * negative or missing from the field map
     */
    protected FMetrics createMetrics(final FieldMap fieldMap)
            throws IllegalArgumentException {
        if (calculatingSummaries) {
            return new SummaryFMetrics(fieldMap);
        } else {
            return new FMetrics(fieldMap);
        }
    }

    /**
     * Increases the number of false negatives for the given type by 1.
     *
     * @param type the type to update
     * @throws IllegalArgumentException if <code>type</code> is
     * <code>null</code> or empty or equal to the reserved {@link #ALL} type
     */
    public void incFalseNeg(final String type)
            throws IllegalArgumentException {
        // update metrics for this type + for all types
        final FMetrics metrics = lookupType(type);
        metrics.incFalseNeg();
        all.incFalseNeg();
    }

    /**
     * Increases the number of false positives for the given type by 1.
     *
     * @param type the type to update
     * @throws IllegalArgumentException if <code>type</code> is
     * <code>null</code> or empty or equal to the reserved {@link #ALL} type
     */
    public void incFalsePos(final String type)
            throws IllegalArgumentException {
        // update metrics for this type + for all types
        final FMetrics metrics = lookupType(type);
        metrics.incFalsePos();
        all.incFalsePos();
    }

    /**
     * Increases the number of true positives for the given type by 1.
     *
     * @param type the type to update
     * @throws IllegalArgumentException if <code>type</code> is
     * <code>null</code> or empty or equal to the reserved {@link #ALL} type
     */
    public void incTruePos(final String type)
            throws IllegalArgumentException {
        // update metrics for this type + for all types
        final FMetrics metrics = lookupType(type);
        metrics.incTruePos();
        all.incTruePos();
    }

    /**
     * Whether {@linkplain FMetricsSummary summaries} are calculated by this
     * type.
     *
     * @return the value of the attribute
     */
    public boolean isCalculatingSummaries() {
        return calculatingSummaries;
    }

    /**
     * Looks up and if necessary initializes the metrics for a given type.
     * Also checks that the given type is valid, i.e. neither <code>null</code>
     * nor empty nor equal to the reserved {@link #ALL} type.
     *
     * @param type the type to check
     * @return the metrics for this type
     * @throws IllegalArgumentException if <code>type</code> is
     * <code>null</code> or empty or equal to the reserved {@link #ALL} type
     */
    protected FMetrics lookupType(final String type)
            throws IllegalArgumentException {
        // check argument
        if (StringUtils.isEmpty(type)) {
            throw new IllegalArgumentException("Type cannot be null or empty");
        }
        if (ALL.equals(type)) {
            throw new IllegalArgumentException("Cannot use the reserved "
                + "identifier " + ALL + " as regular type");
        }

        FMetrics result = typedMetrics.get(type);
        if (result == null) {
            // create and store metrics for this type
            result = createMetrics();
            typedMetrics.put(type, result);
        }
        return result;
    }

    /**
     * Restores items stored in a field container and adds them to this
     * instance. The provided field container must have been filled by calling
     * {@link #storeEntries(FieldContainer)} on a object of this type.
     * Any values already contained in this instance are added to, they are not
     * replaced. The {@link #viewAll() sums and averages} are recalculated from
     * the provided values, it is not directly read from the container.
     * <em>Summary information cannot be restored.</em>
     *
     * @param fContainer the field container to read
     * @throws IllegalArgumentException if <code>fContainer</code> contains
     * a wrong kind of {@link FieldMap}s
     */
    public void restoreEntries(final FieldContainer fContainer)
            throws IllegalArgumentException {
        String currentType;
        FMetrics currentMetrics;
        FieldMap currentFM;
        final Iterator entryIter = fContainer.entryIterator();

        while (entryIter.hasNext()) {
            currentFM = (FieldMap) entryIter.next();
            currentType = Util.asString(currentFM.get(KEY_TYPE));

            if (currentType == null) {
                throw new IllegalArgumentException("Wrong type of field map: "
                    + KEY_TYPE + " field missing");
            }

            // ignore reserved type (sums + averages will be recalculated)
            if (!ALL.equals(currentType)) {
                currentMetrics = createMetrics(currentFM);

                // feed as input to update to avoid overwriting exisiting values
                update(currentType, currentMetrics);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void storeEntries(final FieldContainer fContainer) {
        String currentType;
        FMetrics currentMetrics;
        FieldMap currentFM;
        final Iterator<String> typeIter = typedMetrics.keySet().iterator();

        // store types in alphabetic order
        while (typeIter.hasNext()) {
            currentType = typeIter.next();
            currentMetrics = typedMetrics.get(currentType);
            currentFM = currentMetrics.storeFields();

            // add type field + store in container
            currentFM.put(KEY_TYPE, currentType);
            fContainer.add(currentFM);
        }

        // store averages
        currentFM = all.storeFields();
        currentFM.put(KEY_TYPE, ALL);
        fContainer.add(currentFM);
    }

    /**
     * Returns a string representation of this object.
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("sums/averages", all)
            .append("typed metrics", typedMetrics)
            .append("calculating summaries", calculatingSummaries)
            .toString();
    }

    /**
     * Returns the set of all types (Strings) currently stored in this
     * instance. The set is immutable and cannot modified.
     *
     * @return the set of type names
     */
    public Set types() {
        return Collections.unmodifiableSet(typedMetrics.keySet());
    }

    /**
     * Updates the statistics by adding the contents of the given multi-metrics.
     *
     * @param multiMetrics the metrics to add
     */
    public void update(final MultiFMetricsView multiMetrics) {
        String currentType;
        FMetricsView currentMetrics;
        final Iterator typeIter = multiMetrics.types().iterator();
        final FMetrics newSum = new FMetrics();

        // iterate types and update each one
        while (typeIter.hasNext()) {
            currentType = (String) typeIter.next();
            currentMetrics = multiMetrics.view(currentType);
            updateType(currentType, currentMetrics.getTruePos(),
                currentMetrics.getFalseNeg(), currentMetrics.getFalsePos());
            newSum.update(currentMetrics);
        }

        // update 'all' in a single step because otherwise the summaries
        // (standard deviations etc.) would get mixed up
        all.update(newSum);
    }

    /**
     * Updates the statistics for a specified type, increasing the stored values
     * as specified.
     *
     * @param type the type to update
     * @param input contains the number of true and false positives and
     * false negatives to add
     * @throws IllegalArgumentException if <code>type</code> is
     * <code>null</code> or empty or equal to the reserved {@link #ALL} type or
     * if at least one of the input values is negative
     */
    public void update(final String type, final EvalInput input) throws
            IllegalArgumentException {
        update(type, input.getTruePos(), input.getFalseNeg(),
            input.getFalsePos());
    }

    /**
     * Updates the statistics for a specified type, increasing the stored values
     * as specified.
     *
     * @param type the type to update
     * @param addTruePos the number of new true positives to add
     * @param addFalseNeg the number of new false negatives to add
     * @param addFalsePos the number of new false positives to add
     * @throws IllegalArgumentException f <code>type</code> is
     * <code>null</code> or empty or equal to the reserved {@link #ALL} type or
     * if at least one of the parameters is negative
     */
    public void update(final String type, final long addTruePos,
            final long addFalseNeg, final long addFalsePos)
            throws IllegalArgumentException {
        updateType(type, addTruePos, addFalseNeg, addFalsePos);
        all.update(addTruePos, addFalseNeg, addFalsePos);
    }

    /**
     * Helper method that updates only the metrics of a specified type,
     * but not the {@linkplain #all sums and averages}.
     *
     * @param type the type to update
     * @param addTruePos the number of new true positives to add
     * @param addFalseNeg the number of new false negatives to add
     * @param addFalsePos the number of new false positives to add
     * @throws IllegalArgumentException f <code>type</code> is
     * <code>null</code> or empty or equal to the reserved {@link #ALL} type or
     * if at least one of the parameters is negative
     */
    private void updateType(final String type, final long addTruePos,
            final long addFalseNeg, final long addFalsePos)
            throws IllegalArgumentException {
        // update metrics for this type + for all types
        final FMetrics metrics = lookupType(type);
        metrics.update(addTruePos, addFalseNeg, addFalsePos);
    }

    /**
     * Returns a read-only view of the {@link FMetrics} of the specified type.
     * This is not a snapshot but will change whenever the underlying counts
     * are changed.
     *
     * @param type the type to check
     * @return a view of the counts and evaluation metrics for the given type;
     * or <code>null</code> if no metrics exist for the given type
     */
    public FMetricsView view(final String type) {
        return typedMetrics.get(type);
    }

    /**
     * Returns a read-only view of the {@link FMetrics} containing the sums and
     * averages over all types. This is not a snapshot but will change whenever
     * the underlying counts are changed.
     *
     * @return a view of the sums and averages over all types
     */
    public FMetricsView viewAll() {
        return all;
    }

    /**
     * Optional operation that shows {@link FMetricsSummary statistical
     * summaries of precision, recall, and F1 metrics} over all types,
     * if {@linkplain #isCalculatingSummaries() calculated}. This is not a
     * snapshot but will change whenever the underlying values are changed.
     *
     * @return a statistical summery over the metrics of all types, if
     * calculated
     * @throws UnsupportedOperationException if
     * {@link #isCalculatingSummaries()} is <code>false</code>
     */
    public FMetricsSummary viewAllSummary()
            throws UnsupportedOperationException {
        if (calculatingSummaries) {
            // thanks to the factory method we know that this interface is
            // implemented
            return (FMetricsSummary) all;
        } else {
            throw new UnsupportedOperationException("No summaries calculated");
        }
    }

    /**
     * Optional operation that shows {@link FMetricsSummary statistical
     * summaries of precision, recall, and F1 metrics} of the specified type,
     * if {@linkplain #isCalculatingSummaries() calculated}. This is not a
     * snapshot but will change whenever the underlying values are changed.
     *
     * @param type the type to check
     * @return a statistical summery over the metrics of the given type, if
     * calculated
     * @throws UnsupportedOperationException if
     * {@link #isCalculatingSummaries()} is <code>false</code>
     */
    public FMetricsSummary viewSummary(final String type)
            throws UnsupportedOperationException {
        if (calculatingSummaries) {
            // thanks to the factory method we know that this interface is
            // implemented
            return (FMetricsSummary) typedMetrics.get(type);
        } else {
            throw new UnsupportedOperationException("No summaries calculated");
        }
    }

}
