/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.eval;

import java.util.Set;

import de.fu_berlin.ties.io.StorableContainer;

/**
 * A read-only view of multiple {@link de.fu_berlin.ties.eval.FMetrics}
 * and the sums and averages calculated over them.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/02/19 18:15:51 $, $Author: siefkes $
 */
public interface MultiFMetricsView extends StorableContainer {

    /**
     * Returns the set of all types (Strings) currently stored in this
     * instance. The set is immutable and cannot modified.
     *
     * @return the set of type names
     */
    Set types();

    /**
     * Returns a read-only view of the {@link FMetrics} of the specified type.
     * This is not a snapshot but will change whenever the underlying counts
     * are changed.
     *
     * @param type the type to check
     * @return a view of the counts and evaluation metrics for the given type;
     * or <code>null</code> if no metrics exist for the given type
     */
    FMetricsView view(final String type);

    /**
     * Returns a read-only view of the {@link FMetrics} containing the sums and
     * averages over all types. This is not a snapshot but will change whenever
     * the underlying counts are changed.
     *
     * @return a view of the sums and averages over all types
     */
    FMetricsView viewAll();

    /**
     * Optional operation that shows {@link FMetricsSummary statistical
     * summaries of precision, recall, and F1 metrics} over all types,
     * if calculated by the used implementation. This is not a snapshot but
     * will change whenever the underlying values are changed.
     *
     * @return a statistical summery over the metrics of all types
     * @throws UnsupportedOperationException if this method is not supported by
     * the used implementation
     */
    FMetricsSummary viewAllSummary() throws UnsupportedOperationException;

    /**
     * Optional operation that shows {@link FMetricsSummary statistical
     * summaries of precision, recall, and F1 metrics} of the specified type,
     * if calculated by the used implementation. This is not a snapshot but
     * will change whenever the underlying values are changed.
     *
     * @param type the type to check
     * @return a statistical summery over the metrics of the given type
     * @throws UnsupportedOperationException if this method is not supported by
     * the used implementation
     */
    FMetricsSummary viewSummary(final String type)
            throws UnsupportedOperationException;

}
