/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.eval;

import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.Collection;
import java.util.Iterator;

import org.apache.commons.collections.MultiHashMap;
import org.apache.commons.collections.MultiMap;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TextProcessor;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.extract.EvaluatedExtractionContainer;
import de.fu_berlin.ties.extract.Extraction;
import de.fu_berlin.ties.extract.ExtractionContainer;
import de.fu_berlin.ties.extract.Extractor;
import de.fu_berlin.ties.extract.TargetStructure;
import de.fu_berlin.ties.io.FieldContainer;
import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.util.Util;

/**
 * A processor that can be used to re-evaluate the contents of an
 * {@link de.fu_berlin.ties.extract.EvaluatedExtractionContainer}. This can
 * be used to switch the
 * {@link de.fu_berlin.ties.extract.EvaluatedExtractionContainer#isMatchingAll()
 * match mode}.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.10 $, $Date: 2004/06/17 14:01:41 $, $Author: siefkes $
 */
public class ReEvaluator extends TextProcessor {

    /**
     * Creates a new instance, using a default extension and the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     */
    public ReEvaluator() {
        this(Extractor.EXT_EXTRACTIONS);
    }

    /**
     * Creates a new instance, using the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @param outExt the extension to use for output files
     */
    public ReEvaluator(final String outExt) {
        this(outExt, TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance.
     *
     * @param outExt the extension to use for output files
     * @param conf the configuration to use
     */
    public ReEvaluator(final String outExt, final TiesConfiguration conf) {
        super(outExt, conf);
    }

    /**
     * Created an empty container for evaluating extractions, using the
     * stored configuration to initialize
     * {@linkplain ExtractionContainer#getTargetStructure() target structure}
     * and {@linkplain EvaluatedExtractionContainer#isMatchingAll() match mode}.
     * Subclasses can overwrite this method if the configured values are
     * inadequate.
     *
     * @return the created empty container
     */
    protected EvaluatedExtractionContainer createdEvalContainer() {
        return new EvaluatedExtractionContainer(
            new TargetStructure(getConfig()), getConfig());
    }

    /**
     * Re-evaluates the contents of an extraction container.
     *
     * @param orgExtractions the extractions to re-evaluate
     * @param newEvaluated the re-evaluated extractions are added to this
     * conainer
     */
    public void reEvalulate(final ExtractionContainer orgExtractions,
            final EvaluatedExtractionContainer newEvaluated) {
        // sort into batches based on sources (might include "null" batch)
        final MultiMap batchMap = new MultiHashMap();
        final Iterator extIter = orgExtractions.iterator();
        Extraction currentExt;

        while (extIter.hasNext()) {
            currentExt = (Extraction) extIter.next();
            batchMap.put(currentExt.getSource(), currentExt);
        }

        ExtractionContainer currentPredictions;
        ExtractionContainer currentAnswers;
        final Iterator batchIter = batchMap.keySet().iterator();
        String currentSource;
        Collection currentColl;
        Iterator collIter;
        EvalStatus currentStatus;

        // separated and re-evaluate extractions for each batch
        while (batchIter.hasNext()) {
            currentSource = (String) batchIter.next();
            currentColl = (Collection) batchMap.get(currentSource);
            currentPredictions =
                new ExtractionContainer(newEvaluated.getTargetStructure());
            currentAnswers =
                new ExtractionContainer(newEvaluated.getTargetStructure());
            collIter = currentColl.iterator();

            while (collIter.hasNext()) {
                currentExt = (Extraction) collIter.next();
                currentStatus = currentExt.getEvalStatus();

                // put each extraction into suitable container(s)
                if (currentStatus.isPredictionState()) {
                    currentPredictions.add(currentExt);
                }
                if (currentStatus.isAnswerState()) {
                    currentAnswers.add(currentExt);
                }
            }

            // evaluate batch
            newEvaluated.evaluateBatch(currentPredictions, currentAnswers,
                currentSource);
        }
    }

    /**
     * Re-evaluates the serialized contents of an extraction container,
     * delegating to
     * {@link #reEvalulate(ExtractionContainer, EvaluatedExtractionContainer)}.
     *
     * @param reader reader containg the extractions to re-evaluate in
     * {@link de.fu_berlin.ties.io.DelimSepValues} format; not closed by this
     * method
     * @return the re-evaluated extractions
     * @throws IOException if an I/O error occurs while reading the extractions
     */
    public EvaluatedExtractionContainer reEvalulate(final Reader reader)
            throws IOException {
        final EvaluatedExtractionContainer result = createdEvalContainer();

        // read original extraction
        final FieldContainer fContainer =
            FieldContainer.createFieldContainer(reader);
        final ExtractionContainer orgExtraction =
            new ExtractionContainer(result.getTargetStructure(), fContainer);

        // delegate + return filled result
        reEvalulate(orgExtraction, result);
        return result;
    }

    /**
     * {@inheritDoc}
     */
    protected void doProcess(final Reader reader, final Writer writer,
            final ContextMap context) throws IOException, ProcessingException {
        final EvaluatedExtractionContainer result = reEvalulate(reader);

        // serialize results + metrics
        final FieldContainer storage = FieldContainer.createFieldContainer();
        result.storeEntries(storage);
        storage.store(writer);

        // serialize metrics in same format
        final File metricsFile = IOUtils.createOutFile((File)
            context.get(KEY_DIRECTORY), (String) context.get(KEY_LOCAL_NAME),
            MultiFMetrics.EXT_METRICS);
        final Writer metricsWriter = IOUtils.openWriter(metricsFile,
            getConfig());

        try {
            final FieldContainer metricsStorage =
                FieldContainer.createFieldContainer();
            result.viewMetrics().storeEntries(metricsStorage);
            metricsStorage.store(metricsWriter);
            metricsWriter.flush();
        } finally {
            IOUtils.tryToClose(metricsWriter);
        }

        Util.LOG.info("Stored corresponding metrics in " + metricsFile);
    }

}
