/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.eval;

import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.CollectingProcessor;
import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.text.TextUtils;

/**
 * Arranges all input arguments (for example, files or URLs) in random
 * "shuffles", so they can subsequently processed in random (but fixed) order.
 * This helps generating input data for experiments. Each random shuffle will
 * be stored in a file; each line of the file will contain one of the input
 * arguments; the order of input arguments will be random. If the input
 * arguments are URIs/URLs, this format correspondings to the
 * <code>text/uri-list</code> media type defined in
 * <a href="http://www.faqs.org/rfcs/rfc2483.html">RFC 2483</a>; it can be
 * read using the {@link de.fu_berlin.ties.io.IOUtils#readURIList(CharSequence)}
 * methods.
 *
 * <p>Also contains the static {@link #shuffle(List)} method so it can
 * be used as a static utility class.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2004/11/16 10:41:19 $, $Author: siefkes $
 */
public class ShuffleGenerator extends CollectingProcessor {

    /**
     * Static utility method that "shuffles" a list by re-arranging its elements
     * in random order. The original list will be destroyed by this method
     * (i.e. it might be empty after this method returns)! It is recommended
     * to pass a {@link LinkedList} (and not a {@link ArrayList}) as input
     * argument because iteratively deleting elements from {@link ArrayList}s
     * is very inefficient.
     *
     * @param input the list of elements to shuffle; should be a
     * {@link LinkedList} for efficiency; will be destroyed by this method
     * (i.e. it might be empty after this method returns)
     * @return a new list containing the elements of the <code>input</code>
     * in random order
     */
    public static <T> List<T> shuffle(final List<T> input) {
        final Random random = new Random();
        final List<T> result = new ArrayList<T>(input.size());

        while (!input.isEmpty()) {
            // randomly remove a list element and append it to the result list
            result.add(input.remove(random.nextInt(input.size())));
        }

        return result;
    }

    /**
     * The number used in the name of the first generated shuffle file.
     */
    private final int first;

    /**
     * The number of random shuffles to generate.
     */
    private final int number;

    /**
     * Prefix used for the generated random shuffle files.
     */
    private final String prefix;

    /**
     * Extension used for the generated random shuffle files.
     */
    private final String extension;

    /**
     * Creates a new instance from the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     */
    public ShuffleGenerator() {
        this(TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance from the provided configuration.
     *
     * @param conf used to configure this instance; must not be
     * <code>null</code>
     */
    public ShuffleGenerator(final TiesConfiguration conf) {
        this(conf.getInt("shuffle.first"), conf.getInt("shuffle.num"),
            conf.getString("shuffle.prefix"), conf.getString("shuffle.ext"),
            conf);
    }

    /**
     * Creates a new instance from the provided configuration. Generated
     * shuffle files will be named from <code>filePrefixNUMBER.fileExt</code>,
     * where <code>NUMBER</code> is a number from <em>firstNum</em>
     * to <em>firstNum + num - 1</em>; if necessary, numbers will be padded
     * with leading zeros so all numbers are of the same length
     * (e.g. <em>01</em> to <em>10</em>).
     *
     * @param firstNum The number used in the name of the first generated
     * shuffle file (typically 1)
     * @param num The number of random shuffles to generate
     * @param filePrefix Prefix used for the generated random shuffle files
     * @param fileExt Extension used for the generated random shuffle files
     * @param conf passed to the superclass; if <code>null</code>,
     * the {@linkplain TiesConfiguration#CONF standard configuration} is used
     */
    public ShuffleGenerator(final int firstNum, final int num,
            final String filePrefix, final String fileExt,
            final TiesConfiguration conf) {
        super(conf);
        first = firstNum;
        number = num;
        prefix = filePrefix;
        extension = fileExt;
    }

    /**
     * {@inheritDoc}
     */
    public void process(final List<String> collected, final ContextMap context)
            throws IOException {
        LinkedList<String> inputCopy;
        List resultList;
        Iterator resultIter;
        String item;
        final int last = first + number - 1;
        final File outDirectory = IOUtils.determineOutputDirectory(getConfig());
        File outFile;
        Writer writer;

        // left-pad numbers with 0s if required
        final NumberFormat formatter = NumberFormat.getIntegerInstance();
        formatter.setMinimumIntegerDigits(Integer.toString(last).length());

        // generate + store shuffles
        for (int i = first; i <= last; i++) {
            // create a new copy of the list because it will be destroyed
            inputCopy = new LinkedList<String>(collected);
            resultList = shuffle(inputCopy);

            // write to shuffle file in the output directory
            outFile = IOUtils.createOutFile(outDirectory,
                prefix + formatter.format(i), extension);
            writer = IOUtils.openWriter(outFile, getConfig());

            try {
                resultIter = resultList.iterator();

                // write each item in a single line
                while (resultIter.hasNext()) {
                    item = (String) resultIter.next();
                    writer.write(item);
                    writer.write(TextUtils.LINE_SEPARATOR);
                }
            } finally {
                // close the writer
                IOUtils.tryToClose(writer);
            }
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("first", first)
            .append("number", number)
            .append("prefix", prefix)
            .append("extension", extension)
            .toString();
    }

}
