/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.eval;

import org.apache.commons.math.stat.StatisticalSummary;
import org.apache.commons.math.stat.SummaryStatistics;

import de.fu_berlin.ties.io.FieldMap;

/**
 * {@link de.fu_berlin.ties.eval.FMetrics} extension that additionally
 * calculates a {@link org.apache.commons.math.stat.StatisticalSummary} of the
 * intermediate precision, recall, and F1 metrics resulting from different
 * {@link #update(long, long, long) update} operations. Values added by
 * {@link de.fu_berlin.ties.eval.FMetrics#incFalseNeg() inc...} operations
 * are <em>not</em> considered for summaries.
 *
 * <p>Especially the
 * {@link org.apache.commons.math.stat.StatisticalSummary#getStandardDeviation()
 * standard deviations} might be interested -- thus they are
 * {@linkplain #storeFields() serialized} in addition to the values stored by
 * the superclass. The summaries cannot be restored during
 * {@link #SummaryFMetrics(FieldMap) deserialization} because not all required
 * information is serialized.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/02/20 17:49:18 $, $Author: siefkes $
 */
public class SummaryFMetrics extends FMetrics implements FMetricsSummary {

    /**
     * Prefix of serialization keys for standard deviations.
     */
    public static final String PREFIX_STD_DEVIATION = "Std. Deviation ";

    /**
     * Summary statistics for the precision.
     */
    private final SummaryStatistics precisionSummary =
        SummaryStatistics.newInstance();

    /**
     * Summary statistics for the recall.
     */
    private final SummaryStatistics recallSummary =
        SummaryStatistics.newInstance();

    /**
     * Summary statistics for the F1 measure.
     */
    private final SummaryStatistics f1Summary = SummaryStatistics.newInstance();

    /**
     * Creates a new empty instance.
     */
    public SummaryFMetrics() {
        super();
    }

    /**
     * Creates a new instance from a field map, fulfilling the
     * {@link de.fu_berlin.ties.io.Storable} contract. <strong>Any summary
     * information stored in the initial instance will be lost, thus summaries
     * calculated after restoring will be inaccurate.</strong>
     *
     * @param fieldMap map containing the serialized fields
     * @throws IllegalArgumentException if at least one of the parameters is
     * negative or missing
     */
    public SummaryFMetrics(final FieldMap fieldMap)
            throws IllegalArgumentException {
        super(fieldMap);
    }

    /**
     * Stores all relevant fields of this object in a field map for
     * serialization. An equivalent object can be created by calling
     * {@link de.fu_berlin.ties.io.FieldMap#createObject(Class)} on the created
     * field map. The calculated values precision, recall, and F-measure and
     * their standard deviations are also stored (they are ignored when
     * {@linkplain #SummaryFMetrics(FieldMap) deserializing} a stored instance).
     *
     * @return the created field map
     */
    public FieldMap storeFields() {
        // delegate to super + add standard deviations
        final FieldMap result = super.storeFields();
        result.put(PREFIX_STD_DEVIATION + KEY_PRECISION,
            new Double(precisionSummary.getStandardDeviation()));
        result.put(PREFIX_STD_DEVIATION + KEY_RECALL,
            new Double(recallSummary.getStandardDeviation()));
        result.put(PREFIX_STD_DEVIATION + KEY_F1_MEASURE,
            new Double(f1Summary.getStandardDeviation()));
        return result;
    }

    /**
     * Updates the statistics, increasing the stored values as specified and
     * updating the statistical summaries.
     *
     * @param addTruePos the number of new true positives to add
     * @param addFalseNeg the number of new false negatives to add
     * @param addFalsePos the number of new false positives to add
     * @throws IllegalArgumentException if at least one of the parameters is
     * negative
     */
    public final void update(final long addTruePos, final long addFalseNeg,
            final long addFalsePos) throws IllegalArgumentException {
        // delegate to super + update summaries
        super.update(addTruePos, addFalseNeg, addFalsePos);
        final FMetrics newMetrics =
            new FMetrics(addTruePos, addFalseNeg, addFalsePos);
        precisionSummary.addValue(newMetrics.getPrecision());
        recallSummary.addValue(newMetrics.getRecall());
        f1Summary.addValue(newMetrics.getF1Measure());
    }

    /**
     * Returns a summary view on the F1 values. This is not a snapshot but
     * will change whenever the underlying values are changed.
     *
     * @return a summary view on the F1 value
     */
    public StatisticalSummary viewF1Summary() {
        return f1Summary;
    }

    /**
     * Returns a summary view on the precision values. This is not a snapshot
     * but will change whenever the underlying values are changed.
     *
     * @return a summary view on the precision value
     */
    public StatisticalSummary viewPrecisionSummary() {
        return precisionSummary;
    }

    /**
     * Returns a summary view on the recall values. This is not a snapshot but
     * will change whenever the underlying values are changed.
     *
     * @return a summary view on the recall value
     */
    public StatisticalSummary viewRecallSummary() {
        return recallSummary;
    }

}
