/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract;

import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Document;
import org.dom4j.Element;
import org.dom4j.NodeFilter;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.DocumentReader;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;

import de.fu_berlin.ties.io.FieldContainer;
import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.text.TokenContainer;
import de.fu_berlin.ties.text.TokenDetails;
import de.fu_berlin.ties.text.TokenizerFactory;
import de.fu_berlin.ties.xml.dom.DOMUtils;
import de.fu_berlin.ties.xml.dom.DocumentWalker;
import de.fu_berlin.ties.xml.dom.ElementNameFilter;
import de.fu_berlin.ties.xml.dom.ElementProcessor;

/**
 * Buildings an {@link de.fu_berlin.ties.extract.ExtractionContainer} of answer
 * keys from an annotated text (in XML format).
 *
 * <p>Instances of this class are thread-safe and can process several documents
 * in parallel.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.20 $, $Date: 2004/11/19 14:04:41 $, $Author: siefkes $
 */
public class AnswerBuilder extends DocumentReader implements ElementProcessor {

    /**
     * Context key referring to the extraction container used for storing the
     * answer keys.
     */
    public static final String KEY_ANSWERS = "answer-keys";

    /**
     * The recommended file extension to use for storing answer keys.
     */
    public static final String EXT_ANSWERS = "ans";

    /**
     * Reads back answer keys stored by the
     * {@link #process(Document, Writer, ContextMap)} method of an instance of
     * this class.
     *
     * @param targetStruct the target structure used when creating the answer
     * keys
     * @param file the file containing the answer keys
     * @param config configuration used to determine the character set of the
     * keys (cf. {@link IOUtils#openReader(File, Configuration)}
     * @return an extraction container of the answer keys
     * @throws IllegalArgumentException if the
     * (@linkplain de.fu_berlin.ties.classify.Prediction#getType() type) of
     * some answer keys in the answer keys doesn't fit the target structure
     * @throws IOException if an I/O error occurs while reading the file
     */
    public static ExtractionContainer readAnswerKeys(
            final TargetStructure targetStruct, final File file,
            final Configuration config)
            throws IllegalArgumentException, IOException {
        final Reader reader = IOUtils.openReader(file, config);
        final FieldContainer fContainer =
            FieldContainer.createFieldContainer(reader);
        return new ExtractionContainer(targetStruct, fContainer);
    }

    /**
     * Reads the answer keys corresponding to a file. The answer keys must be
     * in a file ending in {@link AnswerBuilder#EXT_ANSWERS} instead of the
     * extension of the original file.
     *
     * @param targetStruct the target structure used when creating the answer
     * keys
     * @param orgFile the file whose answer keys should be returned
     * @param config configuration used to determine the character set of the
     * keys (cf. {@link IOUtils#openReader(File, Configuration)}
     * @return an extraction container of the answer keys
     * @throws IllegalArgumentException if the
     * (@linkplain de.fu_berlin.ties.classify.Prediction#getType() type) of
     * some answer keys in the answer keys doesn't fit the target structure
     * @throws IOException if an I/O error occurs while reading the file
     */
    public static ExtractionContainer readCorrespondingAnswerKeys(
            final TargetStructure targetStruct, final File orgFile,
            final Configuration config)
            throws IllegalArgumentException, IOException {
        // replace extension + delegate
        final File answerFile = new File(orgFile.getParentFile(),
                IOUtils.getBaseName(orgFile) + IOUtils.EXT_SEPARATOR
                + EXT_ANSWERS);
        return readAnswerKeys(targetStruct, answerFile, config);
    }

    /**
     * The target structure specifying the classes to recognize.
     */
    private final TargetStructure targetStructure;

    /**
     * Used to instantiate tokenizers.
     */
    private final TokenizerFactory factory;

    /**
     * Creates a new instance, using a default extension and configuring the
     * target structure from the {@linkplain TiesConfiguration#CONF standard
     * configuration}.
     */
    public AnswerBuilder() {
        this(EXT_ANSWERS);
    }

    /**
     * Creates a new instance, configuring the target structure from the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @param outExt the extension to use for output files
     */
    public AnswerBuilder(final String outExt) {
        this(outExt, TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance, configuring the target structure from the
     * provided configuration.
     *
     * @param outExt the extension to use for output files
     * @param config the configuration to use
     */
    public AnswerBuilder(final String outExt, final TiesConfiguration config) {
        this(outExt, new TargetStructure(config), new TokenizerFactory(config),
            config);
    }

    /**
     * Creates a new instance.
     * @param outExt the extension to use for output files
     * @param targetStruct the target structure specifying the classes to
     * recognize
     * @param tFactory used to instantiate tokenizers
     * @param config the configuration to use
     */
    public AnswerBuilder(final String outExt,
            final TargetStructure targetStruct, final TokenizerFactory tFactory,
            final TiesConfiguration config) {
        super(outExt, config);
        targetStructure = targetStruct;
        factory = tFactory;
    }

    /**
     * Buildings an {@link de.fu_berlin.ties.extract.ExtractionContainer} of
     * answer keys from from an annotated XML document.
     *
     * @param document the document to read
     * @return a container of the answer keys of this document
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public ExtractionContainer buildAnswers(final Document document)
            throws IOException, ProcessingException {
        // create container of answer keys and store in context
        final ExtractionContainer answerKeys =
            new ExtractionContainer(targetStructure);
        final ContextMap context = new ContextMap();
        context.put(KEY_ANSWERS, answerKeys);

        final NodeFilter filter =
            new ElementNameFilter(targetStructure.getClassNames());
        final DocumentWalker walker = new DocumentWalker(filter, this, factory);

        // the walker will call back (process method) for each element matching
        // a class name
        walker.walk(document, context);

        // return the container of answer keys
        return answerKeys;
    }

    /**
     * Returns the target structure specifying the classes to recognize.
     * @return the used target structure
     */
    public TargetStructure getTargetStructure() {
        return targetStructure;
    }

    /**
     * Buildings an {@link de.fu_berlin.ties.extract.ExtractionContainer} of
     * answer keys from from an annotated XML document.
     *
     * @param document the document to read
     * @param writer the writer to write the processed text to; flushed
     * but not closed by this method
     * @param context a map of objects that are made available for processing
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public void process(final Document document, final Writer writer,
            final ContextMap context) throws IOException, ProcessingException {
        // delegate to build method
        final ExtractionContainer answers = buildAnswers(document);

        // serialize results
        final FieldContainer storage = FieldContainer.createFieldContainer();
        answers.storeEntries(storage);
        storage.store(writer);
    }

    /**
     * Classifies an element in an XML document, building features and
     * delegating to the classifier.
     *
     * @param element the element to process
     * @param tokenContainer a container storing all tokens seen in the
     * document so far; {@link TokenContainer#getLast()} contains the textual
     * content of the element and its child elements
     * @param context a map of objects that are made available for processing;
     * the {@link #KEY_ANSWERS} key must map to an extraction container used for
     * storing the answer keys
     */
    public void processElement(final Element element,
            final TokenContainer tokenContainer, final ContextMap context) {
        // create and store an extraction of the element type
        final TokenDetails details = new TokenDetails(tokenContainer.getLast(),
                tokenContainer.getFirstTokenInLastRep(),
                tokenContainer.getFirstTokenInLastIndex(), false);
        final Extraction extraction =
            new Extraction(DOMUtils.name(element), details);
        final ExtractionContainer answerKeys =
            (ExtractionContainer) context.get(KEY_ANSWERS);
        answerKeys.add(extraction);
    }

    /**
     * Returns a string representation of this object.
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("target structure", targetStructure)
            .toString();
    }

}
