/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.PredictionComparator;
import de.fu_berlin.ties.eval.EvalStatus;
import de.fu_berlin.ties.eval.MultiFMetrics;
import de.fu_berlin.ties.eval.MultiFMetricsView;

/**
 * An extraction container that evaluates containers of predicted extractions
 * against containers of true extractions (answer keys) and merges their
 * contents, setting the {@linkplain de.fu_berlin.ties.eval.EvalStatus
 * evaluation states} accordingly. Extractions in both containers are marked as
 * {@link de.fu_berlin.ties.eval.EvalStatus#CORRECT} (true positives),
 * extractions found only in the container of predictions are marked as
 * {@link de.fu_berlin.ties.eval.EvalStatus#SPURIOUS} (false positives), while
 * those occurring only in the answer keys are
 * {@link de.fu_berlin.ties.eval.EvalStatus#MISSING} (false negatives).
 *
 * <p><a name="match-modes">Two <strong>match modes</strong> are supported:</a>
 * in <em>match-all mode</em>, all extractions are evaluated.
 * In <em>match-best mode</em>, only the recognition with the highest
 * probability of each type is evaluated (in each batch), while the others are
 * {@linkplain de.fu_berlin.ties.eval.EvalStatus#IGNORED ignored} and not
 * counted; it is thus suffienct to match (or fail to match) a single answer key
 * -- the other answer keys are marked as
 * {@link de.fu_berlin.ties.eval.EvalStatus#ALTERNATIVE}s and not counted.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.10 $, $Date: 2004/09/06 17:23:10 $, $Author: siefkes $
 */
public class EvaluatedExtractionContainer extends ExtractionContainer {

    /**
     * Configuration key: whether to use <em>match-all</em> or
     * <em>match-best</em> as <a href="#match-modes">match mode</a>.
     */
    public static final String CONFIG_MATCH_ALL = "eval.match.all";

    /**
     * Configuration key for {@link #isMatchingPosition()}.
     */
    public static final String CONFIG_MATCH_POSITION = "eval.match.pos";

    /**
     * Counts the true positives (correct extractions), false negatives (missing
     * extractions) and false positives (spurious extractions) stored in this
     * container of each type as well as for all types; calculates appropriate
     * statistics.
     */
    private final MultiFMetrics multiMetrics = new MultiFMetrics();

    /**
     * The <a href="#match-modes">match mode</a>: <code>true</code> for
     * <em>match-all mode</em>, <code>false</code> for <em>match-best mode</em>.
     */
    private final boolean matchingAll;

    /**
     * If <code>true</code>, the positions of extraction and answer keys must
     * match; otherwise only their contents must match (string compare).
     */
    private final boolean matchingPosition;

    /**
     * Used to compare the pR values resp. probabilities of predictions.
     */
    private final PredictionComparator predComp = new PredictionComparator();

    /**
     * Creates a new instance.
     *
     * @param targetStruct the target structure specifying the classes to
     * recognize
     * @param config used to configure this instance
     */
    public EvaluatedExtractionContainer(final TargetStructure targetStruct,
            final TiesConfiguration config) {
        this(targetStruct, config.getBoolean(CONFIG_MATCH_ALL),
                config.getBoolean(CONFIG_MATCH_POSITION));
    }

    /**
     * Creates a new instance.
     *
     * @param targetStruct the target structure specifying the classes to
     * recognize
     * @param matchAll sets the <a href="#match-modes">match mode</a>:
     * <code>true</code> for <em>match-all mode</em>, <code>false</code> for
     * <em>match-best mode</em>
     * @param matchPosition if <code>true</code>, the positions of extraction
     * and answer keys must match; otherwise only their contents must match
     * (string compare)
     */
    public EvaluatedExtractionContainer(final TargetStructure targetStruct,
            final boolean matchAll, final boolean matchPosition) {
        super(targetStruct);
        matchingAll = matchAll;
        matchingPosition = matchPosition;
    }

    /**
     * Adds an extraction to this container. Only <em>evaluated</em> extractions
     * can be added to this container, i.e., the evaluation states
     * {@link EvalStatus#UNKNOWN} and {@link EvalStatus#TRUTH} are not allowed.
     * The evaluation states {@link EvalStatus#IGNORED} and
     * {@link EvalStatus#ALTERNATIVE} are only allowed in <em>match-best
     * <a href="#match-modes">mode</a></em>; in <em>match-all mode</em>
     * ({@link #isMatchingAll()} == <code>true</code>) they don't make sense
     * and are thus rejected.
     *
     * @param extraction the extraction to add
     * @throws IllegalArgumentException if the extraction's {@linkplain
     * de.fu_berlin.ties.classify.Prediction#getEvalStatus() evaluation status}
     * is invalid (cf. above); or if the class of the extraction is not
     * in the set of classes accepted by this container
     */
    public void add(final Extraction extraction)
            throws IllegalArgumentException {
        final EvalStatus evalStatus = extraction.getEvalStatus();

        // check argument
        if ((evalStatus == EvalStatus.UNKNOWN)
                || (evalStatus == EvalStatus.TRUTH)) {
            throw new IllegalArgumentException("Illegal eval status: "
                + evalStatus);
        }
        if (isMatchingAll() && ((evalStatus == EvalStatus.IGNORED)
                || (evalStatus == EvalStatus.ALTERNATIVE))) {
            throw new IllegalArgumentException(
                "Illegal eval status for match-all mode: " + evalStatus);
        }

        // delegate to super
        super.add(extraction);

        // update statistics
        if (evalStatus == EvalStatus.CORRECT) {
            multiMetrics.incTruePos(extraction.getType());
        } else if (evalStatus == EvalStatus.MISSING) {
            multiMetrics.incFalseNeg(extraction.getType());
        } else if (evalStatus == EvalStatus.SPURIOUS) {
            multiMetrics.incFalsePos(extraction.getType());
        }
    }

    /**
     * Helper method that adds a bunch of extractions, setting their evaluation
     * status and source as specified.
     *
     * @param extractionColl a collection of {@link Extraction}s to add; the
     * collection will be {@linkplain Collection#clear() cleared} by this
     * method, i.e., it will be empty after the method returned
     * @param status the status of all extractions in the collection will be
     * set to this value prior to adding them
     * @param source the
     * {@linkplain de.fu_berlin.ties.classify.Prediction#getSource() source} to
     * set for all extractions in the collection; ignored if <code>null</code>
     */
    protected void addAllAndClear(final Collection extractionColl,
            final EvalStatus status, final String source) {
        final Iterator iter = extractionColl.iterator();
        Extraction currentExtraction;

        while (iter.hasNext()) {
            currentExtraction = (Extraction) iter.next();

            // adjust values
            currentExtraction.setEvalStatus(status);
            if (source != null) {
                currentExtraction.setSource(source);
            }

            // add extraction
            add(currentExtraction);
        }

        // reset collection
        extractionColl.clear();
    }

    /**
     * Evaluates a container of predicted extractions against a container of
     * true extractions (answer keys) and adds them to this instance.
     * In <em>match-best mode</em>, only the extraction with the highest
     * pR resp. probability of each type in this batch is evaluated.
     * 
     * <p><strong>Warning: If positions are ignored for matching (see
     * {@linkplain #EvaluatedExtractionContainer(TargetStructure, boolean,
     * boolean) constructor}), the two given containers are modified
     * by {@linkplain ExtractionContainer#unsetPositions() unsetting the
     * positions} of all extractions and removing duplicates (extractions with
     * same type + text but different positions).</strong>
     *
     * @param predicted the container of predicted extractions
     * @param expected the container of expected extractions (answer keys)
     * @param source the
     * {@linkplain de.fu_berlin.ties.classify.Prediction#getSource() source}
     * to add to add extractions of this batch; ignored if <code>null</code>
     * @throws IllegalArgumentException if the
     * {@linkplain ExtractionContainer#getTargetStructure() target structures}
     * of the two containers differ from the target structure of this one
     */
    public void evaluateBatch(final ExtractionContainer predicted,
            final ExtractionContainer expected, final String source)
            throws IllegalArgumentException {
        // check arguments
        if (!predicted.getTargetStructure().equals(getTargetStructure())) {
            throw new IllegalArgumentException(
                "Target structures of predictions differ: "
                + predicted.getTargetStructure() + "; "
                + getTargetStructure());
        }
        if (!expected.getTargetStructure().equals(getTargetStructure())) {
            throw new IllegalArgumentException(
                "Target structures of answer keys differs: "
                + expected.getTargetStructure() + "; "
                + getTargetStructure());
        }

        if (!matchingPosition) {
            // unset positions (remove resulting duplicates) in both containers
            predicted.unsetPositions();
            expected.unsetPositions();
        }

        // compare containers and store results for each accepted class
        final Iterator acceptedIter =
            getTargetStructure().getClassNames().iterator();
        String currentClassName;
        Iterator extractionIter;
        Extraction currentExtraction;

        // temporary sets of extractions with different states
        final LinkedHashSet<Extraction> correct =
            new LinkedHashSet<Extraction>();
        final LinkedHashSet<Extraction> missing =
            new LinkedHashSet<Extraction>();
        final LinkedHashSet<Extraction> spurious =
            new LinkedHashSet<Extraction>();

        // sets + variables used in match-best mode
        final LinkedHashSet<Extraction> altern =
            matchingAll ? null : new LinkedHashSet<Extraction>();
        final LinkedHashSet<Extraction> ignored =
            matchingAll ? null : new LinkedHashSet<Extraction>();
        final LinkedHashSet<Extraction> temp =
            matchingAll ? null : new LinkedHashSet<Extraction>();
        Iterator<Extraction> tempIter;
        Extraction bestExtraction = null;
        boolean gotMatch = false;

        while (acceptedIter.hasNext()) {
            currentClassName = (String) acceptedIter.next();

            // iterate all predictions of this class
            extractionIter = predicted.iterator(currentClassName);
            while (extractionIter.hasNext()) {
                currentExtraction = (Extraction) extractionIter.next();

                if (matchingAll) {
                    // match-all mode: initially consider all spurious
                    spurious.add(currentExtraction);
                } else {
                    // match-best mode: find best (most likely) extraction
                    if ((bestExtraction == null) || predComp.compare(
                                    bestExtraction, currentExtraction) < 0) {
                        // most probable extraction seen to far
                        bestExtraction = currentExtraction;
                    }
                    temp.add(currentExtraction);
                }
            }

            if (!matchingAll) {
                // match-best mode: consider best as spurious (initially)
                if (bestExtraction != null) {
                    spurious.add(bestExtraction);
                    temp.remove(bestExtraction);
                }

                // ignore all others + reset variables
                ignored.addAll(temp);
                bestExtraction = null;
                temp.clear();
            }

            // iterate all expected extractions (answer keys)
            extractionIter = expected.iterator(currentClassName);
            while (extractionIter.hasNext()) {
                currentExtraction = (Extraction) extractionIter.next();

                if (spurious.contains(currentExtraction)) {
                    // got a match: move from spurious to correct
                    spurious.remove(currentExtraction);
                    correct.add(currentExtraction);
                    gotMatch = true;
                } else {
                    if (matchingAll) {
                        // match-all mode: extraction is missing
                        missing.add(currentExtraction);
                    } else {
                        // match-best: collect extractions to see if we got one
                        temp.add(currentExtraction);
                    }
                }
            }

            if (!matchingAll) { // match-best mode
                if (!gotMatch) {
                    // no match: mark first one as missing
                    tempIter = temp.iterator();
                    if (tempIter.hasNext()) {
                        currentExtraction = tempIter.next();
                        temp.remove(currentExtraction);
                        missing.add(currentExtraction);
                    }
                }

                // mark others as alternatives + reset variables
                altern.addAll(temp);
                gotMatch = false;
                temp.clear();
            }

            // add all extractions, setting their eval states + resetting sets
            addAllAndClear(correct, EvalStatus.CORRECT, source);
            addAllAndClear(missing, EvalStatus.MISSING, source);
            addAllAndClear(spurious, EvalStatus.SPURIOUS, source);

            if (!matchingAll) { // match-best mode
                addAllAndClear(altern, EvalStatus.ALTERNATIVE, source);
                addAllAndClear(ignored, EvalStatus.IGNORED, source);
            }
        }
    }

    /**
     * Returns the <a href="#match-modes">match mode</a>.
     *
     * @return <code>true</code> for <em>match-all mode</em>,
     * <code>false</code> for <em>match-best mode</em>
     */
    public boolean isMatchingAll() {
        return matchingAll;
    }

    /**
     * If <code>true</code>, the positions of extraction and answer keys must
     * match; otherwise only their contents must match (string compare).
     *
     * @return the value of the attribute
     */
    public boolean isMatchingPosition() {
        return matchingPosition;
    }

    /**
     * Returns a string representation of this object.
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("match-all mode", matchingAll)
            .append("matching positions", matchingPosition)
            .append("statistics", multiMetrics)
            .toString();
    }

    /**
     * Returns a read-only view of the counts of true positives (correct
     * extractions), false negatives (missing extractions) and false positives
     * (spurious extractions) stored in this container of each type as well as
     * for all types; and the metrics that can be calculated from these values.
     * This is not a snapshot but will change whenever the container is changed.
     *
     * @return a view of the counts and evaluation metrics
     */
    public MultiFMetricsView viewMetrics() {
        return multiMetrics;
    }

}
