/*
 * Copyright (C) 2003-2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

import de.fu_berlin.ties.classify.Prediction;
import de.fu_berlin.ties.classify.Probability;
import de.fu_berlin.ties.context.Recognition;
import de.fu_berlin.ties.eval.EvalStatus;
import de.fu_berlin.ties.io.FieldMap;
import de.fu_berlin.ties.text.TextUtils;
import de.fu_berlin.ties.text.TokenDetails;
import de.fu_berlin.ties.text.TokenizerFactory;
import de.fu_berlin.ties.util.Util;

/**
 * Extends a {@link de.fu_berlin.ties.classify.Prediction} by also storing the
 * extracted text and location data.
 *
 * <p>Instances of this class are not thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.12 $, $Date: 2004/11/25 13:36:08 $, $Author: siefkes $
 */
public class Extraction extends Prediction implements Recognition {

    /**
     * Serialization key for the extracted text.
     */
    public static final String KEY_TEXT = "Text";

    /**
     * Serialization key for the repetition of the first token.
     */
    public static final String KEY_FIRST_TOKEN_REP = "FirstTokenRep";

    /**
     * Serialization key for the index.
     */
    public static final String KEY_INDEX = "Index";

    /**
     * A list of {@link TokenDetails} describing the tokens combined in this
     * extraction.
     */
    private final LinkedList<TokenDetails> detailsList =
        new LinkedList<TokenDetails>();

    /**
     * The visible characters of the text fragment (everything except whitespace
     * and control characters).
     */
    private final StringBuilder visibleChars;

    /**
     * Matches characters that are not visible (whitespace and control
     * characters). Used to delete them for updating {@link #visibleChars}.
     */
    private final Matcher nonVisibleMatcher =
        Pattern.compile(TokenizerFactory.WHITESPACE_CONTROL_OTHER).matcher("");

    /**
     * A sealed extraction cannot be changed. This means that
     * {@linkplain #addToken(TokenDetails, Probability, boolean) adding tokens}
     * is not allowed after sealing.
     */
    private boolean sealed = false;

    /**
     * Whether the {@linkplain TokenDetails#getRep() repetition of the first
     * token} should be ignored, comparing only the text but not the position of
     * extractions. Defaults to <code>false</code>.
     */
    private boolean firstTokenRepIgnored = false;

    /**
     * Creates a new instance from a field map, fulfilling the
     * {@link de.fu_berlin.ties.io.Storable} contract. An extraction created
     * this way will be immediately {@link #isSealed() sealed}, thus the
     * extracted text cannot be changed.
     *
     * @param fieldMap map containing the serialized fields
     */
    public Extraction(final FieldMap fieldMap) {
        // delegate to super constructor
        super(fieldMap);

        // read values from map
        final Object rawText = fieldMap.get(KEY_TEXT);
        final String newText = (rawText == null) ? "" : rawText.toString();
        final int firstTokenRep = Util.asInt(fieldMap.get(KEY_FIRST_TOKEN_REP));
        final int index = Util.asInt(fieldMap.get(KEY_INDEX));

        // initialize own fields
        detailsList.add(new TokenDetails(newText, firstTokenRep, index, false));
        visibleChars = new StringBuilder(TextUtils.replaceAll(newText,
            nonVisibleMatcher, ""));

        // seal extraction
        setSealed(true);
    }

    /**
     * Creates a new instance without locating it in a text (using -1 for 
     * first token rep + index), setting the probability to -1 ("confirmed")
     * and the evaluation status to {@link EvalStatus#TRUTH}.
     *
     * @param predicted the predicted class
     * @param extracted the (first part) extracted text fragment; must not be
     * <code>null</code>
     */
    public Extraction(final String predicted, final String extracted) {
        this(predicted, new TokenDetails(extracted, -1, -1, false));
    }

    /**
     * Creates a new instance, setting the probability to -1 ("confirmed")
     * and the evaluation status to {@link EvalStatus#TRUTH}.
     * Use this constructor to build answer keys.
     *
     * @param predicted the predicted class
     * @param details details about the extracted text fragment or its first
     * token
     */
    public Extraction(final String predicted, final TokenDetails details) {
        this(predicted, details, new Probability(-1.0), EvalStatus.TRUTH);
    }

    /**
     * Creates a new instance, setting the evaluation status to
     * {@link EvalStatus#UNKNOWN}.
     *
     * @param predicted the predicted class
     * @param details details about the extracted text fragment or its first
     * token
     * @param prob the probability of the prediction
     */
    public Extraction(final String predicted, final TokenDetails details,
            final Probability prob) {
        this(predicted, details, prob, EvalStatus.UNKNOWN);
    }

    /**
     * Creates a new instance.
     *
     * @param predicted the predicted class
     * @param details details about the extracted text fragment or its first
     * token
     * @param prob the probability of the prediction
     * @param status the {@linkplain EvalStatus evaluation status} of this
     * instance
     */
    public Extraction(final String predicted, final TokenDetails details,
            final Probability prob, final EvalStatus status) {
        super(predicted, prob, status);
        detailsList.add(details);
        visibleChars = new StringBuilder(TextUtils.replaceAll(
                details.getToken(), nonVisibleMatcher, ""));
    }

    /**
     * Adds a token to this extraction, delegating to
     * {@link #addToken(TokenDetails, Probability, boolean)} with a probability
     * of -1 ("confirmed"). Use this method when building answer keys.
     *
     * @param details details about the new token
     * @param atEnd whether to add the new token at the end or at the
     * start
     * @throws IllegalStateException if this extraction
     * {@link #isSealed() is sealed}
     */
    public void addToken(final TokenDetails details, final boolean atEnd)
            throws IllegalStateException {
        addToken(details, new Probability(-1.0), atEnd);
    }

    /**
     * Adds a token to this extraction, recalculating the probability by
     * multiplying the prior probability value with the probability of the
     * new text. Increments the token count by 1.
     *
     * @param details details about the new token
     * @param prob the probability of the new token; might be <code>null</code>
     * if the overall probability of the extraction should not be changed
     * @param atEnd whether to add the new token at the end or at the
     * start
     * @throws IllegalStateException if this extraction
     * {@link #isSealed() is sealed}; or if new and old probabilities/pRs
     * cannot be combined
     */
    public void addToken(final TokenDetails details, final Probability prob,
            final boolean atEnd) throws IllegalStateException {
        // check state
        if (isSealed()) {
            throw new IllegalStateException(
                "Cannot change text of sealed extraction");
        }

        // update superclass + fields
        super.addProb(prob, atEnd);
        final String newVisibleChars =
            TextUtils.replaceAll(details.getToken(), nonVisibleMatcher, "");

        if (atEnd) {
            detailsList.addLast(details);
            visibleChars.append(newVisibleChars);
        } else {
            detailsList.addFirst(details);
            visibleChars.insert(0, newVisibleChars);
        }
    }

    /**
     * Indicates whether some other object is "equal to" this one, fulfulling
     * the {@link Object#equals(java.lang.Object)} contract. The
     * {@linkplain Prediction#getEvalStatus() evaluation status} is ignored when
     * checking equality, thus if all other fields of two extractions are equal,
     * this method will return <code>true</code> even if their evaluation states
     * differ. Only the {@link #getVisibleChars() visible characters} of the
     * extractions are compared, whitespace and control characters are ignored.
     *
     * @param obj the reference object with which to compare
     * @return <code>true</code> iff the specified object is an
     * {@link Extraction} equal to this instance
     */
    public boolean equals(final Object obj) {
        if (obj == this) {
            return true;
        } else if ((obj != null) && (getClass().equals(obj.getClass()))) {
            // used getClass instead of instanceof because otherwise subclasses
            // with additional fields would break the contract
            final Extraction other = (Extraction) obj;
            return new EqualsBuilder()
                .appendSuper(super.equals(obj))
                .append(getVisibleChars(), other.getVisibleChars())
                .append(getFirstTokenRep(), other.getFirstTokenRep())
                .isEquals();
        } else {
            return false;
        }
    }

    /**
     * Returns the repetition of the first token of the extraction in the
     * original text (counting starts with 0, as the first occurrence is the
     * "0th repetition"), -1 if unknown or if {@link #isFirstTokenRepIgnored()}
     * is <code>true</code>. This is useful to locate this extraction in the
     * original text.
     *
     * @return the value of the attribute
     */
    public int getFirstTokenRep() {
        if (firstTokenRepIgnored) {
            return -1;
        } else {
            return detailsList.getFirst().getRep();
        }
    }

    /**
     * Returns the index of the first token in the text (indexing starts with
     * 0); or -1 if unknown/irrelevant.
     *
     * @return the value of the attribute
     */
    public int getIndex() {
        return detailsList.getFirst().getIndex();
    }

    /**
     * Returns the extracted text fragment.
     * @return the extracted text
     */
    public String getText() {
        final StringBuilder text = new StringBuilder();
        final Iterator<TokenDetails> detailsIter = detailsList.iterator();
        TokenDetails details;

        while (detailsIter.hasNext()) {
            details = detailsIter.next();
            if (details.isWhitespaceBefore() && (text.length() > 0)) {
                text.append(' ');
            }
            text.append(details.getToken());
        }
        return text.toString();
    }

    /**
     * Returns the visible characters of the text fragment (everything except
     * whitespace and control characters).
     *
     * @return the visible characters
     */
    public String getVisibleChars() {
        return visibleChars.toString();
    }

    /**
     * Returns a hash code value for this object, fulfulling the
     * {@link Object#hashCode()} contract.
     * @return a hash code value for this object
     */
    public int hashCode() {
        // you pick two hard-coded, randomly chosen, non-zero, odd numbers
        // (preferably primes); ideally different for each class
        return new HashCodeBuilder(7, 11)
            .appendSuper(super.hashCode())
            .append(getVisibleChars())
            .append(getFirstTokenRep())
            .toHashCode();
    }

    /**
     * Whether the {@linkplain TokenDetails#getRep() repetition of the first
     * token} should be ignored, comparing only the text but not the position of
     * extractions. Defaults to <code>false</code>.
     *
     * @return the value of the attribute
     */
    public boolean isFirstTokenRepIgnored() {
        return firstTokenRepIgnored;
    }

    /**
     * Whether this extraction has been sealed. The text of a sealed extraction
     * cannot be changed. This means that
     * {@linkplain #addToken(TokenDetails, Probability, boolean) adding tokens}
     * is not allowed after sealing.
     *
     * @return <code>true</code> iff this extraction is sealed
     */
    public boolean isSealed() {
        return sealed;
    }

    /**
     * Deletes one of the tokens from this prediction. At least one
     * token must always remain, i.e. {@link #tokenCount()} must be 2 or
     * more prior to calling this method.
     *
     * @param atEnd whether to delete the first or the last token
     * @return details describing the removed token
     * @throws IllegalStateException if there is only one token left or
     * if this extraction {@link #isSealed() is sealed}
     */
    public TokenDetails removeToken(final boolean atEnd)
    throws IllegalStateException {
        // check state
        if (isSealed()) {
            throw new IllegalStateException(
                "Cannot change text of sealed extraction");
        }
        // check invariant: there must at many probabilities as there are tokens
        if (tokenCount() != probCount()) {
            throw new IllegalStateException(
                    "Invariant violation: number of tokens " + tokenCount() +
                    " != number of probabilities " + probCount());
        }

        // remove probabilities from superclass
        super.removeProb(atEnd);

        // remove token details
        final TokenDetails removed;
        if (atEnd) {
            removed = detailsList.removeLast();
        } else {
            removed = detailsList.removeFirst();
        }

        // prepare to adapt visibleText
        final String charsToDelete =
            TextUtils.replaceAll(removed.getToken(), nonVisibleMatcher, "");
        final int startIndex, endIndex;

        if (atEnd) {
            // delete last n visible chars
            endIndex = visibleChars.length();
            startIndex = endIndex - charsToDelete.length();
        } else {
            // delete first n visible chars
            endIndex = charsToDelete.length();
            startIndex = 0;
        }

        // delete chars after ensuring that everything is as it should
        if (visibleChars.substring(startIndex, endIndex).equals(
                charsToDelete)) {
            visibleChars.delete(startIndex, endIndex);
        } else {
            throw new RuntimeException("Invariant violation: "
                    + (atEnd ? "last" : "first")
                    + " visible chararacters are "
                    + visibleChars.substring(startIndex, endIndex)
                    + " instead of " + charsToDelete);
        }

        return removed;
    }


    /**
     * Specified whether the {@linkplain TokenDetails#getRep() repetition of
     * the first token} should be ignored, comparing only the text but not the
     * position of extractions.
     *
     * @param firstTokenRepIgnored the new value of the attribute
     */
    public void setFirstTokenRepIgnored(boolean firstTokenRepIgnored) {
        this.firstTokenRepIgnored = firstTokenRepIgnored;
    }

    /**
     * Seals or unseals this extraction. The text of a sealed extraction cannot
     * be changed. This means that
     * {@linkplain #addToken(TokenDetails, Probability, boolean) adding tokens}
     * is not allowed after sealing.
     *
     * @param newSealed the new value of the attribute
     */
    public void setSealed(final boolean newSealed) {
        sealed = newSealed;
    }

    /**
     * Stores all relevant fields of this object in a field map for
     * serialization. An equivalent object can be created by calling
     * {@link de.fu_berlin.ties.io.FieldMap#createObject(Class)} on the created
     * field map.
     *
     * @return the created field map
     */
    public FieldMap storeFields() {
        // delegate to super class and add own fields
        final FieldMap result = super.storeFields();
        result.put(KEY_TEXT, getText());
        final TokenDetails firstDetails = detailsList.getFirst();

        // ignore if ignored or unknown (-1)
        if (!firstTokenRepIgnored && (firstDetails.getRep() >= 0)) {
            result.put(KEY_FIRST_TOKEN_REP, new Integer(firstDetails.getRep()));
        }
        if (firstDetails.getIndex() >= 0) {
            result.put(KEY_INDEX, new Integer(firstDetails.getIndex()));
        }

        return result;
    }

    /**
     * Returns the number of tokens in this extraction. This will only be
     * reliable if a constructor is used to give the first token and operations
     * such as {@link Extraction#addToken(TokenDetails, Probability, boolean)
     * addToken} are used for each further token. Omitted when serializing
     * so it cannot be restored.
     *
     * @return the value of the attribute
     */
    public int tokenCount() {
        return detailsList.size();
    }

}
