/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.collections.IteratorUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.classify.PredictionComparator;
import de.fu_berlin.ties.io.FieldContainer;
import de.fu_berlin.ties.io.FieldMap;
import de.fu_berlin.ties.io.RestorableContainer;
import de.fu_berlin.ties.util.CollectionUtils;
import de.fu_berlin.ties.util.MultiValueMap;
import de.fu_berlin.ties.util.Util;

/**
 * A container of {@link de.fu_berlin.ties.extract.Extraction}s of different
 * classes. Instances of this class are not thread-safe.
 *
 * <p>Extraction containers can be serialized
 * ({@link #storeEntries(FieldContainer)}), but not deserialized.
 *
 * <p>Instances of this class are not thread-safe and have to be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.17 $, $Date: 2004/12/06 17:58:36 $, $Author: siefkes $
 */
public class ExtractionContainer implements RestorableContainer {

    /**
     * A list of all extractions stored in this container.
     */
    private final List<Extraction> allExtractions = new ArrayList<Extraction>();

    /**
     * The target structure specifying the classes to recognize.
     */
    private final TargetStructure targetStructure;

    /**
     * A multi-map storing lists of extractions for each extraction class name.
     */
    private final MultiValueMap<String, Extraction> typedExtractions =
        new MultiValueMap<String, Extraction>();

    /**
     * Creates a new empty instance.
     *
     * @param targetStruct the target structure specifying the classes to
     * recognize
     */
    public ExtractionContainer(final TargetStructure targetStruct) {
        super();
        targetStructure = targetStruct;
    }

    /**
     * Creates a new instance from a field container, delegating to
     * {@link #restoreEntries(FieldContainer)}.
     *
     * @param targetStruct the target structure specifying the classes to
     * recognize
     * @param fContainer the field container to read
     * @throws IllegalArgumentException if <code>fContainer</code> contains
     * extractions of an unsupported
     * (@link de.fu_berlin.ties.classify.Prediction#getType())
     */
    public ExtractionContainer(final TargetStructure targetStruct,
            final FieldContainer fContainer) throws IllegalArgumentException {
        this(targetStruct);
        restoreEntries(fContainer);
    }

    /**
     * Adds an extraction to this container.
     *
     * @param extraction the extraction to add
     * @throws IllegalArgumentException if the type of the extraction
     * (@link de.fu_berlin.ties.classify.Prediction#getType()) is not
     * in the set of class names defined by the target structure
     */
    public void add(final Extraction extraction)
            throws IllegalArgumentException {
        // check class
        if (!targetStructure.getClassNames().contains(extraction.getType())) {
            throw new IllegalArgumentException(
                "Container doesn't allow extractions of type "
                + extraction.getType());
        }

        // store extraction
        allExtractions.add(extraction);
        typedExtractions.put(extraction.getType(), extraction);
    }

    /**
     * Returns an iterator over all extractions in insertion order. The
     * iterator is immutable and cannot be used to modify the underlying
     * collection.
     *
     * @return an iterator over all extractions
     */
    public Iterator iterator() {
        return IteratorUtils.unmodifiableIterator(allExtractions.iterator());
    }

    /**
     * Returns an iterator over the extractions of a specified class, in
     * insertion order. The iterator is immutable and cannot be used to modify
     * the underlying collection.
     *
     * @param className the name of the class of extractions to iterate
     * @return an iterator over the extractions of the given class (if any,
     * otherwise an empty iterator will be returned)
     */
    public Iterator<Extraction> iterator(final String className) {
        Collection<Extraction> typedList = typedExtractions.get(className);
        if (typedList == null) {
            // create empty list
            typedList = new ArrayList<Extraction>(0);
        }

        // return iterator on immutable view
        return Collections.unmodifiableCollection(typedList).iterator();
    }

    /**
     * Returns the target structure specifying the classes to recognize.
     * @return the used target structure
     */
    public TargetStructure getTargetStructure() {
        return targetStructure;
    }

    /**
     * Returns the last extraction added to this container.
     *
     * @return the last extraction added to this container;
     * or <code>null</code> if the container is empty
     */
    public Extraction last() {
        if (allExtractions.isEmpty()) {
            return null;
        } else {
            return allExtractions.get(allExtractions.size() - 1);
        }
    }

    /**
     * Returns a list of the last <em>n</em> extractions added to this
     * container. Modifications of the returned list will not affect this
     * container and vice versa.
     *
     * @param number the number of extractions to copy
     * @return a list containing the last {@link Extraction}s
     */
    public List lastN(final int number) {
        return CollectionUtils.lastN(allExtractions, number);
    }

    /**
     * Returns a list of the last <em>n</em> extractions of a specified class
     * added to this container. Modifications of the returned list will not
     * affect this container and vice versa.
     *
     * @param extractionClass the class of extractions
     * @param number the number of extractions to copy
     * @return a list containing the last {@link Extraction}s of the given
     * class; will be empty if there are no extractions of this class
     */
    public List<Extraction> lastN(final String extractionClass,
            final int number) {
        final List<Extraction> typedList = (List<Extraction>)
            typedExtractions.get(extractionClass);
        return CollectionUtils.lastN(typedList, number);
    }

    /**
     * Removes an extraction from this container, if it is present.
     *
     * @param ext the extraction to remove
     * @return <code>true</code> if the container contained the specified
     * extraction
     */
    public boolean remove(final Extraction ext) {
        // first remove from typed extraction (shorter list)
        final boolean removedFromTyped =
            typedExtractions.remove(ext.getType(), ext) != null;

        if (removedFromTyped) {
            // also remove from all extractions
            final boolean removedFromAll = allExtractions.remove(ext);

            if (!removedFromAll) {
                // this should never happen
                throw new RuntimeException("Implementation error: Removed "
                    + ext + " from typed extractions (" + typedExtractions
                    + ") but not from all extractions (" + allExtractions
                    + ")");
            }
        }

        return removedFromTyped;
    }

    /**
     * Removes the last extraction {@link #add(Extraction) added} to this
     * container.
     *
     * @return the removed extraction
     * @throws IllegalStateException if this method is called while the
     * container is empty
     */
    public Extraction removeLast() throws IllegalStateException {
        if (allExtractions.isEmpty()) {
            throw new IllegalStateException(
                "removeLast called on empty ExtractionContainer");
        } else {
            Extraction removedExt =
                allExtractions.remove(allExtractions.size() - 1);
            final Collection<Extraction> typeCollection =
                typedExtractions.get(removedExt.getType());
            final boolean removedFromTyped = typeCollection.remove(removedExt);

            if (!removedFromTyped) {
                // this should never happen
                throw new RuntimeException("Implementation error: Removed "
                    + removedExt + " from all extractions (" + allExtractions
                    + ") but not from typed extractions (" + typedExtractions
                    + ")");
            }

            return removedExt;
        }
    }

    /**
     * Restores extractions stored in a field container and adds them to this
     * instance. The provided field container must have been filled by calling
     * {@link #storeEntries(FieldContainer)} on an extraction container with
     * identical {@linkplain #getTargetStructure() target structure}, or
     * errors are likely.
     *
     * @param fContainer the field container to read
     * @throws IllegalArgumentException if <code>fContainer</code> contains
     * extractions of an unsupported
     * (@link de.fu_berlin.ties.classify.Prediction#getType())
     */
    public void restoreEntries(final FieldContainer fContainer)
            throws IllegalArgumentException {
        final Iterator entryIter = fContainer.entryIterator();
        FieldMap currentFM;
        Extraction currentExt;
        String currentType;
        final SortedSet<String> ignoredTypes = new TreeSet<String>();

        // deserialize and add extractions
        while (entryIter.hasNext()) {
            currentFM = (FieldMap) entryIter.next();
            currentExt = new Extraction(currentFM);
            currentType = currentExt.getType();

            if (targetStructure.getClassNames().contains(currentType)) {
                add(currentExt);
            } else if (!ignoredTypes.contains(currentType)) {
                // we ignore extractions of other types so it possible to work
                // with a subset of all stored extractions
                ignoredTypes.add(currentType);
            }
        }

        // log debug message if we encountered unknown/ignored types
        if (!ignoredTypes.isEmpty()) {
            Util.LOG.debug("Ignoring extraction(s) of unknown types: "
                + ignoredTypes);
        }
    }

    /**
     * Adds all extractions stored in this instance to a field container for
     * serialization. Extractions are serialized in insertion order, as
     * required for {@linkplain Trainer training}.
     *
     * @param fContainer the field container to fill
     */
    public void storeEntries(final FieldContainer fContainer) {
        // serialize extraction in insertion order (required for training)
        final Iterator extIter = allExtractions.iterator();
        Extraction currentExt;

        while (extIter.hasNext()) {
            currentExt = (Extraction) extIter.next();
            fContainer.add(currentExt);
        }
    }

    /**
     * Returns a string representation of this object.
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("target structure", targetStructure)
            .append("typed extractions", typedExtractions)
            .toString();
    }

    /**
     * Unsets the positions of all stored extractions. In case of duplicates
     * (extractions that only differed in their position), the most probably
     * one is kept.
     */
    public void unsetPositions() {
        final Iterator<String> classIter = typedExtractions.keySet().iterator();
        List<Extraction> extList;
        Iterator<Extraction> extIter;
        String className;
        Extraction ext;
        Extraction otherExt;
        String text;
        Map<String, Extraction> bestExtractions =
            new HashMap<String, Extraction>();
        List<Extraction> duplicates = new ArrayList<Extraction>(); // each class
        List<Extraction> allDuplicates =
            new ArrayList<Extraction>(); // all classes
        PredictionComparator predComp = new PredictionComparator();

        // iterate extractions of each class to detect duplicates
        while (classIter.hasNext()) {
            // reset collections for new class
            bestExtractions.clear();
            duplicates.clear();
            className = classIter.next();
            extList = (List<Extraction>) typedExtractions.get(className);
            extIter = extList.iterator();

            while (extIter.hasNext()) {
                ext = extIter.next();
                text = ext.getText();

                if (bestExtractions.containsKey(text)) {
                    otherExt = bestExtractions.get(text);

                    // found duplicate -- keep more likely one
                    if (predComp.compare(otherExt, ext) < 0) {
                        // new one is more likely
                        bestExtractions.put(text, ext);
                        duplicates.add(otherExt);

                        if (ext.getProbability().getProb() >= 0.0) {
                            Util.LOG.debug("Discarded duplicate " + otherExt
                                    + " in favor of " + ext);
                        }
                    } else {
                        // old one is at least as likely
                        duplicates.add(ext);

                        if (ext.getProbability().getProb() >= 0.0) {
                            Util.LOG.debug("Discarded duplicate " + ext
                                    + " in favor of " + otherExt);
                        }
                    }
                } else {
                    // first extraction with this type and text
                    bestExtractions.put(text, ext);
                }
            } // while extIter

            // remove duplicates from typedExtractions
            extList.removeAll(duplicates);
            allDuplicates.addAll(duplicates);
        } // while classIter

        // remove duplicated from allExtractions
        allExtractions.removeAll(allDuplicates);

        // set first token rep. of all remaining extractions to -1
        final Iterator<Extraction> allIter = allExtractions.iterator();
        while (allIter.hasNext()) {
            ext = allIter.next();
            ext.setFirstTokenRepIgnored(true);
        }
    }

}
