/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract;

import java.util.Iterator;

import org.apache.commons.collections.Bag;
import org.apache.commons.collections.bag.HashBag;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Document;

import de.fu_berlin.ties.text.TextTokenizer;
import de.fu_berlin.ties.util.Util;

/**
 * Locates extractions in a document.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.4 $, $Date: 2004/08/30 17:24:43 $, $Author: siefkes $
 */
public class ExtractionLocator {

    /**
     * An iterator over the extractions to train.
     */
    private Iterator extractionIter;

    /**
     * The extraction that is currently processed or will be processed next;
     * or <code>null</code> if there are no more extractions to process.
     */
    private Extraction currentExtraction;

    /**
     * The first token from the current extaction, utilized together with
     * {@link Extraction#getFirstTokenRep()} to locate the start of the
     * extraction.
     */
    private String firstToken;

    /**
     * <code>true</code> if we are processing the {@link #currentExtraction},
     * <code>false</code> otherwise (we are waiting for it to start or there
     * are no more extractions).
     */
    private boolean inExtraction;

    /**
     * A multiset of the tokens remaing (not yet trained) in the current
     * extraction.
     */
    private Bag remainingTokens = new HashBag();

    /**
     * The tokenizer used to split extractions into tokens.
     */
    private final TextTokenizer tokenizer;


    /**
     * Creates a new instance.
     *
     * @param document the document to use
     * @param extractions the extractions in this document
     * @param textTokenizer the tokenizer used to split extractions into tokens
     */
    public ExtractionLocator(final Document document,
                             final ExtractionContainer extractions,
                             final TextTokenizer textTokenizer) {
        super();
        tokenizer = textTokenizer;

        // initialize state
        extractionIter = extractions.iterator();
        switchToNextExtraction();
    }

    /**
     * Whether we reached the end of the current extraction.
     * @return <code>true</code> iff the current extraction has ended
     */
    public boolean endOfExtraction() {
        return inExtraction && remainingTokens.isEmpty();
    }

    /**
     * Returns the current extraction.
     * @return the current extraction
     */
    public Extraction getCurrentExtraction() {
        return currentExtraction;
    }

    /**
     * Whether we are currently within an extraction.
     *
     * @return <code>true</code> iff are processing the
     * {@link #getCurrentExtraction()}, <code>false</code> otherwise (we are
     * waiting for it to start or there are no more extractions)
     */
    public boolean inExtraction() {
        return inExtraction;
    }

    /**
     * This method must be called at the end of the current document. It will
     * log an error when there are still unprocessed or incompletely processed
     * extractions.
     */
    public void reachedEndOfDocument() {
        if (!remainingTokens.isEmpty()) {
            if (inExtraction) {
                Util.LOG.error("Document ended while processing extraction: "
                    + currentExtraction + ", unprocessed tokens: "
                    + remainingTokens);
            } else {
                Util.LOG.error("Document ended while waiting for start of "
                    + "extraction: " + currentExtraction);
            }

            remainingTokens.clear();
        }
        if (extractionIter.hasNext()) {
            Util.LOG.error("Unprocessed extractions at end of document: "
                + extractionIter.next());
        }
    }

    /**
     * Whether the current token starts a new extraction. <em>This method must
     * be called once for each token in a document, otherwise we might miss
     * extractions.</em>
     *
     * @param token the token to check
     * @param tokenRep the repetition of the <code>token</code> in the document
     * (counting starts with 0, as the first occurrence is the "0th
     * repetition").
     * @return <code>true</code> iff the given token starts a new extraction
     */
    public boolean startOfExtraction(final String token, final int tokenRep) {
        if ((!inExtraction) && (currentExtraction != null)
                && (token.equals(firstToken))
                && (currentExtraction.getFirstTokenRep() <= tokenRep)) {
            // found first element of extraction
            inExtraction = true;
            return true;
        } else {
            return false;
        }
    }

    /**
     * Switches to the next extraction, updating the current
     * extraction and related fields. The prior current extraction must have
     * been fully processed when this method is called, i.e.
     * {@link #endOfExtraction()} must be <code>true</code>.
     *
     * @throws IllegalStateException if {@link #endOfExtraction()} is not
     * <code>true</code> (there are still remaining tokens to process
     */
    public void switchToNextExtraction() throws IllegalStateException {
        // check state
        if (!remainingTokens.isEmpty()) {
            throw new IllegalStateException("Cannot update current extraction "
                    + "while there are remaining tokens: " + remainingTokens);
        }

        inExtraction = false;
        firstToken = null;

        if (extractionIter.hasNext()) {
            currentExtraction = (Extraction) extractionIter.next();
            tokenizer.reset(currentExtraction.getText());
            String token;

            // update first token + remaining tokens
            while ((token = tokenizer.nextToken()) != null) {
                if (firstToken == null) {
                    firstToken = token;
                }
                remainingTokens.add(token);
            }
        } else {
            // no more extractions
            currentExtraction = null;
        }
    }

    /**
     * Returns a string representation of this object.
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("tokenizer", tokenizer)
            .append("in extraction", inExtraction)
            .append("current extraction", currentExtraction)
            .append("first token", firstToken)
            .append("remaining tokens", remainingTokens)
            .toString();
    }

    /**
     * Updates the currently processed extraction. This method must be called
     * once for each token in each extraction.
     *
     * @param token the token to process
     * @param tokenRep the repetition of the <code>token</code> in the document
     * (counting starts with 0, as the first occurrence is the "0th
     * repetition").
     * @return <code>true</code> iff the extraction was successfully updated;
     * <code>false</code> if the token was erroneous (not expected to occur
     * within the current extraction)
     */
    public boolean updateExtraction(final String token, final int tokenRep) {
        boolean removedToken = remainingTokens.remove(token, 1);

        if (!removedToken) {
            // token didn't match thru there are still remaing tokens:
            // this means document and extraction container are out of sync
            Util.LOG.error("Extractions don't match document: "
                    + "still missing tokens " + remainingTokens
                    + " from extraction " + currentExtraction
                    + " but current token '" + token
                    + "' (token rep=" + tokenRep + ") doesn't match");

            // stop handling of current extraction, treat as other/outside
            remainingTokens.clear();
        }

        return removedToken;
    }

}
