/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract;

import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Document;
import org.dom4j.Element;

import de.fu_berlin.ties.combi.CombinationState;
import de.fu_berlin.ties.combi.CombinationStrategy;
import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;

import de.fu_berlin.ties.classify.Classifier;
import de.fu_berlin.ties.classify.Prediction;
import de.fu_berlin.ties.classify.PredictionDistribution;
import de.fu_berlin.ties.classify.Probability;
import de.fu_berlin.ties.classify.Reranker;
import de.fu_berlin.ties.context.Recognition;
import de.fu_berlin.ties.context.Representation;
import de.fu_berlin.ties.eval.FMetricsView;
import de.fu_berlin.ties.filter.EmbeddingElements;
import de.fu_berlin.ties.filter.FilteringTokenWalker;
import de.fu_berlin.ties.filter.TrainableFilter;
import de.fu_berlin.ties.io.FieldContainer;
import de.fu_berlin.ties.text.TokenDetails;
import de.fu_berlin.ties.text.TokenizerFactory;
import de.fu_berlin.ties.util.MathUtils;
import de.fu_berlin.ties.util.Util;

/**
 * An extractor runs a local {@link de.fu_berlin.ties.classify.Classifier}
 * on a list of items/nodes and combines their results using a
 * {@link de.fu_berlin.ties.combi.CombinationStrategy}.
 *
 * <p>Instances of this class are not thread-safe and cannot extract from
 * several documents in parallel.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.41 $, $Date: 2004/12/06 09:21:06 $, $Author: siefkes $
 */
public class Extractor extends ExtractorBase {

    /**
     * The recommended file extension to use for storing extractions.
     */
    public static final String EXT_EXTRACTIONS = "ext";


    /**
     * Helper method that creates a reranker. The reranker is configured from
     * the "extract" prefix. E.g. an entry such as "extract.bias.A = 0.99"
     * could be used to introduce a small malus for the background class used
     * in most combination strategies, thus favoring recall over precision.
     *
     * @param config the configuration to use
     * @return the initialized reranker
     */
    private static Reranker createReranker(final TiesConfiguration config) {
        return new Reranker(config.subset("extract"));        
    }


    /**
     * The last document processed by this instance.
     */
    private Document lastDocument;

    /**
     * The extraction container used for storing the predicted extractions.
     */
    private ExtractionContainer predictedExtractions;

    /**
     * A list of punctuation tokens collected between non-puncutation tokens.
     */
    private final List<TokenDetails> punctuationDetails =
        new ArrayList<TokenDetails>();

    /**
     * An optional reranker that recalculates probabilities to introduce a bias.
     * This can be used to favor recall over precision (by setting a bias &lt;
     * 1 for the background class) etc.
     */
    private final Reranker reranker;


    /**
     * Creates a new instance using a default extension. Delegates to
     * {@link #Extractor(String, TiesConfiguration)} using the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @throws IllegalArgumentException if the combination strategy cannot be
     * initialized (cf. {@link CombinationStrategy#createStrategy(java.util.Set,
     * TiesConfiguration)})
     * @throws ProcessingException if an error occurs during initialization
     */
    public Extractor() throws IllegalArgumentException, ProcessingException {
        this(EXT_EXTRACTIONS);
    }

    /**
     * Creates a new instance. Delegates to
     * {@link #Extractor(String, TiesConfiguration)} using the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @param outExt the extension to use for output files
     * @throws IllegalArgumentException if the combination strategy cannot be
     * initialized (cf. {@link CombinationStrategy#createStrategy(java.util.Set,
     * TiesConfiguration)})
     * @throws ProcessingException if an error occurs during initialization
     */
    public Extractor(final String outExt)
            throws IllegalArgumentException, ProcessingException {
        this(outExt, TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance. Delegates to the corresponding
     * {@linkplain ExtractorBase#ExtractorBase(String, TiesConfiguration) super
     * constructor} to configure the fields.
     *
     * @param outExt the extension to use for output files
     * @param config the configuration to use
     * @throws IllegalArgumentException if the combination strategy cannot be
     * initialized
     * (cf. {@link CombinationStrategy#createStrategy(java.util.Set,
     * TiesConfiguration)})
     * @throws ProcessingException if an error occurs during initialization
     */
    public Extractor(final String outExt, final TiesConfiguration config)
            throws IllegalArgumentException, ProcessingException {
        super(outExt, config);
        reranker = createReranker(config);
    }

    /**
     * Creates a new instance. Delegates to the corresponding
     * {@linkplain ExtractorBase#ExtractorBase(String, File, TiesConfiguration)
     * super constructor} to configure the fields.
     *
     * @param outExt the extension to use for output files
     * @param runDirectory the directory to run the classifier in; used instead
     * of the
     * {@linkplain de.fu_berlin.ties.classify.ExternalClassifier#CONFIG_DIR
     * configured directory} if not <code>null</code>
     * @param config the configuration to use
     * @throws IllegalArgumentException if the combination strategy cannot be
     * initialized
     * (cf. {@link CombinationStrategy#createStrategy(java.util.Set,
     * TiesConfiguration)})
     * @throws ProcessingException if an error occurs during initialization
     */
    public Extractor(final String outExt, final File runDirectory,
            final TiesConfiguration config)
            throws IllegalArgumentException, ProcessingException {
        super(outExt, runDirectory, config);
        reranker = createReranker(config);
    }

    /**
     * Creates a new instance, re-using the components from the provided
     * trainer.
     *
     * @param outExt the extension to use for output files
     * @param trainer trainer whose components should be re-used
     */
    public Extractor(final String outExt, final Trainer trainer) {
        this(outExt, trainer.getTargetStructure(), trainer.getClassifiers(),
            trainer.getRepresentation(), trainer.getStrategy(),
            trainer.getFactory(), trainer.getSentenceFilter(),
            createReranker(trainer.getConfig()),
            trainer.viewRelevantPunctuation(), trainer.getConfig());
    }

    /**
     * Creates a new instance.
     *
     * @param outExt the extension to use for output files
     * @param targetStruct the target structure specifying the classes to
     * recognize
     * @param theClassifiers the classifiers to use for the local classification
     * decisions
     * @param theRepresentation the context representation to use for local
     * classifications
     * @param combiStrat the combination strategy to use
     * @param tFactory used to instantiate tokenizers
     * @param sentFilter the filter used in the first step of a double
     * classification approach ("sentence filtering"); if <code>null</code>,
     * no sentence filtering is used
     * @param rerank a reranker that recalculates probabilities to
     * introduce a bias (can be used to favor recall over precision, by setting
     * a bias &lt; 1 for the background class, etc.); must not be
     * <code>null</code>
     * @param relevantPunct a set of punctuation tokens that have been found to
     * be relevant for token classification; might be empty but not
     * <code>null</code>
     * @param config used to configure superclasses; if <code>null</code>,
     * the {@linkplain TiesConfiguration#CONF standard configuration} is used
     */
    public Extractor(final String outExt, final TargetStructure targetStruct,
            final Classifier[] theClassifiers,
            final Representation theRepresentation,
            final CombinationStrategy combiStrat,
            final TokenizerFactory tFactory, final TrainableFilter sentFilter,
            final Reranker rerank, final Set<String> relevantPunct,
            final TiesConfiguration config) {
        super(outExt, targetStruct, theClassifiers, theRepresentation,
            combiStrat, tFactory, sentFilter, relevantPunct, config);
        reranker = rerank;
    }


    /**
     * Adds an element to the collected punctuation details.
     *
     * @param details the element to add
     */
    protected void addPunctuationDetails(final TokenDetails details) {
        punctuationDetails.add(details);
    }

    /**
     * Appends the collected punctuation details (if any) to the provided
     * extraction. Finally delegates to {@link #clearPunctuation()}
     * to dleetes the processed punctuation.
     *
     * @param ext the extraction to append to
     */
    protected void appendPunctuation(final Extraction ext) {
        if (!punctuationDetails.isEmpty()) {
            final Iterator<TokenDetails> detailsIter =
                punctuationDetails.iterator();
            TokenDetails currentDetails;

            // append
            while (detailsIter.hasNext()) {
                currentDetails = detailsIter.next();
                ext.addToken(currentDetails, null, true);
            }

            // and clear
            clearPunctuation();            
        }
    }

    /**
     * Clears the collected punctuation details.
     */
    protected void clearPunctuation() {
        punctuationDetails.clear();
    }

    /**
     * {@inheritDoc}
     */
    protected FilteringTokenWalker createFilteringTokenWalker(
            final TrainableFilter repFilter) {
        return new FilteringTokenWalker(this, getFactory(), repFilter, this);
    }

    /**
     * Helper method that discards the last extraction, removing it from
     * {@link #getPredictedExtractions() prior recognitions} and
     * {@link #getPredictedExtractions() predicted extractions}.
     */
    private void discardLastExtraction() {
        final Extraction removedFromContainer =
            getPredictedExtractions().removeLast();

        // this one might be null if no prior recognitions are stored
        final Recognition removedFromRecognitions =
            getPriorRecognitions().removeLast();

        // both must be equal, otherwise something would be very wrong
        if ((removedFromRecognitions != null)
                && !removedFromContainer.equals(removedFromRecognitions)) {
            throw new IllegalStateException("Extractions discarded from "
                + "container " + removedFromContainer + " and from prior "
                + "recognitions " + removedFromRecognitions + " differ");
        }

        Util.LOG.debug("Discarded last extraction " + removedFromRecognitions);
    }

    /**
     * Extracts items of interest from the contents of an XML document, based on
     * context representation and local classifier.
     *
     * @param document a document whose contents should be classified
     * @return a container of all extractions from the document, in document
     * order
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public ExtractionContainer extract(final Document document)
            throws IOException, ProcessingException {
        // initialize local fields
        initFields();
        lastDocument = document;
        predictedExtractions = new ExtractionContainer(getTargetStructure());

        // the walker will call back (processToken method) where appropriate
        getWalker().walk(document, null);

        // reset strategy and discard last prediction extraction if requested
        resetStrategy();

        // return the container of predicted extractions
        return getPredictedExtractions();
    }

    /**
     * Evaluates precision and recall for {@linkplain #isSentenceFiltering()
     * sentence filtering} on the last processed document.
     *
     * @param correctExtractions a container of all correct extractions for the
     * document
     * @return the calculated statistics for sentence filtering on the
     * last document; <code>null</code> if {@linkplain
     * #isSentenceFiltering() sentence filtering} is disabled
     */
    public FMetricsView evaluateSentenceFiltering(
            final ExtractionContainer correctExtractions) {
        return evaluateSentenceFiltering(new EmbeddingElements(lastDocument,
                correctExtractions, getFactory()));
    }

    /**
     * Returns the extraction container used for storing the predicted
     * extractions.
     * @return the extraction container
     */
    protected ExtractionContainer getPredictedExtractions() {
        return predictedExtractions;
    }

    /**
     * Extracts items of interest from the contents of an XML document and
     * serializes the extractions.
     *
     * @param document the document to read
     * @param writer the writer to write the extracted items to; flushed
     * but not closed by this method
     * @param context a map of objects that are made available for processing
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public void process(final Document document, final Writer writer,
            final ContextMap context) throws IOException, ProcessingException {
        // delegate to extract
        final ExtractionContainer extractions = extract(document);

        // serialize results
        final FieldContainer storage = FieldContainer.createFieldContainer();
        extractions.storeEntries(storage);
        storage.store(writer);
    }

    /**
     * {@inheritDoc}
     */
    public void processToken(final Element element, final String left,
            final TokenDetails details, final String right,
            final ContextMap context) throws ProcessingException {
        if (isRelevant(details.getToken())) {
            // update context rep + active classes
            updateState(element, left, details.getToken(), right);
    
            final Classifier[] classifiers = getClassifiers();
            // call classifiers and translate result
            final PredictionDistribution[] origDists =
                new PredictionDistribution[classifiers.length];
            final PredictionDistribution[] finalDists =
                new PredictionDistribution[classifiers.length];
            final Prediction[] predictions =
                new Prediction[classifiers.length];
            final String[] predictedTypes = new String[classifiers.length];
            Probability currentProb;
            final double[] probs = new double[classifiers.length];
            final double[] pRs = new double[classifiers.length];
    
            for (int i = 0; i < origDists.length; i++) {
                origDists[i] = classifiers[i].classify(getFeatures(),
                    getActiveClasses()[i]);
                finalDists[i] = reranker.rerank(origDists[i]);
                predictions[i] = finalDists[i].best();
                predictedTypes[i] = predictions[i].getType();
                currentProb = predictions[i].getProbability();
                probs[i] = currentProb.getProb();
                pRs[i] = currentProb.getPR();
            }
    
            final CombinationState translatedState =
                getStrategy().translateResult(finalDists);
            final double meanProb = MathUtils.mean(probs);
            final double meanPR = MathUtils.mean(pRs);
            Util.LOG.debug("Predicted types: '"
                + ArrayUtils.toString(predictedTypes)
                + "'; translated state: " + translatedState + ", mean prob.: "
                + meanProb + ", mean pR: " + meanPR);
    
            if (translatedState.isDiscardPreceding()) {
                discardLastExtraction();
            }
    
            // update extraction container if the result isn't null
            if (translatedState.getType() == null) {
                // outside: seal last extraction if not yet done
                final Extraction lastExtraction =
                    getPredictedExtractions().last();

                if ((lastExtraction != null) && (!lastExtraction.isSealed())) {
                    lastExtraction.setSealed(true);
                }

                // discard any collected punctuation tokens
                clearPunctuation();
            } else if (translatedState.isBegin()) {
                // start of new instance: add to container & recognition buffer
                final Extraction newExtraction =
                    new Extraction(translatedState.getType(), details,
                            new Probability(meanProb, meanPR));
                getPredictedExtractions().add(newExtraction);
                getPriorRecognitions().add(newExtraction);

                // discard any collected punctuation tokens
                clearPunctuation();
            } else {
                // continuation of current instance
                final Extraction currentExtraction =
                    getPredictedExtractions().last();
    
                // append and discard any collected punctuation tokens
                appendPunctuation(currentExtraction);

                // ensure that states match prior to appending
                if (!currentExtraction.getType().equals(
                        translatedState.getType())) {
                    throw new IllegalStateException("Type mismatch: "
                        + translatedState + " cannot continue extraction "
                        + currentExtraction);
                }
                currentExtraction.addToken(details,
                        new Probability(meanProb, meanPR), true);
            }
    
            // update the state of the strategy
            getStrategy().updateState(translatedState);
        } else {
            // irrevalant (punctuation) token:
            final Extraction lastExtraction = getPredictedExtractions().last();

            // remember for later appending if in extraction
            if ((lastExtraction != null) && (!lastExtraction.isSealed())) {
                addPunctuationDetails(details);
                Util.LOG.debug("Keeping irrelevant punctuation token "
                        + details.getToken()
                        + " -- might become part of the current "
                        + lastExtraction.getType() + " extraction");
            } else {
                Util.LOG.debug("Skipping over irrelevant punctuation token "
                        + details.getToken());                
            }
        }
    }

    /**
     * Reset strategy and discard last prediction extraction if requested.
     */
    protected void resetStrategy() {        
        final boolean discardLast = getStrategy().reset();
        if (discardLast) {
            discardLastExtraction();
        }       
    }

    /**
     * Returns a string representation of this object.
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("reranker", reranker)
            .append("predicted extractions", predictedExtractions)
            .toString();
    }

}
