/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract;

import java.io.File;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Element;
import org.dom4j.NodeFilter;

import de.fu_berlin.ties.combi.CombinationStrategy;
import de.fu_berlin.ties.DocumentReader;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;

import de.fu_berlin.ties.classify.Classifier;
import de.fu_berlin.ties.classify.Reranker;
import de.fu_berlin.ties.classify.TrainableClassifier;
import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.context.DefaultRepresentation;
import de.fu_berlin.ties.context.LocalFeature;
import de.fu_berlin.ties.context.PriorRecognitions;
import de.fu_berlin.ties.context.Representation;
import de.fu_berlin.ties.eval.FMetricsView;
import de.fu_berlin.ties.eval.FeatureCount;
import de.fu_berlin.ties.eval.FeatureCountView;
import de.fu_berlin.ties.filter.EmbeddingElements;
import de.fu_berlin.ties.filter.FilterEvaluator;
import de.fu_berlin.ties.filter.FilteringTokenWalker;
import de.fu_berlin.ties.filter.RepresentationFilter;
import de.fu_berlin.ties.filter.SkipHandler;
import de.fu_berlin.ties.filter.TrainableFilter;
import de.fu_berlin.ties.text.TextUtils;
import de.fu_berlin.ties.text.TokenizerFactory;
import de.fu_berlin.ties.util.CollectionUtils;
import de.fu_berlin.ties.util.Util;
import de.fu_berlin.ties.xml.dom.ElementNameFilter;
import de.fu_berlin.ties.xml.dom.TokenProcessor;
import de.fu_berlin.ties.xml.dom.TokenWalker;

/**
 * Common code base shared by {@link de.fu_berlin.ties.extract.Extractor}and
 * {@link de.fu_berlin.ties.extract.Trainer}.
 * <p>
 * Instances of subclasses are not thread-safe and cannot process several
 * documents in parallel.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.39 $, $Date: 2004/12/07 12:01:48 $, $Author: siefkes $
 */
public abstract class ExtractorBase extends DocumentReader implements
        SkipHandler, TokenProcessor {

    /**
     * Configuration key: List of elements to filter.
     */
    public static final String CONFIG_ELEMENTS = "filter.elements";

    /**
     * Configuration key: List of elements that should be avoided when filtering
     * (using parent element instead).
     */
    public static final String CONFIG_AVOID = "filter.avoid";

    /**
     * Configuration key: list of punctuation and symbol tokens that are
     * considered as relevant from the very start.
     */
    public static final String CONFIG_RELEVANT_PUNCTUATION =
        "extract.punctuation.relevant";

    /**
     * Configuration suffix/prefix used for sentence filtering.
     */
    public static final String CONFIG_SENTENCE = "sent";

    /**
     * Helper methat that initializes the filter to be used for the first step
     * of a double classification approach ("sentence filtering").
     * 
     * @param conf the filter is initialized from the "filter" parameters in
     * this configuration
     * @param representation the representation to use
     * @return the created filter; or <code>null</code> if no sentence
     * filtering should be used
     * @throws ProcessingException if an error occurs while creating the filter
     */
    public static TrainableFilter createSentenceFilter(
            final TiesConfiguration conf, final Representation representation)
            throws ProcessingException {
        final RepresentationFilter result;

        if (conf.containsKey(CONFIG_ELEMENTS)) {
            final String[] filteredElements =
                conf.getStringArray(CONFIG_ELEMENTS);

            if (!TiesConfiguration.arrayIsEmpty(filteredElements)) {
                // elements to prefer when filtering
                final NodeFilter positiveFilter = new ElementNameFilter(
                        filteredElements);
                final NodeFilter negativeFilter = new ElementNameFilter(
                        conf.getStringArray(CONFIG_AVOID));

                // configure reranker from "sent" prefix
                // (e.g. sent.bias.true = 1.4)
                final Reranker reranker =
                    new Reranker(conf.subset(CONFIG_SENTENCE));
                result = new RepresentationFilter(conf, positiveFilter,
                        negativeFilter, reranker, representation, "Sentence");
                Util.LOG.debug("Initialized representation filter for sentence "
                        + "filtering: " + result);
            } else {
                result = null;
            }
        } else {
            result = null;
        }

        return result;
    }

    /**
     * The classifier(s) used for the local classification decisions.
     */
    private final Classifier[] classifiers;

    /**
     * The context representation used for local classifications.
     */
    private final Representation representation;

    /**
     * The target structure specifying the classes to recognize.
     */
    private final TargetStructure targetStructure;

    /**
     * Used to instantiate tokenizers.
     */
    private final TokenizerFactory factory;

    /**
     * The filter used in the first step of a double classification approach
     * ("sentence filtering"); if <code>null</code>, no sentence filtering is
     * used.
     */
    private final TrainableFilter sentenceFilter;

    /**
     * The set of candidate classes to consider for the current element for each
     * classifier.
     */
    private Set[] activeClasses;

    /**
     * The feature cache used by the context representation.
     */
    private Map<Element, List<LocalFeature>> featureCache;

    /**
     * The vector of features representing the currently processed element.
     */
    private FeatureVector features;

    /**
     * Used to count documents, contexts, and features and to calculate
     * averages.
     */
    private final FeatureCount featureCount = new FeatureCount();

    /**
     * A set of punctuation tokens that have been found to be relevant for
     * token classification (because they sometimes occur as the first or
     * last token of an extraction).
     */
    private final Set<String> relevantPunctuation;

    /**
     * A buffer of preceding {@link de.fu_berlin.ties.context.Recognition}s
     * from the current document.
     */
    private PriorRecognitions priorRecognitions;

    /**
     * The combination strategy used.
     */
    private final CombinationStrategy strategy;

    /**
     * Used to walk thru documents.
     */
    private TokenWalker walker;

    /**
     * Creates a new instance. Delegates to
     * {@link #ExtractorBase(String, TiesConfiguration)}using the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     * 
     * @param outExt the extension to use for output files
     * @throws IllegalArgumentException if the combination strategy cannot be
     * initialized (cf.
     * {@link CombinationStrategy#createStrategy(Set, TiesConfiguration)})
     * @throws ProcessingException if an error occurs during initialization
     */
    public ExtractorBase(final String outExt) throws IllegalArgumentException,
            ProcessingException {
        this(outExt, TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance, configuring target structure, classifier,
     * {@link DefaultRepresentation}, node filter and combination strategy from
     * the provided configuration.
     * 
     * @param outExt the extension to use for output files
     * @param config the configuration to use
     * @throws IllegalArgumentException if the combination strategy cannot be
     * initialized (cf.
     * {@link CombinationStrategy#createStrategy(Set, TiesConfiguration)})
     * @throws ProcessingException if an error occurs during initialization
     */
    public ExtractorBase(final String outExt, final TiesConfiguration config)
            throws IllegalArgumentException, ProcessingException {
        this(outExt, null, config);
    }

    /**
     * Creates a new instance, configuring target structure, classifier,
     * {@link DefaultRepresentation}, node filter, combination strategy and
     * tokenizer factory from the provided configuration.
     * 
     * @param outExt the extension to use for output files
     * @param runDirectory the directory to run the classifier in; used instead
     * of the
     * {@linkplain de.fu_berlin.ties.classify.ExternalClassifier#CONFIG_DIR
     * configured directory} if not <code>null</code>
     * @param config the configuration to use
     * @throws IllegalArgumentException if the combination strategy cannot be
     * initialized (cf.
     * {@link CombinationStrategy#createStrategy(Set, TiesConfiguration)})
     * @throws ProcessingException if an error occurs during initialization
     */
    public ExtractorBase(final String outExt, final File runDirectory,
            final TiesConfiguration config) throws IllegalArgumentException,
            ProcessingException {
        super(outExt, config);
        targetStructure = new TargetStructure(config);
        representation = new DefaultRepresentation(config);
        strategy = CombinationStrategy.createStrategy(
                targetStructure.getClassNames(), config);
        sentenceFilter = createSentenceFilter(config, representation);

        // initialize classifier(s) suitable fo the combi strategy
        final Set<String>[] allClasses = strategy.allClasses();
        classifiers = new Classifier[allClasses.length];

        for (int i = 0; i < allClasses.length; i++) {
            classifiers[i] = TrainableClassifier.createClassifier(
                    allClasses[i], runDirectory, config);
        }

        // start with configured set of relevant punctuation+symbol tokens
        relevantPunctuation = CollectionUtils.arrayAsSet(
                config.getStringArray(CONFIG_RELEVANT_PUNCTUATION));
        Util.LOG.debug("Initialized set of relevant punctuation + symbol "
                + "tokens to " + relevantPunctuation);
        factory = new TokenizerFactory(config);
    }

    /**
     * Creates a new instance.
     * 
     * @param outExt the extension to use for output files
     * @param targetStruct the target structure specifying the classes to
     * recognize
     * @param theClassifiers the array of classifiers to use for the local
     * classification decisions
     * @param theRepresentation the context representation to use for local
     * classifications
     * @param combiStrat the combination strategy to use
     * @param tFactory used to instantiate tokenizers
     * @param sentFilter the filter used in the first step of a double
     * classification approach ("sentence filtering"); if <code>null</code>,
     * no sentence filtering is used
     * @param relevantPunct a set of punctuation tokens that have been found to
     * be relevant for token classification; might be empty but not
     * <code>null</code>
     * @param config used to configure superclasses; if <code>null</code>,
     * the {@linkplain TiesConfiguration#CONF standard configuration}is used
     */
    public ExtractorBase(final String outExt,
            final TargetStructure targetStruct,
            final Classifier[] theClassifiers,
            final Representation theRepresentation,
            final CombinationStrategy combiStrat,
            final TokenizerFactory tFactory, final TrainableFilter sentFilter,
            final Set<String> relevantPunct, final TiesConfiguration config) {
        super(outExt, config);
        targetStructure = targetStruct;
        classifiers = theClassifiers;
        representation = theRepresentation;
        strategy = combiStrat;
        factory = tFactory;
        sentenceFilter = sentFilter;
        relevantPunctuation = relevantPunct;
        Util.LOG.debug("Initialized set of relevant punctuation + symbol "
                + "tokens to " + relevantPunctuation);
    }

    /**
     * Creates a filtering token walker to be used for walking through a
     * document and sentence classification if a double classification approach
     * is used.
     * 
     * @param repFilter the trainable filter to use
     * @return the created walker
     */
    protected abstract FilteringTokenWalker createFilteringTokenWalker(
            final TrainableFilter repFilter);

    /**
     * Helper method that creates the token walker to use for walking through a
     * document. This walker automatically handles sentence filtering if a
     * double classification approach should be used. Delegates to the abstract
     * {@link #createFilteringTokenWalker(RepresentationFilter)}method if
     * sentence filtering should be used.
     * 
     * @return the created walker
     * @throws ProcessingException if an error occurs while initializing the
     * walker
     */
    private TokenWalker createTokenWalker() throws ProcessingException {
        if (isSentenceFiltering()) {
            // use sentence filter to create a filtering waker
            return createFilteringTokenWalker(sentenceFilter);
        } else {
            // no sentence filtering used -- create simple walker
            return new TokenWalker(this, getFactory());
        }
    }

    /**
     * Evaluates precision and recall for {@linkplain #isSentenceFiltering()
     * sentence filtering} on the last processed document.
     * 
     * @param embeddingElements the correct set of embedding elements
     * @return the calculated statistics for sentence filtering on the last
     * document; <code>null</code> if {@linkplain #isSentenceFiltering()
     * sentence filtering} is disabled
     */
    protected FMetricsView evaluateSentenceFiltering(
            final EmbeddingElements embeddingElements) {
        if (isSentenceFiltering() && (walker != null)) {
            // we know we're using a filtering walker
            return FilterEvaluator.evaluate(embeddingElements,
                    (FilteringTokenWalker) walker);
        } else {
            return null;
        }
    }

    /**
     * Returns the set of candidate classes to consider for the current element
     * for each classifier.
     * 
     * @return the value of the attribute
     */
    protected Set[] getActiveClasses() {
        return activeClasses;
    }

    /**
     * Returns the array of classifiers used for the local classification
     * decisions.
     * 
     * @return the local classifier
     */
    public Classifier[] getClassifiers() {
        return classifiers;
    }

    /**
     * Returns the factory used to instantiate tokenizers.
     * 
     * @return the value of the attribute
     */
    public TokenizerFactory getFactory() {
        return factory;
    }

    /**
     * Returns the object used to count documents, contexts, and features and to
     * calculate averages.
     * 
     * @return the used feature count
     */
    public FeatureCount getFeatureCount() {
        return featureCount;
    }

    /**
     * Returns vector of features representing the currently processed element.
     * 
     * @return the value of the attribute
     */
    protected FeatureVector getFeatures() {
        return features;
    }

    /**
     * Returns the buffer of preceding
     * {@link de.fu_berlin.ties.context.Recognition}s from the current
     * document.
     * 
     * @return the buffer
     */
    public PriorRecognitions getPriorRecognitions() {
        return priorRecognitions;
    }

    /**
     * Returns the context representation used for local classifications.
     * 
     * @return the context representation
     */
    public Representation getRepresentation() {
        return representation;
    }

    /**
     * Returns the filter used in the first step of a double classification
     * approach ("sentence filtering").
     * 
     * @return the node filter, or <code>null</code> if no sentence filtering
     * is used
     */
    protected TrainableFilter getSentenceFilter() {
        return sentenceFilter;
    }

    /**
     * Returns the combination strategy used.
     * 
     * @return the combination strategy
     */
    protected CombinationStrategy getStrategy() {
        return strategy;
    }

    /**
     * Returns the target structure specifying the classes to recognize.
     * 
     * @return the used target structure
     */
    public TargetStructure getTargetStructure() {
        return targetStructure;
    }

    /**
     * Returns the token walker used to walk thru documents.
     * 
     * @return the token walker
     */
    protected TokenWalker getWalker() {
        return walker;
    }

    /**
     * Initializes the fields used for processing a document (feature cache,
     * buffer of prior recognitions, token walker, and statistics) and resets
     * the combination strategy.
     * 
     * @throws ProcessingException if an error occurs while initializing
     */
    protected void initFields() throws ProcessingException {
        featureCache = new HashMap<Element, List<LocalFeature>>();
        priorRecognitions = representation.createRecognitionBuffer();
        walker = createTokenWalker();
        featureCount.countDocument();
        strategy.reset();
    }

    /**
     * Checks whether a token is relevant for training and extraction.
     * Tokens containing only punctuation or symbol characters are considered
     * irrevelant unless they have been {@linkplain #markRelevant(String)
     * marked to be relevant}.
     *
     * @param token the token to check
     * @return <code>true</code> if the is relevant for training and
     * extraction; <code>false</code>  it is can be ignored
     */
    protected boolean isRelevant(final String token) {
        // check the relevantPunctuation set for punctuation tokens
        return !TextUtils.punctuationOrSymbol(token)
            || relevantPunctuation.contains(token);
    }

    /**
     * Whether this instance uses sentence filtering (classification of relevant
     * versus irrelevant sentences in a double classification approach).
     * 
     * @return <code>true</code> if sentence filtering is used
     */
    public boolean isSentenceFiltering() {
        return sentenceFilter != null;
    }

    /**
     * Marks a punctuation token as relevant for classification
     * ((because it did occur as the first or last token of an extraction).
     *
     * @param token the token to mark as relevant
     */
    protected void markRelevant(final String token) {
        relevantPunctuation.add(token);
    }

    /**
     * Reset the combination strategy, handling the boolean result value
     * in an appropriate way.
     */
    protected abstract void resetStrategy();

    /**
     * {@inheritDoc}
     */
    public void skip() {
        // call abstract method to reset the combination strategy
        // -- currently disabled because somehow it degrades F1 measure
        //resetStrategy();
    }

    /**
     * Returns a string representation of this object.
     * 
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this).appendSuper(super.toString())
                .append("classifiers", classifiers)
                .append("representation", representation)
                .append("target structure", targetStructure)
                .append("combination strategy", strategy)
                .append("sentence filter", sentenceFilter)
                .append("active classes", activeClasses)
                .append("prior recognitions", priorRecognitions)
                .toString();
    }

    /**
     * Helper that build the {@linkplain #getFeatures() features}and determines
     * the {@linkplain #getActiveClasses() active classes}for an element.
     * 
     * @param element the element to process
     * @param leftText textual content to the left of (preceding)
     * <code>mainText</code>, might be empty
     * @param mainText the main textual content to represent, might be empty
     * @param rightText textual content to the right of (following)
     * <code>mainText</code>, might be empty
     */
    protected void updateState(final Element element, final String leftText,
            final String mainText, final String rightText) {
        // build context representation + update statistics
        features = getRepresentation().buildContext(element, leftText,
                mainText, rightText, priorRecognitions, featureCache, "Token");
        featureCount.update(features);

        // use strategy to determine candidate classes
        activeClasses = getStrategy().activeClasses();
        //Util.LOG.debug("Determined active classes: " + activeClasses);
    }

    /**
     * Returns a read-only view on the counted documents, contexts, and features
     * and the calculated averages. This is not a snapshot but will change
     * whenever the a document is processed.
     * 
     * @return a view on the counts and averages
     */
    public FeatureCountView viewFeatureCount() {
        return featureCount;
    }

    /**
     * Returns a read-only view on the set of punctuation tokens that have been
     * found to be relevant for token classification (because they sometimes
     * occur as the first or last token of an extraction).
     *
     * @return a read-only view on the relevant punctuation
     */
    public Set<String> viewRelevantPunctuation() {
        return Collections.unmodifiableSet(relevantPunctuation);
    }

}