/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * A class (type) of entities to recognize. Instances of this class are
 * immutable and thus thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/02/03 13:16:42 $, $Author: siefkes $
 */
public class TargetClass {

    /**
     * The name of this class.
     */
    private final String name;

    /**
     * Non-public constructor -- should only be used by subclasses and related
     * classes. Other classes should go thru {@link TargetStructure} to access
     * target classes.
     *
     * @param className the name of this class (outer whitespace will be
     * {@linkplain String#trim() trimmed})
     * @throws IllegalArgumentException if <code>className</code> is null or
     * empty (after trimming)
     */
    protected TargetClass(final String className)
            throws IllegalArgumentException {
        super();
        final String trimmedName = StringUtils.trimToEmpty(className);

        if (StringUtils.isEmpty(trimmedName)) {
            throw new IllegalArgumentException(
                    "Target class name must not be null or empty");
        }
        name = trimmedName;
    }

    /**
     * Indicates whether some other object is "equal to" this one, fulfulling
     * the {@link Object#equals(java.lang.Object)} contract.
     *
     * @param obj the reference object with which to compare
     * @return <code>true</code> iff the specified object is a
     * {@link TargetClass} equal to this instance
     */
    public boolean equals(final Object obj) {
        if (obj == this) {
            return true;
        } else if ((obj != null) && (getClass().equals(obj.getClass()))) {
            // used getClass instead of instanceof because otherwise subclasses
            // with additional fields would break the contract
            final TargetClass other = (TargetClass) obj;
            return new EqualsBuilder()
                .append(getName(), other.getName())
                .isEquals();
        } else {
            return false;
        }
    }

    /**
     * Returns the name of this class.
     * @return the value of the attribute
     */
    public String getName() {
        return name;
    }

    /**
     * Returns a hash code value for this object, fulfulling the
     * {@link Object#hashCode()} contract.
     *
     * @return a hash code value for this object
     */
    public int hashCode() {
        // you pick two hard-coded, randomly chosen, non-zero, odd numbers
        // (preferably primes); ideally different for each class
        return new HashCodeBuilder(13, 17)
            .append(getName())
            .toHashCode();
    }

    /**
     * Returns a string representation of this object.
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("name", name)
            .toString();
    }

}
