/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.util.Util;

/**
 * A target structure manages the classes of entities to recognize.
 * Instances of this class are immutable and thus thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.4 $, $Date: 2004/09/06 17:23:10 $, $Author: siefkes $
 */
public class TargetStructure {

    /**
     * Configuration key: names of the classes to recognize (temporarily).
     */
    protected static final String CONFIG_TARGET_CLASSES = "target.classes";

    /**
     * Maps from class names (String) to {@linkplain TargetClass}es.
     * This map is immutable and cannot modified.
     */
    private final Map<String, TargetClass> classes;

    /**
     * Creates a new instance using the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     */
    public TargetStructure() {
        this(TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance using the provided configuration. Currently,
     * the names of target classes are read from the
     * {@link #CONFIG_TARGET_CLASSES} parameter.
     *
     * @param config used to configure this instance
     */
    public TargetStructure(final Configuration config) {
        super();
        final String[] classNames =
            config.getStringArray(CONFIG_TARGET_CLASSES);
        if ((classNames == null) || (classNames.length == 0)) {
            Util.LOG.warn("No target classes defined ("
                + CONFIG_TARGET_CLASSES + " parameter)");
        }

        // using a linked hash set to preserve the order of definition
        final LinkedHashMap<String, TargetClass> classMap =
            new LinkedHashMap<String, TargetClass>();

        // create classes from specified names
        for (int i = 0; i < classNames.length; i++) {
            classMap.put(classNames[i], new TargetClass(classNames[i]));
        }

        // store an immutable view on the map
        classes = Collections.unmodifiableMap(classMap);
    }

    /**
     * Returns the {@link TargetClass} object for a given class name, if
     * defined in this target structure.
     *
     * @param name the name of the class to look up
     * @return the {@link TargetClass} for the given name; or <code>null</code>
     * if the class is not defined in this target structure
     */
    public TargetClass classForName(final String name) {
        return classes.get(name);
    }

    /**
     * Returns the set of names of the target classes at the top of the
     * inheritance hierarchy. This set is immutable and cannot modified.
     *
     * @return the set of class names
     */
    public Set getBaseClassNames() {
        // currently, all class names are returned
        return classes.keySet();
    }

    /**
     * Returns the {@link TargetClass}es at the top of the inheritance
     * hierarchy. This collection is immutable and cannot modified.
     *
     * @return the defined target classes
     */
    public Collection getBaseTargetClasses() {
        // currently, all classes are returned
        return classes.values();
    }

    /**
     * Indicates whether some other object is "equal to" this one, fulfulling
     * the {@link Object#equals(java.lang.Object)} contract.
     *
     * @param obj the reference object with which to compare
     * @return <code>true</code> iff the specified object is a
     * {@link TargetStructure} equal to this instance
     */
    public boolean equals(final Object obj) {
        if (obj == this) {
            return true;
        } else if ((obj != null) && (getClass().equals(obj.getClass()))) {
            // used getClass instead of instanceof because otherwise subclasses
            // with additional fields would break the contract
            final TargetStructure other = (TargetStructure) obj;
            return new EqualsBuilder()
                .append(classes, other.classes)
                .isEquals();
        } else {
            return false;
        }
    }

    /**
     * Returns the set of names of the defined target classes.
     * This set is immutable and cannot modified.
     *
     * @return the set of class names
     */
    public Set<String> getClassNames() {
        return classes.keySet();
    }

    /**
     * Returns the defined {@link TargetClass}es.
     * This collection is immutable and cannot modified.
     *
     * @return the defined target classes
     */
    public Collection<TargetClass> getTargetClasses() {
        return classes.values();
    }

    /**
     * Returns a hash code value for this object, fulfulling the
     * {@link Object#hashCode()} contract.
     *
     * @return a hash code value for this object
     */
    public int hashCode() {
        // you pick two hard-coded, randomly chosen, non-zero, odd numbers
        // (preferably primes); ideally different for each class
        return new HashCodeBuilder(19, 23)
            .append(classes)
            .toHashCode();
    }

    /**
     * Returns a string representation of this object.
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            // no need to duplicate the class names
            .append("target classes", getTargetClasses())
            .toString();
    }

}
