/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract;

import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.math.stat.univariate.moment.StandardDeviation;
import org.dom4j.Document;
import org.dom4j.DocumentException;

import de.fu_berlin.ties.Closeable;
import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ParsingException;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TextProcessor;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.Prediction;
import de.fu_berlin.ties.eval.AccuracyView;
import de.fu_berlin.ties.eval.FMetrics;
import de.fu_berlin.ties.eval.FMetricsView;
import de.fu_berlin.ties.eval.FeatureCountView;
import de.fu_berlin.ties.eval.MultiFMetrics;
import de.fu_berlin.ties.eval.MultiFMetricsView;
import de.fu_berlin.ties.io.FieldContainer;
import de.fu_berlin.ties.io.FieldMap;
import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.util.Util;
import de.fu_berlin.ties.xml.dom.DOMUtils;

/**
 * Trains an extractor and evaluates extraction quality. Processes shuffle
 * files (as generated by {@link de.fu_berlin.ties.eval.ShuffleGenerator}
 * contain the files to use for training and evaluation.
 * For each of these files, a corresponding answer key (*.ans) must exist.
 *
 * <p>Instances of this class are not thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.67 $, $Date: 2004/11/17 09:17:04 $, $Author: siefkes $
 */
public class TrainEval extends TextProcessor implements Closeable {

    /**
     * An inner class wrapping the results of a
     * {@linkplain TrainEval#trainAndEval(String[], File, File, String, Writer)
     * training + evaluation run}.
     */
    public static class Results {

        /**
         * The map from TUNE iterations to
         * {@link EvaluatedExtractionContainer}s containing the evaluated
         * extractions (after the corresponding TUNE iteration).
         */
        private final SortedMap<Integer, EvaluatedExtractionContainer> evaluated
            = new TreeMap<Integer, EvaluatedExtractionContainer>();

        /**
         * A read-only view on the feature count statistics collected during
         * training.
         */
        private FeatureCountView trainFeatureCV;

        /**
         * A read-only view on the feature count statistics collected during
         * extraction.
         */
        private FeatureCountView extractFeatureCV;

        /**
         * Creates a new instance.
         */
        public Results() {
        }

        /**
         * Adds an evaluated extraction container.
         *
         * @param iteration the number of the TUNE iteration to add
         * @param evaluatedExtractions the container to add
         * @throws IllegalArgumentException if there already is a container
         * stored for <code>iteration</code>
         */
        private void addEvaluated(final int iteration,
                final EvaluatedExtractionContainer evaluatedExtractions) 
        throws IllegalArgumentException {
            final EvaluatedExtractionContainer oldValue =
                evaluated.put(Integer.valueOf(iteration), evaluatedExtractions);

            if (oldValue != null) {
                throw new IllegalArgumentException("Cannot store two "
                        + "extraction containers for same iteration ("
                        + iteration + "): " + oldValue + "; "
                        + evaluatedExtractions);
            }
        }

        /**
         * Returns one of the stored evaluated extraction containers.
         *
         * @param iteration the number of the TUNE iteration to return;
         * must be contained in {@link #iterations()}, otherwise
         * <code>null</code> will be returned
         * @return the specified container
         */
        public EvaluatedExtractionContainer getEvaluated(
                final Integer iteration) {
            return evaluated.get(iteration);
        }

        /**
         * Returns one of the stored evaluated extraction containers.
         *
         * @param iteration the number index of the TUNE iterations to return;
         * must be contained in {@link #iterations()}, otherwise
         * <code>null</code> will be returned
         * @return the specified container
         */
        public EvaluatedExtractionContainer getEvaluated(final int iteration) {
            return getEvaluated(Integer.valueOf(iteration));
        }

        /**
         * Returns a read-only view on the feature count statistics collected
         * during training.
         * @return the stored attribute
         */
        public FeatureCountView getExtractFeatureCV() {
            return extractFeatureCV;
        }

        /**
         * Returns a read-only view on the feature count statistics collected
         * during extraction.
         * @return the stored attribute
         */
        public FeatureCountView getTrainFeatureCV() {
            return trainFeatureCV;
        }

        /**
         * Returns an iterator over the TUNE iterations for which extraction
         * results are stored.
         *
         * @return an iterator over the TUNE iterations
         */
        public Iterator<Integer> iterations() {
            return evaluated.keySet().iterator();
        }

        /**
         * Stores a read-only view on the feature count statistics collected
         * during training.
         *
         * @param extractFeatures a read-only view on the feature count
         * statistics collected during extraction
         */
        private void setExtractFeatureCV(
                final FeatureCountView extractFeatures) {
            extractFeatureCV = extractFeatures;
        }

        /**
         * Stores a read-only view on the feature count statistics collected
         * during extraction.
         *
         * @param trainFeatures a read-only view on the feature count statistics
         * collected during training
         */
        private void setTrainFeatureCV(
                final FeatureCountView trainFeatures) {
            trainFeatureCV = trainFeatures;
        }

        /**
         * Returns a string representation of this object.
         *
         * @return a textual representation
         */
        public String toString() {
            return new ToStringBuilder(this)
                .append("training feature count", trainFeatureCV)
                .append("extraction feature count", extractFeatureCV)
                .append("size", evaluated.size())
                .toString();
        }

    }

    /**
     * Configuration key: The percentage of a corpus to use for training.
     */
    public static final String CONFIG_TRAIN_SPLIT = "eval.train-split";

    /**
     * Configuration key: The percentage of a corpus to use for testing
     * (evaluation).
     */
    public static final String CONFIG_TEST_SPLIT = "eval.test-split";

    /**
     * Configuration key: If <code>true</code>, a fully incremental setup is
     * used where the trainer is trained on each document after the extractor
     * processed it.
     */
    public static final String CONFIG_FEEDBACK = "eval.feedback";

    /**
     * Configuration key: The maximum number of iterations used for TUNE
     * (train until no error) training; if 1, training is incremental.
     */
    public static final String CONFIG_TUNE = "train.tune";

    /**
     * Configuration key: TUNE training is stopped if the training accuracy
     * didn't improve for the specified number of iterations.
     */
    public static final String CONFIG_TUNE_STOP = "train.tune.stop";

    /**
     * Configuration key: The maximum number of iterations used for TUNE
     * training the sentence classifier; if 0 or negative, the value of
     * {@link #CONFIG_TUNE} is used.
     */
    public static final String CONFIG_SENTENCE_TUNE = "sent.tune";

    /**
     * Configuration key: Whether to measure results after each TUNE iteration
     * or only at the end of training.
     */
    public static final String CONFIG_TUNE_EACH = "eval.tune.each";

    /**
     * Configuration key: The training iteration after which to evaluate
     * results for the first time if {@link #CONFIG_TUNE_EACH} is enabled.
     */
    public static final String CONFIG_TUNE_SINCE = "eval.tune.since";

    /**
     * Serialization key for the number of the iteration (when TUNE training).
     */
    public static final String KEY_ITERATION = "Iteration";

    /**
     * Serialization key for the number of the run.
     */
    public static final String KEY_RUN = "Run";

    /**
     * Serialization key for the type (either "Train" or "Eval").
     */
    public static final String KEY_TYPE = "Type";

    /**
     * Serialization value for the "Train" type.
     */
    public static final String TYPE_TRAIN = "Train";

    /**
     * Serialization value for the "Eval" type.
     */
    public static final String TYPE_EVAL = "Eval";

    /**
     * Results are measured after each 10th evaluated document if
     * {@link #feedback} is given.
     */
    private static final int MEASURE_FEEDBACK = 10;

    /**
     * The percentage of a corpus to use for training.
     */
    private final float trainSplit;

    /**
     * The percentage of a corpus to use for testign; if <code>-1</code>,
     * all remaining documents are used.
     */
    private final float testSplit;

    /**
     * If <code>true</code>, a fully incremental setup is used where the
     * trainer is trained on each document after the extractor processed it.
     */
    private final boolean feedback;

    /**
     * The maximum number of iterations used for TUNE (train until no error)
     * training; if 1, training is incremental.
     */
    private final int tuneIterations;

    /**
     * TUNE training is stopped if the training accuracy didn't improve for the
     * specified number of iterations.
     */
    private final int tuneStop;

    /**
     * The maximum number of iterations used for TUNE training the sentence
     * classifier; if 0 or negative, the value of {@link #tuneIterations} is
     * used.
     */
    private final int sentenceIterations;

    /**
     * Whether to measure results after each TUNE iteration or only at the
     * end of training.
     */
    private final boolean tuneEach;

    /**
     * The training iteration after which to evaluate results for the first
     * time if {@link #tuneEach} is enabled.
     */
    private final int tuneSince;

    /**
     * A set of iterations after which to evaluate TUNE training in addition to
     * the last one; ignored if {@link #tuneEach} is <code>true</code>.
     */
    private final Set<Integer> tuneEvaluations = new HashSet<Integer>();

    /**
     * The last TUNE iteration that was actually used for training in any
     * batch of processed files.
     */
    private int lastUsedTuneIteration = -1;

    /**
     * An array of multi-metrics calculating summaries esp. standard deviations
     * over all runs. Each array element corresponds to one TUNE operation.
     */
    private final MultiFMetrics[] averages;

    /**
     * An array of multi-metrics incrementally calculating summaries after
     * evaluating each 10th document if {@link #feedback} is given.
     */
    private final List<MultiFMetrics> feedbackAverages;

    /**
     * Calculate feature count statistics over all runs.
     */
    private final FieldContainer featureCounts =
        FieldContainer.createFieldContainer();

    /**
     * Stores precision and recall values for sentence filtering, if used.
     */
    private FieldContainer sentenceMetricsStore;

    /**
     * The number of the current run.
     */
    private int runNo = 0;

    /**
     * Creates a new instance, using a default extension and the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @throws IllegalArgumentException if the configured values are outside the
     * allowed ranges
     * @throws ClassCastException if the configured numeric values cannot be
     * parsed
     * @throws NoSuchElementException if one of the required values is
     * missing from the configuration
     */
    public TrainEval() throws IllegalArgumentException, ClassCastException,
    NoSuchElementException {
        this("metrics");
    }

    /**
     * Creates a new instance, using the {@linkplain TiesConfiguration#CONF
     * standard configuration}.
     *
     * @param outExt the extension to use for output files
     * @throws IllegalArgumentException if the configured values are outside the
     * allowed ranges
     * @throws ClassCastException if the configured numeric values cannot be
     * parsed
     * @throws NoSuchElementException if one of the required values is
     * missing from the configuration
     */
    public TrainEval(final String outExt) throws IllegalArgumentException,
            ClassCastException, NoSuchElementException {
        this(outExt, TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance.
     *
     * @param outExt the extension to use for output files
     * @param config used to configure this instance
     * @throws IllegalArgumentException if the configured values are outside the
     * allowed ranges
     * @throws ClassCastException if the configured numeric values cannot be
     * parsed
     * @throws NoSuchElementException if one of the required values is
     * missing from the configuration
     */
    public TrainEval(final String outExt, final TiesConfiguration config)
            throws IllegalArgumentException, ClassCastException,
            NoSuchElementException {
        // file names are matched case-sensitive
        this(outExt, config.getFloat(CONFIG_TRAIN_SPLIT),
                config.getFloat(CONFIG_TEST_SPLIT),
                config.getInt(CONFIG_TUNE),
                config.getInt(CONFIG_TUNE_STOP),
                config.getBoolean(CONFIG_TUNE_EACH),
                config.getInt(CONFIG_TUNE_SINCE),
                config.getList("eval.tune.list"),
                config.getInt(CONFIG_SENTENCE_TUNE),
                config.getBoolean(CONFIG_FEEDBACK), config);
    }

    /**
     * Creates a new instance.
     *
     * @param outExt the extension to use for output files
     * @param trainingSplit the percentage of a corpus to use for training
     * @param testingSplit the percentage of a corpus to use for testing
     * (evaluation); if <code>-1</code>, all remaining documents (1 -
     * <code>trainingSplit</code>) are used
     * @param tuneRuns the maximum number of iterations used for TUNE
     * (train until no error) training; if 1, training is incremental
     * @param tuneStopAfter TUNE training is stopped if the training accuracy
     * didn't improve for the specified number of iterations.
     * @param measureEachTUNE whether to measure results after each TUNE
     * iteration or only at the end of training
     * @param startMeasureTUNE he training iteration after which to evaluate
     * results for the first time if <code>measureEachTUNE</code> is enabled
     * (ignored otherwise)
     * @param sentenceTUNE the maximum number of iterations used for TUNE
     * training the sentence classifier (if used); if 0 or negative, the value
     * of <code>tuneRuns</code> is used
     * @param tuneEvalList A list of Integers or int Strings specifying
     * iterations after which to evaluate TUNE training in addition to the last
     * one; ignored if <code>measureEachTUNE</code> is <code>true</code>
     * @param giveFeedback if <code>true</code>, a fully incremental setup is
     * used where the trainer is trained on each document after the extractor
     * processed it; it's not allowed to set this both this and
     * <code>measureEachTUNE</code> to <code>true</code> when training for
     * several <code>tuneRuns</code> because that would mean to evaluate on the
     * training set
     * @param config used to configure superclasses, trainer, and extractor;
     * if <code>null</code>, the {@linkplain TiesConfiguration#CONF standard
     * configuration} is used
     * @throws IllegalArgumentException if <code>trainingSplit</code> is not
     * a percentage (larger than 1 or smaller than 0) or if
     * <code>tuneRuns</code> is non-positive
     */
    public TrainEval(final String outExt, final float trainingSplit,
            final float testingSplit, final int tuneRuns,
            final int tuneStopAfter, final boolean measureEachTUNE,
            final int startMeasureTUNE, final List tuneEvalList,
            final int sentenceTUNE, final boolean giveFeedback,
            final TiesConfiguration config)
    throws IllegalArgumentException {
        super(outExt, config);

        // evaluate arguments prior to storing
        if ((trainingSplit < 0.0) || (trainingSplit > 1.0)) {
            throw new IllegalArgumentException(
                "Train split is not a percentage: " + trainingSplit);
        }
        if (testingSplit > 1.0) {
            throw new IllegalArgumentException(
                "Test split is not a percentage: " + testingSplit);
        }
        if (tuneRuns < 1) {
            throw new IllegalArgumentException(
                "Number of TUNE runs must be at least 1: " + tuneRuns);
        }
        if (tuneStopAfter < 1) {
            throw new IllegalArgumentException(
                "TUNE stopping criterium must be at least 1: " + tuneStopAfter);
        }

        if (giveFeedback && (tuneRuns > 1) && measureEachTUNE) {
            throw new IllegalArgumentException("It's not allowed give feedback "
                    + "when evaluating after each of several iteration because "
                    + "that would mean to evaluate on the training set");
        }

        trainSplit = trainingSplit;
        testSplit = testingSplit;
        tuneIterations = tuneRuns;
        tuneStop = tuneStopAfter;
        tuneEach = measureEachTUNE;
        tuneSince = Math.max(startMeasureTUNE, 1); // must be 1 or higher
        sentenceIterations = sentenceTUNE;
        feedback = giveFeedback;
        averages = new MultiFMetrics[tuneIterations];

        if (!tuneEach) { // otherwise we don't need this
            for (final Iterator iter = tuneEvalList.iterator(); iter.hasNext();) {
                // store iterations after which to evaluate
                tuneEvaluations.add(Integer.valueOf(Util.asInt(iter.next())));
            }            
        }

        // init metrics to store incremental F1 values if feedback is given
        if (feedback) {
            feedbackAverages = new ArrayList<MultiFMetrics>();
        } else {
            feedbackAverages = null;
        }

        // Dummy serialization to preload statistical implementations.
        // This avoid instantiation error in close() method if the JAR file
        // has changed in the meanwhile (which might happen if a nightly build
        // takes place during a long test run).
        final FieldContainer dummy = new FieldContainer();
        final MultiFMetrics dummyMetrics = new MultiFMetrics(true);
        dummyMetrics.storeEntries(dummy);
        StandardDeviation dummyDev = new StandardDeviation();
        dummyDev.clear();
    }

    /**
     * Check whether to disable sentnce training; called after each training
     * iteration.
     * 
     * @param trainer the used trainer
     * @param sentenceMetrics metrics calculated for sentence filtering, if
     * used (<code>null</code> otherwise)
     * @param iteration the current iteration
     */
    private void checkSentenceTraining(final Trainer trainer,
            final FMetrics sentenceMetrics, final int iteration) {
        if ((sentenceIterations > 0) && (sentenceIterations == iteration)) {
            trainer.disableSentenceTraining();
            Util.LOG.info("Disabled sentence training after " + iteration
                    + " iterations");
        }

        if (sentenceMetrics != null) {
            updateSentenceMetricsStore(sentenceMetrics, true, iteration);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void close(final int errorCount)
            throws IOException, ProcessingException {
        // don't do anything if there were errors
        if (errorCount <= 0) {
            final File outDir = IOUtils.determineOutputDirectory(getConfig());

            // store feature counts + average metrics incl. standard deviations
            storeValues(featureCounts, outDir, "FeatureCounts", null);

            final FieldContainer averagesStore =
                FieldContainer.createFieldContainer();
            boolean storeIterations = (averages.length > 1);

            // don't store results for more TUNE iterations than actually used
            final int lastToStore = Math.min(averages.length,
                    lastUsedTuneIteration);

            for (int i = 1; i <= lastToStore; i++) {
                // Some might be null because of eval.tune.since or ...list.
                // If tuneEach is disabled, we only store iterations from the
                // tuneEvaluations list + the very last one.
                if (averages[i-1] != null && (tuneEach
                        || tuneEvaluations.contains(Integer.valueOf(i))
                        || i == lastToStore)) {
                    // add iteration number if we TUNEd
                    if (storeIterations) {
                        averagesStore.backgroundMap().put(KEY_ITERATION,
                                Integer.valueOf(i));
                    }

                    averages[i-1].storeEntries(averagesStore);
                }
            }
            storeValues(averagesStore, outDir, "All",
                MultiFMetrics.EXT_METRICS);

            // store incremental F1 if feedback is enabled
            if (feedbackAverages != null) {
                final FieldContainer feedbackAveragesStore =
                    FieldContainer.createFieldContainer();

                for (int i = 0; i < feedbackAverages.size(); i++) {
                    feedbackAverages.get(i).storeEntries(feedbackAveragesStore);
                }                
                storeValues(feedbackAveragesStore, outDir, "Feedback",
                        MultiFMetrics.EXT_METRICS);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    protected void doProcess(final Reader reader, final Writer writer,
            final ContextMap context) throws IOException, ProcessingException {
        // increase run number + read list of files
        runNo++;
        final String[] files = IOUtils.readURIList(reader);

        if (files.length == 0) {
            Util.LOG.info("No files to process");
        } else {
            final String baseName = IOUtils.getBaseName(
                new File((String) context.get(KEY_LOCAL_NAME)));
            final File inDir = (File) context.get(KEY_DIRECTORY);
            final File outDir = IOUtils.determineOutputDirectory(getConfig());
            Results currentResults;
            FieldMap currentFMap;

            MultiFMetricsView currentMetrics = null;

            //Util.LOG.info("Starting evaluation run " + runNo);
            currentResults =
                trainAndEval(files, inDir, outDir, baseName, writer);

            // store feature counts for training + evaluation
            final Integer runNumber = new Integer(runNo);
            currentFMap = currentResults.getTrainFeatureCV().storeFields();
            currentFMap.put(KEY_RUN, runNumber);
            currentFMap.put(MultiFMetrics.KEY_TYPE, "training");
            featureCounts.add(currentFMap);

            currentFMap =
                currentResults.getExtractFeatureCV().storeFields();
            currentFMap.put(KEY_RUN, runNumber);
            currentFMap.put(MultiFMetrics.KEY_TYPE, "extraction");
            featureCounts.add(currentFMap);

            // update average metrics
            final Iterator<Integer> iterIter = currentResults.iterations();
            Integer iteration;
            int it = 0;

            while (iterIter.hasNext()) {
                iteration = iterIter.next();
                it = iteration.intValue();
                currentMetrics =
                    currentResults.getEvaluated(iteration).viewMetrics();
                if (averages[it-1] == null) {
                    averages[it-1] = new MultiFMetrics(true);
                }
                averages[it-1].update(currentMetrics);
            }

            // if there are additional averages it means that we stopped TUNEing
            // early so the last result is valid for all later iterations too
            for (it++; it <= averages.length; it++) {
                if (averages[it-1] == null) {
                    averages[it-1] = new MultiFMetrics(true);
                }

                // this will copy the last existing result which is what we want
                averages[it-1].update(currentMetrics);
            }
        }
    }

    /**
     * Returns the percentage of a corpus to use for testing (evaluation).
     *
     * @return the percentage to use for evaluation; if negative, all
     * remaining documents (1 - {@link #getTrainSplit()}) are used for
     * evaluation
     */
    public float getTestSplit() {
        return testSplit;
    }

    /**
     * Returns the percentage of a corpus to use for training; the remaining
     * documents (1-x) are used for evaluation.
     *
     * @return the percentage to use for training
     */
    public float getTrainSplit() {
        return trainSplit;
    }

    /**
     * Creates and initializes a extractor to use for an evaluation run,
     * re-using the components of the provided trainer. Subclasses can
     * overwrite this method to provide a different extractor.
     *
     * @param trainer trainer whose components should be re-used
     * @return the created extractor
     */
    protected Extractor initExtractor(final Trainer trainer) {
        // we don't need an output extension
        return new Extractor(null, trainer);
    }

    /**
     * Creates and initializes a trainer to use for an evaluation run,
     * configured from the
     * {@link de.fu_berlin.ties.ConfigurableProcessor#getConfig() stored
     * configuration}. Subclasses can overwrite this method to provide a
     * different trainer.
     *
     * @param runDirectory directory used to run the classifier
     * @return the created trainer
     * @throws ProcessingException if an error occurs during initialization
     */
    protected Trainer initTrainer(final File runDirectory)
            throws ProcessingException {
        // we don't need an output extension
        final Trainer trainer = new Trainer(null, runDirectory, getConfig());
        // ensure the prediction model is empty
        trainer.reset();
        return trainer;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("train split", trainSplit)
            .append("test split", testSplit)
            .append("tune iterations", tuneIterations)
            .append("tune stops after", tuneStop)
            .append("measure after each iteration", tuneEach)
            .append("starting from", tuneSince)
            .append("sentence iterations", sentenceIterations)
            .append("feedback", feedback)
            .toString();
    }

    /**
     * Chooses files to use for training and files to use for evaluation,
     * depending on the configured settings.
     *
     * @param allFiles the array of file names to process
     * @param trainFiles populated with the files to use for training, will
     * be populated with the first <em>{@link #getTrainSplit()} *
     * allFiles.length</em> files; must initially be empty
     * @param evalFiles populated with the files to use for evaluation, will
     * be populated from the next <em>{@link #getTestSplit()} *
     * allFiles.length</em> remaining files (or all remaining files if test
     * split is negative); must initially be empty
     * @throws IllegalArgumentException if the lists aren't empty
     */
    private void selectFiles(final String[] allFiles,
            final List<String> trainFiles, final List<String> evalFiles)
            throws IllegalArgumentException {
        // check arguments
        if (!trainFiles.isEmpty() || !evalFiles.isEmpty()) {
            throw new IllegalArgumentException(
                "Lists of train files and eval files must initially be empty");
        }

        final int numTrainFiles = Math.round(trainSplit * allFiles.length);
        final int filesToUse;

        if (testSplit < 0) {
            // use all remaining files for evaluation
            filesToUse = allFiles.length;
        } else {
            filesToUse = Math.min(allFiles.length,
                Math.round((trainSplit + testSplit) * allFiles.length));
        }

        // add first numTrainFiles to trainFiles, rest to evalFiles
        for (int i = 0; i < filesToUse; i++) {
            if (i < numTrainFiles) {
                // add to training files
                trainFiles.add(allFiles[i]);
            } else {
                // add to evaluation files
                evalFiles.add(allFiles[i]);
            }
        }

        Util.LOG.debug("Using " + filesToUse + " of " + allFiles.length
            + " files: " + trainFiles.size() + " for training, "
            + evalFiles.size() + " for evaluation");
    }

    /**
     * Helper method for serializing the calculated metrics.
     *
     * @param outDirectory directory used to do this run and store the results
     * @param baseName the base name of the files to use for storing
     * all extractions and training statistics
     * @param iteration the number of the current iteration
     * @param evaluated the container of evaluated extractions to tore
     * @param sentenceMetrics metrics calculated for sentence filtering, if
     * used (<code>null</code> otherwise)
     * @throws IOException if an I/O error occurs while serializing
     */
    private void serializeExtractions(final File outDirectory,
            final String baseName, final int iteration,
            final EvaluatedExtractionContainer evaluated,
            final FMetrics sentenceMetrics)
    throws IOException {
        // serialize + store evaluated extractions
        final FieldContainer resultStorage =
            FieldContainer.createFieldContainer();
        evaluated.storeEntries(resultStorage);
        final File evaluatedFile = storeValues(resultStorage,
            outDirectory, baseName, Extractor.EXT_EXTRACTIONS);
        Util.LOG.info("Stored results of training + evaluation run in "
            + evaluatedFile);

        // update metrics of sentence filtering, if used
        if (sentenceMetrics != null) {
            updateSentenceMetricsStore(sentenceMetrics, false, iteration);
        }
    }

    /**
     * Helper method for serializing the calculated metrics.
     *
     * @param outDirectory directory used to do this run and store the results
     * @param baseName the base name of the files to use for storing
     * all extractions and training statistics
     * @param writer used to serialize the calculated metrics
     * @param results contains the evaluated extractions
     * @throws IOException if an I/O error occurs while serializing the metrics
     */
    private void serializeMetrics(final File outDirectory,
            final String baseName, final Writer writer,
            final Results results)
    throws IOException {
        // serialize metrics to output file
        final FieldContainer metricsStore =
            FieldContainer.createFieldContainer();
        Integer iteration;
        EvaluatedExtractionContainer evaluated;
        final boolean storeIterations = (tuneIterations > 1);
        Iterator<Integer> tuneIter = results.iterations();

        while (tuneIter.hasNext()) {
            iteration = tuneIter.next();
            evaluated = results.getEvaluated(iteration);

            // add iteration number, if there are several ones to store
            if (storeIterations) {
                metricsStore.backgroundMap().put(KEY_ITERATION, iteration);
            }

            evaluated.viewMetrics().storeEntries(metricsStore);
        }

        metricsStore.store(writer);
        writer.flush();

        // serialize sentence metrics
        if (sentenceMetricsStore != null) {
            storeValues(sentenceMetricsStore, outDirectory,
                    baseName, "sent");
        }
    }

    /**
     * Helper method for serializing the training accuracies.
     *
     * @param outDirectory directory used to do this run and store the results
     * @param baseName the base name of the files to use for storing
     * all extractions and training statistics
     * @param accContainers a an of containers used to store the accuracy of
     * each classifier, might be <code>null</code> if accuracies aren't measured
     * @param globalAcc an array of accuracies, one for each classifier
     * @throws IOException if an I/O error occurs while serializing
     */
    private void serializeTrainingMetrics(final File outDirectory,
            final String baseName, final FieldContainer[] accContainers,
            final AccuracyView[] globalAcc)
    throws IOException {
        if (accContainers != null) {
            String name;

            for (int j = 0; j < globalAcc.length; j++) {
                if (globalAcc.length > 1) {
                    // append prefix letter (a, b, c etc.)
                    name = baseName + (char) ('a' + j);
                } else {
                    // sufficient to use base name
                    name = baseName;
                }
                storeValues(accContainers[j], outDirectory, name, "train");
            }
        }
    }

    /**
     * Stores the contents of a storable container.
     *
     * @param container the container to store
     * @param directory the directory in which to store the data
     * @param baseName the base name of the file using for storing the data
     * @param extension the file extension to use; if <code>null</code>,
     * the {@linkplain FieldContainer#recommendedExtension() recommended
     * extension} is used
     * @return the file used for storing the data
     * @throws IOException if an I/O error occurs
     */
    private File storeValues(final FieldContainer container,
            final File directory, final String baseName,
            final String extension) throws IOException {

        // use default extension if none is specified
        final String usedExtension = ((extension != null)
                ? extension : FieldContainer.recommendedExtension());
        final File outFile = IOUtils.createOutFile(directory, baseName,
            usedExtension);
        final Writer writer = IOUtils.openWriter(outFile, getConfig());
        container.store(writer);
        writer.flush();
        writer.close();
        return outFile;
    }

    /**
     * Processes an array of files. For each file, a corresponding answer key
     * (*.ans) must exist.
     *
     * @param files the array of file names to process (relative to the
     * <code>inDirectory</code>)
     * @param inDirectory directory containing the files to process
     * @param outDirectory directory used to do this run and store the results
     * @param baseName the base name of the files to use for storing
     * all extractions and training statistics
     * @param writer used to serialize the calculated metrics
     * @return a wrapper of the results of this run
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public Results trainAndEval(final String[] files, final File inDirectory,
            final File outDirectory, final String baseName,
            final Writer writer) throws IOException, ProcessingException {
        // make given directory default dir
        IOUtils.setDefaultDirectory(outDirectory);

        // select files for evaluation + rest for training
        final List<String> trainFiles = new LinkedList<String>();
        final List<String> evalFiles = new LinkedList<String>();
        selectFiles(files, trainFiles, evalFiles);

        final long trainStartTime = System.currentTimeMillis();
        File currentFile;
        Document currentDoc;
        ExtractionContainer currentAnswers;

        // store token accuracies during TOE training
        FieldContainer[] accContainers = null;
        AccuracyView[] currentAcc;
        AccuracyView[] globalAcc = null;
        FieldMap accMap;
        final boolean storeIterations = (tuneIterations > 1);

        // measure accuracies for TUNE training
        double[] currentOverallAcc;
        double[] lastOverallAcc = null;
        boolean allAreOptimal;
        boolean noneGotBetter;
        boolean someGotWorse;
        int noneGotBetterCounter = 0;
        boolean continueTraining = true;
        int evalFileNum, feedbackNum;

        // create new trainer
        final Trainer trainer = initTrainer(outDirectory);
        Extractor extractor = null;

        // measure sentence filtering, if used
        FMetrics sentenceMetrics = null;
        FMetricsView newSentenceMetrics;

        if (trainer.isSentenceFiltering()) {
            sentenceMetricsStore = FieldContainer.createFieldContainer();
        } else {
            sentenceMetricsStore = null;
        }

        // store evaluated extraction containers
        final Results results = new Results();

/*        // pseudo-randonness generator if TUNEing
        final Random pseudoRandom = (tuneIterations > 1) ?
                Util.reproducibleRandom() : null; */

        for (int i = 1; (i <= tuneIterations) && continueTraining; i++) {
            if (tuneIterations > 1) {
                Util.LOG.debug("Starting TUNE iteration " + i + "/" 
                        + tuneIterations + " (will stop if no improvement for "
                        + tuneStop + " iterations)");
            }

/* degrades results:
            // reshuffling train files in 2nd + following TUNE iterations
            if (i > 1) {
                Collections.shuffle(trainFiles, pseudoRandom);
                Util.LOG.debug("Pseudo-randomly reshuffled training files");
            } */

            final Iterator trainIter = trainFiles.iterator();
            trainer.resetGlobalAccuracy();

            // measure sentence filtering, if used
            if (trainer.isSentenceFiltering()) {
                sentenceMetrics = new FMetrics();
            }

            //  run on trainFiles
            while (trainIter.hasNext()) {
                currentFile = new File(inDirectory, (String) trainIter.next());
                Util.LOG.debug("Starting to train " + currentFile);

                try {
                    currentDoc =
                        DOMUtils.readDocument(currentFile, getConfig());
                } catch (DocumentException de) {
                    // wrap exception
                    throw new ParsingException("Error while parsing "
                            + currentFile + ": " + de.toString(), de);
                }

                // read answer keys in DSV format (*.ans file must exist)
                currentAnswers = AnswerBuilder.readCorrespondingAnswerKeys(
                    trainer.getTargetStructure(), currentFile, getConfig());

                // train the trainer
                currentAcc = trainer.train(currentDoc, currentAnswers);

                // store global + local accuracy if in TOE mode
                if (currentAcc != null) {
                    globalAcc = trainer.viewGlobalAccuracy();

                    // initialize containers once
                    if (accContainers == null) {
                        accContainers = new FieldContainer[globalAcc.length];
                        for (int j = 0; j < globalAcc.length; j++) {
                            accContainers[j]
                                = FieldContainer.createFieldContainer();
                        }
                    }

                    // store each accuracy
                    for (int j = 0; j < globalAcc.length; j++) {
                        accMap = globalAcc[j].storeFields();
                        accMap.putAll(currentAcc[j].storeFields());
                        accMap.put(Prediction.KEY_SOURCE,
                            IOUtils.getBaseName(currentFile));
                        if (storeIterations) {
                            accMap.put(KEY_ITERATION, new Integer(i));
                        }
                        accContainers[j].add(accMap);
                    }
                }
                Util.LOG.info("Trained " + currentFile);

                // measure sentence filtering, if used
                if (trainer.isSentenceFiltering()) {
                    newSentenceMetrics = trainer.evaluateSentenceFiltering();
                    Util.LOG.debug("Evaluated sentence filtering for trained "
                            + "document: " + newSentenceMetrics);
                    sentenceMetrics.update(newSentenceMetrics);
                }
            }

            // check whether to continue sentence training + update metrics
            checkSentenceTraining(trainer, sentenceMetrics, i);

            // stop TUNE if all accuracies did not or cannot longer increase
            if (globalAcc != null) {
                currentOverallAcc = new double[globalAcc.length];
                allAreOptimal = true;
                noneGotBetter = true;
                someGotWorse = false;

                for (int j = 0; j < currentOverallAcc.length; j++) {
                    currentOverallAcc[j] = globalAcc[j].getAccuracy();

                    // test whether all accuracies are optimal
                    allAreOptimal = allAreOptimal
                        && (currentOverallAcc[j] >= 1.0);

                    // test whether all accuracies are better than the last ones
                    if (lastOverallAcc != null) {
                        noneGotBetter = noneGotBetter
                            && (currentOverallAcc[j] <= lastOverallAcc[j]);
                        someGotWorse = someGotWorse
                            || (currentOverallAcc[j] < lastOverallAcc[j]);
                    } else {
                        noneGotBetter = false;
                    }
                }

                if (noneGotBetter) {
                    noneGotBetterCounter++;

                    if (noneGotBetterCounter >= tuneStop) {
                        // reached stopping criterium for TUNE training
                        continueTraining = false;
                        Util.LOG.debug("Stopping TUNE training after " + i
                                + " iterations because current accuracies ("
                                + ArrayUtils.toString(currentOverallAcc)
                                + ") aren't higher than last ones ("
                                + ArrayUtils.toString(lastOverallAcc)
                                + ") for the " + tuneStop + ". time");                        
                    } else if (someGotWorse) {
                        // stop TUNE training because accuracy degraded
                        continueTraining = false;
                        Util.LOG.debug("Stopping TUNE training after " + i
                                + " iterations because current accuracies ("
                                + ArrayUtils.toString(currentOverallAcc)
                                + ") are lower than last ones ("
                                + ArrayUtils.toString(lastOverallAcc) + ")");                        
                    }
                }

                if (allAreOptimal) {
                    continueTraining = false;
                    Util.LOG.debug("Stopping TUNE training after " + i
                            + " iterations because all accuracies are already "
                            + "optimal: "
                            + ArrayUtils.toString(currentOverallAcc));
                }

                lastOverallAcc = currentOverallAcc;
            }

            if ((tuneEach && (i >= tuneSince))
                    || !continueTraining
                    || (i == tuneIterations)
                    || tuneEvaluations.contains(Integer.valueOf(i))) {
                // evaluate extractions
                Util.LOG.info("Finished training using "
                    + ArrayUtils.toString(trainer.getClassifiers()) + "; "
                    + Util.showDuration(trainStartTime));
                final long evalStartTime = System.currentTimeMillis();
                evalFileNum = 0;

                // create extractor (first time only)
                if (extractor == null) {
                    extractor = initExtractor(trainer);
                }

                final Iterator evalIter = evalFiles.iterator();
                final EvaluatedExtractionContainer evaluated =
                    new EvaluatedExtractionContainer(
                        extractor.getTargetStructure(), getConfig());
                ExtractionContainer currentResults;
                MultiFMetricsView interimMetrics;
                String currentType;
                Iterator typeIter;

                // measure sentence filtering, if used
                if (extractor.isSentenceFiltering()) {
                    sentenceMetrics = new FMetrics();
                 }

                // run on evalFiles
                while (evalIter.hasNext()) {
                    currentFile = new File(inDirectory,
                        (String) evalIter.next());
                    evalFileNum++;
                    Util.LOG.debug("Starting to extract and evaluate file #"
                            + evalFileNum + ": " + currentFile);

                    try {
                        currentDoc = DOMUtils.readDocument(currentFile,
                            getConfig());
                    } catch (DocumentException de) {
                        // wrap exception
                        throw new ParsingException(de);
                    }

                    // extract results
                    currentResults = extractor.extract(currentDoc);

                    // read answer key
                    currentAnswers = AnswerBuilder.readCorrespondingAnswerKeys(
                        trainer.getTargetStructure(), currentFile, getConfig());

                    // invoke the trainer if feedback should be given
                    if (feedback) {
                        currentAcc = trainer.train(currentDoc, currentAnswers);                        
                    }
                    
                    // measure sentence filtering, if used (we do this prior to
                    // evaluating the predicted extractions because
                    // evaluateBatch destructively modifies the answer keys)
                    if (extractor.isSentenceFiltering()) {
                        newSentenceMetrics =
                            extractor.evaluateSentenceFiltering(currentAnswers);
                        Util.LOG.debug("Evaluated sentence filtering for "
                                + "current document: " + newSentenceMetrics);
                        sentenceMetrics.update(newSentenceMetrics);
                    }

                    // compare with results, storing base name as key
                    evaluated.evaluateBatch(currentResults, currentAnswers,
                        IOUtils.getBaseName(currentFile));

                    // log results
                    interimMetrics = evaluated.viewMetrics();
                    Util.LOG.info("Extracted and evaluated " + currentFile
                        + ", interim results: " + interimMetrics.viewAll());
                    typeIter = interimMetrics.types().iterator();

                    while (typeIter.hasNext()) {
                        currentType = (String) typeIter.next();
                        Util.LOG.debug("Interim results for " + currentType
                            + ": " + interimMetrics.view(currentType));
                    }

                    // store results after each 10th file (and after last file)
                    // if feedback gis iven
                    if (feedback && ((evalFileNum % MEASURE_FEEDBACK == 0)
                            || !evalIter.hasNext())) {
                        feedbackNum = new Double(Math.ceil((double) evalFileNum
                                / MEASURE_FEEDBACK)).intValue();

                        // initialize new metrics if necessary (first run)
                        if (feedbackAverages.size() < feedbackNum) {
                            feedbackAverages.add(new MultiFMetrics(true));
                        }

                        // add current results
                        feedbackAverages.get(feedbackNum - 1).update(
                                evaluated.viewMetrics());
                    }
                }

                serializeExtractions(outDirectory, baseName, i, evaluated,
                        sentenceMetrics);
                results.addEvaluated(i, evaluated);
                Util.LOG.info("Finished extraction and evaluation using "
                    + ArrayUtils.toString(extractor.getClassifiers()) + "; "
                    + Util.showDuration(evalStartTime));
            }

            // update max. number of iterations used in any batch if higher
            lastUsedTuneIteration = Math.max(lastUsedTuneIteration, i);
        }

        // serialize metrics + return extractions and feature counts
        serializeTrainingMetrics(outDirectory, baseName, accContainers,
                globalAcc);
        serializeMetrics(outDirectory, baseName, writer, results);
        results.setTrainFeatureCV(trainer.viewFeatureCount());
        results.setExtractFeatureCV(extractor.viewFeatureCount());
        return results;
    }

    /**
     * Adds the results of a training or evaluation iteration to the statistics
     * calculated for sentence filtering.
     *
     * @param newMetrics the metrics to add
     * @param isTraining whether this is a training or an evaluation result
     * @param iteration the number of the iteration (counting from 1)
     */
    private void updateSentenceMetricsStore(final FMetricsView newMetrics,
            final boolean isTraining, final int iteration) {
        final FieldMap fields = newMetrics.storeFields();

        // type is either "Train" or "Eval"
        fields.put(KEY_TYPE, isTraining ? TYPE_TRAIN : TYPE_EVAL);
        fields.put(KEY_ITERATION, new Integer(iteration));
        sentenceMetricsStore.add(fields);
    }

}
