/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract;

import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.Set;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Document;
import org.dom4j.Element;

import de.fu_berlin.ties.combi.CombinationState;
import de.fu_berlin.ties.combi.CombinationStrategy;
import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;

import de.fu_berlin.ties.classify.Prediction;
import de.fu_berlin.ties.classify.PredictionDistribution;
import de.fu_berlin.ties.classify.Probability;
import de.fu_berlin.ties.classify.TrainableClassifier;
import de.fu_berlin.ties.context.Representation;
import de.fu_berlin.ties.eval.Accuracy;
import de.fu_berlin.ties.eval.AccuracyView;
import de.fu_berlin.ties.eval.FMetricsView;
import de.fu_berlin.ties.filter.EmbeddingElements;
import de.fu_berlin.ties.filter.FilteringTokenWalker;
import de.fu_berlin.ties.filter.Oracle;
import de.fu_berlin.ties.filter.TrainableFilter;
import de.fu_berlin.ties.filter.TrainableFilteringTokenWalker;
import de.fu_berlin.ties.text.TokenDetails;
import de.fu_berlin.ties.text.TokenizerFactory;
import de.fu_berlin.ties.util.CollectionUtils;
import de.fu_berlin.ties.util.Util;
import de.fu_berlin.ties.xml.dom.DOMUtils;

/**
 * A trainer trains a local {@link de.fu_berlin.ties.classify.Classifier}
 * to be used for extraction.
 *
 * <p>Instances of this class are not thread-safe and cannot handle training on
 * several documents in parallel.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.49 $, $Date: 2004/12/07 12:01:48 $, $Author: siefkes $
 */
public class Trainer extends ExtractorBase implements Oracle {

    /**
     * Configuration key for determining the training mode
     * ({@link #isTrainingOnlyErrors()}).
     */
    public static final String CONFIG_TOE = "train.only-errors";

    /**
     * Configuration key determining whether the trainer only ensures that all
     * answer keys exist and can be located in the document instead of doing
     * any training.
     */
    public static final String CONFIG_TEST_ONLY = "train.test-only";

    /**
     * Prefix used for serializing the global (overall) accuracy.
     */
    public static final String PREFIX_GLOBAL_ACC = "Overall ";

    /**
     * Prefix used for serializing the local (document-specific) accuracy.
     */
    public static final String PREFIX_LOCAL_ACC = "Document ";

    /**
     * Training the embedded sentence filter (if used) can be disabled by
     * setting this to <code>false</code>.
     */
    private boolean sentenceTrainingEnabled = true;

    /**
     * The trainalbe classifier(s) used for the local classification decisions.
     */
    private final TrainableClassifier[] trainableClassifiers;

    /**
     * Whether to train only errors (TOE mode, recommmended) or to train all
     * instances (brute-force mode).
     */
    private final boolean trainingOnlyErrors;

    /**
     * If <code>true</code> the trainer only ensures that all answer keys exist
     * and can be located in the document instead of doing any training.
     */
    private final boolean testingOnly;

    /**
     * The token accuracy of all documents trained so far by each classifier,
     * measured in {@linkplain #isTrainingOnlyErrors() TOE mode}.
     */
    private Accuracy[] globalAccuracies = null;

    /**
     * The token accuracy of the current document so far by each classifier,
     * measured in {@linkplain #isTrainingOnlyErrors() TOE mode}.
     */
    private Accuracy[] localAccuracies = null;

    /**
     * Used to locate extractions.
     */
    private ExtractionLocator locator;

    /**
     * Used to determine which elements contain extractions if sentence
     * filtering is used (first step of a double classification approach).
     */
    private EmbeddingElements embeddingElements;

    /**
     * A copy of the current extraction that contains only the part that has
     * already been recognized, to be included in the context representation.
     */
    private Extraction partialExtraction;

    /**
     * Creates a new instance without specifying an output extension (which
     * isn't needed anyway, because this class doesn't produce output).
     * Delegates to {@link #Trainer(String)} using a dummy
     * extension.
     *
     * @throws IllegalArgumentException if the combination strategy cannot be
     * initialized
     * (cf. {@link CombinationStrategy#createStrategy(java.util.Set,
     * TiesConfiguration)})
     * @throws ProcessingException if an error occurs during initialization
     */
    public Trainer() throws IllegalArgumentException, ProcessingException {
        this("tmp");
    }

    /**
     * Creates a new instance. Delegates to
     * {@link #Trainer(String, TiesConfiguration)} using the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     *
     * @param outExt the extension to use for output files
     * @throws IllegalArgumentException if the combination strategy cannot be
     * initialized
     * (cf. {@link CombinationStrategy#createStrategy(java.util.Set,
     * TiesConfiguration)})
     * @throws ProcessingException if an error occurs during initialization
     */
    public Trainer(final String outExt)
            throws IllegalArgumentException, ProcessingException {
        this(outExt, TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance. Delegates to the
     * {@link #Trainer(String, File, TiesConfiguration)} constructor without
     * specifying a <code>runDirectory</code>.
     *
     * @param outExt the extension to use for output files
     * @param config the configuration to use
     * @throws IllegalArgumentException if the combination strategy cannot be
     * initialized
     * (cf. {@link CombinationStrategy#createStrategy(java.util.Set,
     * TiesConfiguration)})
     * @throws ProcessingException if an error occurs during initialization
     */
    public Trainer(final String outExt, final TiesConfiguration config)
            throws IllegalArgumentException, ProcessingException {
        this(outExt, null, config);
    }

    /**
     * Creates a new instance. Sets the training mode
     * ({@link #isTrainingOnlyErrors()}) to the value of the {@link #CONFIG_TOE}
     * configuration key in the provided configuration and delegates to the
     * corresponding {@linkplain
     * ExtractorBase#ExtractorBase(String, TiesConfiguration) super constructor}
     * to configure the fields.
     *
     * @param outExt the extension to use for output files
     * @param runDirectory the directory to run the classifier in; used instead
     * of the
     * {@linkplain de.fu_berlin.ties.classify.ExternalClassifier#CONFIG_DIR
     * configured directory} if not <code>null</code>
     * @param config the configuration to use
     * @throws IllegalArgumentException if the combination strategy cannot be
     * initialized
     * (cf. {@link CombinationStrategy#createStrategy(java.util.Set,
     * TiesConfiguration)})
     * @throws ProcessingException if an error occurs during initialization
     */
    public Trainer(final String outExt, final File runDirectory,
            final TiesConfiguration config)
            throws IllegalArgumentException, ProcessingException {
        super(outExt, runDirectory, config);
        trainingOnlyErrors = config.getBoolean(CONFIG_TOE);
        testingOnly = config.getBoolean(CONFIG_TEST_ONLY);

        // store references to classifiers to allow training
        trainableClassifiers = new TrainableClassifier[getClassifiers().length];

        for (int i = 0; i < trainableClassifiers.length; i++) {
            trainableClassifiers[i] =
                (TrainableClassifier) getClassifiers()[i];
        }

        // initialize global accuracies if training-only-errors (TOE)
        resetGlobalAccuracy();
    }

    /**
     * Creates a new instance, using the
     * {@link TiesConfiguration#CONF standard configuration} to configure the
     * training mode and the superclasses.
     *
     * @param outExt the extension to use for output files
     * @param targetStruct the target structure specifying the classes to
     * recognize
     * @param theClassifiers the array of classifiers to train
     * @param theRepresentation the context representation to use training
     * @param combiStrat the combination strategy to use
     * @param tFactory used to instantiate tokenizers
     * @param sentFilter the filter used in the first step of a double
     * classification approach ("sentence filtering"); if <code>null</code>,
     * no sentence filtering is used
     */
    public Trainer(final String outExt, final TargetStructure targetStruct,
            final TrainableClassifier[] theClassifiers,
            final Representation theRepresentation,
            final CombinationStrategy combiStrat,
            final TokenizerFactory tFactory, final TrainableFilter sentFilter) {
        this(outExt, targetStruct, theClassifiers, theRepresentation,
                combiStrat, tFactory, sentFilter,
                CollectionUtils.arrayAsSet(TiesConfiguration.CONF
                        .getStringArray(CONFIG_RELEVANT_PUNCTUATION)),
                TiesConfiguration.CONF.getBoolean(CONFIG_TOE),
                TiesConfiguration.CONF.getBoolean(CONFIG_TEST_ONLY),
                TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance.
     *
     * @param outExt the extension to use for output files
     * @param targetStruct the target structure specifying the classes to
     * recognize
     * @param theClassifiers the array of classifiers to train
     * @param theRepresentation the context representation to use training
     * @param combiStrat the combination strategy to use
     * @param tFactory used to instantiate tokenizers
     * @param sentFilter the filter used in the first step of a double
     * classification approach ("sentence filtering"); if <code>null</code>,
     * no sentence filtering is used
     * @param relevantPunct a set of punctuation tokens that have been found to
     * be relevant for token classification; might be empty but not
     * <code>null</code>
     * @param trainOnlyErrors whether to train only errors (TOE mode,
     * recommmended) or to train all instances (brute-force mode)
     * @param testOnly if <code>true</code> the trainer only ensures that all
     * answer keys exist and can be located in the document instead of doing
     * any training
     * @param config used to configure superclasses; if <code>null</code>,
     * the {@linkplain TiesConfiguration#CONF standard configuration} is used
     */
    public Trainer(final String outExt, final TargetStructure targetStruct,
            final TrainableClassifier[] theClassifiers,
            final Representation theRepresentation,
            final CombinationStrategy combiStrat,
            final TokenizerFactory tFactory, final TrainableFilter sentFilter,
            final Set<String> relevantPunct,
            final boolean trainOnlyErrors, final boolean testOnly,
            final TiesConfiguration config) {
        super(outExt, targetStruct, theClassifiers, theRepresentation,
            combiStrat, tFactory, sentFilter, relevantPunct, config);
        trainingOnlyErrors = trainOnlyErrors;
        testingOnly = testOnly;
        trainableClassifiers = theClassifiers;

        // initialize global accuracies if training-only-errors (TOE)
        resetGlobalAccuracy();
    }


    /**
     * {@inheritDoc}
     */
    protected FilteringTokenWalker createFilteringTokenWalker(
            final TrainableFilter repFilter) {
        // create trainable walker, passing myself as oracle + skip handler
        return new TrainableFilteringTokenWalker(this, getFactory(),
                repFilter, this, this, sentenceTrainingEnabled);
    }

    /**
     * Disables training the embedded sentence filter, if sentence filtering is
     * used.
     */
    public void disableSentenceTraining() {
        sentenceTrainingEnabled = false;
    }

    /**
     * Re-enables training the embedded filter, if sentence filtering is
     * used.
     */
    public void enableSentenceTraining() {
        sentenceTrainingEnabled = true;
    }

    /**
     * Evaluates precision and recall for {@linkplain #isSentenceFiltering()
     * sentence filtering} on the last processed document.
     *
     * @return the calculated statistics for sentence filtering on the
     * last document; <code>null</code> if {@linkplain
     * #isSentenceFiltering() sentence filtering} is disabled
     */
    public FMetricsView evaluateSentenceFiltering() {
        // use embeddingElements initializes by last "train" method
        return evaluateSentenceFiltering(embeddingElements);
    }

    /**
     * Returns the trainable classifiers used for the local classification
     * decisions. Delegates to {@link ExtractorBase#getClassifiers()} and
     * casts to result (the constructor ensures that only
     * {@link TrainableClassifier}s are accepted).
     *
     * @return the trainable classifier
     */
    private TrainableClassifier[] getTrainableClassifiers() {
        return trainableClassifiers;
    }

    /**
     * Initializes accuracy statistics for each classifier.
     *
     * @param prefix the prefix to use for these accuracy statistics
     * @return an array of ccuracy statistics, one for each classifier
     */
    private Accuracy[] initAccuracies(final String prefix) {
        final Accuracy[] result = new Accuracy[getClassifiers().length];

        for (int i = 0; i < result.length; i++) {
            result[i] = new Accuracy(prefix);
        }

        return result;
    }

    /**
     * If <code>true</code> the trainer only ensures that all answer keys exist
     * and can be located in the document instead of doing any training.
     * @return the value of the attribute
     */
    public boolean isTestingOnly() {
        return testingOnly;
    }

    /**
     * Whether to train only errors (TOE mode, recommmended) or to train all
     * instances (brute-force mode).
     * @return the value of the attribute
     */
    public boolean isTrainingOnlyErrors() {
        return trainingOnlyErrors;
    }

    /**
     * Trains the local classifier with the correct extractions of an XML
     * document, using the provided context representation. In
     * {@linkplain #isTrainingOnlyErrors() TOE mode}, training statistics
     * are serialized to the <code>writer</code>. The answer keys must be
     * in a corresponding file ending in {@link AnswerBuilder#EXT_ANSWERS} in
     * the same directory (when processing a local file) or in the current
     * working directory (when processin an URL).
     *
     * @param document the document to read
     * @param writer ignored by this method
     * @param context a map of objects that are made available for processing
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public void process(final Document document, final Writer writer,
            final ContextMap context) throws IOException, ProcessingException {
        // read answer keys
        final ExtractionContainer answerKeys =
            AnswerBuilder.readCorrespondingAnswerKeys(getTargetStructure(),
                new File((File) context.get(KEY_DIRECTORY),
                    (String) context.get(KEY_LOCAL_NAME)),
                getConfig());

        // delegate to train
        train(document, answerKeys);
    }

    /**
     * {@inheritDoc}
     */
    public void processToken(final Element element, final String left,
            final TokenDetails details, final String right,
            final ContextMap context) throws ProcessingException {
        // update context rep + active classes
        updateState(element, left, details.getToken(), right);
        final String currentType;
        final boolean startOfExtraction =
            locator.startOfExtraction(details.getToken(), details.getRep());
        final boolean endOfExtraction;

        if (startOfExtraction) {
            Util.LOG.debug("Starting extraction (" + details.getToken()
                    + " token)");
        }

        if (locator.inExtraction()) {
            // update extraction (remove current token from remaining tokens)
            boolean updatedExtraction =
                locator.updateExtraction(details.getToken(), details.getRep());

            if (!updatedExtraction) {
                // error: stop handling of current extraction, treat as outside
                currentType = null;
            } else if (startOfExtraction) {
                // first element of current extraction
                // create and store a partial copy of the current extraction
                currentType = locator.getCurrentExtraction().getType();
                final TokenDetails newDetails =
                    new TokenDetails(details.getToken(),
                        locator.getCurrentExtraction().getFirstTokenRep(),
                        locator.getCurrentExtraction().getIndex(), false);
                partialExtraction = new Extraction(currentType, newDetails);
                getPriorRecognitions().add(partialExtraction);
            } else {
                // continuation of current extraction
                currentType = locator.getCurrentExtraction().getType();

                // update partial extraction
                partialExtraction.addToken(details, true);
            }

            // check whether this is the last token of the extraction
            endOfExtraction = locator.endOfExtraction();
        } else {
            // outside any extraction
            currentType = null;
            endOfExtraction = false; // not used
        }

        final CombinationState currentState = (currentType == null)
            ? CombinationState.OUTSIDE
            : new CombinationState(currentType, startOfExtraction,
                endOfExtraction);

        if ((currentType == null) && (partialExtraction != null)
                && (!partialExtraction.isSealed())) {
            // outside: seal last partial extraction
            partialExtraction.setSealed(true);
        }

        // is the token considered relevant?
        boolean relevant = isRelevant(details.getToken());

        if (!relevant && locator.inExtraction()
                && (startOfExtraction || endOfExtraction)) {
            // token is not considered relevant but should be
            markRelevant(details.getToken());
            relevant = true;

            Util.LOG.debug("Marked punctuation token " + details.getToken()
                    + " as relevant since it is the "
                    + (startOfExtraction ? "first" : "last" )
                    + " token of a " + currentType + " extraction");
        }

        // determine input for each classifier, skipping irrelevant punctuation
        if (relevant) {
            final String[] translatedStates =
                getStrategy().translateCurrentState(currentState);
            Util.LOG.debug("Current state: " + currentState
                + "; translated state: '"
                + ArrayUtils.toString(translatedStates) + "'");
            TrainableClassifier[] classifiers = getTrainableClassifiers();
    
            for (int i = 0; i < translatedStates.length; i++) {
                if (!testingOnly) {
                    // train classifiers
                    if (trainingOnlyErrors) {
                        final PredictionDistribution predDist =
                            classifiers[i].trainOnError(getFeatures(),
                                translatedStates[i], getActiveClasses()[i]);
    
                        // update accuracies, logging errors
                        if (predDist == null) {
                            // no training necessary
                            localAccuracies[i].incTrueCount();
                            globalAccuracies[i].incTrueCount();
                        } else {
                            // erroneous prediction
                            localAccuracies[i].incFalseCount();
                            globalAccuracies[i].incFalseCount();
                            final Prediction prediction = predDist.best();
                            Util.LOG.debug(DOMUtils.showToken(element,
                                    details.getToken()) + " classifier " + i
                                    + " misclassified " + prediction
                                    + " instead of " + translatedStates[i]);
                        }
                    } else {
                        classifiers[i].train(getFeatures(),
                                translatedStates[i]);
                        Util.LOG.debug("Trained classifier");
                    }
                } else {
                    // test that combination state can be correctly retranslated
                    final PredictionDistribution[] predDists =
                        new PredictionDistribution[translatedStates.length];
    
                    for (int j = 0; j < predDists.length; j++) {
                        predDists[j] = new PredictionDistribution(
                                new Prediction(translatedStates[j],
                                        new Probability(1.0)));
                    }
    
                    final CombinationState retrans =
                        getStrategy().translateResult(predDists);
    
                    // generous comparison because some strategies cannot handle
                    // end and begin is irrelevant for the null/outside type
                    if ((!StringUtils.equals(currentState.getType(),
                            retrans.getType()))
                            || ((currentState.getType() != null)
                            && (currentState.isBegin() != retrans.isBegin()))) {
                            Util.LOG.error("Error in combination strategy:"
                                + " incorrect re-translation " + retrans
                                + " of current state " + currentState);
                    }
                }
            }

            // update combination strategy
            getStrategy().updateState(currentState);
        } else {
            Util.LOG.debug("Skipping over irrelevant punctuation token "
                    + details.getToken());
        }

        if (locator.endOfExtraction()) {
            // reached the end of the current extraction -- switch to next one
            locator.switchToNextExtraction();
        }
    }

    /**
     * Resets the internal classifer, completely deleting the prediction model.
     * @throws ProcessingException if an error occurs during reset
     */
    public void reset() throws ProcessingException {
        TrainableClassifier[] classifiers = getTrainableClassifiers();
        for (int i = 0; i < classifiers.length; i++) {
            classifiers[i].reset();
        }
    }

    /**
     * Resets the global (overall) accuracies measured so far by each
     * classifier. This can be used to restart accuracy measurements after
     * each round (iteration) of TUNE training, for example. This method
     * is only relevant in TOE (training-only/mainly-errors) mode, otherwise
     * it does nothing.
     */
    public void resetGlobalAccuracy() {
        if (trainingOnlyErrors) {
            globalAccuracies = initAccuracies(PREFIX_GLOBAL_ACC);
        }
    }

    /**
     * Reset the combination strategy, logging a warning if it tells me to
     * discard the last extraction.
     */
    protected void resetStrategy() {        
        final boolean discardLast = getStrategy().reset();
        if (discardLast) {
            Util.LOG.warn("Combination strategy " + getStrategy()
                    + " ordered to discard the last extraction -- "
                    + "this is not supposed to happen when training");
        }       
    }

    /**
     * {@inheritDoc}
     */
    public boolean shouldMatch(final Element element) {
        // ask the embedding elements instance whether this element contains
        // an extraction
        return embeddingElements.containsExtraction(element);
    }

    /**
     * Returns a string representation of this object.
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder result = new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("training only errors", trainingOnlyErrors)
            .append("testing only", testingOnly)
            .append("locator", locator);

        if (getSentenceFilter() != null) {
            result.append("sentence training enabled", sentenceTrainingEnabled);
        }

        return result.toString();
    }

    /**
     * Trains the local classifier with the correct extractions of an XML
     * document, using the provided context representation.
     *
     * @param document a document whose contents should be classified
     * @param correctExtractions a container of all correct extractions for the
     * document
     * @return The token accuracies of each classifier of the trained document
     * if in {@linkplain #isTrainingOnlyErrors() TOE mode}; <code>null</code>
     * otherwise
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    public Accuracy[] train(final Document document,
            final ExtractionContainer correctExtractions)
            throws IOException, ProcessingException {
        // initialize local fields
        initFields();
        locator = new ExtractionLocator(document, correctExtractions,
            getFactory().createTokenizer(""));

        // determine relevant elements if sentence filtering is used
        if (isSentenceFiltering()) {
            embeddingElements = new EmbeddingElements(document,
                    correctExtractions, getFactory());
        }

        // initialize local accuracy if training-only-errors (TOE)
        if (trainingOnlyErrors) {
            localAccuracies = initAccuracies(PREFIX_LOCAL_ACC);
        }

        // the walker will call back (processToken method) where appropriate
        getWalker().walk(document, null);

        // ensure that all extractions have been fully processed
        locator.reachedEndOfDocument();

        if (trainingOnlyErrors) {
            // TOE mode: log + return accuracy in
            Util.LOG.debug("Finished training in TOE mode: "
                + ArrayUtils.toString(localAccuracies) + ", "
                + ArrayUtils.toString(globalAccuracies));
            return localAccuracies;
        } else {
            return null;
        }
    }

    /**
     * Returns a view on the global (overall) accuracies measured so far (or
     * after the last call to {@link #resetGlobalAccuracy()}) by
     * each classifier. This is not a snapshot but will change whenever the
     * underlying values are changed.
     *
     * @return A view on the global accuracies measured so far if in
     * {@linkplain #isTrainingOnlyErrors() TOE mode}; <code>null</code>
     * otherwise
     */
    public AccuracyView[] viewGlobalAccuracy() {
        return globalAccuracies;
    }

}
