/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.extract.amend;

import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.combi.CombinationState;
import de.fu_berlin.ties.text.TokenDetails;

/**
 * Extends the {@link de.fu_berlin.ties.extract.amend.ContextDetails} class by
 * also storing the expected (true) and predicted state on of a token.
 * Instances of this class are immutable and thus thread-safe.
 * 
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/11/19 14:04:51 $, $Author: siefkes $
 */
public class TrainingContextDetails extends ContextDetails {

    /**
     * The expected (true) <code>CombinationState</code> for the token.
     */
    private final CombinationState expectedState;

    /**
     * The <code>CombinationState</code> predicted by the current classification
     * model for the token if there was an classification error; or
     * <code>null</code> if the expected state was correctly predicted.
     */
    private final CombinationState predictedState;


    /**
     * Creates a new instance.
     *
     * @param theToken the token to represent
     * @param tokenRep the repetition of the token in the document (counting
     * starts with 0, as the first occurrence is the "0th repetition")
     * @param tokenIndex the index of the token in the document (counting
     * starts with 0 for the very first token)
     * @param wsBefore whether there is whitespace before the main
     * <code>token</code> (either at the end of <code>left</code> or in the
     * preceding element)
     * @param features a feature vector representing the context of the token
     * @param predicted the <code>CombinationState</code> predicted by the
     * current classification model for this token if there was an
     * classification error; or <code>null</code> if the expected state was
     * correctly predicted
     * @param expected the expected (true) <code>CombinationState</code> for
     * this token
     */
    public TrainingContextDetails(final String theToken, final int tokenRep,
            final int tokenIndex, final boolean wsBefore,
            final FeatureVector features, final CombinationState predicted,
            final CombinationState expected) {
        super(theToken, tokenRep, tokenIndex, wsBefore, features);
        predictedState = predicted;
        expectedState = expected;
    }

    /**
     * Creates a new instance, re-using field values from a {@link TokenDetails}
     * instance.
     *
     * @param orgDetails an instance of the "grandparent" class whose member
     * fields are re-used to initialize this instance
     * @param features a feature vector representing the context of the token
     * @param predicted the <code>CombinationState</code> predicted by the
     * current classification model for this token if there was an
     * classification error; or <code>null</code> if the expected state was
     * correctly predicted
     * @param expected the expected (true) <code>CombinationState</code> for
     * this token
     */
    public TrainingContextDetails(final TokenDetails orgDetails,
            final FeatureVector features, final CombinationState predicted,
            final CombinationState expected) {
        this(orgDetails.getToken(), orgDetails.getRep(), orgDetails.getIndex(),
                orgDetails.isWhitespaceBefore(), features, predicted, expected);
    }


    /**
     * Returns the expected (true) <code>CombinationState</code> for this
     * instance.
     *
     * @return the value of the attribute
     */
    public CombinationState getExpectedState() {
        return expectedState;
    }

    /**
     * Returns the <code>CombinationState</code> predicted by the current
     * classification model for this token if there was an classification error;
     * or <code>null</code> if the expected state was correctly predicted.
     *
     * @return the value of the attribute
     */
    public CombinationState getPredictedState() {
        return predictedState;
    }

}
