/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.filter;

import org.dom4j.Document;
import org.dom4j.Element;

import de.fu_berlin.ties.ProcessingException;

/**
 * Interface for classes that decide whether or not to accept elements in XML
 * documents. This interface cannot extend the {@link org.dom4j.NodeFilter}
 * interface because the {@link #matches(Element)}method is allowed to throw an
 * {@link de.fu_berlin.ties.ProcessingException}.
 * 
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/08/27 15:08:14 $, $Author: siefkes $
 */
public interface ElementFilter {

    /**
     * Resets the filter to start processing a new document. This method must be
     * called for each document to be procesed by this filter prior to calling
     * {@link #matches(Element)}or {@link #prefers(Element)}on elements of
     * this document.
     * 
     * @param document the document to process
     */
    void init(final Document document);

    /**
     * Whether the filter would avoid to filter an element, if possible. Any
     * filter must be able to filter all elements, but some filters like to
     * avoid elements of a certain type, for example.
     * 
     * @param element the element to consider
     * @return <code>true</code> if this would prefer to avoid filtering the
     * element; <code>false</code> otherwise
     */
    boolean avoids(final Element element);

    /**
     * Decides whether an element is accepted by this filter.
     * 
     * @param element the element to test
     * @return <code>true</code> if this filter accepts the element;
     * <code>false</code> otherwise
     * @throws ProcessingException if an error occurs during filtering
     */
    boolean matches(final Element element) throws ProcessingException;

    /**
     * Whether the filter would prefer to filter an element. Any filter must be
     * able to filter all elements, but some filters prefer elements of a
     * certain type, for example.
     * 
     * @param element the element to consider
     * @return <code>true</code> if this would prefer to filter the element;
     * <code>false</code> otherwise
     */
    boolean prefers(final Element element);

}