/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.filter;

import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Document;
import org.dom4j.Element;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.extract.ExtractionContainer;
import de.fu_berlin.ties.extract.ExtractionLocator;
import de.fu_berlin.ties.text.TokenDetails;
import de.fu_berlin.ties.text.TokenizerFactory;
import de.fu_berlin.ties.xml.dom.TokenProcessor;
import de.fu_berlin.ties.xml.dom.TokenWalker;

/**
 * Determines which elements in a document contain extractions (or parts of
 * extractions). This class is used for sentence filtering.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.7 $, $Date: 2004/11/19 14:05:00 $, $Author: siefkes $
 */
public class EmbeddingElements implements TokenProcessor {

    /**
     * The set of elements containing extractions.
     */
    private final Set<Element> elements;

    /**
     * Used to locate extractions.
     */
    private final ExtractionLocator locator;

    /**
     * Creates a new instance.
     *
     * @param document the document to use
     * @param extractions the extractions in this document
     * @param tFactory used to instantiate tokenizers
     */
    public EmbeddingElements(final Document document,
                             final ExtractionContainer extractions,
                             final TokenizerFactory tFactory) {
        super();
        elements = new HashSet<Element>();
        locator = new ExtractionLocator(document, extractions,
                tFactory.createTokenizer(""));

        // walk document to populate elements set
        final TokenWalker walker = new TokenWalker(this, tFactory);

        try {
            // the walker will call back (processToken method) where appropriate
            walker.walk(document, null);
        } catch (IOException ioe) {
            // should never occur since our processToken method doesn't throw
            throw new RuntimeException(ioe);
        } catch (ProcessingException pe) {
            // should never occur since our processToken method doesn't throw
            throw new RuntimeException(pe);
        }

        // ensure that all extractions have been fully processed
        locator.reachedEndOfDocument();
    }

    /**
     * Checks whether the given element contains an extraction (or parts of an
     * extraction).
     *
     * @param element the element to check
     * @return <code>true</code> if the element contains an extraction (fully
     * or partially)
     */
    public boolean containsExtraction(final Element element) {
        return elements.contains(element);
    }

    /**
     * {@inheritDoc}
     */
    public void processToken(final Element element, final String left,
            final TokenDetails details, final String right,
            final ContextMap context) {
        // look for start of extraction
        locator.startOfExtraction(details.getToken(), details.getRep());

        if (locator.inExtraction()) {
            // update extraction (remove current token from remaining tokens)
            final boolean updatedExtraction =
                locator.updateExtraction(details.getToken(), details.getRep());

            if (updatedExtraction) {
                // add this element and all its parents to embedding elements
                // (stop if an element is already known)
                Element currentElement = element;

                while ((currentElement != null)
                        && !elements.contains(currentElement)) {
                    elements.add(currentElement);
                    currentElement = currentElement.getParent();
                }
            }
        }

        if (locator.endOfExtraction()) {
            // reached the end of the current extraction -- switch to next one
            locator.switchToNextExtraction();
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("elements", elements)
            .toString();
    }

}
