/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.filter;

import java.util.Iterator;
import java.util.Set;

import org.dom4j.Element;

import de.fu_berlin.ties.eval.FMetrics;
import de.fu_berlin.ties.eval.FMetricsView;

/**
 * A static class that allows to evaluate the precision, recall, and F1 measure
 * of an {@linkplain de.fu_berlin.ties.filter.ElementFilter element filter}.
 * No instances of this class can be created, only the static members
 * should be used.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/09/01 16:35:45 $, $Author: siefkes $
 */
public final class FilterEvaluator {

    /**
     * Private constructor prevents creation of instances.
     */
    private FilterEvaluator() {
        super();
    }

    /**
     * Evaluates precision, recall, and F1 measure of the last walk performed
     * by a filtering token woken. Delegates to
     * {@link #evaluate(EmbeddingElements, Set, Set)}.
     *
     * @param embeddingElements the correct set of embedding elements
     * @param filteringWalker the filtering token walker to evaluate
     * @return the calculated statistics
     */
    public static FMetricsView evaluate(
            final EmbeddingElements embeddingElements,
            final FilteringTokenWalker filteringWalker) {
        return evaluate(embeddingElements,
                filteringWalker.getAcceptedElements(),
                filteringWalker.getRejectedElements());
    }

    /**
     * Evaluates precision, recall, and F1 measure of an element filter by
     * comparing the accepted and rejected elements of the filter with the
     * correct set of <code>embeddingElements</code>.
     *
     * @param embeddingElements the correct set of embedding elements
     * @param acceptedElements the set of elements accepted by the filter
     * @param rejectedElements the set of elements rejected by the filter
     * @return the calculated statistics
     */
    public static FMetricsView evaluate(
            final EmbeddingElements embeddingElements,
            final Set acceptedElements, final Set rejectedElements) {
        final FMetrics result = new FMetrics();
        Element currentElement;

        /* final Set falsePos = new HashSet(); // only for testing
        final Set falseNeg = new HashSet(); */

        // iterate accepted elements
        final Iterator acceptedIter = acceptedElements.iterator();
        while (acceptedIter.hasNext()) {
            currentElement = (Element) acceptedIter.next();

            if (embeddingElements.containsExtraction(currentElement)) {
                // acceptance was correct: true positive
                result.incTruePos();
            } else {
                // acceptance was erroneous: false positive
                result.incFalsePos();
                //falsePos.add(DOMUtils.showElement(currentElement));
            }
        }

        // iterate rejected elements
        final Iterator rejectedIter = rejectedElements.iterator();
        while (rejectedIter.hasNext()) {
            currentElement = (Element) rejectedIter.next();

            if (embeddingElements.containsExtraction(currentElement)) {
                // reject was erroneous: false negative
                result.incFalseNeg();
                //falseNeg.add(DOMUtils.showElement(currentElement));
            }
            // otherwise reject was correct: true negative (not counted)
        }

        /* if (!falsePos.isEmpty()) {
            Util.LOG.debug("False positives: " + falsePos);
        }
        if (!falseNeg.isEmpty()) {
            Util.LOG.debug("False negatives: " + falseNeg);
        } */

        return result;
    }

}
