/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.filter;

import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Document;
import org.dom4j.Element;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.text.TokenDetails;
import de.fu_berlin.ties.text.TokenizerFactory;
import de.fu_berlin.ties.util.Util;
import de.fu_berlin.ties.xml.dom.DOMUtils;
import de.fu_berlin.ties.xml.dom.TokenProcessor;
import de.fu_berlin.ties.xml.dom.TokenWalker;

/**
 * A token walker that only invokes a provided
 * {@link de.fu_berlin.ties.xml.dom.TokenProcessor} on the subset of tokens
 * that are children of an element accepted by a provided
 * {@link de.fu_berlin.ties.filter.ElementFilter}.
 *
 * <p>Instances of this class are <em>not</em> thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.13 $, $Date: 2004/11/19 14:05:00 $, $Author: siefkes $
 */
public class FilteringTokenWalker extends TokenWalker {

    /**
     * The element filter used by this instance.
     */
    private final ElementFilter filter;

    /**
     * A handler that is called whenever some tokens are skipped; may be
     * <code>null</code>.
     */
    private final SkipHandler skipHandler;

    /**
     * Stores the last decision whether or not to process a token.
     */
    private boolean lastDecision = true;

    /**
     * The elements that so far have been accepted by the filter in the
     * current document.
     */
    private Set<Element> acceptedElements;

    /**
     * The elements that so far have been rejected by the filter in the
     * current document.
     */
    private Set<Element> rejectedElements;

    /**
     * Creates a new instance.
     *
     * @param processor used to process the tokens
     * @param tFactory used to instantiate tokenizers
     * @param elementFilter the element filter to use
     * @param sHandler a handler that is called whenever some tokens are
     * skipped; may be <code>null</code>
     */
    public FilteringTokenWalker(final TokenProcessor processor,
                                final TokenizerFactory tFactory,
                                final ElementFilter elementFilter,
                                final SkipHandler sHandler) {
        super(processor, tFactory);
        filter = elementFilter;
        skipHandler = sHandler;
    }

    /**
     * Returns the set of elements that have been accepted by the filter in the
     * current document.
     *
     * @return the accepted elements
     */
    public Set getAcceptedElements() {
        return acceptedElements;
    }

    /**
     * Returns the set of elements that have been rejected by the filter in the
     * current document.
     *
     * @return the rejected elements
     */
    public Set getRejectedElements() {
        return rejectedElements;
    }

    /**
     * Returns the element filter used by this instance.
     * @return the used element filter
     */
    protected ElementFilter getFilter() {
        return filter;
    }

    /**
     * This method can be overwritten by subclasses to modify decisions of
     * the element filter. The standard behavior is to accept the decision
     * as is.
     *
     * @param element the element to test
     * @param filteredElement the element that was actually filtered
     * (<code>element</code> or a parent), or <code>null</code> if the decision
     * had been cached (no filtering took place)
     * @param decision the decision of the element filer
     * @return the revised decision
     * @throws ProcessingException if an error occurs while revising the
     * decision
     */
    protected boolean handleAccept(final Element element,
            final Element filteredElement, final boolean decision)
    throws ProcessingException {
        return decision;
    }

    /**
     * {@inheritDoc}
     */
    protected void processToken(final Element element, final String left,
            final TokenDetails details, final String right,
            final ContextMap context) throws IOException, ProcessingException {
        // find out whether element is accepted by the filter
        boolean accepted = false;
        boolean foundMatch = false;
        Element currentParent = element;
        Element nearestPreferred = null;  // first element that is preferred
        Element nearestAcceptable = null; // first element that is not avoided

        // did we filter this element or one of its parents already?
        while ((currentParent != null) && !foundMatch) {
            if (acceptedElements.contains(currentParent)) {
                // found accepted
                accepted = true;
                foundMatch = true;
            } else if (rejectedElements.contains(currentParent)) {
                // found rejected
                accepted = false;
                foundMatch = true;
            } else {
                if ((nearestPreferred == null)
                        && (filter.prefers(currentParent))) {
                    // found the nearest parent preferred by the filter --
                    // will use this in case we need to query the filter
                    nearestPreferred = currentParent;
                }
                if ((nearestAcceptable == null)
                        && (!filter.avoids(currentParent))) {
                    // found the nearest parent not avoided by the filter --
                    // will use this in case we need to query the filter and
                    // there is no preferred element
                    nearestAcceptable = currentParent;
                }

                // look at parent element
                currentParent = currentParent.getParent();
            }
        }

        /* Util.LOG.debug((foundMatch ? "Found" : "Didn't find")
                + " match for (parent of) element: "
                + DOMUtils.showElement(element) + "; accepted: " + accepted
                + "; current parent: "
                + DOMUtils.showElement(currentParent)); */

        final Element filteredElement;

        if (!foundMatch) {
            // no match found: need to call the filter

            if (nearestPreferred != null) {
                // filter the first (nearest) element preferred by the filter
                filteredElement = nearestPreferred;

                /*Util.LOG.debug("Filter nearest preferred: "
                        + DOMUtils.showElement(filteredElement)); */
            } else if (nearestAcceptable != null) {
                // filter the first (nearest) element not avoided by the filter
                filteredElement = nearestAcceptable;
                Util.LOG.debug("Filter nearest accepted (no preferred found): "
                        + DOMUtils.showElement(filteredElement));
            } else {
                // no acceptable element -- filter the current element directly
                filteredElement = element;
                Util.LOG.debug("Filter element itself (no preferred or accepted"
                        + " found): " + DOMUtils.showElement(filteredElement));
            }

            // query filter
            accepted = filter.matches(filteredElement);

            // store decision (both for filteredElement and for current element)
            if (accepted) {
                acceptedElements.add(filteredElement);

                if (element != filteredElement) {
                    acceptedElements.add(element);
                }
            } else {
                rejectedElements.add(filteredElement);

                if (element != filteredElement) {
                    rejectedElements.add(element);
               }
           }
        } else {
            // re-use cached decision: no need to filter any
            filteredElement = null;
        }

        // give subclasses a chance to revise the decision
        final boolean finalDecision =
            handleAccept(element, filteredElement, accepted);

        // delegate to super if accepted, discard text otherwise
        if (finalDecision) {
            super.processToken(element, left, details, right, context);
        } else {
            if (lastDecision && (skipHandler != null)) {
                // last token was accepted but this one wasn't --
                // call the skip handler
                skipHandler.skip();
            }
        }

        // update stored decision
        lastDecision = finalDecision;
    }

    /**
     * {@inheritDoc}
     */
    public void walk(final Document document, final ContextMap context)
                           throws IOException, ProcessingException {
        // (re)-initialize maps + filter
        acceptedElements = new HashSet<Element>();
        rejectedElements = new HashSet<Element>();
        filter.init(document);

        // delegate to superclass
        super.walk(document, context);
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder builder = new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("filter", filter);

        if (skipHandler != null) {
            builder.append("skip handler", skipHandler);
        }
        return  builder.toString();
    }

}
