/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.filter;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Document;
import org.dom4j.Element;
import org.dom4j.NodeFilter;

import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.Reranker;
import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.context.LocalFeature;
import de.fu_berlin.ties.context.PriorRecognitions;
import de.fu_berlin.ties.context.Representation;
import de.fu_berlin.ties.extract.Extraction;
import de.fu_berlin.ties.xml.dom.DOMUtils;

/**
 * A trainable filter that uses a
 * {@link de.fu_berlin.ties.context.Representation}to convert elements into
 * feature vectors.
 * 
 * @author Christian Siefkes
 * @version $Revision: 1.9 $, $Date: 2004/11/19 14:05:00 $, $Author: siefkes $
 */
public class RepresentationFilter extends TrainableFilter {

    /**
     * The representation used to convert elements into feature vectors.
     */
    private final Representation representation;

    /**
     * Feature cache used by the representation.
     */
    private final Map<Element, List<LocalFeature>> featureCache =
        new HashMap<Element, List<LocalFeature>>();

    /**
     * The purpose of this filter, e.g. "Sentence" if it is used for sentence
     * filtering.
     */
    private final String purpose;

    /**
     * List of prior recognitions used by the representation.
     */
    private PriorRecognitions priorRecogs;

    /**
     * Creates a new instance.
     * 
     * @param conf used to initialize the trainable classifier by calling {@link
     * de.fu_berlin.ties.classify.TrainableClassifier#createClassifier(
     * java.util.Set, TiesConfiguration)}
     * @param positiveFilter this filter is queried to decide whether this class
     * {@link #prefers(Element) prefers}an element
     * @param negativeFilter this filter is queried to decide whether this class
     * {@link #avoids(Element) avoids}an element
     * @param rerank an optional reranker that recalculates probabilities to
     * introduce a bias (can be used to favor recall over precision, by setting
     * a bias &gt; 1 for the "true" class, etc.); might be <code>null</code>
     * @param rep the representation used to convert elements into feature
     * vectors
     * @param filterPurpose the purpose of this filter, e.g. "Sentence" if it is
     * used for sentence filtering
     * @throws ProcessingException if the initialization of the trainable
     * classifier fails
     */
    public RepresentationFilter(final TiesConfiguration conf,
            final NodeFilter positiveFilter, final NodeFilter negativeFilter,
            final Reranker rerank, final Representation rep,
            final String filterPurpose) throws ProcessingException {
        super(conf, positiveFilter, negativeFilter, rerank);
        representation = rep;
        purpose = filterPurpose;
    }

    /**
     * {@inheritDoc}
     */
    public FeatureVector buildFeatures(final Element element) {
        // collect textual content of element
        final StringBuffer mainText = new StringBuffer();
        DOMUtils.collectText(element, mainText);

        // don't use any left or right text
        return representation.buildContext(element, "", mainText.toString(),
                "", priorRecogs, featureCache, purpose);
    }

    /**
     * {@inheritDoc}
     */
    public void init(final Document document) {
        // reset feature cache + prior recognitions
        featureCache.clear();
        priorRecogs = representation.createRecognitionBuffer();

        // delegate to superclass, just to make sure
        super.init(document);
    }

    /**
     * {@inheritDoc}
     */
    public boolean matches(final Element element) throws ProcessingException {
        // delegate to super class
        final boolean result = super.matches(element);

        // update prior recognitions
        final Extraction newRecognition;

        if (result) {
            // positive result: use element name as type
            final StringBuffer collectedText = new StringBuffer();
            DOMUtils.collectText(element, collectedText);
            newRecognition = new Extraction(DOMUtils.name(element),
                    collectedText.toString());
        } else {
            // negative result: use "<null>" as type and empty string as text
            newRecognition = new Extraction("<null>", "");
        }

        // seal and add recognition
        newRecognition.setSealed(true);
        priorRecogs.add(newRecognition);

        return result;
    }

    /**
     * Returns a string representation of this object.
     * 
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
                .appendSuper(super.toString())
                .append("representation", representation)
                .toString();
    }

}