/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.filter;

import java.util.Collections;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Document;
import org.dom4j.Element;
import org.dom4j.NodeFilter;

import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.classify.PredictionDistribution;
import de.fu_berlin.ties.classify.Reranker;
import de.fu_berlin.ties.classify.TrainableClassifier;
import de.fu_berlin.ties.classify.feature.FeatureVector;
import de.fu_berlin.ties.util.Util;

/**
 * An abstract filter that uses a
 * {@link de.fu_berlin.ties.classify.TrainableClassifier trainable classifier}
 * for training. Subclasses must implement the {@link #buildFeatures(Element)}
 * method to convert elements into feature vectors suitable for the classifier.
 * 
 * <p>
 * Instances of this class are not thread-safe and must be synchronized
 * externally, if required.
 * 
 * @author Christian Siefkes
 * @version $Revision: 1.8 $, $Date: 2004/12/09 18:09:46 $, $Author: siefkes $
 */
public abstract class TrainableFilter implements ElementFilter {

    /**
     * An immutable set of the target classes for the classifier: the strings
     * "false" and "true", in alphabetic order.
     */
    public static final SortedSet<String> BOOLEAN_CLASSES;

    /**
     * Initialize target classes.
     */
    static {
        final SortedSet<String> classes = new TreeSet<String>();
        classes.add(Boolean.FALSE.toString());
        classes.add(Boolean.TRUE.toString());
        BOOLEAN_CLASSES = Collections.unmodifiableSortedSet(classes);
    }

    /**
     * The trainable classifier used by this instance.
     */
    private final TrainableClassifier classifier;

    /**
     * The last element that was classified by this instance, if any.
     */
    private Element lastElement;

    /**
     * The feature vector representing the {@link #lastElement last element
     * classified by this instance}, if any.
     */
    private FeatureVector lastFeatures;

    /**
     * This filter is queried to decide whether this class would like to
     * {@link #avoids(Element) avoid}an element.
     */
    private NodeFilter avoidFilter;

    /**
     * This filter is queried to decide whether this class
     * {@link #prefers(Element) prefers}an element.
     */
    private NodeFilter preferredFilter;

    /**
     * An optional reranker that recalculates probabilities to introduce a bias.
     * This can be used to favor recall over precision (by setting a bias &gt;
     * 1 for the "true" class) etc.
     */
    private final Reranker reranker;

    /**
     * Creates a new instance.
     * 
     * @param conf used to initialize the trainable classifier by calling
     * {@link TrainableClassifier#createClassifier(java.util.Set,
     * TiesConfiguration)}
     * @param positiveFilter this filter is queried to decide whether this class
     * {@link #prefers(Element) prefers}an element
     * @param negativeFilter this filter is queried to decide whether this class
     * {@link #avoids(Element) avoids}an element
     * @param rerank an optional reranker that recalculates probabilities to
     * introduce a bias (can be used to favor recall over precision, by setting
     * a bias &gt; 1 for the "true" class, etc.); might be <code>null</code>
     * @throws ProcessingException if the initialization of the trainable
     * classifier fails
     */
    public TrainableFilter(final TiesConfiguration conf,
            final NodeFilter positiveFilter, final NodeFilter negativeFilter,
            final Reranker rerank) throws ProcessingException {
        super();
        classifier =
            TrainableClassifier.createClassifier(BOOLEAN_CLASSES, conf);
        preferredFilter = positiveFilter;
        avoidFilter = negativeFilter;
        reranker = rerank;
    }

    /**
     * Converts an element into a feature vector to be used by the trainable
     * classifier.
     * 
     * @param element the element to convert
     * @return a feature vector representing the provided element
     */
    public abstract FeatureVector buildFeatures(final Element element);

    /**
     * {@inheritDoc}
     */
    public void init(final Document document) {
        // do nothing -- subclasses can add their own behavior
    }

    /**
     * {@inheritDoc}
     */
    public boolean avoids(final Element element) {
        return avoidFilter.matches(element);
    }

    /**
     * {@inheritDoc}
     */
    public boolean matches(final Element element) throws ProcessingException {
        final FeatureVector features = buildFeatures(element);

        // cache element + features
        lastElement = element;
        lastFeatures = features;

        // invoke classifier + interpret result
        final PredictionDistribution origDist = classifier.classify(features,
                BOOLEAN_CLASSES);

        // rerank results, if supported
        final PredictionDistribution finalDist;
        if (reranker != null) {
            finalDist = reranker.rerank(origDist);
        } else {
            finalDist = origDist;
        }

        final String bestClass = finalDist.best().getType();
        return Util.asBoolean(bestClass);
    }

    /**
     * {@inheritDoc}
     */
    public boolean prefers(final Element element) {
        return preferredFilter.matches(element);
    }

    /**
     * Returns a string representation of this object.
     * 
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
                .append("classifier", classifier)
                .append("preferred filter", preferredFilter)
                .append("reranker", reranker)
                .toString();
    }

    /**
     * Trains the correct decision for an element by calling the
     * {@link TrainableClassifier#trainOnError(FeatureVector, String,
     * java.util.Set)} method on the stored trainable classifier. By using this
     * method, the classifier will train itself only when necessary.
     * 
     * @param element the element to train
     * @param decision the correct decision for the element -- whether or not it
     * should be accepted
     * @return the prediction distribution returned by {@link
     * TrainableClassifier#trainOnError(FeatureVector, String, java.util.Set)}
     * @throws ProcessingException if an error occurs during training
     */
    public PredictionDistribution trainIfNecessary(final Element element,
            final boolean decision) throws ProcessingException {
        final FeatureVector features;

        if (element.equals(lastElement)) {
            // we've seen this element before -- re-use features
            features = lastFeatures;
        } else {
            // build features for element
            features = buildFeatures(element);
        }

        // delegate to classifier and return result
        return classifier.trainOnError(features, Boolean.toString(decision),
                BOOLEAN_CLASSES);
    }

}