/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.filter;

import org.dom4j.Element;

import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.text.TokenizerFactory;
import de.fu_berlin.ties.util.Util;
import de.fu_berlin.ties.xml.dom.DOMUtils;
import de.fu_berlin.ties.xml.dom.TokenProcessor;

/**
 * A filtering token walker that can be trained.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.8 $, $Date: 2004/09/15 15:56:54 $, $Author: siefkes $
 */
public class TrainableFilteringTokenWalker extends FilteringTokenWalker {

    /**
     * Training the embedded filter can be disabled by setting this to
     * <code>false</code>.
     */
    private final boolean trainingEnabled;

    /**
     * Oracle queried to decide which elements should be accepted by the
     * trainable filter.
     */
    private final Oracle oracle;


    /**
     * Creates a new instance, enabling training the filter.
     *
     * @param processor used to process the tokens
     * @param tFactory used to instantiate tokenizers
     * @param elementFilter the trainable element filter to use
     * @param sHandler a handler that is called whenever some tokens are
     * skipped; may be <code>null</code>
     * @param elementOracle oracle queried to decide which elements should be
     * accepted by the trainable filter
     */
    public TrainableFilteringTokenWalker(final TokenProcessor processor,
                                         final TokenizerFactory tFactory,
                                         final TrainableFilter elementFilter,
                                         final SkipHandler sHandler,
                                         final Oracle elementOracle) {
        this(processor, tFactory, elementFilter, sHandler, elementOracle, true);
    }

    /**
     * Creates a new instance.
     *
     * @param processor used to process the tokens
     * @param tFactory used to instantiate tokenizers
     * @param elementFilter the trainable element filter to use
     * @param sHandler a handler that is called whenever some tokens are
     * skipped; may be <code>null</code>
     * @param elementOracle oracle queried to decide which elements should be
     * accepted by the trainable filter
     * @param enableTraining if <code>true</code> the embedded filter is
     * trained from the decisions of the oracle; otherwise the oracle is only
     * queried to log if the filter made a mistake
     */
    public TrainableFilteringTokenWalker(final TokenProcessor processor,
                                         final TokenizerFactory tFactory,
                                         final TrainableFilter elementFilter,
                                         final SkipHandler sHandler,
                                         final Oracle elementOracle,
                                         final boolean enableTraining) {
        super(processor, tFactory, elementFilter, sHandler);
        oracle = elementOracle;
        trainingEnabled = enableTraining;
    }


    /**
     * {@inheritDoc} This implementation relies on the oracle to make the final
     * decision and joins the predicted <code>decision</code> and the correct
     * decision via OR. This allows the next step to view the tokenized text
     * in all necessary cases (if it should view it as determined by the oracle,
     * or it is would view it because of the the trainable classifer's
     * prediction). It also gives the trainable filter to chance to train itself
     * on the correct decision -- even if the original decision was already
     * correct since there are classifiers (e.g. Winnow) that are not purely
     * error-driven but also learn from reinforcement of (some) correct
     * instances.
     */
    protected boolean handleAccept(final Element element,
            final Element filteredElement,  final boolean decision)
    throws ProcessingException {
        final boolean correctDecision = oracle.shouldMatch(element);

        // check prediction if the filter was asked
        if (filteredElement != null) {
            // log if mistake
            if (decision != correctDecision) {
                Util.LOG.debug("Trainable filter predicted " + decision
                        + " instead of " + correctDecision + " for "
                        + DOMUtils.showElement(filteredElement));
            }

            if (trainingEnabled) {
                // give trainable filter a chance to train itself (we know it is
                // trainable because of our constructor)
                final TrainableFilter filter = (TrainableFilter) getFilter();
                filter.trainIfNecessary(filteredElement, correctDecision);
            }
        }

        // allow the next step to view the text in all necessary cases
        return decision || correctDecision;
    }

    /**
     * Returns <code>true</code> if training the embedded filter is enabled
     * (default).
     *
     * @return whether training is enabled
     */
    public boolean isTrainingEnabled() {
        return trainingEnabled;
    }

}
