/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.io;

import java.util.Iterator;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * Convenience implementations of the {@link de.fu_berlin.ties.io.Storable}
 * interface that implements the {@link #toString()} method based on the field
 * map created by {@link de.fu_berlin.ties.io.Storable#storeFields()}.
 *
 * <p>Classes that cannot extend this interface can easily implement this method
 * by delegating to the static {@link #toString(Storable)} method provided by
 * this class, passing itself as argument and returning the result.
 *
 * <p>The {@link java.lang.Object#equals(java.lang.Object)} and
 * {@link java.lang.Object#hashCode()} methods are <em>not</em> overwritten by
 * this class because it would not only be inefficent, but also error-prone
 * to do this based on the field map (e.g. a number could by represented either
 * by a {@link java.lang.Number} object or by a string representation, thus
 * classes might be equal even if their field map serializations are not).
 * Non-abstract {@link de.fu_berlin.ties.io.Storable} implementations should
 * provide their own implementations of these methods, e.g. using the builder
 * tools from <a href="http://jakarta.apache.org/commons/lang/">Jakarta
 * Commons Lang</a>.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/12/06 09:21:18 $, $Author: siefkes $
 */
public abstract class BaseStorable implements Storable {

    /**
     * Returns a String representation of a {@link Storable} object,
     * printing all field name/value pairs in the order used to insert them
     * into the {@link FieldMap}.
     *
     * @param storable the storable to represent
     * @return a string representation of the storable
     */
    public static String toString(final Storable storable) {
        // create builder for this object
        final ToStringBuilder builder = new ToStringBuilder(storable);
        final FieldMap fields = storable.storeFields();
        final Iterator fieldIter = fields.keySet().iterator();
        Object currentKey;
        Object currentValue;

        // append all fields in iteration order
        while (fieldIter.hasNext()) {
            currentKey = fieldIter.next();
            currentValue = fields.get(currentKey);
            builder.append(currentKey.toString(), currentValue);
        }
        return builder.toString();
    }

    /**
     * Creates a new instance.
     */
    protected BaseStorable() {
        super();
    }

    /**
     * Returns a String representation of this object,
     * printing all field name/value pairs in the order used to insert them
     * into the {@link FieldMap}.
     *
     * @return a string representation of this object
     */
    public String toString() {
        // delegate to static method
        return toString(this);
    }

}
