/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.io;

import java.io.File;
import java.net.URL;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.util.Util;

/**
 * A simple wrapper for the MIME type and the character set of a document.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/05/31 19:14:24 $, $Author: siefkes $
 */
public final class ContentType {

    /**
     * Context key recommended to store the MIME type string.
     */
    public static final String KEY_MIME_TYPE = "MIME-Type";

    /**
     * MIME type for HTML.
     */
    public static final String MIME_HTML = "text/html";

    /**
     * MIME type for plain text.
     */
    public static final String MIME_PLAIN = "text/plain";

    /**
     * The regex pattern for parsing the "Content-Type" header field of an URL.
     * The first capturing group captures the MIME type, the second one the
     * character set, if specified.
     */
    private static final Pattern CONTENT_TYPE_PATTERN = Pattern.compile(
        "([^\\s;]+)(?:.*?;\\s*charset\\s*=\\s*\"?([^\\s;\"]*)\"?\\s*;?)?",
        Pattern.CASE_INSENSITIVE);

    /**
     * Determines the content type based on a file name.
     *
     * @param file the file name
     * @param config the configuration to use
     * @return the determined content type
     */
    public static ContentType determineContentType(final File file,
            final TiesConfiguration config) {
        final String extension = (file == null)
            ? "" : IOUtils.getExtension(file);
        return determineContentType(extension, null, config);
    }

    /**
     * Helper method to determines the content type based an "Content-Type"
     * header and/or file name extension. The "Content-Type" header is
     * examined first, if it is specified (not null or empty). If this does
     * not lead to a result, the extension is examined.
     *
     * @param extension the file name extension (might be <code>null</code>)
     * @param contentTypeHeader the "Content-Type" header of an URL, as returned
     * by {@link java.net.URLConnection#getContentType()} (might be
     * <code>null</code>)
     * @param config the configuration to use
     * @return the determined content type
     */
    private static ContentType determineContentType(final String extension,
            final String contentTypeHeader, final TiesConfiguration config) {
        String type = null;
        String charSet = null;

        // parse Content-Type header, if given
        if (StringUtils.isNotEmpty(contentTypeHeader)) {
            final Matcher contentTypeMatcher =
                CONTENT_TYPE_PATTERN.matcher(contentTypeHeader);
            if (contentTypeMatcher.find()) {
                type = contentTypeMatcher.group(1);
                charSet = contentTypeMatcher.group(2);
            } else {
                // otherwise it's not a valid MIME type
                Util.LOG.warn("Not a valid Content-Type header: '"
                        + contentTypeHeader + "'");
            }
        }

        // check extension if no valid Content-Type header given
        if (type == null) {
            // convert extension to lower case
            final String normalizedExtension = (extension == null)
                ? "" : extension.toLowerCase();

            if (!"".equals(normalizedExtension)) {
                // check config for extension, using null as default value
                type = config.getString(TiesConfiguration.joinKey("ext",
                                normalizedExtension), null);
            }
        }

        // check config to replace alternative MIME type by main one
        if (type != null) {
            final String mimeLookupKey =
                TiesConfiguration.joinKey("mime", type);
            if (config.containsKey(mimeLookupKey)) {
                type = config.getString(mimeLookupKey);
            }
        }

        return new ContentType(type, charSet);
    }

    /**
     * Helper method to determines the content type based an "Content-Type"
     * header and/or URL extension. The "Content-Type" header is
     * examined first, if it is specified (not null or empty). If this does
     * not lead to a result, the extension of the URL is examined.
     *
     * @param url the URL (might be <code>null</code>)
     * @param contentTypeHeader the "Content-Type" header of an URL, as returned
     * by {@link java.net.URLConnection#getContentType()} (might be
     * <code>null</code>)
     * @param config the configuration to use
     * @return the determined content type
     */
    public static ContentType determineContentType(final URL url,
            final String contentTypeHeader, final TiesConfiguration config) {
        final String extension = (url == null)
            ? "" : IOUtils.getExtension(url);
        return determineContentType(extension, contentTypeHeader, config);
    }


    /**
     * The character set.
     */
    private final String charset;

    /**
     * The MIME type, <code>null</code> if unknown.
     */
    private final String mimeType;

    /**
     * Non-public constructor for internal use. Use the factory methods instead.
     *
     * @param type the MIME type of this document, <code>null</code> if unknown;
     * will be converted to lower-case
     * @param characterSet the character set of this document,
     * <code>null</code> if unknown
     */
    private ContentType(final String type, final String characterSet) {
        super();
        mimeType = StringUtils.lowerCase(type);
        charset = characterSet;
    }

    /**
     * Returns the character set.
     *
     * @return the character set, or <code>null</code> if unknown
     */
    public String getCharset() {
        return charset;
    }

    /**
     * Returns the MIME type.
     *
     * @return the MIME type, or <code>null</code> if unknown
     */
    public String getMimeType() {
        return mimeType;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("mime type", mimeType)
            .append("charset", charset)
            .toString();
    }

}
