/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;

import de.fu_berlin.ties.text.TextUtils;
import de.fu_berlin.ties.util.Util;

/**
 * A field container that stores and processed its contents as values separated
 * by a pipe character ('|'). Entries are separated by newlines.
 *
 * <p>The very first non-comment line contains the field names, separated by
 * pipes. Each further non-comment lines contains the values of a entry
 * ({@link de.fu_berlin.ties.io.FieldMap}). The first non-comment line must
 * contain as least as most pipes as each further line, so the field names
 * are known for all fields of all field maps. Empty fields are ignored when
 * populating field maps.
 *
 * <p>The '#' character can be used to introduce comment lines.
 * Comments and lines that are empty or contain only whitespace are ignored
 * when reading data.
 *
 * <p>Pipe characters and newlines within fields and '#' at the begin of the
 * first field of an entry are escaped with a backslash.
 * The recommended character set for storing delimiter-separated values is
 * <strong>UTF-8</strong>.
 *
 * <p>Floats and doubles are formatted (rounded) via
 * {@link de.fu_berlin.ties.util.Util#format(double)}.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.4 $, $Date: 2004/09/06 17:23:31 $, $Author: siefkes $
 */
public class DelimSepValues extends FieldContainer {

    /**
     * The recommended file extension for this format: {@value}
     * (delimiter-separated values).
     */
    public static final String FILE_EXT = "dsv";

    /**
     * The character introducing a comment: '#'.
     */
    public static final char COMMENT_START = '#';

    /**
     * The escape character: '\' (a backslash).
     */
    public static final char ESCAPE = '\\';

    /**
     * The escape character as a regular expression.
     */
    private static final Pattern ESCAPE_PATTERN = Pattern.compile("\\\\");

    /**
     * The field delimiter character: '|' (a pipe).
     */
    public static final char DELIM = '|';

    /**
     * The field delimiter character as a regular expression.
     */
    private static final Pattern DELIM_PATTERN = Pattern.compile("\\|");

    /**
     * String specifying the replacement to use for the
     * {@link #globalEscapeMatcher}: prepend the escape character to the match.
     */
    private static final String GLOBAL_REPLACEMENT =
        ESCAPE_PATTERN.pattern() + "$0";

    /**
     * Matches strings that must be escaped anywhere in fields
     * (pipe characters and newlines).
     */
    private final Matcher globalEscapeMatcher = Pattern.compile("(?:"
            + DELIM_PATTERN.pattern() + '|' + TextUtils.NEWLINE_ALTERNATIVES
            + ')').matcher("");

    /**
     * Matches strings that must be escaped when writing the begin
     * of a non-comment line: the comment start character, optionally preceded
     * by any number of backslashes. Escaped by prepending a backslash.
     */
    private final Matcher comstartEscapeMatcher = Pattern.compile(
        ESCAPE_PATTERN.pattern() + "*" + COMMENT_START).matcher("");

    /**
     * Matches strings that must be unescaped when reading the
     * begin of a non-comment line: the comment start character preceded by one
     * or more any number of backslashes. Unescaped by removing the first
     * backslash.
     */
    private final Matcher comstartUnescapeMatcher = Pattern.compile(
        ESCAPE_PATTERN.pattern() + "+" + COMMENT_START).matcher("");

    /**
     * Creates a new empty instance.
     */
    public DelimSepValues() {
        super();
    }

    /**
     * Creates a new instance from serialized delimiter-separated values.
     *
     * @param input the input data to process
     * @throws IllegalArgumentException if the input data contains errors,
     * e.g. when there are insufficient field names given in the first
     * non-comment line
     */
    public DelimSepValues(final CharSequence input)
            throws IllegalArgumentException {
        this();
        // split into into lines
        String[] lines = TextUtils.splitLinesExact(input);
        boolean readFieldNames = false;
        String entry;
        String[] orgFields;
        String currentField;
        int fIndex;
        List<String> fieldList;

        for (int i = 0; i < lines.length; i++) {
            entry = lines[i];

            // jump over comment lines and blank lines
            if (StringUtils.isNotBlank(entry)
                    && (entry.charAt(0) != COMMENT_START)) {
                comstartUnescapeMatcher.reset(entry);

                if (comstartUnescapeMatcher.lookingAt()) {
                    // found escaped '#': remove the initial backslash
                    entry = lines[i].substring(1);
                } else {
                    entry = lines[i];
                }

                // join with next line if ending in odd number of '\'
                while ((TextUtils.countLast(entry, ESCAPE) % 2) == 1) {
                    i++;
                    if (i >= lines.length) {
                        throw new IllegalArgumentException(
                                "Premature end of entry: " + entry);
                    }

                    // replace last backslash by newline + append next line
                    entry = entry.substring(0, entry.length() - 1)
                        + TextUtils.LINE_SEPARATOR + lines[i];
                    Util.LOG.debug("Joint lines to handle escape: " + entry);
                }

                // split line on pipes
                orgFields = DELIM_PATTERN.split(entry);
                fieldList = new ArrayList<String>(orgFields.length);

                // join field with next one if ending in odd number of '\'
                // (we know the last one doesn't because we checked above)
                for (fIndex = 0; fIndex < orgFields.length; fIndex++) {
                    currentField = orgFields[fIndex];

                    while ((TextUtils.countLast(currentField, ESCAPE) % 2)
                            == 1) {
                        fIndex++;

                        // replace last backslash by '|' + append next field
                        currentField = currentField.substring(0,
                                entry.length() - 1) + DELIM + orgFields[fIndex];
                        Util.LOG.debug("Joint fields to handle escape: "
                            + currentField);
                    }

                    fieldList.add(currentField);
                }

                if (readFieldNames) {
                    // create new field map
                    add(fieldList);
                } else {
                    // first line: field headers
                    for (fIndex = 0; fIndex < fieldList.size(); fIndex++) {
                        addKey(fieldList.get(fIndex));
                    }
                    readFieldNames = true;
                }
            }
        }
    }

    /**
     * Creates a new instance from serialized delimiter-separated values.
     *
     * @param in a stream containing the input data to process, must use the
     * UTF-8 charset; the stream is not closed by this method
     * @throws IOException if an I/O error occurs while reading from the stream
     * @throws IllegalArgumentException if the input data contains errors,
     * esp. when there are insufficient field names given in the first
     * non-comment line
     */
    public DelimSepValues(final InputStream in)
            throws IOException, IllegalArgumentException {
        // read data with UTF-8 charset
        this(new InputStreamReader(in, IOUtils.STANDARD_UNICODE_CHARSET));
    }

    /**
     * Creates a new instance from serialized delimiter-separated values.
     *
     * @param reader a reader containing the input data to process; not closed
     * by this method
     * @throws IOException if an I/O error occurs while reading
     * @throws IllegalArgumentException if the input data contains errors,
     * esp. when there are insufficient field names given in the first
     * non-comment line
     */
    public DelimSepValues(final Reader reader)
            throws IOException, IllegalArgumentException {
        this(IOUtils.readToString(reader));
    }

    /**
     * Creates a new instance and populates it from a {@link StorableContainer}.
     *
     * @param contents the contents to add by calling
     * {@link StorableContainer#storeEntries(FieldContainer)}
     */
    public DelimSepValues(final StorableContainer contents) {
        super(contents);
    }

    /**
     * Serializes contents as delimiter-separated values.
     *
     * @param writer the writer to write to; flushed but not closed by this
     * method
     * @throws IOException if an I/O error occurs while writing to the stream
     */
    public void store(final Writer writer) throws IOException {
        // print headers
        Iterator<String> keyIter = keyIterator();
        storeEntry(keyIter, writer);

        // print values of each field map
        final Iterator mapsIter = entryIterator();
        FieldMap currentMap;
        List<Object> currentValues;
        String currentKey;

        while (mapsIter.hasNext()) {
            currentMap = (FieldMap) mapsIter.next();

            // create list of values (might be null) for each globally known key
            // using an array list because we know the number of keys to expect
            currentValues = new ArrayList<Object>(keyCount());
            keyIter = keyIterator();

            while (keyIter.hasNext()) {
                currentKey = keyIter.next();
                // storing value for key (might be null)
                currentValues.add(currentMap.get(currentKey));
            }

            // remove trailing null values (no need to print unnecessary tabs)
            while ((currentValues.size() > 0)
                    && (currentValues.get(currentValues.size() - 1) == null)) {
                currentValues.remove(currentValues.size() - 1);
            }

            // print values of field map
            storeEntry(currentValues.iterator(), writer);
        }

        writer.flush();
    }

    /**
     * Helper method that serializes an entry as a line of delimiter-separated
     * values. Tabs are used to separate fields; a newline is printed after
     * the last fields. Escapes are added as required.
     * <code>null</code> fields are handled by just adding a tab.
     *
     * @param iter an iterator over the fields to add
     * @param writer the writer to write to
     * @throws IOException if an I/O error occurs while writing
     */
    private void storeEntry(final Iterator iter, final Writer writer)
            throws IOException {
        Object rawItem;
        Number numericItem;
        String item;
        boolean isFirst = true;

        while (iter.hasNext()) {
            rawItem = iter.next();

            if (rawItem == null) {
                // convert null to empty string
                item = "";
            } else if ((rawItem instanceof Double)
                    || (rawItem instanceof Float)) {
                // use Util method to format/round floats + doubles
                numericItem = (Number) rawItem;
                item = Util.format(numericItem.doubleValue());
            } else {
                // convert via toString method
                item = rawItem.toString();
            }

            if (isFirst) {
                comstartEscapeMatcher.reset(item);

                if (comstartEscapeMatcher.lookingAt()) {
                    // prepend backslach to escape comment start character
                    item = ESCAPE + item;
                }
                isFirst = false;
            } else {
                // write pipe preceding item
                writer.write(DELIM);
            }

            // no need to bother with an empty string
            if (item.length() > 0) {
                // escape any pipe chars + newlines in item prior to writing
                item = TextUtils.replaceAll(item, globalEscapeMatcher,
                        GLOBAL_REPLACEMENT);
                writer.write(item);
            }
        }

        // print ending newline
        writer.write(TextUtils.LINE_SEPARATOR);
    }

}
