/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.io;

import java.io.File;
import java.io.FileFilter;
import java.util.Collections;
import java.util.Set;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * A file filter that checks the extension to decide whether to accept a file.
 * Don't prepend a dot when specifying extensions -- this happens automatically.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/09/06 17:23:31 $, $Author: siefkes $
 */
public class ExtensionFilter implements FileFilter {

    /**
     * Whether case is ignored when matching the extension.
     */
    private final boolean caseIgnored;

    /**
     * The immutable set of extensions (Strings) to accept.
     */
    private final Set extensions;

    /**
     * Creates a new instance that accepts only a single extension.
     *
     * @param extension the extension to accept
     * @param ignoreCase whether case is ignored when matching the extension;
     * if <code>true</code>, <code>extension</code> must be lower-case
     * or matching will fail (file extensions are converted to lower-case
     * prior to comparing)
     */
    public ExtensionFilter(final String extension, final boolean ignoreCase) {
        this(Collections.singleton(extension), ignoreCase);
    }

    /**
     * Creates a new instance.
     *
     * @param extensionSet the set of extensions (Strings) to accept
     * @param ignoreCase whether case is ignored when matching the extension;
     * if <code>true</code>, <code>extensionSet</code> must contain only
     * lower-case entries or matching will fail (file extensions are converted
     * to lower-case, but the contents of the set are used as is)
     */
    public ExtensionFilter(final Set<String> extensionSet,
            final boolean ignoreCase) {
        super();
        caseIgnored = ignoreCase;

        // make set immutable
        extensions = Collections.unmodifiableSet(extensionSet);
    }

    /**
     * Tests whether or not the specified abstract pathname should be included
     * in a pathname list.
     *
     * @param pathname the abstract pathname to be tested
     * @return <code>true</code> iff the {@linkplain IOUtils#getExtension(File)
     * extension} of the <code>pathname</code> is contained in the
     */
    public boolean accept(final File pathname) {
        String extension = IOUtils.getExtension(pathname);

        if (caseIgnored) {
            // convert to lower case
            extension = extension.toLowerCase();
        }

        return extensions.contains(extension);
    }

    /**
     * The immutable set of extensions (Strings) accepted by this filter.
     * @return the value of the attribute
     */
    public Set getExtensions() {
        return extensions;
    }

    /**
     * Whether case is ignored when matching the extension.
     * If <code>true</code>, file extensions are converted to lower case prior
     * to comparing them.
     *
     * @return the value of the attribute
     */
    public boolean isCaseIgnored() {
        return caseIgnored;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("extension set", extensions)
            .append("ignore case", caseIgnored)
            .toString();
    }

}
