/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.text.TextUtils;

/**
 * A container of {@link de.fu_berlin.ties.io.FieldMap}s. A container stores all
 * field maps added to itself and keeps a set of all keys found in the field
 * maps. This is an "append-only" container; field maps can only be added but
 * never removed.
 *
 * <p>Instances of this class are not thread-safe and must be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.16 $, $Date: 2004/11/09 10:11:29 $, $Author: siefkes $
 */
public class FieldContainer {

    /**
     * Factory method that creates a field container in {@link DelimSepValues}
     * format.
     *
     * @return the created container
     */
    public static FieldContainer createFieldContainer() {
        return new DelimSepValues();
    }

    /**
     * Factory method that creates a field container from serialized data in
     * {@link DelimSepValues} format.
     *
     * @param input the input data to process
     * @throws IllegalArgumentException if the input data contains errors
     * @return the created container
     */
    public static FieldContainer createFieldContainer(
            final CharSequence input) throws IllegalArgumentException {
        return new DelimSepValues(input);
    }

    /**
     * Factory method that creates a field container from serialized data in
     * {@link DelimSepValues} format. Internally uses
     * {@link IOUtils#openCompressableInStream(InputStream)} so
     * <code>gzip</code>-compressed streams are also supported.
     *
     * @param in a stream containing the input data to process, must use the
     * UTF-8 charset; the stream is not closed by this method
     * @throws IOException if an I/O error occurs while reading from the stream
     * @throws IllegalArgumentException if the input data contains errors
     * @return the created container
     */
    public static FieldContainer createFieldContainer(final InputStream in)
            throws IOException, IllegalArgumentException {
        // autodetect compression if used
        final InputStream uncompressedStream =
            IOUtils.openCompressableInStream(in);

        // read data in UTF-8 charset
        return createFieldContainer(new InputStreamReader(uncompressedStream,
                IOUtils.STANDARD_UNICODE_CHARSET));
    }

    /**
     * Factory method that creates a field container from serialized data in
     * {@link DelimSepValues} format.
     *
     * @param reader a reader containing the input data to process; not closed
     * by this method
     * @throws IOException if an I/O error occurs while reading from the stream
     * @throws IllegalArgumentException if the input data contains errors
     * @return the created container
     */
    public static FieldContainer createFieldContainer(final Reader reader)
            throws IOException, IllegalArgumentException {
        return createFieldContainer(IOUtils.readToString(reader));
    }

    /**
     * Returns the file extension recommended for {@link FieldContainer}s.
     *
     * @return the recommended extension: {@link DelimSepValues#FILE_EXT}
     */
    public static String recommendedExtension() {
        return DelimSepValues.FILE_EXT;
    }

    /**
     * This map can be used to extend newly added field maps: at each
     * {@link #add(FieldMap)} operation, any key/value pairs from this map are
     * added to field map prior to storing it. The map is empty by default.
     */
    private final FieldMap backgroundMap = new FieldMap();

    /**
     * A map of attributes stored in this container.
     */
    private final FieldMap attributes = new FieldMap();

   /**
    * A map of nested containers stored in this container.
    */
   private final LinkedHashMap<String, FieldContainer> nested =
       new LinkedHashMap<String, FieldContainer>();

    /**
     * A list of all the field maps stored in this container.
     */
    private final LinkedList<FieldMap> entries = new LinkedList<FieldMap>();

    /**
     * A set containing all keys from the stored field maps.
     */
    private final LinkedHashSet<String> keySet = new LinkedHashSet<String>();

    /**
     * Creates a new empty instance.
     */
    public FieldContainer() {
        super();
    }

    /**
     * Creates a new instance and populates it from a {@link StorableContainer}.
     *
     * @param contents the contents to add by calling
     * {@link StorableContainer#storeEntries(FieldContainer)}
     */
    public FieldContainer(final StorableContainer contents) {
        this();
        contents.storeEntries(this);
    }

    /**
     * Helper method that adds a key to the set of all keys.
     *
     * @param key the key to add
     * @return <code>true</code> if the set did not already contain the
     * specified element
     */
    protected boolean addKey(final String key) {
        return keySet.add(key);
    }

    /**
     * Adds a field map to this container. Any new keys in this set are added
     * at the end of the set of all keys, in the order of the field map's key
     * set.
     *
     * @param map the field map to add
     */
    public void add(final FieldMap map) {
        // we want to check keys
        add(map, true);
    }

    /**
     * Adds the representation of a {@link Storable} to this container, by
     * calling its {@link Storable#storeFields()} method and adding the
     * resulting field map.
     *
     * @param storable the storable whose fields to add
     */
    public void add(final Storable storable) {
        add(storable.storeFields());
    }

    /**
     * Helper method for adding a field map to this container.
     *
     * @param map the field map to add
     * @param checkKeys whether to check the keys of the field map and add any
     * missing keys to the set of all keys; subclasses calling this method can
     * set this to <code>false</code> iff they know for sure that the map does
     * not contain any new keys
     */
    protected void add(final FieldMap map, final boolean checkKeys) {
        // add entries from the background map, if any
        if (!backgroundMap.isEmpty()) {
            map.putAll(backgroundMap);
        }

        entries.add(map);

        if (checkKeys) {
            // append any new keys at the end of the key set, in iteration order
            keySet.addAll(map.keySet());
        }
    }

    /**
     * Adds a field map created from the specified values, using the
     * <em>n</em>-th key from the set of all keys for the <em>n</em>-th
     * specified value. Empty strings and <code>null</code> values are omitted
     * (the corresponding key is skipped).
     *
     * @param values the list of values to store in the field map
     * @throws IllegalArgumentException if the specified array contains more
     * elements than the set of all keys
     */
    public void add(final List values) throws IllegalArgumentException {
        if (values.size() > keySet.size()) {
            throw new IllegalArgumentException(
                "Cannot create FieldMap: more values (" + values.size()
                + ") than keys (" + keySet.size() + ")");
        }

        // create and populate map
        final FieldMap fieldMap = new FieldMap();
        final Iterator<String> keyIter = keyIterator();
        final Iterator valueIter = values.iterator();
        String currentKey;
        Object currentValue;

        while (valueIter.hasNext()) {
            // we checked above that the key iterator has enough elements
            currentKey = keyIter.next();
            currentValue = valueIter.next();

            // skip null and empty strings
            if ((currentValue != null) && (!"".equals(currentValue))) {
                fieldMap.put(currentKey, currentValue);
            }
        }

        // no need to check the keys, as they are taken from the set
        add(fieldMap, false);
    }

    /**
     * Returns the number of attributes stored in this container.
     *
     * @return the number of attributes
     */
    public int attributeCount()  {
        return attributes.size();
    }

    /**
     * Returns an iterator over the names of the attributes stored in this
     * container.
     *
     * @return an iterator over the attribute names
     */
    public Iterator<String> attributeIterator() {
        return attributes.keySet().iterator();
    }

    /**
     * This map can be used to extend newly added field maps: at each
     * {@link #add(FieldMap)} operation, any key/value pairs from this map are
     * added to field map prior to storing it. The map is empty by default.
     *
     * @return the background map
     */
    public FieldMap backgroundMap() {
        return backgroundMap;
    }

    /**
     * Creates and returns a new nested subcontainer.
     *
     * @param name the name of the new subcontainer
     * @return the newly created nested container
     * @throws IllegalArgumentException if the specified name contains
     * whitespace or is null or empty or if a nested container with the
     * given name already exists
     */
    public FieldContainer createNestedContainer(final String name)
    throws IllegalArgumentException {
        TextUtils.ensurePrintableName(name);
        if (nested.containsKey(name)) {
            throw new IllegalArgumentException("Nested container " + name
					       + " already exists");
	}
        final FieldContainer subContainer = new FieldContainer();
        nested.put(name, subContainer);
        return subContainer;
    }

    /**
     * Creates (deserializes) an list of objects of a specified type by calling
     * {@link FieldMap#createObject(Class)} for each of the field maps contained
     * in this container. The contained field maps are deserialized in the order
     * they were added to this container.
     *
     * @param type the class of the objects to create; must have a constructor
     * whose only argument is a <code>FieldMap</code>
     * @return a list of the created object; all elements of this list will be
     * instances of the specified class
     * @throws InstantiationException if instantiation failed
     * @throws SecurityException if access to the required reflection
     * information is denied
     */
    public List<Object> createObjects(final Class type)
    throws InstantiationException, SecurityException {
        final List<Object> result = new LinkedList<Object>();
        final Iterator entryIter = entryIterator();
        FieldMap currentMap;

        // create and store an object from each field map
        while (entryIter.hasNext()) {
            currentMap = (FieldMap) entryIter.next();
            result.add(currentMap.createObject(type));
        }
        return result;
    }

    /**
     * Returns an iterator over the {@link FieldMap}s in this container in the
     * order they were added.
     *
     * @return an iterator over the contained field maps
     */
    public Iterator<FieldMap> entryIterator() {
        return entries.iterator();
    }

   /**
     * Returns the value of an attribute.
     *
     * @param name the name of the attribute to look up
     * @return the value of the attribute; or <code>null</code> if the
     * attribute doesn't exist
     */
    public Object getAttribute(final String name) {
        return attributes.get(name);
    }

   /**
    * Returns a nested subcontainer managed by this instance.
    *
    * @param name the name of the container to retrieve
    * @return the nested container; or <code>null</code> if there is no
    * container with this name
    */
    public FieldContainer getNestedContainer(final String name) {
        return nested.get(name);
    }

    /**
     * Returns the number of keys in this container.
     *
     * @return the number of keys, i.e. the size of the set of all keys
     */
    public int keyCount() {
        return keySet.size();
    }

    /**
     * Returns an iterator over the set of all keys used in contained field
     * maps.
     *
     * @return an iterator over the set of all keys
     */
    public Iterator<String> keyIterator() {
        return keySet.iterator();
    }

   /**
     * Returns the number of nested containers managed by container.
     *
     * @return the number of nested containers
     */
    public int nestedCount() {
        return nested.size();
    }
   
    /**
     * Returns an iterator over the names of nested containers managed by this
     * container.
     *
     * @return an iterator over the names of nested containers
     */
    public Iterator<String> nestedIterator() {
        return nested.keySet().iterator();
    }
   
    /**
     * Returns the number of entries stored in this container.
     *
     * @return the number of entries
     */
    public int size() {
        return entries.size();
    }

    /**
     * Serializes contents by wrapping the stream in a writer with UTF-8
     * character set and delegating to {@link #store(Writer)}.
     *
     * @param out the output stream to write to; flushed but not closed by this
     * method
     * @throws IOException if an I/O error occurs while writing to the stream
     */
    public void store(final OutputStream out) throws IOException {
        // wrap in writer with UTF-8 charset
        store(new OutputStreamWriter(out, IOUtils.STANDARD_UNICODE_CHARSET));
    }

    /**
     * Sets name and value of an attribute.
     *
     * @param name the name of the attribute, must not contain whitespace
     * @param value to value of the attribute
     * @return the previous value associated with the specified attribute name,
     * or <code>null</code> if there was no previous value
     * @throws IllegalArgumentException if the specified name contains
     * whitespace or is null or empty
     */
    public Object setAttribute(final String name, final Object value)
    throws IllegalArgumentException {
        TextUtils.ensurePrintableName(name);
        return attributes.put(name, value);
    }

    /**
     * Subclasses can overwrite this method to serialize their contents
     * in a class-specific format. This class does not prescribe a specific
     * format and thus cannot store the data, throwing an
     * {@link UnsupportedOperationException} instead.
     *
     * @param writer the writer to write to; not closed by this method
     * @throws IOException might be thrown by subclasses if an I/O error occurs
     * while serializing the data
     * @throws UnsupportedOperationException always thrown by instances of this
     * class
     */
    public void store(final Writer writer)
            throws IOException, UnsupportedOperationException {
        throw new UnsupportedOperationException(
                "Storing is not supported by FieldContainer");
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("key set", keySet)
            .toString();
    }

}
