/*
 * Copyright (C) 2003-2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.io;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * A map targeted at serialization and deserialiation of objects in
 * human-readable formats. It that allows creating (deserializing) an object
 * by passing itself to the constructor.
 *
 * <p>This class preserves the guarantees of {@link java.util.LinkedHashMap}:
 * it allows both efficient retrieval of values by key/field name (due to the
 * contained hash map) and iteration of key/value pairs in order of insertion
 * (due to the contained linked list).
 *
 * <p><strong>Note that this implementation is not synchronized.</strong>
 * If multiple threads access a field map concurrently, and at least one of the
 * threads modifies the map structurally, it <em>must</em> be synchronized
 * externally. This is typically accomplished by synchronizing on some object
 * that naturally encapsulates the map. If no such object exists, the map
 * should be "wrapped" using the
 * {@link java.util.Collections#synchronizedMap(java.util.Map)} method.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/09/06 17:23:31 $, $Author: siefkes $
 */
public class FieldMap extends LinkedHashMap<String, Object> {

    /**
     * Creates a new instance.
     */
    public FieldMap() {
        super();
    }

    /**
     * Creates a new instance with the same mappings as the specified map.
     * @param map the map whose mappings are to be placed in this map
     * @throws NullPointerException if the specified map is null
     */
    public FieldMap(final Map<String, Object> map) throws NullPointerException {
        super(map);
    }

    /**
     * Creates (deserializes) an object of a specified type by calling a
     * constructor of the class that accepts a field map as single argument
     * and passing itself as parameter. Only classes that provide a suitable
     * constructor can be instantiated this way.
     *
     * @param type the class of the object to create; must have a constructor
     * whose only argument is a <code>FieldMap</code>
     * @return the created object; the returned object will be an instance of
     * the specified class
     * @throws InstantiationException if instantiation failed
     * @throws SecurityException if access to the required reflection
     * information is denied
     */
    public Object createObject(final Class type) throws InstantiationException,
            SecurityException {
        try {
            // load constructor with FieldMap as only argument
            final Constructor cons = type.getConstructor(
                    new Class[] {FieldMap.class});
            // pass myself as single argument
            return cons.newInstance(new Object[] {this});
        } catch (IllegalAccessException iae) {        // repackage exception
            final InstantiationException ie =
                new InstantiationException(iae.toString());
            ie.initCause(iae);
            throw ie;
        } catch (InvocationTargetException ite) {     // repackage exception
            final InstantiationException ie =
                new InstantiationException(ite.toString());
            ie.initCause(ite);
            throw ie;
        } catch (ExceptionInInitializerError eiie) {  // repackage exception
            final InstantiationException ie =
                new InstantiationException(eiie.toString());
            ie.initCause(eiie);
            throw ie;
        } catch (NoSuchMethodException nsme) {        // repackage exception
            final InstantiationException ie =
                new InstantiationException(nsme.toString());
            ie.initCause(nsme);
            throw ie;
        }
    }

}
