/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.io;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PushbackInputStream;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.lang.StringUtils;

import de.fu_berlin.ties.TextProcessor;
import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.text.TextUtils;
import de.fu_berlin.ties.util.MutableInt;
import de.fu_berlin.ties.util.Util;

/**
 * A static class that provides utility constants and methods for I/O.
 * No instances of this class can be created, only the static members
 * should be used.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.21 $, $Date: 2004/12/09 18:10:32 $, $Author: siefkes $
 */
public final class IOUtils {

    /**
     * Configuration and context key: the character set to use when reading and
     * writing local files. If omitted, the default charset of the current
     * platform is used.
     */
    public static final String KEY_LOCAL_CHARSET = "charset";

    /**
     * Configuration key: whether to compress your data in <code>gzip</code>
     * format. Used by
     * {@link #openCompressableOutStream(OutputStream, TiesConfiguration)}.
     */
    public static final String CONFIG_COMPRESS_GZIP = "compress.gzip";

    /**
     * The standard block size recommended for I/O.
     */
    public static final int STANDARD_BLOCK_SIZE = 64 * 1024;

    /**
     * The standard character set defined by the HTTP standard.
     */
    public static final String STANDARD_HTTP_CHARSET = "ISO-8859-1";

    /**
     * The standard Unicode-compatible character set.
     */
    public static final String STANDARD_UNICODE_CHARSET = "UTF-8";

    /**
     * The character separating the base name from the extension of a file
     * (a dot).
     */
    public static final char EXT_SEPARATOR = '.';

    /**
     * The maximum BOM size.
     */
    private static final int BOM_SIZE = 4;

    /**
     * Default directory that can be given as "parent" when
     * {@linkplain File#File(java.io.File, java.lang.String) constructing} file
     * objects. Initially set to <code>null</code> so the "system-dependent
     * default directory" (typically the working directory) will be used.
     * Synchronized via the {@link #DD_GUARD}.
     */
    private static File defaultDirectory = null;

    /**
     * Used to guard synchronization of the {@link #defaultDirectory}.
     */
    private static final Object DD_GUARD = new Object();

    /**
     * Creates a file for writing output to. Delegates to
     * {@link #createOutFile(File, String, String)} without specifying an
     * extension.
     *
     * @param directory the directory in which the file should be created;
     * if <code>null</code> or not writable, the file is created in the
     * current working directory instead
     * @param localName the local name of the input file --- the extension of
     * this file (if any) is replaced by the specified output extension
     * @return the create file
     */
    public static File createOutFile(final File directory, final
            String localName) {
        return createOutFile(directory, localName, null);
    }

    /**
     * Creates a file for writing output to. The file is created in the
     * specified directory if given and writable; otherwise it is created
     * in the current working directory.  The resulting file can also be used
     * as a directory (by calling its {@link File#mkdir()} method).
     *
     * <p>The name of the file is derivated from the given local name by
     * replacing the extension of the input file (if any) by the specified
     * <code>outExtension</code>. If a file with this name already exists,
     * "2" or the next available number is inserted before the extension unless
     * an unused file name is found.
     *
     * <p>E.g. for input file "test.txt" and output extension "out", the file
     * "test.out" is created. If this file already exists, the file "test2.out"
     * is created instead (or "test3.out" etc. unless an unsed name is found).
     *
     * @param directory the directory in which the file should be created;
     * if <code>null</code> or not writable, the file is created in the
     * current working directory instead
     * @param localName the local name of the input file --- the extension of
     * this file (if any) is replaced by the specified output extension
     * @param outExtension the extension to append to the output file
     * (without a starting dot); ignored if empty or <code>null</code>
     * @return the create file
     */
    public static File createOutFile(final File directory, final
            String localName, final String outExtension) {
        return createOutFile(directory, localName, outExtension, null);
    }

    /**
     * Creates a file for writing output to. The file is created in the
     * specified directory if given and writable; otherwise it is created
     * in the current working directory.  The resulting file can also be used
     * as a directory (by calling its {@link File#mkdir()} method).
     *
     * <p>The name of the file is derivated from the given local name by
     * replacing the extension of the input file (if any) by the specified
     * <code>outExtension</code>. If a file with this name already exists,
     * "2" or the next available number is inserted before the extension unless
     * an unused file name is found.
     *
     * <p>E.g. for input file "test.txt" and output extension "out", the file
     * "test.out" is created. If this file already exists, the file "test2.out"
     * is created instead (or "test3.out" etc. unless an unsed name is found).
     *
     * @param directory the directory in which the file should be created;
     * if <code>null</code> or not writable, the file is created in the
     * current working directory instead
     * @param localName the local name of the input file --- the extension of
     * this file (if any) is replaced by the specified output extension
     * @param outExtension the extension to append to the output file
     * (without a starting dot); ignored if empty or <code>null</code>
     * @param last if not <code>null</code>, assumped to wrap the last numeric
     * prefix already in use, i.e. the first tried prefix will the the value
     * of this + 1; after determining a file name, the value will be set to the
     * numeric prefix used this time
     * @return the create file
     */
    public static File createOutFile(final File directory, final
            String localName, final String outExtension,
            final MutableInt last) {
        final File myDirectory;
        if (directory == null) {
            // no directory given: write to working dir
            myDirectory = null;
        } else if (!directory.canWrite()) {
            // cannot write to specified directory: write to working dir
            myDirectory = null;
        } else {
            myDirectory = directory;
        }

        final String inputExt = getExtension(localName);
        final String baseName;
        if ("".equals(inputExt)) {
            baseName = localName;
        } else {
            // remove input extension + preceding dot
            baseName = localName.substring(0,
                localName.length() - inputExt.length() - 1);
        }

        final String fullExt;
        if (StringUtils.isEmpty(outExtension)) {
            // ignore empty or null extension
            fullExt = "";
        } else {
            // prepend dot
            fullExt = EXT_SEPARATOR + outExtension;
        }

        final String separator;
        if ((baseName.length() > 0) && (Character.isDigit(
                baseName.charAt(baseName.length() - 1)))) {
            // base name ends in a digit: separate with "-" from the counter
            separator = "-";
        } else {
            separator = "";
        }

        File result;
        int counter;
        if (last == null) {
            // initially try without extension
            counter = 1;
            result = new File(myDirectory, baseName + fullExt);
        } else {
            // start with last value + 1
            counter = last.getValue() + 1;
            result = new File(myDirectory, baseName + separator + counter
                + fullExt);
        }

        while (result.exists()) {
            // increate count until a non-existing file has been found
            counter++;
            result = new File(myDirectory, baseName + separator + counter
                + fullExt);
        }

        if (last != null) {
            // store used counter
            last.setValue(counter);
        }

        return result;
    }

    /**
     * Returns the charset used by an InputStreamReader.
     *
     * @param reader the reader to check
     * @return the charset used by the reader
     */
    public static Charset determineCharset(final InputStreamReader reader) {
        return Charset.forName(reader.getEncoding());
    }

    /**
     * Returns the canoncical name of the charset used by an InputStreamReader.
     * This method always returns the <em>canonical</em> (standard) name.
     * It should preferably be used instead of calling
     * {@link InputStreamReader#getEncoding()} directly, because that method
     * often returns a non-standard ("historical") name.
     *
     * @param reader the reader to check
     * @return the canonical name of the charset used by the reader
     */
    public static String determineCharsetName(final InputStreamReader reader) {
        return determineCharset(reader).name();
    }

    /**
     * Returns the charset used by an OutputStreamWriter.
     *
     * @param writer the writer to check
     * @return the charset used by the writer
     */
    public static Charset determineCharset(final OutputStreamWriter writer) {
        return Charset.forName(writer.getEncoding());
    }

    /**
     * Returns the canoncical name of the charset used by an OutputStreamWriter.
     * This method always returns the <em>canonical</em> (standard) name.
     * It should preferably be used instead of calling
     * {@link OutputStreamWriter#getEncoding()} directly, because that method
     * often returns a non-standard ("historical") name.
     *
     * @param writer the writer to check
     * @return the canonical name of the charset used by the writer
     */
    public static String determineCharsetName(final OutputStreamWriter writer) {
        return determineCharset(writer).name();
    }

    /**
     * Determines the output to directory to use, reading it from the
     * {@link TextProcessor#KEY_OUT_DIRECTORY} configuration key in a given
     * configuration.
     *
     * @param config configuration used to determine the output directory
     * (read from the {@link TextProcessor#KEY_OUT_DIRECTORY} key)
     * @return the output directory to use; or <code>null</code> if none
     * is configured
     */
    public static File determineOutputDirectory(
            final TiesConfiguration config) {
        if (config.containsKey(TextProcessor.KEY_OUT_DIRECTORY)) {
            return new File(config.getString(TextProcessor.KEY_OUT_DIRECTORY));
        } else {
            return null;
        }
    }

    /**
     * Returns the base name of a file (the local name without
     * {@link #getExtension(File)} and preceding {@link #EXT_SEPARATOR dot}).
     *
     * @param file the file to check
     * @return the base name of the given file
    */
    public static String getBaseName(final File file) {
        // for file names valid in the current file system, the surrounding
        // call to getLocalName should be unnecessary; but it's helpful to
        // ensure correctness for file names from other OS
        return getBaseName(getLocalName(file.getName(), false));
    }

    /**
     * Helper method that returns the base name of a file name or URL.
     *
     * @param name the local name of the file or URL path to check
     * @return the base name of the given string
     */
    private static String getBaseName(final String name) {
        final int lastDot = name.lastIndexOf(EXT_SEPARATOR);

        if ((lastDot >= 0)) {
            return name.substring(0, lastDot);
        } else {
            // return full name
            return name;
        }
    }

    /**
     * Returns the base name of an URL (the local name without
     * {@link #getExtension(File)} and preceding {@link #EXT_SEPARATOR dot}).
     *
     * @param url the URL to check
     * @return the base name of the given URL
    */
    public static String getBaseName(final URL url) {
        // use local name (files only, no directories)
        final String localName = getLocalName(url, false);
        return getBaseName(localName);
    }

    /**
     * Returns an default directory that can be given as "parent" when
     * {@linkplain File#File(java.io.File, java.lang.String) constructing} file
     * objects. Initially set to <code>null</code> so the "system-dependent
     * default directory" (typically the working directory) will be used.
     *
     * @return the default directory
     */
    public static File getDefaultDirectory() {
        synchronized (DD_GUARD) {
            return defaultDirectory;
        }
    }

    /**
     * Returns the extension of a file.
     *
     * @param file the file to check
     * @return the extension of the given file; or an empty string if no
     * extension exists
    */
    public static String getExtension(final File file) {
        // for file names valid in the current file system, the surrounding
        // call to getLocalName should be unnecessary; but it's helpful to
        // ensure correctness for file names from other OS
        return getExtension(getLocalName(file.getName(), false));
    }

    /**
     * Helper method that returns the file extension of a file name or URL.
     *
     * @param name the local name of the file or URL path to check
     * @return the extension of the given string; or an empty string if no
     * extension exists
     */
    private static String getExtension(final String name) {
        final int lastDot = name.lastIndexOf(EXT_SEPARATOR);

        if ((lastDot >= 0)) {
            return name.substring(lastDot + 1);
        } else {
            return "";
        }
    }

    /**
     * Returns the file extension of an URL.
     *
     * @param url the URL to check
     * @return the extension of the given URL; or an empty string if no
     * extension exists
    */
    public static String getExtension(final URL url) {
        // use local name (files only, no directories)
        final String localName = getLocalName(url, false);
        return getExtension(localName);
    }

    /**
     * Returns the local name of an URL. This is the last existing element of
     * from the path component of the URL, typically the local file name
     * (without directories).
     *
     * <p>For URLs that are recognizable as directories (i.e. end in a slash),
     * the name of the final directory is returned if <code>acceptDir</code>
     * is <code>true</code> -- otherwise the empty string is returned.
     *
     * @param url the URL to check
     * @param acceptDir whether to return a final directory name
     * (see description)
     * @return the local name
     */
    public static String getLocalName(final URL url, final boolean acceptDir) {
        // the path component of an URL is what we need to check
        // (without a final ?-query, #-reference or ;-session info)
        final String localName = getLocalName(url.getPath(), acceptDir);

        // remove parameters after ";" (e.g. session info) if any exist
        final int firstSemicolon = localName.indexOf(';');

        if (firstSemicolon >= 0) {
            return localName.substring(0, firstSemicolon);
        } else {
            return localName;
        }
    }

    /**
     * Helper method that returns the local name of a file or URL path.
     *
     * <p>For paths that are recognizable as directories (i.e. end in a slash),
     * the name of the final directory is returned if <code>acceptDir</code>
     * is <code>true</code> -- otherwise the empty string is returned.
     *
     * @param path the file or URL path to check
     * @param acceptDir whether to return a final directory name
     * (see description)
     * @return the local name
     */
    private static String getLocalName(final String path,
            final boolean acceptDir) {
        // check for both '/' (Unix, URLs) and '\' (Windows, just to make sure)
        final String usedPath;
        if (path.endsWith("/") || path.endsWith("\\")) {
            // this is a directory URL
            if (acceptDir) {
                // remove trailing slash
                usedPath = path.substring(0, path.length() - 1);
            } else {
                // no dir allowed -- return empty string
                return "";
            }
        } else {
            // use complete path
            usedPath = path;
        }

        final int lastSlashOrBackslash = Math.max(path.lastIndexOf('/'),
            path.lastIndexOf('\\'));

        final String result;
        if (lastSlashOrBackslash >= 0) {
            result = usedPath.substring(lastSlashOrBackslash + 1);
        } else {
            // return complete used path
            result = usedPath;
        }
        return result;
    }

    /**
     * Opens an input stream that might have been compressed in
     * <code>gzip</code> format. This method autodetects whether the stream
     * has been compressed and returns a stream that allows accessing the
     * uncompressed data.
     *
     * @param in the raw input stream (either uncompressed or in
     * <code>gzip</code> format
     * @return a stream that allows accessing the uncompressed data
     * @throws IOException if an I/O error has occurred
     */
    public static InputStream openCompressableInStream(InputStream in)
    throws IOException {
        // we need a stream that supports mark + reset
        final InputStream markableStream;
        InputStream result;

        if (in.markSupported()) {
            markableStream = in;
        } else {
            // wrap in buffered stream to get support for mark + reset
            markableStream = new BufferedInputStream(in);
        }

        // mark initial position
        markableStream.mark(1024);

        // try to wrap in GZIP stream
        try {
            result = new GZIPInputStream(markableStream);
        } catch (IOException ioe) {
            // IOexception is thrown if the stream isn't in gzip format:
            // reset stream to marked position and return stream "as is"
            markableStream.reset();
            result = markableStream;
        }
        return result;
    }

    /**
     * Transparently opens an input stream that may use compression to store
     * the data (in <code>gzip</code> format).
     * 
     * @param out the original output stream
     * @param doCompress whether or not to use compression
     * @return to output stream to use for storing data; will be a
     * {@link GZIPOutputStream} wrapping <code>out</code> if
     * <code>doCompress</code> is <code>true</code>; otherwise it will be the
     * raw <code>out</code> stream
     * @throws IOException if an I/O error has occurred
     */
    public static OutputStream openCompressableOutStream(OutputStream out,
            boolean doCompress) throws IOException {
        if (doCompress) {
            // wrap in GZIP stream
            return new GZIPOutputStream(out);
        } else {
            // return stream as is
            return out;
        }
    }

    /**
     * Transparently opens an input stream that may use compression to store
     * the data (in <code>gzip</code> format). Checks the
     * {@link #CONFIG_COMPRESS_GZIP} parameter to decide whether or not to
     * use compression.
     * 
     * @param out the original output stream
     * @param config the configuration to use
     * @return to output stream to use for storing data;
     * @throws IOException if an I/O error has occurred
     */
    public static OutputStream openCompressableOutStream(OutputStream out,
            TiesConfiguration config) throws IOException {
        return openCompressableOutStream(out,
                config.getBoolean(CONFIG_COMPRESS_GZIP));
    }

    /**
     * Opens an reader on a local file. Uses the {@link #KEY_LOCAL_CHARSET
     * configured character set} -- if not specified, the default charset of
     * the current platform is used instead. Compressed files are automatically
     * decompressed (using the
     * {@link #openCompressableInStream(InputStream)} method). Don't forget to
     * finally {@linkplain #tryToClose(Reader) close} any reader you open!
     *
     * @param file the file to read
     * @param config the configuration to use
     * @return a reader on the local file
     * @throws IOException if an I/O error has occurred
     */
    public static InputStreamReader openReader(final File file,
            final Configuration config) throws IOException {
        // delegate
        return openReader(file, config.getString(KEY_LOCAL_CHARSET, null));
    }

    /**
     * Opens an reader on a local file, using a given charset. Compressed files
     * are automatically decompressed (using the
     * {@link #openCompressableInStream(InputStream)} method). Don't forget to
     * finally {@linkplain #tryToClose(Reader) close} any reader you open!
     *
     * @param file the file to read
     * @param charset the character set to use for reading the file;
     * if <code>null</code>, the default charset of the current platform is used
     * @return a reader on the local file
     * @throws IOException if an I/O error has occurred
     */
    public static InputStreamReader openReader(final File file,
            final String charset) throws IOException {
        final InputStream in =
            openCompressableInStream(new FileInputStream(file));
        final InputStreamReader reader;

        if (charset == null) {
            // use default charset
            reader = new InputStreamReader(in);
        } else {
            // use given charset
            reader = new InputStreamReader(in, charset);
        }
        return reader;
    }

    /**
     * Opens a reader on an input stream that uses a Unicode character set
     * (UTF-8, UTF-16, or UTF-32) and optionally a
     * <a href="http://www.unicode.org/unicode/faq/utf_bom.html">BOM
     * (byte order mark)</a> to identify the used charset. UTF-8 is used if BOM
     * is not found.
     *
     * <p>Adapted from the
     * <a href="http://koti.mbnet.fi/akini/java/unicodereader/"
     * ><code>UnicodeReader</code></a> class created by Thomas Weidenfeller and
     * Aki Nieminen.
     *
     * @param in the input stream to wrap
     * @return a reader on the stream
     * @throws IOException if an I/O error has occurred
     */
    public static InputStreamReader openUnicodeReader(final InputStream in)
    throws IOException {
        final PushbackInputStream pushbackIn =
            new PushbackInputStream(in, BOM_SIZE);

        final String encoding;
        final byte bom[] = new byte[BOM_SIZE];
        final int n, unread;
        n = pushbackIn.read(bom, 0, bom.length);

        // check BOM to determine charset
        if (  (bom[0] == (byte)0xEF) && (bom[1] == (byte)0xBB) &&
              (bom[2] == (byte)0xBF) ) {
            encoding = STANDARD_UNICODE_CHARSET; // UTF-8
            unread = n - 3;
        } else if ( (bom[0] == (byte)0xFE) && (bom[1] == (byte)0xFF) ) {
            encoding = "UTF-16BE";
            unread = n - 2;
        } else if ( (bom[0] == (byte)0xFF) && (bom[1] == (byte)0xFE) ) {
            encoding = "UTF-16LE";
            unread = n - 2;
        } else if ( (bom[0] == (byte)0x00) && (bom[1] == (byte)0x00) &&
                    (bom[2] == (byte)0xFE) && (bom[3] == (byte)0xFF)) {
            encoding = "UTF-32BE";
            unread = n - 4;
        } else if ( (bom[0] == (byte)0xFF) && (bom[1] == (byte)0xFE) &&
                    (bom[2] == (byte)0x00) && (bom[3] == (byte)0x00)) {
            encoding = "UTF-32LE";
            unread = n - 4;
        } else {
            // Unicode BOM mark not found, unread all bytes and use UTF-8
            encoding = STANDARD_UNICODE_CHARSET; // UTF-8
            unread = n;
        }

        // push back any read bytes that are not part of the BOM
        if (unread > 0) {
            pushbackIn.unread(bom, (n - unread), unread);
        }

        // Use determined encoding
        return new InputStreamReader(pushbackIn, encoding);
    }

    /**
     * Opens a writer that uses the standard Unicode character set
     * {@linkplain #STANDARD_UNICODE_CHARSET UTF-8}.
     *
     * @param out the output stream to wrap
     * @return a writer on the stream
     */
    public static OutputStreamWriter openUnicodeWriter(final OutputStream out) {
        try {
            return new OutputStreamWriter(out, STANDARD_UNICODE_CHARSET);            
        } catch (UnsupportedEncodingException uee) {
            // not supposed to happen as each virtual machine must support UTF-8
            throw new RuntimeException("Unexpected error: virtual machine does "
                    + "not support the standard charaset set"
                    + STANDARD_HTTP_CHARSET);
        }
    }

    /**
     * Opens an writer on a local file. Uses the {@link #KEY_LOCAL_CHARSET
     * configured character set} -- if not specified, the default charset of
     * the current platform is used instead. Don't forget to finally
     * {@linkplain #tryToClose(Writer) close} any writer you open!
     *
     * @param file the file to write to
     * @param config the configuration to use
     * @return a writer on the local file
     * @throws IOException if the file is a directory or for some
     * other reason cannot be opened for writing
     * @throws UnsupportedEncodingException if the named charset is not
     * supported
     */
    public static Writer openWriter(final File file,
            final Configuration config)
            throws IOException, UnsupportedEncodingException {
        // delegate
        return openWriter(file, config.getString(KEY_LOCAL_CHARSET, null));
    }

    /**
     * Opens an writer on a local file, using a given charset. Don't forget to
     * finally {@linkplain #tryToClose(Writer) close} any writer you open!
     *
     * @param file the file to write to
     * @param charset the character set to use for writing the file;
     * if <code>null</code>, the default charset of the current platform is used
     * @return a writer on the local file
     * @throws IOException if the file is a directory or for some
     * other reason cannot be opened for writing
     * @throws UnsupportedEncodingException if the named charset is not
     * supported
     */
    public static Writer openWriter(final File file, final String charset)
            throws IOException, UnsupportedEncodingException {
        final Writer writer;

        if (charset == null) {
            // use default charset
            writer = new FileWriter(file);
        } else {
            // use given charset
            writer =
                new OutputStreamWriter(new FileOutputStream(file), charset);
        }
        return writer;
    }

    /**
     * Reads the contents of a reader into a string. The reader is <em>not</em>
     * closed by this method.
     *
     * @param reader the reader to use
     * @return the contents of the reader
     * @throws IOException if an I/O error occurs
     */
    public static String readToString(final Reader reader) throws IOException {
        final StringBuffer result = new StringBuffer();

        final char[] buffer = new char[STANDARD_BLOCK_SIZE];
        int charsRead;

        while ((charsRead = reader.read(buffer)) > -1) {
            result.append(buffer, 0, charsRead);
        }

        return result.toString();
    }

    /**
     * Reads the contents of a reader into a writer. Neither reader nor writer
     * are closed by this method.
     *
     * @param reader the reader to read from
     * @param writer the writer to write to write to
     * @return the number of characters read
     * @throws IOException if an I/O error occurs
     */
    public static int readToWriter(final Reader reader, final Writer writer)
            throws IOException {
        final char[] buffer = new char[STANDARD_BLOCK_SIZE];
        int charsRead;
        int allCharsRead = 0;

        while ((charsRead = reader.read(buffer)) > -1) {
            writer.write(buffer, 0, charsRead);
            allCharsRead += charsRead;
        }

        writer.flush();
        return allCharsRead;
    }

    /**
     * Converts an URI list as defined in RFC 2483 (MIME type
     * <code>text/uri-list</code>) into an array of strings. Comment lines
     * in the input are ignored.
     *
     * @param in the URI list to convert
     * @return an array of strings containing the URIs/URLs listed in the input
     */
    public static String[] readURIList(final CharSequence in) {
        final String[] lines = TextUtils.splitLines(in);
        final List<String> uriList = new ArrayList<String>(lines.length);

        // discard comment URLs (starting with #)
        for (int i = 0; i < lines.length; i++) {
            if (!lines[i].startsWith("#")) {
                uriList.add(lines[i]);
            }
        }

        // convert to array
        String[] result = new String[uriList.size()];
        return uriList.toArray(result);
    }

    /**
     * Converts an URI list as defined in RFC 2483 (MIME type
     * <code>text/uri-list</code>) into an array of strings. Comment lines
     * in the input are ignored.
     *
     * @param in a reader containing the URI list to convert
     * @return an array of strings containing the URIs/URLs listed in the input
     * @throws IOException if an I/O error occurs
     */
    public static String[] readURIList(final Reader in) throws IOException {
        final String inSequence = readToString(in);
        return readURIList(inSequence);
    }

    /**
     * Modifies the default directory that can be given as "parent" when
     * {@linkplain File#File(java.io.File, java.lang.String) constructing} file
     * objects. Can be set to <code>null</code> so the "system-dependent
     * default directory" (typically the working directory) will be used.
     *
     * @param directory the new directory to use
     * @throws IllegalArgumentException if the given directory is neither
     * <code>null</code> nor an existing directory
     */
    public static void setDefaultDirectory(final File directory)
            throws IllegalArgumentException {
        if ((directory != null) && !directory.isDirectory()) {
            throw new IllegalArgumentException("Default directory must be "
                + "<null> or an existing directory, but " + directory
                + " is not");
        }

        synchronized (DD_GUARD) {
            defaultDirectory = directory;
        }
    }

    /**
     * Convenience method for closing an input stream. If the specified stream
     * is <code>null</code>, this method does nothing. Any {@link
     * java.io.IOException} thrown during closing is swallowed by this method.
     *
     * @param in the input stream to close (might be <code>null</code>)
     * @return <code>true</code> if the stream was closed successfully,
     * <code>false</code> otherwise (the stream was <code>null</code> or an
     * exception was thrown during closing)
     */
    public static boolean tryToClose(final InputStream in) {
        boolean result = false;

        if (in != null) {
            try {
                in.close();
                result = true;
            } catch (IOException ioe) {
                Util.LOG.warn("Exception while trying to close InputStream "
                    + in + ": " + ioe.toString());
            }
        }

        return result;
    }

    /**
     * Convenience method for closing an output stream. If the specified stream
     * is <code>null</code>, this method does nothing. Any {@link
     * java.io.IOException} thrown during closing is swallowed by this method.
     *
     * @param out the output stream to close (might be <code>null</code>)
     * @return <code>true</code> if the stream was closed successfully,
     * <code>false</code> otherwise (the stream was <code>null</code> or an
     * exception was thrown during closing)
     */
    public static boolean tryToClose(final OutputStream out) {
        boolean result = false;

        if (out != null) {
            try {
                out.close();
                result = true;
            } catch (IOException ioe) {
                Util.LOG.warn("Exception while trying to close OutputStream "
                    + out + ": " + ioe.toString());
            }
        }

        return result;
    }

    /**
     * Convenience method for closing a reader. If the specified reader is
     * <code>null</code>, this method does nothing. Any {@link
     * java.io.IOException} thrown during closing is swallowed by this method.
     *
     * @param reader the reader to close (might be <code>null</code>)
     * @return <code>true</code> if the reader was closed successfully,
     * <code>false</code> otherwise (reader was <code>null</code> or an
     * exception was thrown during closing)
     */
    public static boolean tryToClose(final Reader reader) {
        boolean result = false;

        if (reader != null) {
            try {
                reader.close();
                result = true;
            } catch (IOException ioe) {
                Util.LOG.warn("Exception while trying to close Reader "
                    + reader + ": " + ioe.toString());
            }
        }

        return result;
    }

    /**
     * Convenience method for closing a writer. If the specified writer is
     * <code>null</code>, this method does nothing. Any {@link
     * java.io.IOException} thrown during closing is swallowed by this method.
     *
     * @param writer the writer to close (might be <code>null</code>)
     * @return <code>true</code> if the writer was closed successfully,
     * <code>false</code> otherwise (writer was <code>null</code> or an
     * exception was thrown during closing)
     */
    public static boolean tryToClose(final Writer writer) {
        boolean result = false;

        if (writer != null) {
            try {
                writer.close();
                result = true;
            } catch (IOException ioe) {
                Util.LOG.warn("Exception while trying to close Writer "
                    + writer + ": " + ioe.toString());
            }
        }

        return result;
    }

    /**
     * Writes the contents of a character sequence to a writer. The writer
     * is flushed but not closed by this method.
     *
     * @param input the text to send to the wrier
     * @param writer the writer to write to write to
     * @throws IOException if an I/O error occurs
     */
    public static void writeToWriter(final CharSequence input,
            final Writer writer) throws IOException {
        writer.write(input.toString());
        writer.flush();
    }

    /**
     * Private constructor prevents creation of instances.
     */
    private IOUtils() {
        super();
    }

}
