/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.preprocess;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.w3c.tidy.Configuration;
import org.w3c.tidy.Node;
import org.w3c.tidy.Out;
import org.w3c.tidy.OutImpl;
import org.w3c.tidy.PPrint;
import org.w3c.tidy.Tidy;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ParsingException;
import de.fu_berlin.ties.ProcessingException;
import de.fu_berlin.ties.TextProcessor;
import de.fu_berlin.ties.TiesConfiguration;

import de.fu_berlin.ties.io.ContentType;
import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.text.TextUtils;
import de.fu_berlin.ties.util.ExternalCommand;
import de.fu_berlin.ties.util.Util;
import de.fu_berlin.ties.xml.OtherConstituent;
import de.fu_berlin.ties.xml.TagConstituent;
import de.fu_berlin.ties.xml.XMLAdjuster;
import de.fu_berlin.ties.xml.XMLConstituent;

/**
 * Preprocesses documents by converting them a suitable XML format and adding
 * lingustic information. Instances of this class are thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.16 $, $Date: 2004/12/07 12:02:05 $, $Author: siefkes $
 */
public class PreProcessor extends TextProcessor {

    /**
     * Configuration key prefix: command name and arguments of an external
     * converter from a specified type to HTML.
     */
    public static final String CONFIG_HTMLCONV_COMMAND
        = "html-converter.command";

    /**
     * Configuration key: Whether plain text is preprocessed to recognize
     * and reformat definition lists.
     */
    public static final String CONFIG_PREPROCESS_TEXT
        = "preprocess.text";

    /**
     * Configuration key: A tagger (or a list of taggers) used to annotate a
     * text e.g. with linguistic information. Each tagger must implement the
     * TextProcessor interface and accept a string (the output extension) as
     * single constructor argument.
     */
    public static final String CONFIG_PREPROCESS_TAGGER
        = "preprocess.tagger";

    /**
     * The replacement inserted by {@link #insertLineBreaks(CharSequence)}
     * (an empty <code>br</code> element is appended after each match).
     */
    private static final String NEWLINE_REPLACEMENT = "$0<br/>";

    /**
     * The name of the XHTML tag marking preformatted text.
     */
    private static final String PRE_TAG = "pre";

    /**
     * The name of the XHTML tag marking definitions in a definition list.
     */
    private static final String DD_TAG = "dd";

    /**
     * Pattern fragment matching a term in a definition list (printable
     * characters excluding ':' followed by a colon (':'). Following whitespace
     * is not included in this pattern.
     */
    private static final String DL_TERM = "[^\\s:]+:";

    /**
     * Pattern fragment matching an entry (term + definition) in a definition
     * text in plain text, formatted similar to RFC 822/2882:
     * "Term: Definition\n", optionally followed by deeper indented
     * continuation lines. Group 1 must match the current indentation level
     * (whitespace at the start of each line, if any).
     */
    private static final String DL_ENTRY =
        // term followed by whitespace and (start of) definition
        DL_TERM + TextUtils.SINGLE_LINE_WS.pattern() + "+\\S.*"
        // any number of continuation lines (deeper indented)
        + "(?:" + TextUtils.NEWLINE_PATTERN.pattern() + "\\1"
        + TextUtils.SINGLE_LINE_WS.pattern() + "+.+)*";

    /**
     * Pattern matching a paragraph that should be converted to a description
     * list (when converting plain text to XHTML). A list must have at least
     * two entries to be recognized.
     */
    private static final Pattern DL_PARA = Pattern.compile(
        // preceded by empty line (or start of text)
        "(?:\\A|" + TextUtils.NEWLINE_PATTERN.pattern()
            + TextUtils.SINGLE_LINE_WS.pattern() + "*"
            + TextUtils.NEWLINE_PATTERN.pattern() + ")"
        // group 1: indent level (whitespace that must be present on all lines)
        + "(" + TextUtils.SINGLE_LINE_WS.pattern() + "*)"
        // first entry
        + DL_ENTRY
        // further entries (starting w/possible indent captured in group 1)
        + "(?:" + TextUtils.NEWLINE_PATTERN.pattern() + "\\1" + DL_ENTRY + ")+"
        // we also allow a trailing term
        + "(?:" + TextUtils.NEWLINE_PATTERN.pattern() + "\\1" + DL_TERM
        + TextUtils.SINGLE_LINE_WS.pattern() + "*)?"
        // followed (lookahead) by empty line (or end of text)
            + "(?=\\Z|" + TextUtils.NEWLINE_PATTERN.pattern()
            + TextUtils.SINGLE_LINE_WS.pattern() + "*"
            + TextUtils.NEWLINE_PATTERN.pattern() + ")"
        );

    /**
     * Pattern fragment matching a short line (containing at most 40
     * characters + optionally starting and trailing whitespace).
     */
    private static final String SHORT_LINE =
        // optional whitespace
        TextUtils.SINGLE_LINE_WS.pattern() + "*"
        // opt. whitespace + 1-40 characters (first + last must be visible)
        + "\\S(?:.{1,38}\\S)?"
        // optional whitespace
        + TextUtils.SINGLE_LINE_WS.pattern() + "*";

    /**
     * Pattern matching two consecutive {@link #SHORT_LINE short lines}.
     */
    private static final Pattern TWO_SHORT_LINES = Pattern.compile(
        // first line + line separator
        "^" + SHORT_LINE + TextUtils.NEWLINE_PATTERN.pattern()
        // second line (using lookahead so the next match can match it again)
        + "(?=" + SHORT_LINE + "$)",
        Pattern.MULTILINE);

    /**
     * Pattern used to match an XML declaration not containing an encoding.
     */
    private static final Pattern XML_DECLARATION = Pattern.compile(
            "(<\\?xml\\s+version=\"1\\.0\")\\s*(\\?>)");

    /**
     * Whether plain text is preprocessed to recognize and reformat definition
     * lists.
     */
    private final boolean preprocessingText;

    /**
     * Tool for cleaning up malformed and faulty HTML. Synchronized on itself.
     */
    private final Tidy tidy;

    /**
     * Used for printing the output returned by Tidy. Synchronized on the
     * {@link #tidy} instance.
     */
    private final PPrint tidyPrinter;

    /**
     * An array of taggers used to annotate a text e.g. with linguistic
     * information.
     */
    private final TextProcessor[] tagger;

    /**
     * XML used for {@linkplain #insertLineBreaks(CharSequence) inserting line
     * breaks}.
     */
    private final XMLAdjuster xmlAdjuster;

    /**
     * Creates and configured a new instance, using a default extension and the
     * {@linkplain TiesConfiguration standard configuration}.
     */
    public PreProcessor() {
        this("aug");
    }

    /**
     * Creates and configured a new instance, using the
     * {@linkplain TiesConfiguration standard configuration}.
     *
     * @param outExt the extension to use for output files
     * @throws IllegalArgumentException if the configured linguistic tagger(s)
     * cannot be instantiated
     */
    public PreProcessor(final String outExt) throws IllegalArgumentException {
        this(outExt, TiesConfiguration.CONF);
    }

    /**
     * Creates and configured a new instance.
     *
     * @param outExt the extension to use for output files
     * @param config used to configure superclasses
     * @throws IllegalArgumentException if the configured linguistic tagger(s)
     * cannot be instantiated
     */
    public PreProcessor(final String outExt, final TiesConfiguration config)
    throws IllegalArgumentException {
        // call superclass + store configuration + set parameters
        super(outExt, config);
        preprocessingText = config.getBoolean(CONFIG_PREPROCESS_TEXT);

        // initialize (linguistic) taggers
        String[] taggerNames = config.getStringArray(CONFIG_PREPROCESS_TAGGER);
        if (!TiesConfiguration.arrayIsEmpty(taggerNames)) {
            // each tagger gets a dummy output extension as constructor arg
            final String[] params = new String[] { "tt" };
            tagger = new TextProcessor[taggerNames.length];

            for (int i = 0; i < taggerNames.length; i++) {
                try {
                    tagger[i] = (TextProcessor) Util.createObject(
                            Class.forName(taggerNames[i]), params);                    
                } catch (Exception e) {
                    // repackage exception
                    throw new IllegalArgumentException(
                            "Tagger initialization failed", e);
                }
            }
        } else {
            // no taggers are used
            tagger = null;
        }

        // create and configure TJidy
        tidy = new Tidy();

        // try to change the output log
        try {
            final FileWriter tidyLog = new FileWriter("ties-tidy.log");
            tidy.setErrout(new PrintWriter(tidyLog, true));
        } catch (IOException ioe) {
            Util.LOG.warn(
                "PreProcessor: couldn't redirect Tidy output to ties-tidy.log",
                ioe);
        }

        // adjust Tidy configuration, cf. http://www.w3.org/People/Raggett/tidy/
        tidy.setCharEncoding(Configuration.UTF8);
        tidy.setEncloseBlockText(true);
        tidy.setEncloseText(true);
        tidy.setDocType("omit");
        tidy.setDropEmptyParas(true);
        tidy.setLogicalEmphasis(true);
        tidy.setMakeClean(true);
        tidy.setOnlyErrors(true);
        tidy.setQuoteNbsp(false);
        tidy.setQuiet(true);
        tidy.setRawOut(true);
        tidy.setShowWarnings(false);
        tidy.setTidyMark(false);
        tidy.setWraplen(0);  // wrapping disabled
        tidy.setXmlPi(true);
        tidy.setXmlPIs(true);
        tidy.setXmlOut(true);

        // initialize the Tidy printer
        tidyPrinter = new PPrint(tidy.getConfiguration());

        // initialize XML adjuster without doing any extras
        xmlAdjuster =
            new XMLAdjuster(null, null, null, false, false, false, config);
    }

    /**
     * Converts HTML input to a clean XHTML representation, if necessary.
     * Delegates to <code>JTidy</code> for checking and cleaning the HTML code.
     *
     * @param input the HTML to tidy
     * @param charset the character to be used for storing the resulting XHTML
     * document (required to write the XML Declaration correctly)
     * @return the cleaned-up XHTML
     * @throws IOException if the I/O goes wrong
     */
    public final String cleanHTML(final String input, final String charset)
    throws IOException {
        // convert string to byte array input stream using UTF-8
        final InputStream inStream = new ByteArrayInputStream(
            input.getBytes(IOUtils.STANDARD_UNICODE_CHARSET));

        // store result in byte array output stream, estimating initial size
        final ByteArrayOutputStream outStream =
            new ByteArrayOutputStream(input.length());

        // synchronized on the tidy instance
        synchronized (tidy) {
            // do the actual clean-up + conversion
            final Node document = tidy.parse(inStream, null);

            // in case of some errors Tidy returns a node but doesn't print to
            // the output stream, so we'll handle this ourselves
            final Out out = new OutImpl();
            out.encoding = Configuration.UTF8;
            out.out = outStream;
            tidyPrinter.printTree(out, (short) 0, 0, null, document);
            tidyPrinter.flushLine(out, 0);
        }

        final String rawResult =
            outStream.toString(IOUtils.STANDARD_UNICODE_CHARSET);
        final String result;

        if (StringUtils.isNotEmpty(charset)) {
            // insert encoding in XML declaration
            result = XML_DECLARATION.matcher(rawResult).replaceFirst(
                    "$1 encoding=\"" + charset + "\"$2");
        } else {
            Util.LOG.warn("No character set specified -- cannot fix XML"
                    + " declaration of XHTML document");
            result = rawResult;
        }

        return result;
    }

    /**
     * Preprocesses the contents of a file. Neither input stream nor output
     * writer are closed by this method.
     *
     * @param reader a reader containing the text to preprocess; not closed
     * by this method
     * @param writer a writer used to store the preprocessed text; flushed
     * but not closed by this method
     * @param context a map of objects that are made available for processing;
     * the {@link ContentType#KEY_MIME_TYPE} key should to mapped to the MIME
     * type of the document and the {@link IOUtils#KEY_LOCAL_CHARSET} key to
     * the character set of the <code>writer</code>
     * @throws IOException if an I/O error occurred
     * @throws ProcessingException if the file couldn't be parsed, e.g. due to
     * an error in the XML input
     */
    protected final void doProcess(final Reader reader, final Writer writer,
            final ContextMap context)
            throws IOException, ProcessingException {
        // read contents from provided stream
        String contents = IOUtils.readToString(reader);
        final String mimeType = (String) context.get(ContentType.KEY_MIME_TYPE);
        final boolean html;
        final boolean xml;

        // preprocess plain text if configured
        if (preprocessingText && ContentType.MIME_PLAIN.equals(mimeType)) {
            contents = preprocessText(contents);
        }

        if (ContentType.MIME_HTML.equals(mimeType)) {
            // this is HTML but not necessarily XHTML
            html = true;
            xml = false;
        } else if ((mimeType != null) && (getConfig().containsKey(
                TiesConfiguration.joinKey(CONFIG_HTMLCONV_COMMAND,
                                            mimeType)))) {
            // there is an external XHTML converter configured for this type
            final String[] commandArgs = getConfig().getStringArray(
                TiesConfiguration.joinKey(CONFIG_HTMLCONV_COMMAND, mimeType));
            final ExternalCommand extConverter =
                new ExternalCommand(commandArgs);

            // run external converter and store output
            contents = extConverter.execute(contents);

            // output is already XHTML -- no need to send thru Tidy
            html = true;
            xml = true;
        } else {
            // try to process contents without conversion (XML)
            html = false;
            xml = true;
        }

        if (html) {
            if (!xml) {
                // clean-up HTML markup + convert to XHTML
                contents = cleanHTML(contents,
                        (String) context.get(IOUtils.KEY_LOCAL_CHARSET));
            }

            // insert <br> tags in <pre>
            contents = insertLineBreaks(contents);
        }

        // delegate to taggers, if any
        if (tagger != null) {
            StringWriter stringWriter;

            for (int i = 0; i < tagger.length; i++) {
                stringWriter = new StringWriter(contents.length());
                tagger[i].process(new StringReader(contents), stringWriter,
                        context);
                contents = stringWriter.toString();
            }
        }

        // store results
        writer.write(contents);
        writer.flush();
    }

    /**
     * Adds empty <code>br</code> elements in an XHTML document where
     * appropriate for better recognizing of the physical formatting.
     * Currently this is done at the begin of each line in a <code>pre</code>
     * element; and between short lines (at most 40 visible characters each)
     * in <code>dd</code> element (definitions within definition lists).
     * (Text in CDATA sections is not modified; leading and trailing linebreaks
     * are ignored.)
     *
     * @param input the XHTML to process
     * @return the XHTML with <code>br</code> elements added
     * @throws ParsingException if the file couldn't be parsed, e.g. due to an
     * error in the XML input
     */
    private String insertLineBreaks(final CharSequence input)
            throws ParsingException {
        final StringBuffer result = new StringBuffer();
        final XMLConstituent firstConst =
            xmlAdjuster.rawConstituents(input, false);
        XMLConstituent currentConst = firstConst;
        TagConstituent currentTag;
        int currentType;
        boolean inPre = false;
        boolean inDD = false;
        String output;
        final char namespaceSeparator = ':';
        final Matcher newlineMatcher = TextUtils.NEWLINE_PATTERN.matcher("");
        final Matcher shortLinesMatcher = TWO_SHORT_LINES.matcher("");

        // iterate + insert br elements where required
        while (currentConst != null) {
            currentType = currentConst.getType();

            if ((currentType == TagConstituent.START_TAG)
                    || (currentType == TagConstituent.END_TAG)) {
                currentTag = (TagConstituent) currentConst;

                // quick-and-dirty namespace handling
                if (currentTag.getName().equals(PRE_TAG) || currentTag
                        .getName().endsWith(namespaceSeparator + PRE_TAG)) {
                    // entering or exiting <pre> element
                    inPre = (currentType == TagConstituent.START_TAG);
                } else if (currentTag.getName().equals(DD_TAG) || currentTag
                    .getName().endsWith(namespaceSeparator + DD_TAG)) {
                // entering or exiting <dd> element
                inDD = (currentType == TagConstituent.START_TAG);
                }
            }

            if (inPre && (currentType == OtherConstituent.TEXT)) {
                // insert <br/> tags within text content of <pre> elements
                output = TextUtils.replaceAll(currentConst.getRepresentantion(),
                    newlineMatcher, NEWLINE_REPLACEMENT);

                if (output != currentConst.getRepresentantion()) {
                    Util.LOG.debug("Inserted break tags into preformatted text:"
                        + TextUtils.LINE_SEPARATOR + output);
                }
            } else if (inDD && (currentType == OtherConstituent.TEXT)) {
                // insert <br/> tags within two short lines within <dd> elements
                output = TextUtils.replaceAll(currentConst.getRepresentantion(),
                    shortLinesMatcher, NEWLINE_REPLACEMENT);

                if (output != currentConst.getRepresentantion()) {
                    Util.LOG.debug("Inserted break tags between short lines "
                        + "in <dd> element:" + TextUtils.LINE_SEPARATOR
                        + output);
                }
            } else {
                // output representation as is
                output = currentConst.getRepresentantion();
            }

            result.append(output);
            currentConst = currentConst.nextConstituent();
        }

        return result.toString();
    }

    /**
     * Preprocessed plain text to bring definition lists in a format recognized
     * by <code>txt2html</code>.
     *
     * @param input the plain text to process
     * @return the plain text with converted definition lists; or a reference
     * to <code>input</code> if there was nothing to convert
     */
    private String preprocessText(final String input) {
        // trimming text because a single empty line at start or end could
        // spoil matches
        final String trimmedInput = input.trim();
        final Matcher dlMatcher = DL_PARA.matcher(trimmedInput);
        boolean found = dlMatcher.find();
        int afterLastMatch = 0;
        String orgPara, normalizedPara, convertedPara;
        String indent;  // indentation level (group 1)
        String extraIndent;
        Pattern termToConvert;
        Pattern continuationIndent;

        if (found) {
            final StringBuffer result = new StringBuffer();
            do {
                // append text preceding the match
                result.append(trimmedInput.substring(afterLastMatch,
                    dlMatcher.start()));
                afterLastMatch = dlMatcher.end();
                indent = dlMatcher.group(1);
                orgPara = dlMatcher.group();

                // use extra indent of 2 for (first line of) definition
                extraIndent = indent + "  ";

                // normalize continuation lines to use extra indent
                continuationIndent = Pattern.compile(
                    // must start on new line (group 1)
                    "(\\A|" + TextUtils.NEWLINE_PATTERN.pattern() + ")"
                    // with deeper intentation
                    + indent + TextUtils.SINGLE_LINE_WS.pattern() + "+"
                    );
                normalizedPara = TextUtils.replaceAll(orgPara,
                    continuationIndent, "$1" + extraIndent);

                // term to convert
                termToConvert = Pattern.compile(
                    // must start on new line
                    "((?:\\A|" + TextUtils.NEWLINE_PATTERN.pattern() + ")"
                    // with correct indent
                    + indent + DL_TERM + ")"
                    // followed by whitespace
                    + TextUtils.SINGLE_LINE_WS.pattern() + "+"
                    // and then (lookahead) a printable char
                    + "(?=\\S)");

                // replace whitespace after term by newline + extra indent
                convertedPara = TextUtils.replaceAll(normalizedPara,
                    termToConvert,
                    "$1" + TextUtils.LINE_SEPARATOR + extraIndent);


                Util.LOG.debug(
                    "Reformatted paragraph containing definition list: "
                    + convertedPara);

                // append + look for next one
                result.append(convertedPara);
                found = dlMatcher.find();
            } while (found);

            // append rest of text
            result.append(trimmedInput.substring(afterLastMatch));
            return result.toString();
        } else {
            // nothing to replace
            return input;
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder builder = new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("preprocessing text", preprocessingText);

        if (tagger != null) {
            builder.append("tagger", tagger);
        }

        return builder.toString();
    }

}
