/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.preprocess;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ParsingException;
import de.fu_berlin.ties.TextProcessor;
import de.fu_berlin.ties.TiesConfiguration;

import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.text.TextUtils;
import de.fu_berlin.ties.util.ExternalCommand;
import de.fu_berlin.ties.util.Util;
import de.fu_berlin.ties.xml.TagIsolator;
import de.fu_berlin.ties.xml.XMLAdjuster;
import de.fu_berlin.ties.xml.XMLTokenizerFactory;

/**
 * Integrates the
 * <a href="http://www.ims.uni-stuttgart.de/projekte/corplex/TreeTagger/"
 * >TreeTagger</a>, a linguistic tool for part-of-speech tagging and chunk
 * parsing. This integration brings XML-based input files in a form that can
 * be processed by TreeTagger, runs the external TreeTagger command, converts
 * the output in the augmented text format defined by TIE, inserting tags
 * marking sentences and unifying the original XML markup and the TreeTagger
 * output in a single XML tree. This class is thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.14 $, $Date: 2004/11/08 11:57:35 $, $Author: siefkes $
 */
public class TreeTagger extends TextProcessor {

    /**
     * Configuration key: the name of the TreeTagger command
     * (language-specific).
     */
    private static final String CONFIG_COMMAND = "treetagger.command";

    /**
     * Configuration key: the POS tag marking the end of a sentence
     * (language-specific).
     */
    private static final String CONFIG_END_OF_SENTENCE = "treetagger.eos";

    /**
     * Configuration key: list of POS tags still to include in a previous
     * sentence when they occur after an end-of-sentence tag (language-specific,
     * optional).
     */
    private static final String CONFIG_AFTER_EOS = "treetagger.after-eos";

    /**
     * Tag marking the end of a sentence constituent (XML end tag).
     */
    private static final String END_TAG_CONSTITUENT = "</const>";

    /**
     * Tag marking the end of a sentence (XML end tag).
     */
    private static final String END_TAG_SENTENCE = "</sent>";

    /**
     * The apostrophe character (').
     */
    private static final String APOSTROPHE_CHAR = "'";

    /**
     * A pseudo entity used to replace the apostrophe (') within XML tags,
     * so it is protected from the TreeTagger which separates clitics
     * (English version).
     */
    private static final String PSEUDO_ENTITY_APOSTROPHE = "&';";

    /**
     * A pseudo entity used to replace whitespace within XML tags,
     * so it is protected from the TreeTagger which splits lines at whitespace.
     */
    private static final String PSEUDO_ENTITY_WHITESPACE = "&;";

    /**
     * A map of regexp patterns and the corresponding replacement strings used
     * to make XML tags safe for use with the TreeTagger. We can use a hash map
     * as the order or replacements doesn't matter.
     */
    private static final Map<Pattern, String> REPLACE_WITHIN_TAGS =
        new HashMap<Pattern, String>();

    /**
     * A map of regexp patterns and the corresponding replacement strings to
     * revert the changes of applying the {@link #REPLACE_WITHIN_TAGS} map. We
     * can use a hash map as the order or replacements doesn't matter.
     */
    private static final Map<Pattern, String> RESTORE_WITHIN_TAGS =
        new HashMap<Pattern, String>();

    /**
     * Pattern for recognizing regular XML entities.
     */
    private static final Pattern XML_ENTITY =
        Pattern.compile("(&" + XMLTokenizerFactory.XML_NAME + ");");

    /**
     * Replacement string to protect regular XML entities from being split
     * by the TreeTagger tokenizer.
     */
    private static final String XML_ENTITY_REPLACEMENT = "$1&";

    /**
     * Pattern for recognizing the XML entities in the replaced form used to
     * protect them from TreeTagger.
     */
    private static final Pattern REPLACED_XML_ENTITY =
        Pattern.compile("(&" + XMLTokenizerFactory.XML_NAME + ")&");

    /**
     * Replacement string to restore protected regular XML entities to their
     * original form.
     */
    private static final String XML_ENTITY_RESTORE = "$1;";

    /**
     * An unmodifiable map of properties containg the HTML characters and
     * entities to substitute. Each key is a regular expression
     * {@link java.util.regex.Pattern} of characters and entities and should
     * be replaced by the correspinding value String.
     */
    private static final Map SUBSTITUTES;

    /**
     * Name of the properties file containing the HTML characters and entities
     * to substitute.
     */
    private static final String SUBSTITUTES_FILE = "conf/substitutes.cfg";

    /**
     * Static initialization of the static maps.
     */
    static {
        final Properties substProps = new Properties();
        final InputStream substStream =
            ClassLoader.getSystemResourceAsStream(SUBSTITUTES_FILE);

        if (substStream == null) {
            Util.LOG.error("Class loader returned null for "
                + SUBSTITUTES_FILE);
        }

        final Map<Pattern, String> tempSubst = new HashMap<Pattern, String>();

        try {
            substProps.load(substStream);
            Enumeration keys = substProps.propertyNames();

            // replace each 'name' by '&name;'
            final Matcher nameMatcher = Pattern.compile(
                "\\p{Alpha}+").matcher("");
            final String nameReplacement = "&$0;";

            // replace each 'number' by '&#number;|Unicode char'
            final Matcher numMatcher = Pattern.compile(
                "\\p{Digit}+").matcher("");
            final String numReplacement = "&#$0;";

            // replace whitespace and '|' by a single '|' (to mark alternatives)
            final Matcher whitespaceMatcher =
                Pattern.compile("[\\s|]+").matcher("\\p{Digit}+");
            final String whitespaceReplacement = "|";

            String currentPropKey, currentPropValue;
            StringBuffer currentPatternValue;
            Pattern currentPattern;
            int entityNumber; // the numeric value of the entity
            char entityChar;

            while (keys.hasMoreElements()) {
                currentPropKey = (String) keys.nextElement();
                currentPropValue =
                    substProps.getProperty(currentPropKey).trim();

                // replace whitespace
                currentPropValue = TextUtils.replaceAll(currentPropValue,
                    whitespaceMatcher, whitespaceReplacement);

                currentPropValue = TextUtils.replaceAll(currentPropValue,
                    nameMatcher, nameReplacement);

                numMatcher.reset(currentPropValue);
                currentPatternValue = new StringBuffer();

                while (numMatcher.find()) {
                    numMatcher.appendReplacement(currentPatternValue,
                        numReplacement);
                    // append raw Unicode character as third option
                    entityNumber = Integer.parseInt(numMatcher.group());
                    entityChar = (char) entityNumber;
                    currentPatternValue.append("|").append(entityChar);
                }
                numMatcher.appendTail(currentPatternValue);

                // create final pattern
                currentPattern = Pattern.compile("(?:"
                    + currentPatternValue + ")");
                tempSubst.put(currentPattern, currentPropKey);
            }
        } catch (RuntimeException rte) {
            Util.LOG.error("Could not initialize substitution table from "
                + SUBSTITUTES_FILE + " = ", rte);
        } catch (IOException ioe) {
            Util.LOG.error("Could not initialize substitution table from "
                + SUBSTITUTES_FILE + ": ", ioe);
        }
        SUBSTITUTES = Collections.unmodifiableMap(tempSubst);

        REPLACE_WITHIN_TAGS.put(Pattern.compile(APOSTROPHE_CHAR),
            PSEUDO_ENTITY_APOSTROPHE);
        RESTORE_WITHIN_TAGS.put(Pattern.compile(PSEUDO_ENTITY_APOSTROPHE),
            APOSTROPHE_CHAR);
        REPLACE_WITHIN_TAGS.put(Pattern.compile("\\s+"),
            PSEUDO_ENTITY_WHITESPACE);
        RESTORE_WITHIN_TAGS.put(Pattern.compile(PSEUDO_ENTITY_WHITESPACE),
            " ");
    }

    /**
     * XML adjuster for merging the original XML tree with the markup added by
     * the TreeTagger. Neither missing root nor emptiable tags are used,
     * control characters and pseudo-tags are not deleted, only illegal "&amp;"
     * are escaped; no file extension is required.
     */
    private final XMLAdjuster xmlAdjuster;

    /**
     * Used to isolate XML tags so the TreeTagger can handle a resulting file.
     */
    private final TagIsolator tagIsolator = new TagIsolator();

    /**
     * Creates a new instance, using the
     * {@linkplain TiesConfiguration standard configuration}.
     *
     * @param outExt the extension to use for output files
     */
    public TreeTagger(final String outExt) {
        this(outExt, TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance.
     *
     * @param outExt the extension to use for output files
     * @param config used to configure superclasses
     */
    public TreeTagger(final String outExt, final TiesConfiguration config) {
        super(outExt, config);

        // initialize XML adjuster: Neither missing root nor emptiable tags are
        // used, control characters and pseudo-tags are not deleted, only
        // illegal "&amp;" are escaped; no file extension is required
        xmlAdjuster =
            new XMLAdjuster(null, null, null, false, false, false, config);
    }

    /**
     * Augments the <code>input</code> text with the output of the TreeTagger.
     *
     * @param in reader containing the text to process; must contain the textual
     * representation of a well-formed XML document
     * @param out the writer to write the processed text to; the text will
     * be augmented with part-of-speech, lemma, and chunk information,
     * it will be a well-formed XML document (if the input was well-formed)
     * @param context a map of objects that are made available for processing
     * @throws IOException if an I/O error occurred
     * @throws ParsingException if the file couldn't be parsed, e.g. due to an
     * error in the XML input
     */
    protected void doProcess(final Reader in, final Writer out,
            final ContextMap context) throws IOException, ParsingException {
        final String input = IOUtils.readToString(in);
        //Util.LOG.debug("Un-augmented XML: " + input);

        // replaces some special characters/entities with a simpler
        // representation (e.g., different types of quotes, dashes,
        // and special whitespace characters)
        final String simplifiedXML = TextUtils.multipleReplaceAll(
            input, SUBSTITUTES);

        // protect entities from TreeTagger
        final String protectedXML = TextUtils.replaceAll(simplifiedXML,
            XML_ENTITY, XML_ENTITY_REPLACEMENT);

        // isolate tags for TreeTagger
        Writer isolateOut = new StringWriter();
        tagIsolator.isolateTags(new StringReader(protectedXML), isolateOut,
            REPLACE_WITHIN_TAGS);

        // delegate to TreeTagger
        final String taggerCommandName = getConfig().getString(
                getConfig().localizeKey(CONFIG_COMMAND));
        final ExternalCommand taggerCommand = new ExternalCommand(
            new String[] {taggerCommandName});
        final String treeTagged =
            taggerCommand.execute(null, isolateOut.toString());

        // restore entities to original form
        final String unprotectedXML = TextUtils.replaceAll(treeTagged,
            REPLACED_XML_ENTITY, XML_ENTITY_RESTORE);

        // restore whitespace + apostrophes in tags
        // We're using pseudo-entities  (illegal XML) instead of actual
        // ones, so we can replace them without fear of replacing too much
        final String unisolatedXML = TextUtils.multipleReplaceAll(
            unprotectedXML, RESTORE_WITHIN_TAGS);

        //Util.LOG.debug("Un-adjusted XML: " + unisolatedXML);

        // add end-of-sentence markers (</sent>)
        final String sentenceTagged = tagSentences(unisolatedXML);

        // adjusting XML structure (merging original + TreeTagger trees)
        xmlAdjuster.adjust(new StringReader(sentenceTagged), out);
    }

    /**
     * Adds tags to mark the sentences in a document. Only the ends of
     * sentences are tagged by this method by inserted &lt;/sent&gt; tags --
     * the corresponding start tags are later added by the XML adjuster.
     *
     * @param input the text to process
     * @return the processed tag with &lt;/sent&gt; tags added
     */
    protected final String tagSentences(final String input) {
        // build pattern of text to match
        final String eosType = getConfig().getString(getConfig().localizeKey(
                        CONFIG_END_OF_SENTENCE));
        final String[] afterEOSTypes = getConfig().getStringArray(
                getConfig().localizeKey(CONFIG_AFTER_EOS));
        // start pattern
        final StringBuffer patternString = new StringBuffer(typedPos(eosType));

        // open non-capturing group, accept preceding whitespace
        patternString.append("(?:\\s*");
        final String trailingTags;

        if ((afterEOSTypes != null) && (afterEOSTypes.length > 0)) {
            final String afterEOSPattern =
                TextUtils.joinAlternatives(afterEOSTypes);
            // these POS tags and ending constituents should be included
            trailingTags = TextUtils.joinAlternatives(new String[] {
                    typedPos(afterEOSPattern), END_TAG_CONSTITUENT
                }
            );
        } else {
            // ending constituents should be included
            trailingTags = END_TAG_CONSTITUENT;
        }

        // append trailing tags + close group (matching any number of times)
        patternString.append(trailingTags);
        patternString.append(")*");

        /* rough debugging only--very inefficient
        final Pattern eosPattern = Pattern.compile(patternString.toString());
        Log.TIES.info("End-of-sentence pattern '" + patternString.toString()
            + "' matches " + (eosPattern.split(input).length - 1) + " times");
        */

        // append sentence end tag after each match of the created pattern
        return TextUtils.replaceAll(input,
            Pattern.compile(patternString.toString()), "$0" + END_TAG_SENTENCE);
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("tag isolator", tagIsolator)
            .append("XML adjuster", xmlAdjuster)
            .toString();
    }

    /**
     * Helper method that create a regular expression string for matching a
     * part-of-speech element of a specified type.
     *
     * @param type the type string in {@link Pattern} format
     * @return a string in {@link Pattern} format matching the POS element
     */
    private String typedPos(final String type) {
        // we rely on the format written by our TreeTagger adaption (type is
        // the first attribute, quotes are " not ')
        return "<pos\\s+type=\"" + type + ".+?</pos>";
    }

}
