/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.text;

import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * Splits a text into a sequence of tokens.
 *
 * <p>This class is not thread-safe, so if you should want to share
 * a tokenizer between threads you have to ensure adequate synchronization.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.9 $, $Date: 2004/08/24 18:00:28 $, $Author: siefkes $
 */
public class TextTokenizer {

    /**
     * The index of the last character matched by the penultimate
     * token, plus one; or <code>0</code> if there was no match yet.
     */
    private int afterPenultimateMatch;

    /**
     * The normalized whitespace representation prepended if {@link
     * #normalizedWhitespacePrepended} is <code>true</code>. Defaults to a
     * space character.
     */
    private String normalizedWhitespace = " ";

    /**
     * Whether whitespace is prepended in a normalized form ({@link
     * #normalizedWhitespace}) to those tokens where {@link
     * #hasPrecedingWhitespace()} would return <code>true</code>.
     * Defaults to <code>false</code>.
     */
    private boolean normalizedWhitespacePrepended = false;

    /**
     * The whitespace preceding the last read token. Might be the empty
     * string if there is no preceding whitespace, or <code>null</code>
     * if the whitespace has not yet been determined.
     */
    private String precedingWhitespace;

    /**
     * The index of the first character matched by the last
     * token; or <code>-1</code> if there was no match yet, or
     * <code>textToTokenize.length()</code> if the last match wasn't
     * succeesful.
     */
    private int startOfLastMatch;

    /**
     * The text to tokenize.
     */
    private CharSequence textToTokenize;

    /**
     * The matcher used to split the text into matching tokens.
     */
    private final Matcher tokenMatcher;

    /**
     * The whitespace (non-token) matcher.
     */
    private final Matcher whitespaceMatcher;

    /**
     * Whether whitespace (the text between patterns) is checked to
     * ensure that the defined whitespace pattern is matched. If set to
     * true (default), a call to {@link #hasPrecedingWhitespace()} or
     * {@link #precedingWhitespace()} will throw an {@link
     * java.lang.IllegalArgumentException} if the whitespace preceding the
     * last read token does not match.
     *
     * <p>It might not be a good idea to enable this, because some characters
     * that are theoretically not in use in a charset might cause problems.
     * E.g. the French starting and ending quotes from the Windows charset
     * (at unused positions in Latin1 + Unicode) are recognized as {@link
     * java.lang.Character#INITIAL_QUOTE_PUNCTUATION} (<code>Pi</code>
     * category) or {@link java.lang.Character#FINAL_QUOTE_PUNCTUATION}
     * (<code>Pf</code>), but in some cases they do not seem to match any
     * regex category patterns.
     */
    private boolean whitespacePatternEnsured = false;

    /**
     * Creates a new instance. Only use this constructor if you know what you
     * are doing! Usually it should be sufficient to use one of the factory
     * methods provided by {@link TokenizerFactory}.
     *
     * @param patterns a list of patterns to accept as tokens; patterns
     * jointed and compiled with the {@link java.util.regex.Pattern#DOTALL}
     * flag activated
     * @param whitespacePattern a pattern that should match all text
     * between tokens ("whitespace"), to ensure that no text is left out
     * by mistake; the pattern is compiled with the
     * {@link java.util.regex.Pattern#DOTALL} flag activated
     * @param text the text to tokenize
     * @throws PatternSyntaxException if the syntax of the provided
     * patterns is invalid
     */
    public TextTokenizer(final String[] patterns,
            final String whitespacePattern, final CharSequence text)
            throws PatternSyntaxException {
        super();
        final StringBuffer patternBuffer;

        // combine in non-capturing group, if more than one pattern
        if (patterns.length > 1) {
            patternBuffer = new StringBuffer("(?:");
        } else {
            patternBuffer = new StringBuffer();
        }

        // build complete pattern
        for (int i = 0; i < patterns.length; i++) {
            patternBuffer.append(patterns[i]);

            if (i == (patterns.length - 1)) {
                if (patterns.length > 1) {
                    // last of several patterns: close non-capturing group
                    patternBuffer.append(")");
                }
            } else {
                patternBuffer.append("|");
            }
        }

        textToTokenize = text;

        final Pattern tokenPattern = Pattern.compile(
            patternBuffer.toString(), Pattern.DOTALL);
        tokenMatcher = tokenPattern.matcher(text);

        final Pattern wsPattern = Pattern.compile(whitespacePattern,
            Pattern.DOTALL);
        whitespaceMatcher = wsPattern.matcher("");

        // initialize fields
        reinit();
    }

    /**
     * Returns the text captured within "capturing groups" in the last token.
     * All captured text sequences are joint in a single string.
     * If there were no capturing groups involved in the last match, the
     * empty string is returned.
     *
     * @return the joint text matched within captured groups in the last token
     * match
     */
    public final String capturedText() {
        final StringBuffer result = new StringBuffer();
        String currentGroup;
        for (int i = 1; i <= tokenMatcher.groupCount(); i++) {
            currentGroup = tokenMatcher.group(i);
            if (currentGroup != null) {
                result.append(currentGroup);
            }
        }
        return result.toString();
    }

    /**
     * Helper method to find the preceding whitespace.
     *
     * @throws IllegalStateException if this method is called
     * without a prior call to {@link #nextToken()}
     */
    private void doFindPrecedingWhitespace() throws IllegalStateException {
        if (startOfLastMatch < 0) {
            throw new IllegalStateException(
                "Prior call to nextToken() required");
        }

        if (precedingWhitespace == null) {
            // otherwise there is nothing to do

            /* rough debugging
            System.out.print("/" + afterPenultimateMatch + "/"
                + startOfLastMatch + "/");
            */

            precedingWhitespace = textToTokenize.subSequence(
                afterPenultimateMatch, startOfLastMatch).toString();
        }
    }

    /**
     * Helper method to find and optionally validate the preceding whitespace.
     *
     * @throws IllegalStateException if this method is called
     * without a prior call to {@link #nextToken()}
     * @throws IllegalArgumentException if
     * {@link #isWhitespacePatternEnsured()} is <code>true</code> and the
     * whitespace preceding the last read token does not match the defined
     * whitespace pattern
     */
    private void findPrecedingWhitespace() throws IllegalStateException,
            IllegalArgumentException {
        doFindPrecedingWhitespace();

        // validate
        if (isWhitespacePatternEnsured() && (!precedingWhitespaceIsValid())) {
            throw new IllegalArgumentException(
                "Supposed whitespace '" + precedingWhitespace
                + "' between position " + afterPenultimateMatch + " and "
                + startOfLastMatch + " doesn't match specified pattern "
                + whitespaceMatcher.pattern().pattern());
        }
    }

    /**
     * Returns the normalized whitespace representation prepended if {@link
     * #isNormalizedWhitespacePrepended()} is <code>true</code>. Defaults to
     * a space character.
     *
     * @return the normalized representation
     */
    public final String getNormalizedWhitespace() {
        return normalizedWhitespace;
    }

    /**
     * Whether the token returned by the last call to {@link #nextToken()}
     * is preceded by whitespace (i.e., text not matched by any token).
     * If we arrived at the end of the text to tokenize (last call to
     * {@link #nextToken()} returned <code>null</code>), this is the whitespace
     * between the last existing token and the end of the text.
     *
     * @return whether the last token is preceded by whitespace
     * @throws IllegalStateException if this method is called
     * without a prior call to {@link #nextToken()}
     * @throws IllegalArgumentException if
     * {@link #isWhitespacePatternEnsured()} is <code>true</code> and the
     * whitespace preceding the last read token does not match the defined
     * whitespace pattern
     */
    public final boolean hasPrecedingWhitespace() throws IllegalStateException,
            IllegalArgumentException {
        // determine whitespace, if not yet done
        findPrecedingWhitespace();

        return precedingWhitespace.length() > 0;
    }

    /**
     * Convenience method that counts the number of whitespace characters at the
     * begin of a string, according to the defined whitespace pattern.
     *
     * @param text the text to check
     * @return the number of whitespace characters at the begin, of 0 if there
     * are none
     */
    public int initialWhitespaceCount(final String text) {
        whitespaceMatcher.reset(text);
        if (whitespaceMatcher.lookingAt()) {
            return whitespaceMatcher.end();
        } else {
            // no initial whitespace
            return 0;
        }
    }

    /**
     * Returns whether whitespace is prepended in a normalized form ({@link
     * #getNormalizedWhitespace()}) to those tokens where {@link
     * #hasPrecedingWhitespace()} would return <code>true</code>.
     * Defaults to <code>false</code>.
     *
     * @return whether whitespace is prepended
     */
    public final boolean isNormalizedWhitespacePrepended() {
        return normalizedWhitespacePrepended;
    }

    /**
     * Convenience method that checks whether a string matches the defined
     * whitespace pattern.
     *
     * @param text the text to match
     * @return <code>true</code> iff the given text matches the
     * defined whitespace pattern or is the empty string
     */
    public boolean isValidWhitespace(final String text) {
        if (text.length() > 0) {
            whitespaceMatcher.reset(text);
            return whitespaceMatcher.matches();
        } else {
            // empty string always matches
            return true;
        }
    }

    /**
     * Whether whitespace (the text between patterns) is checked to
     * ensure that the defined whitespace pattern is matched. If set to
     * true (default), a call to {@link #hasPrecedingWhitespace()} or
     * {@link #precedingWhitespace()} will throw an {@link
     * java.lang.IllegalArgumentException} if the whitespace preceding the
     * last read token does not match.
     *
     * @return the value of this property
     */
    public final boolean isWhitespacePatternEnsured() {
        return whitespacePatternEnsured;
    }

    /**
     * Returns the complete text to the left (preceding) the token returned
     * by the last call to {@link #nextToken()}. This includes any
     * {@link #precedingWhitespace()}.
     *
     * @return the complete text to the left of the last token
     * @throws IllegalStateException if this method is called
     * without a prior call to {@link #nextToken()}
     */
    public CharSequence leftText() throws IllegalStateException {
        if (startOfLastMatch < 0) {
            throw new IllegalStateException(
                "Prior call to nextToken() required");
        }

        return textToTokenize.subSequence(0, startOfLastMatch);
    }

    /**
     * Returns the next token, or <code>null</code> if there are no
     * more tokens left in the provided text. When the tokenizer arrived
     * at the end of the text, all subsequent calls to this method
     * will return <code>null</code> until you call one of the
     * {@link #reset() reset} methods. If the token is preceded by whitespace
     * and {@link #isNormalizedWhitespacePrepended()} is <code>true</code>,
     * the returned token will start with the normalized whitespace
     * representation ({@link #getNormalizedWhitespace()}).
     *
     * @return the next token read from the provided text (with or without
     * prepended whitespace), or <code>null</code> if no tokens are left
     * @throws IllegalArgumentException if
     * {@link #isWhitespacePatternEnsured()} and
     * {@link #isNormalizedWhitespacePrepended()} are <code>true</code> and the
     * whitespace preceding this token does not match the defined whitespace
     * pattern
     */
    public final String nextToken() throws IllegalArgumentException {
        final String theToken;
        final String result;

        // reset whitespace
        precedingWhitespace = null;

        if ((startOfLastMatch >= 0)
                && (startOfLastMatch < textToTokenize.length())) {
            // there was a successful match before, so we can safely call
            // this method
            afterPenultimateMatch = tokenMatcher.end();
        }

        if (tokenMatcher.find()) {
            theToken = tokenMatcher.group();

            // remember the start of this match
            startOfLastMatch = tokenMatcher.start();
        } else {
            theToken = null;

            // non-match: matches "after text"
            startOfLastMatch = textToTokenize.length();
        }

        if ((theToken != null) && isNormalizedWhitespacePrepended()
                && hasPrecedingWhitespace()) {
            // prepend whitespace
            result = getNormalizedWhitespace() + theToken;
        } else {
            // return as is
            result = theToken;
        }

        return result;
    }

    /**
     * Returns the whitespace (i.e., text not matched by any token) preceding
     * the token returned by the last call to {@link #nextToken()}.
     * If we arrived at the end of the text to tokenize (last call to
     * {@link #nextToken()} returned <code>null</code>), this is the
     * whitespace between the last existing token and the end of the text.
     *
     * @return the whitespace preceding the last token, or the empty
     * string if there is no preceding whitespace (i.e.
     * {@link #hasPrecedingWhitespace()} would return <code>false</code>)
     * @throws IllegalStateException if this method is called
     * without a prior call to {@link #nextToken()}
     * @throws IllegalArgumentException if
     * {@link #isWhitespacePatternEnsured()} is <code>true</code> and the
     * whitespace preceding the last read token does not match the defined
     * whitespace pattern
     */
    public final String precedingWhitespace() throws IllegalStateException,
            IllegalArgumentException {
        // determine whitespace, if not yet done
        findPrecedingWhitespace();
        return precedingWhitespace;
    }

    /**
     * Checks whether the whitespace (i.e., text not matched by any token)
     * preceding the token returned by the last call to {@link #nextToken()}
     * matches the defined whitespace pattern. This method is called
     * automatically if {@link #isWhitespacePatternEnsured()} is
     * <code>true</code>. Otherwise it can be called externally to check
     * whether the whitespace is valid and take appropriate action if required.
     *
     * @return <code>true</code> iff the preceding whitespace matches the
     * specified whitespace pattern or if there is no preceding whitespace
     * @throws IllegalStateException if this method is called
     * without a prior call to {@link #nextToken()}
     */
    public boolean precedingWhitespaceIsValid() throws IllegalStateException {
        doFindPrecedingWhitespace();
        return isValidWhitespace(precedingWhitespace);
    }

    /**
     * Initializes or re-initializes the state of this tokenizer
     * at construction or when resetting the text.
     */
    private void reinit() {
        startOfLastMatch = -1;
        afterPenultimateMatch = 0;
        precedingWhitespace = null;
    }

    /**
     * Resets this tokenizer, so it will restart at the begin of the
     * current text.
     */
    public final void reset() {
        tokenMatcher.reset();

        // re-initialize fields
        reinit();
    }

    /**
     * Resets this tokenizer, so it will restart at the begin of the
     * provided text.
     *
     * @param newText the new text to tokenize
     */
    public final void reset(final CharSequence newText) {
        tokenMatcher.reset(newText);
        textToTokenize = newText;

        // re-initialize fields
        reinit();
    }

    /**
     * Returns the complete text to the right (following) the token returned
     * by the last call to {@link #nextToken()}. This includes any following
     * whitespace.
     *
     * @return the complete text to the right of the last token
     * @throws IllegalStateException if this method is called
     * without a prior call to {@link #nextToken()}
     */
    public CharSequence rightText() throws IllegalStateException {
        if (startOfLastMatch < 0) {
            throw new IllegalStateException(
                "Prior call to nextToken() required");
        }

        // return everything after the matched sequence
        return textToTokenize.subSequence(tokenMatcher.end(),
            textToTokenize.length());
    }

    /**
     * Changes the normalized whitespace representation prepended if {@link
     * #isNormalizedWhitespacePrepended()} is <code>true</code>.
     *
     * @param newValue the new value
     */
    public final void setNormalizedWhitespace(final String newValue) {
        normalizedWhitespace = newValue;
    }

    /**
     * Changes whether whitespace is prepended in a normalized form ({@link
     * #getNormalizedWhitespace()}) to those tokens where {@link
     * #hasPrecedingWhitespace()} would return <code>true</code>.
     *
     * @param newValue the new value
     */
    public final void setNormalizedWhitespacePrepended(final boolean newValue) {
        normalizedWhitespacePrepended = newValue;
    }

    /**
     * Specifies whether whitespace (the text between patterns) is checked to
     * ensure that the defined whitespace pattern is matched. If set to
     * true (default), a call to {@link #hasPrecedingWhitespace()} or
     * {@link #precedingWhitespace()} will throw an {@link
     * java.lang.IllegalArgumentException} if the whitespace preceding the
     * last read token does not match.
     *
     * @param ensured the new value of this property
     */
    public final void setWhitespacePatternEnsured(final boolean ensured) {
        whitespacePatternEnsured = ensured;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder builder = new ToStringBuilder(this).
            append("token pattern", tokenMatcher.pattern().pattern()).
            append("whitespace pattern", whitespaceMatcher.pattern().pattern()).
            append("whitespace pattern ensured", whitespacePatternEnsured);

        if (normalizedWhitespacePrepended) {
            builder.append("normalized whitespace (is prepended)",
                normalizedWhitespace);
        }
        return builder.toString();
    }

    /**
     * Convenience method that counts the number of whitespace characters at the
     * end of a string, according to the defined whitespace pattern.
     *
     * @param text the text to check
     * @return the number of whitespace characters at the end, of 0 if there
     * are none
     */
    public int trailingWhitespaceCount(final String text) {
        // append "\z" (end of input) to whitespace pattern
        final Pattern trailingWSPattern = Pattern.compile(
            whitespaceMatcher.pattern().pattern() + "\\z");
        final Matcher trailingWSMatcher = trailingWSPattern.matcher(text);

        if (trailingWSMatcher.find()) {
            return trailingWSMatcher.group().length();
        } else {
            // no trailing whitespace
            return 0;
        }
    }

}
