/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.text;

import java.util.Iterator;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringUtils;

/**
 * A static class that provides utility constants and methods for working with
 * texts and regular expressions.
 * No instances of this class can be created, only the static members
 * should be used.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.10 $, $Date: 2004/12/06 17:59:28 $, $Author: siefkes $
 */
public final class TextUtils {

    /**
     * The line separator on the current operating system ("\n" on Unix).
     */
    public static final String LINE_SEPARATOR =
        System.getProperty("line.separator");

    /**
     * Regex fragment listing the newline alternatives used by differents
     * systems: "\r\n" (Windows), "\n" (Unix) or "\r" (Mac).
     */
    public static final String NEWLINE_ALTERNATIVES = "\\r\\n|\\n|\\r";

    /**
    * A regular expression matching a non-line-breaking whitespace character
    * (character class containing space and tab).
    */
    public static final Pattern SINGLE_LINE_WS =
           Pattern.compile("[ \\t]");

    /**
    * A regular expression matching a single newlines (build by enclosing
    * {@link #NEWLINE_ALTERNATIVES} in a non-capturing group).
    */
    public static final Pattern NEWLINE_PATTERN =
           Pattern.compile("(?:" + NEWLINE_ALTERNATIVES + ")");

    /**
     * A regular expression matching newlines, including surrounding whitespace.
     * Will match several newlines if they immediately follow each other or
     * are separated by whitespace only.
     */
    public static final Pattern NEWLINES_PATTERN =
            Pattern.compile("\\s*(?:[\\r\\n]+\\s*)+");

    /**
     * A simple regular expression for strings that contain only
     * punctuation characters.
     */
    public static final Pattern PUNCTUATION_PATTERN =
        Pattern.compile("\\p{P}+");

    /**
     * A simple regular expression for strings that contain only
     * punctuation and symbol characters.
     */
    public static final Pattern PUNCTUATION_SYMBOL_PATTERN =
        Pattern.compile("[\\p{P}\\p{S}]+");

    /**
     * A simple regular expression for whitespace.
     */
    public static final Pattern WHITESPACE_PATTERN =
        Pattern.compile("\\s+");

    /**
     * Counts how often a character is repeated at the begin of a string.
     *
     * @param str the string to check
     * @param ch the character to count
     * @return how often the character is repeated at the begin of the string
     * (0 if the string starts with another character or is empty)
     */
    public static int countFirst(final String str, final char ch) {
        for (int i = 0; i < str.length(); i++) {
            if (str.charAt(i) != ch) {
                // found first non-match
                return i;
            }
        }

        // whole string matches
        return str.length();
    }

    /**
     * Counts how often a character is repeated at the end of a string.
     *
     * @param str the string to check
     * @param ch the character to count
     * @return how often the character is repeated at the end of the string
     * (0 if the string ends with another character or is empty)
     */
    public static int countLast(final String str, final char ch) {
        final int length = str.length();

        for (int i = 0; i < length; i++) {
            if (str.charAt(length - i - 1) != ch) {
                // found last non-match
                return i;
            }
        }

        // whole string matches
        return length;
    }

    /**
     * Checks that a string is a printable name, meaning it has at at least
     * one character and does not contain any whitespace.
     *
     * @param string the string to check
     * @throws IllegalArgumentException if the given string null or empty
     * or contains whitespace
     */
    public static void ensurePrintableName(final String string)
    throws IllegalArgumentException {
        if (StringUtils.isEmpty(string)) {
            throw new IllegalArgumentException("Name is null or empty: "
                    + string);
        }
        if (WHITESPACE_PATTERN.matcher(string).find()) {
            throw new IllegalArgumentException("Name contains whitespace: '"
                    + string + "'");
        }
    }

    /**
     * Helper method for building a regular expression {@link Pattern} by
     * combining several alternatives.
     *
     * @param alternatives the alternatives to combine
     * @return a pattern string containing the joined alternatives; two or
     * more alternatives are combined in a non-capturing group; a single
     * alternative is just returned as is; if the array is empty, an empty
     * string is returned
     */
    public static String joinAlternatives(final String[] alternatives) {
        final int length = alternatives.length;
        if (length >= 2) {
            // open non-capturing group
            final StringBuffer result = new StringBuffer("(?:");
            for (int i = 0; i < length; i++) {
                result.append(alternatives[i]);
                if (i < length - 1) {
                    result.append("|");   // separate alternatives
                } else {
                    result.append(")");   // close group
                }
            }
            return result.toString();
        } else if (length == 1) {
            // just return the single element
            return alternatives[0];
        } else {
            // return empty string
            return "";
        }
    }

    /**
     * Performs multiple replace-all operations on a text. The replacements
     * are performed in the order of the key-set iterator of the given map.
     *
     * @param input the character sequence to perform the replacements on
     * @param replacements a mapping of regular expression
     * {@link java.util.regex.Pattern}s to replacement {@link java.lang.String}s
     * @return the string constructed by performing all replacements
     */
    public static String multipleReplaceAll(final CharSequence input,
            final Map replacements) {
        final Iterator patternIterator = replacements.keySet().iterator();
        Pattern currentPattern;
        String currentReplace;
        String text = input.toString();

        while (patternIterator.hasNext()) {
            currentPattern = (Pattern) patternIterator.next();
            currentReplace = (String) replacements.get(currentPattern);
            text = replaceAll(text, currentPattern, currentReplace);
        }
        return text;
    }

    /**
     * Normalizes the whitespace in a string, replacing all internal whitespace
     * sequences with a single space character and trimming any leading and
     * trailing whitespace.
     *
     * @param input the string to normalize
     * @return the normalized string
     */
    public static String normalize(final String input) {
        // replace whitespace by space char
        final String result = replaceAll(input, WHITESPACE_PATTERN, " ");
        // trim result
        return result.trim();
    }

    /**
     * Helper method that handles the actual replacement process.
     *
     * @param input the string to process
     * @param matcher a matcher on the pattern
     * @param replacement the replacement string
     * @param doReset whether to reset this matcher
     * @return the resulting string; or a reference to the <code>input</code>
     * string if no replacements were made
     */
    private static String replaceAll(final String input, final Matcher matcher,
            final String replacement, final boolean doReset) {
        if (doReset) {
            matcher.reset(input);
        }
        boolean found = matcher.find();

        if (found) {
            final StringBuffer result = new StringBuffer();
            do {
                matcher.appendReplacement(result, replacement);
                found = matcher.find();
            } while (found);

            matcher.appendTail(result);
            return result.toString();
        } else {
            // nothing to replace
            return input;
        }
    }

    /**
     * Replaces each substring of the <code>input</code> matched by the
     * given {@linkplain Pattern pattern} matcher with the given
     * replacement. See {@link Matcher#replaceAll(java.lang.String)} for details
     * of the replacement process and special characters in the
     * <code>replacement</code> string.
     *
     * <p>This method only returns a new string if there is at least one match
     * to replace. Otherwise the reference to the <code>input</code> object is
     * returned. Thus you can use the <code>==</code> operator to find out
     * whether replacements have been made, it is not necessary to use
     * {@link String#equals(java.lang.Object)}. When there is nothing to
     * replace, it might be more efficient than
     * {@link Matcher#replaceAll(java.lang.String)} (and certainly than
     * {@link String#replaceAll(java.lang.String, java.lang.String)}, because
     * (as of JDK 1.4.2) these methods always create and return new objects.
     *
     * <p>Matchers are stateful and not thread-safe. It is not necessary to
     * {@link Matcher#reset()} the matcher prior to calling this method but you
     * should reset it if you want to used it in other matching operations
     * afterwards.
     *
     * @param input the string to process
     * @param matcher a matcher on the pattern
     * @param replacement the replacement string
     * @return the resulting string; or a reference to the <code>input</code>
     * string if no replacements were made
     */
    public static String replaceAll(final String input, final Matcher matcher,
            final String replacement) {
        return replaceAll(input, matcher, replacement, true);
    }

    /**
     * Shortens a string, inserting an ellipsis ("...") in the middle if the
     * string is too long. Specifically:
     *
     * <ul>
     * <li>If the length of the <code>input</code> String isn't larger than
     * <code>startChars + endChars + 3</code>, return it.
     * <li>Otherwise return the first <code>startChars</code> characters of the
     * <code>input</code> String, followed by "..." (an ellipsis) and the last
     * <code>endChars</code> characters of the String
     * </ul>
     *
     * <p>This method is similar to
     * {@link org.apache.commons.lang.StringUtils#abbreviate(String, int)},
     * but the ellipsis is inserted in the middle of the string, not at the
     * end.
     *
     * @param input the input string
     * @param startChars the number of characters to include before the ellipsis
     * @param endChars the number of characters to include after the ellipsis
     * @return a shortened string, as described above
     */
    public static String shorten(final String input, final int startChars,
            final int endChars) {
        final int length = input.length();

        if (length <= startChars + endChars + 3) {
            // string is short enough -- return as is
            return input;
        } else {
            // return first <startChars> characters + ellipsis + last
            // <endChars> characters
            return input.substring(0, startChars) + "..."
            + input.substring(length - endChars);
        }
    }

    /**
     * Delegates to {@link #shorten(String, int, int)}, using the same number
     * of characters at the start and the end of the shortened string.
     *
     * @param input the input string
     * @param numChars the number of characters to to use for
     * both <code>startChars</code> and <code>endChars</code> parameter
     * @return the shortened string
     */
    public static String shorten(final String input, final int numChars) {
        return shorten(input, numChars, numChars);
    }

    /**
     * Checks whether a string contains only punctuation characters.
     *
     * @param text the test to check
     * @return <code>true</code> iff the text contains one or more
     * punctuation characters and no other characters
     */
    public static boolean punctuation(final CharSequence text) {
        return PUNCTUATION_PATTERN.matcher(text).matches();
    }

    /**
     * Checks whether a string contains only punctuation and symbol characters.
     *
     * @param text the test to check
     * @return <code>true</code> iff the text contains one or more
     * punctuation or symbol characters and no other characters
     */
    public static boolean punctuationOrSymbol(final CharSequence text) {
        return PUNCTUATION_SYMBOL_PATTERN.matcher(text).matches();
    }


    /**
     * Delegates to {@link #shorten(String, int, int)}, showing up to
     * <strong>24</strong> characters at the start and the end of the shortened
     * string.
     *
     * @param input the input string
     * @return the shortened string
     */
    public static String shorten(final String input) {
        return shorten(input, 24);
    }

    /**
     * Replaces each substring of the <code>input</code> that matches the given
     * {@link Pattern} with the given replacement. See
     * {@link Matcher#replaceAll(java.lang.String)} for details of the
     * replacement process and special characters in the
     * <code>replacement</code> string.
     *
     * <p>This method only returns a new string if there is at least one match
     * to replace. Otherwise the reference to the <code>input</code> object is
     * returned. Thus you can use the <code>==</code> operator to find out
     * whether replacements have been made, it is not necessary to use
     * {@link String#equals(java.lang.Object)}.
     *
     * <p>This method is thread-safe since pattern objects are stateless. On the
     * other hand, it needs to create a new {@link Matcher} object, thus
     * {@link #replaceAll(String, Matcher, String)} is more efficient for
     * multiple replacements on the same pattern.
     *
     * @param input the string to process
     * @param pattern the regular expression {@link Pattern} to replace
     * @param replacement the replacement string
     * @return the resulting string; or a reference to the <code>input</code>
     * string if no replacements were made
     */
    public static String replaceAll(final String input, final Pattern pattern,
            final String replacement) {
        final Matcher matcher = pattern.matcher(input);
        // no need to reset newly created matcher
        return replaceAll(input, matcher, replacement, false);
    }

    /**
     * Splits a text into an array of lines. Only the textual contents of
     * non-empty lines are retained; empty lines and training and leading
     * whitespace are removed.
     *
     * @param input the text to split
     * @return an array of the lines contained in the text; each line is trimmed
     * (trailing and leading whitespace is removed) and empty lines are
     * suppressed
     */
    public static String[] splitLines(final CharSequence input) {
        // remove whitespace at the very beginning and end
        final String trimmed = input.toString().trim();
        return NEWLINES_PATTERN.split(trimmed);
    }

    /**
     * Splits a text into an array of lines, without trimming lines and
     * discarding empty lines.
     *
     * @param input the text to split
     * @return an array of the lines contained in the text
     */
    public static String[] splitLinesExact(final CharSequence input) {
        return NEWLINE_PATTERN.split(input);
    }

    /**
     * Splits a string around whitespace.
     *
     * @param input the string to split
     * @return an array of strings computed by splitting the input
     */
    public static String[] splitString(final String input) {
        return splitString(input, -1);
    }

    /**
     * Splits a string around whitespace. The number of returned subsequences
     * won't be higher than the specified <code>splitMaximum</code>.
     * If splitting results in more subsequences, only the <em>last</em>
     * <code>splitMaximum</code> are kept, while the other ones are discarded.
     * This implementation splits around the {@link #WHITESPACE_PATTERN}.
     *
     * @param input the string to split
     * @param splitMaximum the maximum number of subsequences to keep;
     * or <code>-1</code> if all subsequences should be kept
     * @return an array of strings computed by splitting the input; will
     * contain at least 1 and at most <code>splitMaximum</code> elements
     */
    public static String[] splitString(final String input,
            final int splitMaximum) {
        return splitString(input, WHITESPACE_PATTERN, splitMaximum);
    }

    /**
     * Splits a string around whitespace. The number of returned subsequences
     * won't be higher than the specified <code>splitMaximum</code>.
     * If splitting results in more subsequences, only the <em>last</em>
     * <code>splitMaximum</code> are kept, while the other ones are discarded.
     *
     * @param input the string to split
     * @param whitespacePattern the pattern around which to split
     * @param splitMaximum the maximum number of subsequences to keep;
     * or <code>-1</code> if all subsequences should be kept
     * @return an array of strings computed by splitting the input; will
     * contain at most <code>splitMaximum</code> elements
     */
    public static String[] splitString(final String input,
            final Pattern whitespacePattern, final int splitMaximum) {
        final String[] results;
        final String[] allSplits = whitespacePattern.split(input);
        final int surplus;
        if (splitMaximum >= 0) {
            surplus = allSplits.length - splitMaximum;
        } else {
            // return all matches
            surplus = -1;
        }

        if (surplus > 0) {
            // keep only last splitMaximum splits
            results = new String[splitMaximum];
            for (int i = 0; i < results.length; i++) {
                results[i] = allSplits[i + surplus];
            }
        } else {
            // keep all splits
            results = allSplits;
        }

        return results;
    }

    /**
     * Private constructor prevents creation of instances.
     */
    private TextUtils() {
        super();
    }

}
