/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.text;

import java.util.Iterator;

import org.apache.commons.collections.Bag;
import org.apache.commons.collections.bag.HashBag;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * A container that keeps track of the tokens in a document. Instances of this
 * class are not thread-safe; if you want to share a single instance between
 * different thread, you have to ensure proper synchronization.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2004/11/17 15:25:11 $, $Author: siefkes $
 */
public class TokenContainer {

    /**
     * Returns the repetition of the first token of the {@link #getLast() last}
     * {@link #add(String) added} string in the original text (counting starts
     * with 0, as the first occurrence is the "0th repetition").
     */
    private int firstTokenInLastRep;

    /**
     * Returns the index of the first token of the {@link #getLast() last}
     * {@link #add(String) added} string in the original text (indexing starts
     * with 0).
     */
    private int firstTokenInLastIndex;

    /**
     * Whether there is whitespace before/at the start of the string
     * added this container by the last {@link #add(String)} operation.
     */
    private boolean whitespaceBeforeLast;

    /**
     * Whether there is whitespace after/at the end of the string
     * added this container by the last {@link #add(String)} operation.
     */
    private boolean whitespaceAfterLast;

    /**
     * A multi-set of all the tokens added to this container.
     */
    private final Bag tokens = new HashBag();

    /**
     * A multi-set of the last batch of tokens added to this container.
     */
    private final Bag lastTokens = new HashBag();

    /**
     * The last string added to this container.
     */
    private String last = null;

    /**
     * The tokenizer used to split strings into tokens.
     */
    private final TextTokenizer tokenizer;

    /**
     * Creates a new instance.
     *
     * @param tFactory used to instantiate the employed tokenizer
     */
    public TokenContainer(final TokenizerFactory tFactory) {
        super();
        tokenizer = tFactory.createTokenizer("");
    }

    /**
     * Adds text to this container. The specified string is split into a series
     * of tokens, the token count is increased accordingly for all contained
     * tokens.
     *
     * @param text the text to add
     */
    public void add(final String text) {
        // reset lastTokens
        lastTokens.clear();

        // tokenize text and update token count (in both multi-sets)
        StringBuffer lastBuffer = new StringBuffer();
        String token;
        tokenizer.reset(text);
        boolean isFirst = true;

        while ((token = tokenizer.nextToken()) != null) {
            if (isFirst) {
                // very first token. Determine repetition+index: count prior to
                // adding the token, as counting starts with 0
                firstTokenInLastRep = tokens.getCount(token);
                firstTokenInLastIndex = tokens.size();

                // whitespace before: true iff there was whitespace after the
                // previous "last" or if there is whitespace at the begin of
                // this one
                whitespaceBeforeLast =
                    whitespaceAfterLast || tokenizer.hasPrecedingWhitespace();
                isFirst = false;
            } else {
                // already got token(s)
                if (tokenizer.hasPrecedingWhitespace()) {
                    lastBuffer.append(' ');
                }
            }

            // add token to lastBuffer; count it in boths sets
            lastBuffer.append(token);
            tokens.add(token);
            lastTokens.add(token);
        }

        // remember if there is whitespace after the last token
        whitespaceAfterLast = tokenizer.hasPrecedingWhitespace();

        // store normalized string as "last"
        last = lastBuffer.toString();
    }

    /**
     * Returns the cardinality of the given token in this container.
     *
     * @param token the token to check
     * @return the number of copies of the specified token in this container,
     * <code>&gt;= 0</code>
     */
    public int getCount(final String token) {
        return tokens.getCount(token);
    }

    /**
     * Returns the index of the first token of the {@link #getLast() last}
     * {@link #add(String) added} string in the original text (indexing starts
     * with 0).
     *
     * @return the value of the attribute
     */
    public int getFirstTokenInLastIndex() {
        return firstTokenInLastIndex;
    }

    /**
     * Returns the repetition of the first token of the {@link #getLast() last}
     * {@link #add(String) added} string in the original text (counting starts
     * with 0, as the first occurrence is the "0th repetition").
     *
     * @return the value of the attribute
     */
    public int getFirstTokenInLastRep() {
        return firstTokenInLastRep;
    }

    /**
     * Returns the cardinality of the given token in the text added by the last
     * {@link #add(String)} operation.
     *
     * @param token the token to check
     * @return the number of copies of the specified token in the text added
     * last, <code>&gt;= 0</code>
     */
    public int getLastCount(final String token) {
        return lastTokens.getCount(token);
    }

    /**
     * Returns a trimmed and whitespace-normalized representation of the string
     * added this container by the last {@link #add(String)} operation.
     * Starting and trailing whitespace is removed; each internal whitespace
     * is converted into a single space charater.
     *
     * @return the normalized representation of the last string added to this
     * container
     */
    public String getLast() {
        return last;
    }

    /**
     * Whether there is whitespace after the {@link #getLast() last} added
     * string.
     *
     * @return <code>true</code> iff there is whitespace after/at the end of
     * the string
     */
    public boolean isWhitespaceAfterLast() {
        return whitespaceAfterLast;
    }

    /**
     * Whether there is whitespace before the {@link #getLast() last} added
     * string.
     *
     * @return <code>true</code> iff there is whitespace before/at the start of
     * the string
     */
    public boolean isWhitespaceBeforeLast() {
        return whitespaceBeforeLast;
    }

    /**
     * Whether the text added by the last {@link #add(String)} operation
     * contains the specified token.
     *
     * @param token the token to check
     * @return <code>true</code> iff the specified argument is contained as a
     * word or number token in the last added string.
     */
    public boolean lastContains(final String token) {
        return lastTokens.contains(token);
    }

    /**
     * Returns an iterator over the word and number tokens added by the last
     * {@link #add(String)} operation. The iterator contains each token only
     * once (no matter how often it occurred in the last string); the tokens
     * are iterated in no particular order.
     *
     * @return an iterator over the last added tokens
     */
    public Iterator lastIterator() {
        return lastTokens.uniqueSet().iterator();
    }

    /**
     * Returns the token number of tokens counted by this instances (including
     * duplicates).
     * 
     * @return the number of tokens counted
     */
    public int size() {
        return tokens.size();
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("tokens", tokens)
            .append("last", last)
            .append("whitespace before last", whitespaceBeforeLast)
            .append("whitespace after last", whitespaceAfterLast)
            .append("repetition of first token in last", firstTokenInLastRep)
            .append("index of first token in last", firstTokenInLastIndex)
            .append("tokenizer", tokenizer)
            .toString();
    }

}
