/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.text;

import org.apache.commons.collections.Bag;
import org.apache.commons.collections.bag.HashBag;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * A simple container that keeps track of the tokens in a document. This class
 * differs from {@link de.fu_berlin.ties.text.TokenContainer} by not doing
 * any tokenization itself, relying on external tokenization instead.
 *
 * <p>Instances of this class are not thread-safe; if you want to share a single
 * instance between different thread, you have to ensure proper synchronization.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2004/11/09 16:29:34 $, $Author: siefkes $
 */
public class TokenCounter {

    /**
     * Whether there is whitespace before the {@link #getLast() last} added
     * token.
     */
    private boolean whitespaceBeforeLast;

    /**
     * Whether there is whitespace after the {@link #getLast() last} added
     * token.
     */
    private boolean whitespaceAfterLast;

    /**
     * A multi-set of all the tokens added to this container.
     */
    private final Bag tokens = new HashBag();

    /**
     * The last token added to this container.
     */
    private String last = null;

    /**
     * Creates a new instance.
     */
    public TokenCounter() {
        super();
    }

    /**
     * Adds a token to this instance.
     *
     * @param whitespaceBefore whether there is whitespace before the token
     * @param token the token to add
     */
    public void add(final boolean whitespaceBefore, final String token) {
        // whitespace before: true iff there was whitespace after the
        // previous "last" or if there is whitespace before this one
        whitespaceBeforeLast = whitespaceAfterLast || whitespaceBefore;

        // count + store token
        tokens.add(token);
        last = token;
    }

    /**
     * Adds whitespace to this instance, setting
     * {@link #isWhitespaceAfterLast()} to <code>true</code>.
     */
    public void addWhitespace() {
        whitespaceAfterLast = true;
    }

    /**
     * Returns the cardinality of the given token in this container.
     *
     * @param token the token to check
     * @return the number of copies of the specified token in this container,
     * <code>&gt;= 0</code>
     */
    public int getCount(final String token) {
        return tokens.getCount(token);
    }

    /**
     * Returns the repetition of the {@link #getLast() last} added token in the
     * original text (counting starts with 0, as the first occurrence is the
     * "0th repetition").
     *
     * @return the value of the attribute
     */
    public int getLastRep() {
        // subtract 1, since counting starts with 0
        return tokens.getCount(last) - 1;
    }

    /**
     * Returns the last added token.
     *
     * @return the value of the attribute
     */
    public String getLast() {
        return last;
    }

    /**
     * Whether there is whitespace after the {@link #getLast() last} added
     * token.
     *
     * @return <code>true</code> iff there is whitespace after the token
     */
    public boolean isWhitespaceAfterLast() {
        return whitespaceAfterLast;
    }

    /**
     * Whether there is whitespace before the {@link #getLast() last} added
     * token.
     *
     * @return <code>true</code> iff there is whitespace before the token
     */
    public boolean isWhitespaceBeforeLast() {
        return whitespaceBeforeLast;
    }

    /**
     * Returns the token number of tokens counted by this instances (including
     * duplicates).
     * 
     * @return the number of tokens counted
     */
    public int size() {
        return tokens.size();
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("tokens", tokens)
            .append("last", last)
            .append("whitespace before last", whitespaceBeforeLast)
            .append("whitespace after last", whitespaceAfterLast)
            .toString();
    }

}
