/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.text;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * Stores details on a token in a document. Instances of this class are
 * immutable and thus thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/11/17 15:25:11 $, $Author: siefkes $
 */
public class TokenDetails {

    /**
     * The token itself.
     */
    private final String token;

    /**
     * The repetition of the token in the original text (counting starts with 0,
     * as the first occurrence is the "0th repetition").
     */
    private final int rep;

    /**
     * The index of the token in the original text (indexing starts with 0).
     */
    private final int index;

    /**
     * Whether there is whitespace before the token.
     */
    private final boolean whitespaceBefore;

    /**
     * Creates a new instance.
     *
     * @param theToken the token to represent
     * @param tokenRep the repetition of the token in the document (counting
     * starts with 0, as the first occurrence is the "0th repetition")
     * @param tokenIndex the index of the token in the document (counting
     * starts with 0 for the very first token)
     * @param wsBefore whether there is whitespace before the main
     * <code>token</code> (either at the end of <code>left</code> or in the
     * preceding element)
     */
    public TokenDetails(final String theToken, final int tokenRep,
            final int tokenIndex, final boolean wsBefore) {
        super();
        token = theToken;
        rep = tokenRep;
        index = tokenIndex;
        whitespaceBefore = wsBefore;
    }

    /**
     * Returns the index of the token in the original text (indexing starts
     * with 0).
     *
     * @return the value of the attribute
     */
    public int getIndex() {
        return index;
    }

    /**
     * Returns the repetition of the token in the original text (counting starts
     * with 0, as the first occurrence is the "0th repetition").
     *
     * @return the value of the attribute
     */
    public int getRep() {
        return rep;
    }

    /**
     * Returns the token itself.
     * @return the value of the attribute
     */
    public String getToken() {
        return token;
    }

    /**
     * Whether there is whitespace before the token.
     * @return the value of the attribute
     */
    public boolean isWhitespaceBefore() {
        return whitespaceBefore;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("token", token)
            .append("repetition", rep)
            .append("index", index)
            .append("whitespace before", whitespaceBefore)
            .toString();
    }

}
