/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.text;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.TiesConfiguration;

/**
 * Factory for creating {@link de.fu_berlin.ties.text.TextTokenizer}s of
 * different types.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2004/04/13 07:08:35 $, $Author: siefkes $
 */
public class TokenizerFactory {

    /**
     * Configuration key for the array of regular expressions defining the token
     * types accepted by the tokenizer.
     */
    public static final String CONFIG_TOKEN_PATTERNS = "tokenizer.pattern";

    /**
     * Configuration key for the regular expression giving the whitespace
     * accepted by the tokenizer.
     */
    public static final String CONFIG_WHITESPACE_PATTERN =
        "tokenizer.whitespace";

    /**
     * Pattern string capturing whitespace and control/other characters.
     */
    public static final String WHITESPACE_CONTROL_OTHER = "[\\p{Z}\\p{C}]*";

    /**
     * Static factory method to create an instance for tokenizing alphanumeric
     * and symbol sequences and puntuation. Token types:
     *
     * <ul>
     * <li>sequences of letters and digits (incl. marks, Unicode categories
     * "L", "M", "N") -- full token is captured for
     * {@link TextTokenizer#capturedText()}</li>
     * <li>sequences of symbols ("S" category) -- nothing is captured</li>
     * <li>a single punctuation sign, possibly repeated ("P" category) --
     * nothing is captured</li>
     * </ul>
     *
     * <p>When you are only interested in words and numbers (e.g. for indexing),
     * you can use the {@link TextTokenizer#capturedText() captured text} --
     * it will contain the full token for alphanumeric sequences, it will be
     * empty for symbols and punctuation.
     *
     * <p>The whitespace pattern comprised a sequence of whitespace and
     * control/other characters ("C" and "Z" categories).
     *
     * @param text the text to tokenize
     * @return the created tokenizer
     */
    public static TextTokenizer createAlnumTokenizer(final CharSequence text) {
        final TextTokenizer result = new TextTokenizer(new String[] {
                "([\\p{L}\\p{M}\\p{N}]+)", // alphanumeric
                "\\p{S}+",        // symbols
                "(\\p{P})\\1*",   // single punctuation sign, possibly repeated
            },
            WHITESPACE_CONTROL_OTHER, // whitespace + control/other characters
            text);
        return result;
    }

    /**
     * Static factory method to create an instance for tokenizing according to
     * Unicode categories. Token types:
     *
     * <ul>
     * <li>sequences of letters (incl. marks, "L" and "M" categories) -- full
     * token is captured for {@link TextTokenizer#capturedText()}</li>
     * <li>sequences of digits (integral numbers, "N" category) -- full token is
     * captured</li>
     * <li>sequences of symbols ("S" category) -- nothing is captured</li>
     * <li>sequences of punctuation ("P" category) -- nothing is captured</li>
     * </ul>
     *
     * <p>When you are only interested in words and numbers (e.g. for indexing),
     * you can use the {@link TextTokenizer#capturedText() captured text} --
     * it will contain the full token for letter and digit sequences, it will
     * be empty for symbols and punctuation.
     *
     * <p>The whitespace pattern comprised a sequence of whitespace and
     * control/other characters ("C" and "Z" categories).
     *
     * @param text the text to tokenize
     * @return the created tokenizer
     */
    public static TextTokenizer createCategoryTokenizer(
                final CharSequence text) {
        final TextTokenizer result = new TextTokenizer(new String[] {
                "([\\p{L}\\p{M}]+)",    // letters incl. marks, fully captured
                "(\\p{N}+)",            // numbers, fully captured
                "\\p{S}+",              // symbols
                "\\p{P}+",              // punctuation
            },
            WHITESPACE_CONTROL_OTHER,   // whitespace + control/other characters
            text);
        return result;
    }

    /**
     * Static factory method to create an instance that uses the "thorough"
     * patterns listed below.
     *
     * <ul>
     * <li>sequences of letters (incl. marks)</li>
     * <li>sequences of digits (integral numbers)</li>
     * <li>sequences of math symbols</li>
     * <li>sequences of currency symbols</li>
     * <li>sequences of other symbols (modifiers and misc)</li>
     * <li>a single punctuation sign, possibly repeated</li>
     * </ul>
     *
     * <p>These patterns don't contain any useful information for
     * {@link TextTokenizer#capturedText()}.
     *
     * <p>The whitespace pattern comprised a sequence of whitespace and
     * control/other characters.
     *
     * @param text the text to tokenize
     * @return the created tokenizer
     */
    public static TextTokenizer createThoroughTokenizer(
            final CharSequence text) {
        final TextTokenizer result = new TextTokenizer(new String[] {
                "[\\p{L}\\p{M}]+",    // letters incl. marks
                "\\p{N}+",            // numbers
                "\\p{Sm}+",           // math symbols
                "\\p{Sc}+",           // currency symbols
                "[\\p{Sk}\\p{So}]+",  // modifier and other symbols
                "(\\p{P})\\1*",   // single punctuation sign, possibly repeated
            },
            WHITESPACE_CONTROL_OTHER, // whitespace + control/other characters
            text);
        return result;
    }

    /**
     * Array of regular expressions strings defining the token types accepted by
     * the tokenizer.
     */
    private final String[] tokenPatterns;

    /**
     * Regular expression giving the whitespace accepted by the tokenizer.
     */
    private final String whitespacePattern;

    /**
     * Creates a new instance from the {@link #CONFIG_TOKEN_PATTERNS} and
     * {@link #CONFIG_WHITESPACE_PATTERN} keys of the provided configuration.
     *
     * @param config the configuration to use
     */
    public TokenizerFactory(final TiesConfiguration config) {
        this(config, null);
    }

    /**
     * Creates a new instance from the {@link #CONFIG_TOKEN_PATTERNS} and
     * {@link #CONFIG_WHITESPACE_PATTERN} keys of the provided configuration,
     * {@linkplain TiesConfiguration#adaptKey(String, String) adapted} by
     * appending the <code>suffix</code>.
     *
     * @param config the configuration to use
     * @param suffix the suffix to append to the keys
     */
    public TokenizerFactory(final TiesConfiguration config,
                            final String suffix) {
        super();
        final String tokenKey = config.adaptKey(CONFIG_TOKEN_PATTERNS, suffix);
        final String whitespaceKey =
            config.adaptKey(CONFIG_WHITESPACE_PATTERN, suffix);
        tokenPatterns = config.getStringArray(tokenKey);
        whitespacePattern = config.getString(whitespaceKey);
    }

    /**
     * Factory method to create an instance from the configured token
     * and whitespace patterns.
     *
     * @param text the text to tokenize
     * @return the created tokenizer
     */
    public TextTokenizer createTokenizer(final CharSequence text) {
        final TextTokenizer result =
            new TextTokenizer(tokenPatterns, whitespacePattern, text);
        return result;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("token patterns", ArrayUtils.toString(tokenPatterns))
            .append("whitespace pattern", whitespacePattern)
            .toString();
    }

}
