/*
 * Copyright (C) 2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.util;

import java.util.Iterator;
import java.util.Map;

import org.apache.commons.collections.map.AbstractLinkedMap;
import org.apache.commons.collections.map.LRUMap;

/**
 * A fixed-size map that uses an flexible adaptable strategy for pruning
 * entries based on LRU (pruning one the least recently used entries).
 * Instances of this class are not thread-safe and must be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.7 $, $Date: 2004/06/04 17:13:22 $, $Author: siefkes $
 */
public class AdaptableLRUMap extends LRUMap {

    /**
     * Used to decide which entries to prune.
     */
    private final Pruner pruner;

    /**
     * The number of candidates to consider for each pruning operation.
     */
    private final int candidateNumber;

    /**
     * The number of elements to remove by each pruning operation.
     */
    private final int pruneNumber;

    /**
     * Constructs a new, empty map with the specified initial capacity and
     * the default load factor. If the number of candidates is one, this
     * class behaves like a normal {@link LRUMap}; otherwise the
     * {@link Pruner} is consulted to decide which entry to prune.
     *
     * @param maxSize the maximum size of the map
     * @param prun Used to decide which entries to prune
     * @param candidates The number of candidates to consider for each pruning
     * operation
     * @param pruneNum The number of elements to remove by each pruning
     * operation, must not be larger than <code>candidates</code>
     * @throws IllegalArgumentException if one of the input parameters is
     * invalid
     */
    public AdaptableLRUMap(final int maxSize, final Pruner prun,
                           final int candidates, final int pruneNum)
                           throws IllegalArgumentException {
        super(maxSize);
        checkArguments(maxSize, candidates, pruneNum);

        pruner = prun;
        candidateNumber = candidates;
        pruneNumber = pruneNum;
    }

    /**
     * Constructs a new, empty map with the specified initial capacity and
     * load factor. If the number of candidates is one, this
     * class behaves like a normal {@link LRUMap}; otherwise the
     * {@link Pruner} is consulted to decide which entry to prune.
     *
     * @param maxSize the maximum size of the map
     * @param loadFactor the load factor
     * @param prun Used to decide which entries to prune
     * @param candidates The number of candidates to consider for each pruning
     * operation
     * @param pruneNum The number of elements to remove by each pruning
     * operation, must not be larger than <code>candidates</code>
     * @throws IllegalArgumentException if one of the input parameters is
     * invalid
     */
    public AdaptableLRUMap(final int maxSize, final float loadFactor,
                           final Pruner prun, final int candidates,
                           final int pruneNum)
                           throws IllegalArgumentException {
        super(maxSize, loadFactor);
        checkArguments(maxSize, candidates, pruneNum);

        pruner = prun;
        candidateNumber = candidates;
        pruneNumber = pruneNum;
    }

    /**
     * Helper method that checks whether constructor arguments are valid and
     * throws an exception if this is not the case.
     *
     * @param maxSize the maximum size of the map
     * @param candidates The number of candidates to consider for each pruning
     * operation
     * @param pruneNum The number of elements to remove by each pruning
     * operation, must not be larger than <code>candidates</code>
     * @throws IllegalArgumentException if one of the input parameters is
     * invalid
     */
    private void checkArguments(final int maxSize, final int candidates,
        final int pruneNum) throws IllegalArgumentException {
        if ((candidates < 1) || (candidates > maxSize)) {
            throw new IllegalArgumentException("Number of pruning candidates "
                    + "must be in range from 1 to maxSize (" + maxSize + "): "
                    + candidates);
        }

        if ((pruneNum < 1) || (pruneNum > candidates)) {
            throw new IllegalArgumentException("Number of elements to prun "
                    + "must be in range from 1 to number of candidates ("
                    + candidates + "): " + pruneNum);
        }
    }

    /**
     * Returns the  number of candidates considered for each pruning operation.
     * @return the value of the attribute
     */
    public int getCandidateNumber() {
        return candidateNumber;
    }

    /**
     * Returns the number of elements removed by each pruning operation.
     * @return the value of the attribute
     */
    public int getPruneNumber() {
        return pruneNumber;
    }

    /**
     * Controls removal of one of the least recently used entries from the map.
     * Lets the pruner decide if there are multiple candidates to consider.
     *
     * @param entry the least recently used entry
     * @return <code>true</code> if the superclass should delete the
     * specified entry; <code>false</code> if this class took the necessary
     * action by deleting the first {@link #getPruneNumber()} entries selected
     * by the pruner
     */
    protected boolean removeLRU(final AbstractLinkedMap.LinkEntry entry) {
        if (candidateNumber > 1) {
            // collect candidates for the pruner
            final int consideredCandidates = Math.min(candidateNumber, size());
            final Map.Entry[] candidates = new Map.Entry[consideredCandidates];
            final Iterator entryIter = entrySet().iterator();

            for (int i = 0; i < candidates.length; i++) {
                //Util.LOG.debug("Calling entryIter.next");
                candidates[i] = (Map.Entry) entryIter.next();
                //Util.LOG.debug("Called entryIter.next");
            }

            // let the pruner decide + delete the chosen entry
            final Map.Entry[] resortedEntries =
                pruner.sortForPruning(candidates);

            // remove specified number of entries
            final int entriesToPrune =
                Math.min(pruneNumber, resortedEntries.length);
            for (int i = 0; i < entriesToPrune; i++) {
                remove(resortedEntries[i].getKey());
            }

            // deleted chosen entries, no need for further action
            return false;
        } else {
            // standard LRU behavior: let superclass delete entry
            return true;
        }
    }

}
