/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * A static class that provides utility methods for working with
 * {@link java.util.Collection}s and arrays. No instances of this class can be
 * created, only the static members should be used.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/12/06 17:59:41 $, $Author: siefkes $
 */
public final class CollectionUtils {

    /**
     * Default separator used to {@linkplain #flatten(Object[]) flatten} array
     * if no other separator is specified: {@value} (a single space).
     */
    public static final String SEPARATOR = " ";

    /**
     * Wraps an array into a set. Duplicates will be ignored.
     *
     * @param array the array to wrap
     * @return a set containing the contents of the set (in random order,
     * without duplicates)
     */
    public static <T> Set<T> arrayAsSet(final T[] array) {
        return new HashSet<T>(Arrays.asList(array));
    }

    /**
     * Converts an object array into a boolean array, calling
     * {@link Util#asBoolean(Object)} on each element.
     *
     * @param objArray the array of objects to convert
     * @return the converted boolean array
     * @throws IllegalArgumentException if an object's <code>toString()</code>
     * output cannot be parsed as a boolean
     */
    public static boolean[] asBooleanArray(final Object[] objArray)
    throws IllegalArgumentException {
        final boolean[] result = new boolean[objArray.length];
        for (int i = 0; i < objArray.length; i++) {
            result[i] = Util.asBoolean(objArray[i]);
        }
        return result;
    }

    /**
     * Converts an object array into a byte array, calling
     * {@link Util#asByte(Object)} on each element.
     *
     * @param objArray the array of objects to convert
     * @return the converted byte array
     * @throws NumberFormatException if an object's <code>toString()</code>
     * output does not contain a parsable byte
     */
    public static byte[] asByteArray(final Object[] objArray)
    throws NumberFormatException {
        final byte[] result = new byte[objArray.length];
        for (int i = 0; i < objArray.length; i++) {
            result[i] = Util.asByte(objArray[i]);
        }
        return result;
    }

    /**
     * Converts an object array into a char array, calling
     * {@link Util#asChar(Object)} on each element.
     *
     * @param objArray the array of objects to convert
     * @return the converted char array
     * @throws IndexOutOfBoundsException if the object's <code>toString()</code>
     * output is the empty string (after trimming outer whitespace)
     */
    public static char[] asCharArray(final Object[] objArray)
    throws IndexOutOfBoundsException {
        final char[] result = new char[objArray.length];
        for (int i = 0; i < objArray.length; i++) {
            result[i] = Util.asChar(objArray[i]);
        }
        return result;
    }

    /**
     * Converts an object array into a double array, calling
     * {@link Util#asDouble(Object)} on each element.
     *
     * @param objArray the array of objects to convert
     * @return the converted double array
     * @throws NumberFormatException if an object's <code>toString()</code>
     * output does not contain a parsable double
     */
    public static double[] asDoubleArray(final Object[] objArray)
    throws NumberFormatException {
        final double[] result = new double[objArray.length];
        for (int i = 0; i < objArray.length; i++) {
            result[i] = Util.asDouble(objArray[i]);
        }
        return result;
    }

    /**
     * Converts an object array into a float array, calling
     * {@link Util#asFloat(Object)} on each element.
     *
     * @param objArray the array of objects to convert
     * @return the converted float array
     * @throws NumberFormatException if an object's <code>toString()</code>
     * output does not contain a parsable float
     */
    public static float[] asFloatArray(final Object[] objArray)
    throws NumberFormatException {
        final float[] result = new float[objArray.length];
        for (int i = 0; i < objArray.length; i++) {
            result[i] = Util.asFloat(objArray[i]);
        }
        return result;
    }

    /**
     * Converts an object array into an integer array, calling
     * {@link Util#asInt(Object)} on each element.
     *
     * @param objArray the array of objects to convert
     * @return the converted int array
     * @throws NumberFormatException if an object's <code>toString()</code>
     * output does not contain a parsable int
     */
    public static int[] asIntArray(final Object[] objArray)
    throws NumberFormatException {
        final int[] result = new int[objArray.length];
        for (int i = 0; i < objArray.length; i++) {
            result[i] = Util.asInt(objArray[i]);
        }
        return result;
    }

    /**
     * Converts an object array into a long array, calling
     * {@link Util#asLong(Object)} on each element.
     *
     * @param objArray the array of objects to convert
     * @return the converted long array
     * @throws NumberFormatException if an object's <code>toString()</code>
     * output does not contain a parsable long
     */
    public static long[] asLongArray(final Object[] objArray)
    throws NumberFormatException {
        final long[] result = new long[objArray.length];
        for (int i = 0; i < objArray.length; i++) {
            result[i] = Util.asLong(objArray[i]);
        }
        return result;
    }

    /**
     * Converts an object array into a short array, calling
     * {@link Util#asShort(Object)} on each element.
     *
     * @param objArray the array of objects to convert
     * @return the converted short array
     * @throws NumberFormatException if an object's <code>toString()</code>
     * output does not contain a parsable short
     */
    public static short[] asShortArray(final Object[] objArray)
    throws NumberFormatException {
        final short[] result = new short[objArray.length];
        for (int i = 0; i < objArray.length; i++) {
            result[i] = Util.asShort(objArray[i]);
        }
        return result;
    }

    /**
     * Converts an object array into a String array, calling
     * {@link Util#asString(Object)} on each element.
     *
     * @param objArray the array of objects to convert
     * @return the converted String array
     */
    public static String[] asStringArray(final Object[] objArray) {
        final String[] result = new String[objArray.length];
        for (int i = 0; i < objArray.length; i++) {
            result[i] = Util.asString(objArray[i]);
        }
        return result;
    }

    /**
     * Combines two array into a target array, inserting all elements of the
     * first array and then all elements of the second array in the target
     * array.
     *
     * @param array1 the first array to copy
     * @param array2 the second array to copy
     * @param targetArray the array to copy the two other array into; the
     * type of this array must be suitable to accept elements from both array;
     * the length of this array must be equal or greater than
     * <code>array1.length + array2.length</code>
     */
    public static void combineArrays(final Object[] array1,
                                    final Object[] array2,
                                    final Object[] targetArray) {
        final int lengthOfFirst = array1.length;
        int i;
        // copy first array into target array
        for (i = 0; i < lengthOfFirst; i++) {
            targetArray[i] = array1[i];
        }
    
        // copy second array into target array
        for (i = 0; i < array2.length; i++) {
            targetArray[i + lengthOfFirst] = array2[i];
        }
    }

    /**
     * Flattens the elements of the provided array into a single string,
     * separating elements by a space character.
     *
     * @param array the array of values to join
     * @return the flattened string
     */
    public static String flatten(final boolean[] array) {
        final StringBuilder result = new StringBuilder();

        for (int i = 0; i < array.length; i++) {
            if (i > 0) {
                result.append(SEPARATOR);
            }
            result.append(array[i]);
        }

        return result.toString();
    }

    /**
     * Flattens the elements of the provided array into a single string,
     * separating elements by a space character.
     *
     * @param array the array of values to join
     * @return the flattened string
     */
    public static String flatten(final byte[] array) {
        final StringBuilder result = new StringBuilder();

        for (int i = 0; i < array.length; i++) {
            if (i > 0) {
                result.append(SEPARATOR);
            }
            result.append(array[i]);
        }

        return result.toString();
    }

    /**
     * Flattens the elements of the provided array into a single string,
     * separating elements by a space character.
     *
     * @param array the array of values to join
     * @return the flattened string
     */
    public static String flatten(final char[] array) {
        final StringBuilder result = new StringBuilder();

        for (int i = 0; i < array.length; i++) {
            if (i > 0) {
                result.append(SEPARATOR);
            }
            result.append(array[i]);
        }

        return result.toString();
    }

    /**
     * Flattens the elements of the provided array into a single string,
     * separating elements by a space character.
     *
     * @param array the array of values to join
     * @return the flattened string
     */
    public static String flatten(final double[] array) {
        final StringBuilder result = new StringBuilder();

        for (int i = 0; i < array.length; i++) {
            if (i > 0) {
                result.append(SEPARATOR);
            }
            result.append(array[i]);
        }

        return result.toString();
    }

    /**
     * Flattens the elements of the provided array into a single string,
     * separating elements by a space character.
     *
     * @param array the array of values to join
     * @return the flattened string
     */
    public static String flatten(final float[] array) {
        final StringBuilder result = new StringBuilder();

        for (int i = 0; i < array.length; i++) {
            if (i > 0) {
                result.append(SEPARATOR);
            }
            result.append(array[i]);
        }

        return result.toString();
    }

    /**
     * Flattens the elements of the provided array into a single string,
     * separating elements by a space character.
     *
     * @param array the array of values to join
     * @return the flattened string
     */
    public static String flatten(final int[] array) {
        final StringBuilder result = new StringBuilder();

        for (int i = 0; i < array.length; i++) {
            if (i > 0) {
                result.append(SEPARATOR);
            }
            result.append(array[i]);
        }

        return result.toString();
    }

    /**
     * Flattens the elements of the provided array into a single string,
     * separating elements by a space character.
     *
     * @param array the array of values to join
     * @return the flattened string
     */
    public static String flatten(final long[] array) {
        final StringBuilder result = new StringBuilder();

        for (int i = 0; i < array.length; i++) {
            if (i > 0) {
                result.append(SEPARATOR);
            }
            result.append(array[i]);
        }

        return result.toString();
    }

    /**
     * Flattens the elements of the provided array into a single string,
     * separating elements by a space character.
     *
     * @param array the array of values to join
     * @return the flattened string
     */
    public static String flatten(final Object[] array) {
        return flatten(array, SEPARATOR);
    }

    /**
     * Flattens the elements of the provided array into a single string,
     * separating elements by the provided separator.
     *
     * @param array the array of values to join
     * @param separator the separator string to use
     * @return the flattened string
     */
    public static String flatten(final Object[] array, final String separator) {
        final StringBuilder result = new StringBuilder();

        for (int i = 0; i < array.length; i++) {
            if (i > 0) {
                result.append(separator);
            }
            result.append(array[i]);
        }

        return result.toString();
    }

    /**
     * Flattens the elements of the provided array into a single string,
     * separating elements by a space character.
     *
     * @param array the array of values to join
     * @return the flattened string
     */
    public static String flatten(final short[] array) {
        final StringBuilder result = new StringBuilder();

        for (int i = 0; i < array.length; i++) {
            if (i > 0) {
                result.append(SEPARATOR);
            }
            result.append(array[i]);
        }

        return result.toString();
    }

    /**
     * Copied the last <em>n</em> elements from a list into a new list (or all
     * elements, if the size of the input list is smaller or equal to
     * <em>n</em>). Modifications of the returned list will not affect the
     * original list and vice versa.
     *
     * <p>If <code>number</code> is 0 or negative or if the original list is
     * <code>null</code> or empty, an empty list is returned.
     *
     * <p>Note that this is somewhat inefficient if the input list is a
     * {@link java.util.LinkedList} because repeated calls to
     * {@link List#get(int)} are necessary (unless the whole list is copied).
     *
     * @param list the input list
     * @param number the number of elements to copy
     * @return an ArrayList containing the last elements from the original list;
     * or an empty ArrayList iff the <code>list</code> is <code>null</code>
     */
    public static <T> ArrayList<T> lastN(final List<? extends T> list,
            final int number) {
        final ArrayList<T> result;
        if (list != null) {
            final int size = list.size();
            if (number <= size) {
                // copy the whole collection
                result = new ArrayList<T>(list);
            } else {
                // add last N elements
                result = new ArrayList<T>(number);
                for (int i = number; i > 0; i--) {
                    result.add(list.get(size - number));
                }
            }
        } else {
            // return empty list if input list is null
            result = new ArrayList<T>(0);
        }
        return result;
    }


    /**
     * Private constructor prevents creation of instances.
     */
    private CollectionUtils() {
        super();
    }

}
