/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.util;

import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.io.IOUtils;

/**
 * A simple wrapper for external scripts or programs. Calls an external program
 * with arguments and environment settings as specified and returns the output
 * of the program.
 *
 * <p>This class is thread-safe as each call to {@link #execute()} creates
 * another instances of the program. It uses the default
 * {@link de.fu_berlin.ties.util.TaskRunner}. <em>To allow efficient thread
 * re-use, it is highly recommended to initially
 * {@linkplain de.fu_berlin.ties.util.TaskRunner#registerInterest() register
 * your interest} in the default task runner and to finally
 * {@linkplain de.fu_berlin.ties.util.TaskRunner#deregisterInterest()
 * deregister} when you no longer need it.</em>
 *
 * @author Christian Siefkes
 * @version $Revision: 1.6 $, $Date: 2004/12/06 17:59:41 $, $Author: siefkes $
 */
public class ExternalCommand {

    /**
     * Array containing the command to call and its general arguments.
     */
    private final String[] commandArray;

    /**
     * An array of strings, each element of which has
     * environment variable settings in format <em>name=value</em>.
     */
    private final String[] envp;

    /**
     * The working directory of the process.
     */
    private final File workingDir;

    /**
     * Creates a new instance, without specifying environment parameters and
     * working directory.
     * Communication with the external process occurs in several parallel
     * threads using the default {@link TaskRunner} for asynchronous tasks.
     *
     * @param cmdArray array containing the command to call and its general
     * arguments, i.e. arguments that should be used used every time the
     * program is called
     * @throws IllegalArgumentException if the <code>cmdArray</code> is empty,
     * i.e. the command name is missing
     */
    public ExternalCommand(final String[] cmdArray)
                           throws IllegalArgumentException {
        this(cmdArray, null);
    }

    /**
     * Creates a new instance, without specifying environment parameters.
     * Communication with the external process occurs in several parallel
     * threads using the default {@link TaskRunner} for asynchronous tasks.
     *
     * @param cmdArray array containing the command to call and its general
     * arguments, i.e. arguments that should be used used every time the
     * program is called
     * @param workDir the working directory of the process; or <code>null</code>
     * if the command should inherit the working directory of the current
     * process
     * @throws IllegalArgumentException if the <code>cmdArray</code> is empty,
     * i.e. the command name is missing
     */
    public ExternalCommand(final String[] cmdArray, final File workDir)
                           throws IllegalArgumentException {
        this(cmdArray, null, workDir);
    }

    /**
     * Creates a new instance. Communication with the external process occurs
     * in several parallel threads using the default {@link TaskRunner} for
     * asynchronous tasks.
     *
     * @param cmdArray array containing the command to call and its general
     * arguments, i.e. arguments that should be used used every time the
     * program is called
     * @param envparams an array of strings, each element of which has
     * environment variable settings in format <em>name=value</em>; might be
     * <code>null</code> or empty
     * @param workDir the working directory of the process; or <code>null</code>
     * if the command should inherit the working directory of the current
     * process
     * @throws IllegalArgumentException if the <code>cmdArray</code> is empty,
     * i.e. the command name is missing
     */
    public ExternalCommand(final String[] cmdArray, final String[] envparams,
                           final File workDir) throws IllegalArgumentException {
        super();
        if (cmdArray.length == 0) {
            throw new IllegalArgumentException(
                "Command name missing (command array is empty)");
        }
        commandArray = cmdArray;
        envp = envparams;
        workingDir = workDir;
    }

    /**
     * Executed the external command any further arguments and without sending
     * input and returns its output (standard out).
     * Messages written to the standard error stream (stderr) are logged as
     * info statements. The system-specific default character set is used for
     * conversion between strings and streams.
     *
     * @return the output returned by the command's standard output stream.
     * @throws IOException if an I/O error occurs
     */
    public String execute() throws IOException {
        return execute(null, null);
    }

    /**
     * Executed the external command without any further arguments and returns
     * its output (standard out).
     * Messages written to the standard error stream (stderr) are logged as
     * info statements. The system-specific default character set is used for
     * conversion between strings and streams.
     *
     * @param input the input to send to the command's standard input stream;
     * or <code>null</code> if no input should be sent
     * @return the output returned by the command's standard output stream.
     * @throws IOException if an I/O error occurs
     */
    public String execute(final CharSequence input) throws IOException {
        return execute(null, input);
    }

    /**
     * Executed the external command without sending input and returns
     * its output (standard out).
     * Messages written to the standard error stream (stderr) are logged as
     * info statements. The system-specific default character set is used for
     * conversion between strings and streams.
     *
     * @param furtherArguments an array of further arguments to use for this
     * single call, might be <code>null</code>; appended <em>after</em> the
     * general arguments specified in the constructor
     * @return the output returned by the command's standard output stream.
     * @throws IOException if an I/O error occurs
     */
    public String execute(final String[] furtherArguments) throws IOException {
        return execute(furtherArguments, null);
    }

    /**
     * Executed the external command and returns its output (standard out).
     * Messages written to the standard error stream (stderr) are logged as
     * info statements. The system-specific default character set is used for
     * conversion between strings and streams.
     *
     * @param furtherArguments an array of further arguments to use for this
     * single call, might be <code>null</code>; appended <em>after</em> the
     * general arguments specified in the constructor
     * @param input the input to send to the command's standard input stream;
     * or <code>null</code> if no input should be sent
     * @return the output returned by the command's standard output stream.
     * @throws IOException if an I/O error occurs
     */
    public String execute(final String[] furtherArguments,
            final CharSequence input) throws IOException {
        final String[] fullCmdArray;
        if ((furtherArguments != null) && (furtherArguments.length > 0)) {
            // add further arguments after general arguments
            fullCmdArray =
                new String[commandArray.length + furtherArguments.length];
            CollectionUtils.combineArrays(commandArray, furtherArguments,
                    fullCmdArray);
        } else {
            // just use general arguments
            fullCmdArray = commandArray;
        }

        final String commandName = commandArray[0];
        Util.LOG.debug("Invoking " + commandName);
        final Process process =
            Runtime.getRuntime().exec(fullCmdArray, envp, workingDir);
        InputStreamReader readFromProcess = null;
        String result = null;

        TaskRunner.registerInterest();
        try {
            // asynchronously feed input to the process, if any
            if (input != null) {
                TaskRunner.invokeDefault(new Runnable() {
                        public final void run() {
                            OutputStreamWriter feedToProcess = null;
                            try {
                                feedToProcess = new OutputStreamWriter(
                                    process.getOutputStream());
                                IOUtils.writeToWriter(input, feedToProcess);
                            } catch (IOException ioe) {
                                Util.LOG.error("Error while feeding input to "
                                    + commandName + " process",
                                    ioe);
                            } finally {
                                IOUtils.tryToClose(feedToProcess);
                            }
                        }
                    },
                    commandName + " stdin feeder"
                );
            }

            // asynchronously read and log errors messages from stream (if any)
            TaskRunner.invokeDefault(new Runnable() {
                    public final void run() {
                        InputStreamReader errorsFromProcess = null;
                        try {
                            errorsFromProcess = new InputStreamReader(
                                process.getErrorStream());
                            // check the error/warn message of the process
                            final String errorsMessages =
                                IOUtils.readToString(errorsFromProcess);
                            // log as debug
                            if ((errorsMessages != null)
                                    && (!"".equals(errorsMessages))) {
                                Util.LOG.debug(commandName
                                    + " wrote to stderr: " + errorsMessages);
                            }
                        } catch (IOException ioe) {
                            Util.LOG.error("Error while feed input to "
                                + commandName + " process",
                                ioe);
                        } finally {
                            IOUtils.tryToClose(errorsFromProcess);
                        }
                    }
                },
                commandName + " stderr reader"
            );

            try {
                // read the data returned by the process
                readFromProcess =
                    new InputStreamReader(process.getInputStream());
                result = IOUtils.readToString(readFromProcess);

                // wait for process to finish and get return code
                try {
                    final int exitValue = process.waitFor();
                    if (exitValue != 0) {
                        Util.LOG.warn(commandName + " returned exit code "
                                + exitValue);
                    }
                } catch (InterruptedException ie) {
                    Util.LOG.warn("Interrupted while waiting for " + commandName
                        + " to finish");
                }

            } finally {
                IOUtils.tryToClose(readFromProcess);
            }
        } finally {
            TaskRunner.deregisterInterest();
        }
        return result;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this).
            append("command array", commandArray).
            append("environment", envp).
            append("working dir", workingDir).
            toString();
    }

}
