/*
 * Copyright (C) 2003-2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.util;

import java.util.Comparator;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * Compares objects, inverting the order of a wrapped comparator. This means
 * that numeric objects will be sorted in decreasing order etc.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2004/09/07 10:01:37 $, $Author: siefkes $
 */
public class InvertedComparator<T> implements Comparator<T> {

    /**
     * The comparator wrapped and inverted by this instance.
     */
    private final Comparator<T> wrappedComparator;

    /**
     * Creates a new instance, inverting the order returned by the provided
     * comparator.
     * @param wrappedComp the comparator to wrap and invert; must not be
     * <code>null</code>
     */
    public InvertedComparator(final Comparator<T> wrappedComp) {
        super();
        wrappedComparator = wrappedComp;
    }

    /**
     * Compares its two arguments for order. Inverts the natural ordering
     * of the two arguments resp. if order of the {@link #getWrappedComparator()
     * wrapped comparator}, if any. So this method returns a negative integer,
     * zero, or a positive integer as the first argument is greater than,
     * equal to, or less than the second.
     *
     * @param o1 the first object to be compared
     * @param o2 the second object to be compared
     * @return the inverted ordering of the arguments
     */
    public int compare(final T o1, final T o2) {
        return -wrappedComparator.compare(o1, o2);
    }

    /**
     * Returns the comparator wrapped and inverted by this instance.
     *
     * @return the value of the attribute
     */
    public Comparator getWrappedComparator() {
        return wrappedComparator;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
        .append("wrapped comparator", wrappedComparator)
        .toString();
    }

}
