/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.util;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * An entry in a doubly linked list. Useful when you cannot use the
 * {@link java.util.List} implementations, e.g. because you need direct
 * references to list entries. This class only stores a reference to the
 * previous and next entry for list management -- you must extend it so store
 * any useful data.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/08/24 18:00:38 $, $Author: siefkes $
 */
public class ListEntry {

    /**
     * Reference to the previous element in this list; <code>null</code> if this
     * is the first element of a list.
     */
    private ListEntry previous = null;

    /**
     * Reference to the next element in this list; <code>null</code> if this
     * is the final element of a list.
     */
    private ListEntry next = null;

    /**
     * Creates a new instance, without setting references to surrounding
     * entries.
     */
    public ListEntry() {
        this(null, null);
    }

    /**
     * Creates a new instance.
     *
     * @param prevEntry a reference to the previous entry, must not yet have
     * a next element; or <code>null</code> if this is the first entry in a list
     * @param nextEntry a reference to the next entry, must not yet have
     * a next element; or <code>null</code> if this is the final entry in a list
     * @throws IllegalArgumentException if <code>prevEntry</code> already has
     * a next element ({@link #hasNext()} == <code>true</code>) or if
     * <code>nextEntry</code> already has a previous element
     * ({@link #hasPrevious()} == <code>true</code>)
     */
    public ListEntry(final ListEntry prevEntry, final ListEntry nextEntry)
            throws IllegalArgumentException {
        super();

        if (prevEntry != null) {
            if (prevEntry.hasNext()) {
                throw new IllegalArgumentException(
                    "Previous entry already has a next element!");
            }
            previous = prevEntry;
            prevEntry.next = this;
        }

        if (nextEntry != null) {
            if (nextEntry.hasPrevious()) {
                throw new IllegalArgumentException(
                    "Next entry already has a previous element!");
            }
            next = nextEntry;
            nextEntry.previous = this;
        }
    }

    /**
     * Returns <code>true</code> if there is a next element. (In other words,
     * returns <code>true</code> if {@link #next()} would return an element
     * rather than <code>null</code>.)
     *
     * @return <code>true</code> iff the list has more elements
     */
    public boolean hasNext() {
        return (next != null);
    }

    /**
     * Returns <code>true</code> if there is a previous element. (In other
     * words, returns <code>true</code> if {@link #previous()} would return an
     * element rather than <code>null</code>.)
     *
     * @return <code>true</code> iff the list has more elements
     */
    public boolean hasPrevious() {
        return (previous != null);
    }

    /**
     * Inserts a new element after the current one, adjusting the list structure
     * accordingly. The new entry must not yet have a previous element (i.e.,
     * calling {@link #hasPrevious()} must return <code>false</code>).
     * The {@link #next()} element of either this entry or the new
     * entry must be <code>null</code>.
     * Inserting <code>null</code> elements is not allowed.
     *
     * @param newEntry the new list entry to insert
     * @throws IllegalArgumentException if <code>newEntry</code> already has a
     * previous element ({@link #hasPrevious()} == <code>true</code>) or
     * both this entry and <code>newEntry</code> have a {@link #next()}
     * element
     * @throws NullPointerException if <code>newEntry</code> is
     * <code>null</code>
     */
    public void insertAfter(final ListEntry newEntry)
            throws IllegalArgumentException, NullPointerException {
        if (newEntry.hasPrevious()) {
            throw new IllegalArgumentException(
                "Second entry already has a previous element!");
        }
        if (hasNext() && newEntry.hasNext()) {
            throw new IllegalArgumentException(
                "Both entries have a next element!");
        }

        // adjust list structure: insert between this element and next one
        if (next != null) {
            next.previous = newEntry;
            newEntry.next = next;
        }
        newEntry.previous = this;
        next = newEntry;
    }

    /**
     * Inserts a new element before the current one, adjusting the list
     * structure accordingly. If this entry has a {@link #previous()} element,
     * the {@link #next()} element of either the new entry or the
     * {@link #previous()} element must be <code>null</code>; otherwise the
     * {@link #next()} element of either this entry or the new entry must be
     * <code>null</code>.
     * Inserting <code>null</code> elements is not allowed.
     *
     * @param newEntry the new list entry to insert
     * @throws IllegalArgumentException if the conditions stated above are
     * violated
     * @throws NullPointerException if <code>newEntry</code> is
     * <code>null</code>
     */
    public void insertBefore(final ListEntry newEntry)
            throws IllegalArgumentException, NullPointerException {
        if (previous != null) {
            previous.insertAfter(newEntry);
        } else {
            newEntry.insertAfter(this);
        }
    }

    /**
     * Returns the next element in the list, if any.
     *
     * @return the next element in the list, or <code>null</code> if there is
     * no next element
     */
    public ListEntry next() {
        return next;
    }

    /**
     * Returns the previous element in the list, if any.
     *
     * @return the previous element in the list, or <code>null</code> if there
     * is no previous element
     */
    public ListEntry previous() {
        return previous;
    }

    /**
     * Removes this entry from this list. The list structure
     * is adjusted accordingly. Afterwards this entry won't be part of a
     * list, i.e. both {@link #hasPrevious()} and {@link #hasNext()} will
     * return <code>false</code>.
     */
    public void remove() {
        // adjusting list structure
        if (next != null) {
            next.previous = previous;
        }
        if (previous != null) {
            previous.next = next;
        }
        next = null;
        previous = null;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this).
            append("has previous", hasPrevious()).
            append("has next", hasNext()).
            toString();
    }

}
