/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.util;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * A {@link de.fu_berlin.ties.util.MultiValueMap} that sorts the values
 * stored for each key, discarding duplicates. For this purpose,
 * {@link java.util.TreeSet}s are used for inner collections.
 *
 * By default, the keys are sorted as well, by wrapping a
 * {@link java.util.TreeMap}. This can be changed by specifying a different kind
 * of map to wrap in the constructor.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/11/05 12:16:16 $, $Author: siefkes $
 */
public class SortedMultiValueMap<K, V> extends MultiValueMap<K, V> {

    /**
     * Creates a new instance, wrapping a {@link TreeMap}.
     */
    public SortedMultiValueMap() {
        this(new TreeMap<K, Collection<V>>());
    }

    /**
     * Creates a new instance.
     *
     * @param wrappedMap wrapped map used as storage, e.g. a {@link HashMap} or
     * a {@link java.util.TreeMap}
     */
    public SortedMultiValueMap(Map<K, Collection<V>> wrappedMap) {
        super(wrappedMap);
    }

    /**
     * {@inheritDoc}
     * This implementation returns a {@link TreeSet}.
     */
    protected Collection<V> createCollection(
            final Collection<? extends V> coll) {
        if (coll == null) {
            return new TreeSet<V>();
        } else {
            return new TreeSet<V>(coll);
        }
    }

}
