/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package de.fu_berlin.ties.util;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.pool.BasePoolableObjectFactory;
import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.impl.StackObjectPool;

import de.fu_berlin.ties.Closeable;


/**
 * Asynchronously executes any number of {@link java.lang.Runnable} tasks.
 * Internally manages a pool of worker threads that are invoked to
 * asychronously execute tasks as requested. The number of worker threads is
 * updated on demand so every task is executed without delay.
 *
 * <p>Usually it should be sufficient to invoke tasks through the static
 * {@link #invokeDefault(Runnable, String) invokeDefault} methods; in most
 * cases there should be no reason to create instances of this class.
 * <em>To avoid the creation of unnecessary instances, it is highly recommended
 * to register your interest ({@link #registerInterest()}) in the default task
 * runner prior to using it and to deregister
 * ({@link #deregisterInterest()}) when you no longer need it.
 * A good idea is to do this at the begin and end of your main method.
 * You should register in a <code>finally</code> block and you <strong>must not
 * forget</strong> to deregister, otherwise your program might run forever
 * (because the worker threads continue waiting for tasks even after all other
 * threads have terminated).</em>
 *
 * <p>When creating your own runner, you <strong>must</strong> finally call
 * {@link #close(int)} to release each task runner you have created, or your
 * program might run forever.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.11 $, $Date: 2004/10/11 17:08:24 $, $Author: siefkes $
 */
public class TaskRunner implements Closeable {

    /**
     * A helper threads that executes any tasks as requested and then sleeps
     * until a new task is required or it is told to terminate.
     */
    private final class WorkerThread extends Thread {

        /**
         * Indicates whether this thread should terminate (after executing any
         * specified task).
         */
        private boolean shouldTerminate = false;

        /**
         * The task to be execute by this thread.
         */
        private Runnable task = null;

        /**
         * A name or short description of the current task -- used for debugging
         * purposes.
         */
        private String taskName = null;

        /**
         * Used to guard synchrization of this thread.
         */
        private final Object threadGuard = new Object();

        /**
         * Creates a new instance.
         *
         * @param name the name of the new thread
         */
        private WorkerThread(final String name) {
            super(name);
        }

        /**
         * The run method of this thread: executing the task provided by
         * calling {@link #setTask(Runnable, String)} and then puts itself back
         * into the pool and waits until a new task arrives or it
         * {@link #shouldTerminate()}. Any throwables thrown by the executed
         * {@link java.lang.Runnable} are catched and logged so this thread
         * cannot be crashed by careless Runnables.
         */
        public void run() {
            // local copies that not require synchronization
            boolean doTerminate = false;
            Runnable currentTask;
            String currentTaskName;

            while (!doTerminate) {
                // get task, if any
                synchronized (threadGuard) {
                    if ((task == null) && (!shouldTerminate) && (!closed)) {
                        // nothing to do--wait until notified of status change
                        try {
                            threadGuard.wait();
                        } catch (InterruptedException ie) {
                            Util.LOG.warn(
                                "Interrupted while waiting for status change"
                                + "  -- this was not supposed to happen");
                        }
                    }

                    // Terminate when requested or when the outer class has
                    // been closed.
                    doTerminate = shouldTerminate || closed;
                    currentTask = task;
                    currentTaskName = taskName;

                    // set member fields to null so a new task can arrive
                    task = null;
                    taskName = null;
                }

                // execute task, if any
                if (currentTask != null) {
                    try {
                        currentTask.run();
                    } catch (Throwable t) {
                        Util.LOG.error("Error while executing "
                            + currentTaskName,
                            t);
                    }

                    // now put myself back into the thread pool (if still open)
                    if (!closed) {
                        try {
                            threadPool.returnObject(this);
                        } catch (Exception e) {
                            doTerminate = true;
                            Util.LOG.error("Could not return myself into the "
                                + "pool -- will now terminate myself",
                                e);
                        }                        
                    }
                }
            }
        }

        /**
         * Sets the task to be executed by this thread and notifies the
         * thread that new work has arrived.
         *
         * @param newTask the task to be executed by this thread
         * @param newTaskName A name or short description of the new task --
         * used for debugging purposes
         * @throws IllegalStateException if there already is another task set
         * (a thread can only execute a single given task at any time)
         */
        private void setTask(final Runnable newTask,
            final String newTaskName) throws IllegalStateException {
            synchronized (threadGuard) {
                if (task != null) {
                    throw new IllegalStateException("Task of " + this.toString()
                        + "already set to " + taskName
                        + " -- must be null when setting a new task");
                }
                task = newTask;
                taskName = newTaskName;

                // wake the thread if it's waiting
                threadGuard.notify();
            }
        }

        /**
         * Tells this thread to terminate and notifies the thread of the status
         * change. The thread will only terminate after executing the given
         * task, if any is left.
         */
        private void shouldTerminate() {
            synchronized (threadGuard) {
                shouldTerminate = true;

                // wake the thread if it's waiting
                threadGuard.notify();
            }
        }

        /**
         * Returns a string representation of this thread.
         *
         * @return a textual representation
         */
        public String toString() {
            return new ToStringBuilder(this).
                appendSuper(super.toString()).
                append("task name", taskName).
                append("should terminate", shouldTerminate).
                toString();
        }

    }


    /**
     * The base name of worker threads used by the default instance.
     */
    public static final String DEFAULT_NAME = "worker";

    /**
     * The default instance used by the static methods. This instance should be
     * sufficient for most purposes, so usually there should be no reason to
     * create any other instances.
     */
    private static TaskRunner defaultRunner;

    /**
     * The default executor used for tasks that can be queued (executed later).
     */

    /**
     * The number of callers that have registered their interest in the
     * {@link #defaultRunner}. The default runner is alive while this number
     * is greater than 0.
     */
    private static int interestCount = 0;

    /**
     * Deregisters interest to use the default runner. You <strong>must</strong>
     * have called {@link #registerInterest()} prior to calling this method.
     * You should calls this method in a <code>finally</code> block to ensure
     * it is always executed.
     */
    public static synchronized void deregisterInterest() {
        if (interestCount > 0) {
            interestCount--;

            if (interestCount == 0) {
                if (defaultRunner != null) {
                    // no more interest: destroy the runner
                    defaultRunner.close(0);
                    defaultRunner = null;
                    Util.LOG.debug("Terminated default task runner");
                }
            }
        } else {
            Util.LOG.error("TaskRunner.deregisterInterest called but interest "
                + "count is already " + interestCount);
        }
    }

    /**
     * Invokes a task to be executed asynchronously, using the default task
     * runner. This method returns immediately; the given task is executed in
     * another thread. The task is executed with the default priority
     * ({@link Thread#NORM_PRIORITY}).
     *
     * @param task the Runnable to be executed asynchronously
     * @param taskName a name or short description of the new task --
     * used for debugging purposes
     */
    public static void invokeDefault(final Runnable task,
            final String taskName) {
        invokeDefault(task, taskName, Thread.NORM_PRIORITY);
    }

    /**
     * Invokes a task to be executed asynchronously using the default task
     * runner. This method returns immediately; the given task is executed in
     * another thread.
     *
     * @param task the Runnable to be executed asynchronously
     * @param taskName a name or short description of the new task --
     * used for debugging purposes
     * @param priority the priority to use for the thread executing the task;
     * should be in the range of {@link java.lang.Thread#MIN_PRIORITY} to
     * {@link java.lang.Thread#MAX_PRIORITY}
     * @throws IllegalArgumentException if the priority is not in the range
     * {@link java.lang.Thread#MIN_PRIORITY} to
     * {@link java.lang.Thread#MAX_PRIORITY}
     */
    public static synchronized void invokeDefault(final Runnable task,
            final String taskName, final int priority)
            throws IllegalArgumentException {
        registerInterest();

        // initialize runner if not yet done
        if (defaultRunner == null) {
            defaultRunner = new TaskRunner(DEFAULT_NAME);
            Util.LOG.debug("Initialized default task runner");
        }

        try {
            defaultRunner.invoke(task, taskName, priority);
        } finally {
            deregisterInterest();
        }
    }

    /**
     * Registers interest to use the default runner. You <strong>must</strong>
     * {@link #deregisterInterest()} some time after calling this method,
     * otherwise your program will not terminate regularly.
     */
    public static synchronized void registerInterest() {
        interestCount++;
    }


    /**
     * The base name of worker threads. The actual names of worker threads
     * are formed by appending {@link #nextFreeNumber} to this name.
     */
    private final String baseName;

    /**
     * Whether this task runner has been closed.
     */
    private boolean closed = false;

    /**
     * The default priority to use for threads.
     */
    private final int defaultPriority;

    /**
     * The next available number to use when naming worker threads. Should
     * be synchronized on {@link #threadPool}.
     */
    private int nextFreeNumber = 1;

    /**
     * The pool of worker thread used to execute the requested tasks.
     */
    private final ObjectPool threadPool;


    /**
     * Creates a new instance, using the default priority
     * ({@link java.lang.Thread#NORM_PRIORITY}) for threads.
     *
     * @param baseThreadName the base name of worker threads -- the actual
     * names are formed by appending the next available number
     */
    public TaskRunner(final String baseThreadName) {
        this(baseThreadName, Thread.NORM_PRIORITY);
    }

    /**
     * Creates a new instance.
     *
     * @param baseThreadName the base name of worker threads -- the actual
     * names are formed by appending the next available number
     * @param defaultPrio the priority to use for threads;
     * should be in the range of {@link java.lang.Thread#MIN_PRIORITY} to
     * {@link java.lang.Thread#MAX_PRIORITY}
     * @throws IllegalArgumentException if the priority is not in the range
     * {@link java.lang.Thread#MIN_PRIORITY} to
     * {@link java.lang.Thread#MAX_PRIORITY}
     */
    public TaskRunner(final String baseThreadName, final int defaultPrio)
            throws IllegalArgumentException {
        super();
        checkPriority(defaultPrio);
        baseName = baseThreadName;
        defaultPriority = defaultPrio;

        // define a factory that returns WorkerThreads and tells them to stop
        // when they are dropped from the pool
        threadPool = new StackObjectPool(new BasePoolableObjectFactory() {
            public final void destroyObject(final Object object) {
                //Log.TIES.debug("Got request to destroy worker thread object");
                // must be a worker thread -- tell it to stop
                ((WorkerThread) object).shouldTerminate();
                //Log.TIES.debug("Ordered worker thread " + object.toString()
                //    + " to terminate");
            }

            public final Object makeObject() {
                final String threadName;
                synchronized (threadPool) {
                    // determine thread name and increase then next free number
                    threadName = baseName + nextFreeNumber++;
                }
                //Log.TIES.debug("Creating worker thread " + threadName);
                final WorkerThread worker = new WorkerThread(threadName);
                if (worker.getPriority() != defaultPriority) {
                    worker.setPriority(defaultPriority);
                }
                worker.start();
                //Log.TIES.debug("Started worker thread " + threadName);
                return worker;
            }
        });
    }

    /**
     * Helper methods that checks whether a priority is in the valid range
     * and throws an exception otherwise.
     *
     * @param priority the priority to check
     * @throws IllegalArgumentException if the priority is outside the valid
     * range
     */
    private void checkPriority(final int priority)
            throws IllegalArgumentException {
        if ((priority < Thread.MIN_PRIORITY)
                || (priority > Thread.MAX_PRIORITY)) {
            throw new IllegalArgumentException("Priority " + priority
                + "is outside the valid range " + Thread.MIN_PRIORITY
                + " to " + Thread.MAX_PRIORITY);
        }
    }

    /**
     * Closes this task runner. All thread will be terminated after executing
     * their current tasks, so no tasks will be lost. You must call this
     * method to release each task runner you created, or your program might
     * run forever (because the worker threads continue waiting for tasks
     * even after all other threads have terminated). You must not call
     * {@link #invoke(Runnable, String)} after calling this method.
     *
     * @param errorCount the number of errors (exceptions) that occurred during
     * calls to this instance (0 if none); ignored by this method
     */
    public final void close(final int errorCount) {
        if (!closed) { // otherwise there is nothing to do
            closed = true;
            try {
                // delegate to thread pool
                threadPool.clear();
                threadPool.close();
            } catch (RuntimeException re) {
                // rethrow "as is"
                throw re;
            } catch (Exception e) {
                // normally everything should work fine so we just wrap the
                // exception in a RuntimeException
                throw new RuntimeException(
                    "Error while closing task runner", e);
            }
        }
    }

    /**
     * Called by the garbage collector on an object when garbage collection
     * determines that there are no more references to the object. Delegates
     * to {@link #close(int)} to terminate all worker threads if not yet done.
     * <b>You should not rely on this method but always call {@link #close(int)}
     * yourself.</b>
     */
    protected final void finalize() {
        close(0);
    }

    /**
     * Returns the base name of worker threads. The actual names of worker
     * threads are formed by appending the next available number to this name
     * (baseName + 1, baseName + 2 etc.)
     *
     * @return the base name
     */
    public final String getBaseName() {
        return baseName;
    }

    /**
     * Returns the default priority to use for threads.
     *
     * @return the priority
     */
    public final int getDefaultPriority() {
        return defaultPriority;
    }

    /**
     * Invokes a task to be executed asynchronously. This method returns
     * immediately; the given task is executed in another thread. The task
     * is executed with the default priority of this task runner
     * ({@link #getDefaultPriority()}).
     *
     * @param task the Runnable to be executed asynchronously
     * @param taskName a name or short description of the new task --
     * used for debugging purposes
     */
    public final void invoke(final Runnable task, final String taskName) {
        invoke(task, taskName, getDefaultPriority());
    }

    /**
     * Invokes a task to be executed asynchronously. This method returns
     * immediately; the given task is executed in another thread.
     *
     * @param task the Runnable to be executed asynchronously
     * @param taskName a name or short description of the new task --
     * used for debugging purposes
     * @param priority the priority to use for the thread executing the task;
     * should be in the range of {@link java.lang.Thread#MIN_PRIORITY} to
     * {@link java.lang.Thread#MAX_PRIORITY}
     * @throws IllegalArgumentException if the priority is not in the range
     * {@link java.lang.Thread#MIN_PRIORITY} to
     * {@link java.lang.Thread#MAX_PRIORITY}
     */
    public final void invoke(final Runnable task, final String taskName,
            final int priority) throws IllegalArgumentException {
        checkPriority(priority);
        final WorkerThread worker;
        try {
            worker = (WorkerThread) threadPool.borrowObject();
        } catch (RuntimeException re) {
            // rethrow "as is"
            throw re;
        } catch (Exception e) {
            // normally everything should work fine so we just wrap the
            // exception in a RuntimeException
            throw new RuntimeException(
                "Could not borrow worker thread from pool", e);
        }

        // adjust priority
        if (worker.getPriority() != priority) {
            worker.setPriority(priority);
        }

        // call worker thread to execute the task
        worker.setTask(task, taskName);
    }

    /**
     * Returns whether this task runner has been closed.
     *
     * @return <code>true</code> iff {@link #close(int)} has been called on this
     * object
     */
    public final boolean isClosed() {
        return closed;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this).
            append("default priority", defaultPriority).
            append("is closed", closed).
            append("base name of worker thread", baseName).
            append("next free thread number", nextFreeNumber).
            append("thread pool", threadPool).
            toString();
    }

}
