/*
 * Copyright (C) 2003-2004 Christian Siefkes &lt;christian@siefkes.net&gt;.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.util;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.text.NumberFormat;
import java.util.Arrays;
import java.util.Locale;
import java.util.Random;

import org.apache.commons.lang.BooleanUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.StandardToStringStyle;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.log.Hierarchy;
import org.apache.log.LogTarget;
import org.apache.log.Logger;
import org.apache.log.Priority;
import org.apache.log.filter.PriorityFilter;
import org.apache.log.format.ExtendedPatternFormatter;
import org.apache.log.output.io.FileTarget;
import org.apache.log.output.io.StreamTarget;

import de.fu_berlin.ties.TiesConfiguration;

/**
 * A static class that provides general utility methods.
 * No instances of this class can be created, only the static members
 * should be used.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.27 $, $Date: 2004/12/06 17:59:41 $, $Author: siefkes $
 */
public final class Util {

    /**
     * Filter used by the logger to write messages to standard output.
     */
    private static final PriorityFilter STD_FILTERED;

    /**
     * The logger used in the TIE system. The
     * <a href="http://avalon.apache.org/logkit/">Avalon LogKit</a> (bundled
     * with Velocity) is used for logging. All log messages are written to a log
     * file; message with priority <code>INFO</code> or higher are also written
     * to standard out.
     */
    public static final Logger LOG =
        Hierarchy.getDefaultHierarchy().getLoggerFor("ties");

    /**
     * Configuration key: Only messages with this priority or higher are logged.
     */
    public static final String CONFIG_LOGGER_LOG = "log.log";

    /**
     * Configuration key: Only messages  with this priority or higher are
     * written to standard output (but only if covered by
     * {@link #CONFIG_LOGGER_LOG}).
     */
    public static final String CONFIG_LOGGER_SHOW = "log.show";

    /**
     * String recognized as Not-a-Number when parsing floating-point numbers:
     * {@value} (ignoring case).
     */
    public static final String NAN = "NaN";

    /**
     * Number format used by {@link #showDuration(long)} to format seconds.
     * This is meant for log methods so we always use the English locale.
     * Should be synchronized on itself.
     */
    private static final NumberFormat SEC_FORMAT =
        NumberFormat.getNumberInstance(Locale.ENGLISH);

    /**
     * Number format used by the {@link #format(double) format} methods.
     * Uses the English locale. Should be synchronized on itself.
     */
    private static final NumberFormat NUM_FORMAT =
        NumberFormat.getNumberInstance(Locale.ENGLISH);

    /**
     * Static initializer: modify style used by {@link ToStringBuilder} and
     * configure number format + logger.
     */
    static {
        // modify ToStringStyle: short class names, no hash code, separator: ";"
        final StandardToStringStyle tiesStyle = new StandardToStringStyle();
        tiesStyle.setUseShortClassName(true);
        tiesStyle.setUseIdentityHashCode(false);
        tiesStyle.setFieldSeparator(";");
        ToStringBuilder.setDefaultStyle(tiesStyle);

        // configure number format: round to milliseconds
        SEC_FORMAT.setMaximumFractionDigits(3);

        // configure logger:
        // currently no category is printed as "ties" is used globally
        final String basePattern =
            "%{priority}: %{message}\\n%{throwable}";
        // full pattern also showns time + thread name
        final String fullPattern =
            "%{time:yyyy-MM-dd HH:mm:ss.SSS} [%{thread}] " + basePattern;

        // log everything to "ties.log" in full format
        final String logFileName = "ties.log";
        LogTarget fileLog;
        IOException caughtException = null;
        Util.LOG.setPriority(Priority.DEBUG);

        // INFO and higher goes to standard out in compact format
        final LogTarget stdRaw = new StreamTarget(System.out,
            new ExtendedPatternFormatter(basePattern));
        STD_FILTERED = new PriorityFilter(Priority.INFO);
        STD_FILTERED.addTarget(stdRaw);

        try {
            fileLog = new FileTarget(new File(logFileName), false,
                new ExtendedPatternFormatter(fullPattern));
        } catch (IOException ioe) {
            // delay logging until format is ok
            caughtException = ioe;
            fileLog = null;
        }

        final LogTarget[] targets;
        if (fileLog != null) {
            targets = new LogTarget[] {fileLog, STD_FILTERED};
        } else {
            // file cannot be written -- log everything to stdout instead
            targets = new LogTarget[] {stdRaw};
        }

        Util.LOG.setLogTargets(targets);

        if (caughtException != null) {
            Util.LOG.error("Log initialization: could not open " + logFileName
                + " -- will log to standard out only: " + caughtException);
        }
    }

    /**
     * Appends a non-negative number to a string buffer.
     *
     * @param appender the string buffer to append to
     * @param number the number to append; must not be negative
     * @param minDigits the minimum number of digits to print
     * @param prefix a prefix to print in front of the number, if not
     * <code>null</code>
     * @param appendAlways if <code>true</code>, the number is always printed;
     * otherwise only positive numbers are printed -- in this case the appender
     * is not modified if the number is 0
     * @throws IllegalArgumentException if <code>number &lt; 0</code>
     */
    public static void appendPositiveNumber(final StringBuffer appender,
            final long number, final int minDigits, final String prefix,
            final boolean appendAlways)
    throws IllegalArgumentException {
        // no support for negative numbers
        ensureNonNegative(number, "Number");

        if (appendAlways || (number != 0)) {
            // append prefix if given
            if (StringUtils.isNotEmpty(prefix)) {
                appender.append(prefix);
            }

            // convert number to a string
            final String numberString = Long.toString(number);
            final int lengthDiff = minDigits - numberString.length();

            // left-pad with zeros, if necessary
            for (int i = 0; i < lengthDiff; i++) {
                appender.append('0');
            }

            // append actual number
            appender.append(numberString);
        }
    }

    /**
     * Converts an object into a boolean primitive. For a {@link Boolean}
     * object, the contained boolean primitive is returned.
     *
     * <p>Otherwise the object's <code>toString()</code> output is parsed
     * calling {@link BooleanUtils#toBooleanObject(java.lang.String)}, i.e.
     * The strings "true", "on", and "yes" are accepted for a <code>true</code>
     * value; "false", "off", and "no" are accepted for <code>false</code>;
     * case and surrounding whitespace are ignored. An exception is thrown for
     * all other strings.
     *
     * <p>If the specified object is <code>null</code>, <code>false</code>
     * is returned.
     *
     * @param obj the object to check
     * @return the converted boolean
     * @throws IllegalArgumentException if the object's <code>toString()</code>
     * output cannot be parsed as a boolean
     */
    public static boolean asBoolean(final Object obj)
            throws IllegalArgumentException {
        final boolean result;
        if (obj instanceof Boolean) {
            // use contained value
            result = ((Boolean) obj).booleanValue();
        } else if (obj != null) {
            // parse string value
            final String raw = obj.toString().trim();
            final Boolean parsed = BooleanUtils.toBooleanObject(raw);

            if (parsed == null) {
                // invalid string value
                throw new IllegalArgumentException("String '" + raw
                        + "' does not specify a boolean value");
            } else {
                result = parsed.booleanValue();
            }
        } else {
            // null object: return false
            result = false;
        }
        return result;
    }

    /**
     * Converts an object into a byte primitive. For a {@link Number}
     * object, the contained number is converted to a byte -- this may involve
     * rounding or truncation.
     * <p>Otherwise the object's <code>toString()</code> output is parsed
     * using the {@link Byte#decode(java.lang.String)} method.
     * <p>If the specified object is <code>null</code>, -1 is returned.
     *
     * @param obj the object to check
     * @return the converted byte
     * @throws NumberFormatException if the object's <code>toString()</code>
     * output does not contain a parsable byte
     */
    public static byte asByte(final Object obj) throws NumberFormatException {
        final byte result;
        if (obj instanceof Number) {
            // use contained number
            result = ((Number) obj).byteValue();
        } else if (obj != null) {
            // parse string value
            final String raw = obj.toString().trim();
            result = Byte.decode(raw).byteValue();
        } else {
            // null object
            result = -1;
        }
        return result;
    }

    /**
     * Converts an object into a char primitive. For a {@link Character}
     * object, the contained char primitive is returned.
     * <p>Otherwise the first non-whitespace character of the
     * object's <code>toString()</code> output is returned. An exception is
     * thrown if the string is empty or contains only whitespace.
     * <p>If the specified object is <code>null</code>,
     * {@link Character#MIN_VALUE} (the smallest character value) is returned.
     *
     * @param obj the object to check
     * @return the converted char
     * @throws IndexOutOfBoundsException if the object's <code>toString()</code>
     * output is the empty string (after trimming outer whitespace)
     */
    public static char asChar(final Object obj)
            throws IndexOutOfBoundsException {
        final char result;
        if (obj instanceof Character) {
            // use contained character
            result = ((Character) obj).charValue();
        } else if (obj != null) {
            // return first character of trimmed string value
            final String raw = obj.toString().trim();
            result = raw.charAt(0);
        } else {
            // null object: return smallest character value
            result = Character.MIN_VALUE;
        }
        return result;
    }

    /**
     * Converts an object into a double primitive. For a {@link Number}
     * object, the contained number is converted to a double -- this may involve
     * rounding.
     *
     * <p>Otherwise the object's <code>toString()</code> output is parsed
     * using the {@link Double#parseDouble(java.lang.String)} method.
     * If the string value to equal to {@link #NAN} (ignoring case),
     * {@link Double#NaN} is returned.
     *
     * <p>If the specified object is <code>null</code>, -1.0 is returned.
     *
     * @param obj the object to check
     * @return the converted double
     * @throws NumberFormatException if the object's <code>toString()</code>
     * output does not contain a parsable double
     */
    public static double asDouble(final Object obj)
            throws NumberFormatException {
        final double result;
        if (obj instanceof Number) {
            // use contained number
            result = ((Number) obj).doubleValue();
        } else if (obj != null) {
            final String raw = obj.toString().trim();

            if (NAN.equalsIgnoreCase(raw)) {
                // Not-a-Number
                result = Double.NaN;
            } else {
                // parse string value
                result = Double.parseDouble(raw);
            }
        } else {
            // null object
            result = -1.0;
        }
        return result;
    }

    /**
     * Converts an object into a float primitive. For a {@link Number}
     * object, the contained number is converted to a float -- this may involve
     * rounding.
     *
     * <p>Otherwise the object's <code>toString()</code> output is parsed
     * using the {@link Float#parseFloat(java.lang.String)} method.
     * If the string value to equal to {@link #NAN} (ignoring case),
     * {@link Float#NaN} is returned.
     *
     * <p>If the specified object is <code>null</code>, -1 is returned.
     *
     * @param obj the object to check
     * @return the converted float
     * @throws NumberFormatException if the object's <code>toString()</code>
     * output does not contain a parsable float
     */
    public static float asFloat(final Object obj)
            throws NumberFormatException {
        final float result;
        if (obj instanceof Number) {
            // use contained number
            result = ((Number) obj).floatValue();
        } else if (obj != null) {
            final String raw = obj.toString().trim();

            if (NAN.equalsIgnoreCase(raw)) {
                // Not-a-Number
                result = Float.NaN;
            } else {
                // parse string value
                result = Float.parseFloat(raw);
            }
        } else {
            // null object
            result = -1.0f;
        }
        return result;
    }

    /**
     * Converts an object into an integer primitive. For a {@link Number}
     * object, the contained number is converted to a int -- this may involve
     * rounding or truncation.
     * <p>Otherwise the object's <code>toString()</code> output is parsed
     * using the {@link Integer#decode(java.lang.String)} method.
     * <p>If the specified object is <code>null</code>, -1 is returned.
     *
     * @param obj the object to check
     * @return the converted int
     * @throws NumberFormatException if the object's <code>toString()</code>
     * output does not contain a parsable int
     */
    public static int asInt(final Object obj) throws NumberFormatException {
        final int result;
        if (obj instanceof Number) {
            // use contained number
            result = ((Number) obj).intValue();
        } else if (obj != null) {
            // parse string value
            final String raw = obj.toString().trim();
            result = Integer.decode(raw).intValue();
        } else {
            // null object
            result = -1;
        }
        return result;
    }

    /**
     * Converts an object into a long primitive. For a {@link Number}
     * object, the contained number is converted to a long -- this may involve
     * rounding or truncation.
     * <p>Otherwise the object's <code>toString()</code> output is parsed
     * using the {@link Long#decode(java.lang.String)} method.
     * <p>If the specified object is <code>null</code>, -1 is returned.
     *
     * @param obj the object to check
     * @return the converted long
     * @throws NumberFormatException if the object's <code>toString()</code>
     * output does not contain a parsable long
     */
    public static long asLong(final Object obj) throws NumberFormatException {
        final long result;
        if (obj instanceof Number) {
            // use contained number
            result = ((Number) obj).longValue();
        } else if (obj != null) {
            // parse string value
            final String raw = obj.toString().trim();
            result = Long.decode(raw).longValue();
        } else {
            // null object
            result = -1;
        }
        return result;
    }

    /**
     * Converts an object into a short primitive. For a {@link Number}
     * object, the contained number is converted to a short -- this may involve
     * rounding or truncation.
     * <p>Otherwise the object's <code>toString()</code> output is parsed
     * using the {@link Short#decode(java.lang.String)} method.
     * <p>If the specified object is <code>null</code>, -1 is returned.
     *
     * @param obj the object to check
     * @return the converted short
     * @throws NumberFormatException if the object's <code>toString()</code>
     * output does not contain a parsable short
     */
    public static short asShort(final Object obj) throws NumberFormatException {
        final short result;
        if (obj instanceof Number) {
            // use contained number
            result = ((Number) obj).shortValue();
        } else if (obj != null) {
            // parse string value
            final String raw = obj.toString().trim();
            result = Short.decode(raw).shortValue();
        } else {
            // null object
            result = -1;
        }
        return result;
    }

    /**
     * Converts an object into a String. For non-null objects, the output
     * of the {@link Object#toString()} is returned; otherwise,
     * <code>null</code> is returned.
     *
     * @param obj the object to check
     * @return the output of the object's {@link Object#toString()}; or
     * <code>null</code> if <code>obj</code> is <code>null</code>
     */
    public static String asString(final Object obj) {
        if (obj != null) {
            return obj.toString();
        } else {
            return null;
        }
    }

    /**
     * Reconfigures the {@link #LOG logger} from the {@link #CONFIG_LOGGER_LOG}
     * and {@link #CONFIG_LOGGER_SHOW} values in the provided configuration.
     * This method should be used with care since it affects the global logging.
     * It is called by {@link de.fu_berlin.ties.Main#main(String[])} after
     * configurations and command-line propertis have been read.
     *
     * @param config the configuration to use
     */
    public static void configureLog(final TiesConfiguration config) {
        // ignoring case of configured values by converting them to upper case
        LOG.setPriority(Priority.getPriorityForName(config.getString(
            CONFIG_LOGGER_LOG).toUpperCase()));
        STD_FILTERED.setPriority(Priority.getPriorityForName(config.getString(
        CONFIG_LOGGER_SHOW).toUpperCase()));
    }

    /**
     * Delegates to {@link Util#createObject(Class, String[])}, reading the
     * class name from the first element in the array.
     *
     * @param classDef defines the class to create: the first element must be
     * a string giving the fully qualified class name; any further elements
     * are passed as constructor parameters
     * @return the created object; the returned object will be an instance of
     * the specified type
     * @throws IllegalArgumentException if the class name is missing (array is
     * empty)
     * @throws ClassNotFoundException if the class cannot be located
     * @throws InstantiationException if instantiation failed
     * @throws SecurityException if access to the required reflection
     * information is denied
     */
    public static Object createObject(final String[] classDef)
            throws IllegalArgumentException, ClassNotFoundException,
            InstantiationException, SecurityException {
        final int length = classDef.length;

        // check arguments
        if (length == 0) {
            throw new IllegalArgumentException(
                "Class name missing (empty class definition)");
        }

        // first element is class name, further elements are constructor args
        final String type = classDef[0];
        final String[] params = new String[length - 1];

        for (int i = 0; i < params.length; i++) {
            params[i] = classDef[i + 1];
        }

        return createObject(Class.forName(type), params);
    }

    /**
     * Delegates to {@link Util#createObject(Class, Object[], Class)}, setting
     * the <code>paramType</code> to the {@link String} class.
     *
     * @param type the class of the object to create
     * @param params the objects in this array are passed to the constructor;
     * if <code>null</code> or empty, the default constructor is used
     * @return the created object; the returned object will be an instance of
     * <code>type</code>
     * @throws InstantiationException if instantiation failed
     * @throws SecurityException if access to the required reflection
     * information is denied
     */
    public static Object createObject(final Class type, final String[] params)
            throws InstantiationException, SecurityException {
        return createObject(type, params, String.class);
    }

    /**
     * Creates an object of a specified type. The object must have a public
     * constructor that accepts <code>params.length</code> arguments of the
     * specified <code>paramType</code>.
     *
     * @param type the class of the object to create
     * @param params the objects in this array are passed to the constructor;
     * if <code>null</code> or empty, the default constructor is used
     * @param paramType the type of the elements stored in the
     * <code>params</code> array
     * @return the created object; the returned object will be an instance of
     * <code>type</code>
     * @throws InstantiationException if instantiation failed
     * @throws SecurityException if access to the required reflection
     * information is denied
     */
    public static Object createObject(final Class type, final Object[] params,
            final Class paramType)
            throws InstantiationException, SecurityException {
        final Class[] paramClasses = new Class[params.length];
        Arrays.fill(paramClasses, paramType);
        return createObject(type, params, paramClasses);
    }

    /**
     * Creates an object of a specified type. The object must have a public
     * constructor that accepts <code>params.length</code> arguments of the
     * specified types.
     *
     * @param type the class of the object to create
     * @param params the objects in this array are passed to the constructor;
     * if <code>null</code> or empty, the default constructor is used
     * @param paramTypes the types of the elements stored in the
     * <code>params</code> list
     * @return the created object; the returned object will be an instance of
     * <code>type</code>
     * @throws IllegalArgumentException if <code>params</code> and
     * <code>paramTypes</code> are of different lengths
     * @throws InstantiationException if instantiation failed
     * @throws SecurityException if access to the required reflection
     * information is denied
     */
    public static Object createObject(final Class type, final Object[] params,
            final Class[] paramTypes) throws IllegalArgumentException,
            InstantiationException, SecurityException {
        // check that array lengths match
        if ((params != null) && (params.length != paramTypes.length)) {
            throw new IllegalArgumentException("Lengths of given arrays of "
                + " parameters and parameter types differ: " + params.length
                + ", " + paramTypes.length);
        }

        Object result;
        try {
            if ((params == null) || (params.length == 0)) {
                // try default constructor
                result = type.newInstance();
            } else {
                final Constructor cons = type.getConstructor(paramTypes);

                // pass each of the list elements as constructor argument
                result = cons.newInstance(params);
            }
        } catch (IllegalAccessException iae) {        // repackage exception
            final InstantiationException ie =
                new InstantiationException(iae.toString());
            ie.initCause(iae);
            throw ie;
        } catch (InvocationTargetException ite) {     // repackage exception
            final InstantiationException ie =
                new InstantiationException(ite.toString());
            ie.initCause(ite);
            throw ie;
        } catch (ExceptionInInitializerError eiie) {  // repackage exception
            final InstantiationException ie =
                new InstantiationException(eiie.toString());
            ie.initCause(eiie);
            throw ie;
        } catch (NoSuchMethodException nsme) {        // repackage exception
            final InstantiationException ie =
                new InstantiationException(nsme.toString());
            ie.initCause(nsme);
            throw ie;
        }
        return result;
    }

    /**
     * Formats a duration. The generated string has the form
     * "<i>days</i>D+<i>hours</i>:<i>minutes</i>:<i>secs</i>.<i>millisecs</i>".
     * Days and milliseconds are omitted if 0; the time fields are omitted if
     * all of them are 0, but only if days are present (to avoid returning an
     * empty string). Samples: 2D+03:02:43.666, 01:00:00, 1D, 00:05:00.123,
     * 1D+04:19:03, 00:00:00.
     *
     * @param msecs the number of milliseconds
     * @return a formatted representation as specified above
     */
    public static String formatDurationInMillisecs(final long msecs) {
        // check argument
        if (msecs < 0.0) {
            throw new IllegalArgumentException(
                "Negative durations are not supported: " + msecs);
        }

        final int milliRange = 1000;
        final int secMinRange = 60;
        final int hourRange = 24;

        final int millis = (int) (msecs % milliRange);
        long rest = msecs / milliRange;
        final int seconds = (int) (rest % secMinRange);
        rest = rest / secMinRange;
        final int minutes = (int) (rest % secMinRange);
        rest /= secMinRange;
        final int hours = (int) (rest % hourRange);
        rest /= hourRange;

        final StringBuffer result = new StringBuffer();

        // start with days, if any
        if (rest > 0) {
            result.append(rest).append("D");
        }

        // don't print anything else if days are present and all time fields
        // are 0
        if ((rest == 0) || (hours != 0) || (minutes != 0) || (seconds != 0)
                || (millis != 0)) {
            // constant for padding + printing
            // append "+" separator if days are present
            if (rest > 0) {
                result.append("+");
            }

            // always append hours:minutes:secs
            appendPositiveNumber(result, hours, 2, null, true);
            appendPositiveNumber(result, minutes, 2, ":", true);
            appendPositiveNumber(result, seconds, 2, ":", true);

            // append milliseconds if not 0
            appendPositiveNumber(result, millis, 3, ".", false);
        }

        return result.toString();
    }

    /**
     * Formats a duration. The generated string has the form
     * "<i>days</i>D+<i>hours</i>:<i>minutes</i>:<i>secs</i>.<i>millisecs</i>".
     * Days and milliseconds are omitted if 0; the time fields are omitted if
     * all of them are 0, but only if days are present (to avoid returning an
     * empty string). Samples: 2D+03:02:43.666, 01:00:00, 1D, 00:05:00.123,
     * 1D+04:19:03, 00:00:00.
     *
     * @param secs the number of seconds; must be nonnegative
     * @return a formatted representation as specified above
     * @throws IllegalArgumentException if <code>secs &lt; 0</code>
     */
    public static String formatDurationInSeconds(final double secs)
            throws IllegalArgumentException {
        return formatDurationInMillisecs(Math.round(secs * 1000));
    }

    /**
     * Ensures that a number is positive or 0, throwing an exception if this
     * condition is violated.
     *
     * @param number the number to check
     * @param identifier string used to describe the number in the thrown
     * exception
     * @throws IllegalArgumentException if <code>number &lt; 0.0</code>
     */
    public static void ensureNonNegative(final double number,
            final String identifier) throws IllegalArgumentException {
        if (number < 0.0) {
            throw new IllegalArgumentException(identifier
                + " must not be negative: " + number);
        }
    }

    /**
     * Ensures that a number is positive or 0, throwing an exception if this
     * condition is violated.
     *
     * @param number the number to check
     * @param identifier string used to describe the number in the thrown
     * exception
     * @throws IllegalArgumentException if <code>number &lt; 0</code>
     */
    public static void ensureNonNegative(final long number,
            final String identifier) throws IllegalArgumentException {
        if (number < 0) {
            throw new IllegalArgumentException(identifier
                + " must not be negative: " + number);
        }
    }

    /**
     * Formats the given number, using at most 4 fraction digits.
     *
     * @param number the number for format
     * @return the formatted number
     */
    public static String format(final double number) {
        return format(number, 4);
    }

    /**
     * Formats the given number, using at most the specified number of
     * fraction digits.
     *
     * @param number the number for format
     * @param maxFractDigits maximum number of digits allowed in the fraction
     * portion of the number
     * @return the formatted number
     */
    public static String format(final double number, final int maxFractDigits) {
        synchronized (NUM_FORMAT) {
            NUM_FORMAT.setMaximumFractionDigits(maxFractDigits);
            return NUM_FORMAT.format(number);
        }
    }

    /**
     * Returns an instance of a pseudo-random number generator that uses a
     * fixed seed, so the same sequence of calls to each object created
     * by this method will yield the same sequence of pseudo-random numbers.
     *
     * @return a new source of pseudo-randomness that will return
     * reproducable pseudo-random numbers
     */
    public static Random reproducibleRandom() {
        // using a fixed seed which was randomly generated and will never change
        return new Random(4127208836813110743L);
    }

    /**
     * Calculated and prints the time passed since a given start time
     * (in English).
     *
     * @param startTime the start time in milliseconds
     * @return a formatted string containing the time difference between the
     * start time and the {@linkplain System#currentTimeMillis() current time}
     */
    public static String showDuration(final long startTime) {
        final long timeInMS = System.currentTimeMillis() - startTime;
        final double timeInSecs = ((double) timeInMS) / 1000;
        String result;

        synchronized (SEC_FORMAT) {
            result = "elapsed time: " + SEC_FORMAT.format(timeInSecs) + "s";
        }

        // also convert to XML Schema duration etc. if 5min or more
        if (timeInSecs >= 300) {
            result += " = " + formatDurationInSeconds(timeInSecs);
        }

        return result;
    }

    /**
     * <code>Util</code> instances should NOT be constructed in standard
     * programming. Instead, the method of this class should be called in a
     * static way. This constructor is public to permit tools that require a
     * JavaBean instance to operate.
     */
    public Util() {
        super();
    }

}
