/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.util;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.SortedSet;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;
import org.apache.velocity.context.Context;
import org.apache.velocity.exception.VelocityException;
import org.apache.velocity.runtime.RuntimeConstants;

import de.fu_berlin.ties.TiesConfiguration;
import de.fu_berlin.ties.io.IOUtils;

/**
 * A static class that provides a convenience interface to a Velocity
 * singletons. No instances of this class can be created, only the static
 * members should be used.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.17 $, $Date: 2004/11/08 11:57:48 $, $Author: siefkes $
 */
public final class VelocityService {

    /**
     * The changes to the Velocity configuration made by default.
     */
    private static final Properties DEFAULT_PROPERTIES;

    /**
     * The template directory.
     */
    public static final String TEMPLATE_DIR = "velocity" + File.separator;

    /**
     * The extension of template files.
     */
    public static final String TEMPLATE_EXT = ".vm";

    /**
     * Name of the default template used to print configuration parameters.
     */
    public static final String CONFIG_TEMPLATE = completeTemplateName("config");

    /**
     * Name of the default template used to print goals.
     */
    public static final String GOAL_TEMPLATE = completeTemplateName("goals");

    /**
     * Name of the main template used for Anakia XML format.
     */
    public static final String ANAKIA_TEMPLATE = completeTemplateName("xdoc");

    /**
     * Static initializer populates the default properties and configures the
     * Velocity singleton.
     */
    static {
        // init and populate properties to change
        DEFAULT_PROPERTIES = new Properties();
        // load resources from both filesystem (default) and classpath
        DEFAULT_PROPERTIES.setProperty(
            RuntimeConstants.RESOURCE_LOADER, "file,classpath");
        // configure file loader to use template caching (check every 2min)
        DEFAULT_PROPERTIES.setProperty(
            RuntimeConstants.FILE_RESOURCE_LOADER_CACHE, "true");
        DEFAULT_PROPERTIES.setProperty(
            "file.resource.loader.modificationCheckInterval", "120");
        // specify classpath loader
        DEFAULT_PROPERTIES.setProperty("classpath.resource.loader.class",
            "org.apache.velocity.runtime.resource.loader."
                + "ClasspathResourceLoader");
        DEFAULT_PROPERTIES.setProperty("classpath.resource.loader.description",
            "Velocity Classpath Resource Loader");

        try {
            /* use shared logger
            Velocity.setProperty(RuntimeConstants.RUNTIME_LOG_LOGSYSTEM,
                Log.TIES);
            Velocity.setProperty(RuntimeConstants.RUNTIME_LOG_LOGSYSTEM_CLASS,
                "org.apache.velocity.runtime.log.AvalonLogSystem");
            Velocity.setProperty("runtime.log.logsystem.avalon.logger",
                Log.TIES);
            */
            // configure with properties
            Velocity.init(DEFAULT_PROPERTIES);
        } catch (Exception e) {
            Util.LOG.error(
                "VelocityService: could not initialize the Velocity singleton",
                e);
        }
    }

    /**
     * Completes the short form of a template name, by prepending the
     * {@link #TEMPLATE_DIR} and appending the {@link #TEMPLATE_EXT}.
     *
     * @param shortName the short form of the template name (without directory
     * and extension)
     * @return the full form of the template name, which is assumed to be
     * in the default directory and have the default extension
     */
    public static String completeTemplateName(final String shortName) {
        return TEMPLATE_DIR + shortName + TEMPLATE_EXT;
    }

    /**
     * Returns properties listing the changes to the Velocity configuration
     * made by default.
     *
     * @return a {@link java.util.Properties} object containing the changed
     * parameters
     */
    public static Properties getDefaultProperties() {
        return DEFAULT_PROPERTIES;
    }

    /**
      * Main methods: prints the list of config parameters and
      * {@linkplain #printGoals(TiesConfiguration, Writer) goals} from the
      * {@link TiesConfiguration#CONF standard configuration} to specified
      * files (or standard out). The generated files are in Anakia XML format
      * and uses the platform's default character set.
      *
      * @param args the command-line arguments: should contain the filename
      * for writing the config parameters as first element and the filename
      * for writing the goals as second element; if no filenames are given, the
      * corresponding output is written to standard out
      */
     public static void main(final String[] args) {
         Writer confWriter = null;
         Writer goalWriter = null;

         try {
             // write config parameters
             confWriter = (args.length > 0)
                 ? new FileWriter(args[0]) // use first file name
                 : new OutputStreamWriter(System.out); // write to stdout
             final Map<String, Object> confContextObjects =
                 new HashMap<String, Object>();
             confContextObjects.put("contents", CONFIG_TEMPLATE);
             confContextObjects.put("title", "TIE Configuration Parameters");
             confContextObjects.put("keywords", "TIE, configuration, "
                 + "configuration parameters, properties, default values");
             printConfigProperties(TiesConfiguration.CONF, ANAKIA_TEMPLATE,
                IOUtils.STANDARD_UNICODE_CHARSET, confContextObjects,
                    confWriter);

             // write goal processors
             goalWriter = (args.length > 1)
                 ? new FileWriter(args[1]) // use second file name
                 : new OutputStreamWriter(System.out); // write to stdout
             final Map<String, Object> goalContextObjects =
                 new HashMap<String, Object>();
             goalContextObjects.put("contents", GOAL_TEMPLATE);
             goalContextObjects.put("title", "TIE Processing Goals");
             goalContextObjects.put("keywords", "TIE, goals, "
                 + "processors, processing goals, goal processors");
             printGoals(TiesConfiguration.CONF, ANAKIA_TEMPLATE,
                IOUtils.STANDARD_UNICODE_CHARSET, goalContextObjects,
                goalWriter);
         } catch (Exception e) {
             e.printStackTrace();
         } finally {
             // close writers
             IOUtils.tryToClose(confWriter);
             IOUtils.tryToClose(goalWriter);
         }
     }

    /**
     * Prints all properties contained in a configuration, using the
     * specified template for rendering. In addition to the user-specified
     * context objects (if any), the template has access to a collection
     * <code>$allParameters</code> of all config parameters.
     *
     * @param config the configuration to print
     * @param templateName name of template to be used in merge, either in
     * short form (without directory and extension, when default directory and
     * extension are used) or in complete form
     * @param charset the character set used in the template
     * @param contextObjects a map of objects that are made available in the
     * context of the template; might be <code>null</code>
     * @param writer the writer to write the output to; the writer is flushed
     * but not closed by this method
     * @throws IOException if an I/O error occurred
     * @throws VelocityException if rendering failed
     */
    public static void printConfigProperties(final TiesConfiguration config,
            final String templateName, final String charset,
            final Map<String, Object> contextObjects, final Writer writer)
            throws IOException, VelocityException {
        // populate from user-specified map, if not null
        final Map<String, Object> contextMap = (contextObjects == null)
            ? new HashMap<String, Object>()
            : new HashMap<String, Object>(contextObjects);

        // create and store descriptors for all key, incl. non-set keys
        final SortedSet keySet = config.sortedKeys(true);
        final Iterator keys = keySet.iterator();
        final List<TiesConfiguration.EntryDescriptor> allParams =
            new ArrayList<TiesConfiguration.EntryDescriptor>(keySet.size());
        String key;
        TiesConfiguration.EntryDescriptor descriptor;

        while (keys.hasNext()) {
            key = (String) keys.next();
            descriptor = config.getDescriptor(key);

            if (descriptor != null) {
                allParams.add(descriptor);
            } else {
                Util.LOG.warn("Missing description for key " + key);
            }
        }

        contextMap.put("allParameters", allParams);
        VelocityService.renderTemplate(templateName, charset, contextMap,
            writer);
    }

    /**
     * Prints all properties contained in a configuration, using the
     * default template for printing them. This templates generates an
     * (X)HTML fragment listing all configuration parameters with their
     * types and default values in a table.
     *
     * @param config the configuration to print
     * @param writer the writer to write the output to; the writer is flushed
     * but not closed by this method
     * @throws IOException if an I/O error occurred
     * @throws VelocityException if rendering failed
     */
    public static void printConfigProperties(final TiesConfiguration config,
            final Writer writer) throws IOException, VelocityException {
        printConfigProperties(config, CONFIG_TEMPLATE,
            IOUtils.STANDARD_UNICODE_CHARSET, null, writer);
    }

    /**
     * Prints the goals contained in a configuration, using the default
     * template for printing them. In addition to the user-specified
     * context objects (if any), the template has access to a collection
     * <code>$allParameters</code> of all goals.
     *
     * @param config the configuration to print
     * @param templateName name of template to be used in merge, either in
     * short form (without directory and extension, when default directory and
     * extension are used) or in complete form
     * @param charset the character set used in the template
     * @param contextObjects a map of objects that are made available in the
     * context of the template; might be <code>null</code>
     * @param writer the writer to write the output to; the writer is flushed
     * but not closed by this method
     * @throws IOException if an I/O error occurred
     * @throws VelocityException if rendering failed
     */
    public static void printGoals(final TiesConfiguration config,
            final String templateName, final String charset,
            final Map<String, Object> contextObjects, final Writer writer)
            throws IOException, VelocityException {
        final TiesConfiguration goalSubset = (TiesConfiguration)
            config.subset(TiesConfiguration.CONFIG_GOAL_PREFIX);
        printConfigProperties(goalSubset, templateName,
            charset, contextObjects, writer);
    }

    /**
     * Prints the goals contained in a configuration, using the default
     * template for printing them. This templates generates an (X)HTML fragment
     * listing all goals with their output extension (if given) and any further
     * arguments in a table.
     *
     * @param config the configuration to print
     * @param writer the writer to write the output to; the writer is flushed
     * but not closed by this method
     * @throws IOException if an I/O error occurred
     * @throws VelocityException if rendering failed
     */
    public static void printGoals(final TiesConfiguration config,
            final Writer writer) throws IOException, VelocityException {
        printGoals(config, GOAL_TEMPLATE, IOUtils.STANDARD_UNICODE_CHARSET,
            null, writer);
    }

    /**
     * Renders a template using the given context objects. The template must
     * be in the standard Unicode character set (UFT-8).
     *
     * @param templateName name of template to be used in merge, either in
     * short form (without directory and extension, when default directory and
     * extension are used) or in complete form
     * @param contextObjects a map of objects that are made available in the
     * context of the template; mappings from "escapeTool" to a
     * {@link StringEscapeUtils} and from "util" to a {@link Util} are added to
     * allow XML entity escaping, String conversion etc.
     * @param writer writer to write template into; the writer is flushed
     * but not closed by this method
     * @throws IOException if an I/O error occurred
     * @throws VelocityException if rendering failed
     */
    public static void renderTemplate(final String templateName,
            final Map contextObjects, final Writer writer)
            throws IOException, VelocityException {
        renderTemplate(templateName, IOUtils.STANDARD_UNICODE_CHARSET,
            contextObjects, writer);
    }

    /**
     * Renders a template using the given context objects.
     *
     * @param templateName name of template to be used in merge, either in
     * short form (without directory and extension, when default directory and
     * extension are used) or in complete form
     * @param charset the character set used in template
     * @param contextObjects a map of objects that are made available in the
     * context of the template; mappings from "escapeTool" to a
     * {@link StringEscapeUtils} and from "util" to a {@link Util} are added to
     * allow XML entity escaping, String conversion etc.
     * @param writer writer to write template into; the writer is flushed
     * but not closed by this method
     * @throws IOException if an I/O error occurred
     * @throws VelocityException if rendering failed
     */
    public static void renderTemplate(final String templateName,
            final String charset, final Map contextObjects, final Writer writer)
            throws IOException, VelocityException {
        final Context context = new VelocityContext(contextObjects);

        // "embed" specified name within default directory and extension
        final String embeddedTemplateName = completeTemplateName(templateName);

        // use the embedded name if it exists, otherwise use the name "as is"
        final String actualTemplateName =
            Velocity.resourceExists(embeddedTemplateName)
                ? embeddedTemplateName : templateName;

        // add utilities
        context.put("escapeTool", new StringEscapeUtils());
        context.put("util", new Util());
        boolean result = false;

        try {
            result = Velocity.mergeTemplate(actualTemplateName, charset,
                context, writer);
            writer.flush();
        } catch (IOException ioe) {
            throw ioe; // can rethow "as is"
        } catch (VelocityException ve) {
            throw ve; // can rethow "as is"
        } catch (RuntimeException re) {
            throw re; // can rethow "as is"
        } catch (Exception e) {
            // repackaging the exception in a more suitable type
            final VelocityException ve = new VelocityException(
                "Exception during rendering: " + e.toString());
            ve.initCause(e);
            throw ve;
        }

        // something went wrong, but we don't know what
        if (!result) {
            throw new VelocityException("Velocity could not render the template"
                + "-- check log for errors");
        }
    }

    /**
     * Private constructor prevents creation of instances.
     */
    private VelocityService() {
        super();
    }
}
