/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml;

import java.util.LinkedList;
import java.util.List;
import java.util.SortedMap;

import org.apache.commons.collections.MultiHashMap;
import org.apache.commons.collections.MultiMap;
import org.apache.commons.collections.map.LinkedMap;

/**
 * A container used to manage the currently open start tags during XML
 * adjustment. Adds stack-like push and pop functionality to the superclass
 * and allows checking specifically for
 * {@link de.fu_berlin.ties.xml.TagVariety#TENTATIVE tentative} tag.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2004/08/23 17:11:06 $, $Author: siefkes $
 */
public class OpenTags extends TagContainer {

    /**
     * A dummy object stored as map value where no value is required.
     */
    protected static final Object DUMMY = new Object();

    /**
     * Contains all tentative tags (checking the variety at insertion
     * time), mapping from tag name Strings to Collections of TagConstituents.
     */
    private final MultiMap tentativeTags = new MultiHashMap();

    /**
     * Contains all non-tentative tags (checking the variety at insertion
     * time), mapping from tag name Strings to Collections of TagConstituents.
     */
    private final MultiMap nonTentativeTags = new MultiHashMap();

    /**
     * Additionally preserves order of insertion for all TagConstituents,
     * for push/pop/peek operations. We would prefer to use an ordered
     * <em>set</em>, but there is no "LinkedSet" class and
     * {@link java.util.LinkedHashSet} doesn't provide fast access to the
     * <em>last</em> element (only to the first ones). Thus we just store
     * a {@link #DUMMY} reference in each value, effectively using the set
     * part of this map.
     */
    private final LinkedMap tagSequence = new LinkedMap();

    /**
     * Creates a new instance.
     */
    public OpenTags() {
        super();
    }

    /**
     * Checks whether this instance contains at least one appearance of the
     * specified tag that is not {@link TagVariety#TENTATIVE}.
     *
     * @param tagname the name of the tag to check
     * @return <code>true</code> iff the container contains one or more
     * non-tenatative (at insertion time) appearances of the specified tag
     */
    public boolean containsNonTentative(final String tagname) {
        // check tentative tags entries for tag name
        final List appearances = (List) nonTentativeTags.get(tagname);
        return (appearances != null) && !appearances.isEmpty();
    }

    /**
     * Finds a tentative appearance of a tag, if any exists.
     * If several tentative appearances exist, the <em>last</em> of them is
     * returned.
     *
     * @param tagname the name of the tag to check
     * @return the found apearance of the tag, or <code>null</code> if no
     * matching appearance was found
     */
    public TagConstituent findTentativeTag(final String tagname) {
        // check tentative tags entries for tag name
        final List appearances = (List) tentativeTags.get(tagname);
        final TagConstituent result;

        if ((appearances != null) && !appearances.isEmpty()) {
            // return last entry
            result =
                (TagConstituent) appearances.get(appearances.size() - 1);
            // Log.TIES.info("Found tentative tag: " + result);
        } else {
            result = null;
        }
        return result;
    }

    /**
     * Checks whether a tag is the root tag (the first tag inserted into this
     * container).
     *
     * @param tag the tag to check
     * @return <code>true</code> if the tag is the root tag
     */
    public boolean isRoot(final TagConstituent tag) {
        return (tag != null) && (tag == peekFirst());
    }

    /**
     * Returns the first tag appearance that was initially
     * {@link #push(TagConstituent) push}ed into this container.
     *
     * @return the first undeleted tag pushed into this container,
     * or <code>null</code> if the container is empty
     */
    protected TagConstituent peekFirst() {
        return (tagSequence.isEmpty())
            ? null : (TagConstituent) tagSequence.firstKey();
    }

    /**
     * Returns the last tag appearance that was
     * {@link #push(TagConstituent) push}ed into this container.
     *
     * @return the last undeleted tag pushed into this container,
     * or <code>null</code> if the container is empty
     */
    public TagConstituent peek() {
        return (tagSequence.isEmpty())
            ? null : (TagConstituent) tagSequence.lastKey();
    }

    /**
     * Removes and returns the last tag appearance that was
     * {@link #push(TagConstituent) push}ed into this container.
     *
     * @return the removed tag
     * @throws NullPointerException if the container is empty (i.e. a
     * call to {@link #peek()} would return <code>null</code>)
     */
    public TagConstituent pop() throws NullPointerException {
        final TagConstituent tagToRemove = peek();
        final String tagname = tagToRemove.getName();
        final Integer seriesNo = new Integer(tagToRemove.getMarkupSeriesNo());
        tagSequence.remove(tagToRemove);

        // remove it from (non)tentative tags, if contained there
        tentativeTags.remove(tagname, tagToRemove);
        nonTentativeTags.remove(tagname, tagToRemove);

        // we know the map must exist
        final SortedMap nestedMap = (SortedMap) getTagMap().get(tagname);

        // we know the list must exist and the tag must be the last entry
        final LinkedList tagsInSeries =
            (LinkedList) nestedMap.get(seriesNo);
        final TagConstituent removedTag =
            (TagConstituent) tagsInSeries.removeLast();

        if (tagsInSeries.isEmpty()) {
            // inner list is now empty -- remove it
            nestedMap.remove(seriesNo);
            if (nestedMap.isEmpty()) {
                // nested map is now empty -- remove it too
                getTagMap().remove(tagname);
            }
        }

        if (removedTag != tagToRemove) {
            // should never happen
            throw new RuntimeException("Implementation error: "
                + tagToRemove + " and " + removedTag
                + " should be the same object, but the aren't!");
        }
        return tagToRemove;
    }

    /**
     * Removes and returns the last tag appearance that was
     * {@link #push(TagConstituent) push}ed into this container. The variety of
     * the removed tag is set to {@link TagVariety#REGULAR} (independently of
     * its previous value).
     *
     * @return <code>true</code> if the tag was irregular and thus had to be
     * regularized; <code>false</code> if it was a regular tag
     * @throws NullPointerException if the container is empty (i.e. a
     * call to {@link #peek()} would return <code>null</code>)
     */
    public boolean popAndRegularize() throws NullPointerException {
        final TagConstituent popped = pop();
        final boolean result;
        if (popped.getVariety() != TagVariety.REGULAR) {
            // adjust variety
            popped.setVariety(TagVariety.REGULAR);
            result = true;
        } else {
            result = false;
        }
        return result;
    }

    /**
     * Inserts an appearance of a tag into the container. The name of the
     * tag is used as key; the whole constituent is appended to the list of
     * appearances stored for this tag name. This implementation doesn't support
     * inserting at the begin of af the list because it would hinder the stack
     * operations. Instances of this class accept only start tags.
     *
     * @param tag the tag to insert
     * @param appendAtEnd must always be <code>true</code>
     * @throws IllegalArgumentException if the specified tag is not a start tag
     * @throws UnsupportedOperationException if <code>appendAtEnd</code> is
     * <code>false</code> (prepending is not supported)
     */
    public void push(final TagConstituent tag, final boolean appendAtEnd)
            throws IllegalArgumentException, UnsupportedOperationException {
        // check arguments
        if (tag.getType() != TagConstituent.START_TAG) {
            throw new IllegalArgumentException(
                "OpenTags allows only start tags, not tags of type "
                    + tag.getType() + " (" + tag + ")");
        }
        if (!appendAtEnd) {
            throw new UnsupportedOperationException("OpenTags.push cannot "
                + "handle prepending -- would destroy stack operations");
        }

        // delegate to superclass
        super.push(tag, appendAtEnd);

        // add tag to tag sequence
        tagSequence.put(tag, DUMMY);

        // also add to tentative resp. non-tenative tags
        if (tag.getVariety() == TagVariety.TENTATIVE) {
            tentativeTags.put(tag.getName(), tag);
        } else {
            nonTentativeTags.put(tag.getName(), tag);
        }
    }

    /**
     * Removes a single appearance of a tag from the container.
     *
     * @param tag the tag to remove
     * @return <code>true</code> if the specified TagConstituent was
     * removed successfully; <code>false</code> otherwise (the specified
     * constituent wasn't found in the container)
     */
    public boolean remove(final TagConstituent tag) {
        // remove from tentative resp. non-tenative tags, if contained there
        // (we do this always for both because the tag variety could have been
        // changed after the tag was added)
        tentativeTags.remove(tag.getName(), tag);
        nonTentativeTags.remove(tag.getName(), tag);

        // delegate to superclass
        final boolean superResult = super.remove(tag);

        // also remove from tag sequence
        return superResult && (tagSequence.remove(tag) == DUMMY);
    }

}
