/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * A constituent in an XML document that is not a tag. An instance of this class
 * represents a document or markup declaration, XML prolog, processing
 * instruction, comment, CDATA section, textual content or outer whitespace.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/02/02 18:50:16 $, $Author: siefkes $
 */
public class OtherConstituent extends XMLConstituent {

    /**
     * Type constant: a document type declaration.
     */
    public static final short DOCTYPE = 3;

    /**
     * Type constant: the XML prolog.
     */
    public static final short XML_PROLOG = 4;

    /**
     * Type constant: a processing instruction.
     */
    public static final short PI = 5;

    /**
     * Type constant: a comment.
     */
    public static final short COMMENT = 6;

    /**
     * Type constant: a CDATA section.
     */
    public static final short CDATA_SECTION = 7;

    /**
     * Type constant: textual content. Textual content is trimmed, preceding
     * and trailing whitespace are stored as {@link #OUTER_WHITESPACE}
     * constituents.
     */
    public static final short TEXT = 8;

    /**
     * Type constant: Whitespace between elements.
     */
    public static final short OUTER_WHITESPACE = 9;

    /**
     * Creates a new instance, without setting a reference to a next
     * constituent.
     *
     * @param constType the typ of this constituent (must be one of the static
     * type constants defined in this class)
     * @param rep the representation of this constituent within the XML
     * document, i.e. the string fragment from the document representing this
     * constituent
     * @throws IllegalArgumentException if <code>constType</code> differs from
     * the static type constants defined in this class
     */
    public OtherConstituent(final short constType, final String rep)
            throws IllegalArgumentException {
        this(constType, rep, null, null);
    }

    /**
     * Creates a new instance.
     *
     * @param constType the typ of this constituent (must be one of the static
     * type constants defined in this class)
     * @param rep the representation of this constituent within the XML
     * document, i.e. the string fragment from the document representing this
     * constituent
     * @param prevEntry a reference to the next constituent
     * @param nextEntry a reference to the next constituent
     * @throws IllegalArgumentException if <code>constType</code> differs from
     * the static type constants defined in this class
     */
    public OtherConstituent(final short constType, final String rep,
            final XMLConstituent prevEntry,
            final XMLConstituent nextEntry)
            throws IllegalArgumentException {
        super(constType, rep, prevEntry, nextEntry);

        // ensure range of type constant
        if ((constType < DOCTYPE) || (constType > OUTER_WHITESPACE)) {
            throw new IllegalArgumentException("Specified tag type must "
                + "be one of the static type constants defined in "
                + "TagConstituent, but it is " + constType);
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this).
            appendSuper(super.toString()).
            toString();
    }

}
