/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * A tag in an XML document. An instance of this class represents a single start
 * tag, end tag or empty tag.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/02/02 18:50:16 $, $Author: siefkes $
 */
public class TagConstituent extends XMLConstituent {

    /**
     * Type constant: a start tag.
     */
    public static final short START_TAG = 0;

    /**
     * Type constant: an end tag.
     */
    public static final short END_TAG = 1;

    /**
     * Type constant: an empty tag.
     */
    public static final short EMPTY_TAG = 2;

    /**
     * The name of this tag.
     */
    private final String name;

    /**
     * The number of the markup series this tag is part of. Markup series are
     * series of markup (tags, PIs, comments) not interrupted by textual
     * content.
     */
    private int markupSeriesNo = -1;

    /**
     * The {@link TagVariety} of this tag.
     * Defaults to {@link TagVariety#REGULAR}.
     */
    private TagVariety variety = TagVariety.REGULAR;

    /**
     * Creates a new instance, automatically generating a suitable
     * representation without attributes (i.e. "&lt;<em>tagName</em>&gt;" for
     * a start tag, "&lt;/<em>tagName</em>&gt;" for an end tag,
     * "&lt;<em>tagName</em>/&gt;" for an empty tag).
     *
     * @param constType the typ of this constituent (must be one of the static
     * type constants defined in this class)
     * @param tagName the name of this tag
     * @throws IllegalArgumentException if <code>constType</code> differs from
     * the static type constants defined in this class
     */
    public TagConstituent(final short constType, final String tagName)
            throws IllegalArgumentException {
        this(constType, tagName, -1);
    }

    /**
     * Creates a new instance, automatically generating a suitable
     * representation without attributes (i.e. "&lt;<em>tagName</em>&gt;" for
     * a start tag, "&lt;/<em>tagName</em>&gt;" for an end tag,
     * "&lt;<em>tagName</em>/&gt;" for an empty tag).
     *
     * @param constType the typ of this constituent (must be one of the static
     * type constants defined in this class)
     * @param tagName the name of this tag
     * @param markupSeries the number of the markup series this tag is part of
     * @throws IllegalArgumentException if <code>constType</code> differs from
     * the static type constants defined in this class
     */
    public TagConstituent(final short constType, final String tagName,
            final int markupSeries)
            throws IllegalArgumentException {
        this(constType, tagName,
            '<' + (constType == END_TAG ? "/" : "") + tagName
                + (constType == EMPTY_TAG ? "/" : "") + '>',
            markupSeries, null, null);
    }

    /**
     * Creates a new instance, without storing a markup series number.
     *
     * @param constType the typ of this constituent (must be one of the static
     * type constants defined in this class)
     * @param tagName the name of this tag
     * @param rep the representation of this constituent within the XML
     * document, i.e. the string fragment from the document representing this
     * constituent
     * @throws IllegalArgumentException if <code>constType</code> differs from
     * the static type constants defined in this class
     */
    public TagConstituent(final short constType, final String tagName,
            final String rep) throws IllegalArgumentException {
        this(constType, tagName, rep,  -1, null, null);
    }

    /**
     * Creates a new instance, without setting a reference to a next
     * constituent.
     *
     * @param constType the typ of this constituent (must be one of the static
     * type constants defined in this class)
     * @param tagName the name of this tag
     * @param rep the representation of this constituent within the XML
     * document, i.e. the string fragment from the document representing this
     * constituent
     * @param markupSeries the number of the markup series this tag is part of
     * @throws IllegalArgumentException if <code>constType</code> differs from
     * the static type constants defined in this class
     */
    public TagConstituent(final short constType, final String tagName,
            final String rep, final int markupSeries)
            throws IllegalArgumentException {
        this(constType, tagName, rep,  markupSeries, null, null);
    }

    /**
     * Creates a new instance.
     *
     * @param constType the typ of this constituent (must be one of the static
     * type constants defined in this class)
     * @param tagName the name of this tag
     * @param rep the representation of this constituent within the XML
     * document, i.e. the string fragment from the document representing this
     * constituent
     * @param markupSeries the number of the markup series this tag is part of
     * @param prevEntry a reference to the previous constituent
     * @param nextEntry a reference to the next constituent
     * @throws IllegalArgumentException if <code>constType</code> differs from
     * the static type constants defined in this class
     */
    public TagConstituent(final short constType, final String tagName,
            final String rep, final int markupSeries,
            final XMLConstituent prevEntry, final XMLConstituent nextEntry)
            throws IllegalArgumentException {
        super(constType, rep, prevEntry, nextEntry);

        // ensure range of type constant
        if ((constType < START_TAG) || (constType > EMPTY_TAG)) {
            throw new IllegalArgumentException("Specified tag type must "
                + "be one of the static type constants defined in "
                + "TagConstituent, but it is " + constType);
        }
        name = tagName;
        markupSeriesNo = markupSeries;
    }

    /**
     * Returns the name of this tag.
     *
     * @return the tag name
     */
    public final String getName() {
        return name;
    }

    /**
     * Returns the number of the markup series this tag is part of.
     *
     * @return the series number; or <code>-1</code> if the markup series number
     * is not known
     */
    public int getMarkupSeriesNo() {
        return markupSeriesNo;
    }

    /**
     * Returns the {@link TagVariety} of this tag.
     * Defaults to {@link TagVariety#REGULAR}.
     *
     * @return the value
     */
    public TagVariety getVariety() {
        return variety;
    }

    /**
     * Sets the number of the markup series this tag is part of. Tag series are
     * series of tags not interrupted by textual content.
     *
     * @param i the new number
     */
    public void setMarkupSeriesNo(final int i) {
        markupSeriesNo = i;
    }

    /**
     * Specifies the {@link TagVariety} of this tag.
     *
     * @param newValue the new value to set
     */
    public void setVariety(final TagVariety newValue) {
        variety = newValue;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .appendSuper(super.toString())
            .append("name", name)
            .append("markup series no", markupSeriesNo)
            .append("variety", variety)
            .toString();
    }

}
