/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * This class manages tag sequences by keeping track of names and appearances
 * of tags. It is a helper class used for XML adjustment.
 * Instances of this class are <em>not</em> thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/09/07 10:01:48 $, $Author: siefkes $
 */
public class TagContainer {

    /**
     * Stores the mappings from tag names Strings to nested TreeMaps
     * mapping markup series Integers to LinkedLists of TagConstituents.
     */
    private final Map<String, SortedMap<Integer, LinkedList<TagConstituent>>>
        tagMap =
        new HashMap<String, SortedMap<Integer, LinkedList<TagConstituent>>>();

    /**
     * Creates a new instance.
     */
    public TagContainer() {
    }

    /**
     * Checks whether this instance contains at least one appearance of the
     * specified tag.
     *
     * @param tagname the name of the tag to check
     * @return <code>true</code> iff the container contains one or more
     * appearances of the specified tag
     */
    public boolean contains(final String tagname) {
        return tagMap.containsKey(tagname);
    }

    /**
     * Finds the first appearance of a tag.
     * If the specified tag did never appear within the requested markup series,
     * <code>null</code> is returned instead.
     *
     * @param tagname the name of the tag to check
     * @return the found apearance of the tag, or <code>null</code> if no
     * matching appearance was found
     */
    public TagConstituent findFirst(final String tagname) {
        // check entries for tag name
        final SortedMap nestedMap = tagMap.get(tagname);

        if ((nestedMap != null) && !nestedMap.isEmpty()) {
            final Iterator keyIter = nestedMap.keySet().iterator();
            Integer currentKey;
            LinkedList currentTags;

            while (keyIter.hasNext()) {
                currentKey = (Integer) keyIter.next();
                currentTags = (LinkedList) nestedMap.get(currentKey);

                if ((currentTags != null) && !currentTags.isEmpty()) {
                    // return first appearance
                    return (TagConstituent) currentTags.getFirst();
                }
            }
        }

        // no appearance found
        return null;
    }

    /**
     * Finds the appearance of a tag within a specified markup series.
     * If the specified tag didn't appear within the requested markup series,
     * <code>null</code> is returned instead.
     *
     * @param tagname the name of the tag to check
     * @param markupSeriesNo the markup series number to match
     * @param returnFirst whether to return the first or the last
     * appearance of the tag within the specified series, if several
     * appearances are found
     * @return the found apearance of the tag, or <code>null</code> if no
     * matching appearance was found
     */
    public TagConstituent findInSeries(final String tagname,
            final int markupSeriesNo, final boolean returnFirst) {
        // check entries for tag name
        final SortedMap nestedMap = tagMap.get(tagname);
        final TagConstituent result;

        if ((nestedMap != null) && !nestedMap.isEmpty()) {
            // check entries of specified markup series
            final Integer keyToUse = new Integer(markupSeriesNo);
            final LinkedList tagsInSeries =
                (LinkedList) nestedMap.get(keyToUse);

            if ((tagsInSeries != null) && !tagsInSeries.isEmpty()) {
                // return first resp. last appearance
                if (returnFirst) {
                    result = (TagConstituent) tagsInSeries.getFirst();
                } else {
                    result = (TagConstituent) tagsInSeries.getLast();
                }
            } else {
                result = null;
            }
        } else {
            result = null;
        }
        return result;
    }

    /**
     * Forces the removal of a single appearance of a tag from the
     * container. Throws a RuntimeException if the appearance could not be
     * removed, i.e. it didn't not exist in the container. Use this method
     * if you are sure that the specified appearance exists.
     *
     * @param tag the tag to remove
     * @throws RuntimeException stating an "Implementation error" if the
     * specified tag appearance could <em>not</em> be removed
     */
    public final void forceRemove(final TagConstituent tag)
            throws RuntimeException {
        if (!remove(tag)) {
            // remove failed
            throw new RuntimeException("Implementation error: forced removal "
                + "failed for tag " + tag);
        }
    }

    /**
     * Whether this tag container is empty.
     *
     * @return <code>true</code> iff this container is empty, i.e. doesn't
     * contain any tags
     */
    public boolean isEmpty() {
        return tagMap.isEmpty();
    }

    /**
     * Grants subclasses direct access to the map that is used internally for
     * managing the tags. Maps tag names Strings to nested TreeMaps mapping
     * markup series Integers to LinkedLists of TagConstituents.
     *
     * @return the tag map manages by this class
     */
    protected Map getTagMap() {
        return tagMap;
    }

    /**
     * Inserts an appearance of a tag into the container. The name of the
     * tag is used as key; the whole constituent is appended at the end of the
     * list of appearances stored for this tag name.
     *
     * @param tag the tag to insert
     */
    public final void push(final TagConstituent tag) {
        push(tag, true);
    }

    /**
     * Inserts an appearance of a tag into the container. The name of the
     * tag is used as key; the whole constituent is inserted in the list of
     * appearances stored for this tag name.
     *
     * @param tag the tag to insert
     * @param appendAtEnd whether to insert the tag at the end or at the begin
     * of appearances within a series
     * @throws UnsupportedOperationException might be thrown by child classes
     * that don't support prepending if <code>appendAtEnd</code> is
     * <code>false</code>
     */
    public void push(final TagConstituent tag, final boolean appendAtEnd)
            throws UnsupportedOperationException {
        final String tagname = tag.getName();
        final Integer seriesNo = new Integer(tag.getMarkupSeriesNo());

        // check entries for tag name
        SortedMap<Integer, LinkedList<TagConstituent>> nestedMap =
            tagMap.get(tagname);
        if (nestedMap == null) {
            // create and insert new map
            nestedMap = new TreeMap<Integer, LinkedList<TagConstituent>>();
            tagMap.put(tagname, nestedMap);
        }

        // check entries within series no.
        LinkedList<TagConstituent> tagsInSeries = nestedMap.get(seriesNo);
        if (tagsInSeries == null) {
            // create and insert new list
            tagsInSeries = new LinkedList<TagConstituent>();
            nestedMap.put(seriesNo, tagsInSeries);
        }

        // add tag to list
        if (appendAtEnd) {
            tagsInSeries.addLast(tag);
        } else {
            tagsInSeries.addFirst(tag);
        }
    }

    /**
     * Removes a single appearance of a tag from the container.
     *
     * @param tag the tag to remove
     * @return <code>true</code> if the specified TagConstituent was
     * removed successfully; <code>false</code> otherwise (the specified
     * constituent wasn't found in the container)
     */
    public boolean remove(final TagConstituent tag) {
        final String tagname = tag.getName();
        final Integer seriesNo = new Integer(tag.getMarkupSeriesNo());

        // check entries for tag name
        SortedMap<Integer, LinkedList<TagConstituent>> nestedMap =
            tagMap.get(tagname);
        if (nestedMap == null) {
            return false;
        }

        // check entries within series no.
        LinkedList<TagConstituent> tagsInSeries = nestedMap.get(seriesNo);
        if (tagsInSeries == null) {
            return false;
        }
        final boolean removedFromSeries = tagsInSeries.remove(tag);

        if (removedFromSeries && tagsInSeries.isEmpty()) {
            // inner list is now empty -- remove it
            nestedMap.remove(seriesNo);
            if (nestedMap.isEmpty()) {
                // nested map is now empty -- remove it too
                tagMap.remove(tagname);
            }
        }
        return removedFromSeries;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("tag map", tagMap)
            .toString();
    }

}