/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.text.TextUtils;

/**
 * "Isolates" XML tags and textual contents in an XML/HTML document by printing
 * each XML/HTML tag on a single line. This class is thread-safe.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/02/02 18:50:16 $, $Author: siefkes $
 */
public class TagIsolator {

    /**
     * The character class of line separators.
     */
    public static final String LINE_SEPARATOR_CLASS = "[\\r\\n\\p{Zl}]";

    /**
     * The class of whitespace characters that are <i>not</i> line separators.
     */
    public static final String SINGLE_LINE_WHITESPACE = "[ \\t\\x0B\\f\\p{Zs}]";

    /**
     * Matches one or more line-separator characters, including any surrounding
     * whitespace.
     */
    private static final Pattern LINE_SEP_PATTERN = Pattern.compile(
        SINGLE_LINE_WHITESPACE + "*"       // optional preceding whitespace
            + LINE_SEPARATOR_CLASS + "+"   // line-separators
            + SINGLE_LINE_WHITESPACE + "*" // optional following whitespace
        );

    /**
     * Matches the whole contents of an XML tag (anything between &lt; and
     * &gt;), optionally followed resp. preceded by a line separator.
     */
    private static final Pattern TAG_PATTERN = Pattern.compile(
        "(" + LINE_SEPARATOR_CLASS + "?)"     // optional preceding line-sep
        + SINGLE_LINE_WHITESPACE + "*"        // optional preceding whitespace
        + "(<.*?>)"                           // the actual tag
        + SINGLE_LINE_WHITESPACE + "*"        // optional following whitespace
        + "(" + LINE_SEPARATOR_CLASS + "?)", // optional following line-sep
        Pattern.DOTALL); // "." should also match newlines

    /**
     * Creates a new instance.
     */
    public TagIsolator() {
        super();
    }

    /**
     * "Isolates" XML tags and textual contents in an XML document by
     * printing each XML tag on a single line. This might introduce
     * additional whitespace, so it should be used with care in situations
     * where whitespace is significant. Neither reader nor writer are
     * closed by this method.
     *
     * @param reader the reader to read the original XML file from
     * @param writer the writer to write the modified XML file to
     * @throws IOException if an I/O error occurs while reading or writing
     * the data
     */
    public final void isolateTags(final Reader reader, final Writer writer)
            throws IOException {
        isolateTags(reader, writer, null);
    }


    /**
     * "Isolates" XML tags and textual contents in an XML document by
     * printing each XML tag on a single line. This might introduce
     * additional whitespace, so it should be used with care in situations
     * where whitespace is significant. Neither reader nor writer are
     * closed by this method.
     *
     * @param reader the reader to read the original XML file from
     * @param writer the writer to write the modified XML file to
     * @param replacements a map of additional replacements to perform within
     * the XML tags; maps regular expression {@link java.util.regex.Pattern}s
     * to replacement {@link java.lang.String}s; might be <code>null</code>
     * @throws IOException if an I/O error occurs while reading or writing
     * the data
     */
    public final void isolateTags(final Reader reader, final Writer writer,
            final Map replacements) throws IOException {
        final String contents = IOUtils.readToString(reader);
        final Matcher tagMatcher = TAG_PATTERN.matcher(contents);
        final Matcher lineSepMatcher = LINE_SEP_PATTERN.matcher("");

        int endOfLastMatch = 0;
        String precedingLineSep;
        String singleLineTagContents, finalTagContents;
        String followingLineSep;
        String textBetweenTags;
        boolean textBetweenExists;

        // iterate over XML tags
        while (tagMatcher.find()) {
            // write everything between last and current match
            textBetweenTags = contents.substring(endOfLastMatch,
                tagMatcher.start());
            textBetweenExists = textBetweenTags.length() > 0;
            if (textBetweenExists) {
                writer.write(textBetweenTags);
            }
            endOfLastMatch = tagMatcher.end();

            // first group: optional preceding line-separator
            precedingLineSep = tagMatcher.group(1);

            // insert preceding line-separator, if none exists and if there was
            // text after the last tag (no need to insert two line-separators)
            if (precedingLineSep.length() > 0) {
                writer.write(precedingLineSep);
            } else if (textBetweenExists) {
                writer.write(TextUtils.LINE_SEPARATOR);
            }

            // replace all line-seps in the actual tag by a single whitespace
            singleLineTagContents = TextUtils.replaceAll(tagMatcher.group(2),
                lineSepMatcher, " ");

            // do additional replacements, if specified
            if (replacements != null) {
                finalTagContents = TextUtils.multipleReplaceAll(
                    singleLineTagContents, replacements);
            } else {
                finalTagContents = singleLineTagContents;
            }
            writer.write(finalTagContents);

            // third group: optional following line-separator
            followingLineSep = tagMatcher.group(3);

            // insert following line-separator, if none exists
            if (followingLineSep.length() > 0) {
                writer.write(followingLineSep);
            } else {
                writer.write(TextUtils.LINE_SEPARATOR);
            }
        }

        // append tail and flush output
        writer.write(contents.substring(endOfLastMatch));
        writer.flush();
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this).toString();
    }
}
