/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml;

/**
 * Type-safe enumeration of the three variaties of tags employed for XML
 * adjustment.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/02/02 18:50:16 $, $Author: siefkes $
 */
public final class TagVariety {

    /**
     * Constant for the most common tag variety, marking all tags that
     * pre-existed in the input data.
     */
    public static final TagVariety REGULAR = new TagVariety("regular");

    /**
     * Constant for start tags that were created when splitting a tag. Tags of
     * this variety that be freely moved within a tag series, but the cannot
     * be moved to the next series (if required, they must be split again).
     */
    public static final TagVariety CONTINUATION =
        new TagVariety("continuation");

    /**
     * Constant for tentative start tags that were created when an end tag is
     * followed by another end tag of the same type without a start tag of this
     * type between them. The tentative tag replaces the missing start tag,
     * it can be freely moved between tag series when required.
     */
    public static final TagVariety TENTATIVE = new TagVariety("tentative");

    /**
     * The name of this tag variety, printed by the {@link #toString()} method.
     */
    private final String name;

    /**
     * Private constructor to prevent creation of further instances.
     * The static constants defined in this class are the only instantiations.
     *
     * @param varietyName the name of this tag variety, to be printed by the
     * {@link #toString()} method
     */
    private TagVariety(final String varietyName) {
        super();
        name = varietyName;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return name;
    }

}
