/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml;

import org.apache.commons.collections.Bag;
import org.apache.commons.collections.bag.HashBag;

/**
 * A container that keeps track of unprocessed start and end tags during XML
 * adjustment.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/05/03 09:21:38 $, $Author: siefkes $
 */
public class UnprocessedTags extends TagContainer {

    /**
     * A bag (multi-set) that counts how often each start tag appears in the
     * container.
     */
    private Bag startTagBag = new HashBag();

    /**
     * A bag (multi-set) that counts how often each end tag appears in the
     * container.
     */
    private Bag endTagBag = new HashBag();

    /**
     * Creates a new instance.
     */
    public UnprocessedTags() {
        super();
    }

    /**
     * Returns the number of end tags with the specified name contained in
     * this container.
     *
     * @param tagName the name of the tags to count
     * @return the number of matching end tags contained (0 or more)
     */
    public int endTagCount(final String tagName) {
        return endTagBag.getCount(tagName);
    }

    /**
     * Inserts an appearance of a tag into the container. The name of the
     * tag is used as key; the whole constituent is inserted in the list of
     * appearances stored for this tag name. Instances of this class accept only
     * start and end tags.
     *
     * @param tag the tag to insert
     * @param appendAtEnd whether to insert the tag at the end or at the begin
     * of appearances within a series
     * @throws IllegalArgumentException if the specified tag is not a start or
     * end tag
     */
    public void push(final TagConstituent tag, final boolean appendAtEnd)
            throws IllegalArgumentException {
        final String tagName = tag.getName();
        final int tagType = tag.getType();

        // increase count in appropriate bag
        switch (tagType) {
            case TagConstituent.START_TAG :
                startTagBag.add(tagName);
                break;

            case TagConstituent.END_TAG :
                endTagBag.add(tagName);
                break;

            default :
                throw new IllegalArgumentException("UnprocessedTags allows "
                    + " only start or end tags, not tags of type "
                    + tag.getType() + " (" + tag + ")");
        }

        // delegate to superclass
        super.push(tag, appendAtEnd);
    }

    /**
     * Removes a single appearance of a tag from the container.
     *
     * @param tag the tag to remove
     * @return <code>true</code> if the specified TagConstituent was
     * removed successfully; <code>false</code> otherwise (the specified
     * constituent wasn't found in the container)
     */
    public boolean remove(final TagConstituent tag) {
        if (super.remove(tag)) {
            // removed from superclass: remove from appropriate bag
            final String tagName = tag.getName();
            final int tagType = tag.getType();
            final boolean myResult;

            // increase count in appropriate bag
            switch (tagType) {
                case TagConstituent.START_TAG :
                    myResult = startTagBag.remove(tagName, 1);
                    break;

                case TagConstituent.END_TAG :
                    myResult = endTagBag.remove(tagName, 1);
                    break;

                default :
                    // normally this should be prevented by push + super.remove
                    // -- just a sanity check
                    throw new IllegalArgumentException("UnprocessedTags allows "
                        + " only start or end tags, not tags of type "
                        + tag.getType() + " (" + tag + ")");
            }

            return myResult;
        } else {
            // not found by superclass
            return false;
        }
    }

    /**
     * Returns the number of start tags with the specified name contained in
     * this container.
     *
     * @param tagName the name of the tags to count
     * @return the number of matching start tags contained (0 or more)
     */
    public int startTagCount(final String tagName) {
        return startTagBag.getCount(tagName);
    }
}
