/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ParsingException;
import de.fu_berlin.ties.TextProcessor;
import de.fu_berlin.ties.TiesConfiguration;

import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.text.TextTokenizer;
import de.fu_berlin.ties.text.TextUtils;
import de.fu_berlin.ties.util.CollectionUtils;
import de.fu_berlin.ties.util.Util;

/**
 * This class tries to fix corrupt XML documents, especially documents
 * containing nesting errors. Instances of this class are thread-safe and
 * can fix several documents in parallel.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.14 $, $Date: 2004/12/06 18:00:05 $, $Author: siefkes $
 */
public class XMLAdjuster extends TextProcessor {

    /**
     * Configuration key: the name to use for the root element if missing.
     */
    public static final String CONFIG_MISSING_ROOT = "adjust.missing-root";

    /**
     * Configuration key: Set of names of tags that can be converted empty tags
     * when required.
     */
    public static final String CONFIG_EMPTIABLE_TAGS = "adjust.emptiable-tags";

    /**
     * Configuration key: whether to delete
     * {@link #CONTROL_CHARS control characters} (which are not allowed in
     * XML 1.0 and discouraged in XML 1.1).
     */
    public static final String CONFIG_DELETE_CONTROL_CHARS =
        "adjust.delete.control-chars";

    /**
     * Configuration key: whether to delete "pseudo-tags".
     */
    public static final String CONFIG_DELETE_PSEUDO_TAGS =
        "adjust.delete.pseudo-tags";

    /**
     * Configuration key: whether to escape "&amp;" starting a possible
     * nonstandard entity reference ("&amp;" at the start of one of the 5
     * predefined entity references or a character reference is never escaped,
     * all other "&amp;" are always escaped).
     */
    public static final String CONFIG_ESCAPE_PSEUDO_ENTITIES =
        "adjust.escape.pseudo-entities";

    /**
     * Pattern string specifying characters that can occur at the start of end
     * of an unquoted attribute value: everything except '&lt;', '&gt;', '='
     * and whitespace (whitespace is also allowed, but only in the middle of a
     * value). Evaluated lazily (reluctant) to avoid missing the "/&gt;" at
     * the end of an empty tag.
     */
    public static final String UNQUOTED_ATTRIB_CHARS =
        "[^<>=" + XMLTokenizerFactory.XML_WHITESPACE_CHARS + "]+?";

    /**
     * Pattern string specifying an XML attribute without proper quotes.
     * Equal sign and value are captured in groups; the value is matched lazily
     * (reluctant) so we won't miss the start of the next attribute.
     */
    public static final String UNQUOTED_ATTRIBUTE = XMLTokenizerFactory.XML_NAME
        + XMLTokenizerFactory.XML_OPT_WHITESPACE + "(=)"
        + XMLTokenizerFactory.XML_OPT_WHITESPACE + "((?:"
        + UNQUOTED_ATTRIB_CHARS + XMLTokenizerFactory.XML_WHITESPACE + ")*?"
        + UNQUOTED_ATTRIB_CHARS + ")";

    /**
     * Pattern specifying of a "lax" XML start or empty tag that can contain
     * unquoted (invalid) attributes (combined into a single pattern to avoid
     * unnecessary backtracking). Element name, equal signs and values of
     * the (last) unquoted attribute, and '/' (for empty tags) are captured.
     */
    public static final Pattern LAX_START_OR_EMPTY_TAG = Pattern.compile(
        "<(" + XMLTokenizerFactory.XML_NAME + ")" + "(?:"
        + XMLTokenizerFactory.XML_WHITESPACE + "(?:"
        + XMLTokenizerFactory.XML_ATTRIBUTE + "|" + UNQUOTED_ATTRIBUTE + "))*"
        + XMLTokenizerFactory.XML_OPT_WHITESPACE + "(/)?>");

    /**
     * A "&amp;" that is not the start of an predefined entity reference or
     * a character reference and thus should be escaped if
     * {@link #isEscapingPseudoEntities()} is <code>true</code>.
     * (A pattern matching the rest of predefined entity or character reference
     * is included via negative lookahead.)
     */
    public static final Pattern PSEUDO_AMP = Pattern.compile(
        "&(?!(?:amp|lt|gt|apos|quot|#[0-9]+|#x[0-9a-fA-F]+);)");

    /**
     * A "&amp;" that is not the start of an entity and thus must be escaped.
     * (A pattern matching the rest of legal entity or character reference is
     * included via negative lookahead.)
     */
    public static final Pattern SPURIOUS_AMP = Pattern.compile("&(?!(?:"
        + XMLTokenizerFactory.XML_NAME + "|#[0-9]+|#x[0-9a-fA-F]+);)");

    /**
     * Escape sequence for the "&amp;" character.
     */
    public static final String ESCAPED_AMP = "&amp;";

    /**
     * Pattern specifying sequences of control characters (character codes
     * below the space character, except tab, line feed and carriage return).
     */
    public static final Pattern CONTROL_CHARS =
        Pattern.compile("[\u0001-\u0008\u000B-\u000C\u000E-\u001F]+");

    /**
     * Event constant: Converted to empty tag.
     */
    protected static final String EVENT_CONVERTED_TO_EMPTY_TAG =
        "Converted to empty tag";

    /**
     * Event constant: Inserted missing end tag.
     */
    protected static final String EVENT_INSERTED_MISSING_END_TAG  =
        "Inserted missing end tag";

    /**
     * Event constant: Inserted missing root element.
     */
    protected static final String EVENT_INSERTED_MISSING_ROOT_ELEMENT  =
        "Inserted missing root element";

    /**
     * Event constant: Inserted missing start tag.
     */
    protected static final String EVENT_INSERTED_MISSING_START_TAG  =
        "Inserted missing start tag";

    /**
     * Event constant: Moved end tag up.
     */
    protected static final String EVENT_MOVED_END_TAG_UP =
        "Moved end tag up";

    /**
     * Event constant: Moved start tag dow.
     */
    protected static final String EVENT_MOVED_START_TAG_DOWN  =
        "Moved start tag down";

    /**
     * Event constant: Split tag.
     */
    protected static final String EVENT_SPLIT_TAG  =
        "Split tag";

    /**
     * Event constant: Deleted control characters.
     */
    protected static final String EVENT_DELETED_CONTROL_CHARS  =
        "Deleted control characters";

    /**
     * Event constant: Deleted pseudo-tag.
     */
    protected static final String EVENT_DELETED_PSEUDO_TAG  =
        "Deleted pseudo-tag";

    /**
     * Event constant: Escaped characters that are illegal or unwanted.
     */
    protected static final String EVENT_ESCAPED_CHARS  =
        "Escaped characters";

    /**
     * Event constant: Quoted attribute values.
     */
    protected static final String EVENT_QUOTED_ATTRIBUTE_VALUES  =
        "Quoted attribute values";

    /**
     * The name to use for the root element if missing. A root element
     * with this name is created when not all elements and textual
     * content are inclosed within a single element (the root).
     * If <code>null</code>, processing stops with an exception if the root
     * element is missing.
     */
    private final String missingRootName;

    /**
     * Contains the names (Strings) of tags that can be converted an empty tags
     * when required for fixing a document (e.g. "br" when
     * <code>&lt;br&gt;</code> may be converted to <code>&lt;br/&gt;</code>
     * during repair). Might be <code>null</code>.
     */
    private final Set<String> emptiableTags;

    /**
     * Whether to delete {@link #CONTROL_CHARS control characters} (which are
     * not allowed in XML 1.0 and discouraged in XML 1.1).
     */
    private final boolean deletingControlChars;

    /**
     * Whether to delete "pseudo-tags", i.e., sequences that cannot be parsed as
     * tags but look similar to them. "Pseudo-tags" start with '&lt;' and end
     * with '&gt;', contain a printable character after the initial '&lt;', and
     * do not contain any inner '&lt;' or '&gt;'). If <code>true</code>, such
     * sequences will be deleted; otherwise (default) the starting '&lt;' will
     * be escaped.
     */
    private final boolean deletingPseudoTags;

    /**
     * Whether to escape "&amp;" starting a possible nonstandard entity
     * reference ("&amp;" at the start of one of the 5 predefined entity
     * references or a character reference is never escaped, all other "&amp;"
     * are always escaped).
     */
    private final boolean escapingPseudoEntities;

    /**
     * Creates a new instance using a default extension and the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     */
    public XMLAdjuster() {
        this("xml");
    }

    /**
     * Creates a new instance, configured from the
     * {@linkplain TiesConfiguration#CONF standard configuration}.
     * @param outExt the extension to use for output files
     */
    public XMLAdjuster(final String outExt) {
        this(outExt, TiesConfiguration.CONF);
    }

    /**
     * Creates a new instance from the provided configuration.
     * @param outExt the extension to use for output files
     * @param config used to configure this instance
     */
    public XMLAdjuster(final String outExt, final TiesConfiguration config) {
        this(outExt, config.getString(CONFIG_MISSING_ROOT, null),
            config.containsKey(CONFIG_EMPTIABLE_TAGS)
                    ? CollectionUtils.arrayAsSet(config.getStringArray(
                            CONFIG_EMPTIABLE_TAGS))
                    : null,
            config.getBoolean(CONFIG_DELETE_CONTROL_CHARS),
            config.getBoolean(CONFIG_DELETE_PSEUDO_TAGS),
            config.getBoolean(CONFIG_ESCAPE_PSEUDO_ENTITIES), config);
    }

    /**
     * Creates a new instance.
     *
     * @param outExt the extension to use for output files
     * @param missingRoot the name to use for the root element if missing, i.e.
     * if not all elements and textual content are inclosed within a single
     * element (the root); if <code>null</code>, processing stops with an
     * exception if the root element is missing
     * @param emptiableTagSet contains the names (Strings) of tags that can be
     * converted an empty tags when required for fixing a document (e.g. "br"
     * when <code>&lt;br&gt;</code> may be converted to
     * <code>&lt;br/&gt;</code> during repair); might be <code>null</code> if
     * there are none
     * @param deleteControlChars whether to delete
     * {@link #CONTROL_CHARS control characters} (which are not allowed in XML
     * 1.0 and discouraged in XML 1.1)
     * @param deletePseudoTags whether to
     * {@linkplain #isDeletingPseudoTags() delete "pseudo-tags"}
     * @param escapePseudoEntities whether to escape "&amp;" starting a possible
     * nonstandard entity reference ("&amp;" at the start of one of the 5
     * predefined entity references or a character reference is never escaped,
     * all other "&amp;" are always escaped)
     * @param config used to configure superclasses; if <code>null</code>,
     * the {@linkplain TiesConfiguration#CONF standard configuration} is used
     */
    public XMLAdjuster(final String outExt, final String missingRoot,
            final Set<String> emptiableTagSet, final boolean deleteControlChars,
            final boolean deletePseudoTags,
            final boolean escapePseudoEntities,
            final TiesConfiguration config) {
        super(outExt, config);
        missingRootName = missingRoot;
        emptiableTags = emptiableTagSet;
        deletingControlChars = deleteControlChars;
        deletingPseudoTags = deletePseudoTags;
        escapingPseudoEntities = escapePseudoEntities;
    }

    /**
     * Tries to fix corrupt XML documents, especially documents
     * containing nesting errors. Delegates to
     * {@link #fixedConstituents(CharSequence)} and writes the result to the
     * specified writer.
     *
     * @param input the corrupt XML document
     * @param out the writer to write the corrected XML document to; flushed but
     * not closed by this method
     * @throws IOException if an I/O error occurs while using the writer
     * @throws ParsingException if the XML input contains an uncorrectable error
     */
    public final void adjust(final CharSequence input, final Writer out)
            throws IOException, ParsingException {
        final XMLConstituent firstConst = fixedConstituents(input);

        // write corrected data to out
        XMLConstituent currentConst = firstConst;
        while (currentConst != null) {
            out.write(currentConst.getRepresentantion());
            currentConst = currentConst.nextConstituent();
        }
        out.flush();
    }

    /**
     * Tries to fix corrupt XML documents, especially documents containing
     * nesting errors. Delegates to {@link #adjust(CharSequence, Writer)}.
     *
     * @param in the reader to read the corrupt XML document from; not closed
     * by this method
     * @param out the writer to write the corrected XML document to; flushed but
     * not closed by this method
     * @throws IOException if an I/O error occurs while using the reader or
     * writer
     * @throws ParsingException if the XML input contains an uncorrectable error
     */
    public final void adjust(final Reader in, final Writer out)
            throws IOException, ParsingException {
        final String input = IOUtils.readToString(in);
        adjust(input, out);
    }

    /**
     * Method called by the {@link #logEvent(String, String)} methods whenever
     * an event occurred to ensure the event is acceptable. Subclasses that want
     * to prevent certain events can overwrite this method and throw an
     * exception if an "illegal" event is encountered.
     * This implementation does nothing, letting all events pass.
     *
     * @param eventType the event that occurred; should be one of the
     * EVENT constants defined in this class.
     * @throws ParsingException could be thrown by subclasses if the event is
     * considered illicit
     */
    protected void checkEvent(final String eventType) throws ParsingException {
    }

    /**
     * Helper method called after completing an end tag with a missing start tag
     * or processing a tentative start tag to check whether the next appearance
     * of this tag type is a start tag or another end tag.
     * The second case means that another start tag is missing
     * -- the missing start tag is created as a tentative tag and inserted.
     *
     * @param endTag the end tag whose next appearance should be checked
     * @param insertAfter the tag after which the newly created tentative
     * tag should be inserted
     * @param unprocessedTags the container of unprocessed start and end tags
     * @throws ParsingException might be thrown by {@link #checkEvent(String)}
     * implementations in subclasses if an "illicit" event occurred
     */
    private void checkNextAppearance(final TagConstituent endTag,
            final TagConstituent insertAfter,
            final UnprocessedTags unprocessedTags) throws ParsingException {
        // check whether the next appearance is a start tag (ok) or end tag
        // (insert tentative start tag)
        if (correspondingEndTag(endTag.getName(), -1, unprocessedTags,
                false) != null) {
            // this means next appearance is another end tag:
            // create tentative start tag after current end tag
            final TagConstituent tentative = new TagConstituent(
                TagConstituent.START_TAG, endTag.getName(),
                insertAfter.getMarkupSeriesNo());
            tentative.setVariety(TagVariety.TENTATIVE);

            // insert after specified tag + following whitespace, if any
            XMLConstituent nextConst = insertAfter.nextConstituent();
            if (nextConst.getType() == OtherConstituent.OUTER_WHITESPACE) {
                nextConst.insertAfter(tentative);
            } else {
                insertAfter.insertAfter(tentative);
            }

            // insert at begin of unprocessed tags and log event
            unprocessedTags.push(tentative, false);
            logEvent(EVENT_INSERTED_MISSING_START_TAG, tentative);
        }
    }

    /**
     * Helper method that creates an constituent that is part of a markup series
     * (i.e., that is neither text nor a CDATA section).
     *
     * @param currentToken the string representation of the constituent
     * @param capturedText the text captured by the XML text tokenizer, used to
     * determine the token type
     * @param markupSeriesNo the number of the markup series of the constituent
     * @param startAndEndTags all start and end tags are added to this
     * container, if it isn't <code>null</code>
     * @return the created constituent
     */
    private XMLConstituent createMarkupConstituent(final String currentToken,
            final String capturedText, final int markupSeriesNo,
            final UnprocessedTags startAndEndTags) {
        // used to determine the exact token type
        final char firstChar = capturedText.charAt(0);
        final char lastChar = capturedText.charAt(capturedText.length() - 1);
        final XMLConstituent result;

        if (firstChar == '/') {
            result = new TagConstituent(TagConstituent.END_TAG,
                capturedText.substring(1), currentToken, markupSeriesNo);
            if (startAndEndTags != null) {
                startAndEndTags.push((TagConstituent) result);
            }
        } else if (lastChar == '/') {
            result = new TagConstituent(TagConstituent.EMPTY_TAG,
                capturedText.substring(0, capturedText.length() - 1),
                currentToken, markupSeriesNo);
        } else if (firstChar == '!') {
            result = new OtherConstituent(OtherConstituent.DOCTYPE,
                currentToken);
        } else if (firstChar == '?') {
            // prolog ("<?xml...") or processing instruction?
            if (capturedText.substring(1).equals("xml")) {
                result = new OtherConstituent(OtherConstituent.XML_PROLOG,
                    currentToken);
            } else {
                result = new OtherConstituent(OtherConstituent.PI,
                    currentToken);
            }
        } else if (firstChar == '<') {
            result = new OtherConstituent(OtherConstituent.COMMENT,
                currentToken);
        } else {
            result = new TagConstituent(TagConstituent.START_TAG, capturedText,
                currentToken, markupSeriesNo);
            if (startAndEndTags != null) {
                startAndEndTags.push((TagConstituent) result);
            }
        }
        return result;
    }

    /**
     * Checks whether <code>unprocessedTags</code> contains a
     * matching tag within the specified <code>markupSeriesNo</code>, not
     * preceeded by a matching start tag. Returns the found end tag, if it
     * exists; returns <code>null</code> otherwise.
     *
     * @param tagname the name of the tag to check
     * @param markupSeriesNo the markup series number to match;
     * or <code>-1</code> if the markup series number doesn't matter
     * @param unprocessedTags the container of start and end tags to check
     * @param remove whether to remove the found end tag from the
     * <code>unprocessedTags</code> container
     * @return the found corresponding end tag,
     * or <code>null</code> if no matching appearance was found
     */
    private TagConstituent correspondingEndTag(final String tagname,
            final int markupSeriesNo, final UnprocessedTags unprocessedTags,
            final boolean remove) {
        final TagConstituent result;
        final TagConstituent suitableTag;

        if (markupSeriesNo >= 0) {
            suitableTag =
                unprocessedTags.findInSeries(tagname, markupSeriesNo, true);
        } else {
            suitableTag = unprocessedTags.findFirst(tagname);
        }

        // check whether we found a start tag or end tag (if any)
        if ((suitableTag != null)
            && (suitableTag.getType() == TagConstituent.END_TAG)) {
                // found suitable end tag
                result = suitableTag;
                if (remove) {
                    unprocessedTags.forceRemove(suitableTag);
                }
        } else {
            // no tag or start tag
            result = null;
        }

        return result;
    }

    /**
     * Checks whether <code>openTags</code> contains a matching start tag
     * (ignoring the root tag), either within the specified
     * <code>markupSeriesNo</code> or a tentative tag anywhere.
     * Returns and remove the found end tag, if it exists.
     * Returns <code>null</code> otherwise.
     *
     * @param tagname the name of the tag to check
     * @param markupSeriesNo the markup series number to match
     * @param openTags the container of currently open start tags to check
     * @return the found and removed matching start tag,
     * or <code>null</code> if no matching non-root appearance was found
     */
    private TagConstituent correspondingOpenTag(final String tagname,
            final int markupSeriesNo, final OpenTags openTags) {
        // use last match within specified markup series, if any
        TagConstituent result =
            openTags.findInSeries(tagname, markupSeriesNo, false);

        // otherwise look for tentative tag
        if (result == null) {
            result = openTags.findTentativeTag(tagname);
        }

        if (result != null) {
            if (openTags.isRoot(result)) {
                // the root tag is not accepted -- return null instead
                result = null;
            } else {
                // remove found tag
                openTags.forceRemove(result);
            }
        }
        return result;
    }

    /**
     * Tries to fix corrupt XML documents, especially documents containing
     * nesting errors. Delegates to {@link #adjust(Reader, Writer)}, ignoring
     * the <code>context</code>.
     *
     * @param reader reader containing the text to process; should not be closed
     * by this method
     * @param writer the writer to write the processed text to; might be flushed
     * but not closed by this method
     * @param context a map of objects that are made available for processing
     * @throws IOException if an I/O error occurs while using the reader or
     * writer
     * @throws ParsingException if the XML input contains an uncorrectable error
     */
    protected void doProcess(final Reader reader, final Writer writer,
            final ContextMap context) throws IOException, ParsingException {
        adjust(reader, writer);
    }

    /**
     * Helper method that checks whether an end tag is certain to be missing for
     * a specified open tag. This is the case if (a) the next appearance of
     * this tag type is NOT an end tag and if (b) there are at least as much
     * unprocessed start tags than unprocessed end tags for the given type.
     *
     * @param openTagName the name of the open tag check
     * @param unprocessedTags the container of unprocessed start and end tags
     * @return <code>true</code> iff both conditions stated above are fulfulled
     */
    private boolean endTagMissing(final String openTagName,
            final UnprocessedTags unprocessedTags) {
        // true if the next tag is not an tag
        final boolean result = (correspondingEndTag(openTagName, -1,
                unprocessedTags, false) == null)
        // and if there are at least as many start tags as end tags
            && (unprocessedTags.startTagCount(openTagName)
                >= unprocessedTags.endTagCount(openTagName));
        return result;
    }

    /**
     * Escapes every of a specified pattern in the representation of a
     * constituent.
     *
     * @param constituent the constituent to check
     * @param pattern the pattern to replace
     * @param replacement the replacement text
     * @param eventType the event to log when replacing was necessary
     * @throws ParsingException might be thrown by {@link #checkEvent(String)}
     * implementations in subclasses if an "illicit" event occurred
     */
    private void replaceInRep(final XMLConstituent constituent,
            final Pattern pattern, final String replacement,
            final String eventType) throws ParsingException {
        final String oldRep = constituent.getRepresentantion();
        final String newRep =
            TextUtils.replaceAll(oldRep, pattern, replacement);

        if (oldRep != newRep) {
            // representation changed: log event and update representation
            logEvent(eventType, "old: '" + oldRep + "', new: '" + newRep + "'");
            constituent.setRepresentantion(newRep);
        }
    }

    /**
     * Returns the constituents of an XML-like document after fixing possible
     * nesting errors etc.
     *
     * @param input the XML-like input data
     * @return a reference to the first contained constituent; the list
     * of constituents can be traversed by calling
     * {@link XMLConstituent#nextConstituent()} on each constituent untill
     * <code>null</code> is returned
     * @throws ParsingException if the XML input contains an uncorrectable error
     */
    public final XMLConstituent fixedConstituents(final CharSequence input)
            throws ParsingException {
        // first run: create list of all tags + container of start and end tags;
        // fix character-level errors
        final UnprocessedTags unprocessedTags = new UnprocessedTags();
        XMLConstituent firstConst =
            rawConstituents(input, true, unprocessedTags);

        // second run: ensure corrent nesting of start + end tags
        XMLConstituent currentConst = firstConst;
        XMLConstituent lastConst = currentConst;
        TagConstituent currentTag;
        final OpenTags openTags = new OpenTags();

        // all root content (tags, non-whitespace text, CDATA sections) must
        // be enclosed in a root element
        XMLConstituent firstRootContent = null;
        boolean insertedMissingRoot = false;

        while (currentConst != null) {
            if ((firstRootContent == null) && isRootContent(currentConst)) {
                // this is the first root content
                firstRootContent = currentConst;

                if (currentConst.getType() != TagConstituent.START_TAG) {
                    // not a start tag: try to insert missing root element
                    insertMissingRoot(firstRootContent, openTags);
                    insertedMissingRoot = true;
                }
            } else if (openTags.isEmpty() && (isRootContent(currentConst))) {
                // all tags have been closed but this is root content:
                // try to insert missing root element
                insertMissingRoot(firstRootContent, openTags);
                insertedMissingRoot = true;
            }

            if (currentConst instanceof TagConstituent) {
                currentTag = (TagConstituent) currentConst;
                if (currentTag.getType() == TagConstituent.START_TAG) {
                    // move from unprocessed tags to open tags
                    unprocessedTags.forceRemove(currentTag);
                    openTags.push(currentTag);
                } else if (currentTag.getType() == TagConstituent.END_TAG) {
                    // remove from unprocessed tags and do handling
                    unprocessedTags.forceRemove(currentTag);
                    handleEndTag(currentTag, openTags, unprocessedTags);
                }
            }

            lastConst = currentConst; // so we don't loose the very last one
            currentConst = currentConst.nextConstituent();
        }

        // handle end-of-file
        handleEOF(lastConst, openTags, unprocessedTags, insertedMissingRoot);

        // first constituent might have changed (e.g. due to moves in first tag
        // series, insertion of root element)
        while (firstConst.hasPrevious()) {
            firstConst = firstConst.previousConstituent();
        }
        return firstConst;
    }

    /**
     * Helper method that fixes character errors in the representation of a
     * constituent, if any.
     *
     * @param constituent the constituent to check
     * @throws ParsingException might be thrown by {@link #checkEvent(String)}
     * implementations in subclasses if an "illicit" event occurred
     */
    private void fixRepresentation(final XMLConstituent constituent)
            throws ParsingException {
        // delete control characters if configured
        if (deletingControlChars) {
            replaceInRep(constituent, CONTROL_CHARS, "",
                EVENT_DELETED_CONTROL_CHARS);
        }

        if (needsAmpEscape(constituent)) {
            if (escapingPseudoEntities) {
                // escape every '&' unless starting predefined entity or
                // character reference
                replaceInRep(constituent, PSEUDO_AMP, ESCAPED_AMP,
                    EVENT_ESCAPED_CHARS);
            } else {
                // escape only illegal '&' (not start of any entity)
                replaceInRep(constituent, SPURIOUS_AMP, ESCAPED_AMP,
                    EVENT_ESCAPED_CHARS);
            }
        }
    }

    /**
     * Helper method for handling an end tag. Here the actual work of the
     * algorithm takes place, because we have to ensure a match.
     *
     * @param endTag the end tag to handle
     * @param openTags must contain all currently open tags
     * @param unprocessedTags must contain all unprocessed start and end tags
     * @throws ParsingException might be thrown by {@link #checkEvent(String)}
     * implementations in subclasses if an "illicit" event occurred
     */
    protected void handleEndTag(final TagConstituent endTag,
            final OpenTags openTags, final UnprocessedTags unprocessedTags)
            throws ParsingException {
        TagConstituent lastOpenTag;
        TagVariety lastOpenTagVariety;
        String lastOpenTagName;
        TagConstituent correspondingTag;
        boolean matchFound = false;

        // iterate until end tag has been processed by finding a match
        while (!matchFound) {
            lastOpenTag = openTags.peek();
            lastOpenTagVariety = lastOpenTag.getVariety();
            lastOpenTagName = lastOpenTag.getName();
            //Log.TIES.debug("Mismatch: " + lastOpenTag + " / " + endTag);

            if (lastOpenTagName.equals(endTag.getName())) {
                // last open tag matched: remove it and ensure it's regular
                // (checking next appearance for tentative tags)
                if (openTags.popAndRegularize()
                        && (lastOpenTagVariety == TagVariety.TENTATIVE)) {
                    checkNextAppearance(endTag, endTag, unprocessedTags);
                }
                matchFound = true; // done
            } else if (lastOpenTagVariety == TagVariety.TENTATIVE) {
                // last open tag is tentative: insert after the end tag; removed
                // it from open tags and re-added to unprocessed tags
                moveTentativeTag(lastOpenTag, endTag, openTags,
                    unprocessedTags);
                // now retry
            } else if (openTags.containsNonTentative(endTag.getName())
                    && (correspondingTag = correspondingEndTag(
                    lastOpenTagName, endTag.getMarkupSeriesNo(),
                    unprocessedTags, true)) != null) {
                // A non-tentative start tag for the current end tag exists and
                // there is a corresponding end tag for the last open tag in the
                // current markup series. Remove the last open tag and make it
                // regular if necessary
                openTags.popAndRegularize();

                // move the corresponding end tag before the current end tag
                correspondingTag.remove();
                endTag.insertBefore(correspondingTag);

                // check next appearance if tentative (if necessary inserting
                // new tentative tag after the current end tag)
                if (lastOpenTagVariety == TagVariety.TENTATIVE) {
                    checkNextAppearance(correspondingTag, endTag,
                        unprocessedTags);
                }

                // log event and retry
                logEvent(EVENT_MOVED_END_TAG_UP, correspondingTag);
            } else if ((correspondingTag = correspondingOpenTag(
                    endTag.getName(), lastOpenTag.getMarkupSeriesNo(),
                    openTags)) != null) {
                // Open tags contained a non-root tag corresponding to the end
                // tag, either within the markup series of the last open tag or
                // a tentative tag anywhere.
                // Move the corresponding start after the last open tag
                correspondingTag.remove();
                lastOpenTag.insertAfter(correspondingTag);

                if (correspondingTag.getVariety() == TagVariety.REGULAR) {
                    // log event
                    logEvent(EVENT_MOVED_START_TAG_DOWN, correspondingTag);
                } else {
                    if (correspondingTag.getVariety() == TagVariety.TENTATIVE) {
                        // tentative tag: check whether the next appearance is
                        // a start tag (ok) or end tag (another start tag
                        // missing: insert tentative start tag)
                        checkNextAppearance(endTag, endTag, unprocessedTags);
                    }

                    // convert to regular tag
                    correspondingTag.setVariety(TagVariety.REGULAR);
                    // event was already logged when creating the irregular tag
                }
                matchFound = true; // done
            } else if (!openTags.contains(endTag.getName())) {
                // no open tag matching the current end tag -- we have to
                // create one
                final TagConstituent missingStartTag =
                    new TagConstituent(TagConstituent.START_TAG,
                        endTag.getName(), lastOpenTag.getMarkupSeriesNo());

                // insert new tag after the last open tag
                lastOpenTag.insertAfter(missingStartTag);
                logEvent(EVENT_INSERTED_MISSING_START_TAG, missingStartTag);

                // Check whether the next appearance is a start tag (ok) or end
                // tag (another start tag missing: insert tentative start tag)
                checkNextAppearance(endTag, endTag, unprocessedTags);
                matchFound = true; // done
            } else if ((lastOpenTag.getMarkupSeriesNo()
                    == endTag.getMarkupSeriesNo())
                    && !endTagMissing(lastOpenTagName, unprocessedTags)) {
                // the last open tag is within the current markup series and a
                // corresponding end tag exists for it (in any markup series):
                // Move it after the current end tag, removing it from open tags
                // and from current place in list
                openTags.forceRemove(lastOpenTag);
                lastOpenTag.remove();
                endTag.insertAfter(lastOpenTag);

                // append at the start of unprocessed tags of this series
                unprocessedTags.push(lastOpenTag, false);

                // log event if this is a regular tag
                if (lastOpenTagVariety == TagVariety.REGULAR) {
                    logEvent(EVENT_MOVED_START_TAG_DOWN, lastOpenTag);
                }
                // now retry
            } else {
                // no way out: we have to "get rid of" the last open tag
                final boolean endTagMissing =
                    endTagMissing(lastOpenTagName, unprocessedTags);

                if (endTagMissing && isAnEmptiableTag(lastOpenTagName)) {
                    // convert last open tag to empty tag
                    final TagConstituent newEmptyTag =
                        replaceWithEmptyCopy(lastOpenTag);
                    logEvent(EVENT_CONVERTED_TO_EMPTY_TAG, newEmptyTag);
                } else {
                    // create and insert suitable end tag and insert it before
                    // the current end tag
                    final TagConstituent matchingEndTag = new TagConstituent(
                        TagConstituent.END_TAG, lastOpenTagName,
                        endTag.getMarkupSeriesNo());
                    endTag.insertBefore(matchingEndTag);

                    if (endTagMissing) {
                        // end tag of lastOpenTag type is missing: created it.
                        // Log event for regular tags
                        if (lastOpenTagVariety == TagVariety.REGULAR) {
                            logEvent(EVENT_INSERTED_MISSING_END_TAG,
                                matchingEndTag);
                        }
                    } else {
                        // sufficient end tags available:
                        // split lastOpenTag by creating continuation copy
                        final TagConstituent continuation = new TagConstituent(
                            TagConstituent.START_TAG, lastOpenTagName,
                            lastOpenTag.getRepresentantion(),
                            endTag.getMarkupSeriesNo());
                        continuation.setVariety(TagVariety.CONTINUATION);

                        // insert copy after current end tag and log event
                        endTag.insertAfter(continuation);
                        unprocessedTags.push(continuation, false);
                        logEvent(EVENT_SPLIT_TAG, lastOpenTag);
                    }

                    // Remove from open tags and ensure it's regular; retry
                    openTags.popAndRegularize();
                }
            }
        }
    }

    /**
     * Helper method for handling an the end of a file. Suitable end tags are
     * created to close any left-over open tags.
     *
     * @param lastConst the last constituent in the original input (must not
     * have a successor)
     * @param openTags must contain all currently open tags
     * @param unprocessedTags must contain all unprocessed start and end tags;
     * should better be empty
     * @param insertedMissingRoot whether the start tag of a missing root
     * element was created (in this case we'll insert the corresponding end tag
     * without logging another event)
     * @throws ParsingException might be thrown by {@link #checkEvent(String)}
     * implementations in subclasses if an "illicit" event occurred
     */
    protected void handleEOF(final XMLConstituent lastConst,
            final OpenTags openTags, final UnprocessedTags unprocessedTags,
            final boolean insertedMissingRoot) throws ParsingException {
        // check state
        if (lastConst.nextConstituent() != null) {
            Util.LOG.error("Implementation error: constituents after last "
                + "constituent: " + lastConst);
        }
        if (!unprocessedTags.isEmpty()) {
            Util.LOG.error("Implementation error: still unprocessed tags at "
                + "end-of-file -- last constituent: " + lastConst
                + " unprocessed tags: " + unprocessedTags);
        }

        XMLConstituent currentConst = lastConst;
        TagConstituent lastOpenTag = openTags.peek();

        if (lastOpenTag != null) {
            // there are unclosed elements -- look for last root content
            while (!isRootContent(currentConst)) {
                currentConst = currentConst.previousConstituent();
            }

            // insert missing end tags after last root content
            TagConstituent missingEndTag;
            while (lastOpenTag != null) {
                // create suitable end tag for each unclosed open tag
                missingEndTag = new TagConstituent(TagConstituent.END_TAG,
                    lastOpenTag.getName());

                // insert after last root content
                currentConst.insertAfter(missingEndTag);
                currentConst = missingEndTag;
                // remove handled tag and peek the previous one
                openTags.pop();
                lastOpenTag = openTags.peek();

                // log an event unless this is the end of a missing root
                if (!(insertedMissingRoot && openTags.isEmpty())) {
                    logEvent(EVENT_INSERTED_MISSING_END_TAG, missingEndTag);
                }
            }
        }
    }

    /**
     * Logs the occurance of an event necessary for fixing a document.
     * This method variant is used for character errors.
     *
     * @param eventType the event that occurred; should be one of the
     * EVENT constants defined in this class.
     * @param details a detailed description of the event
     * @throws ParsingException might be thrown by {@link #checkEvent(String)}
     * implementations in subclasses if the event is considered illicit
     */
    protected void logEvent(final String eventType, final String details)
            throws ParsingException {
        checkEvent(eventType);
        Util.LOG.debug("Modified document: " + eventType
                + " (" + details.toString() + ')');
    }

    /**
     * Logs the occurance of an event necessary for fixing a document.
     * This method variant is used for nesting errors and missing root elements.
     *
     * @param eventType the event that occurred; should be one of the
     * EVENT constants defined in this class.
     * @param tag the involved tag
     * @throws ParsingException might be thrown by {@link #checkEvent(String)}
     * implementations in subclasses if the event is considered illicit
     */
    protected void logEvent(final String eventType, final TagConstituent tag)
            throws ParsingException {
        checkEvent(eventType);
        Util.LOG.debug("Modified document: " + eventType
                + " (" + tag.toString() + ')');
    }

    /**
     * Inserts a root element with the configured name
     * ({@link #missingRootName}); or throws a parsing exception to finish
     * processing if inserting a root element is not allowed. This method
     * only creates and inserts a start tag of the configured type; the
     * matching end tag will automatically be added by
     * {@link #handleEOF(XMLConstituent, OpenTags, UnprocessedTags, boolean)}.
     *
     * @param firstRootContent the root start will be inserted before this
     * content
     * @param openTags the container of open tags to push the root tag to,
     * assumed to be empty
     * @throws ParsingException if inserting a missing root tag is now allowed
     * (no name for a missing root tag specified)
     */
    private void insertMissingRoot(final XMLConstituent firstRootContent,
            final OpenTags openTags) throws ParsingException {
        if (missingRootName != null) {
            // create start tag
            final TagConstituent rootStartTag = new TagConstituent(
                    TagConstituent.START_TAG, missingRootName, 0);

            // insert before first root content
            firstRootContent.insertBefore(rootStartTag);

            // push to open tags and log event
            openTags.push(rootStartTag);
            logEvent(EVENT_INSERTED_MISSING_ROOT_ELEMENT, rootStartTag);
        } else {
            throw new ParsingException("Root tag missing (resp. tags or "
                    + "textual content outside root element)");
        }
    }

    /**
     * Whether the specified tag is one of the tags that can be converted an
     * empty tags when required for fixing a document. For example, "br" when
     * <code>&lt;br&gt;</code> may be converted to <code>&lt;br/&gt;</code>
     * during repair.
     *
     * @param tag the name of the tag to look up
     * @return <code>true</code> iff this tag is contained in the set of
     * emptiable tags
     */
    protected boolean isAnEmptiableTag(final String tag) {
        return (emptiableTags != null) && emptiableTags.contains(tag);
    }

    /**
     * Whether {@link #CONTROL_CHARS control characters} are deleted (these
     * characters are not allowed in XML 1.0 and discouraged in XML 1.1).
     * @return <code>true</code> iff control characters are deleted
     */
    public boolean isDeletingControlChars() {
        return deletingControlChars;
    }

    /**
     * Whether "pseudo-tags" are deleted, i.e., sequences that cannot be parsed
     * as tags but look similar to them. "Pseudo-tags" start with '&lt;' and end
     * with '&gt;', contain a printable character after the initial '&lt;', and
     * do not contain any inner '&lt;' or '&gt;'). Disabled by default.
     *
     * @return <code>true</code> iff pseudo-tags are be deleted (otherwise the
     * starting '&lt;' is escaped)
     */
    public boolean isDeletingPseudoTags() {
        return deletingPseudoTags;
    }

    /**
     * Whether to escape "&amp;" starting a possible nonstandard entity
     * reference ("&amp;" at the start of one of the 5 predefined entity
     * references or a character reference is never escaped, all other "&amp;"
     * are always escaped).
     *
     * @return <code>true</code> iff "pseudo entites" are escaped
     */
    public boolean isEscapingPseudoEntities() {
        return escapingPseudoEntities;
    }

    /**
     * Checks whether a constituent is content that must be enclosed in the
     * root element. Tags, textual content, and CDATA sections are root
     * content; other constituent types can occur outside of the root element.
     *
     * @param constituent the constituent to check
     * @return <code>true</code> iff the constituent is root content
     */
    private boolean isRootContent(final XMLConstituent constituent) {
        final boolean result;
        if (constituent instanceof TagConstituent) {
            // all tags are root content
            result = true;
        } else {
            if ((constituent.getType() == OtherConstituent.CDATA_SECTION)
                    || (constituent.getType() == OtherConstituent.TEXT)) {
                // text and CDATA sections are root content
                result = true;
            } else {
                result = false;
            }
        }
        return result;
    }

    /**
     * Helper method that moves a tentative tag after a specified end tag.
     * The tentative tag must originally be contained in the container of open
     * tags; it is popped from this container and re-added to the container of
     * unprocessed tags.
     *
     * @param tentativeTag the tag to move, should be a tentative tag
     * @param endTag the tentative tag is moved after this tag
     * @param openTags must contain all currently open tags, including the tag
     * to move
     * @param unprocessedTags must contain all unprocessed start and end tags
     */
    private void moveTentativeTag(final TagConstituent tentativeTag,
            final TagConstituent endTag, final OpenTags openTags,
            final UnprocessedTags unprocessedTags) {
        // remove from open tags and from current place in list
        openTags.forceRemove(tentativeTag);
        tentativeTag.remove();

        // adjust series no. and re-add after current tag
        tentativeTag.setMarkupSeriesNo(endTag.getMarkupSeriesNo());
        endTag.insertAfter(tentativeTag);

        // prepend at the begin of unprocessed tags of this series
        unprocessedTags.push(tentativeTag, false);
    }

    /**
     * Checks whether escapes for the "&amp;" character are required in the
     * representation of a constituent.
     *
     * @param constituent the constituent to check
     * @return <code>true</code> if the {@linkplain XMLConstituent#getType()
     * type} of the constituent is {@link TagConstituent#START_TAG},
     * {@link TagConstituent#EMPTY_TAG}, or {@link OtherConstituent#TEXT};
     * <code>false</code> for all other types (which either cannot contain any
     * "&amp;" or do not need to escape "&amp;")
     */
    private boolean needsAmpEscape(final XMLConstituent constituent) {
        final short constType = constituent.getType();
        return (constType == TagConstituent.START_TAG)
            || (constType == TagConstituent.EMPTY_TAG)
            || (constType == OtherConstituent.TEXT);
    }

    /**
     * Returns the raw constituents of an XML-like document. The constituents
     * are returned "raw" as they occur in the input, without fixing possible
     * nesting errors etc.
     *
     * @param input the XML-like input data
     * @param fixCharacterErrors whether to try to fix character errors, i.e.
     * unescaped "&lt;" and "&amp;" and tags with unquoted attribute values; if
     * <code>false</code>, unescaped "&lt;" in textual content and unquoted
     * attribute values will yield an exception, while any unescaped "&amp"
     * and unescaped "&lt;" in attribute values will be ignored
     * @return a reference to the first contained constituent; the list
     * of constituents can be traversed by calling
     * {@link XMLConstituent#nextConstituent()} on each constituent untill
     * <code>null</code> is returned
     * @throws ParsingException if the XML input contains an uncorrectable error
     */
    public final XMLConstituent rawConstituents(final CharSequence input,
            final boolean fixCharacterErrors) throws ParsingException {
        return rawConstituents(input, fixCharacterErrors, null);
    }

    /**
     * Returns the raw constituents of an XML-like document. The constituents
     * are returned "raw" as they occur in the input, without fixing possible
     * nesting errors etc.
     *
     * @param input the XML-like input data
     * @param fixCharacterErrors whether to try to fix character errors, i.e.
     * unescaped "&lt;" and "&amp;" and tags with unquoted attribute values; if
     * <code>false</code>, unescaped "&lt;" in textual content and unquoted
     * attribute values will yield an exception, while any unescaped "&amp"
     * and unescaped "&lt;" in attribute values will be ignored
     * @param startAndEndTags all start and end tags are added to this
     * container, if it isn't <code>null</code>
     * @return a reference to the first contained constituent; the list
     * of constituents can be traversed by calling
     * {@link XMLConstituent#nextConstituent()} on each constituent untill
     * <code>null</code> is returned
     * @throws ParsingException if the XML input contains an uncorrectable error
     */
    protected final XMLConstituent rawConstituents(final CharSequence input,
            final boolean fixCharacterErrors,
            final UnprocessedTags startAndEndTags) throws ParsingException {
        // tokenizer for XML-like input -- if fixCharacterErrors we'll ensure
        // whitespace validity outselves, otherwise let the tokenizer validate
        final TextTokenizer tokenizer =
            XMLTokenizerFactory.createXMLTokenizer(input, !fixCharacterErrors);
        String tokenText;
        String precWS;
        String captured = null;
        boolean done = false;
        XMLConstituent firstConst = null;
        XMLConstituent priorConst = null;
        XMLConstituent currentConst;
        // whether we're currently in a markup series (any kind of tags and
        // declarations) or outside (textual content or CDATA sections)
        boolean inMarkupSeries = true;
        // markup series are sequentially numbered, starting at 0
        int markupSeriesNo = 0;

        // variables for checking and fixing character errors
        boolean fixedCharError;
        XMLConstituent extraConst = null;
        String extraWS = null;

        try {
            while (!done) {
                tokenText = tokenizer.nextToken();
                if (tokenText == null) {
                    done = true;    // reached end of input
                } else {
                    // create suitable type
                    captured = tokenizer.capturedText();
                }

                if (tokenizer.hasPrecedingWhitespace()) {
                    precWS = tokenizer.precedingWhitespace();
                    if (fixCharacterErrors
                            && (!tokenizer.precedingWhitespaceIsValid())) {
                        // "whitespace" is invalid: try to fix it
                        fixedCharError = false;
                        if (precWS.endsWith("<") && (captured.length() == 0)) {
                            // trailing '<' and text token is TEXT:
                            final String remainingWS = precWS.substring(0,
                                precWS.length() - 1);
                            final int gtPos = tokenText.indexOf('>');

                            if (tokenizer.isValidWhitespace(remainingWS)
                                    && gtPos > 0) {
                                // trailing '<' is only problem and
                                // text contains '>' after at least one char
                                final Object[] fixed
                                    = tryToFixTag('<' + tokenText);
                                if (fixed != null) {
                                    // store tag + set token to unused rest
                                    extraConst = (TagConstituent) fixed[0];
                                    tokenText = (String) fixed[1];
                                    fixedCharError = true;
                                    precWS = remainingWS;
                                } else if (deletingPseudoTags) {
                                    // delete "pseudo-tags" + set token to rest
                                    logEvent(EVENT_DELETED_PSEUDO_TAG, '<'
                                        + tokenText.substring(0, gtPos + 1));
                                    tokenText = tokenText.substring(gtPos + 1);
                                    fixedCharError = true;
                                    precWS = remainingWS;
                                }
                            }
                        }

                        if (fixedCharError) {
                            // fixed error: check if next token (TEXT)
                            // now starts with whitespace
                            final int initWS =
                                tokenizer.initialWhitespaceCount(tokenText);
                            if (initWS > 0) {
                                if (extraConst != null) {
                                    // store after extra constituent
                                    extraWS = tokenText.substring(0, initWS);
                                } else {
                                    // append to current whitespace
                                    precWS += tokenText.substring(0, initWS);
                                }
                                tokenText = tokenText.substring(initWS);
                            }
                        } else {
                            // escape illegal characters
                            String escape = StringEscapeUtils.escapeXml(precWS);
                            final int initWS =
                                tokenizer.initialWhitespaceCount(escape);

                            // separate initial whitespace + log event
                            if (initWS > 0) {
                                precWS = escape.substring(0, initWS);
                                escape = escape.substring(initWS);
                            } else {
                                precWS = "";
                            }
                            logEvent(EVENT_ESCAPED_CHARS, escape);

                            if (captured.length() == 0) {
                                // next token is TEXT: just prepend
                                tokenText = escape + tokenText;
                            } else {
                                final int trailingWSChars =
                                    tokenizer.trailingWhitespaceCount(escape);
                                if (trailingWSChars > 0) {
                                    // store trailing whitespace separately
                                    extraWS = escape.substring(
                                        escape.length() - trailingWSChars);
                                    escape = escape.substring(0,
                                        escape.length() - trailingWSChars);
                                }
                                // create text constituent
                                extraConst = new OtherConstituent(
                                    OtherConstituent.TEXT, escape);
                            }
                        }
                    }

                    // create whitespace constituent (if whitespace remains
                    // after fixing possible character errors)
                    if (precWS.length() > 0) {
                        currentConst = new OtherConstituent(
                            OtherConstituent.OUTER_WHITESPACE, precWS);
                        // add to list
                        if (priorConst == null) {
                            firstConst = currentConst;
                        } else {
                            priorConst.insertAfter(currentConst);
                        }
                        priorConst = currentConst;
                    }

                    if (extraConst != null) {
                        // handle extra constituent from fixed character error
                        if ((extraConst.getType() == OtherConstituent.TEXT)
                                || (extraConst.getType()
                                    == OtherConstituent.CDATA_SECTION)) {
                            inMarkupSeries = false;   // non-markup
                        } else {    // markup
                            if (!inMarkupSeries) {    // start new series
                                inMarkupSeries = true;
                                markupSeriesNo++;
                            }
                            if (extraConst instanceof TagConstituent) {
                                // store markup series in tag constituents
                                ((TagConstituent) extraConst)
                                    .setMarkupSeriesNo(markupSeriesNo);

                                // add to unprocessed tags if start or end tag
                                if ((startAndEndTags != null)
                                        && (extraConst.getType()
                                                != TagConstituent.EMPTY_TAG)) {
                                    startAndEndTags.push(
                                            (TagConstituent) extraConst);
                                }
                            }
                        }
                        if (fixCharacterErrors) {
                            // fix illegal characters in representation
                            fixRepresentation(extraConst);
                        }
                        // add extra constituent to list
                        if (priorConst == null) {
                            firstConst = extraConst;
                        } else {
                            priorConst.insertAfter(extraConst);
                        }
                        priorConst = extraConst;
                        extraConst = null;  // reset variable

                        if (extraWS != null) {
                            // add extra whitespace to list
                            currentConst = new OtherConstituent(
                                OtherConstituent.OUTER_WHITESPACE,
                                extraWS);
                            priorConst.insertAfter(currentConst);
                            priorConst = currentConst;
                            extraWS = null; // reset variable
                        }
                    }
                }

                // tokenText lenght might be 0 if we fixed an character error
                if ((!done) && (tokenText.length() > 0)) {
                    // create suitable type
                    if ((captured.length() == 0)
                            || (captured.equals("[CDATA"))) {
                        // textual content (character data) or CDATA section
                        // now we're outside a markup series
                        if (inMarkupSeries) {
                            inMarkupSeries = false;
                        }
                        if (captured.length() == 0) {
                            currentConst = new OtherConstituent(
                                OtherConstituent.TEXT, tokenText);
                        } else {
                            currentConst = new OtherConstituent(
                                OtherConstituent.CDATA_SECTION, tokenText);
                        }
                    } else {
                        // now we're inside a markup series
                        if (!inMarkupSeries) {
                            inMarkupSeries = true;
                            markupSeriesNo++;
                        }
                        currentConst = createMarkupConstituent(tokenText,
                            captured, markupSeriesNo, startAndEndTags);
                    }
                    if (fixCharacterErrors) {
                        // fix illegal characters in representation
                        fixRepresentation(currentConst);
                    }
                    // add to list
                    if (priorConst == null) {
                        firstConst = currentConst;
                    } else {
                        priorConst.insertAfter(currentConst);
                    }
                    priorConst = currentConst;
                }
            }
        } catch (IllegalArgumentException iae) {
            // convert unchecked into checked exception
            throw new ParsingException(
                "Uncorrectable error in XML input: " + iae.getMessage());
        }
        return firstConst;
    }

    /**
     * Replaces a start tag with an copy that is an empty tag. This means,
     * the representation of the created copy ends in "/&gt;" instead of "&gt;"
     * and its type is {@link TagConstituent#EMPTY_TAG} instead of
     * {@link TagConstituent#START_TAG}.
     *
     * <p>If the original tag is part of a list, it is removed and the copy
     * is inserted instead.
     *
     * @param startTag the start tag to replace
     * @return the created copy
     * @throws IllegalArgumentException if the specified tag is not a valid
     * start tag
     */
    private TagConstituent replaceWithEmptyCopy(final TagConstituent startTag)
            throws IllegalArgumentException {
        if (startTag.getType() != TagConstituent.START_TAG) {
            throw new IllegalArgumentException(
                "Tag to replace must be a start tag (actual type: "
                    + startTag.getType() + ')');
        }
        final StringBuffer representation =
            new StringBuffer(startTag.getRepresentantion());
        final int endMarker = representation.lastIndexOf(">");
        if (endMarker < 0) {
            throw new IllegalArgumentException(
                "Start tag representation is invalid: '>' missing!");
        }

        // insert '/' before closing '>'
        representation.insert(endMarker, '/');
        final TagConstituent result =
            new TagConstituent(TagConstituent.EMPTY_TAG, startTag.getName(),
                representation.toString(), startTag.getMarkupSeriesNo());

        // remove original tag from list (if any) and insert new one instead
        final XMLConstituent prevConst = startTag.previousConstituent();
        final XMLConstituent nextConst = startTag.nextConstituent();
        startTag.remove();

        if (prevConst != null) {
            prevConst.insertAfter(result);
        } else if (nextConst != null) {
            // this is the first tag in a list
            result.insertAfter(nextConst);
        }

        // copy will always be of the "regular" variety--that's ok, after
        // processing start tag would be "regularized" too
        return result;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("missing root name", missingRootName)
            .append("emptiable tags", emptiableTags)
            .append("delete control characters", deletingControlChars)
            .append("delete pseudo-tags", deletingPseudoTags)
            .append("escape pseudo-entities", escapingPseudoEntities)
            .toString();
    }

    /**
     * Helper method that tries to parse a string as an XML start or empty tag
     * that contains unquoted attribute values (might be followed by other
     * text). If this is the case, the unquoted values are fixed and a suitable
     * tag constituent is created.
     *
     * <p>Only call this method when you're sure that the input does not contain
     * a <em>valid</em> tag, i.e. if there must be at least one unquoted value.
     * This method doesn't store a markup series in the tag.
     *
     * @param text the text to parse
     * @return an array of two elements: the created {@link TagConstituent}
     * and a String containing the unused rest of the input text (might be empty
     * but not <code>null</code>; or <code>null</code> if the text couldn't be
     * parsed as a text.
     * @throws ParsingException might be thrown by {@link #checkEvent(String)}
     * implementations in subclasses if an "illicit" event occurred
     */
    private Object[] tryToFixTag(final String text) throws ParsingException {
        final Object[] result;
        final Matcher laxMatcher = LAX_START_OR_EMPTY_TAG.matcher(text);

        if (laxMatcher.lookingAt()) {
            // it's a start or empty tag with unquoted
            // attribute values -- fix them
            final String tagName = laxMatcher.group(1);
            final short tagType;

            result = new Object[2];
            // split tag + store rest (if any) to be returned
            result[1] = text.substring(laxMatcher.end());
            String oldTagRep = laxMatcher.group();

            // determine tag type
            final String lastGroup = laxMatcher.group(laxMatcher.groupCount());

            if (lastGroup == null) {
                tagType = TagConstituent.START_TAG;
            } else if ("/".equals(lastGroup)) {
                tagType = TagConstituent.EMPTY_TAG;                
            } else {
                // not supposed to happen
                throw new RuntimeException("Implementation error: last group of"
                        + " lax tag '" + laxMatcher.group() + "' is '"
                        + lastGroup + "' instead of '/' or null");
            }

            StringBuilder newTagRep = null;
            String unquotedValue;

            // is there still an unquoted attrib (whole value will be captured)?
            while (laxMatcher.groupCount() > 2
                    && (laxMatcher.group(3) != null)) {
                // create representation, starting with text until unquoted val.
                // (the 2nd groups contain '=' so we have to check the 3rd one)
                newTagRep = new StringBuilder(oldTagRep.substring(0,
                        laxMatcher.start(3)));
                unquotedValue = laxMatcher.group(3);

                // remove starting or ending quote
                if (unquotedValue.startsWith("\"")
                        || unquotedValue.startsWith("'")) {
                    unquotedValue = unquotedValue.substring(1);
                }
                if (unquotedValue.endsWith("\"")
                        || unquotedValue.endsWith("'")) {
                    unquotedValue = unquotedValue.substring(0,
                        unquotedValue.length() - 1);
                }

                // escape any embedded quotes
                unquotedValue = TextUtils.replaceAll(unquotedValue,
                    Pattern.compile("\""), "&quot;");

                // append the value within quotes
                newTagRep.append('"');
                newTagRep.append(unquotedValue);
                newTagRep.append('"');

                // append rest of tag
                newTagRep.append(oldTagRep.substring(laxMatcher.end(3),
                    laxMatcher.end()));

                // match again to check if there are more unquoted attribs
                laxMatcher.reset(newTagRep);
                if (!laxMatcher.matches()) {
                    // not supposed to happen
                    throw new RuntimeException("Implementation error while "
                            + "trying to fix unquoted attribute values: '"
                            + newTagRep + "' is no longer parsable as a tag");
                }
                oldTagRep = newTagRep.toString();
            }

            // create and store TagConstituent + log event
            result[0] = new TagConstituent(tagType, tagName,
                    newTagRep.toString());
            logEvent(EVENT_QUOTED_ATTRIBUTE_VALUES, newTagRep.toString());
        } else {
            result = null;
        }
        return result;
    }

}
