/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

import de.fu_berlin.ties.util.ListEntry;

/**
 * A constituent in an XML document.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/02/02 18:50:17 $, $Author: siefkes $
 */
public abstract class XMLConstituent extends ListEntry {

    /**
     * The number of characters from the representations of neighboring
     * constituents to include in {@link #toString()} call (including ellipses,
     * if required).
     */
    private static final int NEIGHBOR_LENGTH = 15;

    /**
     * The maximum number of characters from the representation to include in
     * {@link #toString()} call (including ellipses, if required).
     */
    private static final int REPRESENTATION_LENGTH = 33;

    /**
     * The representation of this constituent within the XML document, i.e. the
     * string fragment from the document representing this constituent.
     */
    private String representantion;

    /**
     * The type of this constituent. Suitable type constants are defined in
     * subclasses of this class.
     */
    private final short type;

    /**
     * Creates a new instance, without setting a reference to a next
     * constituent.
     *
     * @param constType the typ of this constituent (one of the type constants
     * defined in subclasses of this type)
     * @param rep the representation of this constituent within the XML
     * document, i.e. the string fragment from the document representing this
     * constituent
     */
    public XMLConstituent(final short constType, final String rep) {
        this(constType, rep, null, null);
    }

    /**
     * Creates a new instance.
     *
     * @param constType the typ of this constituent (one of the type constants
     * defined in subclasses of this type)
     * @param rep the representation of this constituent within the XML
     * document, i.e. the string fragment from the document representing this
     * constituent
     * @param prevEntry a reference to the next constituent
     * @param nextEntry a reference to the next constituent
     */
    public XMLConstituent(final short constType, final String rep,
            final XMLConstituent prevEntry,
            final XMLConstituent nextEntry) {
        super(prevEntry, nextEntry);
        type = constType;
        representantion = rep;
    }

    /**
     * Returns the representation of this tag within the XML document.
     *
     * @return the representation, i.e. the string fragment from the document
     * representing this constituent
     */
    public String getRepresentantion() {
        return representantion;
    }

    /**
     * Returns the type of this constituent. Suitable type constants are
     * defined in subclasses of this class.
     *
     * @return the constituent type
     */
    public short getType() {
        return type;
    }

    /**
     * Returns the next constituent in the list, if any. A convience alternative
     * to calling {@link ListEntry#next()} and casting the result to an
     * <code>XMLConstituent</code>.
     *
     * @return the next constituent in the file
     */
    public XMLConstituent nextConstituent() {
        return (XMLConstituent) next();
    }

    /**
     * Returns the previous constituent in the list, if any. A convience
     * alternative to calling {@link ListEntry#previous()} and casting the
     * result to an <code>XMLConstituent</code>.
     *
     * @return the previous constituent in the file
     */
    public XMLConstituent previousConstituent() {
        return (XMLConstituent) previous();
    }

    /**
     * Sets the representation of this tag within the XML document.
     *
     * @param string the representation, i.e. the string fragment from the
     * document representing this constituent
     */
    public void setRepresentantion(final String string) {
        representantion = string;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        final ToStringBuilder builder = new ToStringBuilder(this)
            .append("type", type)
            .append("representation",
                StringUtils.abbreviate(representantion, REPRESENTATION_LENGTH));

        // include fragments of neighbor representations instead of including
        // output from super.toString
        if (hasPrevious()) {
            final String prevRep = previousConstituent().getRepresentantion();
            builder.append("previous", (prevRep.length() > NEIGHBOR_LENGTH)
                ? "..." + StringUtils.right(prevRep, NEIGHBOR_LENGTH - 3)
                : prevRep);
        }
        if (hasNext()) {
            builder.append("next", StringUtils.abbreviate(
                nextConstituent().getRepresentantion(), NEIGHBOR_LENGTH));
        }

        return builder.toString();
    }

}
