/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml;

import de.fu_berlin.ties.text.TextTokenizer;

/**
 * Static factory for creating a {@link de.fu_berlin.ties.text.TextTokenizer}s
 * for XML-like input.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2004/04/08 16:38:26 $, $Author: siefkes $
 */
public final class XMLTokenizerFactory {

    /**
     * Pattern fragment listing allowed whitespace characters in an XML
     * document. Space, tab, line feed, carriage return, and some other
     * line-ending characters (according to XML 1.1) are allowed.
     */
    public static final String XML_WHITESPACE_CHARS = " \\t\\n\\r\u0085\u2028";

    /**
     * Pattern specifying whitespace in an XML document (one or more
     * {@linkplain #XML_WHITESPACE_CHARS whitespace characters}).
     */
    public static final String XML_WHITESPACE =
        "[" + XML_WHITESPACE_CHARS + "]+";

    /**
     * Pattern specifying optional whitespace in an XML document (zero or more
     * {@linkplain #XML_WHITESPACE_CHARS whitespace characters}).
     */
    public static final String XML_OPT_WHITESPACE =
        "[" + XML_WHITESPACE_CHARS + "]*";

    /**
     * Pattern fragment listing valid start characters of XML names (according
     * to XML 1.1).
     */
    private static final String XML_NAME_START_CHARS =
        ":A-Z_a-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D"
        + "\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF"
        + "\uF900-\uFDCF\uFDF0-\uFFFD";

    /**
     * Pattern fragment listing valid start characters of XML names that can
     * only be represented as surrogate pairs in Java: characters "0x10000" to
     * "0xEFFFF", encoded in the UFT-16 surrogate pairs "0xD800 0xDC00" to
     * "0xDB7F 0xDFFF".
     */
    private static final String XML_NAME_SURROGATE_PAIRS =
        "[\uD800-\uDB7F][\uDC00-\uDFFF]";

    /**
     * Pattern string specifying the class of valid start characters of XML
     * names.
     */
    public static final String XML_NAME_START_CHAR =
        "(?:[" + XML_NAME_START_CHARS + "]|" + XML_NAME_SURROGATE_PAIRS + ")";

    /**
     * Pattern string for XML names (according to XML 1.1).
     */
    public static final String XML_NAME = XML_NAME_START_CHAR + "(?:[-"
        + XML_NAME_START_CHARS + ".0-9\u00B7\u0300-\u036F\u203F-\u2040]|"
        + XML_NAME_SURROGATE_PAIRS + ")*";

    /**
     * Pattern string for strings enclosed in full or half quotes, e.g. XML
     * attribute values.
     */
    public static final String XML_QUOTED_STRING =
        "(?:\"[^\"]*\"|'[^']*')";

    /**
     * Pattern string for the '=' sign, optionally surrounded by whitespace.
     */
    public static final String XML_EQUAL_SIGN =
        XML_OPT_WHITESPACE + "=" + XML_OPT_WHITESPACE;

    /**
     * Pattern string specifying an XML attribute (name = quoted-value pair).
     */
    public static final String XML_ATTRIBUTE = XML_NAME + XML_EQUAL_SIGN
        + XML_QUOTED_STRING;

    /**
     * Pattern string for a visible textual token in XML documents (contains
     * neither whitespace nor markup).
     */
    public static final String XML_CDATA_TOKEN =
        "[^<" + XML_WHITESPACE_CHARS + "]+";

    /**
     * Pattern string specifying an XML start or empty tag (combined into a
     * single pattern to avoid unnecessary backtracking).
     */
    public static final String XML_START_OR_EMPTY_TAG =
        "<(" + XML_NAME + ")" + "(?:" + XML_WHITESPACE + XML_ATTRIBUTE + ")*"
        + XML_OPT_WHITESPACE + "(/)?>";

    /**
     * Pattern string specifying an XML end tag.
     */
    public static final String XML_END_TAG = "<(/" + XML_NAME
        + ")" + XML_OPT_WHITESPACE + ">";

    /**
     * Helper pattern string used to construct a P.I.
     */
    private static final String XML_PI_START = "\\?" + XML_NAME;

    /**
     * Helper pattern string used to construct a P.I.
     */
    private static final String XML_PI_REST = "[^<>]*?\\?";

    /**
     * Pattern string specifying an XML prolog or processing instruction.
     */
    public static final String XML_PROLOG_OR_PI =
        "<(" + XML_PI_START + ")" + XML_PI_REST + ">";

    /**
     * Helper pattern string used to construct a comment.
     */
    private static final String XML_COMMENT_START = "!--";

    /**
     * Helper pattern string used to construct a comment.
     */
    private static final String XML_COMMENT_END = ".*?--";

    /**
     * Pattern string specifying an XML comment.
     */
    public static final String XML_COMMENT = "(<" + XML_COMMENT_START
        + ")" + XML_COMMENT_END + ">";

    /**
     * Pattern string specifying a PE reference within a doctype declaration.
     */
    public static final String PE_REFERENCE = "%" + XML_NAME + ";";

    /**
     * Pattern string specifying an markup declaration within a doctype
     * declaration. A markup declaration either declares an entity, element,
     * attribute, or notation; or it is a processing instruction or a comment.
     */
    public static final String MARKUP_DECL =
        "<(?:(!?:ELEMENT|ATTLIST|ENTITY|NOTATION)(?:[^\"'>]*"
            + XML_QUOTED_STRING + ")*[^\"'>]*|" + XML_PI_START + XML_PI_REST
            + "|" + XML_COMMENT_START + XML_COMMENT_END + ")>";

    /**
     * Pattern string specifying an XML document type declaration.
     */
    public static final String XML_DOCTYPE =
        "<(!DOCTYPE)(?:" + XML_WHITESPACE + XML_NAME + ")*(?:" + XML_WHITESPACE
        + XML_QUOTED_STRING + "){1,2}" + XML_OPT_WHITESPACE + "(?:\\[(?:"
        + MARKUP_DECL + "|" + PE_REFERENCE + "|" + XML_WHITESPACE
        + ")*\\])?" + XML_OPT_WHITESPACE + ">";

    /**
     * Pattern string specifying a CDATA section in an XML document.
     */
    public static final String XML_CDATA_SECTION =
        "<!(\\[CDATA).*?\\]\\]>";

    /**
     * Pattern string specifying textual content (character data) in an XML
     * document. Starting and trailing whitespace is not included.
     */
    public static final String XML_TEXTUAL_CONTENT = "(?:" + XML_CDATA_TOKEN
        + XML_WHITESPACE + ")*" + XML_CDATA_TOKEN;

    /**
     * The array of patterns used for shallow XML parsing.
     */
    public static final String[] XML_PATTERNS = {
        XML_START_OR_EMPTY_TAG, XML_END_TAG,
        XML_DOCTYPE, XML_PROLOG_OR_PI,
        XML_COMMENT, XML_CDATA_SECTION, XML_TEXTUAL_CONTENT
    };

    /**
     * Factory method to create an instance for parsing files in XML syntax.
     * Creates a shallow XML parser that splits XML input into a series of tags
     * and textual content. The main difference from regular XML parsers is
     * that tags can occur in any order; nesting constrains are not enforced.
     * This can be useful to repair XML-like files.
     *
     * <p>The type of token returns can be determined by calling
     * {@link TextTokenizer#capturedText()}:
     *
     * <dl>
     * <dt><em>tagname</em><dt>
     * <dd>for start tags of type <em>tagname</em></dd>
     * <dt>/<em>tagname</em><dt>
     * <dd>for end tags</dd>
     * <dt><em>tagname/</em>/<dt>
     * <dd>for empty tags</dd>
     * <dt>!DOCTYPE<dt>
     * <dd>for doctype declarations</dd>
     * <dt>?<em>targetname</em><dt>
     * <dd>for prolog ("?xml") and processing instructions</dd>
     * <dt>&lt;!--<dt>
     * <dd>for comments</dd>
     * <dt>[CDATA<dt>
     * <dd>for CDATA section</dd>
     * <dt>"" (empty string)<dt>
     * <dd>for textual content (character data)</dd>
     * </dl>
     *
     * <p>Whitespace between tags and before and after textual content can be
     * retrieved using the {@link TextTokenizer#precedingWhitespace()} method.
     *
     * @param text the text to tokenize
     * @param ensureWhitespace whether to validate whitespace
     * ({@link TextTokenizer#isWhitespacePatternEnsured()}), throwing an
     * exception if a document contains serious errors (i.e. an unescaped "&lt;"
     * within textual content); if <code>false</code>, the caller is responsible
     * for validating whitespace
     * @return a new instance suitable for parsing XML
     */
    public static TextTokenizer createXMLTokenizer(final CharSequence text,
            final boolean ensureWhitespace) {
        final TextTokenizer result = new TextTokenizer(XML_PATTERNS,
            XML_WHITESPACE, text);
        result.setWhitespacePatternEnsured(ensureWhitespace);
        return result;
    }

    /**
     * Private constructor to prevent creation of instances.
     */
    private XMLTokenizerFactory() {
        super();
    }

}
