/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml.dom;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.List;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.lang.StringUtils;
import org.dom4j.Attribute;
import org.dom4j.Branch;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.Node;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.SAXReader;
import org.dom4j.io.XMLWriter;
import org.dom4j.io.aelfred2.SAXDriver;

import de.fu_berlin.ties.io.IOUtils;
import de.fu_berlin.ties.text.TextUtils;

/**
 * A static class that provides utility constants and methods for working with
 * DOM-like XML representations, focussing especially on
 * <a href="http://www.dom4j.org/">dom4j</a>.
 * No instances of this class can be created, only the static members
 * should be used.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.14 $, $Date: 2004/11/02 11:36:25 $, $Author: siefkes $
 */
public final class DOMUtils {

    /**
     * The parser used to read XML documents.
     */
    private static final SAXReader XML_READER = new SAXReader(new SAXDriver());

    /**
     * Private constructor prevents creation of instances.
     */
    private DOMUtils() {
        super();
    }

    /**
     * Returns the attribute with the given name, compatible to the name format
     * returned by {@link #name(Attribute)}.  If there are more than one
     * attributes with the given name (e.g. in different namespaces) then the
     * first one is returned.
     *
     * @param element the element whose attribute to return
     * @param name the name of the attribute, compatible to the name format
     * returned by {@link #name(Attribute)}
     * @return the (first) matching attribute or <code>null</code> if none
     * exists
     */
    public static Attribute attributeByName(final Element element,
            final String name) {
        return element.attribute(name);
    }

    /**
     * Recursively collects the complete textual content of a branch, i.e.
     * a document or element.
     *
     * @param branch the branch to recurse
     * @return the collected text of the branch and all its child elements
     */
    public static String collectText(final Branch branch) {
        final StringBuffer appender = new StringBuffer();

        // delegate + return result as string
        collectText(branch, appender);
        return appender.toString();
    }

    /**
     * Recursively collects the complete textual content of a branch, i.e.
     * a document or element.
     *
     * @param branch the branch to recurse
     * @param appender the collected text of the branch and all its child
     * elements is appended to this string buffer
     */
    public static void collectText(final Branch branch,
                                   final StringBuffer appender) {
       // delegate to helper method
       try {
           collectText(branch, appender, null);
        } catch (IOException ioe) {
            // this should never occur when using a string buffer
            throw new RuntimeException(
                "Implementation error: " + ioe.toString(), ioe);
        }
    }

    /**
     * Helper method that recursively collects the complete textual content of a
     * branch and sends the result to a string buffer or writer.
     *
     * @param branch the branch to recurse
     * @param appender the collected text of the branch and all its child
     * elements is appended to this buffer, if not <code>null</code>
     * @param writer the collected text of the branch and all its child
     * elements is appended to this writer if <code>appender</code> is
     * <code>null</code>
     * @throws IOException if an I/O error occurs while writing to the writer
     */
    private static void collectText(final Branch branch,
            final StringBuffer appender, final Writer writer)
            throws IOException {
        Node currentChild;
        int currentType;

        // process child nodes
        for (int i = 0; i < branch.nodeCount(); i++) {
            currentChild = branch.node(i);
            currentType = currentChild.getNodeType();

            if ((currentType == Node.TEXT_NODE)
                    || (currentType == Node.CDATA_SECTION_NODE)
                    || (currentType == Node.ENTITY_REFERENCE_NODE)) {
                // send textual content to buffer resp. writer
                if (appender != null) {
                    appender.append(currentChild.getText());
                } else {
                    writer.write(currentChild.getText());
                }
            } else if (currentType == Node.ELEMENT_NODE) {
                // recursively collect text of child elements
                collectText((Element) currentChild, appender, writer);
            }
        }
    }

    /**
     * Recursively collects the complete textual content of a branch, i.e.
     * a document or element.
     *
     * @param branch the branch to recurse
     * @param writer the collected text of the branch and all its child
     * elements is appended to this writer; flushed but not closed by this
     * method
     * @throws IOException if an I/O error occurs while writing to the writer
     */
    public static void collectText(final Branch branch, final Writer writer)
            throws IOException {
        // delegate to helper method + flush
        collectText(branch, null, writer);
        writer.flush();
    }

    /**
     * Returns the child elements with the given name, compatible to the name
     * format returned by {@link #name(Element)}. If no elements are found
     * then this method returns an empty list.
     *
     * @param element the element whose child elements to return
     * @param name the name of the child elements, compatible to the name format
     * returned by {@link #name(Attribute)}
     * @return a list of all the child {@link Element}s for the given name
     */
    public static List elementsByName(final Element element,
            final String name) {
        return element.elements(name);
    }

    /**
     * Static method that returns a String representing the name of an attribute
     * in an XML document. This method should always be used when building
     * context representations and related structures to ensure that
     * attributes are represented in a unified way. See {@link #name(Element)}
     * for details.
     *
     * @param attrib the element to name
     * @return the name to use for this element
     */
    public static String name(final Attribute attrib) {
        return attrib.getName();
    }

    /**
     * Static method that returns a String representing the name of an element
     * in an XML document. This method should always be used when building
     * context representations and related structures to ensure that
     * elements are represented in a unified way. Please don't call
     * {@link org.dom4j.Node#getName()} or {@link Element#getQualifiedName()}
     * or similar methods directly in such cases.
     *
     * <p>Currently, only the local name if used, namespace URIs and namespace
     * prefixes are ignored. Including namespace prefixes in context
     * representations would be quite useless, because in different document
     * different prefixes can represent the same namespace and vice versa.
     *
     * <p>Including namespace URIs might lead to higher precision by avoiding
     * the risk of confusing elements from totally different namespaces. On
     * other other hand it might lead to lower recall and slower learning
     * because elements from similar namespaces (e.g. different versions of the
     * HTML standard) are all considered separated from each other.
     *
     * @param element the element to name
     * @return the name to use for this element
     */
    public static String name(final Element element) {
        return element.getName();
    }

    /**
     * Reads an XML document from a local file, using a configured charset.
     * Delegates to {@link IOUtils#openReader(File, Configuration)} to determine
     * the character set. Compressed files are automatically decompressed (cf.
     * {@link IOUtils#openCompressableInStream(InputStream)}).
     *
     * @param file the file to read
     * @param config the configuration to use
     * @return the newly created document
     * @throws DocumentException if an error occurs during parsing
     * @throws IOException if an I/O error occurrs
     */
    public static Document readDocument(final File file,
            final Configuration config)
    throws DocumentException, IOException {
        Reader reader = null;
        try {
            // open reader and delegate
            reader = IOUtils.openReader(file, config);
            return readDocument(reader);
        } finally {
            // close the reader
            IOUtils.tryToClose(reader);
        }
    }

    /**
     * Reads an XML document from a local file, using a given charset.
     * Compressed files are automatically decompressed (cf.
     * {@link IOUtils#openCompressableInStream(InputStream)})
     *
     * @param file the file to read
     * @param charset the character set to use for reading the file;
     * if <code>null</code>, the default charset of the current platform is used
     * @return the newly created document
     * @throws DocumentException if an error occurs during parsing
     * @throws IOException if an I/O error occurrs
     */
    public static Document readDocument(final File file, final String charset)
    throws DocumentException, IOException {
        Reader reader = null;
        try {
            // open reader and delegate
            reader = IOUtils.openReader(file, charset);
            return readDocument(reader);
        } finally {
            // close the reader
            IOUtils.tryToClose(reader);
        }
    }

    /**
     * Reads an XML document from a given stream.
     * Compressed files are automatically decompressed (cf.
     * {@link IOUtils#openCompressableInStream(InputStream)})
     *
     * @param in stream containing the text to parse; not closed by this method
     * @return the newly created document
     * @throws DocumentException if an error occurs during parsing
     * @throws IOException if an I/O error occurrs
     */
    public static Document readDocument(final InputStream in)
            throws DocumentException, IOException {
        return XML_READER.read(IOUtils.openCompressableInStream(in));
    }

    /**
     * Reads an XML document from a given reader.
     *
     * @param reader reader containing the text to parse; not closed by this
     * method
     * @return the newly created document
     * @throws DocumentException if an error occurs during parsing
     */
    public static Document readDocument(final Reader reader)
            throws DocumentException {
        return XML_READER.read(reader);
    }

    /**
     * Builds a simple partial representation of an element, containing the
     * {@linkplain #name(Element) name of the element} and its normalized
     * and {@linkplain TextUtils#shorten(String) shortened} textual content.
     * Useful for logging.
     *
     * @param element the element to show (may be <code>null</code>)
     * @return a simple partial representation of the element
     */
    public static String showElement(final Element element) {
        if (element != null) {
            return '[' + name(element) + ": "
            + TextUtils.shorten(TextUtils.normalize(collectText(element)))
            + ']';
        } else {
            return null;
        }
    }

    /**
     * Builds a simple partial representation of a textual token in an element,
     * containing the {@linkplain #name(Element) name of the element} and the
     * normalized and {@linkplain TextUtils#shorten(String) shortened} text of
     * the token. Useful for logging.
     *
     * @param element the element containing the token; must not be
     * <code>null</code>
     * @param token the token to show (may be <code>null</code>)
     * @return a simple representation joining element and token
     */
    public static String showToken(final Element element, final String token) {
        return '['
        + (token == null ? null : TextUtils.shorten(TextUtils.normalize(token)))
        + '/' + name(element) + ']';
    }

    /**
     * Writes an XML document to a given stream.
     *
     * @param document the document to write
     * @param out the stream to write the document text to; flushed
     * but not closed by this method
     * @throws IOException if an I/O error occurs during writing
     */
    public static void writeDocument(final Document document,
            final OutputStream out) throws IOException {
        final OutputFormat outFormat = OutputFormat.createPrettyPrint();
        final XMLWriter xmlWriter = new XMLWriter(out, outFormat);
        xmlWriter.write(document);
        xmlWriter.flush();
    }

    /**
     * Writes an XML document to a given writer, using the character set of the
     * underlying output stream.
     *
     * @param document the document to write
     * @param writer the writer to write the document text to; flushed
     * but not closed by this method
     * @throws IOException if an I/O error occurs during writing
     */
    public static void writeDocument(final Document document,
            final OutputStreamWriter writer) throws IOException {
        writeDocument(document, writer, IOUtils.determineCharsetName(writer));
    }

    /**
     * Writes an XML document to a given writer, using the given character set.
     *
     * @param document the document to write
     * @param writer the writer to write the document text to; flushed
     * but not closed by this method
     * @param charset the character set of the writer; this must be a valid
     * charset name (not <code>null</code> or empty etc.), it should be
     * the <em>canonical</em> (standard) name of the used charset
     * @throws IllegalArgumentException if the specific charset is
     * <code>null</code> or empty
     * @throws IOException if an I/O error occurs during writing
     */
    public static void writeDocument(final Document document,
            final Writer writer, final String charset)
    throws IllegalArgumentException, IOException {
        if (StringUtils.isEmpty(charset)) {
            throw new IllegalArgumentException("Charset name must not be "
                    + ((charset == null) ? "null" : "empty"));
        }

        final OutputFormat outFormat = OutputFormat.createPrettyPrint();
        outFormat.setEncoding(charset);

        final XMLWriter xmlWriter = new XMLWriter(writer, outFormat);
        xmlWriter.write(document);
        xmlWriter.flush();
    }

}
