/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml.dom;

import java.io.IOException;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Document;
import org.dom4j.Element;
import org.dom4j.Node;
import org.dom4j.NodeFilter;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;

import de.fu_berlin.ties.text.TokenContainer;
import de.fu_berlin.ties.text.TokenizerFactory;

/**
 * Walks through a document, handing the elements matched by a
 * {@link org.dom4j.NodeFilter} over to an
 * {@link de.fu_berlin.ties.xml.dom.ElementProcessor}.
 * The textual contents of the document are tokenized; the resulting tokens
 * are stored in a multi-set ({@link org.apache.commons.collections.Bag}).
 *
 * @author Christian Siefkes
 * @version $Revision: 1.6 $, $Date: 2004/11/04 15:27:06 $, $Author: siefkes $
 */
public class DocumentWalker {

    /**
     * A filter used to decide which elements to hand over to the element
     * processor.
     */
    private final NodeFilter elementFilter;

    /**
     * Used to process the elements selected by the element filter.
     */
    private final ElementProcessor elementProcessor;

    /**
     * Used to instantiate tokenizers.
     */
    private final TokenizerFactory factory;

    /**
     * Creates a new instance.
     *
     * @param filter a filter used to decide which elements to hand over to the
     * element processor
     * @param processor used to process the elements selected by the filter
     * @param tFactory used to instantiate tokenizers
     */
    public DocumentWalker(final NodeFilter filter,
                          final ElementProcessor processor,
                          final TokenizerFactory tFactory) {
        super();
        elementFilter = filter;
        elementProcessor = processor;
        factory = tFactory;
    }

    /**
     * Walks through the contents of an XML document, tokenizing the textual
     * contents. The resulting tokens are stored in a {@link TokenContainer}.
     *
     * @param document the document to walk through
     * @param context a map of objects that are made available for processing;
     * might be <code>null</code> if not requred by the element processor
     * @throws IOException might be thrown by the element processor
     * @throws ProcessingException might be thrown by the element processor
     */
    public final void walk(final Document document, final ContextMap context)
                           throws IOException, ProcessingException {
        final TokenContainer tokenContainer = new TokenContainer(factory);
        final Element root = document.getRootElement();
        walk(root, tokenContainer, context);
    }

    /**
     * Walks through the contents of a node, tokenizing textual contents
     * and recursing through nested elements. For elements matched by the
     * registered node filter, the registered element processor is called
     * -- the full textual content of the matched element and its children is
     * available via {@link TokenContainer#getLast()}. For other elements,
     * the textual contents are stored and child elements are walked through
     * and matched recursively.
     *
     * <p><i>A successful match stops recursion, i.e. child elements of a
     * matching element are never handed over to the node filter for
     * testing (in the case, only the textual contents are recursively
     * collected).</i>
     *
     * @param element the element to walk through
     * @param tokenContainer container storing all tokens
     * @param context a map of objects that are made available for processing;
     * might be <code>null</code> if not required by the element processor
     * @throws IOException might be thrown by the element processor
     * @throws ProcessingException might be thrown by the element processor
     */
    protected void walk(final Element element,
                        final TokenContainer tokenContainer,
                        final ContextMap context)
                        throws IOException, ProcessingException {
        final boolean matched = elementFilter.matches(element);
        StringBuffer collectedText = new StringBuffer();
        Node currentChild;
        int currentType;

        // process child nodes
        for (int i = 0; i < element.nodeCount(); i++) {
            currentChild = element.node(i);
            currentType = currentChild.getNodeType();

            if ((currentType == Node.TEXT_NODE)
                    || (currentType == Node.CDATA_SECTION_NODE)
                    || (currentType == Node.ENTITY_REFERENCE_NODE)) {
                // collect textual content
                collectedText.append(currentChild.getText());
            } else if (currentType == Node.ELEMENT_NODE) {
                if (matched) {
                    // recursively collect text of child elements
                    DOMUtils.collectText((Element) currentChild, collectedText);
                } else {
                    // flush text
                    if (collectedText.length() > 0) {
                        tokenContainer.add(collectedText.toString());
                        collectedText.setLength(0); // reset StringBuffer
                    }

                    // walk though child elements for matching
                    walk((Element) currentChild, tokenContainer, context);
                }
            }
        }

        // tokenize (remaining) text
        tokenContainer.add(collectedText.toString());

        if (matched) {
            // delegate to element processor
            elementProcessor.processElement(element, tokenContainer, context);
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("element filer", elementFilter)
            .append("element processor", elementProcessor)
            .append("factory", factory)
            .toString();
    }

}
