/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml.dom;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Element;
import org.dom4j.Node;
import org.dom4j.NodeFilter;

/**
 * A node filter that only accepts elements with a specified name.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2004/09/07 10:02:00 $, $Author: siefkes $
 */
public class ElementNameFilter implements NodeFilter {

    /**
     * An immutable set of element name strings whose format must be compatible
     * to {@link DOMUtils#name(Element)}. An elements is accepted iff its name
     * is contained in this set.
     */
    private final Set<String> names;

    /**
     * Creates a new instance.
     *
     * @param nameSet set of element name strings to accept, format must be
     * compatible to {@link DOMUtils#name(Element)}
     */
    public ElementNameFilter(final Set<String> nameSet) {
        super();
        // make set immutable
        names = Collections.unmodifiableSet(nameSet);
    }

    /**
     * Creates a new instance.
     *
     * @param nameList array of element name strings to accept, format must be
     * compatible to {@link DOMUtils#name(Element)}
     */
    public ElementNameFilter(final List<String> nameList) {
        // convert list to set
        this(new HashSet<String>(nameList));
    }

    /**
     * Creates a new instance.
     *
     * @param nameArray array of element name strings to accept, format must be
     * compatible to {@link DOMUtils#name(Element)}
     */
    public ElementNameFilter(final String[] nameArray) {
        // convert array to list
        this(Arrays.asList(nameArray));
    }

    /**
     * Creates a new instance that accepts only elements of a single type.
     *
     * @param elementName Only elements with this name are accepted, format must
     * be compatible to {@link DOMUtils#name(Element)}
     */
    public ElementNameFilter(final String elementName) {
        // wrap in set
        this(Collections.singleton(elementName));
    }

    /**
     * Returns the set of element names accepted by this filter.
     *
     * @return an immutable set containing the element names (Strings) accepted
     * by this filter
     */
    public Set getNames() {
        return names;
    }

    /**
     * Tests whether a node is accected by this filter. To match a node must
     * fulfill two conditions: (a) it must be an {@link Element} and (b)
     * {@link #getNames()} must contain the name of the element as returned by
     * {@link DOMUtils#name(Element)}.
     *
     * @param node the node to test
     * @return <code>true</code> iff this filter matches the given node
     */
    public boolean matches(final Node node) {
        final boolean result;
        if (node instanceof Element) {
            final Element element = (Element) node;
            result = names.contains(DOMUtils.name(element));
        } else {
            result = false;
        }
        return result;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("names", names)
            .toString();
    }

}
