/*
 * Copyright (C) 2003-2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml.dom;

import java.io.IOException;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.dom4j.Document;
import org.dom4j.Element;
import org.dom4j.Node;

import de.fu_berlin.ties.ContextMap;
import de.fu_berlin.ties.ProcessingException;

import de.fu_berlin.ties.text.TextTokenizer;
import de.fu_berlin.ties.text.TokenCounter;
import de.fu_berlin.ties.text.TokenDetails;
import de.fu_berlin.ties.text.TokenizerFactory;

/**
 * Walks through a document, handing all textual tokens over to a
 * {@link de.fu_berlin.ties.xml.dom.TokenProcessor}.
 *
 * <p>Instances of this class are thread-safe iff the provided
 * <code>TokenProcessor</code> is -- but subclass implementations might be not.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.9 $, $Date: 2004/11/19 14:05:09 $, $Author: siefkes $
 */
public class TokenWalker {

    /**
     * Used to process the tokens.
     */
    private final TokenProcessor tokenProcessor;

    /**
     * Used to instantiate tokenizers.
     */
    private final TokenizerFactory factory;

    /**
     * Creates a new instance.
     *
     * @param processor used to process the tokens
     * @param tFactory used to instantiate tokenizers
     */
    public TokenWalker(final TokenProcessor processor,
            final TokenizerFactory tFactory) {
        super();
        tokenProcessor = processor;
        factory = tFactory;
    }

    /**
     * Helper method that tokenizes the collected textual contents of an
     * element and delegates to the token processor for each of them.
     *
     * @param element the element to walk through
     * @param collectedText the collected textual contents (limited to the
     * text between/before/after child elements in case of mixed content)
     * @param tokenCounter keeps track of the encountered tokens
     * @param tokenizer used to tokenize text
     * @param context a map of objects that are made available for processing
     * @throws IOException might be throws by the token processor
     * @throws ProcessingException might be throws by the token processor
     */
    protected void processCollectedText(final Element element,
                        final CharSequence collectedText,
                        final TokenCounter tokenCounter,
                        final TextTokenizer tokenizer,
                        final ContextMap context)
                        throws IOException, ProcessingException {
        String currentToken;
        tokenizer.reset(collectedText);
        TokenDetails details;

        while ((currentToken = tokenizer.nextToken()) != null) {
            // add each token to container (incl. whitespace) + call processor
            tokenCounter.add(tokenizer.hasPrecedingWhitespace(), currentToken);
            details = new TokenDetails(currentToken, tokenCounter.getLastRep(),
                    tokenCounter.size() - 1,
                    tokenCounter.isWhitespaceBeforeLast());
            processToken(element, tokenizer.leftText().toString(), details,
                tokenizer.rightText().toString(), context);
        }

        // add trailing whitespace, if any
        if (tokenizer.hasPrecedingWhitespace()) {
            tokenCounter.addWhitespace();
        }
    }

    /**
     * Processes a token in an XML element by delegating to the configured
     * {@link TokenProcessor}.
     *
     * @param element the element containing the token
     * @param left the textual contents of the element to the left of the
     * <code>token</code> (in case of mixed contents, only up to the last
     * preceding child element, if any)
     * @param details details about the token to process
     * @param right the textual contents of the element to the right of the
     * <code>token</code> (in case of mixed contents, only up to the next
     * following child element, if any)
     * @param context a map of objects that are made available for processing
     * @throws IOException if an I/O error occurs
     * @throws ProcessingException if an error occurs during processing
     */
    protected void processToken(final Element element, final String left,
            final TokenDetails details, final String right,
            final ContextMap context) throws IOException, ProcessingException {
        // simply delegate to tokenProcessor
        tokenProcessor.processToken(element, left, details, right, context);
    }

    /**
     * Walks through the contents of an XML document, tokenizing the textual
     * contents. The resulting tokens are handed over to the stored
     * {@link TokenProcessor}.
     *
     * @param document the document to walk through
     * @param context a map of objects that are made available for processing;
     * might be <code>null</code> if not requred by the token processor
     * @throws IOException might be throws by the token processor
     * @throws ProcessingException might be throws by the token processor
     */
    public void walk(final Document document, final ContextMap context)
                           throws IOException, ProcessingException {
        final TokenCounter tokenCounter = new TokenCounter();
        final TextTokenizer tokenizer = factory.createTokenizer("");
        final Element root = document.getRootElement();
        walk(root, tokenCounter, tokenizer, context);
    }

    /**
     * Walks through the contents of a node, tokenizing textual contents
     * and recursing through nested elements. The registered token processor
     * is called for each token.
     *
     * @param element the element to walk through
     * @param tokenCounter keeps track of the encountered tokens
     * @param tokenizer used to tokenize text
     * @param context a map of objects that are made available for processing
     * @throws IOException might be throws by the token processor
     * @throws ProcessingException might be throws by the token processor
     */
    protected void walk(final Element element,
                        final TokenCounter tokenCounter,
                        final TextTokenizer tokenizer,
                        final ContextMap context)
                        throws IOException, ProcessingException {
        StringBuffer collectedText = new StringBuffer();
        Node currentChild;
        int currentType;

        // process child nodes
        for (int i = 0; i < element.nodeCount(); i++) {
            currentChild = element.node(i);
            currentType = currentChild.getNodeType();

            if ((currentType == Node.TEXT_NODE)
                    || (currentType == Node.CDATA_SECTION_NODE)
                    || (currentType == Node.ENTITY_REFERENCE_NODE)) {
                // collect textual content
                collectedText.append(currentChild.getText());
            } else if (currentType == Node.ELEMENT_NODE) {
                // process any text collected so far
                if (collectedText.length() > 0) {
                    processCollectedText(element, collectedText,
                        tokenCounter, tokenizer, context);
                    collectedText.setLength(0); // reset StringBuffer
                }

                // walk though child elements for matching
                walk((Element) currentChild, tokenCounter, tokenizer,
                    context);
            }
        }

        // process any remaining collected text
        if (collectedText.length() > 0) {
            processCollectedText(element, collectedText,
                tokenCounter, tokenizer, context);
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("token processor", tokenProcessor)
            .append("factory", factory)
            .toString();
    }

}
