/*
 * Copyright (C) 2004 Christian Siefkes <christian@siefkes.net>.
 * Development of this software is supported by the German Research Society,
 * Berlin-Brandenburg Graduate School in Distributed Information Systems
 * (DFG grant no. GRK 316).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, visit
 * http://www.gnu.org/licenses/lgpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */
package de.fu_berlin.ties.xml.io;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import org.dom4j.Attribute;
import org.dom4j.Element;
import org.dom4j.Namespace;
import org.dom4j.QName;
import org.dom4j.tree.DefaultElement;

/**
 * An XML element that is targeted at storing an Java object.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.1 $, $Date: 2004/12/06 18:00:41 $, $Author: siefkes $
 */
public class ObjectElement extends DefaultElement {

    /**
     * The name of the attribute used to store the Java class of an object:
     * {@value}.
     */
    public static final String JAVA_CLASS_ATTRIBUTE = "java";


    /**
     * Creates (deserializes) an object of a specified type by calling a
     * constructor of the class that accepts an XML element as single argument
     * and passing itself as parameter. Only classes that provide a suitable
     * constructor can be instantiated this way.
     *
     * @param type the class of the object to create; must have a constructor
     * whose only argument is an {@link Element}
     * @return the created object; the returned object will be an instance of
     * the specified class
     * @throws InstantiationException if instantiation failed
     * @throws SecurityException if access to the required reflection
     * information is denied
     */
    public static final Object createObject(final Element element,
            final Class type)
    throws InstantiationException, SecurityException {
        try {
            // load constructor with Element as only argument
            final Constructor cons = type.getConstructor(
                    new Class[] {Element.class});
            // pass element as single argument
            return cons.newInstance(new Object[] {element});
        } catch (IllegalAccessException iae) {        // repackage exception
            final InstantiationException ie =
                new InstantiationException(iae.toString());
            ie.initCause(iae);
            throw ie;
        } catch (InvocationTargetException ite) {     // repackage exception
            final InstantiationException ie =
                new InstantiationException(ite.toString());
            ie.initCause(ite);
            throw ie;
        } catch (ExceptionInInitializerError eiie) {  // repackage exception
            final InstantiationException ie =
                new InstantiationException(eiie.toString());
            ie.initCause(eiie);
            throw ie;
        } catch (NoSuchMethodException nsme) {        // repackage exception
            final InstantiationException ie =
                new InstantiationException(nsme.toString());
            ie.initCause(nsme);
            throw ie;
        }
    }


    /**
     * Creates a new instance.
     *
     * @param name the name of the element
     * @param javaClass the class of the object to store
     */
    public ObjectElement(final String name, final Class javaClass) {
        super(name);
        addAttribute(JAVA_CLASS_ATTRIBUTE, javaClass.getName());
    }

    /**
     * Creates a new instance.
     *
     * @param qname the qualified name of the element
     * @param javaClass the class of the object to store
     */
    public ObjectElement(final QName qname, final Class javaClass) {
        super(qname);
        addAttribute(JAVA_CLASS_ATTRIBUTE, javaClass.getName());
    }

    /**
     * Creates a new instance.
     *
     * @param name the local name of the element
     * @param namespace the namespace of the element
     * @param javaClass the class of the object to store
     */
    public ObjectElement(final String name, final Namespace namespace,
            final Class javaClass) {
        super(name, namespace);
        addAttribute(JAVA_CLASS_ATTRIBUTE, javaClass.getName());
    }

    /**
     * Creates (deserializes) an object of a specified type by calling a
     * constructor of the class that accepts an XML element as single argument
     * and passing itself as parameter. Only classes that provide a suitable
     * constructor can be instantiated this way.
     *
     * @return the created object
     * @throws InstantiationException if instantiation failed
     * @throws SecurityException if access to the required reflection
     * information is denied
     * @throws ClassNotFoundException if the class specified in the
     * {@link #JAVA_CLASS_ATTRIBUTE} is not available on this system
     * @throws IllegalStateException if the java class attribute has been
     * removed (by calling {@link #unsetJavaClassAttrib()}
     */
    public final Object createObject() throws InstantiationException,
    SecurityException, ClassNotFoundException, IllegalStateException {
        return createObject(this, javaClass());
    }

    /**
     * Returns the name of the {@link Class} of the stored object.
     *
     * @return the class name of the stored object; or <code>null</code> if
     * {@link #unsetJavaClassAttrib()} has been called before
     */
    public String javaClassName() {
        return attributeValue(JAVA_CLASS_ATTRIBUTE);
    }

    /**
     * Returns the {@link Class} of the stored object.
     *
     * @return the class of the stored object
     * @throws ClassNotFoundException if the class cannot be located
     * @throws IllegalStateException if the java class attribute has been
     * removed (by calling {@link #unsetJavaClassAttrib()}
     */
    public Class javaClass()
    throws ClassNotFoundException, IllegalStateException {
        final String className = javaClassName();

        if (className != null) {
            return Class.forName(className);            
        } else {
            throw new IllegalStateException(
                    "Java class attribute has been unset");
        }
    }

    /**
     * Unsets the attribute representing the {@link Class} of the stored object.
     * @return <code>true</code> if the attribute was successfully removed
     * (first call to this method); <code>false</code> if there was no such
     * attribute (later calls to this method)
     */
    public boolean unsetJavaClassAttrib() {
        // remove the java attrib
        final Attribute javaAttrib = attribute(JAVA_CLASS_ATTRIBUTE);

        if (javaAttrib != null) {
            return remove(javaAttrib);
        } else {
            return false;
        }
    }

}
